/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.configuration;

import java.util.Collections;
import java.util.Dictionary;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import org.osgi.framework.BundleContext;
import org.osgi.framework.Constants;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.ServiceRegistration;
import org.osgi.service.cm.ConfigurationException;
import org.osgi.service.cm.ManagedService;
import org.osgi.util.tracker.ServiceTracker;
import org.osgi.util.tracker.ServiceTrackerCustomizer;

public class ConfigurationServiceImpl implements ManagedService, ConfigurationService {
    public static final String CONFIG_SERVICE_PID = "config-service";
    private static final Logger LOG = Logger.getLogger(ConfigurationServiceImpl.class.getName());

    boolean configServiceRegistered;
    Map<String, String> configuration;
    Set<ConfiguredService> services = new HashSet<ConfiguredService>();

    final BundleContext context;
    final ServiceTracker tracker;
    ServiceRegistration registration;
    ServiceRegistration registration2;
    private HttpConfiguration httpConfiguration; 
    
    public ConfigurationServiceImpl(BundleContext ctx) {
        context = ctx;
                
        tracker = new ServiceTracker(ctx, ConfiguredService.class.getName(), getCustomizer());
        tracker.open();
        
        httpConfiguration = new HttpConfigurationImpl(ctx);
        
    }

    ConfiguredServiceTrackerCustomizer getCustomizer() {
        return new ConfiguredServiceTrackerCustomizer();
    }
    
    public synchronized void start() {
        registration = context.registerService(ManagedService.class.getName(), this, getDefaults());
        LOG.info("Configuration Service started and registered");
    }
        
    public Map<String, String> getConfiguration() {
        return Collections.unmodifiableMap(configuration);
    }

    public HttpConfiguration getHttpConfiguration() {
        return httpConfiguration;
    }

    private Dictionary<String, String> getDefaults() {
        Dictionary<String, String> defaults = new Hashtable<String, String>();
        defaults.put(Constants.SERVICE_PID, CONFIG_SERVICE_PID);        
        return defaults;
    }
    
    @SuppressWarnings("unchecked")
    public synchronized void updated(Dictionary properties) throws ConfigurationException {        
        LOG.info("Configuration updated.");

        Map<String, String> c = new HashMap<String, String>();
        if (properties != null) {            
            for (Enumeration e = properties.keys(); e.hasMoreElements();) {
                String key = e.nextElement().toString();
                String value = properties.get(key).toString();
                c.put(key, value);
                LOG.finer("  " + key + ":" + value);
            }             
        } else {
            LOG.finer("Update revert to defaults");
            properties = getDefaults();
        }
        configuration = c;
        LOG.info("New configuration: " + c);
        
        if (c.size() > 0 && !configServiceRegistered) {
            configServiceRegistered = true;
            registration2 = context.registerService(ConfigurationService.class.getName(), this, 
                    new Hashtable<String, String>());
            LOG.info("ConfigurationService Registered.");
        }
        
        if (registration != null) {
            registration.setProperties(properties);
        }
        
        if (registration2 != null) {
            registration2.setProperties(properties);
        }
        
        Set<ConfiguredService> list = new HashSet<ConfiguredService>(services.size());
        synchronized (services) { 
            list.addAll(services);
        }        
        
        for (ConfiguredService cs : list) {
            cs.configurationUpdated(configuration);
            LOG.finer("Configuration update relayed to " + cs);
        }
    }

    public void shutdown() {
        registration.unregister();
        if (registration2 != null) {
            registration2.unregister();
        }
        tracker.close();
        LOG.info("Configuration Service shut down");
    }
    
    class ConfiguredServiceTrackerCustomizer implements ServiceTrackerCustomizer {

        public Object addingService(ServiceReference sr) {
            Object svc = context.getService(sr);
            LOG.fine("Adding Configured Service: " + sr + " - " + svc);
            
            synchronized (services) {
                if (svc instanceof ConfiguredService) {
                    services.add((ConfiguredService) svc);                    
                }                
            }
            return svc;
        }

        public void modifiedService(ServiceReference sr, Object svc) { }
        public void removedService(ServiceReference sr, Object svc) {
            LOG.fine("Removing Configured Service: " + sr + " - " + svc);
            
            synchronized (services) {
                services.remove(svc);
            }
        }        
    }

}

