/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.configuration;

import java.io.File;
import java.io.FileInputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;
import org.osgi.service.cm.Configuration;
import org.osgi.service.cm.ConfigurationAdmin;
import org.osgi.util.tracker.ServiceTracker;

public class FileConfigurationProvider {
    private static final Logger LOG = Logger.getLogger(FileConfigurationProvider.class.getName());
    
    ConfigurationAdmin configAdminService;
    final ServiceTracker tracker;
    
    public FileConfigurationProvider(BundleContext ctx) {
        tracker = new ServiceTracker(ctx, ConfigurationAdmin.class.getName(), null) {
            @Override
            public Object addingService(ServiceReference reference) {
                Object svc = super.addingService(reference);
                if (svc instanceof ConfigurationAdmin) {
                    configAdminService = (ConfigurationAdmin) svc;
                    initConfiguration();
                }
                return svc;
            }            
        };
        tracker.open();
    }

    private void initConfiguration() {
        LOG.info("Providing Configuration Admin Service with config data");
        File [] configFiles = new File(System.getProperty("user.dir")).listFiles(new FilenameFilter() {
            public boolean accept(File dir, String name) {
                return name.endsWith(".cfg");
            }
        });
        if (configFiles == null) {
            LOG.fine("No configuration files found in current directory");
            return;
        }
        
        for (File configFile : configFiles) {
            LOG.info("Handling configuration file: " + configFile.getAbsolutePath());
            String pid = configFile.getName();
            pid = pid.substring(0, pid.length() - 4); // eat off the .cfg bit

            Properties p = new Properties();
            FileInputStream fis = null;
            
            try {
                fis = new FileInputStream(configFile);
                p.load(fis);
                handleSubstitution(p);
                
                Configuration config = configAdminService.getConfiguration(pid);
                config.update(p);
                LOG.info("Updated configuration of " + pid + " with: " + p);
            } catch (IOException e) {
                LOG.log(Level.WARNING, "Problem handling configuration for pid:" + pid, e);
            } finally {
                if (fis != null) {
                    try {
                        fis.close();
                    } catch (IOException e) {
                        LOG.log(Level.WARNING, "Problem closing configuration file.", e);
                    }
                }
            }
        }
    }

    private void handleSubstitution(Properties p) {
        for (Enumeration<?> e = new Properties(p).propertyNames(); e.hasMoreElements();) {
            String propName = e.nextElement().toString();
            
            String value = p.getProperty(propName);
            value = handleSubstitution(p, value);
            p.setProperty(propName, value);
        }
    }

    private String handleSubstitution(Properties p, String value) {
        int idx = value.indexOf("${");
        if (idx < 0) {
            return value;
        }
        int idx2 = value.indexOf('}', idx);
        if (idx2 < 0) {
            return value;
        }

        String variable = value.substring(idx + 2, idx2);
        String toReplace = p.getProperty(variable);
        if (toReplace == null) {
            toReplace = System.getProperty(variable);
        }
        if (toReplace == null) {
            return value;
        }            
        toReplace = handleSubstitution(p, toReplace);
        
        StringBuilder sb = new StringBuilder(value);
        sb.replace(idx, idx2 + 1, toReplace);
        return sb.toString();
    }

    public void shutdown() {
        tracker.close();
    }    
}
