/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.configuration;

import java.io.IOException;
import java.net.Inet4Address;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.osgi.framework.BundleContext;


public class HttpConfigurationImpl implements HttpConfiguration {
    
    private static final Logger LOG = Logger.getLogger(HttpConfigurationImpl.class.getName());

    // TODO: PaxWeb default ports (8080 & 8443) are different from the OSGi specification (80 & 443).
    private static final int DEFAULT_HTTP_PORT = 8080;
    private static final int DEFAULT_SECURE_HTTP_PORT = 8443;
    private static final boolean DEFAULT_HTTP_ENABLED = true;
    private static final boolean DEFAULT_SECURE_HTTP_ENABLED = false;
    private static final String DEFAULT_HOSTNAME = "localhost";
    
    private int httpPort;
    private int secureHttpPort; 
    private boolean httpEnabled;
    private boolean secureHttpEnabled;
    private String hostname;

    public HttpConfigurationImpl(BundleContext ctx) {
        
        httpEnabled = str2Bool(ctx.getProperty("org.osgi.service.http.enabled"), 
                               DEFAULT_HTTP_ENABLED);
        secureHttpEnabled = str2Bool(ctx.getProperty("org.osgi.service.http.secure.enabled"), 
                                     DEFAULT_SECURE_HTTP_ENABLED);

        secureHttpPort = str2Int(ctx.getProperty("org.osgi.service.http.port.secure"), 
                                 secureHttpEnabled ? DEFAULT_SECURE_HTTP_PORT : 0);

        httpPort = str2Int(ctx.getProperty("org.osgi.service.http.port"), 
                           httpEnabled ? DEFAULT_HTTP_PORT : 0);

        hostname = getHostName(ctx.getProperty("depot.hostname"));
        
        

        LOG.info("hostname: " + hostname);
        if (httpEnabled) {
            LOG.info("HTTP enabled on port " + httpPort);
        } else {
            LOG.info("HTTP not enabled");
        }
        if (secureHttpEnabled) {
            LOG.info("Secure HTTP enabled on port " + secureHttpPort);
        } else {
            LOG.info("Secure HTTP not enabled");
        }
    }



    public int getHttpPort() {
        if (!isHttpEnabled()) {
            return 0;
        }
        
        return httpPort;
    }
    
    public int getSecureHttpPort() {
        if (!isSecureHttpEnabled()) {
            return 0;
        }
        
        return secureHttpPort;        
    }
    
    public boolean isHttpEnabled() {
        return httpEnabled;
    }
    
    public boolean isSecureHttpEnabled() {
        return secureHttpEnabled;
    }
    
    public String getHostName() {
        return hostname;
    }
    
    private boolean str2Bool(String value, boolean defValue) {
        if (value == null) {
            return defValue;
        }
        return Boolean.parseBoolean(value);
    }
    
    private int str2Int(String value, int defValue) {
        if (value == null) {
            return defValue;
        }
        return Integer.parseInt(value);
    }

    private String getHostName(String value) {
        if (value != null) {
            return value;
        }
        
        try {
            return Inet4Address.getLocalHost().getCanonicalHostName();
        } catch (IOException e) {
            LOG.log(Level.WARNING, "exception getting host name" , e);
            return DEFAULT_HOSTNAME;
        }
    }

}
