/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.configgen.util;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.List;
import java.util.Properties;
import java.util.logging.Logger;

import com.iona.soa.repository.ModelRegistry;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.configuration.HttpConfiguration;
import com.iona.soa.repository.jdbc.EMFSQLUtil;
import com.iona.soa.repository.jdbc.JDBCConnectionManager;
import com.iona.soa.repository.jdbc.JDBCResourceRepository;
import com.iona.soa.repository.jdbc.PooledConnection;
import com.iona.soa.repository.util.SQLHelper;
import com.iona.soa.repository.util.VersionType;

import org.eclipse.emf.ecore.EObject;

/**
 * @author ohurley
 * 
 */
public final class DBCreate {

    private static final Logger LOG = Logger
            .getLogger(DBCreate.class.getName());

    private static final String JDBC_PROPERTIES_FILE = "jdbcProps";

    private static final String DB_URL_PROP = "jdbcURL";
    private static final String DB_DRIVER_PROP = "jdbcDriver";
    private static final String USERNAME_PROP = "username";
    private static final String BASE_STORAGE_URL_PROP = "storageURL";
    private static final String PASS_PROP = "password";

    protected DBCreate() {
    }

    public static void createNetwork(VersionType schemaVersion,
            JDBCConnectionManager connMgr, ModelRegistry registry,
            HttpConfiguration httpConfig) throws Exception {
        SOARepository soa = null;

        try {
            soa = new JDBCResourceRepository(connMgr, registry);
            List<EObject> objects = GenerateModel.generate(schemaVersion,
                    httpConfig);

            for (EObject obj : objects) {
                if (obj.eContainer() == null) {
                    soa.persist(obj);
                }
            }
        } finally {
            if (soa != null) {
                soa.close();
            }
        }
    }

    public static void createTables(JDBCConnectionManager connMgr,
            ModelRegistry registry) throws SQLException {
        PooledConnection pc = connMgr.getConnection();
        EMFSQLUtil.createAllModelTables(registry, pc.getConnection());
        connMgr.returnConnection(pc);
    }

    public static Connection connect(Properties props) throws SQLException {

        try {
            Class.forName(props.getProperty(DB_DRIVER_PROP)).newInstance();
        } catch (Exception e) {
            throw new SQLException(
                    "Error trying to load the JDBC driver class; details: "
                            + e.getMessage());
        }

        Properties connectionProperties = new Properties();
        connectionProperties.put("user", props.getProperty(USERNAME_PROP, ""));
        connectionProperties.put("password", props.getProperty(PASS_PROP, ""));

        String url = props.getProperty(DB_URL_PROP);
        Connection connection = DriverManager.getConnection(url,
                connectionProperties);

        LOG.fine("Successfully initialised connection to JDBC at url '" + url
                + "'");
        return connection;
    }

    private static Properties getProperties(String[] args) {
        Properties props = new Properties();
        for (int i = 0; i < args.length; i++) {
            String arg = args[i];
            if (arg.startsWith("-")) {
                String key = arg.substring(1);
                if (i + 1 < args.length && !args[i + 1].startsWith("-")) {
                    props.put(key, args[i + 1]);
                } else {
                    props.put(key, "true");
                }
            }
        }

        String propsFile = props.getProperty(JDBC_PROPERTIES_FILE);

        // this takes precedence
        if (propsFile != null) {
            InputStream is = null;
            try {
                is = new FileInputStream(propsFile);
                props.load(is);
            } catch (Exception ex) {
                LOG.warning("Error loading properties");
            }
            if (is != null) {
                try {
                    is.close();
                } catch (Exception ex) {
                    // ignore
                }
            }
        }

        return props;
    }

    private static void validateProperties(Properties props) {
        String problem = null;

        String dbUrl = props.getProperty(DB_URL_PROP);
        if (dbUrl == null) {
            problem = "Missing database storage URL";
        }

        String dbDriver = props.getProperty(DB_DRIVER_PROP);
        if (dbDriver == null) {
            problem = "Missing database driver name";
        }

        String storageURL = props.getProperty(BASE_STORAGE_URL_PROP);
        if (storageURL == null) {
            problem = "Missing base storage URL";
        }

        if (problem != null) {
            LOG.warning(problem);
            System.exit(1);
        }
    }

    public static boolean tablesAlreadyCreated(JDBCConnectionManager connMgr) {
        PooledConnection pc = connMgr.getConnection();
        try {
            return tablesAlreadyCreated(pc.getConnection());
        } finally {
            connMgr.returnConnection(pc);
        }
    }

    public static boolean tablesAlreadyCreated(Connection conn) {
        boolean alreadyCreated = false;

        Statement stmt = null;
        try {
            // let's just hand code the select for now...
            stmt = conn.createStatement();
            stmt.execute("select * from "
                    + SQLHelper.getNonSQLKeyword("MODELMETADATA"));
            alreadyCreated = true;
        } catch (Exception ex) {
            // assume tables are already created
        } finally {
            SQLHelper.close(stmt);
        }
        return alreadyCreated;
    }

    protected static String readTextFile(Reader r) throws IOException {
        BufferedReader br = new BufferedReader(r);
        try {
            StringBuilder sb = new StringBuilder();
            String line = null;
            line = br.readLine();
            while (line != null) {
                sb.append(line);
                line = br.readLine();
            }
            return sb.toString();
        } catch (IOException e) {
            e.printStackTrace();
            return null;
        } finally {
            br.close();
            r.close();
        }
    }
}
