/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.pac;

import java.text.MessageFormat;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.repository.event.DataChangeAdapter;
import com.iona.repository.event.DataChangeEventObserver;
import com.iona.soa.model.Container;
import com.iona.soa.model.InstalledApplication;
import com.iona.soa.model.SOA_NetworkPackage;
import com.iona.soa.model.events.DataChangeEvent;
import com.iona.soa.model.events.EventsPackage;
import com.iona.soa.model.repository.AuditEvent;
import com.iona.soa.model.repository.AuditLevel;
import com.iona.soa.model.repository.Property;
import com.iona.soa.model.repository.RepositoryFactory;
import com.iona.soa.model.repository.RepositoryPackage;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.SOARepositoryService;
import com.iona.soa.repository.TransactionData;
import com.iona.soa.resourceset.ResourceSetQueryHandler;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EEnumLiteral;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EOperation;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EParameter;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.ocl.OCL;
import org.eclipse.ocl.ParserException;
import org.eclipse.ocl.ecore.CallOperationAction;
import org.eclipse.ocl.ecore.Constraint;
import org.eclipse.ocl.ecore.EcoreEnvironmentFactory;
import org.eclipse.ocl.ecore.SendSignalAction;
import org.eclipse.ocl.helper.OCLHelper;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.InitializingBean;

public class AuditLogPublisher implements InitializingBean, DisposableBean, DataChangeEventObserver {

    private static final Logger LOG = Logger.getLogger(AuditLogPublisher.class.getName());

    private static final int MAX_MSG_LEN = 255;
    
    final ExecutorService executor = Executors.newSingleThreadExecutor();
    
    // OCL 1.1.0
    private final OCL<EPackage, EClassifier, EOperation, EStructuralFeature, EEnumLiteral,
        EParameter, EObject, CallOperationAction, SendSignalAction, Constraint, EClass, EObject> ocl;
   
    private final OCLHelper<EClassifier, ?, ?, Constraint> oclHelper;

    private SOARepository repository; // Spring injected
    
    
    AuditLogPublisher() {
        ocl = OCL.newInstance(EcoreEnvironmentFactory.INSTANCE);
        
        oclHelper = ocl.createOCLHelper();
        oclHelper.setContext(EventsPackage.Literals.DATA_CHANGE_EVENT);
    }
    
    /**
     * Spring Injected
     */
    public final void setRepositoryService(SOARepositoryService repoService) {
        repository = repoService.getRepository();
    }
    
    /**
     * InitializingBean implementation
     */
    public void afterPropertiesSet() throws Exception {
        LOG.fine("Registering self as DataChangeEventObserver");
        DataChangeAdapter.getInstance().registerObserver(this);
    }

    /**
     * DisposableBean implementation
     */
    public void destroy() throws Exception {
        LOG.fine("Removing self as DataChangeEventObserver");
        DataChangeAdapter.getInstance().unregisterObserver(this);
    }


    public synchronized void dataChanged(DataChangeEvent event) {
        // Currently assuming that an event cannot contain more
        // than one change we are interested in.  This may change.
        @SuppressWarnings("unused")
        boolean eventCreated = 
            handleAgentStateEvent(event) 
            || handleInstalledApplicationsChangeEvent(event);
    }    
    
    private boolean handleAgentStateEvent(DataChangeEvent event) {
        String containerGuid = null;
        String msg = null;
        
        // This is for a Container with no properties that just has had the
        // liveness property added
        String expr = "changeKind = DataChangeKind::MODIFY and targetType.name = 'Container' and " 
            + "updatedFields->one(d: DataChangeFieldDelta | d.fieldDesc.name = 'status' and " 
            + "d.newValues->includes('liveness'))";
        boolean matchingEvent = evalOclConstraint(expr, event);
        if (matchingEvent) {
            containerGuid = event.getTargetGuid();
            try {
                Container container = getObjectWithGuid(SOA_NetworkPackage.eINSTANCE.getContainer(), 
                                               event.getTargetGuid());
                Property p = ProvisioningDataProvider.findProperty(container.getStatus(), "liveness");
                msg = MessageFormat.format("Run status set to \"{0}\"", p.getValue());
            } catch (Exception ex) {
                LOG.log(Level.SEVERE, "Failed to read changed container", ex);
                msg = ex.getMessage();
            }
        } else {
            // Existing property set
            expr = "changeKind = DataChangeKind::MODIFY and targetType.name = 'Property' and " 
                + "updatedFields->one(d: DataChangeFieldDelta | d.newValue = 'alive' " 
                + "or d.newValue = 'dead')";
            matchingEvent = evalOclConstraint(expr, event);
            if (matchingEvent) {
                try {
                    Property p = getObjectWithGuid(RepositoryPackage.eINSTANCE.getProperty(), 
                                              event.getTargetGuid());
                    containerGuid = EcoreUtil.getID(p.eContainer());
                    msg = MessageFormat.format("Run status changed to \"{0}\"", 
                                               new Object[] {p.getValue()});
                } catch (Exception ex) {
                    LOG.log(Level.SEVERE, "Failed to read changed liveness property", ex);
                    msg = ex.getMessage();
                }
            }
        }
        
        if (matchingEvent) {
            AuditEvent auditEvent = createAuditEvent(event);
            
            auditEvent.setTargetGuid(containerGuid);
            auditEvent.setMessage(msg);
            persist(auditEvent, event.getUserId(), event.getUserToken());
        }        

        return matchingEvent;
    }

    private boolean handleInstalledApplicationsChangeEvent(DataChangeEvent event) {
        String containerGuid = null;
        String msg = null;
        
        // This is for a Container with no properties that just has had the
        // liveness property added
        String expr = "changeKind = DataChangeKind::MODIFY and targetType.name = 'Container' and " 
            + "updatedFields->one(d: DataChangeFieldDelta | d.fieldDesc.name = 'applications')"; 
        boolean matchingEvent = evalOclConstraint(expr, event);
        if (matchingEvent) {
            containerGuid = event.getTargetGuid();
            try {
                Container container = getObjectWithGuid(SOA_NetworkPackage.eINSTANCE.getContainer(), 
                                               event.getTargetGuid());
                StringBuffer appDetails = new StringBuffer();
                if (container.getApplications().isEmpty()) {
                    appDetails.append("<none>");
                } else {
                    int i = 0; 
                    for (InstalledApplication app : container.getApplications()) {
                        appDetails.append(app.getName());
                        if (++i < container.getApplications().size()) {
                            appDetails.append(", ");
                        }
                    }
                }
                msg = MessageFormat.format("Installed applications changed to \"{0}\"", 
                                           appDetails);
                
            } catch (Exception ex) {
                LOG.log(Level.SEVERE, "Failed to read changed container", ex);
                msg = ex.getMessage();
            }
        }       
        if (matchingEvent) {
            AuditEvent auditEvent = createAuditEvent(event);
            
            auditEvent.setTargetGuid(containerGuid);
            auditEvent.setMessage(msg);
            persist(auditEvent, event.getUserId(), event.getUserToken());
        }        

        return matchingEvent;
    }
    
    private AuditEvent createAuditEvent(DataChangeEvent event) {
        AuditEvent auditEvent = RepositoryFactory.eINSTANCE.createAuditEvent();
        auditEvent.setLevel(AuditLevel.INFO);
        auditEvent.setTimestamp(event.getTimestamp());
        return auditEvent;
    }
    
    void persist(final AuditEvent auditEvent, final String userId, final String ssoToken) {
        executor.execute(new Runnable() {
            public void run() {
                boolean success = false;
                repository.startWrite(new TransactionData(userId, 
                                                          ssoToken));
                try {
                    if (auditEvent.getMessage() != null 
                            && auditEvent.getMessage().length() > MAX_MSG_LEN) {
                        String trimmedMsg = auditEvent.getMessage().substring(0, 
                                                                              MAX_MSG_LEN - 3).concat("...");
                        auditEvent.setMessage(trimmedMsg);
                    }
                    repository.persist(auditEvent);
                    success = true;
                } catch (Exception ex) {
                    LOG.log(Level.SEVERE, "Failed to log AuditEvent for changed Agent", ex);
                } finally {
                    repository.endWrite(success);
                }
            }
        });
    }
    
    private boolean evalOclConstraint(String expr, DataChangeEvent event) {
        try {
            Constraint constraint = oclHelper.createInvariant(expr);
            org.eclipse.ocl.Query<EClassifier, EClass, EObject> constraintEval 
                = ocl.createQuery(constraint);
    
            return constraintEval.check(event);
        } catch (ParserException ex) {
            throw new RuntimeException(ex);
        }
    }
    
    /*
    private DataChangeFieldDelta runOclQuery(String expr, DataChangeEvent event) {
        try {
            OCLExpression<EClassifier> query = oclHelper.createQuery(expr);
            Query<EClassifier, EClass, EObject> queryEval = ocl.createQuery(query);
            return (DataChangeFieldDelta)queryEval.evaluate(event);
        } catch (ParserException ex) {
            throw new RuntimeException(ex);
        }
    }
    */
    
    @SuppressWarnings("unchecked")
    <T extends EObject> T getObjectWithGuid(EClass type, String guid) throws Exception {
        return (T)ResourceSetQueryHandler.executeQuery(repository.getResourceSet(), 
                                                       type, 
                                                       guid);
    }
}
