/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.pac;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.soa.expression.Equals;
import com.iona.soa.expression.Expression;
import com.iona.soa.model.Application;
import com.iona.soa.model.Artifact;
import com.iona.soa.model.ArtifactType;
import com.iona.soa.model.ConfigurationOverride;
import com.iona.soa.model.Container;
import com.iona.soa.model.DeploymentProfile;
import com.iona.soa.model.InstalledApplication;
import com.iona.soa.model.SOA_NetworkFactory;
import com.iona.soa.model.SOA_NetworkPackage;
import com.iona.soa.model.VersionedProfile;
import com.iona.soa.model.repository.Property;
import com.iona.soa.model.repository.RepositoryFactory;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.SOARepositoryService;
import com.iona.soa.repository.TransactionData;
import com.iona.soa.repository.configuration.ConfigurationHelper;
import com.iona.soa.repository.configuration.ConfigurationKeys;
import com.iona.soa.repository.configuration.ConfigurationService;
import com.iona.soa.repository.configuration.HttpConfiguration;
import com.iona.soa.repository.model.util.VersionHelper;
import com.iona.soa.repository.util.InputStreamHelper;
import com.iona.soa.resourceset.ResourceSetQueryHandler;
import com.iona.soa.security.SecurityContext;
import com.iona.soa.security.SecurityContextImpl;
import com.iona.soa.security.SecurityService;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.fusesource.cloudmix.common.ControllerDataProvider;
import org.fusesource.cloudmix.common.GridController;
import org.fusesource.cloudmix.common.controller.AgentController;
import org.fusesource.cloudmix.common.controller.FeatureController;
import org.fusesource.cloudmix.common.controller.ProfileController;
import org.fusesource.cloudmix.common.dto.AgentCfgUpdate;
import org.fusesource.cloudmix.common.dto.AgentDetails;
import org.fusesource.cloudmix.common.dto.Dependency;
import org.fusesource.cloudmix.common.dto.FeatureDetails;
import org.fusesource.cloudmix.common.dto.ProfileDetails;
import org.fusesource.cloudmix.common.dto.ProvisioningHistory;
import org.osgi.service.http.HttpService;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.InitializingBean;

public class ProvisioningDataProvider implements ControllerDataProvider, InitializingBean, DisposableBean {
    public static final String LOCALHOST_MARKER = "@{LOCALHOST}";

    static boolean asyncMode = true;
    
    private static final String CONTAINER_GUID_PREFIX = "Container-";
    
    private static final Logger LOG = Logger.getLogger(ProvisioningDataProvider.class.getName());
    private static final String FEATURE_RESOURCES_CONTEXT = "/controller_features";
    private static final char FEATURE_ID_SEPARATOR = '_';
    private static final String DEFAULT_PACKAGE_TYPE = "osgi";

    private static final SecurityContext PROVISIONER_SECURITY_CONTEXT
        = new SecurityContextImpl(SecurityService.INTERNAL_USER, SecurityService.INTERNAL_TOKEN);

    RepositoryChangeListener listener;
    File storageDir;
    long nextAgentUpdate;
    int repoWriteInterval = 10000; // ms - TODO make configurable

    private ConcurrentHashMap<String, AgentController> agents = 
        new ConcurrentHashMap<String, AgentController>();    
    private Map<String, ProfileController> profiles = new ConcurrentHashMap<String, ProfileController>();    
    private Map<String, FeatureController> features = new ConcurrentHashMap<String, FeatureController>();
    private Map<String, ArtifactController> artifacts = new ConcurrentHashMap<String, ArtifactController>();
    
    
    private final ExecutorService executor = Executors.newSingleThreadExecutor();
    private HttpService httpService;
    private SOARepositoryService repositoryService;
    private GridController grid;
    private SOARepository repository;
    private boolean injectionComplete;
    private Object mutex = new Object();
    private boolean updateInProgress;
    private String protocol;
    private String hostname;
    private int port;
    
    ProvisioningDataProvider() {
        // Complete.
    }
    
    /**
     * Spring Injected
     */
    public final void setConfiguration(ConfigurationService configService) {
        LOG.log(Level.FINE, "Setting configuration: {0}", configService.getConfiguration());
        
        storageDir = new File(ConfigurationHelper.getValue(
                configService.getConfiguration(), ConfigurationKeys.PROVISIONING_DATA_DIR,   
            System.getProperty("java.io.tmpdir") + "/provisioning"));
        LOG.log(Level.FINE, "Set 'storageDir' to: {0}", storageDir);
        
        HttpConfiguration httpConf = configService.getHttpConfiguration();
        if (httpConf.isHttpEnabled()) {
            protocol = "http";
            port = httpConf.getHttpPort();
        } else {
            protocol = "https";
            port = httpConf.getSecureHttpPort();
        }
        hostname = httpConf.getHostName();
    }    
    
    /**
     * Spring Injected
     */
    public final void setHttpService(HttpService hs) {
        httpService = hs;
    }
    
    /**
     * Spring Injected
     */
    public final void setRepositoryService(SOARepositoryService repoService) {
        repositoryService = repoService;
        repository = repoService.getRepository();
        LOG.log(Level.FINE, "Injected SOARepository: {0}", repository == null ? "null" : "ok");
    }    

    public void afterPropertiesSet() throws Exception {
        injectionComplete = true;
        init();
    }

    public void destroy() throws Exception {
        if (listener != null) {
            listener.destroy();
            listener = null;
        }
    }

    //-------------------------------------------------------------------------
    // ControllerDataProvider implementation
    //-------------------------------------------------------------------------
    
    /**
     * used to register agents (new or otherwise) when they become available.
     * update the depot with the complete agent configuration so that:
     *  - runtime data is accurate
     *  - changes to the agent config file(s) are picked up
     *  
     * @see org.fusesource.cloudmix.common.ControllerDataProvider
     */
    public AgentController addAgent(String id, AgentController agent) {
        LOG.log(Level.FINE, "Registering agent: {0}", id);
        storeAgentDetailsAsync(id, agent.getDetails());
        return agents.putIfAbsent(id, agent);
    }

    /**
     * used by the regular polls of the agent, therefore only update the Depot's data with the current status
     * of the feature installations in that agent.
     * general agent configuration only gets pushed to the Depot when the agent starts up as those are 
     * consistent during the life time of the current agent's process
     *  
     * @see org.fusesource.cloudmix.common.ControllerDataProvider
     */
    public AgentController updateAgent(String id, AgentController agent) {
        LOG.log(Level.FINE, "updating agent data following polling: {0}", id);
        AgentDetails details = agent.getDetails();
        AgentController controller = agents.get(id);
        if (controller == null) {
            LOG.warning("Could not find agent with id " + id);
            return null;
        } 
        controller.setDetails(details);
        
        storeAgentFeaturesStatusAsync(id, details);
        return controller;
    }    
    
    public AgentController getAgent(String agentId) {
//        LOG.info("Retrieving agent: " + agentId);
        return agents.get(agentId);
    }

    public AgentController removeAgent(String agentId) {
        // TODO - should associated container get removed?
        LOG.log(Level.FINE, "Removing agent \"{0}\"", agentId);
        return agents.remove(agentId);
    }
    
    public Collection<AgentController> getAgents() {
        LOG.log(Level.FINE,
                "Retrieving all agents ({0})",
                agents.isEmpty() ? "none" : agents.size());
        return agents.values();
    }

    public FeatureController addFeature(String id, FeatureController fc) {
        throw new UnsupportedOperationException();
    }

    public FeatureController getFeature(String id) {
        return features.get(id);
    }
    
    public Collection<FeatureController> getFeatures() {
        LOG.log(Level.FINE,
                "Retrieving all features ({0})", 
                features.isEmpty() ? "none" : features.size());
        return features.values();
    }

    public ProfileController addProfile(String arg0, ProfileController arg1) {
        throw new UnsupportedOperationException();
    }

    public ProfileController getProfile(String id) {
        LOG.log(Level.FINE,
                "Retrieving profile \"{0}\"",
                id);
        return profiles.get(id);
    }

    public Collection<ProfileController> getProfiles() {
        LOG.log(Level.FINE,
                "Retrieving all profiles ({0})", 
                profiles.isEmpty() ? "none" : profiles.size());
        updateAgentLiveness();
        return profiles.values();
    }

    

    public GridController getGrid() {
        return grid;
    }

    public void setGrid(GridController g) {
        grid = g;
        init();
    }

    // ------------------------------------------------------------------------
    // Init methods - will be called as soon as a GridController is registered
    // ------------------------------------------------------------------------
    void init() {
        if (!injectionComplete || grid == null) {
            return; // Wait for the other
        }
        
        LOG.fine("Provisioning Data Provider initializing.");
        
        listener = new RepositoryChangeListener(this);
        
        try {
            if (!storageDir.exists()) {
                LOG.log(Level.FINE, "Creating cache directory {0}", storageDir.getAbsolutePath());
                if (!storageDir.mkdirs()) {
                    throw new IOException("Could not create cache directory " 
                                          + storageDir.getAbsolutePath());
                }
            } else {
                LOG.log(Level.FINE, "Using cache directory {0}", storageDir.getAbsolutePath());
            }
            RepositoryHttpContext context =
                new RepositoryHttpContext(repositoryService.getSecurityService(), storageDir);
            httpService.registerResources(FEATURE_RESOURCES_CONTEXT, "", context);

            initFeatures();
            initProfiles();
        } catch (Throwable th) {
            LOG.log(Level.SEVERE, "Problem during initialization", th);
        }         
    }
        
    // Called from RepositoryChangeListener
    void refreshProfile(String guid, boolean isDeleted) throws Exception {
        
        if (isDeleted) {
            removeProfile(guid);
            return;
        }
        
        DeploymentProfile dp = getObjectWithGuid(SOA_NetworkPackage.eINSTANCE.getDeploymentProfile(), guid);
        if (dp != null) {
            refreshProfile(dp);
        } else {
            LOG.warning("Could not refresh deployment profile " + guid);
        }
    }

    // Called from RepositoryChangeListener
    void refreshFeature(String guid, boolean isDeleted) throws Exception {
        
        if (isDeleted) {
            removeFeature(guid);
            return;
        }
        
        Application app = getObjectWithGuid(SOA_NetworkPackage.eINSTANCE.getApplication(), guid);
        if (app != null) {
            refreshFeature(app);
        } else {
            LOG.warning("Could not refresh feature " + guid);
        }
    }
    
    // Called from RepositoryChangeListener
    void refreshArtifact(String guid, boolean isDeleted) throws Exception {
        
        if (isDeleted) {
            removeArtifact(guid);
            return;
        }
        
        Artifact artifact = getObjectWithGuid(SOA_NetworkPackage.eINSTANCE.getArtifact(), guid);
        if (artifact != null) {
            refreshArtifact(artifact);
        } else {
            LOG.warning("Could not refresh artifact " + guid);
        }

    }

    // Called from RepositoryChangeListener
    void refreshAgent(String guid, boolean isDeleted) throws Exception {
        
        if (isDeleted) {
            AgentController ac = agents.get(getAgentIdFromGuid(guid));
            if (ac != null) {
                ac.deActivate();
            }
            return;
        }
        
        Container container = getObjectWithGuid(SOA_NetworkPackage.eINSTANCE.getContainer(), guid);
        if (container != null) {
            refreshAgent(container);
        } else {
            LOG.warning("Could not refresh Agent " + guid);
        }
    }

    
    String getFeatureIdByBareName(String name) {
        for (String id : features.keySet()) {
            // Need to search backwards in case the separator is used 
            // as part of the applications's name.
            int idx = id.lastIndexOf(FEATURE_ID_SEPARATOR);
            if (idx > 0) {
                idx = id.lastIndexOf(FEATURE_ID_SEPARATOR, idx - 1);
                if (idx > 0) { 
                    String n = id.substring(0, idx);
                    if (name.equals(n)) {
                        return id;
                    }
                }
            }
        }
        return null;
    }
    
    protected Container findContainer(String id) throws Exception {
        return getObjectWithGuid(SOA_NetworkPackage.eINSTANCE.getContainer(),
                            getContainerId(id));
    }
    
    protected Container createContainer(String id) {
        Container c = SOA_NetworkFactory.eINSTANCE.createContainer();
        c.setName(id);
        c.setGuid(getContainerId(id));
        return c;
    }
    
    void refreshAgent(Container aContainer) {
        
        AgentController ac = agents.get(getAgentId(aContainer));
        LOG.log(Level.FINE, "Provisioning Data Provider: refreshing agent: {0}", ac);
        
        if (ac == null || ac.getDetails() == null) {
            LOG.warning("Provisioning Data Provider: no controller found for container: "
                        + aContainer.getName());
            return;
        }
        
        String oldName = ac.getDetails().getName();
        String oldProfileID = ac.getDetails().getProfile();
        
        String newAgentName = aContainer.getName();
        if (changed(oldName, newAgentName)) {
            ac.getDetails().setName(newAgentName);
            ProvisioningHistory history = ac.getHistory();
            LOG.log(Level.FINE, "Provisioning Data Provider: refreshing agent name to: {0}", newAgentName);
            if (history != null) {
                history.addCfgUpdate(new AgentCfgUpdate(AgentCfgUpdate.PROPERTY_AGENT_NAME,
                                                        newAgentName));
            }            
        }
        
        String newProfileName = (aContainer.getProfile() == null) ? null : aContainer.getProfile().getName();
        if (changed(oldProfileID, newProfileName)) {
            LOG.log(Level.FINE, "Provisioning Data Provider: refreshing agent profile to: {0}", 
                                newProfileName);
            ac.getDetails().setProfile(newProfileName);
            ac.getHistory().addCfgUpdate(new AgentCfgUpdate(AgentCfgUpdate.PROPERTY_PROFILE_ID,
                                                            newProfileName));
        }
        
        if (ac.getHistory() != null) {
            ac.getHistory().removeCfgUpdate(AgentCfgUpdate.PROPERTY_AGENT_FORCE_REGISTER);
        }
        
    }

    void storeAgentLiveness() throws Exception {
        List<Container> containers = getObjectsOfType(SOA_NetworkPackage.eINSTANCE.getContainer());
        
        for (Container container : containers) {
            String agentId = getAgentId(container);
            AgentController agent = agents.get(agentId);
            String agentState;
            long now = System.currentTimeMillis();
            if (agent != null) {
                agentState = agent.isActive(now) ? "alive" : "dead";
            } else {
                agentState = "dead";                    
            }
            String oldAgentState = getProperty(container.getStatus(), "liveness", "unknown");
            if (!oldAgentState.equals(agentState)) {
                LOG.log(Level.FINE,
                        "Agent \"{0}\", running state changed to \"{1}\"",
                        new Object[] {agentId, agentState});
                
                boolean passed = false;
                repository.startWrite(createTransactionData(null));
                try {
                    setProperty(container.getStatus(), "liveness", agentState);
                    setProperty(container.getStatus(), "livenessUpdated", Long.toString(now));
                    passed = true;
                } finally {
                    repository.endWrite(passed);
                }
            }
        }        
    }

    private FeatureController refreshFeature(Application app) throws IOException {
        
        FeatureController fc = impl2FeatureController(app);

        String oldId = getFeatureIdByBareName(encodeURL(app.getName()));
        if (oldId != null && !oldId.equals(getFeatureId(app))) {
            removeFeatureFromCache(oldId);
            refreshProfilesWithFeature(oldId);
        }

        features.put(fc.getId(), fc);

        // Deal with dependencies after updating features map so cycles are handled correctly.
        for (Application depApp : app.getDepends()) {
            String id = getFeatureIdByBareName(depApp.getName());

            FeatureController depFc = null;
            if (id == null) {
                depFc = refreshFeature(depApp);
            } else {
                depFc = getFeature(id);
                if (depFc == null) {
                    depFc = refreshFeature(depApp);
                }
            }
            id = depFc.getId();
            if (id != null) {
                LOG.log(Level.FINE,
                        "adding dependency {0} to application {1}", 
                        new Object[] {id, app.getName()});
                fc.getDetails().depends(id);
            }            
        }
        
        LOG.log(Level.FINEST, "Updated feature {0}", app);
        return fc;
    }
    
    public FeatureController removeFeature(String guid) {
        
        removeFeatureFromCache(guid);
        refreshProfilesWithFeature(guid);
        FeatureController fc = features.remove(guid);

        LOG.log(Level.FINEST,
                "disregarded feature {0}",
                ((fc == null) || (fc.getDetails() == null) || (fc.getDetails().getId() == null))
                    ? "<null>"
                    : fc.getDetails().getId());
        return fc;
    }
    
    private void refreshArtifact(Artifact artifact) throws IOException {
        List<String> appGuids = new ArrayList<String>();
        EList<Application> apps = artifact.getApplications();
        if (apps != null) {
            for (Application app : apps) {
                refreshFeature(app);
                appGuids.add(app.getGuid());
            }
        }
        artifacts.put(artifact.getGuid(),
                      new ArtifactController(artifact.getGuid(), artifact.getName(), appGuids));
    }

    private void removeArtifact(String guid) throws Exception {
        ArtifactController ac = artifacts.remove(guid);
        
        Collection<String> appGuids = ac != null ? ac.getApplicationGuids() : null;
        if (appGuids != null) {
            for (String appGuid : appGuids) {
                refreshFeature(appGuid, false);
            }
        }
    }

    
    private void refreshProfilesWithFeature(String featureId) {
        for (ProfileController pc : findProfilesWithFeature(featureId)) {
            try {
                Expression expr = new Equals("name", decodeURL(pc.getDetails().getId()));
                EClass eCls = SOA_NetworkPackage.eINSTANCE.getDeploymentProfile();
                List<DeploymentProfile> dps = getObjectsOfType(eCls, expr);
                for (DeploymentProfile dp : dps) {
                    refreshProfile(dp);
                }
            } catch (Exception e) {
                LOG.log(Level.WARNING, "Could not refresh Deployment Profile " + pc.getDetails().getId(), e);
            }
        }
    }   

    private DeploymentProfile findDeploymentProfile(String profileName) {
        try {
            EClass eCls = SOA_NetworkPackage.eINSTANCE.getDeploymentProfile();
            List<DeploymentProfile> dps = getObjectsOfType(eCls, new Equals("name", profileName));
            for (DeploymentProfile dp : dps) {
                return dp;
            }
        } catch (Exception e) {
            LOG.log(Level.WARNING, "Could not find Deployment Profile " + profileName, e);
        }
        return null;
    }
    
    private Set<ProfileController> findProfilesWithFeature(String featureId) {
        Set<ProfileController> result = new HashSet<ProfileController>();
        
        for (ProfileController prof : getProfiles()) {
            for (Dependency details : prof.getDetails().getFeatures()) {
                if (details.getFeatureId().equals(featureId)) {
                    result.add(prof);
                }
            }
        }
        return result;
    }
    
    private void refreshProfile(DeploymentProfile dp) {
        VersionedProfile activeProfile = dp.getActive();
        if (activeProfile != null) {
            ProfileController pc = profile2ProfileController(dp, activeProfile);
            
            if (!pc.compare(profiles.get(pc.getDetails().getId()))) {
                profiles.put(pc.getDetails().getId(), pc);
                LOG.log(Level.FINE, "Updated profile {0}", dp.getName());
            }
            
        } else {
            LOG.log(Level.FINE, "Profile {0} has no active version", dp.getName());
        }
    }
    
    public ProfileController removeProfile(String guid) {
        return profiles.remove(guid); 
        
    }

    private void storeAgentDetailsAsync(final String id, final AgentDetails details) {
        
        final SecurityContext ctx = ProvisioningAuthenticator.getCurrentSecurityContext();
        if (ctx == null) {
            LOG.severe("cannot get security context for this request (storeAgentDetails)");            
            return;
        }
        Runnable r = new Runnable() {
            public void run() {
                try {
                    storeAgentDetails(ctx, id, details);
                } catch (Exception e) {
                    LOG.log(Level.SEVERE, "Problem communicating with the repository", e);
                }
            }            
        };
        submitRepoWrite(r);
    }

    private void storeAgentDetails(SecurityContext ctx, String id, AgentDetails details) 
        throws Exception {

        Container c = findContainer(id);
        boolean passed = false;
        repository.startWrite(createTransactionData(ctx));
        try {
            
            if (c == null) {
                LOG.log(Level.FINE, 
                        "Creating container for automatically registered agent \"{0}\"", 
                        id);
                c = createContainer(id);
                repository.persist(c);
            } else {
                LOG.log(Level.FINE, 
                        "Storing details for previously registered agent \"{0}\"", 
                        id);
            }
            
            List<Property> props = c.getProperties();
            setProperty(props, "hostname", details.getHostname());
            setProperty(props, "pid", "" + details.getPid());
            setProperty(props, "profile", details.getProfile());
            setProperty(props, "max features", "" + details.getMaximumFeatures());
            setProperty(props, "os", details.getOs());
            
            if (details.getAgentLink() != null) {
                setProperty(props, "agent link", details.getAgentLink());
            }
            
            c.setName(details.getName() != null ? details.getName() : id);
            
            c.setType(details.getContainerType());
            
            if (details.getSupportPackageTypes() != null) {
                c.getSupportedPackageTypes().addAll(Arrays.asList(details.getSupportPackageTypes()));
            }
            
            DeploymentProfile dp = findDeploymentProfile(details.getProfile());
            
            if (dp == null) {
                dp = SOA_NetworkFactory.eINSTANCE.createDeploymentProfile();
                dp.setName(details.getProfile());
                dp.setDescription("This profile was automatically created because container '" 
                        + id + "' is using it.");

                VersionedProfile vp = SOA_NetworkFactory.eINSTANCE.createVersionedProfile();
                vp.setName(dp.getName());
                dp.getVersions().add(vp);
                dp.setCurrent(vp);
                repository.persist(dp);
            }
            c.setProfile(dp);
            
            if (LOG.isLoggable(Level.FINE)) {
                LOG.fine("Storing Agent Details:");
                LOG.fine("id:               " + c.getName());
                LOG.fine("type:             " + c.getType());
                LOG.fine("profile:          " + (c.getProfile() != null ? c.getProfile().getName() : null));
            }
            storeAgentFeaturesStatus(details, c);
            
            passed = true;
        } finally {
            repository.endWrite(passed);
        }
    }

    
    private TransactionData createTransactionData(SecurityContext ctx) {
        // TODO: Add TransactionData ctor that takes a SecurityContext.
        if (ctx == null) {
            // Assume this is an internal update.
            ctx = PROVISIONER_SECURITY_CONTEXT;
        }
        return new TransactionData(ctx.getUser(), ctx.getSSOToken());
    }

    private boolean changed(String oldValue, String newValue) {
        return !((oldValue == null && newValue == null) || (oldValue != null && oldValue.equals(newValue)));
    }

    private void storeAgentFeaturesStatusAsync(final String id, final AgentDetails details) {
        Runnable r = new Runnable() {
            public void run() {
                try {
                    storeAgentFeaturesStatus(id, details);
                } catch (Exception e) {
                    LOG.log(Level.SEVERE, "Problem communicating with the repository", e);
                }
            }            
        };
        submitRepoWrite(r);
    }

    private void storeAgentFeaturesStatus(String id, AgentDetails details) throws Exception {

        Container c = findContainer(id);
        boolean passed = false;
        repository.startWrite(createTransactionData(null));
        try {
            if (c == null) {
                // TODO - should this ever happen?
                // If so, what about the properties and other details 
                // that are supplied via addAgent?
                LOG.log(Level.WARNING, "Creating container during storeAgentFeatures, id={0}", id);
                c = createContainer(id);
                repository.persist(c);
            }
            storeAgentFeaturesStatus(details, c);
            passed = true;
        } finally {
            repository.endWrite(passed);
        }
    }

    
    private void storeAgentFeaturesStatus(AgentDetails details, Container c) {

        List<String> fl = details.getCurrentFeatures() == null
                ? new ArrayList<String>() 
                : new ArrayList<String>(Arrays.asList(details.getCurrentFeatures()));
                
        for (Iterator<InstalledApplication> iia = c.getApplications().iterator(); iia.hasNext();) {
            InstalledApplication ia = iia.next();
            String n = ia.getName();
            if (fl.contains(n)) {
                fl.remove(n); // remove it from the features list
            } else {
                iia.remove(); // its not on the list so remove here
            }            
        }
        for (String fn : fl) {
            InstalledApplication ia = SOA_NetworkFactory.eINSTANCE.createInstalledApplication();            
            ia.setName(fn);
            c.getApplications().add(ia);
        }
    }
    
    private void updateAgentLiveness() {
        synchronized (mutex) {
            if (nextAgentUpdate > System.currentTimeMillis() || updateInProgress) {
                return;
            }
            updateInProgress = true;
        }
        
        
        Runnable r = new Runnable() {
            public void run() {
                try {                   
                    storeAgentLiveness();
                } catch (Exception e) {
                    LOG.log(Level.WARNING, 
                            "Could not store agent liveness information in the repository",
                            e);
                } finally {
                    synchronized (mutex) {
                        updateInProgress = false;
                        nextAgentUpdate = System.currentTimeMillis() + repoWriteInterval;
                    }
                }
            }
        };        
        submitRepoWrite(r);        
    }

    private void submitRepoWrite(Runnable runnable) {
        if (asyncMode) {
            executor.submit(runnable);
        } else {
            runnable.run();
        }                
    }

    private FeatureController impl2FeatureController(Application app) throws IOException {
        FeatureDetails fd = new FeatureDetails(getFeatureId(app));
        setPackageType(fd, app);
        

        
        String featureId = fd.getId();
        
        createFeaturesFilesIfNeeded(app, featureId);

        fd.setResource(protocol + "://" + hostname + ":" + port + FEATURE_RESOURCES_CONTEXT 
                       + "/" + featureId + ".xml");
        fd.setId(featureId);
        return new FeatureController(grid, fd);
    }

    private ProfileController profile2ProfileController(DeploymentProfile dp, 
                                                        VersionedProfile activeProfile) {
        ProfileDetails pd = new ProfileDetails(encodeURL(dp.getName()));
        for (Application app : activeProfile.getApplications()) {
            
            Map<String, String> cfgOverridesProps = new HashMap<String, String>();
            for (ConfigurationOverride co : activeProfile.getConfigurationOverride()) {
                if (app.getGuid().equals(co.getApplication().getGuid())) {
                    for (Property  cfgOverrideProp : co.getConfiguration()) {
                        cfgOverridesProps.put(cfgOverrideProp.getName(), cfgOverrideProp.getValue());
                    }
                }
            }
            
            pd.addFeature(getFeatureId(app), cfgOverridesProps);  
        }        
        
        return new ProfileController(grid, pd);
    }

    private void removeFeatureFromCache(String id) {
        features.remove(id);
        File file = FileRegistry.getFile(storageDir, id + ".xml");
        synchronized (file) {
            if (!file.delete()) {
                LOG.warning("Could not remove file from cache for feature " + id);
            }
        }
    }            

    private void createFeaturesFilesIfNeeded(Application app, String id) throws IOException {
        File file = FileRegistry.getFile(storageDir, id + ".xml");
        synchronized (file) {
            if (file.exists()) {
                file.delete();
            }
            
            File tmpFile = new File(storageDir, id + ".xml.tmp");
    
            String content = featureToXML(app, id);
            OutputStream os = new FileOutputStream(tmpFile);
            try {
                InputStreamHelper.drain(new ByteArrayInputStream(content.getBytes()), os);
            } finally {
                InputStreamHelper.close(os);
            }
            tmpFile.renameTo(file);
        }
    }

    private void initProfiles() throws Exception {
        LOG.fine("Initializing Profile Data");
        profiles.clear();
        
        List<DeploymentProfile> dps = 
            getObjectsOfType(SOA_NetworkPackage.eINSTANCE.getDeploymentProfile());
        
        for (DeploymentProfile dp : dps) {
            refreshProfile(dp);
        }
        LOG.log(Level.FINEST, "  initProfiles() returned {0}", profiles); 
    }
    
    private void initFeatures() throws Exception {
        LOG.fine("Initializing Feature Data");
        features.clear();

        List<Application> apps = getObjectsOfType(SOA_NetworkPackage.eINSTANCE.getApplication());
        
        for (Application app : apps) {
            refreshFeature(app);
        }
        LOG.log(Level.FINEST, "  initFeatures() obtained {0}", features); 
    }

    SOARepository getRepository() {
        return repository;
    }
    
    @SuppressWarnings("unchecked")
    <T extends EObject> T getObjectWithGuid(EClass type, String guid) throws Exception {
        repository.startRead(null);
        try {
            return (T)ResourceSetQueryHandler.executeQuery(repository.getResourceSet(), 
                                                           type, 
                                                           guid);
        } finally {
            repository.endRead(true);
        }
    }
    
    @SuppressWarnings("unchecked")
    <T extends EObject> List<T> getObjectsOfType(EClass type) throws Exception {
        boolean resolveProxies = true;
        repository.startRead(null);
        try {
            return (List<T>)ResourceSetQueryHandler.executeQuery(repository.getResourceSet(), 
                                                                 type, 
                                                                 resolveProxies);
        } finally {
            repository.endRead(true);
        }
    } 

    @SuppressWarnings("unchecked")
    <T extends EObject> List<T> getObjectsOfType(EClass type, Expression expr) throws Exception {
        boolean resolveProxies = true;
        repository.startRead(null);
        try {
            return (List<T>)ResourceSetQueryHandler.executeQuery(repository.getResourceSet(), 
                                                                 type,
                                                                 expr,
                                                                 resolveProxies);
        } finally {
            repository.endRead(true);
        }
    } 
    
    //-------------------------------------------------------------------------
    // Static helper methods
    //-------------------------------------------------------------------------
    static Property findProperty(List<Property> props, String key) {
        for (Property prop : props) {
            if (key.equals(prop.getName())) {
                return prop;
            }
        }
        return null;
    }

    private static String getProperty(List<Property> props, String key, String defaultValue) {
        Property prop = findProperty(props, key);
        if (prop != null && prop.getValue() != null) {
            return prop.getValue();
        }
        return defaultValue;
    }
    
    private static void addProperties(StringBuilder sb,
                                      String name,
                                      EList<Property> properties) {
        if (properties == null) {
            return;
        }
        // TODO: need to escape characters in name
        sb.append("<config name=\"").append(name).append("\">\n");
        for (Property p : properties) {
            sb.append(p.getName()).append("=").append(p.getValue()).append("\n");
        }
        sb.append("</config>");
        
    }

    static void setProperty(List<Property> props, String key, String value) {
        if (key == null) {
            return;
        }        
        
        Property prop = findProperty(props, key);
        if (value == null) {
            if (prop != null) {
                props.remove(prop);
            }
            return;            
        }
        
        if (prop == null) {
            prop = RepositoryFactory.eINSTANCE.createProperty();
            prop.setName(key);
            props.add(prop);
        }
        
        prop.setValue(value);
    }
    
    private static String getContainerId(String name) {
        // temporary workaround as the container ID currently needs to be prefixed with 'Container-'
        return CONTAINER_GUID_PREFIX + name;
    }

    private static String getAgentId(Container cont) {
        return getAgentIdFromGuid(cont.getGuid());
    }

    private static String getAgentIdFromGuid(String name) {
        if (name.startsWith(CONTAINER_GUID_PREFIX)) {
            return name.substring(CONTAINER_GUID_PREFIX.length());
        }
        return name;
    }


    private static void setPackageType(FeatureDetails fd, Application app) {

        Set<String> types = new HashSet<String>();
        EList<Artifact> artifacts = app.getArtifacts();
        if (artifacts != null) {
            for (Artifact artifact : app.getArtifacts()) {
                types.add(getPackageType(artifact));
            }
        }
        
        fd.setPackageTypes(types.toArray(new String[types.size()]));
    }
    
    private static String getPackageType(Artifact artifact) {       
        
        // TODO:
        // The packageType is taken from the should put the "packageType" as an attribute of ArtifactType.
        ArtifactType type = artifact.getType();
        if (type != null) {
            String id = type.getGuid();
            int i = id.indexOf("-");
            if (i != -1) {
                return id.substring(i + 1);
            }
        }
        
        // Have a guess.
        String filename = artifact.getFileName();
        if (filename == null || "".equals(filename)) {
            URI uri = artifact.getUri();
            if (uri != null) {
                filename = artifact.getUri().toString();    
            }
        }   
        // TODO: configurable mapping?
        if (filename != null && filename.endsWith(".war")) {
            return "war";
        }
        return DEFAULT_PACKAGE_TYPE;              
    }
    
    private static String getFeatureId(Application app) {
        String featureName = encodeURL(app.getName());
        String version = VersionHelper.versionToString(app.getVersion());
        if (version == null) {
            version = "0";
        }
        long digest = getImplDigest(app);
        if (digest < 0L) {
            digest = -digest;
        }
        String featureId = featureName + FEATURE_ID_SEPARATOR + version 
                           + FEATURE_ID_SEPARATOR + Long.toString(digest, 36);
        return featureId;
    }

    private static long getImplDigest(Application app) {
        long rc = 17L * (long) app.getName().hashCode();
        
        for (Artifact a : app.getArtifacts()) {
            URI uri = a.getUri();
            String uriStr = "";            
            if (uri != null) {
                uriStr = uri.toString();
            }
            rc = 37L * (long) uriStr.hashCode();
        }
        return rc;
    }

    private static String decodeURL(String url) {
        try {
            return URLDecoder.decode(url, "UTF-8");
        } catch (UnsupportedEncodingException e) {
            LOG.log(Level.WARNING, "Problem decoding URL", e);
            return url;
        }
    }
    
    private static String encodeURL(String name) {
        try {
            return URLEncoder.encode(name, "UTF-8");
        } catch (UnsupportedEncodingException e) {
            LOG.log(Level.WARNING, "Problem encoding URL", e);
            return name;
        }
    }
    
    private static String featureToXML(Application app, String id) {
        StringBuilder sb = new StringBuilder();
        
        sb.append("<features><feature name=\"" + id + "\">\n");
        
        addProperties(sb, "applicationProperties", app.getConfiguration());
        
        Map<String, Artifact> stillToAdd = new HashMap<String, Artifact>(app.getArtifacts().size());
        for (Artifact artifact : app.getArtifacts()) {
            stillToAdd.put(artifact.getGuid(), artifact);
        }
        for (Artifact artifact : app.getArtifacts()) {
            doAddBundleToFeature(artifact, sb, stillToAdd);
        }

        sb.append("</feature></features>\n");
        return sb.toString();
    }

    /**
     * add an artifact to the feature's bundles but make sure its deps get added 1st.
     * @param artifact
     * @param sb
     * @param stillToAdd
     */
    private static String doAddBundleToFeature(Artifact artifact,
                                               StringBuilder sb,
                                               Map<String, Artifact> stillToAdd) {
        String uriStr = null;
        
        StringBuffer deprefs = new StringBuffer();
        for (Artifact dep : artifact.getDepends()) {
            if (stillToAdd.get(dep.getGuid()) != null) {
                String depUri = doAddBundleToFeature(stillToAdd.get(dep.getGuid()), sb, stillToAdd);
                if (depUri != null) {
                    deprefs.append("<depends>").append(depUri).append("</depends>\n");
                }
            }
        }

        if ((stillToAdd.get(artifact.getGuid()) != null) && artifact.getUri() != null) {
            
            
            String filename = artifact.getFileName();
            sb.append("  <bundle ");
            
            if ((filename != null) && (!"".equals(filename))) {
                sb.append(" name=\"").append(filename).append("\"");
            }
            if (artifact.getType() != null) {
                sb.append(" type=\"").append(getArtifactTypeId(artifact.getType())).append("\"");
            }
            URI uri = artifact.getUri();
            if (artifact.isExternal()) {
                uriStr = uri.toString();
            } else {
                uriStr = uri.toString().replaceFirst("^http[s]?://[^/]*", LOCALHOST_MARKER);
            }
            sb.append(" uri=\"").append(uriStr).append("\"");
            
            if (deprefs.length() > 0) {
                sb.append(">\n").append(deprefs).append("</bundle>\n");
            } else {
                sb.append("/>\n");
            }
        }
        stillToAdd.remove(artifact.getGuid());
        
        return uriStr;
    }

    private static String getArtifactTypeId(ArtifactType artifactType) {
        String id = artifactType.getGuid();
        int i = id.indexOf('-');
        if (i == -1) {
            return id;
        } else {
            return id.substring(i + 1);
        }
    }
}
