/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.pac;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.net.URL;
import java.util.Collection;
import java.util.Collections;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import com.iona.soa.repository.util.InputStreamHelper;
import com.iona.soa.security.AuthenticationException;
import com.iona.soa.security.SecurityContext;
import com.iona.soa.security.SecurityService;
import com.iona.soa.security.SecurityUtil;

import org.osgi.service.http.HttpContext;


public class RepositoryHttpContext implements HttpContext {
    private static final Collection<String> AGENT_AREAS = 
        Collections.singletonList(SecurityUtil.AGENT_SECAREA_ID);
    private static final Logger LOG = Logger.getLogger(RepositoryHttpContext.class.getName());
    private static final ThreadLocal<RequestInfo> THREAD_LOCAL = new ThreadLocal<RequestInfo>();
    
    File storageDirectory;
    private SecurityService securityService;
    
    RepositoryHttpContext(SecurityService secSrvc, File storageDir) {
        securityService = secSrvc;
        storageDirectory = storageDir;
        
    }
    
    public boolean handleSecurity(
            HttpServletRequest request,
            HttpServletResponse response
    ) throws IOException {
        LOG.info("Provisioning request " + request.getMethod() + " " + request.getRequestURI());
        THREAD_LOCAL.set(null);
        try {
            String authHeader = request.getHeader("Authorization");
            SecurityContext ctx = 
                securityService.authenticate(authHeader);            
            if (!securityService.authorizeRead(ctx.getSSOToken(), null, AGENT_AREAS)) {
                String msg = "User " + ctx.getUser() + " is not authorized to "
                           + request.getMethod() + " " + request.getRequestURI();  
                LOG.severe(msg);
                response.sendError(401, msg);
                return false;
            }
            String[] data = SecurityUtil.parseAuthHeader(authHeader);
            if (data != null && SecurityService.BASIC_AUTH_TYPE.equals(data[0])) {
                RequestInfo ri = 
                    new RequestInfo(request.getScheme(), request.getHeader("Host"), data[1], data[2]);
                THREAD_LOCAL.set(ri);
            }
        } catch (AuthenticationException e) {
            String msg = "Authentication failed for request " + request.getMethod()
                       + " " + request.getRequestURI();
            LOG.severe(msg);
            response.sendError(401, msg);
            return false;
            
        }
        return true;
    }
    
    public URL getResource(String name) {
        LOG.fine("Obtaining resource file: " + name);
        
        
        File file = FileRegistry.getFile(storageDirectory, name);
        synchronized (file) {
            try {
                if (!file.exists()) {
                    LOG.warning("Resource not found: " + name);
                    return null;
                } else {
                    rewriteFile(file, THREAD_LOCAL.get());
                    return file.toURL();
                }
            } catch (IOException e) {
                LOG.log(Level.WARNING, "Problem obtaining resource file for " + name, e);
                return null;
            }
        }
    }
    
    private void rewriteFile(File in, RequestInfo ri) {
        if (ri == null || ri.getUser() == null || ri.getPassword() == null) {
            return;
        }
        String replaceStr = ri.getScheme() + "://" 
                          + ri.getUser() + ":" + ri.getPassword() 
                          + "@" + ri.getHost();
        File out = new File(in.getAbsolutePath() + ".tmp");
        BufferedReader reader = null;
        BufferedWriter writer = null;
        try {
            reader = new BufferedReader(new FileReader(in));
            writer = new BufferedWriter(new FileWriter(out));
            String line;
            while ((line = reader.readLine()) != null) {
                writer.write(line.replace(ProvisioningDataProvider.LOCALHOST_MARKER, replaceStr));
                writer.newLine();
            }
        } catch (IOException ioe) {
            InputStreamHelper.closeAll(writer);
            out.delete();
            return;
        } finally {
            InputStreamHelper.closeAll(reader, writer);
        }
        in.delete();
        out.renameTo(in);
    }

    public String getMimeType(String name) {
        return "text/xml";
    }
    
    private static class RequestInfo {
        private String scheme;
        private String host;
        private String user;
        private String password;
        public RequestInfo(String scheme, String host, String user, String password) {
            this.scheme = scheme;
            this.host = host;
            this.user = user;
            this.password = password;
        }
        public String getScheme() {
            return scheme;
        }
        public String getHost() {
            return host;
        }
        public String getUser() {
            return user;
        }
        public String getPassword() {
            return password;
        }
        public String toString() {
            return super.toString() + " [Host=" + host + ", user=" + user + ", passwd=********]";
        }
    }
};

