/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.adapters;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.logging.Logger;

import com.iona.repository.util.i18n.Message;
import com.iona.soa.repository.EMFUtil;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.TransactionAdapter;
import com.iona.soa.repository.TransactionData;
import com.iona.soa.repository.notification.AbstractRepositoryAdapter;
import com.iona.soa.security.AuthorizationException;
import com.iona.soa.security.MessageBundle;
import com.iona.soa.security.SecurityService;
import com.iona.soa.security.SecurityUtil;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;

public class ChangeAuthorizationAdapter extends AbstractRepositoryAdapter implements TransactionAdapter {
    private static final ResourceBundle BUNDLE = MessageBundle.getInstance();
    private static final Logger LOG = Logger.getLogger(ChangeAuthorizationAdapter.class.getName());

    private final SecurityService securityService;
    private TransactionData transactionData;
    private EObject ignoredContainerObj;
    private Collection<EObject> deletedObjects = new LinkedList<EObject>(); 
    
    public ChangeAuthorizationAdapter(SecurityService ss) {
        securityService = ss;
    }

    public int getPriority() {
        // almost the highest value possible as this adapter must be one of the 
        // first that is called when notifications are dispatched.
        return Integer.MAX_VALUE - 1000;
    }

    public void startTransaction(SOARepository repo, TransactionData txData) throws Exception {
        transactionData = txData == null ? new TransactionData() : txData;
        securityService.startTransaction(txData.getUserId());
    }

    public void commitTransaction() throws Exception {
        securityService.endTransaction();
        deletedObjects.clear();
        transactionData = null;
    }

    public void rollbackTransaction() throws Exception {
        securityService.endTransaction();
        deletedObjects.clear();
        transactionData = null;
    }

    @Override
    public void attachedObject(EObject object) {
        checkTxInProgress();
        Map<EObject, Boolean> deliverInfo = null;
        EObject container = object.eContainer();
        EReference ref = object.eContainmentFeature();
        Object newValue = null;
        if (container != null) {
            if (ref.isMany()) {
                newValue = Collections.singletonList(object);
            } else {
                newValue = object;
            }
            deliverInfo = undoRefChange(container, ref, null, newValue, true);
        }
        try {
            if (!securityService.authorizeCreate(transactionData.getSsoToken(), 
                                                 object.eClass(),
                                                 container)) {
                throwException(transactionData.getUserId(), "create", "type", 
                               object.eClass().getName(), object);
            }
            ignoredContainerObj = container;
        } finally {
            if (deliverInfo != null) {
                undoRefChange(container, ref, newValue, null, false);
                reenableDeliver(deliverInfo);
            }
        }
    }

    @Override
    public void changedAttribute(EObject object, EAttribute attribute, Object oldValue, Object newValue) {
        checkTxInProgress();
        if (deletedObjects.contains(object)) {
            return;
        }
        if (!securityService.authorizeUpdate(transactionData.getSsoToken(), object, attribute)) {
            String name = EMFUtil.getName(object);
            if (attribute.equals(object.eClass().getEStructuralFeature("name")) 
                && oldValue instanceof String) {
                name = (String)oldValue;
            }
            throwException(transactionData.getUserId(), "update", "object", name, object);  
        }
    }

    @Override
    public void changedReference(EObject object, EReference ref, Object oldValue, Object newValue) {
        checkTxInProgress();
        if (deletedObjects.contains(object)) {
            return;
        }
        if (ignoredContainerObj != null) {
            if (object == ignoredContainerObj) {
                ignoredContainerObj = null;
                return;
            }
            LOG.warning("Unexpected reference change after attach/detach event: " + ignoredContainerObj);
            ignoredContainerObj = null;
        }
        Map<EObject, Boolean> deliverInfo = 
            undoRefChange(object, ref, oldValue, newValue, true);
        try {
            if (!securityService.authorizeUpdate(transactionData.getSsoToken(), object, ref)) {
                throwException(transactionData.getUserId(), "update", "object", 
                               EMFUtil.getName(object), object);  
            }
        } finally {
            undoRefChange(object, ref, newValue, oldValue, false);
            reenableDeliver(deliverInfo);
        }
    }

    @Override
    public void detachedObject(EObject object, EObject oldContainer, EReference containerRef) {
        checkTxInProgress();
        deletedObjects.add(object);
        if (oldContainer != null && deletedObjects.contains(oldContainer)) {
            return;
        }
        Map<EObject, Boolean> deliverInfo = null;
        Object oldValue = null;
        if (oldContainer != null) {
            if (containerRef.isMany()) {
                oldValue = Collections.singletonList(object);
            } else {
                oldValue = object;
            }
            deliverInfo = undoRefChange(oldContainer, containerRef, oldValue, null, true);
        }
        try {
            if (!securityService.authorizeDelete(transactionData.getSsoToken(), object)) {
                throwException(transactionData.getUserId(), "delete", "object", 
                               EMFUtil.getName(object), object);
            }
            ignoredContainerObj = oldContainer;
        } finally {
            if (deliverInfo != null) {
                undoRefChange(oldContainer, containerRef, null, oldValue, false);
                reenableDeliver(deliverInfo);
            }
        }
    }

    @SuppressWarnings("unchecked")
    private Map<EObject, Boolean> undoRefChange(EObject eObj, EReference ref, 
                                                Object oldValue, Object newValue, boolean disableDeliver) {
        Map<EObject, Boolean> ret = null;
        if (disableDeliver) {
            ret = new HashMap<EObject, Boolean>();
            ret.put(eObj, eObj.eDeliver());
            eObj.eSetDeliver(false);
        }
        if (ref.isMany()) {
            Collection<EObject> noObjects = Collections.emptyList();
            Collection<EObject> oldObjs = oldValue == null ? noObjects : (Collection<EObject>)oldValue;
            Collection<EObject> newObjs = newValue == null ? noObjects : (Collection<EObject>)newValue;
            if (disableDeliver && ref.getEOpposite() != null) {
                for (EObject refObj : oldObjs) {
                    ret.put(refObj, refObj.eDeliver());
                    refObj.eSetDeliver(false);
                }
                for (EObject refObj : newObjs) {
                    ret.put(refObj, refObj.eDeliver());
                    refObj.eSetDeliver(false);
                }
            }
            ((Collection<EObject>)eObj.eGet(ref)).addAll(oldObjs);
            ((Collection<EObject>)eObj.eGet(ref)).removeAll(newObjs);
        } else {
            if (disableDeliver && ref.getEOpposite() != null) {
                if (oldValue != null) {
                    EObject refObj = (EObject)oldValue;
                    ret.put(refObj, refObj.eDeliver());
                    refObj.eSetDeliver(false);
                }
                if (newValue != null) {
                    EObject refObj = (EObject)newValue;
                    ret.put(refObj, refObj.eDeliver());
                    refObj.eSetDeliver(false);
                }
            }
            if (oldValue == null) {
                eObj.eUnset(ref);
            } else {
                eObj.eSet(ref, oldValue);
            }
        }
        return ret;
    }

    private void reenableDeliver(Map<EObject, Boolean> deliverInfo) {
        for (EObject eObj : deliverInfo.keySet()) {
            eObj.eSetDeliver(deliverInfo.get(eObj));
        }
    }

    private void checkTxInProgress() {
        if (transactionData == null) {
            Message msg = new Message(BUNDLE, "NO_WRITE_TX");
            LOG.warning(msg.toString());
            throw new AuthorizationException(msg);
        }
    }

    private void throwException(String userId, String action, String type, String id, EObject eObj) {
        Collection<String> areas = SecurityUtil.getSecurityAreas(eObj, true);
        String area;
        if (areas.size() == 0) {
            area = "???";
        } else if (areas.size() == 1) {
            area = areas.iterator().next();
        } else {
            StringBuilder sb = new StringBuilder();
            sb.append("[");
            boolean first = true;
            for (String s : areas) {
                if (first) {
                    first = false;
                } else {
                    sb.append(", ");
                }
                sb.append(s);
            }
            sb.append("]");
            area = sb.toString();
        }
        Message msg = new Message(BUNDLE, "NOT_AUTHORIZED", userId, action, type, id, area);
        LOG.info(msg.toString());
        throw new AuthorizationException(msg);
    }
}
