/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.adapters;

import java.util.Collection;
import java.util.Collections;
import java.util.logging.Logger;

import com.iona.soa.expression.Equals;
import com.iona.soa.model.folder.FolderPackage;
import com.iona.soa.model.folder.IFolder;
import com.iona.soa.model.folder.IFolderContent;
import com.iona.soa.model.repository.IPersistableObject;
import com.iona.soa.model.repository.IRepositoryObject;
import com.iona.soa.model.repository.IVersionedObject;
import com.iona.soa.model.repository.RepositoryPackage;
import com.iona.soa.model.security.Role;
import com.iona.soa.model.security.User;
import com.iona.soa.model.userattributes.Dictionary;
import com.iona.soa.model.userattributes.UserAttributeDefinition;
import com.iona.soa.repository.EMFUtil;
import com.iona.soa.repository.notification.AbstractRepositoryAdapter;
import com.iona.soa.repository.status.StatusMessage;
import com.iona.soa.repository.status.ValidationException;
import com.iona.soa.resourceset.ResourceSetQueryHandler;
import com.iona.soa.security.SecurityUtil;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.resource.Resource;

public class ChangeValidationAdapter extends AbstractRepositoryAdapter {
    protected static final EAttribute MOD_ATTR = 
        RepositoryPackage.eINSTANCE.getIPersistableObject_Modifiable();
    protected static final EAttribute USERID_ATTR = 
        RepositoryPackage.eINSTANCE.getIPersistableObject_LastEditedBy();
    protected static final EAttribute TIMESTAMP_ATTR = 
        RepositoryPackage.eINSTANCE.getIPersistableObject_Timestamp();
    protected static final String UNKNOWN_USERID = "unknown";
    private static final Logger LOG = Logger.getLogger(ChangeValidationAdapter.class.getName());
    protected Boolean enableChangeValidator;
    
    public ChangeValidationAdapter(Boolean enableChangeValidator) {
        this.enableChangeValidator = enableChangeValidator;
    }

    @Override
    public void changedAttribute(EObject eObject, EAttribute attribute, Object oldValue, Object newValue) {
        
        if (!enableChangeValidator) {
            return;
        }
        
        EAttribute genAttr = EMFUtil.getGenerationNumberAttribute(eObject.eClass());
        
        if (attribute == MOD_ATTR) {
            if (Boolean.FALSE.equals(oldValue) && Boolean.TRUE.equals(newValue)) {
                IPersistableObject obj = upCast(eObject, IPersistableObject.class);
                obj.setModifiable(false);
                LOG.warning(StatusMessage.CANT_MODIFY_UNMODIFIABLE_MSG.getMessage(
                        EMFUtil.getLongName(eObject)));
                throw new RuntimeException(new ValidationException(
                        StatusMessage.CANT_MODIFY_UNMODIFIABLE_MSG, EMFUtil.getLongName(eObject)));
            }
        } else if (attribute != genAttr && attribute != USERID_ATTR && attribute != TIMESTAMP_ATTR) {
            checkModifiable(upCast(eObject, IPersistableObject.class));

            if (attribute.getName() == RepositoryPackage.Literals.IREPOSITORY_OBJECT__NAME.getName()) {
                checkDuplicateEntry(upCast(eObject, IRepositoryObject.class));
            }
        }

        checkChangeInFolder(eObject, attribute, newValue);
    }

    @Override
    public void changedReference(EObject eObject, EReference ref, Object oldValue, Object newValue) {

        if (!enableChangeValidator) {
            return;
        }
        checkModifiable(upCast(eObject, IPersistableObject.class));

        checkChangeInFolder(eObject, ref, newValue);
    }

    @Override
    public void attachedObject(EObject eObject) {

        if (!enableChangeValidator) {
            return;
        }
        
        checkDuplicateEntry(upCast(eObject, IRepositoryObject.class));
        checkDuplicateUserEntry(upCast(eObject, User.class));
        checkUniqueInFolder(upCast(eObject, IFolder.class));
        checkUniqueInFolder(upCast(eObject, IFolderContent.class));
    }

    @Override
    public void detachedObject(EObject eObject, EObject oldContainer, EReference containerRef) {

        if (!enableChangeValidator) {
            return;
        }

        checkModifiable(upCast(eObject, IPersistableObject.class));
        checkNotAdmin(upCast(eObject, Role.class), upCast(eObject, User.class));
    }

    private void checkNotAdmin(Role role, User user) {
        if (role != null && SecurityUtil.ADMIN_ROLE_GUID.equals(role.getGuid())
            || user != null && SecurityUtil.ADMIN_USER_GUID.equals(user.getGuid())) {
            IPersistableObject obj = role != null ? role : user;
            LOG.warning(StatusMessage.CANT_DELETE_USR_MSG.getMessage(EMFUtil.getName(obj)));
            throw new RuntimeException(new ValidationException(
                    StatusMessage.CANT_DELETE_USR_MSG, EMFUtil.getName(obj)));            
        }
    }

    private void checkModifiable(IPersistableObject ipObject) {
        if (ipObject != null && !ipObject.isModifiable()) {
            LOG.warning(StatusMessage.CANT_MODIFY_UNMODIFIABLE_MSG.getMessage(
                    EMFUtil.getLongName(ipObject)));
            throw new RuntimeException(new ValidationException(
                    StatusMessage.CANT_MODIFY_UNMODIFIABLE_MSG, EMFUtil.getLongName(ipObject)));
        }
    }
    
    @SuppressWarnings("unchecked")
    private void checkDuplicateEntry(IRepositoryObject obj) {
        
        if (obj == null || obj.eResource() == null) {
            return;
        }
      
        if (obj instanceof IVersionedObject) {
            return;
        }
        
        if (obj instanceof UserAttributeDefinition) {            
            UserAttributeDefinition defn = (UserAttributeDefinition) obj;
            Dictionary dict = (Dictionary) defn.eContainer();
            checkForDuplicateFromCollection(defn, dict.getAttributeDefs());
            
        } else if (obj instanceof Dictionary) {            
            IRepositoryObject iObj = (IRepositoryObject) obj;
            
            Collection<EObject> existing;
            try {
                existing = ResourceSetQueryHandler.executeQuery(obj.eResource().getResourceSet(), 
                    iObj.eClass(), new Equals("name", iObj.getName()), false);
            } catch (Exception e) {
                LOG.warning("Error when querying Repository " + e.getMessage());
                throw new RuntimeException(e);                        
            }                            

            for (EObject found : existing) {
                if (((IRepositoryObject)found).getGuid() != iObj.getGuid()) {
                    LOG.warning(StatusMessage.DUPLICATE_NAMED_OBJECT_FOUND.getMessage(
                            iObj.getName()));
                    throw new RuntimeException(new ValidationException(
                            StatusMessage.DUPLICATE_NAMED_OBJECT_FOUND, 
                            iObj.getName()));                        
                }
            }
        }        
    }

    private void checkForDuplicateFromCollection(IRepositoryObject iObj, 
                                                 Collection<? extends IRepositoryObject> existing) {
        for (IRepositoryObject found : existing) {
            if (found.getGuid() != iObj.getGuid() && found.getName().equals(iObj.getName())) {
                LOG.warning(StatusMessage.DUPLICATE_NAMED_OBJECT_FOUND.getMessage(
                        iObj.getName()));
                throw new RuntimeException(new ValidationException(
                        StatusMessage.DUPLICATE_NAMED_OBJECT_FOUND, 
                        iObj.getName()));                        
            }
        }
    }

    private void checkDuplicateUserEntry(User user) {
        
        if (user == null || user.eResource() == null) {
            return;
        }
      
        if (user instanceof IVersionedObject) {
            return;
        }

        Resource resource = user.eResource();
        for (Object item : resource.getContents()) {
            User otherUser = upCast(item, User.class);
            if (otherUser != null 
                && !otherUser.getGuid().equals(user.getGuid())
                && otherUser.getUserid().equals(user.getUserid())) {
                
                LOG.warning(StatusMessage.DUPLICATE_NAMED_OBJECT_FOUND.getMessage(
                        user.getName()));
                throw new RuntimeException(new ValidationException(
                        StatusMessage.DUPLICATE_NAMED_OBJECT_FOUND, 
                        user.getName()));
            }
        }
    }
    
    private void checkUniqueInFolder(IFolderContent folderContent) {
        if (folderContent != null) {
            checkUniqueInFolder(folderContent.getBelongsTo(), Collections.singletonList(folderContent));
        }
    }

    private void checkUniqueInFolder(IFolder folder) {
        if (folder != null) {
            checkUniqueInFolder(folder.getParent(), Collections.singletonList(folder));
        }
    }

    private void checkChangeInFolder(EObject eObject, EAttribute attr, Object newValue) {
        if (attr == RepositoryPackage.eINSTANCE.getIRepositoryObject_Name()) {
            checkUniqueInFolder(upCast(eObject, IFolder.class));
            checkUniqueInFolder(upCast(eObject, IFolderContent.class));
        }
        
    }

    private void checkChangeInFolder(EObject object, EReference ref, Object newValue) {
        if (ref == FolderPackage.eINSTANCE.getIFolder_Parent()
            || ref == FolderPackage.eINSTANCE.getIFolderContent_BelongsTo()) {
            checkUniqueInFolder(upCast(newValue, IFolder.class), Collections.singletonList(object));
        } else if (ref == FolderPackage.eINSTANCE.getIFolder_Children()
            || ref == FolderPackage.eINSTANCE.getIFolder_Content()) {
            checkUniqueInFolder(upCast(object, IFolder.class), upCastList(newValue, EObject.class));
        }
        
    }

    private void checkUniqueInFolder(IFolder parent, Collection<? extends EObject> eObjects) {
        if (parent != null && eObjects != null) {
            for (EObject eObject : eObjects) {
                if (duplicateNameInList(parent.getChildren(), eObject) 
                    || duplicateNameInList(parent.getContent(), eObject)) {
                    LOG.warning(StatusMessage.DUPLICATE_NAMED_OBJECT_IN_FOLDER.getMessage(
                                EMFUtil.getName(eObject), EMFUtil.getName(parent)));
                    throw new RuntimeException(new ValidationException(
                                               StatusMessage.DUPLICATE_NAMED_OBJECT_IN_FOLDER, 
                                               EMFUtil.getName(eObject), EMFUtil.getName(parent)));
                }
            }
        }
    }
    
    private boolean duplicateNameInList(Collection<? extends EObject> objs, EObject namedEObject) {
        String name = EMFUtil.getName(namedEObject);
        for (EObject obj : objs) {
            if (obj != namedEObject && obj.eClass() == namedEObject.eClass() 
                && name.equals(EMFUtil.getName(obj))) {
                return true;
            }
        }
        return false;
    }
}
