/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.adapters;

import java.util.LinkedHashSet;
import java.util.Set;

import com.iona.soa.model.repository.RepositoryPackage;
import com.iona.soa.repository.EMFUtil;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.TransactionAdapter;
import com.iona.soa.repository.TransactionData;
import com.iona.soa.repository.notification.AbstractRepositoryAdapter;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;

/**
 * This adapter updates the Timestamp, userid and generation number of all modified objects.
 * This is done once per db transaction.
 *   
 * @author GREHEINE
 */
public class PerTransactionUpdatesAdapter extends AbstractRepositoryAdapter implements TransactionAdapter {
    private static final String UNKNOWN_USERID = "unknown";
    private static final EAttribute CREATED_BY_ATTR = 
        RepositoryPackage.eINSTANCE.getIPersistableObject_CreatedBy();
    private static final EAttribute LAST_USER_ATTR = 
        RepositoryPackage.eINSTANCE.getIPersistableObject_LastEditedBy();
    private static final EAttribute TIMESTAMP_ATTR = 
        RepositoryPackage.eINSTANCE.getIPersistableObject_Timestamp();
    private static final boolean GEN_NUMBERS_DISABLED = 
        Boolean.parseBoolean(System.getProperty("disable.generation.numbers", "false"));

    private final Set<EObject> updatedObjects = new LinkedHashSet<EObject>();
    private TransactionData transactionData;
    private State state = State.INACTIVE;
    
    public int getPriority() {
        // any adapter with a priority higher than 100 must *not* change any objects during commit
        return 100;
    }

    public synchronized void startTransaction(SOARepository repo, TransactionData txData) throws Exception {
        if (state != State.INACTIVE) {
            state = State.INACTIVE;
            throw new RuntimeException("Transaction already in progress");
        }
        updatedObjects.clear();
        transactionData = txData == null ? new TransactionData() : txData;
        state = State.TX_IN_PROGRESS;
    }
    
    public synchronized void commitTransaction() throws Exception {
        endTransaction();
    }

    public synchronized void rollbackTransaction() throws Exception {
        endTransaction();
    }
    
    private void endTransaction() {
        if (state != State.TX_IN_PROGRESS) {
            state = State.INACTIVE;
            throw new RuntimeException("No transaction in progress");
        }
        updatedObjects.clear();
        state = State.INACTIVE;
    }
    
    @Override
    public void attachedObject(EObject object) {
        if (state == State.TX_IN_PROGRESS) {
            try {
                state = State.UPDATING;
                updateObject(object, true);
            } finally {
                state = State.TX_IN_PROGRESS;
                updatedObjects.add(object);
            }
        } else if (state != State.UPDATING) {
            throw new RuntimeException("No transaction in progress");
        }
    }

    @Override
    public void changedAttribute(EObject object, EAttribute attribute,
            Object oldValue, Object newValue) {
        if (oldValue == null && newValue != null || oldValue != null && !oldValue.equals(newValue)) {
            changedObject(object);
        }
    }

    @Override
    public void changedReference(EObject object, EReference ref,
            Object oldValue, Object newValue) {
        if (oldValue == null && newValue != null || oldValue != null && !oldValue.equals(newValue)) {
            changedObject(object);
        }
    }

    @Override
    public void detachedObject(EObject object, EObject oldContainer, EReference containerRef) {
        if (state == State.TX_IN_PROGRESS) {
            updatedObjects.add(object);
        } else if (state != State.UPDATING) {
            throw new RuntimeException("No transaction in progress");
        }
    }

    private void changedObject(EObject object) {
        if (state == State.TX_IN_PROGRESS) {
            try {
                state = State.UPDATING;
                updateObject(object, false);
            } finally {
                state = State.TX_IN_PROGRESS;
                updatedObjects.add(object);
            }
        } else if (state != State.UPDATING) {
            throw new RuntimeException("No transaction in progress");
        }
    }
    
    private void updateObject(EObject eObj, boolean isNewObject) {
        if (!updatedObjects.contains(eObj) 
            && RepositoryPackage.eINSTANCE.getIPersistableObject().isSuperTypeOf(eObj.eClass())) {
            String userId = transactionData.getUserId() == null 
                            ? UNKNOWN_USERID 
                            : transactionData.getUserId();
            eObj.eSet(LAST_USER_ATTR, userId);
            eObj.eSet(TIMESTAMP_ATTR, transactionData.getTimestamp());
            if (isNewObject) {
                eObj.eSet(CREATED_BY_ATTR, userId);
            } else if (!GEN_NUMBERS_DISABLED) {
                EAttribute genNrAttr = EMFUtil.getGenerationNumberAttribute(eObj.eClass());
                if (genNrAttr != null) {
                    Long storedVersion = (Long)eObj.eGet(genNrAttr);
                    eObj.eSet(genNrAttr, storedVersion + 1);
                }
            }
        }
    }
    
    private enum State {
        TX_IN_PROGRESS,
        UPDATING,
        INACTIVE;
    }
}
