/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.event;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.soa.expression.Equals;
import com.iona.soa.expression.Expression;
import com.iona.soa.expression.Operand;
import com.iona.soa.model.events.DataChangeEvent;
import com.iona.soa.model.events.DataChangeFieldDelta;
import com.iona.soa.model.events.DataChangeKind;
import com.iona.soa.model.events.EventsFactory;
import com.iona.soa.model.repository.IRepositoryObject;
import com.iona.soa.model.repository.RepositoryPackage;
import com.iona.soa.model.security.SecurityPackage;
import com.iona.soa.model.security.User;
import com.iona.soa.repository.EMFUtil;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.TransactionAdapter;
import com.iona.soa.repository.TransactionData;
import com.iona.soa.repository.notification.AbstractRepositoryAdapter;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

public class DataChangeAdapter extends AbstractRepositoryAdapter implements TransactionAdapter {
    private static DataChangeAdapter instance;
    private static final Logger LOG = Logger.getLogger(DataChangeAdapter.class.getName());
    
    private SOARepository soaRepo;
    private TransactionData transactionData;
    private List<DataChangeEventObserver> observers = new ArrayList<DataChangeEventObserver>();
    private Map<String, DataChangeEvent> processedEvents = new HashMap<String, DataChangeEvent>();
    private boolean asycDispatch;
    
    protected DataChangeAdapter(boolean asycDispatch) { 
        this.asycDispatch = asycDispatch;
    }
    
    public static DataChangeAdapter getInstance() {
        if (instance == null) {
            instance = new DataChangeAdapter(true);
        }
        
        return instance;
    }
    
    public boolean registerObserver(DataChangeEventObserver newObserver) {
        return newObserver != null && observers.add(newObserver);
    }

    public boolean unregisterObserver(DataChangeEventObserver newObserver) {
        return newObserver != null && observers.remove(newObserver);
    }
    
    /**
     * Intended for use in testing only.
     */
    public void clearObservers() {
        observers.clear();
    }
    
    @Override
    public void attachedObject(EObject object) {
        DataChangeEvent evt = createEvent(object, DataChangeKind.ADD_LITERAL);
        if (evt == null) {
            return;
        }

        for (EStructuralFeature feature : object.eClass().getEAllStructuralFeatures()) {

            Object featureData = object.eGet(feature);

            if (feature.isMany()) {
                List<?> newValuesObj = (List<?>)featureData;

//              only add a field delta if the field actually contains some data
                if (newValuesObj != null && newValuesObj.size() > 0) {

                    List<String> newValues = new ArrayList<String>(newValuesObj.size());
                    for (Object newValueObj  : newValuesObj) {
                        newValues.add(serializeFieldValueToString(newValueObj));
                    }
                    DataChangeFieldDelta delta = EventsFactory.eINSTANCE.createDataChangeFieldDelta();
                    delta.setFieldDesc(feature);
                    delta.getNewValues().addAll(newValues);
                    evt.getUpdatedFields().add(delta);
                }

            } else if (featureData != null) {
                DataChangeFieldDelta delta = EventsFactory.eINSTANCE.createDataChangeFieldDelta();
                delta.setFieldDesc(feature);
                delta.setNewValue(serializeFieldValueToString(featureData));
                evt.getUpdatedFields().add(delta);
            }
        }

        processedEvents.put(evt.getTargetGuid(), evt);
    }

    @Override
    public void changedAttribute(EObject object, EAttribute attribute, Object oldValue, Object newValue) {
        changedFeature(object, attribute, oldValue, newValue);
    }

    @Override
    public void changedReference(EObject object, EReference ref, Object oldValue, Object newValue) {
        changedFeature(object, ref, oldValue, newValue);
    }

    public void changedFeature(EObject object, EStructuralFeature feature, Object oldValue, Object newValue) {
        DataChangeEvent evt;
        if (processedEvents.get(EMFUtil.getGuid(object)) != null) {
            evt = processedEvents.get(EMFUtil.getGuid(object));
            
            // if the obj is about to be removed then updates to it do not really matter 
            if (evt != null && evt.getChangeKind() == DataChangeKind.REMOVE_LITERAL) {
                return;
            }
        } else {
            evt = createEvent(object, DataChangeKind.MODIFY_LITERAL);
            if (evt == null) {
                return;
            } else if (RepositoryPackage.eINSTANCE.getIRepositoryObject_Name().equals(feature)) {
                evt.setTargetName((String)oldValue);
            } else if (RepositoryPackage.eINSTANCE.getIRepositoryObject_Namespace().equals(feature)) {
                evt.setTargetNamespace((String)oldValue);
            }
        }
        
        DataChangeFieldDelta delta = getDelta(evt.getUpdatedFields(), feature.getName());
        boolean newDelta = delta == null;
        if (delta == null) {
            delta = EventsFactory.eINSTANCE.createDataChangeFieldDelta();
            delta.setFieldDesc(feature);
            evt.getUpdatedFields().add(delta);
        }

        if (feature.isMany()) {
            List<String> currentValues = new ArrayList<String>();
            if (object.eIsSet(feature)) {
                for (Object val : (List<?>)object.eGet(feature)) {
                    currentValues.add(serializeFieldValueToString(val));
                }
            }
            delta.getNewValues().clear();
            delta.getNewValues().addAll(currentValues);
            
            if (newDelta) {
                List<String> oldValues = new ArrayList<String>(currentValues);
                if (oldValue instanceof List) {
                    for (Object val  : (List<?>)oldValue) {
                        oldValues.add(serializeFieldValueToString(val));
                    }
                } else if (oldValue != null) {
                    oldValues.add(serializeFieldValueToString(oldValue));
                }
                if (newValue instanceof List) {
                    for (Object val  : (List<?>)newValue) {
                        oldValues.remove(serializeFieldValueToString(val));
                    }
                } else if (newValue != null) {
                    oldValues.remove(serializeFieldValueToString(newValue));
                }

                delta.getOldValues().addAll(oldValues);
            }
        } else {
            if (newDelta) {
                delta.setOldValue(serializeFieldValueToString(oldValue));
            }
            delta.setNewValue(serializeFieldValueToString(newValue));
        }

        processedEvents.put(evt.getTargetGuid(), evt);
    }

    @Override
    public void detachedObject(EObject object, EObject oldContainer, EReference containerRef) {
        if (processedEvents.get(EMFUtil.getGuid(object)) != null) {
            DataChangeEvent processedEvent = processedEvents.get(EMFUtil.getGuid(object));
            
            // since the obj is about to be removed, other events don't really matter
            processedEvents.remove(processedEvent.getTargetGuid());
            
            // if the already processed event was for an addition then the 2 cancel each other out
            if (processedEvent.getChangeKind() == DataChangeKind.ADD_LITERAL) {
                return;
            }
        }
        
        DataChangeEvent evt = createEvent(object, DataChangeKind.REMOVE_LITERAL);
        if (evt == null) {
            return;
        }


        for (EStructuralFeature feature : object.eClass().getEAllStructuralFeatures()) {

            Object featureData = object.eGet(feature);
            
            
            if (feature.isMany()) {
                List<?> oldValuesObj = (List<?>)featureData;
                
                // only add a field delta if the field actually contains some data
                if (oldValuesObj != null && oldValuesObj.size() > 0) {
                    
                    List<String> oldValues = new ArrayList<String>(oldValuesObj.size());
                    for (Object oldValueObj  : oldValuesObj) {
                        oldValues.add(serializeFieldValueToString(oldValueObj));
                    }
                    
                    DataChangeFieldDelta delta = EventsFactory.eINSTANCE.createDataChangeFieldDelta();
                    delta.setFieldDesc(feature);
                    delta.getOldValues().addAll(oldValues);
                    delta.getNewValues().addAll(new ArrayList<String>(0));
                    evt.getUpdatedFields().add(delta);
                }
            } else if (featureData != null) {
                DataChangeFieldDelta delta = EventsFactory.eINSTANCE.createDataChangeFieldDelta();
                delta.setFieldDesc(feature);
                delta.setOldValue(serializeFieldValueToString(featureData));
                evt.getUpdatedFields().add(delta);
            }
        }
        
        processedEvents.put(evt.getTargetGuid(), evt);
    }

    /**
     * @param target
     * @param userId
     * @param timestamp
     * @param kind
     * @return
     */
    private DataChangeEvent createEvent(EObject target, DataChangeKind kind) {
        if (transactionData == null) {
            return null;
        }
        
        DataChangeEvent evt = EventsFactory.eINSTANCE.createDataChangeEvent();

        String guid = EMFUtil.getGuid(target);
        evt.setTargetGuid(guid);
        evt.setTargetType(target.eClass());
        evt.setChangeKind(kind);
        evt.setUserId(transactionData.getUserId());
        evt.setUserToken(transactionData.getSsoToken());
        evt.setTimestamp(transactionData.getTimestamp());

        if (target instanceof IRepositoryObject) {
            evt.setTargetName(((IRepositoryObject)target).getName());
            evt.setTargetNamespace(((IRepositoryObject)target).getNamespace());
        }
        
        User user = getUser();
        if (user != null) {
            evt.setUserName(user.getName());
            evt.setUserEMail(user.getEmail());
        }

        return evt;
    }

    private String serializeFieldValueToString(Object featureData) {
        
        String newValue = featureData == null
                            ? null
                            : featureData instanceof String
                                ? (String)featureData
                                : featureData instanceof EObject
                                    ? getFeatureValue((EObject)featureData, "name") != null
                                        ? getFeatureValue((EObject)featureData, "name")
                                        : getFeatureValue((EObject)featureData, "description")
                                    : featureData.toString();
        return newValue;
    }

    /**
     * @param repositoryObject
     * @param featureName TODO
     * @return
     */
    private String getFeatureValue(final EObject repositoryObject, String featureName) {
        EStructuralFeature feature = repositoryObject.eClass().getEStructuralFeature(featureName);
        if (feature != null && feature instanceof EAttribute) {
            final Object value = repositoryObject.eGet(feature);
            if (value != null) {
                return value.toString();
            }
        }
        return null;
    }

    private DataChangeFieldDelta getDelta(List<DataChangeFieldDelta> pool, String fieldName) {
        if (fieldName != null && fieldName.trim().length() > 0 && pool != null && pool.size() > 0) {
            for (DataChangeFieldDelta delta : pool) {
                if (fieldName.equals(delta.getFieldDesc().getName())) {
                    return delta;
                }
            }
        }
        return null;
    }

    private User getUser() {
        if (soaRepo != null && transactionData.getUserId() != null) {
            try {
                Operand idOper = new Operand(SecurityPackage.Literals.USER__USERID.getName());
                Operand valueOper = new Operand(transactionData.getUserId());
                Expression expr = new Equals(idOper, valueOper);
                EClass userCls = SecurityPackage.Literals.USER;
                Collection<EObject> objects = soaRepo.getResourceSet().executeQuery(userCls, 
                                                                                    expr, 
                                                                                    true);
                if (!objects.isEmpty()) {
                    return (User)objects.iterator().next();
                }
            } catch (Exception ex) {
                LOG.log(Level.SEVERE, 
                        "Error occured retrieving user <" + transactionData.getUserId() + ">", ex);
            }
        }
        return null;
    }

    public void startTransaction(SOARepository repo, TransactionData txData) throws Exception {
        soaRepo = repo;
        transactionData = txData == null ? new TransactionData() : txData;
    }
    
    public void commitTransaction() throws Exception {
        final Collection<DataChangeEvent> events = 
            new ArrayList<DataChangeEvent>(processedEvents.values());
        final Collection<DataChangeEventObserver> watchers = 
            new ArrayList<DataChangeEventObserver>(observers);
        if (events.size() > 0 && watchers.size() > 0) {
            if (asycDispatch) {
                // dispatch events asynchronously to work around a 
                // deadlock in the ProvisioningDataProvider
                new Thread() {
                    public void run() {
                        notifyObservers(watchers, events);
                    }
                } .start();
            } else {
                notifyObservers(watchers, events);
            }
        }
        processedEvents.clear();
        transactionData = null;
    }


    public void rollbackTransaction() throws Exception {
        processedEvents.clear();
        transactionData = null;
    }

    public int getPriority() {
        return 0;
    }


    public static String eventAsExternalString(DataChangeEvent evt) {
        StringBuffer buf = new StringBuffer();
        buf.append("\nData change event:");
        buf.append("\n      Target GUID: ");
        buf.append(evt.getTargetGuid());
        buf.append("\n      Target Type: ");
        buf.append(evt.getTargetType().getName());
        buf.append("\n      Target name: ");
        buf.append(evt.getTargetName());
        buf.append("\n      Target namespace: ");
        buf.append(evt.getTargetNamespace());
        buf.append("\n      Change Type: ");
        buf.append(evt.getChangeKind());
        buf.append("\n      Change Timestamp: ");
        buf.append(evt.getTimestamp());
        buf.append("\n      User ID: ");
        buf.append(evt.getUserId());
        buf.append("\n      User Name: ");
        buf.append(evt.getUserName());
        buf.append("\n      field deltas: ");
        
        for (DataChangeFieldDelta delta : (List<DataChangeFieldDelta>)evt.getUpdatedFields()) {
            final boolean isMany = delta.getFieldDesc().isMany();
            final String oldVal = isMany ? delta.getOldValues().toString() : delta.getOldValue();
            final String newVal = isMany ? delta.getNewValues().toString() : delta.getNewValue();

            buf.append(delta.getFieldDesc().getName());
            buf.append(" [ ");
            buf.append(oldVal == null ? "***" : oldVal);
            buf.append(" --> ");
            buf.append(newVal == null ? "***" : newVal);
            buf.append(" ] ");
            buf.append("\n                    ");

        }

        return buf.toString();
    }
    
    private static void notifyObservers(Collection<DataChangeEventObserver> observers, 
                                        Collection<DataChangeEvent> events) {
        for (DataChangeEvent event : events) {
            for (DataChangeEventObserver watcher : observers) {
                watcher.dataChanged(event);
            }
        }
    }
}
