/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.web.repository.agentprovider.providers;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Logger;

import com.iona.soa.repository.configuration.ConfigurationService;

public abstract class AbstractAgentFileProvider implements IAgentFileProvider {

    protected static final Logger LOG = Logger.getLogger(AbstractAgentFileProvider.class.getName());
    private static final String BUNDLE_ARTIFACTS_KEY = ".agent.artifacts";
    private static final String PASSWORD_FILE_LOCATION_KEY = ".password.file.location";
    private static final String PROPERTIES_FILE_LOCATION_KEY = ".properties.file.location";
    private static final String AGENT_DEPLOY_DIR_KEY = ".deploy.dir.location";
    private static final String AGENT_CONTROLER_URL_PROPERTY_KEY = ".controller.url.property";
    private static final String AGENT_USER_PROPERTY_KEY = ".agent.user.property";

    
    protected String passwordFileLocation = "etc/agent.password";
    protected String propertiesFileLocation = "etc/agent.properties";
    protected String agentDeployDir = "hotdeploy/";
    protected String agentBundleLocation = "";
    protected String controllerUrlProperty = "controllerRootUri";
    protected String agentUserProperty = "agent.user";
    protected String configPrefix;
    protected ConfigurationService configService;

    public AbstractAgentFileProvider(ConfigurationService configService, String configPrefix) {
        this.configPrefix = configPrefix;
        this.configService = configService;
        this.agentBundleLocation = getConfigValue(agentBundleLocation, BUNDLE_ARTIFACTS_KEY);
        this.passwordFileLocation = getConfigValue(passwordFileLocation, PASSWORD_FILE_LOCATION_KEY);
        this.propertiesFileLocation = getConfigValue(propertiesFileLocation, PROPERTIES_FILE_LOCATION_KEY);
        this.agentDeployDir = getConfigValue(agentDeployDir, AGENT_DEPLOY_DIR_KEY);
        this.controllerUrlProperty = getConfigValue(controllerUrlProperty, AGENT_CONTROLER_URL_PROPERTY_KEY);
        this.agentUserProperty = getConfigValue(agentUserProperty, AGENT_USER_PROPERTY_KEY);
    }

    private String getConfigValue(String defaultValue, String configKey) {
        String configValue = configService.getConfiguration().get(configPrefix + configKey);
        if (configValue == null) {
            configValue = defaultValue;
        }
        return configValue;
    }

    public Map<String, byte[]> getFiles(Map<String, String> params,
                                        String username,
                                        String password,
                                        String controllerUrl) {
        Map<String, byte[]> files = new HashMap<String, byte[]>();
        addAgentArtifacts(files);
        addAgentPasswordFile(files, password);
        addAgentPropertiesFile(files, params, username, controllerUrl);
        return files;
    }

    public void addAgentPasswordFile(Map<String, byte[]> destination, String password) {
        String location = getPasswordFileLocation();
        destination.put(location, password.getBytes());
    }

    public void addAgentPropertiesFile(Map<String, byte[]> destination, Map<String, String> properties,
                                       String username,
                                       String controllerUrl) {
        try {
            Properties props = new Properties();
            props.putAll(properties);
            props.put(agentUserProperty, username);
            props.put(controllerUrlProperty, controllerUrl);
            ByteArrayOutputStream out = new ByteArrayOutputStream();
            props.store(out, "");
            destination.put(getPropertiesFileLocation(), out.toByteArray());
        } catch (IOException ioe) {
            LOG.warning("Error generating properties file: " + ioe.toString());
        }
    }

    public void addAgentArtifacts(Map<String, byte[]> destination) {
        File[] agentFiles = getAgentBundles();
        for (File agentFile : agentFiles) {
            byte[] data = new byte[(int)agentFile.length()];
            try {
                FileInputStream is = new FileInputStream(agentFile);
                is.read(data);
                destination.put(getAgentDeployDir() + agentFile.getName(), data);
            } catch (IOException e) {
                LOG.warning(e.toString());
            }
        }

    }

    protected File[] getAgentBundles() {
        String[] locations = getAgentBundleLocation().split(",");
        File[] files = new File[locations.length];
        for (int i = 0; i < locations.length; i++) {
            String loc = locations[i].trim();
            LOG.info("adding artifact " + loc);
            files[i] = new File(loc);
        }
        return files;
    }

    protected String getControllerUrlProperty() {
        return controllerUrlProperty;
    }

    protected void setControllerUrlProperty(String controllerUrlProperty) {
        this.controllerUrlProperty = controllerUrlProperty;
    }

    protected String getPasswordFileLocation() {
        return passwordFileLocation;
    }

    protected void setPasswordFileLocation(String passwordFileLocation) {
        this.passwordFileLocation = passwordFileLocation;
    }

    protected String getPropertiesFileLocation() {
        return propertiesFileLocation;
    }

    protected void setPropertiesFileLocation(String propertiesFileLocation) {
        this.propertiesFileLocation = propertiesFileLocation;
    }

    protected String getAgentDeployDir() {
        return agentDeployDir;
    }

    protected void setAgentDeployDir(String agentDeployDir) {
        this.agentDeployDir = agentDeployDir;
    }

    protected String getAgentBundleLocation() {
        return agentBundleLocation;
    }

    protected void setAgentBundleLocation(String agentBundleLocation) {
        this.agentBundleLocation = agentBundleLocation;
    }

}
