/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.web.repository.agentprovider.servlet;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import com.iona.soa.repository.configuration.ConfigurationService;
import com.iona.soa.repository.configuration.HttpConfiguration;
import com.iona.soa.repository.util.Base64;
import com.iona.soa.security.AuthenticationException;
import com.iona.soa.security.SecurityContext;
import com.iona.soa.security.SecurityService;
import com.iona.soa.security.SecurityUtil;
import com.iona.soa.web.repository.agentprovider.providers.IAgentFileProvider;
import com.iona.soa.web.repository.agentprovider.providers.IAgentFileProviderFactory;

public class AgentProviderServlet extends HttpServlet {
    private static final long serialVersionUID = 1L;

    private static final Logger LOG = Logger.getLogger(AgentProviderServlet.class.getName());
    private static final String BUNDLE_PREFIX = "fuse-depot";

    private static final String AGENT_TYPE_PARAM = "agentType";
    private static final Collection<String> AGENT_AREAS = Collections
        .singletonList(SecurityUtil.AGENT_SECAREA_ID);
    private IAgentFileProviderFactory agentFileProviderFactory;
    private SecurityService securityService;
    private HttpConfiguration httpConfiguration;
    private String version;

    public AgentProviderServlet(IAgentFileProviderFactory agentFileProviderFactory,
                                SecurityService securityService, 
                                ConfigurationService configService) {
        this.agentFileProviderFactory = agentFileProviderFactory;
        this.securityService = securityService;
        this.httpConfiguration = configService.getHttpConfiguration();
        version = configService.getConfiguration().get("depot.build.version");
    }

    @Override
    protected void doGet(HttpServletRequest req, HttpServletResponse resp) throws ServletException,
        IOException {
        Map<String, String> authDetails = authenticate(req, resp);
        if ((authDetails != null) && (authDetails.size() == 2)) {
            String username = authDetails.get("username");
            String password = authDetails.get("password");
            String agentType = req.getParameter(AGENT_TYPE_PARAM);
            Map<String, String> params = new HashMap<String, String>();
            Enumeration<?> paramNames = req.getParameterNames();
            while (paramNames.hasMoreElements()) {
                String paramName = (String)paramNames.nextElement();
                String param = req.getParameter(paramName);
                if (!paramName.equals(AGENT_TYPE_PARAM)) {
                    params.put(paramName, param);
                }
            }

            String filename = BUNDLE_PREFIX + "-" + agentType.toLowerCase() + "-agent-" + version + ".zip";
            File file = getAgentBundle(agentType, params, username, password, getControllerUrl(req));
            if (file != null) {
                InputStream is = new FileInputStream(file);
                ServletOutputStream os = resp.getOutputStream();
                resp.setContentType("application/octet-stream");
                resp.setContentLength((int)file.length());
                resp.setHeader("Content-Disposition", "attachment; filename=\"" + filename + "\"");
                byte[] data = new byte[(int)file.length()];
                is.read(data);
                os.write(data);

                os.close();
                is.close();
                if (!file.delete()) {
                    System.out.println("failed to delete file ");
                }
            } else {
                resp.setStatus(501);
            }
        }
    }

    private String getControllerUrl(HttpServletRequest req) {
        String url = req.getRequestURL().toString();
        url = url.replaceFirst("agents", "controller");
        if (!httpConfiguration.isHttpEnabled()) {
            return url;
        }
        // Force the use of insecure port.
        url = url.replaceFirst("https", "http");
        url = url.replace(Integer.toString(httpConfiguration.getSecureHttpPort()), 
                                  Integer.toString(httpConfiguration.getHttpPort()));
        return url;
    }

    private Map<String, String> authenticate(HttpServletRequest req, HttpServletResponse resp) {
        String authHeader = req.getHeader(SecurityService.AUTHENTICATE_REQUEST_HEADER);
        if (authHeader != null) {
            try {
                SecurityContext secCon = securityService.authenticate(authHeader);
                if (securityService.authorizeRead(secCon.getSSOToken(), null, AGENT_AREAS)) {

                    String token = authHeader.replace("Basic ", "");
                    String creds = new String(Base64.decode(token));
                    String[] usernamePass = creds.split(":");
                    if (usernamePass.length > 1) {
                        Map<String, String> tokens = new HashMap<String, String>();
                        tokens.put("username", usernamePass[0]);
                        tokens.put("password", usernamePass[1]);
                        return tokens;
                    }
                }
            } catch (AuthenticationException e) {
                LOG.warning("Error during client auth: " + e.toString());
            }
        }
        resp.setHeader(SecurityService.AUTHENTICATE_RESPONSE_HEADER, SecurityService.BASIC_AUTH_TYPE);
        resp.setStatus(401);
        return null;
    }

    private File getAgentBundle(String agentType, Map<String, String> params, String username,
                                String password, String controllerUrl) {
        IAgentFileProvider provider = agentFileProviderFactory.getAgentFileProvider(agentType);
        if (provider != null) {
            return zipFiles(provider.getFiles(params, username, password, controllerUrl));
        } else {
            return null;
        }

    }

    private File zipFiles(Map<String, byte[]> files) {
        File zipFile = null;

        if (files.size() < 1) {
            return null;
        }

        ZipOutputStream zos;
        try {
            zipFile = File.createTempFile("agentZip", ".zip");
            zos = new ZipOutputStream(new FileOutputStream(zipFile));
            for (String location : files.keySet()) {
                byte[] data = files.get(location);
                zos.putNextEntry(new ZipEntry(location));
                zos.write(data);
                zos.closeEntry();
            }
            zos.close();
        } catch (IOException ioe) {
            LOG.warning("ERROR Creating zipfile: " + ioe.toString());
        }
        return zipFile;
    }
}
