/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.expression;

import java.util.Arrays;

import com.iona.soa.repository.EMFUtil;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

public class Expression implements Term {
    private final Operator operator;
    private final Term [] operands;
    
    public Expression(Operator operator, Object ... operands) {
        this.operator = operator;                
        this.operands = new Term[operands.length];
        for (int i = 0; i < operands.length; i++) {
            if (operands[i] instanceof Term) {
                this.operands[i] = (Term) operands[i];
            } else if (operands[i] instanceof EStructuralFeature) {
                this.operands[i] = new Operand(((EStructuralFeature)operands[i]).getName());
            } else if (operands[i] instanceof EObject) {
                this.operands[i] = new Operand(EMFUtil.getGuid((EObject)operands[i]));
            } else {
                this.operands[i] = new Operand(operands[i]);
            }
        }
        
        if (operands.length < operator.minOperands
            || operands.length > operator.maxOperands) {
            throw new IllegalArgumentException(toString() 
                + " - Incorrect number of arguments passed to operation " + operator + ". Between " 
                + operator.minOperands + " and " + operator.maxOperands + " expected.");  
        }        
    }
    
    public Operator getOperator() {
        return operator;
    }
    
    public Term [] getOperands() {
        return operands;
    }
    
    public void execute(OperationDelegate delegate) {
        delegate.handle(this);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;            
        }
        
        if (!(obj instanceof Expression)) {
            return false;
        }
        
        Expression e = (Expression) obj;
        if (operands.length != e.operands.length) {
            return false;
        }
        for (int i = 0; i < operands.length; i++) {
            if (!(operands[i].equals(e.operands[i]))) {
                return false;
            }
        }
        
        return operator == e.operator; 
    }

    @Override
    public int hashCode() {
        int hc = operator.hashCode();
        for (Term operand : operands) {
            hc += 17 * operand.hashCode();
        }
        return hc;
    }

    @Override
    public final String toString() {
        String result = operator.toString();
        
        StringBuilder sb = new StringBuilder(Arrays.asList(operands).toString());
        sb.setCharAt(0, '(');
        sb.setCharAt(sb.length() - 1, ')');
        return result + sb.toString();
    }
}
