/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.expression.delegate;

import java.beans.BeanInfo;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

import com.iona.soa.expression.Expression;
import com.iona.soa.expression.Operand;
import com.iona.soa.expression.Operation;
import com.iona.soa.expression.OperationDelegate;
import com.iona.soa.expression.Term;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

public class BeanExpressionHandler implements OperationDelegate {
    private final Collection<EObject> objects;
    private final Set<EObject> qualifying;
    
    public BeanExpressionHandler(Collection<EObject> objects) { 
        this.objects = objects;
        qualifying = new HashSet<EObject>();
    }
    
    public boolean handle(Operation op) {
        throw new UnsupportedOperationException();
    }    

    public boolean handle(Expression op, Map context) {
        throw new UnsupportedOperationException();
    }    

    public boolean handle(Operand op, Map context) {
        throw new UnsupportedOperationException();
    }    

    public boolean handle(Expression expr) {
        for (EObject obj : objects) {
            if (perform(obj, expr).size() > 0) {
                qualifying.add(obj);
            }
        }
        return true;
    }    

    public boolean handle(Operand op) { 
        // This handlers needs to do nothing here
        return true;
    }
    
    public Set<EObject> getQualifying() {
        return qualifying;
    }

    public List<EObject> perform(EObject obj, Term term) {
        if (term instanceof Expression) {
            return perform(obj, (Expression) term);
        } else if (term instanceof Operand) {
            Object val = ((Operand) term).getValue();
            if (val instanceof String
                && Boolean.parseBoolean((String) val)) {
                return Arrays.asList(new EObject[] {obj});
            }
        }
        return null;
    }
    
    public List<EObject> perform(EObject obj, Expression expr) {
        switch (expr.getOperator()) {
        case AND:
            return performAnd(obj, expr.getOperands());
        case EQUALS:
            return performEquals(obj, expr.getOperands());
        default:
            break;
        }
        return null;
    }
    
    private Object getValue(EObject obj, Term term) {
        if (term instanceof Operand) {
            return ((Operand) term).getValue();
        } else if (term instanceof Expression) {
            return perform(obj, (Expression) term);
        } else {
            return null;
        }
    }

    private List<EObject> performEquals(EObject obj, Term ... operands) {
        String address = (String) getValue(obj, operands[0]);
        Object value = getValue(obj, operands[1]);
        
        List<String> addressPath = new LinkedList<String>();
        StringTokenizer tok = new StringTokenizer(address, ".");
        while (tok.hasMoreTokens()) {
            addressPath.add(tok.nextToken());
        }
        
        return performEquals(obj, addressPath, value);
    }
    
    @SuppressWarnings("unchecked")
    private List<EObject> performEquals(EObject obj, List<String> address, Object value) {
        // Clone the list to allow backtracking
        address = new LinkedList<String>(address);
        
        String current = address.remove(0);
        if (address.size() == 0) {
            return performEquals(obj, current, value);
        }

        EStructuralFeature feat = obj.eClass().getEStructuralFeature(current);
        if (feat instanceof EReference) {
            if (feat.isMany()) {
                List<EObject> refList = (List<EObject>) obj.eGet(feat);
                for (EObject refObj : refList) {
                    List<EObject> found = performEquals(refObj, address, value);
                    if (found.size() > 0) {
                        return found;
                    }
                }
                return Collections.emptyList();
            } else {
                EObject refObj = (EObject) obj.eGet(feat);
                if (refObj != null) {
                    return performEquals(refObj, address, value);
                } else {
                    return Collections.emptyList();
                }
            }
        }        
        return Collections.emptyList();
    }

    private List<EObject> performEquals(EObject obj, String key, Object value) {
        try {
            BeanInfo bi = Introspector.getBeanInfo(obj.getClass());
            PropertyDescriptor pd = findPropertyDescriptor(bi, key);                
            if (pd != null) {
                Object actualValue = pd.getReadMethod().invoke(obj, new Object [] {});
                
                String s1 = "" + value;
                String s2 = "" + actualValue;
                
                if (s1.equals(s2)) {
                    return Arrays.asList(new EObject[] {obj});
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return Collections.emptyList();
    }

    private PropertyDescriptor findPropertyDescriptor(BeanInfo bi, Object key) {
        for (PropertyDescriptor pd : bi.getPropertyDescriptors()) {
            if (pd.getName().equals(key)) {
                return pd;
            }
        }
        return null;
    }

    private List<EObject> performAnd(EObject obj, Term[] operands) {
        List<List<EObject>> results = new ArrayList<List<EObject>>(operands.length);
        for (Term term : operands) {
            List<EObject> ctx = perform(obj, term);
            if (ctx.size() == 0) {
                return Collections.emptyList();
            }
            results.add(ctx);
        }
        
        Set<EObject> total = new HashSet<EObject>();
        for (List<EObject> sub : results) {
            total.addAll(sub);
        }
        return new ArrayList<EObject>(total);
        
        /*
        // Found a number of objects, make sure that the ones with the same type have the same ID
        List<EObject> firstResults = new ArrayList<EObject>(results.remove(0));
        for (List<EObject> result : results) {
            for (Iterator it = firstResults.iterator(); it.hasNext();) {
                if (!result.contains(it.next())) {
                    it.remove();
                }
            }
        }        
        return firstResults;
        */
    }
}
