/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.expression.serialization;

import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

import com.iona.soa.expression.And;
import com.iona.soa.expression.Equals;
import com.iona.soa.expression.IsNull;
import com.iona.soa.expression.Operand;
import com.iona.soa.expression.Operation;
import com.iona.soa.expression.OperationType;
import com.iona.soa.expression.Term;

public final class DeserializationHandler {
    private DeserializationHandler() { }
    
    public static Term deserialize(String s) {
        if (s == null || s.length() == 0) {
            return null;
        }
        
        return deserializeAnds(s);
    }

    private static Term deserializeAnds(String s) {
        StringTokenizer tok = new StringTokenizer(s, "&");
        
        List<Term> subExpressions = new LinkedList<Term>();
        while (tok.hasMoreTokens()) {            
            String token = tok.nextToken();
            Operation operation = getOperation(token);
            if (operation != null) {
                return operation;
            }
            Term subExpression = deserializeEqualsOrIsNull(token);
            if (subExpression == null) {
                try {
                    subExpression = new Operand(URLDecoder.decode(token, "UTF-8"));
                } catch (UnsupportedEncodingException e) {
                    subExpression = new Operand(token);
                }
            }
            subExpressions.add(subExpression);
        }           
        
        if (subExpressions.size() == 1) {
            return subExpressions.get(0);
        } else {
            return new And(subExpressions.toArray());
        }
    }



    private static Operation getOperation(String token) {
        if (!token.startsWith("_")) {
            return null;
        }
        Equals subExpression = deserializeEquals(token);
        if (subExpression == null) {
            return null;
        }
        Term[] parts = subExpression.getOperands();
        if (parts.length != 2) {
            return null;
        }
        Operand opName = (Operand)parts[0];
        OperationType opType = null;
        for (OperationType ot : OperationType.values()) {
            if (ot.name().equals(opName.getValue().toString())) {
                opType = OperationType.valueOf(opName.getValue().toString());
                break;
            }
        }
        if (opType == null) {
            return null;
        }
        String opArgStr = ((Operand)parts[1]).getValue().toString();
        Object[] opArgs = opArgStr.split(";");
        return new Operation(opType, opArgs);        
    }

    private static Term deserializeEqualsOrIsNull(String s) {
        
        int i = s.indexOf('=');
        int j = s.indexOf(':');
        
        if (i == -1 && j == -1) {
            return null;
        }
        
        if (i == -1 && j != -1) {
            return deserializeIsNull(s);
        }
        
        if (i != -1 && j == -1) {
            return deserializeEquals(s); 
        }
        
        if (i < j) {
            return deserializeEquals(s);             
        } else {
            return deserializeIsNull(s);
        }        
    }
    
    private static Equals deserializeEquals(String s) {
                
        if (s.indexOf('=') < 0) {
            return null;
        }        
        StringTokenizer tok = new StringTokenizer(s, "=", true);
        
        List<Operand> ops = new LinkedList<Operand>();
        boolean lastOneWasEquals = false;
        while (tok.hasMoreTokens()) {
            String token = tok.nextToken();
            lastOneWasEquals = "=".equals(token);
            if (lastOneWasEquals) {
                if (ops.size() == 0) {
                    // Equals without a token
                    ops.add(new Operand(""));
                }
                continue;
            }
                
            
            Operand operand;
            try {
                operand = new Operand(URLDecoder.decode(token, "UTF-8"));
            } catch (UnsupportedEncodingException e) {
                operand = new Operand(token);
            }
            ops.add(operand);
        }        
        
        if (lastOneWasEquals) {
            // if last one token was the equals, add an empty operand to it
            ops.add(new Operand(""));
        }
        
        return new Equals(ops.toArray());
    }

    private static IsNull deserializeIsNull(String s) {
        String a[] = s.split(":");
        if (a.length != 2) {
            return null;
        }
        
        if (!"isnull".equals(a[0])) {
            return null;
        }
        
        return new IsNull(a[1]);
    }
}
