/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.expression.serialization;

import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import com.iona.soa.expression.And;
import com.iona.soa.expression.DanglingReferences;
import com.iona.soa.expression.Env;
import com.iona.soa.expression.Equals;
import com.iona.soa.expression.Expression;
import com.iona.soa.expression.If;
import com.iona.soa.expression.In;
import com.iona.soa.expression.InstanceOf;
import com.iona.soa.expression.Not;
import com.iona.soa.expression.Operand;
import com.iona.soa.expression.Operator;
import com.iona.soa.expression.Or;
import com.iona.soa.expression.References;
import com.iona.soa.expression.Select;
import com.iona.soa.expression.Term;
import com.iona.soa.repository.util.ConstructorResolver;
import com.iona.soa.repository.util.Out;

public final class ModelQueryDeserializationHandler {
    private static Map<Operator, Class<? extends Expression>> registry =  
        new HashMap<Operator, Class<? extends Expression>>();
    
    static {
        registry.put(Operator.ENV, Env.class);
        registry.put(Operator.IF, If.class);
        registry.put(Operator.AND, And.class);
        registry.put(Operator.EQUALS, Equals.class);
        registry.put(Operator.OR, Or.class);
        registry.put(Operator.NOT, Not.class);
        registry.put(Operator.IN, In.class);
        registry.put(Operator.INSTANCEOF, InstanceOf.class);
        registry.put(Operator.SELECT, Select.class);
        registry.put(Operator.REFERENCES, References.class);
        registry.put(Operator.DANGLING_REFERENCES, DanglingReferences.class);
    }
    
    private ModelQueryDeserializationHandler() {
        // complete
    }
    
    public static Collection<Expression> deserialize(String value) {
        if (value == null || value.length() == 0) {
            return null;
        }
        return parseExpressions(value);
    }
    
    protected static Collection<Expression> parseExpressions(String s) {
        Collection<Expression> exps = new ArrayList<Expression>();
        Out<String> value = new Out<String>(s);
        for (;;) {
            value.set(value.get().trim());
            Expression e = parseExpression(value);
            if (e == null) {
                throw new ParseException(value.get(), 
                                         "Unable to resolve corresponding"
                                         + " function implementation for " 
                                         + value.get());
            }
            exps.add(e);
            if (value.get().indexOf(",") == -1 
                || value.get().length() == value.get().indexOf(",") + 1) {
                return exps;
            }
            value.set(value.get().substring(value.get().indexOf(",") + 1));
        }
    }

    @SuppressWarnings("unchecked")
    protected static Expression parseExpression(Out<String> value) {
        Expression exp = null;
        for (Operator op : Operator.values()) {
            if (isExpression(value.get(), op)) {
                Class c = registry.get(op);
                try {
                    Constructor ctor = ConstructorResolver.resolve(c, new Class[]{Object[].class});
                    if (ctor == null) {
                        throw new ParseException(value.get(), 
                                                 "Unable to resolve corresponding"
                                                 + " function implementation for " 
                                                 + value.get().substring(0, 
                                                                         value.get().indexOf("(")).trim());
                    }
                    if (value.get().indexOf("(") == -1 
                        || value.get().length() == value.get().indexOf("(") + 1) {
                        throw new ParseException("(", 
                                                 "Could not find expected token '('"
                                                 + " while parsing '" 
                                                 + value.get() + "'");

                    }
                    value.set(value.get().substring(value.get().indexOf("(") + 1).trim());
                    Term[] args = parseArguments(value);
                    if (value.get().indexOf(")") == -1) {
                        throw new ParseException(")", 
                                                 "Could not find expected token ')'"
                                                 + " while parsing '" 
                                                 + value.get() + "'");
                    }                    
                    exp = (Expression)ctor.newInstance(new Object[]{args});
                    if (value.get().length() == value.get().indexOf(")") + 1) {
                        break;
                    }
                    value.set(value.get().substring(value.get().indexOf(")") + 1));
                } catch (Throwable t) {
                    throw new ParseException(value.get(), 
                                             "Exception while parsing '"
                                             + value.get() + "'", t);
                }
            }
        }
        return exp;
    }

    protected static boolean isExpression(String value, Operator op) {
        if (value.startsWith(op.getLiteral())) {
            if (value.length() == op.getLiteral().length()) {
                return true;
            }
            value = value.substring(op.getLiteral().length());
            if (value.startsWith(" ") || value.startsWith("(")) {
                return true;
            }
        }
        return false;
    }
    
    protected static Term[] parseArguments(Out<String> value) {
        Collection<Term> terms = new ArrayList<Term>();
        if (value.get().length() == 0  || value.get().charAt(0) == ',') {
            return null;
        }
        for (;;) {
            Term term = parseArgument(value);
            if (term == null) {
                break;
            }
            terms.add(term);
            if (value.get().length() == 0 || value.get().charAt(0) != ',') {
                break;
            }
            value.set(value.get().substring(1).trim());
        }
        return terms.toArray(new Term[]{});
    }
    
    protected static Term parseArgument(Out<String> value) {
        Term term = parseExpression(value);
        if (term != null) {
            return term;
        }
        StringBuffer operand = new StringBuffer();
        int i = 0;
        while (i < value.get().length()) {
            char c = value.get().charAt(i++);
            if (c == ',' || c == ')') {
                --i;
                break;
            }
            if (c != ' ') {
                operand.append(c);
            }
            if (c == '\'') {
                while (i < value.get().length()) {
                    c = value.get().charAt(i++);
                    if (c == ',' || c == ')') {
                        throw new ParseException(value.get(), 
                                                 "Exception while parsing '"
                                                 + value.get() + "'"
                                                 + " expected \"'\"" 
                                                 + "character but found \"" 
                                                 + c + "\"");
                    }
                    operand.append(c);
                    if (c == '\'') {
                        break;
                    }
                }
            }
        }
        if (operand.length() == 0) {
            return null;
        }
        value.set(value.get().substring(i));
        return new Operand(operand.toString());
    }
}
