/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.emf.common.notify.Adapter;

public abstract class AbstractSOARepositoryImpl implements SOARepository {
    private final List<Adapter> emfAdapters = new LinkedList<Adapter>();
    private final List<TransactionAdapter> txAdapters = new LinkedList<TransactionAdapter>();
    
    public AbstractSOARepositoryImpl(Adapter ... adapters) {
        for (Adapter adapter : adapters) {
            addEMFAdapter(adapter);
            if (adapter instanceof TransactionAdapter) {
                registerTransactionAdapter((TransactionAdapter)adapter);
            }
        }
    }

    public void startRead(TransactionData txData) {
    }
    
    public void endRead(boolean success) {
    }
    
    public void startWrite(TransactionData txData) {
        // highest prio TA is last in txAdapters list
        for (int i = txAdapters.size() - 1; i >= 0; i--) {
            try {
                txAdapters.get(i).startTransaction(this, txData);
            } catch (Exception ex) {
                for (int j = i + 1; j < txAdapters.size(); j++) {
                    try {
                        txAdapters.get(j).rollbackTransaction();
                    } catch (Exception ex2) {
                        // ignore
                    }
                }
                if (ex instanceof RuntimeException) {
                    throw (RuntimeException)ex;
                }
                throw new RuntimeException(ex);
            }
        }
    }
    
    public void endWrite(boolean success) {
        try {
            Exception caughtEx = null;
            for (TransactionAdapter ta : txAdapters) {
                try {
                    if (success) {
                        ta.commitTransaction();
                    } else {
                        ta.rollbackTransaction();
                    }
                } catch (Exception ex) {
                    success = false;
                    if (caughtEx == null) {
                        caughtEx = ex;
                    }
                }
            }
            if (caughtEx != null) {
                if (caughtEx instanceof RuntimeException) {
                    throw (RuntimeException)caughtEx;
                }
                throw new RuntimeException(caughtEx);
            }
        } finally {
            if (!success) {
                close();
            }
        }
    }
    
    public void registerTransactionAdapter(TransactionAdapter ta) {
        if (ta instanceof Adapter && addEMFAdapter((Adapter)ta)) {
            close();
        }
        for (int i = 0; i < txAdapters.size(); i++) {
            TransactionAdapter current = txAdapters.get(i);
            if (current.getPriority() > ta.getPriority()) {
                // TA's are stored in ascending priority
                txAdapters.add(i, ta);
                return;
            }
        }
        txAdapters.add(ta);
    }
    
    private boolean addEMFAdapter(Adapter adapter) {
        if (emfAdapters.contains(adapter)) {
            return false;
        }
        int prio = adapter instanceof TransactionAdapter 
                   ? ((TransactionAdapter)adapter).getPriority() 
                   : Integer.MIN_VALUE;
        for (int i = 0; i < emfAdapters.size(); i++) {
            Adapter current = emfAdapters.get(i);
            int currentPrio = current instanceof TransactionAdapter 
                              ? ((TransactionAdapter)current).getPriority() 
                              : Integer.MIN_VALUE;
            if (currentPrio < prio) {
                // TA's are stored in ascending priority
                emfAdapters.add(i, adapter);
                return true;
            }
        }
        emfAdapters.add(adapter);
        return true;
    }

    public Collection<Adapter> getResourceAdapters() {
        return Collections.unmodifiableCollection(emfAdapters);
    }
}
