/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAnnotation;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EModelElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.InternalEObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.emf.ecore.util.EcoreUtil.Copier;
import org.eclipse.emf.ecore.util.InternalEList;

/**
 * This class contains some general EMF utility functions. Generally they are used by other 
 * classes in this package.
 */
public final class EMFUtil {

    public static final String DELETE_ANNOTATION = "delete";
    public static final String DELETE_PREDICATE = "predicate";
    public static final String DELETE_DEPENDENCIES = "dependencies";

    private static String genNumberAttrName = "generationNumber";
    private static Map<EClass, Map<EClass, Collection<EReference>>> backReferenceCache 
        = new HashMap<EClass, Map<EClass, Collection<EReference>>>();

    private EMFUtil() {
        // emtpy
    }

    public static EObject createEObject(EClass eClass, Object id) {
        EObject ret = eClass.getEPackage().getEFactoryInstance().create(eClass);
        if (id != null) {
            ret.eSet(eClass.getEIDAttribute(), id);
        }
        return ret;
    }

    public static URI getProxyURI(EObject obj) {
        EClass eclass = obj.eClass();
        return getProxyURI(eclass, getGuid(obj));
    }

    public static URI getProxyURI(EClass actualClass, String id) {
        return URI.createURI(actualClass.getEPackage().getNsURI() + "#"
                + getProxyURIFragment(actualClass, id));
    }

    public static URI getProxyVersionedURI(EObject obj) {
        return URI.createURI(getProxyURI(obj) + "/" + getGenerationNumber(obj).toString());
    }

    // query parameter is suggestive, if object is a proxy and has its own
    // URI with a query fragment, then this fragment will be used instead
    public static URI getProxyVersionedURI(EObject obj, String query) {
        if (obj.eIsProxy()) {
            URI uri = ((InternalEObject) obj).eProxyURI();
            if (null != uri && uri.hasFragment()) {
                String fragment = uri.fragment();
                int index = fragment.indexOf("?");
                if (index > 0) {
                    query = fragment.substring(index + 1, fragment.length());
                }
            }
        }
        return URI.createURI(getProxyURI(obj, query) + "/" + getGenerationNumber(obj).toString());
    }

    public static URI getProxyURILessVersion(String versionedURI) {
        if (-1 != versionedURI.lastIndexOf('/') && !versionedURI.endsWith("/")) {
            return URI.createURI(versionedURI.substring(0, versionedURI.lastIndexOf('/')));
        }
        return URI.createURI(versionedURI);
    }

    public static URI getProxyURI(EObject obj, String query) {
        return (null != query) ? getProxyURI(obj.eClass(), query, getGuid(obj)) : getProxyURI(obj.eClass(),
                                                                                              getGuid(obj));
    }

    public static URI getProxyURI(EClass eclass, String query, String id) {
        URI uri = URI.createURI(eclass.getEPackage().getNsURI() + '?'
                + query
                + "#"
                + getProxyURIFragment(eclass, id));
        return uri;
    }

    public static long getObjectVersion(String versionedURI) {
        if (-1 != versionedURI.lastIndexOf('/') && !versionedURI.endsWith("/")) {
            return Long.valueOf(versionedURI.substring(versionedURI.lastIndexOf('/') + 1,
                                                       versionedURI.length())).longValue();
        }
        return 0;
    }

    public static EAttribute getGenerationNumberAttribute(EClass eClass) {
        EStructuralFeature verAttr = eClass.getEStructuralFeature(genNumberAttrName);
        if (verAttr instanceof EAttribute) {
            return (EAttribute) verAttr;
        }
        return null;
    }

    public static Long getGenerationNumber(EObject eObject) {
        return getGenerationNumber(eObject, 0L);
    }

    public static Long getGenerationNumber(EObject eObject, Long defaultValue) {
        EAttribute verAttr = getGenerationNumberAttribute(eObject.eClass());
        if (verAttr != null) {
            // No need to cause proxy resolution as we include generationNumber
            // in proxy objects
            Object version = eObject.eGet(verAttr, false);
            if (version instanceof Long) {
                return (Long) version;
            }
        }
        return defaultValue;
    }

    public static String getProxyURIFragment(EObject obj) {
        return getProxyURIFragment(obj.eClass(), getGuid(obj));
    }

    private static String getProxyURIFragment(EClass actualClass, String id) {
        return "/" + actualClass.getName() + "/" + id;
    }

    public static List<EClass> getAllSubClasses(EClass eclass, ModelRegistry registry) {
        List<EClass> subClasses = getSubClasses(eclass, registry);
        List<EClass> subSubClasses = new LinkedList<EClass>();
        for (EClass subClass : subClasses) {
            subSubClasses.addAll(getAllSubClasses(subClass, registry));
        }
        subClasses.addAll(subSubClasses);
        return subClasses;
    }

    public static List<EClass> getSubClasses(EClass eclass, ModelRegistry registry) {
        List<EClass> subClasses = new LinkedList<EClass>();
        getSubClasses(eclass.getEPackage(), eclass, subClasses);
        if (registry != null) {
            for (EPackage pkg : registry.getPackages()) {
                if (eclass.getEPackage() != pkg) {
                    getSubClasses(pkg, eclass, subClasses);
                }
            }
        }
        return subClasses;
    }

    @SuppressWarnings("unchecked")
    private static void getSubClasses(EPackage pkg, EClass eclass, List<EClass> subClasses) {
        for (EClassifier classifier : (List<EClassifier>) pkg.getEClassifiers()) {
            if (classifier instanceof EClass) {
                EClass subclass = (EClass) classifier;
                if (subclass.getESuperTypes().contains(eclass)) {
                    subClasses.add(subclass);
                }
            }
        }
    }

    public static String getGuid(EObject object) {
        if (object == null) {
            return null;
        } else if (object.eIsProxy()) {
            URI uri = ((InternalEObject) object).eProxyURI();
            if (uri.hasFragment()) {
                String fragment = uri.fragment();
                int index = Math.max(fragment.lastIndexOf('/'), fragment.lastIndexOf('#'));
                if (index >= 0) {
                    int end = fragment.indexOf("?");
                    if (end > 0 && end > index + 1) {
                        fragment = fragment.substring(index + 1, end);
                    } else {
                        fragment = fragment.substring(index + 1);
                    }
                }
                return fragment;
            } else {
                return uri.toString();
            }
        } else {
            return EcoreUtil.getID(object);
        }
    }

    /**
     * Convenience method to return the "name" attribute of an eObject if it exists
     * and is set. Otherwise returns the object's guid.
     * 
     * @param eObject The eObject
     * @return The name or guid of the eObject
     */
    public static String getName(EObject eObject) {
        EStructuralFeature feature = eObject.eClass().getEStructuralFeature("name");
        if (feature instanceof EAttribute && eObject.eIsSet(feature)) {
            return eObject.eGet(feature).toString();
        } else {
            return getGuid(eObject);
        }
    }

    /**
     * Convenience method to return the "name" and guid of an eObject.
     * If there is no name then we just return the gui.
     * 
     * @param eObject The eObject
     * @return The name plus guid of the eObject
     */
    public static String getLongName(EObject eObject) {
        EStructuralFeature feature = eObject.eClass().getEStructuralFeature("name");
        if (feature instanceof EAttribute && eObject.eIsSet(feature)) {
            return eObject.eGet(feature).toString() + " (" + getGuid(eObject) + ")";
        } else {
            return getGuid(eObject);
        }
    }

    /** A very thin wrapper around {@link org.eclipse.emf.ecore.EClass#getEIDAttribute()}.
     * This one throws a {@link RuntimeException} when the ID cannot be found.
     * @param eclass The class to find the ID attribute for.
     * @return The ID attribute.
     * @throws RuntimeException when the class does not have an ID attribute. 
     */
    @SuppressWarnings("unchecked")
    public static EAttribute findID(EClass eclass) {
        EAttribute attr = eclass.getEIDAttribute();
        if (attr != null) {
            return attr;
        }
        throw new RuntimeException("Could not find ID in class " + eclass);
    }

    @SuppressWarnings("unchecked")
    public static EObject copyObjectWithFreshGuids(EObject original) {
        
        CopierEscapingGuids copier = new CopierEscapingGuids();
        EObject copy = copier.copy(original);
        copier.copyReferences();
        
        EClass clazz = original.eClass();
        
        for (EStructuralFeature feature : (List<EStructuralFeature>) clazz.getEAllStructuralFeatures()) {
            if (feature instanceof EReference && original.eIsSet(feature)) {
                if (((EReference) feature).isContainment()) {
                    copyReferenceWithFreshGuids(original, copy, (EReference) feature, true);
                } else if (!copy.eIsSet(feature)) {
                    copyReferenceWithFreshGuids(original, copy, (EReference) feature, false);
                }
            }
        }
        return copy;
    }

    @SuppressWarnings("unchecked")
    private static void copyReferenceWithFreshGuids(EObject original,
                                                    EObject copy,
                                                    EReference reference,
                                                    boolean deepCopy) {
        if (reference.isMany()) {
            List<EObject> source = (List<EObject>) original.eGet(reference);
            List<EObject> target = (List<EObject>) copy.eGet(reference);
            target.clear();
            for (EObject refOrig : source) {
                EObject refCopy =
                    deepCopy ? copyObjectWithFreshGuids(refOrig) : EMFUtil.createEProxy(refOrig);
                target.add(refCopy);
            }
        } else {
            EObject refOrig = (EObject) original.eGet(reference);
            EObject refCopy = deepCopy ? copyObjectWithFreshGuids(refOrig) : EMFUtil.createEProxy(refOrig);
            copy.eSet(reference, refCopy);
        }
    }
    
    private static class CopierEscapingGuids extends Copier {
        
        @Override
        protected void copyAttribute(EAttribute attribute, EObject object, EObject copyEObject) {
            if (attribute.isID()) { //if (attribute.equals(object.eClass().getEIDAttribute())) {
                return;
            }
            super.copyAttribute(attribute, object, copyEObject);
        }
    }

    @SuppressWarnings("unchecked")
    public static EObject copyObject(EObject original) {
        EObject copy = EcoreUtil.copy(original);
        EClass clazz = original.eClass();
        for (EStructuralFeature feature : (List<EStructuralFeature>) clazz.getEAllStructuralFeatures()) {
            if (feature instanceof EReference && original.eIsSet(feature)) {
                if (((EReference) feature).isContainment()) {
                    copyReference(original, copy, (EReference) feature, true);
                } else if (!copy.eIsSet(feature)) {
                    copyReference(original, copy, (EReference) feature, false);
                }
            }
        }
        return copy;
    }

    @SuppressWarnings("unchecked")
    private static void copyReference(EObject original,
                                      EObject copy,
                                      EReference reference, 
                                      boolean deepCopy) {
        if (reference.isMany()) {
            List<EObject> source = (List<EObject>) original.eGet(reference);
            List<EObject> target = (List<EObject>) copy.eGet(reference);
            target.clear();
            for (EObject refOrig : source) {
                EObject refCopy = deepCopy ? copyObject(refOrig) : createEProxy(refOrig);
                target.add(refCopy);
            }
        } else {
            EObject refOrig = (EObject) original.eGet(reference);
            EObject refCopy = deepCopy ? copyObject(refOrig) : createEProxy(refOrig);
            copy.eSet(reference, refCopy);
        }
    }

    @SuppressWarnings("unchecked")
    public static EObject deepObjectCopy(EObject original) {
        Map<EObject, EObject> visited = new HashMap<EObject, EObject>();
        return copyObjectWithAggregates(original, visited);
    }

    @SuppressWarnings("unchecked")
    private static EObject copyObjectWithAggregates(EObject original, Map<EObject, EObject> visited) {
        EObject copy = visited.get(original);
        if (null != copy) {
            return copy;
        }
        copy = EcoreUtil.copy(original);
        visited.put(original, copy);
        EClass clazz = original.eClass();
        for (EStructuralFeature feature : (List<EStructuralFeature>) clazz.getEAllStructuralFeatures()) {
            if (feature instanceof EReference && original.eIsSet(feature)) {
                if (feature.isMany()) {
                    List<EObject> source = (List<EObject>) original.eGet(feature);
                    List<EObject> target = (List<EObject>) copy.eGet(feature);
                    target.clear();
                    for (EObject refOrig : source) {
                        EObject refCopy = copyObjectWithAggregates(refOrig, visited);
                        target.add(refCopy);
                    }
                } else {
                    EObject refOrig = (EObject) original.eGet(feature);
                    EObject refCopy = copyObjectWithAggregates(refOrig, visited);
                    copy.eSet(feature, refCopy);
                }
            }
        }
        return copy;
    }

    @SuppressWarnings("unchecked")
    public static EObject copyObjectWithParent(EObject theObject, URI resourceUri) {
        EObject copy = copySingleObject(theObject, null, resourceUri);
        EObject parent = theObject.eContainer();
        if (parent != null) {
            EObject parentCopy = copyObjectWithParent(parent, resourceUri);
            EReference ref = theObject.eContainmentFeature();
            if (ref.isMany()) {
                int index = ((List) parent.eGet(ref, false)).indexOf(theObject);
                if (index < 0) {
                    ((List) parentCopy.eGet(ref)).add(copy);
                } else {
                    ((List) parentCopy.eGet(ref)).set(index, copy);
                }

            } else {
                parentCopy.eSet(ref, copy);
            }
        }
        return copy;
    }

    public static EObject copyObjectWithParent(EObject theObject) {
        return copyObjectWithParent(theObject, null);
    }

    @SuppressWarnings("unchecked")
    public static EObject copySingleObject(EObject original,
                                           Collection<EStructuralFeature> excluded,
                                           URI resourceUri) {
        if (excluded == null) {
            excluded = Collections.emptySet();
        }

        EClass eClass = original.eClass();
        EObject copy = eClass.getEPackage().getEFactoryInstance().create(eClass);
        for (EStructuralFeature feature : (List<EStructuralFeature>) eClass.getEAllStructuralFeatures()) {
            if (excluded.contains(feature)) {
                continue;
            }

            Object origVal = original.eGet(feature, false);
            Object copyVal = origVal;
            if (original.eIsSet(feature) && feature instanceof EReference) {
                if (feature.isMany()) {
                    BasicEList children = (BasicEList) origVal;
                    List<EObject> childCopies = new LinkedList<EObject>();
                    for (int i = 0; i < children.size(); i++) {
                        EObject child = (EObject) children.basicGet(i);
                        childCopies.add(createEProxy(child, resourceUri));
                    }
                    copyVal = childCopies;
                } else {
                    copyVal = createEProxy((EObject) origVal, resourceUri);
                }
            }
            copy.eSet(feature, copyVal);
        }
        return copy;
    }

    public static EObject copySingleObject(EObject original) {
        return copySingleObject(original, null, null);
    }

    public static EObject copySingleObject(EObject original, Collection<EStructuralFeature> excluded) {
        return copySingleObject(original, excluded, null);
    }

    public static EObject createEProxy(EObject obj, URI resourceUri) {
        EClass eClass = obj.eClass();
        InternalEObject proxy = (InternalEObject) eClass.getEPackage().getEFactoryInstance().create(eClass);
        if (obj.eIsProxy()) {
            proxy.eSetProxyURI(((InternalEObject) obj).eProxyURI());
        } else {
            if (null != resourceUri && 0 == resourceUri.toString().indexOf(eClass.getEPackage().getNsURI())) {
                String fragment = EMFUtil.getProxyURIFragment(obj);
                proxy.eSetProxyURI(URI.createURI(resourceUri.toString() + '#' + fragment));
            } else {
                proxy.eSetProxyURI(getProxyURI(obj));
            }
        }
        proxy.eSet(eClass.getEIDAttribute(), getGuid(obj));
        return proxy;
    }

    public static EObject createEProxy(EObject obj) {
        return createEProxy(obj, null);
    }

    @SuppressWarnings("unchecked")
    public static List<EReference> findOneToManyReferences(EClass eclass) {
        List<EReference> refList = new LinkedList<EReference>();
        List<String> nameList = new LinkedList<String>();
        for (EReference ref : (List<EReference>) eclass.getEAllReferences()) {
            if (!ref.isMany() && !nameList.contains(ref.getName())) {
                refList.add(ref);
                nameList.add(ref.getName());
            }
        }
        return refList;
    }

    public static EClass getEClass(String nsURI, String entityName) {
        EPackage ePackage = EPackage.Registry.INSTANCE.getEPackage(nsURI);
        return (EClass) ePackage.getEClassifier(entityName);
    }

    /**
     * Adds a reference from one EObject to another EObject, if the reference isn't already set.
     * @param object The origin of the reference (the referencor)
     * @param ref The reference
     * @param relObj The target of the reference (the referencee)
     */
    @SuppressWarnings("unchecked")
    public static void addReferenceToMemoryObject(EObject object, EReference ref, EObject relObj) {
        if (ref.isMany()) {
            Object o = object.eGet(ref);

            if (o instanceof List) {
                List l = (List) o;
                // only add to list if object isn't in there yet.
                if (!l.contains(relObj)) {
                    // Shouldn't use plain List.add() here as it will cause the proxy to resolve, which
                    // can result in an endless loop of resolution and will result in big performance penalty
                    if (l instanceof InternalEList) {
                        // Suggested by Ed Merks on EMF newsgroup
                        ((InternalEList) l).addUnique(relObj);
                    } else {
                        l.add(relObj);
                    }
                }
            }
        } else if (object.eGet(ref) == null) {
            // only set reference if it isn't set already
            object.eSet(ref, relObj);
        }
    }

    public static boolean isBuiltInType(EAttribute attr) {
        if (attr.getEType() instanceof EEnum) {
            return true;
        }

        Class<?> instanceClass = attr.getEType().getInstanceClass();
        return isBuiltInType(instanceClass);
    }

    public static boolean isBuiltInType(Class<?> instanceClass) {
        boolean ret = false;
        if (instanceClass == null) {
            ret = false;
        } else if (String.class.equals(instanceClass)) {
            ret = true;
        } else if (boolean.class.equals(instanceClass)) {
            ret = true;
        } else if (Boolean.class.equals(instanceClass)) {
            ret = true;
        } else if (int.class.equals(instanceClass)) {
            ret = true;
        } else if (Integer.class.equals(instanceClass)) {
            ret = true;
        } else if (long.class.equals(instanceClass)) {
            ret = true;
        } else if (Long.class.equals(instanceClass)) {
            ret = true;
        } else if (byte.class.equals(instanceClass)) {
            ret = true;
        } else if (Byte.class.equals(instanceClass)) {
            ret = true;
        } else if (Date.class.isAssignableFrom(instanceClass)) {
            ret = true;
        } else if (short.class.equals(instanceClass)) {
            ret = true;
        } else if (Short.class.equals(instanceClass)) {
            ret = true;
        }
        return ret;
    }

    /** Copies attributes but does not copy the ID attribute 
     * @param source The source EObject 
     * @param dest The destination EObject 
     */
    @SuppressWarnings("unchecked")
    public static void copySimpleAttributes(EObject source, EObject dest) {
        if (source.eClass().getName() != dest.eClass().getName()) {
            return;
        }

        for (EAttribute attr : (List<EAttribute>) source.eClass().getEAllAttributes()) {
            if (attr.isID()) {
                // Don't copy the ID
                continue;
            }

            dest.eSet(attr, source.eGet(attr));
        }
    }

    public static String getAnnotationValue(EModelElement el, String annotationName, String key) {
        EAnnotation ann = el.getEAnnotation(annotationName);
        if (ann != null) {
            return ann.getDetails().get(key);
        }
        return null;
    }

    /**
     * get multi-line / multi word annotations split up into it's parts. 
     * All whitespace is treated as a separator
     *   
     * @param el
     * @param annotationName
     * @param key
     * @return
     */
    public static Collection<String> getAnnotationValues(EModelElement el, 
                                                         String annotationName, 
                                                         String key) {
        String annoVal = getAnnotationValue(el, annotationName, key);
        if (annoVal != null) {
            String[] parts = annoVal.split("[\b\t\n\f\r]");
            Collection<String> ret = new ArrayList<String>(parts.length);
            for (String part : parts) {
                if (part.length() > 0) {
                    ret.add(part);
                }
            }
            return ret;
        }
        return Collections.emptyList();
    }

    public static Map<EClass, Collection<EReference>> getBackReferences(EClass eClass, 
                                                                        ModelRegistry registry) {
        synchronized (backReferenceCache) {
            if (!backReferenceCache.containsKey(eClass)) {
                backReferenceCache.put(eClass, discoverBackReferences(eClass, registry));
            }
            return backReferenceCache.get(eClass);
        }
    }

    @SuppressWarnings("unchecked")
    private static Map<EClass, Collection<EReference>> discoverBackReferences(EClass eClass,
                                                                              ModelRegistry registry) {

        Map<EClass, Collection<EReference>> backRefs = new HashMap<EClass, Collection<EReference>>();
        if (registry == null) {
            registry = getDefaultModelRegistry();
        }
        for (EPackage ePackage : registry.getPackages()) {
            for (Object obj : (List<EObject>) ((EPackage) ePackage).eContents()) {
                if (obj instanceof EClass) {
                    for (EReference reference : (List<EReference>) ((EClass) obj).getEAllReferences()) {
                        if (reference.getEReferenceType().isSuperTypeOf(eClass)) {
                            if (!backRefs.containsKey(obj)) {
                                backRefs.put((EClass) obj, new ArrayList<EReference>(1));
                            }
                            backRefs.get(obj).add(reference);
                        }
                    }
                }
            }
        }
        return backRefs;
    }

    public static ModelRegistry getDefaultModelRegistry() {
        Collection<Object> pkgs = EPackage.Registry.INSTANCE.values();

        List<EPackage> packages = new ArrayList<EPackage>(pkgs.size());
        for (Object o : pkgs) {
            if (o instanceof EPackage) {
                packages.add((EPackage) o);
            }
        }

        return new ModelRegistry(packages);
    }

    public static class URIEClassEntry implements Map.Entry<URI, EClass> {
        private URI uri;

        private EClass eclass;

        public URIEClassEntry(URI uri, EClass eclass) {
            this.uri = uri;
            this.eclass = eclass;
        }

        public URI getKey() {
            return uri;
        }

        public EClass getValue() {
            return eclass;
        }

        public EClass setValue(EClass value) {
            throw new UnsupportedOperationException();
        }
    };

    public static URIEClassEntry getObjectIdentity(String href) {
        if (href.contains("#")) {
            String[] packageTypedId = href.split("#");
            if (packageTypedId.length != 2 && !packageTypedId[1].startsWith("/")) {
                throw new IllegalArgumentException("Malformed object URI = " + href);
            }
            String[] typedId = packageTypedId[1].substring(1).split("/");
            if (typedId.length != 2) {
                throw new IllegalArgumentException("Malformed object URI = " + href);
            }
            EClass eClass = EMFUtil.getEClass(packageTypedId[0], typedId[0]);
            if (eClass == null) {
                throw new IllegalArgumentException("Malformed object URI = " + href);
            }
            URI uri = EMFUtil.getProxyURI(eClass, typedId[1]);
            if (uri == null) {
                throw new IllegalArgumentException("URI '" + href + "' represents an unknown entity");
            }
            return new URIEClassEntry(uri, eClass);
        }
        String[] typedId = href.split("/");
        if (typedId.length != 2) {
            throw new IllegalArgumentException("Malformed object URI = " + href);
        }
        EClass eClass = ModelRegistry.getDefaultRegistry().getEClass(typedId[0]);
        if (eClass == null) {
            throw new IllegalArgumentException("Malformed object URI = " + href);
        }
        URI uri = EMFUtil.getProxyURI(eClass, typedId[1]);
        if (uri == null) {
            throw new IllegalArgumentException("URI '" + href + "' represents an unknown entity");
        }
        return new URIEClassEntry(uri, eClass);
    }
}
