/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.notification;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.notify.Notifier;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;

public abstract class AbstractRepositoryAdapter extends AbstractNotificationAdapter {
    private List<EObject> movedContainments = new LinkedList<EObject>();

    public AbstractRepositoryAdapter() {
    }
    
    public AbstractRepositoryAdapter(Notifier notifier) {
        setTarget(notifier);
    }
    
    public AbstractRepositoryAdapter(List<Notifier> notifiers) {
        for (Notifier notifier : notifiers) {
            setTarget(notifier);
        }
    }
    
    public abstract void changedAttribute(EObject eObject, EAttribute attribute, 
                                          Object oldValue, Object newValue);
    
    public abstract void changedReference(EObject eObject, EReference ref, 
                                          Object oldValue, Object newValue);
    
    public abstract void attachedObject(EObject eObject);
    
    public abstract void detachedObject(EObject eObject, EObject oldContainer, EReference containerRef);
    
    @Override
    public void notifyChangedObject(int event, EObject eObject, EStructuralFeature feature, 
                                       Object oldValue, Object newValue) {
        if (feature.isMany()) {
            switch (event) {
            case Notification.ADD:
                newValue = Collections.singletonList(newValue);
             // fallthrough
            case Notification.ADD_MANY: 
                oldValue = Collections.emptyList();
                break;
            case Notification.REMOVE:
                oldValue = Collections.singletonList(oldValue);
                // fallthrough
            case Notification.REMOVE_MANY:
                newValue = Collections.emptyList();
                break;
            case Notification.SET:
                newValue = Collections.singletonList(newValue);
                oldValue = Collections.singletonList(oldValue);
                break;
            default:
            }
        }
        if (feature instanceof EAttribute) {
            changedAttribute(eObject, (EAttribute)feature, oldValue, newValue);
        } else {
            EReference ref = (EReference)feature;
            if (ref.isContainment()) {
                boolean objectsMoved = containmentChange(eObject, ref, oldValue, newValue);
                if (!objectsMoved) {
                    movedContainments.clear();
                }
            }
            changedReference(eObject, ref, oldValue, newValue);
        }
    }

    @Override
    public void notifyChangedResource(int event, Resource resource, 
                                         Object oldValue, Object newValue) {
        boolean objectsMoved = containmentChange(null, null, oldValue, newValue);
        if (!objectsMoved) {
            movedContainments.clear();
        }

    }

    @Override
    public void notifyChangedResourceSet(int event, ResourceSet resource, 
                                            Object oldValue, Object newValue) {
        // nothing to do
    }
    
    protected static <T> T upCast(Object obj, Class<T> clazz) {
        if (clazz != null && clazz.isInstance(obj)) {
            return clazz.cast(obj);
        }
        return null;
    }
    
    @SuppressWarnings("unchecked")
    protected static <T> Collection<T> upCastList(Object obj, Class<T> clazz) {
        if (clazz != null && obj instanceof Collection) {
            Collection<?> c = (Collection<?>)obj;
            for (Object o : c) {
                if (o != null && !clazz.isInstance(o)) {
                    return null;
                }
            }
            return (Collection<T>)c;
        }
        return null;
    }

    @SuppressWarnings("unchecked")
    private boolean containmentChange(EObject oldContainer, EReference ref, 
                                      Object oldValue, Object newValue) {
        boolean objectsMoved = false;
        
        List<? extends Object> oldVals = oldValue instanceof List 
                       ? (List<? extends Object>)oldValue 
                       : (oldValue instanceof EObject 
                          ? Collections.singletonList(oldValue) 
                          : Collections.emptyList());
        
        for (Object obj : oldVals) {
            if (obj instanceof EObject) {
                EObject eObj = (EObject)obj;
                if (eObj.eContainer() != null || eObj.eResource() != null) {
                    movedContainments.add(eObj);
                    objectsMoved = true;
                } else {
                    detachedObject(eObj, oldContainer, ref);
                }
            }
        }

        List<? extends Object> newVals = newValue instanceof List 
                       ? (List<? extends Object>)newValue 
                       : (newValue instanceof EObject 
                          ? Collections.singletonList(newValue) 
                          : Collections.emptyList());

        for (Object obj : newVals) {
            if (obj instanceof EObject) {
                EObject eObj = (EObject)obj;
                if (!movedContainments.remove(eObj)) {
                    attachedObject(eObj);
                }
            }
        }
        
        return objectsMoved;
    }
}
