/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.resourceset;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.soa.expression.Expression;
import com.iona.soa.expression.Operand;
import com.iona.soa.expression.Operation;
import com.iona.soa.repository.EMFUtil;
import com.iona.soa.repository.EMFUtil.URIEClassEntry;
import com.iona.soa.resourceset.handler.PlainResourceQueryHandler;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceImpl;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.emf.ecore.xmi.impl.XMIResourceImpl;
import org.eclipse.emf.ecore.xmi.impl.XMLResourceImpl;

public final class ResourceSetQueryHandler {

    private static final Logger LOG = Logger.getLogger(ResourceSetQueryHandler.class.getName());

    static {
        QueryHandlerRegistry.getInstance().registerQueryHandler(ResourceImpl.class,
                                                                new PlainResourceQueryHandler());
        QueryHandlerRegistry.getInstance().registerQueryHandler(XMLResourceImpl.class,
                                                                new PlainResourceQueryHandler());
        QueryHandlerRegistry.getInstance().registerQueryHandler(XMIResourceImpl.class,
                                                                new PlainResourceQueryHandler());
    }

    private ResourceSetQueryHandler() {
    }

    @SuppressWarnings("unchecked")
    public static List<EObject> executeQuery(ResourceSet rs, EClass type, boolean resolve)
        throws Exception {
        List<EObject> result = new ArrayList<EObject>();
        List<String> resultIds = new ArrayList<String>();
        Resource resource = null;
        try {
            resource = getResource(rs, type);
            if (resource != null) {
                ResourceQueryHandler handler = getQueryHandler(resource);
                if (handler != null) {
                    Collection<EObject> l = handler.executeQuery(resource, type);
                    addObjects(resource, l, result, resultIds, resolve);
                }
            } else {
                for (Resource r : (List<Resource>)rs.getResources()) {
                    resource = r;
                    ResourceQueryHandler handler = getQueryHandler(resource);
                    if (handler != null) {
                        Collection<EObject> l = handler.executeQuery(resource, type);
                        addObjects(resource, l, result, resultIds, resolve);
                    }
                }
            }
        } finally {
            logWarningsAndErrors(resource, type);
        }
        return result;
    }

    @SuppressWarnings("unchecked")
    public static List<EObject> executeQuery(ResourceSet rs, EClass type, Expression expression,
                                                   boolean resolve) throws Exception {
        List<EObject> result = new ArrayList<EObject>();
        List<String> resultIds = new ArrayList<String>();
        Resource resource = null;
        try {
            resource = getResource(rs, type);
            if (resource != null) {
                ResourceQueryHandler handler = getQueryHandler(resource);
                if (handler != null) {
                    Collection<EObject> l = handler.executeQuery(resource, type, expression);
                    addObjects(resource, l, result, resultIds, resolve);
                }
            } else {
                for (Resource r : (List<Resource>)rs.getResources()) {
                    resource = r;
                    ResourceQueryHandler handler = getQueryHandler(resource);
                    if (handler != null) {
                        if (handler != null) {
                            Collection<EObject> l = handler.executeQuery(resource, type, expression);
                            addObjects(resource, l, result, resultIds, resolve);
                        }
                    }
                }
            }
        } finally {
            logWarningsAndErrors(resource, type);
        }
        return result;
    }
    
    @SuppressWarnings("unchecked")
    public static List<EObject> executeQuery(ResourceSet rs, EClass type, Operation operation,
                                                   boolean resolve) throws Exception {
        List<EObject> result = new ArrayList<EObject>();
        List<String> resultIds = new ArrayList<String>();
        Resource resource = null;
        try {
            resource = getResource(rs, type);
            if (resource != null) {
                ResourceQueryHandler handler = getQueryHandler(resource);
                if (handler != null) {
                    Collection<EObject> l = handler.executeQuery(resource, type, operation);
                    addObjects(resource, l, result, resultIds, resolve);
                }
            } else {
                for (Resource r : (List<Resource>)rs.getResources()) {
                    ResourceQueryHandler handler = getQueryHandler(r);
                    if (handler != null) {
                        resource = r;
                        Collection<EObject> l = handler.executeQuery(resource, type, operation);
                        addObjects(resource, l, result, resultIds, resolve);
                    }
                }
            }
        } finally {
            logWarningsAndErrors(resource, type);
        }
        return result;
    }

    @SuppressWarnings("unchecked")
    public static List<EObject> executeQuery(ResourceSet rs, Operation operation, 
                                                   boolean resolve) throws Exception {
        List<EObject> result = new ArrayList<EObject>();
        List<String> resultIds = new ArrayList<String>();
        List<Resource> resources = new ArrayList<Resource>();
        
        for (Operand o : operation.getArguments()) {
            URIEClassEntry entry = EMFUtil.getObjectIdentity(o.getValue().toString());
            Resource r = getResource(rs, entry.getValue());
            if (r != null && !resources.contains(r)) {
                resources.add(r);
            }
        }
        if (resources.isEmpty()) {
            resources = rs.getResources();
        }
        Resource resource = null;
        try {
            for (Resource r : resources) {
                ResourceQueryHandler handler = getQueryHandler(r);
                if (handler != null) {
                    resource = r;
                    Collection<EObject> l = handler.executeQuery(resource, operation);
                    addObjects(resource, l, result, resultIds, resolve);
                }
            }
        } finally {
            logWarningsAndErrors(resource);
        }
        return result;
    }

    @SuppressWarnings("unchecked")
    public static List<EObject> executeQuery(ResourceSet rs, EClass type, String query, char language,
                                                   boolean resolve) throws Exception {
        List<EObject> result = new ArrayList<EObject>();
        List<String> resultIds = new ArrayList<String>();
        Resource resource = null;
        try {
            resource = getResource(rs, type);
            if (resource != null) {
                ResourceQueryHandler handler = getQueryHandler(resource);
                if (handler != null) {
                    Collection<EObject> l = handler.executeQuery(resource, type, query, language);
                    addObjects(resource, l, result, resultIds, resolve);
                }
            } else {
                for (Resource r : (List<Resource>)rs.getResources()) {
                    ResourceQueryHandler handler = getQueryHandler(r);
                    if (handler != null) {
                        resource = r;
                        Collection<EObject> l = handler.executeQuery(resource, type, query, language);
                        addObjects(resource, l, result, resultIds, resolve);
                    }
                }
            }
        } finally {
            logWarningsAndErrors(resource, type);
        }
        return result;
    }

    @SuppressWarnings("unchecked")
    public static List<EObject> executeQuery(ResourceSet rs, URI resourceURI, EClass type, 
                                                   String query, char language, 
                                                   boolean resolve) throws Exception {
        
        List<EObject> result = new ArrayList<EObject>();
        List<String> resultIds = new ArrayList<String>();
        Resource resource = null;
        try {
            resource = rs.getResource(resourceURI, true);
            if (resource != null) {
                ResourceQueryHandler handler = getQueryHandler(resource);
                if (handler != null) {
                    Collection<EObject> l = handler.executeQuery(resource, type, query, language);
                    addObjects(resource, l, result, resultIds, resolve);
                }
            }
        } finally {
            logWarningsAndErrors(resource, type);
        }
        return result;
    }

    @SuppressWarnings("unchecked")
    public static EObject executeQuery(ResourceSet rs, EClass type, Object id) throws Exception {
        Resource resource = null;
        try {
            resource = getResource(rs, type);
            if (resource != null) {
                ResourceQueryHandler handler = getQueryHandler(resource);
                if (handler != null) {
                    EObject result = handler.executeQuery(resource, type, id);
                    if (result != null) {
                        if (result.eIsProxy()) {
                            return EcoreUtil.resolve(result, resource);
                        } else {
                            return result;
                        }
                    }
                }
            } else {
                for (Resource r : (List<Resource>)rs.getResources()) {
                    ResourceQueryHandler handler = getQueryHandler(r);
                    if (handler != null) {
                        resource = r;
                        EObject result = handler.executeQuery(resource, type, id);
                        if (result != null) {
                            if (result.eIsProxy()) {
                                return EcoreUtil.resolve(result, resource);
                            } else {
                                return result;
                            }
                        }
                    }
                }
            }
        } finally {
            logWarningsAndErrors(resource, type);
        }
        return null;
    }

    private static void addObjects(Resource resource, Collection<EObject> objects,
                                   Collection<EObject> result, Collection<String> resultIds, 
                                   boolean resolve) {
        for (EObject obj : objects) {
            EObject resolved = obj;
            if (resolve && obj.eIsProxy()) {
                resolved = EcoreUtil.resolve(obj, resource);
            }
            String guid = EMFUtil.getGuid(resolved);
            if (resolved == null) {
                LOG.log(Level.WARNING, "Failed to resolve proxy object {0}", obj);
            } else if (!resultIds.contains(guid)) {
                result.add(resolved);
                resultIds.add(guid);
            }
        }
    }

    private static ResourceQueryHandler getQueryHandler(Resource resource) {
        return QueryHandlerRegistry.getInstance().getQueryHandler(resource);
    }

    // TO DO: Change JDBCResourceSetImpl to use URIConverter
    private static Resource getResource(ResourceSet rs, EClass type) {
        Resource resource = null;
        String packageUri = type.getEPackage().getNsURI();
        URI resourceUri = (URI)rs.getURIConverter().getURIMap().get(URI.createURI(packageUri));
        if (null != resourceUri) {
            resource = rs.getResource(resourceUri, true);
        }
        return resource;
    }

    private static void logWarningsAndErrors(Resource resource) {
        if (resource.getWarnings() != null && !resource.getWarnings().isEmpty()) {
            for (Iterator it = resource.getWarnings().iterator(); it.hasNext();) {
                log(Level.WARNING, it.next());
            }
        }

        if (resource.getErrors() != null && !resource.getErrors().isEmpty()) {
            for (Iterator it = resource.getErrors().iterator(); it.hasNext();) {
                log(Level.SEVERE, it.next());
            }
        }
    }
    
    private static void logWarningsAndErrors(Resource resource, EClass type) {
        if (resource == null) {
            log(Level.WARNING, "No Resource was found for Class: " + type.getName()
                               + "(" + type.getEPackage().getNsURI() + ")");
            return;
        }
        logWarningsAndErrors(resource);
    }

    private static void log(Level level, Object diagnostic) {
        // The Resource API docco states that items in the getErrors() and
        // getWarnings() lists
        // will be of type Resource.Diagnostic. It's a big fat whopper!
        //
        if (diagnostic instanceof Resource.Diagnostic) {
            Resource.Diagnostic rd = (Resource.Diagnostic)diagnostic;
            LOG.log(level, "Message={0}, Location={1}, Line={2}, Column={3}", new Object[] {rd.getMessage(),
                                                                                            rd.getLocation(),
                                                                                            rd.getLine(),
                                                                                            rd.getColumn()});
        } else if (diagnostic instanceof Resource.IOWrappedException) {
            Resource.IOWrappedException ioex = (Resource.IOWrappedException)diagnostic;
            LOG.log(level, ioex.getLocalizedMessage());
        } else {
            LOG.log(level, diagnostic.toString());
        }
    }
}
