/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.resourceset.handler;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.iona.soa.expression.Expression;
import com.iona.soa.expression.Operand;
import com.iona.soa.expression.Operation;
import com.iona.soa.expression.OperationType;
import com.iona.soa.expression.Term;
import com.iona.soa.expression.delegate.BeanExpressionHandler;
import com.iona.soa.expression.serialization.ExpressionSerializer;
import com.iona.soa.repository.EMFUtil;
import com.iona.soa.repository.EMFUtil.URIEClassEntry;
import com.iona.soa.resourceset.QueryConstants;
import com.iona.soa.resourceset.ResourceQueryHandler;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.resource.Resource;

public class PlainResourceQueryHandler implements ResourceQueryHandler {
    @SuppressWarnings("unchecked")
    public Collection<EObject> executeQuery(Resource resource, EClass type) throws Exception {
        return getObjectsByType(resource.getContents(), type);
    }

    @SuppressWarnings("unchecked")
    public Collection<EObject> executeQuery(Resource resource, EClass type, Operation operation) {
        Operand[] arguments = operation.getArguments();
        if (OperationType._GETXREFS.equals(operation.getName())
            && arguments.length == 1
            && arguments[0].getValue() instanceof String) {
            String guid = (String)arguments[0].getValue();
            return getXRefs(resource, type, guid);
        }
        return Collections.EMPTY_LIST;
    }
    
    @SuppressWarnings("unchecked")
    public Collection<EObject> executeQuery(Resource resource, Operation operation) {
        if (OperationType._xrefs.equals(operation.getName())) {
            Collection<EObject> results = new LinkedList<EObject>();
            Collection<EObject> visited = new LinkedList<EObject>();
            for (Operand o : operation.getArguments()) {
                URIEClassEntry entry = EMFUtil.getObjectIdentity(o.getValue().toString());
                EObject eo = null;
                for (Iterator<EObject> iter = (Iterator<EObject>)resource.getAllContents(); iter.hasNext();) {
                    EObject eObject = iter.next();
                    if (EMFUtil.getProxyURI(eObject).equals(entry.getKey())) {
                        eo = eObject;
                        break;
                    }
                }
                if (eo == null) {
                    return results;
                }
                visited.add(eo);
                for (EObject contained : (List<EObject>)eo.eContents()) {
                    visited.add(contained);
                }
            }
            
            for (Operand o : operation.getArguments()) {
                URIEClassEntry entry = EMFUtil.getObjectIdentity(o.getValue().toString());
                String id = entry.getKey().fragment().substring(1).split("/")[1];
                Collection<EObject> result = getXRefs(resource, entry.getValue(), id);
                for (EObject eo : result) {
                    // String guid = EMFUtil.getGuid(eo);
                    if (!results.contains(eo)) {
                        results.add(eo);
                    }
                }
                for (EObject eo : visited) {
                    results.remove(eo);
                }
            }
            return results;
        } else if (OperationType._deps.equals(operation.getName())) {
            throw new UnsupportedOperationException("Unsupported operation: " + operation.getName());
        }

        throw new UnsupportedOperationException("Unsupported operation: " + operation.getName());
    }
    
    @SuppressWarnings("unchecked")
    public Collection<EObject> executeQuery(Resource resource, EClass type, Expression expr) {
        Collection objs = getObjectsByType(resource.getContents(), type);
        BeanExpressionHandler beh = new BeanExpressionHandler(objs);
        beh.handle(expr);
        return beh.getQualifying();
    }
    
    @SuppressWarnings("unchecked")
    private Collection<EObject> getObjectsByType(Collection<? extends EObject> eObjs, EClassifier type) {
        Collection<EObject> ret = new LinkedList<EObject>();
        if (eObjs == null) {
            return ret;
        }
        
        for (EObject eObj :  eObjs) {
            if (type.isInstance(eObj)) {
                ret.add(eObj);
            }
            if (eObj.eContents().size() > 0) {
                ret.addAll(getObjectsByType(eObj.eContents(), type));
            }
        }
        return ret;
    }

    public Collection<EObject> executeQuery(Resource resource, EClass type, String query, char language) 
        throws Exception {
        switch (language) {
        case QueryConstants.QUERY_LANGUAGE_IT_EMFQL:
            Term expr = ExpressionSerializer.deserialize(query);
            if (expr instanceof Expression) {
                return executeQuery(resource, type, (Expression) expr);                
            } else if (expr instanceof Operation) {
                return executeQuery(resource, type, (Operation) expr);
            } else {
                throw new UnsupportedOperationException("Cannot handle the following query: " + query);
            }
        default:
            throw new UnsupportedOperationException("Unsupported query language: " + language);        
        }
    }

    public EObject executeQuery(Resource resource, EClass type, Object id) throws Exception {
        EAttribute idAttr = type.getEIDAttribute();
        if (idAttr == null) {
            return null;
        }
        if (id == null) {
            throw new NullPointerException("Entity ID cannot be null: " + idAttr.getName());
        }
        
        for (EObject obj : executeQuery(resource, type)) {
            if (id.equals(obj.eGet(idAttr))) {
                return obj;
            }
        }
        return null;
    }

    
    @SuppressWarnings("unchecked")
    private Collection<EObject> getXRefs(Resource resource, EClass type, String guid) {
        List<EObject> result = new LinkedList<EObject>();
        Map<EClass, Collection<EReference>> backReferences = EMFUtil.getBackReferences(type, null);
        Set<EReference> allRefs = new HashSet<EReference>();
        for (EClass eClass : backReferences.keySet()) {
            allRefs.addAll(backReferences.get(eClass));
        }
        
        
        // find backrefs in resource
        //
        for (Iterator<EObject> iter = (Iterator<EObject>)resource.getAllContents(); iter.hasNext();) {
            EObject eObject = iter.next();
            if (hasReference(eObject, allRefs, guid)) {
                result.add(eObject);
            }
        }
        return result;
    }

    @SuppressWarnings("unchecked")
    private boolean hasReference(EObject eObject, Collection<EReference> references, String guid) {
        for (EReference reference : references) {
            if (eObject.eClass().getEAllReferences().contains(reference)
                && !reference.isContainment() && eObject.eIsSet(reference)) {
                Object value = eObject.eGet(reference);
                if (value instanceof List) {
                    for (EObject referent : (List<EObject>)value) {
                        if (guid.equals(EMFUtil.getGuid(referent))) {
                            return true;
                        }
                    }
                } else if (value instanceof EObject
                    && guid.equals(EMFUtil.getGuid((EObject)value))) {
                    return true;
                }
            }
        }
        return false;
    }
}
