/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.expression.delegate;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import com.iona.soa.expression.And;
import com.iona.soa.expression.Expression;
import com.iona.soa.expression.Operand;
import com.iona.soa.expression.Operation;
import com.iona.soa.expression.OperationType;
import com.iona.soa.expression.runtime.DelegatingFactory;
import com.iona.soa.expression.runtime.ExpressionRuntime;
import com.iona.soa.expression.runtime.model.ContextKey;
import com.iona.soa.expression.serialization.ExpressionSerializer;
import com.iona.soa.repository.EMFUtil;
import com.iona.soa.repository.EMFUtil.URIEClassEntry;
import com.iona.soa.repository.jdbc.resource.JDBCResourceImpl;
import com.iona.soa.repository.util.Out;
import com.iona.soa.resourceset.ResourceSetQueryHandler;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;

public class ReadOnlyOperationDelegate implements JDBCOperationDelegate {

    public static class Factory implements JDBCOperationDelegate.Factory {
        public JDBCOperationDelegate create(JDBCResourceImpl resource) {
            return new ReadOnlyOperationDelegate(resource);
        }
    };

    private JDBCResourceImpl resource;
    private Exception ex;
    private Collection<EObject> results = new LinkedList<EObject>();
    
    public ReadOnlyOperationDelegate(JDBCResourceImpl resource) {
        this.resource  = resource;
    }

    @SuppressWarnings("unchecked")
    public boolean handle(Operation op) {
        try {
            // find out all dependencies
            Operation depsOp  = new Operation(OperationType._deps, (Object[])op.getArguments());
            Collection<EObject> deps =  
                ResourceSetQueryHandler.executeQuery(resource.getResourceSet(), 
                                                     depsOp, true);
            
            // resolve root query objects and add them into dependency list
            for (Operand o : op.getArguments()) {
                URIEClassEntry entry = EMFUtil.getObjectIdentity(o.getValue().toString());
                EObject eobj = resource.getResourceSet().getEObject(entry.getKey(), true);
                if (eobj != null) {
                    deps.add(eobj);
                }
            }
            // for each dependency, find its delete predicate using model hierarchy;
            // if evaluated to true then skip this object else add it into the resultset
            //
            // TO DO: Extend the query framework to include parent's class
            // expressions against the same annotation 
            //
            // Then Document's delete predicate query that looks like this now:
            // or(equals(modifiable, true), and(equals(modifiable, false), equals(external, true)))
            // can be rewritten to look like this:
            // or(super(), and(equals(modifiable, false), equals(external, true)))
            
            for (EObject o : deps) {
                Out<Collection<Expression>> exps = 
                    new Out<Collection<Expression>>(new LinkedList<Expression>());
                getReadOnlyExpression(o.eClass(), exps);
                if (exps.get().size() > 0) {
                    Expression exp = null;
                    if (exps.get().size() > 1) {
                        exp = new And(exps.get().toArray());
                    } else {
                        exp = exps.get().iterator().next();
                    }
                    ExpressionRuntime runtime = DelegatingFactory.create(exp.getOperator());
                    Map<ContextKey, Object> context = new HashMap<ContextKey, Object>();
                    context.put(ContextKey.EOBJECT, o);
                    context.put(ContextKey.SET, deps);
                    runtime.run(exp, context);
                    Boolean result = (Boolean)context.get(ContextKey.RESULT);
                    if (!result.equals(Boolean.TRUE)) {
                        results.add(o);
                    }
                }
            }
        } catch (Exception e) {
            ex = e;
            return false;
        } 
        return true;
    }
    
    // starting from this class walk up the type hierarchy and collect expressions, 
    // if an expression is found then stop 
    protected void getReadOnlyExpression(EClass eclass, 
                                         Out<Collection<Expression>> exps) {
        Collection<EClass> visited = new LinkedList<EClass>();
        getReadOnlyExpression(eclass, exps, visited);
    }
    
    @SuppressWarnings("unchecked")
    protected void getReadOnlyExpression(EClass eclass, 
                                         Out<Collection<Expression>> exps,
                                         Collection<EClass> visited) {
        if (visited.contains(eclass)) {
            return;
        }
        visited.add(eclass);
        
        String value = EMFUtil.getAnnotationValue(eclass, 
                                                  EMFUtil.DELETE_ANNOTATION, 
                                                  EMFUtil.DELETE_PREDICATE);
        if (value != null) {
            exps.get().addAll(ExpressionSerializer.deserializeModelQuery(value));
        } else {
            for (EClass c : (List<EClass>)eclass.getESuperTypes()) {
                getReadOnlyExpression(c, exps, visited);
            }
        }
    } 
    
    public boolean handle(Expression e) {
        // This can be implemented so that unwanted dependencies 
        // are filtered via this expression
        throw new UnsupportedOperationException();
    }

    public boolean handle(Operand op) {
        throw new UnsupportedOperationException();
    }
    
    public Collection<EObject> getResult() {
        return results;
    }
    public Exception getException() {
        return ex;
    }
}
