/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.expression.delegate;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;

import com.iona.soa.expression.Expression;
import com.iona.soa.expression.Operand;
import com.iona.soa.expression.Operation;
import com.iona.soa.expression.OperationDelegate;
import com.iona.soa.expression.Term;
import com.iona.soa.repository.jdbc.EMFSQLUtil;
import com.iona.soa.repository.util.SQLHelper;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

public class SQLRenderer implements OperationDelegate {
    private final EClass resultType; 
    private WhereClause whereClause = new WhereClause();
    
    public SQLRenderer(EClass type) {
        resultType = type;
    }

    public boolean handle(Operation op) {
        throw new UnsupportedOperationException();
    }
    
    public boolean handle(Expression expr) {
        switch (expr.getOperator()) {
        case AND: 
            return performAnd(expr.getOperands());
        case EQUALS:
            return performEquals(expr.getOperands());
        case ISNULL:
            return performIsNull(expr.getOperands());
        default:
            return false;
        }
    }



    public boolean handle(Operand op) {
        if (whereClause.whereString.length() > 0 
                && whereClause.whereString.charAt(whereClause.whereString.length() - 1) == '=') {
            // the rhs of an equation
            whereClause.whereString.append('?');
            whereClause.values.add(op.getValue());
            return true;
        } else {
            return handleEqualsLHS(op.getValue());
        }
    }
    
    private boolean handleEqualsLHS(Object value) {
        String address = value.toString();
        
        List<String> indirectRefs = new LinkedList<String>();
        StringTokenizer tok = new StringTokenizer(address, ".");
        while (tok.hasMoreTokens()) {
            indirectRefs.add(tok.nextToken());
        }

        String field = indirectRefs.remove(indirectRefs.size() - 1);
        EClass lastType = handleEqualsIndirections(indirectRefs);
        if (field == null || lastType == null) {
            return false;
        }
        whereClause.whereString.append(EMFSQLUtil.getTableName(lastType));
        whereClause.whereString.append('.');
        whereClause.whereString.append(SQLHelper.getNonSQLKeyword(field));
        return true;
    }

    private EClass handleEqualsIndirections(List<String> indirectRefs) {
        EClass curType = resultType;
        for (String ref : indirectRefs) {
            curType = handleEqualsIndirection(ref, curType);
            if (curType == null) {
                return null;
            }
        }
        return curType;
    }

    private EClass handleEqualsIndirection(String ref, EClass curType) {
        EStructuralFeature feature = curType.getEStructuralFeature(ref);
        if (feature instanceof EReference) {
            EReference reference = (EReference) feature;

            if (reference.getEOpposite() != null) {
                return handleToManyBidirectional(curType, reference);
            } else {
                if (reference.isMany()) {                        
                    return handleToManyUnidirectional(curType, reference);
                } else {
                    return handleToOneUnidirectional(curType, reference);
                }
            }
        } else {
            return null;
        }
    }

    private EClass handleToManyBidirectional(EClass curType, EReference reference) {
        EClass oppositeType = reference.getEReferenceType();

        String curTableName = EMFSQLUtil.getTableName(curType);
        String oppositeTableName = EMFSQLUtil.getTableName(oppositeType);
        
        EReference opposite = reference.getEOpposite();
        String joinTableName = EMFSQLUtil.getJoinTableName(reference, opposite);            
        
        whereClause.extraTables.add(oppositeTableName);
        whereClause.extraTables.add(joinTableName);
        
        whereClause.whereString.append(joinTableName);
        whereClause.whereString.append(".");
        whereClause.whereString.append(EMFSQLUtil.getFKColumnName(opposite));
        whereClause.whereString.append("=");
        whereClause.whereString.append(curTableName);
        whereClause.whereString.append(".");
        whereClause.whereString.append(curType.getEIDAttribute().getName());
        whereClause.whereString.append(" AND ");
        whereClause.whereString.append(joinTableName);
        whereClause.whereString.append(".");
        whereClause.whereString.append(EMFSQLUtil.getFKColumnName(reference));
        whereClause.whereString.append("=");
        whereClause.whereString.append(oppositeTableName);
        whereClause.whereString.append(".");
        whereClause.whereString.append(oppositeType.getEIDAttribute().getName());
        whereClause.whereString.append(" AND ");
        
        return oppositeType;
    }

    private EClass handleToManyUnidirectional(EClass curType, EReference reference) {
        EClass oppositeType = reference.getEReferenceType();

        String curTableName = EMFSQLUtil.getTableName(curType);
        String oppositeTableName = EMFSQLUtil.getTableName(oppositeType);

        String joinTableName = EMFSQLUtil.getJoinTableName(reference);            
        String curClassJoinFKColumnName = EMFSQLUtil.getFKColumnName(
                reference.getEContainingClass().getEIDAttribute());
        
        whereClause.extraTables.add(oppositeTableName);
        whereClause.extraTables.add(joinTableName);
        
        whereClause.whereString.append(joinTableName);
        whereClause.whereString.append(".");
        whereClause.whereString.append(curClassJoinFKColumnName);
        whereClause.whereString.append("=");
        whereClause.whereString.append(curTableName);
        whereClause.whereString.append(".");
        whereClause.whereString.append(curType.getEIDAttribute().getName());
        whereClause.whereString.append(" AND ");
        whereClause.whereString.append(joinTableName);
        whereClause.whereString.append(".");
        whereClause.whereString.append(EMFSQLUtil.getFKColumnName(reference));
        whereClause.whereString.append("=");
        whereClause.whereString.append(oppositeTableName);
        whereClause.whereString.append(".");
        whereClause.whereString.append(oppositeType.getEIDAttribute().getName());
        whereClause.whereString.append(" AND ");
        
        return oppositeType;
    }

    private EClass handleToOneUnidirectional(EClass curType, EReference reference) {
        EClass oppositeType = reference.getEReferenceType();

        String curTableName = EMFSQLUtil.getTableName(curType);
        String oppositeTableName = EMFSQLUtil.getTableName(oppositeType);

        whereClause.extraTables.add(oppositeTableName);
        
        whereClause.whereString.append(curTableName);
        whereClause.whereString.append(".");
        whereClause.whereString.append(EMFSQLUtil.getFKColumnName(reference));
        whereClause.whereString.append("=");
        whereClause.whereString.append(oppositeTableName);
        whereClause.whereString.append(".");
        whereClause.whereString.append(oppositeType.getEIDAttribute().getName());
        whereClause.whereString.append(" AND ");
        
        return oppositeType;
    }

    private boolean handle(Term term) {
        if (term instanceof Expression) {
            return handle((Expression) term);
        } else if (term instanceof Operand) {
            return handle((Operand) term);
        }
        return false;
    }
    
    private boolean performAnd(Term[] operands) {
        boolean ret = true;
        boolean firstTime = true;
        for (Term operand : operands) {
            if (firstTime) {
                firstTime = false;
            } else {
                whereClause.whereString.append(" INTERSECT ");
            }
            
            ret &= handle(operand);
        }
        return ret;
    }

    private boolean performEquals(Term[] operands) {
        boolean ret = handle(operands[0]);
        whereClause.whereString.append("=");
        ret &= handle(operands[1]);
        return ret;
    }
    
    private boolean performIsNull(Term[] operands) {
        boolean ret = handle(operands[0]);
        whereClause.whereString.append(" IS NULL");
        return ret;
    }

    public WhereClause getWhereClause() {
        return whereClause;
    }
    
    public static final class WhereClause {
        private final Set<String> extraTables = new HashSet<String>();
        private final StringBuffer whereString = new StringBuffer();
        private final List<Object> values = new LinkedList<Object>();
        
        WhereClause() { }
        
        public WhereClause(String where) {
            whereString.append(where);
        }
        
        private String getExtraTableFromClause() {
            StringBuilder sb = new StringBuilder();
            for (String table : extraTables) {
                sb.append(',');
                sb.append(table);
            }
            return sb.toString();
        }
        
        public String getFullQuery(EClass type) {
            StringBuilder query = new StringBuilder();
            boolean firstTime = true;
            for (String t : getWhereClause().split(" INTERSECT ")) {
                if (firstTime) {
                    firstTime = false;
                } else {
                    query.append(" INTERSECT ");
                }

                query.append(getSelect(type));
                query.append(t);                
            }
            return query.toString();
        }
        
        public String getSelect(EClass type) {
            return "SELECT " + EMFSQLUtil.getIdentifyingColumns(type)
                + " FROM " + EMFSQLUtil.getTableName(type) + getExtraTableFromClause()
                + " WHERE ";            
        }
        
        public String getWhereClause() {
            return whereString.toString().toUpperCase();
        }
        
        public Object[] getParams() {
            return values.toArray();
        }
    }
}
