/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.jdbc;

import java.io.InputStream;
import java.io.PrintStream;
import java.io.Reader;
import java.math.BigDecimal;
import java.net.URL;
import java.sql.Array;
import java.sql.Blob;
import java.sql.Clob;
import java.sql.Connection;
import java.sql.Date;
import java.sql.ParameterMetaData;
import java.sql.PreparedStatement;
import java.sql.Ref;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.SQLWarning;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Map;
import java.util.StringTokenizer;

class AuditedPreparedStatement implements PreparedStatement {
    private final String preparedSQL;
    private PreparedStatement preparedStatement;
    private final Map<Integer, Object> values = new HashMap<Integer, Object>();
    private final PrintStream logger;

    AuditedPreparedStatement(PreparedStatement ps, String s, PrintStream l) {
        preparedStatement = ps;
        preparedSQL = s;
        logger = l;
    }
    
    private void log(String sql) {
        logger.println("[STMT] " + fillPreparedString(sql));
    }

    private String fillPreparedString(String sql) {
        StringBuilder sb = new StringBuilder();
        
        StringTokenizer tok = new StringTokenizer(sql, "?", true);
        // really stupid workaround for Checkstyle
        // ModifiedControlVarialbe rule.  Will remove it from the 
        // iona checkstyle eventually but that's a PIA
        int i = -1;
        for (i = 1; tok.hasMoreTokens();) {
            String cur = tok.nextToken();
            if ("?".equals(cur)) {
                sb.append("" + values.get(i++));
            } else {
                sb.append(cur);
            }
        }
        
        return sb.toString();
    }

    public void addBatch() throws SQLException {
        preparedStatement.addBatch();
    }

    public void addBatch(String sql) throws SQLException {
        preparedStatement.addBatch(sql);
    }

    public void cancel() throws SQLException {
        preparedStatement.cancel();
    }

    public void clearBatch() throws SQLException {
        preparedStatement.clearBatch();
    }

    public void clearParameters() throws SQLException {
        preparedStatement.clearParameters();
    }

    public void clearWarnings() throws SQLException {
        preparedStatement.clearWarnings();
    }

    public void close() throws SQLException {
        preparedStatement.close();
    }

    public boolean execute() throws SQLException {
        return preparedStatement.execute();
    }

    public boolean execute(String sql, int autoGeneratedKeys) throws SQLException {
        log(sql);
        return preparedStatement.execute(sql, autoGeneratedKeys);
    }

    public boolean execute(String sql, int[] columnIndexes) throws SQLException {
        log(sql);
        return preparedStatement.execute(sql, columnIndexes);
    }

    public boolean execute(String sql, String[] columnNames) throws SQLException {
        log(sql);
        return preparedStatement.execute(sql, columnNames);
    }

    public boolean execute(String sql) throws SQLException {
        log(sql);
        return preparedStatement.execute(sql);
    }

    public int[] executeBatch() throws SQLException {
        return preparedStatement.executeBatch();
    }

    public ResultSet executeQuery() throws SQLException {
        log(preparedSQL);
        return preparedStatement.executeQuery();
    }

    public ResultSet executeQuery(String sql) throws SQLException {
        log(sql);
        return preparedStatement.executeQuery(sql);
    }

    public int executeUpdate() throws SQLException {
        log(preparedSQL);
        return preparedStatement.executeUpdate();
    }

    public int executeUpdate(String sql, int autoGeneratedKeys) throws SQLException {
        log(sql);
        return preparedStatement.executeUpdate(sql, autoGeneratedKeys);
    }

    public int executeUpdate(String sql, int[] columnIndexes) throws SQLException {
        log(sql);
        return preparedStatement.executeUpdate(sql, columnIndexes);
    }

    public int executeUpdate(String sql, String[] columnNames) throws SQLException {
        log(sql);
        return preparedStatement.executeUpdate(sql, columnNames);
    }

    public int executeUpdate(String sql) throws SQLException {
        log(sql);
        return preparedStatement.executeUpdate(sql);
    }

    public Connection getConnection() throws SQLException {
        return preparedStatement.getConnection();
    }

    public int getFetchDirection() throws SQLException {
        return preparedStatement.getFetchDirection();
    }

    public int getFetchSize() throws SQLException {
        return preparedStatement.getFetchSize();
    }

    public ResultSet getGeneratedKeys() throws SQLException {
        return preparedStatement.getGeneratedKeys();
    }

    public int getMaxFieldSize() throws SQLException {
        return preparedStatement.getMaxFieldSize();
    }

    public int getMaxRows() throws SQLException {
        return preparedStatement.getMaxRows();
    }

    public ResultSetMetaData getMetaData() throws SQLException {
        return preparedStatement.getMetaData();
    }

    public boolean getMoreResults() throws SQLException {
        return preparedStatement.getMoreResults();
    }

    public boolean getMoreResults(int current) throws SQLException {
        return preparedStatement.getMoreResults(current);
    }

    public ParameterMetaData getParameterMetaData() throws SQLException {
        return preparedStatement.getParameterMetaData();
    }

    public int getQueryTimeout() throws SQLException {
        return preparedStatement.getQueryTimeout();
    }

    public ResultSet getResultSet() throws SQLException {
        return preparedStatement.getResultSet();
    }

    public int getResultSetConcurrency() throws SQLException {
        return preparedStatement.getResultSetConcurrency();
    }

    public int getResultSetHoldability() throws SQLException {
        return preparedStatement.getResultSetHoldability();
    }

    public int getResultSetType() throws SQLException {
        return preparedStatement.getResultSetType();
    }

    public int getUpdateCount() throws SQLException {
        return preparedStatement.getUpdateCount();
    }

    public SQLWarning getWarnings() throws SQLException {
        return preparedStatement.getWarnings();
    }

    public void setArray(int i, Array x) throws SQLException {
        values.put(i, x);
        preparedStatement.setArray(i, x);
    }

    public void setAsciiStream(int parameterIndex, InputStream x, int length) throws SQLException {
        preparedStatement.setAsciiStream(parameterIndex, x, length);
    }

    public void setBigDecimal(int parameterIndex, BigDecimal x) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setBigDecimal(parameterIndex, x);
    }

    public void setBinaryStream(int parameterIndex, InputStream x, int length) throws SQLException {
        preparedStatement.setBinaryStream(parameterIndex, x, length);
    }

    public void setBlob(int i, Blob x) throws SQLException {
        values.put(i, x);
        preparedStatement.setBlob(i, x);
    }

    public void setBoolean(int parameterIndex, boolean x) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setBoolean(parameterIndex, x);
    }

    public void setByte(int parameterIndex, byte x) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setByte(parameterIndex, x);
    }

    public void setBytes(int parameterIndex, byte[] x) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setBytes(parameterIndex, x);
    }

    public void setCharacterStream(int parameterIndex, Reader reader, int length) throws SQLException {
        preparedStatement.setCharacterStream(parameterIndex, reader, length);
    }

    public void setClob(int i, Clob x) throws SQLException {
        values.put(i, x);
        preparedStatement.setClob(i, x);
    }

    public void setCursorName(String name) throws SQLException {
        preparedStatement.setCursorName(name);
    }

    public void setDate(int parameterIndex, Date x, Calendar cal) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setDate(parameterIndex, x, cal);
    }

    public void setDate(int parameterIndex, Date x) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setDate(parameterIndex, x);
    }

    public void setDouble(int parameterIndex, double x) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setDouble(parameterIndex, x);
    }

    public void setEscapeProcessing(boolean enable) throws SQLException {
        preparedStatement.setEscapeProcessing(enable);
    }

    public void setFetchDirection(int direction) throws SQLException {
        preparedStatement.setFetchDirection(direction);
    }

    public void setFetchSize(int rows) throws SQLException {
        preparedStatement.setFetchSize(rows);
    }

    public void setFloat(int parameterIndex, float x) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setFloat(parameterIndex, x);
    }

    public void setInt(int parameterIndex, int x) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setInt(parameterIndex, x);
    }

    public void setLong(int parameterIndex, long x) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setLong(parameterIndex, x);
    }

    public void setMaxFieldSize(int max) throws SQLException {
        preparedStatement.setMaxFieldSize(max);
    }

    public void setMaxRows(int max) throws SQLException {
        preparedStatement.setMaxRows(max);
    }

    public void setNull(int paramIndex, int sqlType, String typeName) throws SQLException {
        values.put(paramIndex, null);
        preparedStatement.setNull(paramIndex, sqlType, typeName);
    }

    public void setNull(int parameterIndex, int sqlType) throws SQLException {
        values.put(parameterIndex, null);
        preparedStatement.setNull(parameterIndex, sqlType);
    }

    public void setObject(int parameterIndex, Object x, int targetSqlType, int scale) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setObject(parameterIndex, x, targetSqlType, scale);
    }

    public void setObject(int parameterIndex, Object x, int targetSqlType) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setObject(parameterIndex, x, targetSqlType);
    }

    public void setObject(int parameterIndex, Object x) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setObject(parameterIndex, x);
    }

    public void setQueryTimeout(int seconds) throws SQLException {
        preparedStatement.setQueryTimeout(seconds);
    }

    public void setRef(int i, Ref x) throws SQLException {
        preparedStatement.setRef(i, x);
    }

    public void setShort(int parameterIndex, short x) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setShort(parameterIndex, x);
    }

    public void setString(int parameterIndex, String x) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setString(parameterIndex, x);
    }

    public void setTime(int parameterIndex, Time x, Calendar cal) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setTime(parameterIndex, x, cal);
    }

    public void setTime(int parameterIndex, Time x) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setTime(parameterIndex, x);
    }

    public void setTimestamp(int parameterIndex, Timestamp x, Calendar cal) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setTimestamp(parameterIndex, x, cal);
    }

    public void setTimestamp(int parameterIndex, Timestamp x) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setTimestamp(parameterIndex, x);
    }

    @SuppressWarnings("deprecation")
    public void setUnicodeStream(int parameterIndex, InputStream x, int length) throws SQLException {
        preparedStatement.setUnicodeStream(parameterIndex, x, length);
    }

    public void setURL(int parameterIndex, URL x) throws SQLException {
        values.put(parameterIndex, x);
        preparedStatement.setURL(parameterIndex, x);
    }
}
