/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.jdbc;

import java.sql.SQLException;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.soa.repository.jdbc.resource.JDBCResource;

import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.notify.Notifier;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.xmi.XMIException;

/** This adapter is used as a listener to changes to an EMF object. It will automatically serialize these 
 * changes to the database. 
 * <p/>
 * This class is used by the {@link com.iona.soa.repository.jdbc.resource.JDBCResourceImpl} class to 
 * achieve automatic persistence.
 */
public class EMFDatabaseSerializationAdapter implements Adapter {
    private static final Logger LOG = Logger.getLogger(EMFDatabaseSerializationAdapter.class.getName());
    private final JDBCResource resource;

    /** Constructor. What is needed is: an <tt>EObject</tt>, a database connection and an instance of
     * a {@link Persister}.
     * @param res The JDBCResource to use
     */
    public EMFDatabaseSerializationAdapter(JDBCResource res) {
        resource = res;
    }
    
    /** Implements the <tt>Adapter</tt> API.
     * @param notification The actual notification.
     * @see org.eclipse.emf.common.notify.Adapter#notifyChanged(org.eclipse.emf.common.notify.Notification)
     */
    @SuppressWarnings("unchecked")
    public void notifyChanged(Notification notification) {
        if (notification.getNewValue() instanceof Throwable) {
            // This gets called when addException() is called;
            return;
        }

        try {
            if (notification.getNotifier() == resource) {
                notifyChangeResource(notification);
            } else if (notification.getNotifier() instanceof EObject) {
                notifyChangedObject(notification);
            }
        } catch (SQLException e) {
            resource.getErrors().add(new XMIException(e));
            LOG.log(Level.WARNING, "Database error: {0}", e.toString());
            LOG.log(Level.FINE, "Database error", e);
            throw new RuntimeException(e);
        }
    }

    private void notifyChangeResource(Notification notification) throws SQLException {
        switch (notification.getEventType()) {
        case Notification.ADD:
            if (notification.getNewValue() instanceof EObject) {
                EObject newObject = (EObject) notification.getNewValue();
                resource.persist(newObject);
            }
            break;                
        case Notification.REMOVE:
            if (notification.getOldValue() instanceof EObject) {
                EObject deletedObj = (EObject) notification.getOldValue();
                if (deletedObj.eContainer() == null) {
                    resource.unPersist(deletedObj, true);
                }
            }
            break;
        case Notification.ADD_MANY:
            for (Object obj : (List<?>)notification.getNewValue()) {
                if (obj instanceof EObject) {
                    EObject newObject = (EObject) notification.getNewValue();
                    resource.persist(newObject);
                }
            }
            break;                
        case Notification.REMOVE_MANY:
            for (Object obj : (List<?>)notification.getOldValue()) {
                if (obj instanceof EObject) {
                    EObject deletedObj = (EObject) obj;
                    if (deletedObj.eContainer() == null) {
                        resource.unPersist(deletedObj, true);
                    }
                }
            }
            break;
        default:
                
        }
    }

    private void notifyChangedObject(Notification notification) throws SQLException {
        switch (notification.getEventType()) {
        case Notification.SET:
            notifyUpdateField(notification);
            break;
        case Notification.UNSET:
            // Todo, when does this typically happen?
            // notifyUpdateField(notification);
            break;
        case Notification.ADD:
            notifyAdd(notification);
            break;
        case Notification.REMOVE:
            notifyRemove(notification);
            break;
        case Notification.ADD_MANY:
            notifyAdd(notification);
            break;
        case Notification.REMOVE_MANY:
            notifyRemove(notification);
            break;
        case Notification.MOVE:
            // Move can be dealed with in the same way as add
            notifyAdd(notification); 
            break;
//          Had to comment the ones below out as checkstyle was complaining about the amount of 
//            case clauses.
//          case Notification.REMOVING_ADAPTER:
//          break;
//          case Notification.RESOLVE:
//          break;
        default: 
            break;
        }
    }

    private void notifyUpdateField(Notification notification) throws SQLException {
        EObject notifier = (EObject)notification.getNotifier();
        Object feature = notification.getFeature();
        if (feature instanceof EAttribute) {
            EAttribute attr = (EAttribute)feature;
            if (attr.isID() && notification.getNewValue() == null) {
                // this is a cue to remove the object from the database

                // Set the ID value back and then delete the object
                boolean prevDeliver = notifier.eDeliver();
                try {
                    notifier.eSetDeliver(false);
                    notifier.eSet(attr, notification.getOldValue());
                } finally {
                    notifier.eSetDeliver(prevDeliver);
                }

                if (resource.getContents().contains(notifier)) {
                    boolean oldDeliver = resource.eDeliver();
                    try {
                        resource.eSetDeliver(false);
                        resource.getContents().remove(notifier);
                    } finally {
                        resource.eSetDeliver(oldDeliver);
                    }
                }
                resource.unPersist(notifier, false);
            } else {
                PooledConnection pc = resource.getRepository().getConnectionManager().getConnection();
                try {
                    EMFSQLUtil.dbUpdateSingleAttribute(notifier, attr, pc.getConnection());
                } finally {
                    resource.getRepository().getConnectionManager().returnConnection(pc);
                }
            }
        } else if (feature instanceof EReference) {
            addToResource(notification.getNewValue());

            EReference ref = (EReference) feature;
            PooledConnection pc = resource.getRepository().getConnectionManager().getConnection();
            try {
                EMFSQLUtil.dbStoreReferences(notifier, ref, pc.getConnection());            
            } finally {
                resource.getRepository().getConnectionManager().returnConnection(pc);
            }
            if (ref.isContainment()) {
                EObject oldEObject = (EObject) notification.getOldValue();
                resource.unPersist(oldEObject, true);
            }
        }
    }

    private void notifyAdd(Notification notification) throws SQLException {
        Object feature = notification.getFeature();
        EObject notifier = (EObject)notification.getNotifier();
        PooledConnection pc = resource.getRepository().getConnectionManager().getConnection();
        try {
            if (feature instanceof EReference) {        
                if (notification.getNewValue() instanceof EObject) {
                    addToResource(notification.getNewValue());
                } else if (notification.getNewValue() instanceof List) {
                    for (Object o : (List<?>)notification.getNewValue()) {
                        addToResource(o);
                    }
                }

                EMFSQLUtil.dbStoreReferences(notifier, (EReference) feature, pc.getConnection());
            } else if (feature instanceof EAttribute) {
                EMFSQLUtil.dbStoreMultiValue(notifier, (EAttribute) feature, pc.getConnection());
            }
        } finally {
            resource.getRepository().getConnectionManager().returnConnection(pc);
        }
    }

    @SuppressWarnings("unchecked")
    private void addToResource(Object newValue) throws SQLException {
        if (newValue instanceof EObject) {
            EObject newEObject = (EObject) newValue;
            if (newEObject.eContainer() == null) {
                resource.getContents().add(newEObject);
            } else {
                resource.persist(newEObject);
            }
        }
    }
    
    private void notifyRemove(Notification notification) throws SQLException {
        Object feature = notification.getFeature();
        EObject notifier = (EObject)notification.getNotifier();
        PooledConnection pc = resource.getRepository().getConnectionManager().getConnection();
        try {
            if (feature instanceof EReference) {
                EReference reference = (EReference) feature;
                EMFSQLUtil.dbStoreReferences(notifier, reference, pc.getConnection());

                if (reference.isContainment()) {
                    if (notification.getOldValue() instanceof EObject) {
                        EObject value = (EObject) notification.getOldValue();
                        if (value.eContainer() == null && value.eResource() == null) {
                            resource.unPersist(value, true);
                        }
                    } else if (notification.getOldValue() instanceof List) {
                        for (Object o : (List<?>)notification.getOldValue()) {
                            if (o instanceof EObject) {
                                EObject value = (EObject) o;
                                resource.unPersist(value, true);
                            }
                        }
                    }
                }
            } else if (feature instanceof EAttribute) {
                EMFSQLUtil.dbStoreMultiValue(notifier, (EAttribute) feature, pc.getConnection());            
            }
        } finally {
            resource.getRepository().getConnectionManager().returnConnection(pc);
        }
    }

    public Notifier getTarget() {
        return null;
    }

    public void setTarget(Notifier newTarget) {
    }

    public boolean isAdapterForType(Object type) {
        return type.equals(EObject.class);
    }
}
