/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.jdbc;

import java.beans.IntrospectionException;
import java.beans.PropertyDescriptor;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import com.iona.soa.repository.EMFUtil;
import com.iona.soa.repository.ModelRegistry;
import com.iona.soa.repository.jdbc.EMFSQLUtil.HierarchyStrategy;
import com.iona.soa.repository.util.SQLHelper;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

/** This class contains the database code that can persist, read back and delete an EMF model to a database.
 * All SQL code is written according to SQL-92.
 * The class contains static utility methods only, the actual database persistence should be
 * achieved by using
 * the {@link com.iona.soa.repository.jdbc.resource.JDBCResourceFactoryImpl} as the resource factory.
 * <p/>
 * One useful method to be used outside of the resource factory is the one that creates all the SQL's
 * needed to create the database schema. See {@link #getSchemaCreationSQLs(EClass, String))}.
 */
public final class EMFSQLCreationUtil {
    private EMFSQLCreationUtil() {
    }
    
    ///
    /// --- Schema creation stuff here
    ///

    /** This method produces a list of SQL statements that can be issued against a database to create
     * the schema needed to persist an EMF model.
     * @param ePackage The EPackage for which to create a database schema.
     * @return A list of <tt>String</tt> SQL statements that should be executed against the database.
     * @throws SQLException
     */
    public static List<String> getSchemaCreationSQLs(EPackage ePackage, 
                                                     ModelRegistry registry, 
                                                     String dbName) {
        List<EClass> related = new LinkedList<EClass>();
        for (Object obj : ePackage.eContents()) {
            if (obj instanceof EClass) {
                related.add((EClass)obj);
            }
        }

        List<String> sqls = new LinkedList<String>();
        for (EClass cls : related) {
            sqls.addAll(getTableCreationSQL(cls, registry, dbName));
        }

        // Join tables are added separately here. They can only be created after the ordinary tables are done
        for (EClass cls : related) {
            sqls.addAll(getReferenceCreationSQL(cls, dbName));
        }

        tidyUpTableCreationSQLs(sqls);
        return sqls;
    }

    public static void tidyUpTableCreationSQLs(Collection<String> sqls) {
        // In certain cases empty lines were added, which we'll remove here.
        // Also, move the alter table statements to the end
        List<String> alterSQLs = new LinkedList<String>();
        for (Iterator<String> i = sqls.iterator(); i.hasNext();) {
            String sql = i.next();

            if ("".equals(sql)) {
                i.remove();
            } else if (sql.startsWith("ALTER")) {
                i.remove();
                alterSQLs.add(sql);
            }
        }
        sqls.addAll(alterSQLs);
    }

    public static void createAllModelTables(ModelRegistry registry, Connection conn) throws SQLException {
        for (EPackage ePackage : registry.getPackages()) {
            Collection<String> sqls = getSchemaCreationSQLs(ePackage, 
                                                            registry,
                                                            EMFSQLUtil.getDatabaseName(conn));
            Statement stmt = DBFactory.createStatement(conn);
            try {
                for (String sql : sqls) {
                    stmt.executeUpdate(sql);
                }
            } finally {
                stmt.close();
            }
        }
    }
    
    public static List<String> getTableCreationSQL(EClass eclass, ModelRegistry registry, String dbName) {
        // Annotation of table=none implies that the class is not for persistence, i.e.
        // it's an API class
        if (!EMFSQLUtil.isTableRequired(eclass)) {
            return Collections.emptyList();            
        }
        
        // skip sub-classes in a hierarchy with a top-class storage strategy,
        // they are handled at the hierarchy-top level
        EList<EClass> superTypes = eclass.getESuperTypes();
        if (superTypes.size() > 0) {
            if (EMFSQLUtil.getHierarchyStrategy(superTypes.get(0)) == HierarchyStrategy.TOP_CLASS) {
                return Collections.emptyList();
            }
        }

        if (!EMFSQLUtil.isTableClass(eclass)) {
            return Collections.emptyList();
        }

        List<String> sqls = new LinkedList<String>();
        StringBuilder sb = new StringBuilder();

        sb.append("CREATE TABLE ");
        sb.append(EMFSQLUtil.getTableName(eclass));
        sb.append(" (");

        appendPrimaryKey(eclass, sb, dbName);

        boolean addSubclassColumns = 
            eclass.isAbstract() || EMFSQLUtil.getHierarchyStrategy(eclass) == HierarchyStrategy.TOP_CLASS;
        Collection<EStructuralFeature> handledFeatures = new LinkedList<EStructuralFeature>();
        sb.append(addEClassColumns(eclass, registry, dbName, sqls, handledFeatures, addSubclassColumns));

        if (addSubclassColumns) {
            sb.append(", " + EMFSQLUtil.DISCRIMINATOR_COL_NAME + " VARCHAR(255)");
        }

        sb.append(")");
        sqls.add(sb.toString().toUpperCase());
        return sqls;
    }

    private static String addEClassColumns(EClass eclass, 
                                           ModelRegistry registry, 
                                           String dbName, 
                                           List<String> sqls, 
                                           Collection<EStructuralFeature> handledFeatures,
                                           boolean addSubclassColumns) {
        StringBuilder sb = new StringBuilder();
        for (EAttribute attr : (List<EAttribute>)eclass.getEAllAttributes()) {
            if (attr.isID() || handledFeatures.contains(attr)) {
                // The ID is already handled for the primary key
                continue;
            }
            
            handledFeatures.add(attr);

            if (attr.isMany()) {
                // Attributes that can have multiple values are stored as 1-many relations
                getMultiValueSubTableSQL(attr, sqls, dbName);
            } else {
                addColumnDef(sb, attr, dbName);
            }
        }
        sb.append(createOneToManySQL(eclass, sqls, dbName, handledFeatures));
        if (addSubclassColumns) {
            for (EClass subclass : EMFUtil.getSubClasses(eclass, registry)) {
                sb.append(addEClassColumns(subclass, registry, dbName, sqls, handledFeatures, true));
            }
        }
        return sb.toString();
    }
    
    public static List<String> getReferenceCreationSQL(EClass cls, String dbName) {
        // Annotation of table=none implies that the class is not for persistence, i.e.
        // it's an API class
        if (!EMFSQLUtil.isTableRequired(cls)) {
            return Collections.emptyList();            
        }
        
        List<String> sqls = new LinkedList<String>();
        for (EReference ref : (List<EReference>)cls.getEReferences()) {
            EReference opposite = ref.getEOpposite();
            if (opposite == null) {
                if (ref.isMany()) {
                    sqls.add(getOneToManyTableSQL(ref, dbName));
                }
            } else {
                if (ref.isMany() && opposite.isMany()) {
                    if (!EMFSQLUtil.isReverseOfJoinTableName(ref, opposite)) {
                        sqls.add(getManyToManyTableSQL(ref, opposite, dbName));
                    }
                }
            }
        }
        return sqls;
    }

    public static String getFKConstraint(String ... parts) {
        String[] allParts = new String[parts.length + 1];
        allParts[0] = "FK";
        System.arraycopy(parts, 0, allParts, 1, parts.length);
        return SQLHelper.getNonSQLKeyword(allParts);
    }

    static void getMultiValueSubTableSQL(EAttribute attr, Collection<String> sqls, String dbName) {
        EClass cls = (EClass)attr.eContainer();
        EAttribute idAttr = cls.getEIDAttribute();
        String tableName = EMFSQLUtil.getMultiValueTableName(cls, attr);

        StringBuilder sb = new StringBuilder();
        sb.append("CREATE TABLE ");
        sb.append(tableName);
        sb.append(" (");
        String field1 = EMFSQLUtil.getFKColumnName(idAttr);
        sb.append(field1);
        sb.append(" ");
        sb.append(EMFSQLUtil.getSQLDataType(idAttr, dbName));
        sb.append(" NOT NULL");
        addColumnDef(sb, attr, dbName);
        sb.append(", " + EMFSQLUtil.ORDERING_COL_NAME + " INTEGER NOT NULL");
        sb.append(")");
        sqls.add(sb.toString().toUpperCase());

        getMultiValueSubTableFkSQL(attr, sqls);
    }

    static void getMultiValueSubTableFkSQL(EAttribute attr, Collection<String> sqls) {
        EClass cls = (EClass)attr.eContainer();
        if (EMFSQLUtil.isTableClass(cls)) {
            EAttribute idAttr = cls.getEIDAttribute();
            StringBuilder sb = new StringBuilder();
            sb.append("ALTER TABLE " + EMFSQLUtil.getMultiValueTableName(cls, attr));
            sb.append(" ADD CONSTRAINT ");
            sb.append(getFKConstraint(EMFSQLUtil.getUnmangledTableName(cls), attr.getName()));
            sb.append(" FOREIGN KEY (" + EMFSQLUtil.getFKColumnName(idAttr) + ") ");
            sb.append(" REFERENCES " + EMFSQLUtil.getTableName(cls) + "(" + idAttr.getName() + ")");
            sqls.add(sb.toString().toUpperCase());
        }
    }
    
    private static String createOneToManySQL(EClass eclass, Collection<String> sqls, String dbName,
            Collection<EStructuralFeature> handledFeatures) {
        StringBuilder sb = new StringBuilder();
        for (EReference ref : EMFUtil.findOneToManyReferences(eclass)) {
            if (handledFeatures.contains(ref)) {
                continue;
            }
            handledFeatures.add(ref);

            sb.append(", ");
            sb.append(EMFSQLUtil.getFKColumnName(ref));
            sb.append(" ");
            sb.append(EMFSQLUtil.getSQLDataType(EMFUtil.findID(ref.getEReferenceType()), dbName));

            createOneToManyFkSQL(ref, eclass, sqls);
        }
        return sb.toString();
    }

    static void createOneToManyFkSQL(EReference ref, EClass eclass, Collection<String> sqls) {
        EClass refClass = ref.getEReferenceType();
        if (EMFSQLUtil.isTableClass(refClass)) {
            EAttribute refID = EMFUtil.findID(refClass);

            StringBuilder sb2 = new StringBuilder();
            sb2.append("ALTER TABLE " + EMFSQLUtil.getTableName(eclass));
            sb2.append(" ADD CONSTRAINT ");
            sb2.append(getFKConstraint(EMFSQLUtil.getUnmangledTableName(eclass), 
                                       EMFSQLUtil.normalizeReferenceName(ref)));
            sb2.append(" FOREIGN KEY (" + EMFSQLUtil.getFKColumnName(ref) + ") REFERENCES");
            sb2.append(" " + EMFSQLUtil.getTableName(refClass) + "(" + refID.getName() + ")");
            sqls.add(sb2.toString().toUpperCase());
        }
    }

    static void addColumnDef(StringBuilder sb, EAttribute attr, String dbName) {
        sb.append(", ");
        addSingleColumnDef(sb, attr, dbName);
    }

    static void addSingleColumnDef(StringBuilder sb, EAttribute attr, String dbName) {
        if (!EMFUtil.isBuiltInType(attr)) {

            try {
                String serializationHint = EMFUtil.getAnnotationValue(
                                                      attr.getEType(),
                                                      EMFSQLUtil.DBHINTS_ANNOTATION,
                                                      EMFSQLUtil.SERIALIZATION_KEY);

                if (EMFSQLUtil.SERIALIZATION_TO_STRING.equals(serializationHint)) {
                    sb.append(attr.getName() + " " + EMFSQLUtil.DEFAULT_SQL_TYPE);
                } else {
                    Class<?> instanceClass = attr.getEType().getInstanceClass();
                    List<PropertyDescriptor> pds = EMFSQLUtil.getPropertyDescriptors(instanceClass);
                    if (pds.size() > 0) {
                        boolean first = true;
                        for (PropertyDescriptor pd : pds) {
                            if (!first) {
                                sb.append(", ");
                            }
                            first = false;
                            sb.append(attr.getName() + "_" + pd.getName() + " "
                                      + EMFSQLUtil.getSQLDataType(pd.getPropertyType(), dbName));
                        }
                    } else {
                        sb.append(EMFSQLUtil.getColumnForAttribute(attr) + " " + EMFSQLUtil.DEFAULT_SQL_TYPE);
                    }
                }
            } catch (IntrospectionException e) {
                throw new RuntimeException(e);
            }

        } else {
            sb.append(EMFSQLUtil.getColumnForAttribute(attr));
            sb.append(" ");
            sb.append(EMFSQLUtil.getSQLDataType(attr, dbName));
        }
    }

    private static void appendPrimaryKey(EClass eclass, StringBuilder sb, String dbName) {
        EAttribute attr = EMFUtil.findID(eclass);

        sb.append(EMFSQLUtil.getColumnForAttribute(attr));
        sb.append(" ");
        sb.append(EMFSQLUtil.getSQLDataType(attr, dbName));
        sb.append(" NOT NULL PRIMARY KEY");
    }

    static String getOneToManyTableSQL(EReference ref, String dbName) {
        if (ref == null || !ref.isMany()) {
            return "";
        }

        StringBuilder sb = new StringBuilder();
        String tableName = EMFSQLUtil.getJoinTableName(ref);

        sb.append("CREATE TABLE ");
        sb.append(tableName);
        sb.append(" (");
        EAttribute refID = EMFUtil.findID(ref.getEReferenceType());
        String field1 = EMFSQLUtil.getFKColumnName(ref);
        sb.append(field1);
        sb.append(" ");
        sb.append(EMFSQLUtil.getSQLDataType(refID, dbName));
        sb.append(" NOT NULL, ");

        EAttribute id = EMFUtil.findID(ref.getEContainingClass());
        String field2 = EMFSQLUtil.getFKColumnName(id);
        sb.append(field2);
        sb.append(" ");
        sb.append(EMFSQLUtil.getSQLDataType(id, dbName));
        sb.append(" NOT NULL, ");

        sb.append(" CONSTRAINT ");
        sb.append(SQLHelper.getNonSQLKeyword("PK", tableName));
        sb.append(" PRIMARY KEY(");
        sb.append(field1);
        sb.append(", ");
        sb.append(field2);
        if (ref.isContainment()) {
            // if a one-to many relation is a containment one, this object should be the
            // only one to refer to the dependent objects
            sb.append("), CONSTRAINT ");
            sb.append(SQLHelper.getNonSQLKeyword("UQ", tableName));
            sb.append(" UNIQUE(");
            sb.append(field1);
        }
        sb.append(")");

        if (EMFSQLUtil.isTableClass(ref.getEReferenceType())) {
            sb.append(", CONSTRAINT ");
            sb.append(getFKConstraint(EMFSQLUtil.getUnmangledTableName(ref.getEContainingClass()), 
                                      id.getEContainingClass().getName(), 
                                      EMFSQLUtil.normalizeReferenceName(ref)));
            sb.append(" FOREIGN KEY (" + field1 + ") ");
            sb.append("REFERENCES ");
            sb.append(EMFSQLUtil.getTableName(ref.getEReferenceType()) + "(" + refID.getName() + ")");
        }

        if (EMFSQLUtil.isTableClass(ref.getEContainingClass())) {
            // Only store the foreign key to the containing class if its table is actually created.
            sb.append(", CONSTRAINT ");
            sb.append(getFKConstraint(EMFSQLUtil.getUnmangledTableName(ref.getEContainingClass()), 
                                      EMFSQLUtil.normalizeReferenceName(ref), 
                                      id.getEContainingClass().getName()));
            sb.append(" FOREIGN KEY (" + field2 + ") ");
            sb.append("REFERENCES " + EMFSQLUtil.getTableName(ref.getEContainingClass())
                      + "(" + id.getName() + ")");
        }
        sb.append(")");

        return sb.toString().toUpperCase();
    }

    static String getManyToManyTableSQL(EReference ref1, EReference ref2, String dbName) {
        if (ref1 == null || ref2 == null) {
            //          Containment is handled elsewhere
            return "";
        }

        if (!ref1.isMany() || !ref2.isMany()) {
            //          Only create SQL for many-to-many tables
            return "";
        }

        if (ref1.isContainment() || ref2.isContainment()) {
            return "";
        }

        StringBuilder sb = new StringBuilder();
        String tableName = EMFSQLUtil.getJoinTableName(ref1, ref2);

        sb.append("CREATE TABLE ");
        sb.append(tableName);
        sb.append(" (");
        EAttribute id1 = EMFUtil.findID(ref1.getEReferenceType());
        String field1 = EMFSQLUtil.getFKColumnName(ref1);
        sb.append(field1);
        sb.append(" ");
        sb.append(EMFSQLUtil.getSQLDataType(id1, dbName));
        sb.append(" NOT NULL, ");

        EAttribute id2 = EMFUtil.findID(ref2.getEReferenceType());
        String field2 = EMFSQLUtil.getFKColumnName(ref2);
        sb.append(field2);
        sb.append(" ");
        sb.append(EMFSQLUtil.getSQLDataType(id2, dbName));
        sb.append(" NOT NULL, ");

        sb.append(" CONSTRAINT ");
        sb.append(SQLHelper.getNonSQLKeyword("PK", tableName));
        sb.append(" PRIMARY KEY (");
        sb.append(field1);
        sb.append(", ");
        sb.append(field2);
        sb.append(")");
        if (EMFSQLUtil.isTableClass(ref1.getEReferenceType())) {
            sb.append(", CONSTRAINT ");
            sb.append(getFKConstraint(EMFSQLUtil.getUnmangledTableName(ref1.getEContainingClass()), 
                                      EMFSQLUtil.normalizeReferenceName(ref2), 
                                      EMFSQLUtil.normalizeReferenceName(ref1)));
            sb.append(" FOREIGN KEY (" + field1 + ") ");
            sb.append("REFERENCES ");
            sb.append(EMFSQLUtil.getTableName(ref1.getEReferenceType()) + "(" + id1.getName() + ")");
        }
        if (EMFSQLUtil.isTableClass(ref2.getEReferenceType())) {
            sb.append(", CONSTRAINT ");
            sb.append(getFKConstraint(EMFSQLUtil.getUnmangledTableName(ref2.getEContainingClass()), 
                                      EMFSQLUtil.normalizeReferenceName(ref1), 
                                      EMFSQLUtil.normalizeReferenceName(ref2)));
            sb.append(" FOREIGN KEY (" + field2 + ") ");
            sb.append("REFERENCES ");
            sb.append(EMFSQLUtil.getTableName(ref2.getEReferenceType()) + "(" + id2.getName() + ")");
        }
        sb.append(")");

        return sb.toString().toUpperCase();
    }

}
