/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.jdbc;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.sql.Clob;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import com.iona.soa.repository.EMFUtil;
import com.iona.soa.repository.ModelRegistry;
import com.iona.soa.repository.STPRuntimeException;
import com.iona.soa.repository.jdbc.resource.JDBCResource;
import com.iona.soa.repository.jdbc.resource.JDBCResourceImpl;
import com.iona.soa.repository.util.InputStreamHelper;
import com.iona.soa.repository.util.Out;
import com.iona.soa.repository.util.SQLHelper;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAnnotation;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.InternalEObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;

/** This class contains the database code that can persist, read back and delete an EMF model to a database.
 * All SQL code is written according to SQL-92.
 * The class contains static utility methods only, the actual database persistence should be
 * achieved by using
 * the {@link com.iona.soa.repository.jdbc.resource.JDBCResourceFactoryImpl} as the resource factory.
 * <p/>
 * One useful method to be used outside of the resource factory is the one that creates all the SQL's
 * needed to create the database schema. See {@link #getSchemaCreationSQLs(EClass, String))}.
 */
public final class EMFSQLUtil {
    public static final String DISCRIMINATOR_COL_NAME = "CLASS_";
    public static final String ORDERING_COL_NAME = "ORDNR_";

    static final String DBHINTS_ANNOTATION = "dbhints";
    static final String SERIALIZATION_KEY = "serialization";
    static final String SERIALIZATION_TO_STRING = "toString";
    static final String DEFAULT_SQL_TYPE = "VARCHAR(255)";

    public enum HierarchyStrategy {
        TOP_CLASS, SUB_CLASS
    };

    private EMFSQLUtil() {
    }

    static String getMultiValueTableName(EClass cls, EAttribute attr) {
        return SQLHelper.getNonSQLKeyword(getUnmangledTableName(cls), attr.getName());
    }

    public static String getColumnForAttribute(EAttribute attr) {
        return SQLHelper.getNonSQLKeyword(attr.getName());
    }

    public static boolean isTableRequired(EClass cls) {
        String hint = EMFUtil.getAnnotationValue(cls, DBHINTS_ANNOTATION, "table");
        if (hint != null) {
            return !"none".equals(hint);
        }
        return true;
    }
        
    private static boolean isClobType(EAttribute attribute) {
        EAnnotation annotation = attribute.getEAnnotation(DBHINTS_ANNOTATION);
        if (annotation != null) {
            Object type = annotation.getDetails().get("type");
            if (type instanceof String) {
                return "CLOB".equalsIgnoreCase((String)type);
            }
        }
        return false;
    }

    static String getSQLDataType(EAttribute attribute, String dbName) {
        EAnnotation annotation = attribute.getEAnnotation(DBHINTS_ANNOTATION);
        if (annotation != null) {
            Object type = annotation.getDetails().get("type");
            if (type instanceof String) {
                return handleDialect((String)type, dbName);
            }
        }
        return getSQLDataType(attribute.getEType().getInstanceClass(), dbName);
    }

    static String getSQLDataType(Class<?> attrClass, String dbName) {
        // TODO cater for other non-string data types too!
        if (attrClass.equals(Boolean.class) || attrClass.equals(boolean.class)) {
            return dbName.contains("POSTGRESQL") ? "BOOLEAN" : "VARCHAR(5)";
        } else if (attrClass.equals(Long.class) || attrClass.equals(long.class)) {
            return "NUMERIC(19,0)";
        } else if (attrClass.equals(Integer.class) || attrClass.equals(int.class)) {
            return "INTEGER";
        } else if (attrClass.equals(Short.class) || attrClass.equals(short.class)) {
            return "SMALLINT";
        } else if (Date.class.isAssignableFrom(attrClass)) {
            return "TIMESTAMP";
        }

        return DEFAULT_SQL_TYPE;
    }

    public static String handleDialect(String type, String dbName) {
        if ("CLOB".equals(type.toUpperCase())) {
            if (dbName.contains("HSQL")) {
                return "LONGVARCHAR";
            } else if (dbName.contains("MYSQL") || dbName.contains("POSTGRESQL")) {
                return "TEXT";
            }
        }
        return type;
    }

    public static HierarchyStrategy getHierarchyStrategy(EClass cls) {
        return getHierarchyStrategy(cls, new Out<EClass>());
    }

    public static HierarchyStrategy getHierarchyStrategy(EClass cls, Out<EClass> definingClass) {
        String hint = EMFUtil.getAnnotationValue(cls, DBHINTS_ANNOTATION, "hierarchy");
        if (hint != null) {
            definingClass.set(cls);
            return "subclass".equals(hint) ? HierarchyStrategy.SUB_CLASS : HierarchyStrategy.TOP_CLASS;
        }

        EList<EClass> superTypes = cls.getESuperTypes();
        if (superTypes.size() > 0) {
            return getHierarchyStrategy(superTypes.get(0), definingClass);
        } else {
            // the default
            definingClass.set(null);
            return HierarchyStrategy.TOP_CLASS;
        }
    }

    /** Returns the table name of the table in which a particular class is stored.
     * The returned string is mangled to not include any SQL keywords nd not extend
     * the maximum length for tables and keys
     * @param cls The class to return the table name for.
     * @return The name of the table.
     */
    public static String getTableName(EClass cls) {
        return SQLHelper.getNonSQLKeyword(getUnmangledTableName(cls));
    }

    /** Returns the table name of the table in which a particular class is stored.
     * @param cls The class to return the table name for.
     * @return The name of the table.
     */
    static String getUnmangledTableName(EClass cls) {
        // davidb TODO, need to improve for bigger hierarchies, probably
        Out<EClass> definingClass = new Out<EClass>();
        if (getHierarchyStrategy(cls, definingClass) == HierarchyStrategy.TOP_CLASS) {
            if (definingClass.get() != null) {
                return definingClass.get().getName().toUpperCase();
            } else {
                for (Iterator<EClass> i = cls.getEAllSuperTypes().iterator(); i.hasNext();) {
                    return i.next().getName().toUpperCase();
                }
            }
        }
        return cls.getName().toUpperCase();
    }
    
    public static boolean isTableClass(EClass eClass) {
        return eClass != null 
            && (EMFSQLUtil.getHierarchyStrategy(eClass) == EMFSQLUtil.HierarchyStrategy.TOP_CLASS 
                || !eClass.isAbstract());
    }

    public static boolean isTableHierarchy(EClass cls) {
        Out<EClass> definingClass = new Out<EClass>();
        // a non abstract classes can also have a "top-class" strategy
        if (getHierarchyStrategy(cls, definingClass) == HierarchyStrategy.TOP_CLASS
            && null != definingClass.get()
            && definingClass.get().getName().equals(cls.getName())) {
            return true;
        }
        String tblName = SQLHelper.cleanSQLKeyword(getTableName(cls));
        return !tblName.equalsIgnoreCase(cls.getName());
    }
    
    public static Collection<EClass> getTableClasses(EClass eClass, ModelRegistry registry) {
        Collection<EClass> ret = new LinkedList<EClass>();

        Out<EClass> definingClass = new Out<EClass>();
        if (EMFSQLUtil.getHierarchyStrategy(eClass, definingClass) 
                == EMFSQLUtil.HierarchyStrategy.TOP_CLASS) {
            if (definingClass.get() != null) {
                ret.add(definingClass.get());
            }
        } else {
            if (!eClass.isAbstract()) {
                ret.add(eClass);
            }
            for (EClass subClass : EMFUtil.getSubClasses(eClass, registry)) {
                ret.addAll(getTableClasses(subClass, registry));
            }
        }
        return ret;
    }

    public static String getJoinTableName(EReference ref) {
        return SQLHelper.getNonSQLKeyword(
                getUnmangledTableName(ref.getEContainingClass()), normalizeReferenceName(ref));
    }

    public static String getJoinTableName(EReference ref, EReference opposite) {
        String tbl1 = normalizeReferenceName(ref);
        String tbl2 = normalizeReferenceName(opposite);

        // Ensure that regardless of the order passed in, the join table name will always be the same
        String str1;
        String str2;
        if (tbl1.compareTo(tbl2) < 0) {
            str1 = tbl1;
            str2 = tbl2;
        } else {
            str1 = tbl2;
            str2 = tbl1;
        }

        return SQLHelper.getNonSQLKeyword(str1, "TO", str2);
    }

    private static String bareJoinTableName(EReference containingClass, EReference containingClass2) {
        return SQLHelper.getNonSQLKeyword(normalizeReferenceName(containingClass), 
                                          "TO", 
                                          normalizeReferenceName(containingClass2));
    }

    public static EStructuralFeature normalizeReference(EReference ref) {
        if (isReversed(ref)) {
            return ref.getEOpposite();
        } else {
            return ref;
        }
    }
    
    public static boolean isOneOneRef(EReference ref) {
        EReference opposite = ref.getEOpposite();
        if (opposite == null || ref.isMany()) {
            return false;
        } else {
            return !opposite.isMany();
        }
    }
    
    public static boolean isReversed(EReference ref) {
        EReference opposite = ref.getEOpposite();
        if (opposite == null) {
            return false;
        } else if (ref.isMany()) {
            if (opposite.isMany()) {
                return isReverseOfJoinTableName(ref, opposite);
            } else {
                return true;
            }
        } else {
            return false;
        }
    }

    static boolean isReverseOfJoinTableName(EReference containingClass, EReference containingClass2) {
        return !bareJoinTableName(containingClass, containingClass2)
            .equals(getJoinTableName(containingClass, containingClass2));
    }
    
    public static String getFKColumnName(EAttribute attr) {
        return SQLHelper.getNonSQLKeyword(getTableName(attr.getEContainingClass()));
    }

    public static String getFKColumnName(EReference ref) {
        return SQLHelper.getNonSQLKeyword(normalizeReferenceName(ref));
    }

    static String normalizeReferenceName(EReference ref) {
        String refName = ref.getName().toUpperCase();
        if (refName.endsWith("S")) {
            refName = refName.substring(0, refName.length() - 1);
            if (refName.endsWith("IE")) {
                refName = refName.substring(0, refName.length() - 2) + "Y";
            }
        }
        return refName;
    }

    ///
    /// --- Schema creation stuff (moved to EMFSQLCreationUtil). Stubs left here for backward compatibility
    ///
    
    public static void createAllModelTables(ModelRegistry registry, Connection conn) throws SQLException {
        EMFSQLCreationUtil.createAllModelTables(registry, conn);
    }

    public static List<String> getSchemaCreationSQLs(EPackage ePackage, 
                                                     ModelRegistry registry, 
                                                     String dbName) {
        return EMFSQLCreationUtil.getSchemaCreationSQLs(ePackage, registry, dbName);
    }
    
    ///
    /// --- Storing
    ///

    /**
     * Stores a reference or the references of an EObject. When a reference is 1-many or many-many
     * all references are stored.
     * @param eobject The object for which the referenc to store.
     * @param ref The reference to store
     * @param connection The JDBC connection to use.
     * @return <tt>true</tt> if any changes were made to the database. In certain cases (esp. bidirectional
     * associations) database changes are only performed on one of the directions to avoid double-updates
     * in the database.
     * @throws SQLException in case of trouble with the database.
     */
    @SuppressWarnings("unchecked")
    public static void dbStoreReferences(EObject eobject, EReference ref, Connection connection)
        throws SQLException {
        EReference opposite = ref.getEOpposite();

        if (opposite == null) {
            if (!ref.isMany()) {
                dbStoreReference(eobject, ref, connection);
            } else {
                dbStoreOneToManyReference(eobject, ref, connection);
            }
        } else {
            if (!ref.isMany()) {
                dbStoreReference(eobject, ref, connection);
            } else if (opposite.isMany()) {
                dbStoreManyToManyReferences(eobject, ref, opposite, connection);
            }
        }
    }

    private static boolean dbStoreReference(EObject eobject, EReference ref, Connection connection)
        throws SQLException {
        EAttribute id = EMFUtil.findID(eobject.eClass());
        Object idValue = eobject.eGet(id);

        EObject other = (EObject)eobject.eGet(ref);
        Object otherIdValue = null;
        if (other != null) {
            EAttribute otherID = EMFUtil.findID(other.eClass());
            otherIdValue = other.eGet(otherID);
        }

        String sql = "UPDATE " + getTableName(eobject.eClass()) + " SET " + getFKColumnName(ref)
                     + "= ? WHERE " + id.getName() + " = ?";
        PreparedStatement ps = DBFactory.prepareStatement(connection, sql);
        try {
            setValueOnPreparedStatement(connection, ps, 1, otherIdValue);
            setValueOnPreparedStatement(connection, ps, 2, idValue);
            int updated = ps.executeUpdate();
            if (updated != 1) {
                throw new STPRuntimeException("Incorrect number of records updated " + updated + " SQL:"
                                              + sql);
            }
            return true;
        } finally {
            ps.close();
        }
    }

    @SuppressWarnings("unchecked")
    private static boolean dbStoreOneToManyReference(EObject eobject, EReference ref, Connection connection)
        throws SQLException {
        boolean prevAutoCommit = connection.getAutoCommit();
        try {
            if (prevAutoCommit) {
                connection.setAutoCommit(false);
            }
            final EAttribute id = EMFUtil.findID(eobject.eClass());

            StringBuilder sb = new StringBuilder("DELETE FROM " + getJoinTableName(ref));
            sb.append(" WHERE " + getFKColumnName(id) + " = ?");
            PreparedStatement ps = DBFactory.prepareStatement(connection, sb.toString());
            setValueOnPreparedStatement(connection, ps, 1, eobject.eGet(id));
            ps.executeUpdate();
            ps.close();

            StringBuilder sb2 = new StringBuilder("INSERT INTO " + getJoinTableName(ref) + "(");
            sb2.append(getFKColumnName(id));
            sb2.append(",");
            sb2.append(getFKColumnName(ref));
            sb2.append(") VALUES (?,?)");
            for (EObject refObj : (List<EObject>)eobject.eGet(ref)) {
                PreparedStatement ps2 = DBFactory.prepareStatement(connection, sb2.toString());
                setValueOnPreparedStatement(connection, ps2, 1, eobject.eGet(id));
                setValueOnPreparedStatement(connection, ps2, 2, refObj.eGet(EMFUtil.findID(refObj.eClass())));
                ps2.executeUpdate();
                ps2.close();
            }

            if (prevAutoCommit) {
                connection.commit();
            }
            return true;
        } catch (SQLException e) {
            if (prevAutoCommit) {
                connection.rollback();
            }
            throw e;
        } finally {
            if (prevAutoCommit) {
                connection.setAutoCommit(prevAutoCommit);
            }
        }
    }

    @SuppressWarnings("unchecked")
    private static boolean dbStoreManyToManyReferences(EObject eobject, EReference ref, EReference opposite,
                                                       Connection connection) throws SQLException {
        // only store one direction of the references
        if (isReverseOfJoinTableName(ref, opposite)) {
            // If this is a 2-way relation it only needs to be put in the database once
            // return false to inform that no changes were made
            return false;
        }

        Object objectID = eobject.eGet(EMFUtil.findID(eobject.eClass()));
        boolean prevAutoCommit = connection.getAutoCommit();
        try {
            if (prevAutoCommit) {
                connection.setAutoCommit(false);
            }

            // delete the current references first
            StringBuilder sb = new StringBuilder("DELETE FROM " + getJoinTableName(ref, opposite));
            sb.append(" WHERE " + getFKColumnName(opposite) + " = ?");
            PreparedStatement ps = DBFactory.prepareStatement(connection, sb.toString());
            setValueOnPreparedStatement(connection, ps, 1, objectID);
            ps.executeUpdate();
            ps.close();

            StringBuilder sb2 = new StringBuilder("INSERT INTO " + getJoinTableName(ref, opposite) + "(");
            sb2.append(getFKColumnName(opposite));
            sb2.append(",");
            sb2.append(getFKColumnName(ref));
            sb2.append(") VALUES (?,?)");
            for (EObject refObj : (List<EObject>)eobject.eGet(ref)) {
                PreparedStatement ps2 = DBFactory.prepareStatement(connection, sb2.toString());
                setValueOnPreparedStatement(connection, ps2, 1, objectID);
                setValueOnPreparedStatement(connection, ps2, 2, refObj.eGet(EMFUtil.findID(refObj.eClass())));
                ps2.executeUpdate();
                ps2.close();
            }

            if (prevAutoCommit) {
                connection.commit();
            }
            return true;
        } catch (SQLException e) {
            if (prevAutoCommit) {
                connection.rollback();
            }
            throw e;
        } finally {
            if (prevAutoCommit) {
                connection.setAutoCommit(prevAutoCommit);
            }
        }
    }

    ///
    /// --- Loading Code
    ///

    /** Load a particular object from the resource. This method is generally called internally
     * by the framework
     * to resolve proxy objects. The object will be added to the resource's contents.
     * @param resource The resource to load from.
     * @param className The class name of the object.
     * @param id The ID, this is the value of the ID attribute for the object.
     * @param conn The database connection to use.
     * @return The object loaded, or <tt>null</tt> if no such object can be found.
     * @throws SQLException in case of trouble with the database.
     */
    public static EObject loadEObject(JDBCResource resource, String className, String id, Connection conn)
        throws SQLException {
        EClass cls = resource.getRepository().getModelRegistry().getEClass(className);
        if (cls == null) {
            return null;
        }
        return loadEObject(resource, cls, id, conn);
    }
    
    /** Load a particular object from the resource. This method is generally called internally
     * by the framework
     * to resolve proxy objects. The object will be added to the resource's contents.
     * @param resource The resource to load from.
     * @param className The class name of the object.
     * @param id The ID, this is the value of the ID attribute for the object.
     * @param conn The database connection to use.
     * @return The object loaded, or <tt>null</tt> if no such object can be found.
     * @throws SQLException in case of trouble with the database.
     */
    public static EObject loadEObject(JDBCResource resource, EClass eCls, String id, Connection conn)
        throws SQLException {
        String sql = getSelectSQL(getTableName(eCls), eCls.getEIDAttribute().getName(), null, null, null);
        PreparedStatement ps = DBFactory.prepareStatement(conn, sql);
        try {
            setValueOnPreparedStatement(conn, ps, 1, id);
            ResultSet rs = ps.executeQuery();
            if (rs.next()) {
                return loadEObject(eCls, resource, rs, conn);
            } else {
                return null;
            }
        } finally {
            ps.close();
        }
    }

    public static EObject loadEObject(JDBCResource resource, String className,
                                      String id, String ver, Connection conn) throws SQLException {
        EClass cls = resource.getRepository().getModelRegistry().getEClass(className);
        if (cls == null) {
            return null;
        }
        EStructuralFeature ef = EMFUtil.getGenerationNumberAttribute(cls);
        if (null == ef) {
            return null;
        }
        String sql = getSelectSQL(getTableName(cls), cls.getEIDAttribute().getName(), null, null, null);
        sql += " AND " + ef.getName() + "= ?";
        PreparedStatement ps = DBFactory.prepareStatement(conn, sql);
        try {
            setValueOnPreparedStatement(conn, ps, 1, id);
            setValueOnPreparedStatement(conn, ps, 2, ver);
            ResultSet rs = ps.executeQuery();
            if (rs.next()) {
                return loadEObject(cls, resource, rs, conn);
            } else {
                return null;
            }
        } finally {
            ps.close();
        }
    }

    static Object prepForESet(EAttribute attr, Object val) throws SQLException {
        if (val == null) {
            if (attr != null) {
                if (boolean.class.equals(attr.getEType().getInstanceClass())) {
                    return attr.getDefaultValue();
                }
            }
            return null;
        } else if (val instanceof BigDecimal) {
            val = val.toString();
        } else if (attr.getEType() instanceof EEnum) {
            return attr.getEContainingClass().getEPackage().
                getEFactoryInstance().createFromString(
                    (EEnum)attr.getEType(), (String)val);
        } else if ("oracle.sql.TIMESTAMP".equals(val.getClass().getName())) {
            try {
                return val.getClass().getDeclaredMethod("timestampValue").invoke(val);
            } catch (InvocationTargetException ite) {
                if (ite.getTargetException() instanceof SQLException) {
                    throw (SQLException)ite.getTargetException();
                } else {
                    throw new STPRuntimeException(ite.getTargetException());
                }
            } catch (Exception ex) {
                throw new STPRuntimeException(ex);
            }
        }


        Class<?> clazz = attr.getEType().getInstanceClass();

        if (boolean.class.equals(clazz)) {
            if ("1".equals(val)) {
                val = "true";
            }
            val = createInstance(Boolean.class, val);
        } else if (byte.class.equals(clazz)) {
            val = createInstance(Byte.class, val);
        } else if (long.class.equals(clazz)) {
            val = createInstance(Long.class, val);
        } else if (short.class.equals(clazz)) {
            val = createInstance(Short.class, val);
        } else if (Number.class.isAssignableFrom(clazz)) {
            val = createInstance(clazz, val);            
        }
        
        return val;
    }

    private static Object createInstance(Class<?> clazz, Object val) {
        if (clazz.isInstance(val)) {
            return val;
        }
        try {
            Constructor<?> ctor = clazz.getConstructor(new Class[] {val.getClass()});
            return ctor.newInstance(new Object[] {val});
        } catch (Exception e) {
            if (val instanceof String) {
                throw new STPRuntimeException("Could not recreate instance of " + clazz.getName() + ": "
                                              + val, e);
            } else {
                return createInstance(clazz, val.toString());
            }
        }
    }

    private static EObject loadEObject(EClass eClass, JDBCResource resource, ResultSet rs,
                                       Connection connection) throws SQLException {
        EClass actualClass = getActualClass(rs, eClass, resource);
        if (!eClass.isAbstract() && !actualClass.equals(eClass)) {
            return null;
        }
        EAttribute idAttr = actualClass.getEIDAttribute();
        String id = EcoreUtil.convertToString(idAttr.getEAttributeType(), rs.getObject(idAttr.getName()));
        URI proxyURI = EMFUtil.getProxyURI(actualClass, id);

        // loadObject is thread unsafe as it modifies the object in place in few methods
        // but addToResource(resource, ret); is also unsafe if two, possibly the same objects are attempted
        // to be added to the resource at the same time, so the lock need to be atomic
        // it doesn't seem right...
        synchronized (EMFSQLUtil.class) {
            EObject ret = loadEObject(proxyURI, eClass, resource, rs, connection);
            dbLoadContainer(ret, resource, connection);
            addToResource(resource, ret);
            return ret;
        }
    }

    private static void addToResource(JDBCResource resource, EObject eObj) {
        if (eObj.eContainer() != null) {
            addToResource(resource, eObj.eContainer());
        } else if (eObj.eResource() == null) {
            boolean oldDeliver = resource.eDeliver();
            try {
                resource.eSetDeliver(false);
                resource.getContents().add(eObj);
            } finally {
                resource.eSetDeliver(oldDeliver);
            }
        }
    }

    @SuppressWarnings("unchecked")
    private static EObject loadEObject(URI proxyURI, EClass eClass, JDBCResource resource, ResultSet rs,
                                       Connection connection) throws SQLException {

        EObject obj = resource.findEObject(proxyURI);

        if (obj == null) {
            obj = EMFUtil.createEObject(eClass, null);
        }

        boolean prevDeliver = obj.eDeliver();
        try {
            //          disable notifications while we're initializing the object,
            //          they are enabled back in the finally block

            obj.eSetDeliver(false);
            List<EAttribute> multiValues = new LinkedList<EAttribute>();
            for (EAttribute attr : (List<EAttribute>)obj.eClass().getEAllAttributes()) {
                if (attr.isMany()) {
                    multiValues.add(attr);
                } else if (!EMFUtil.isBuiltInType(attr)) {
                    Object co = createCustomObject(attr, rs);
                    obj.eSet(attr, co);
                } else {
                    Object val = rs.getObject(getColumnForAttribute(attr));

                    if (val instanceof Clob) {
                        val = readClob((Clob) val);
                    }

                    obj.eSet(attr, prepForESet(attr, val));
                }
            }

            for (EAttribute mvAttr : multiValues) {
                List<Object> l = (List<Object>)obj.eGet(mvAttr);
                l.clear();
                l.addAll(dbLoadMultiValue(obj, mvAttr, connection));
            }

            boolean isProxy = obj.eIsProxy();
            if (isProxy && obj instanceof InternalEObject) {
                ((InternalEObject)obj).eSetProxyURI(null);
            }

            makeAutoPersistent(obj, resource);

            if (resource instanceof JDBCResourceImpl) {
                ((JDBCResourceImpl)resource).registerEObject(proxyURI, obj);
            }

            EMFSQLUtil.dbLoadRelatedObjects(obj, resource, connection, isProxy);

            return obj;
        } finally {
            //          enable notifications (if they were originally enabled)
            obj.eSetDeliver(prevDeliver);
        }
    }

    static String readClob(Clob clob) throws SQLException {
        InputStream is = clob.getAsciiStream();
        try {
            return new String(InputStreamHelper.drain(is));
        } catch (IOException e) {
            SQLException se = new SQLException("Problem reading clob from database.");
            e.initCause(e);
            throw se;
        } finally {
            InputStreamHelper.close(is);
        }
    }

    private static Object createCustomObject(EAttribute attr, ResultSet rs) throws SQLException {

        String serializationHint = EMFUtil.getAnnotationValue(
                                                attr.getEType(),
                                                DBHINTS_ANNOTATION,
                                                SERIALIZATION_KEY);
        try {
            Object co = null;
            boolean valueSet = false;
            Class<?> cls = attr.getEType().getInstanceClass();

            if (SERIALIZATION_TO_STRING.equals(serializationHint)) {
                Object val = rs.getObject(attr.getName());
                if (val != null) {
                    Constructor<?> ctr = cls.getConstructor(new Class[]{String.class});
                    if (ctr != null) {
                        co = ctr.newInstance(new Object[]{val.toString()});
                        valueSet = true;
                    }
                }
            } else {
                List<PropertyDescriptor> pds = getPropertyDescriptors(cls);
                if (pds.size() > 0) {
                    co = cls.newInstance();
                    for (PropertyDescriptor pd : pds) {
                        Object val = rs.getObject(attr.getName() + "_" + pd.getName());
                        if (val == null) {
                            continue;
                        }
                        Method wm = pd.getWriteMethod();
                        wm.invoke(co, new Object[] {val});
                        valueSet = true;
                    }
                } else {
                    Object val = rs.getObject(attr.getName());
                    if (val != null) {
                        Constructor<?> ctr = cls.getConstructor(new Class[]{String.class});
                        if (ctr != null) {
                            co = ctr.newInstance(new Object[]{val.toString()});
                            valueSet = true;
                        }
                    }
                }
            }
            //If no data for the object is found, just return null
            return valueSet ? co : null;
        } catch (Exception e) {
            SQLException se = new SQLException(e.getMessage());
            se.initCause(e);
            throw se;
        }
    }

    private static void makeAutoPersistent(EObject obj, Resource resource)
        throws SQLException {
        if (resource instanceof JDBCResourceImpl) {
            JDBCResourceImpl jdbcResource = (JDBCResourceImpl)resource;
            jdbcResource.addSerializationAdapter(obj, false);
        }
    }

    private static Collection<Object> dbLoadMultiValue(EObject obj, EAttribute mvAttr, Connection connection)
        throws SQLException {
        EAttribute id = EMFUtil.findID(obj.eClass());
        Object objectID = obj.eGet(id);
        // eClass may be different to obj.eClass() as attribute may be defined in super-class
        EClass eClass = (EClass)mvAttr.eContainer();

        String sql = getSelectSQL(getMultiValueTableName(eClass, mvAttr), 
                                  getFKColumnName(id), null, null, null);
        sql +=  " ORDER BY " + ORDERING_COL_NAME;
        PreparedStatement ps = DBFactory.prepareStatement(connection, sql);
        try {
            setValueOnPreparedStatement(connection, ps, 1, objectID);

            List<Object> values = new LinkedList<Object>();
            ResultSet rs = ps.executeQuery();
            while (rs.next()) {
                if (!EMFUtil.isBuiltInType(mvAttr)) {
                    values.add(createCustomObject(mvAttr, rs));
                } else {
                    values.add(prepForESet(mvAttr, rs.getObject(mvAttr.getName())));
                }
            }
            return values;
        } finally {
            ps.close();
        }
    }

    private static void dbLoadRelatedObjects(EObject eObject, JDBCResource resource, Connection connection,
                                             boolean forceLoadReferences) throws SQLException {
        for (EReference ref : (List<EReference>)eObject.eClass().getEAllReferences()) {
            //          avoid loading relation if its already loaded
            if (!eObject.eIsSet(ref) || forceLoadReferences) {
                EReference opposite = ref.getEOpposite();
                if (opposite == null) {
                    if (!ref.isMany()) {
                        dbLoadRelatedObject(eObject, ref, resource, connection);
                    } else {
                        dbLoadJoinTableOneToManyObjects(eObject, ref, resource, connection);
                    }
                } else {
                    if (ref.isMany()) {
                        if (opposite.isMany()) {
                            dbLoadJoinTableMany2ManyObjects(eObject, ref, opposite, resource, connection);
                        } else {
                            dbLoadRelatedOneToManyObjects(eObject, ref, opposite, resource, connection);
                        }
                    } else {
                        dbLoadRelatedObject(eObject, ref, resource, connection);
                    }
                }
            }
        }
    }

    private static void dbLoadRelatedObject(EObject object, EReference ref, JDBCResource resource,
                                            Connection connection) throws SQLException {
        Collection<EClass> relTableClasses;
        if (isTableClass(ref.getEReferenceType())) {
            relTableClasses = Collections.singleton(ref.getEReferenceType());
        } else {
            relTableClasses = getTableClasses(ref.getEReferenceType(), 
                                              resource.getRepository().getModelRegistry());
        }
        
        for (EClass relClass : relTableClasses) {
            // TODO, in this case (and maybe others) the key to the foreign object is obtained
            // with a SELECT. We actually have this Foreign Key already, so an optimization would
            // to instantly create a proxy object from that without the additional SELECT.
            String sql = getSelectSQL(getTableName(relClass), EMFUtil.findID(relClass).getName(), 
                                      getTableName(object.eClass()), getFKColumnName(ref), 
                                      EMFUtil.findID(object.eClass()).getName());
            PreparedStatement ps = DBFactory.prepareStatement(connection, sql);

            try {
                setValueOnPreparedStatement(connection, ps, 1, object.eGet(EMFUtil.findID(object.eClass())));
                ResultSet rs = ps.executeQuery();
                dbLoadRelatedObjects(rs, object, ref, relClass, resource, connection);
            } finally {
                ps.close();
            }
        }
    }

    private static void dbLoadRelatedObjects(ResultSet rs, EObject eobject, EReference ref,
                                             EClass relClass, JDBCResource resource, Connection connection)
        throws SQLException {

        while (rs.next()) {
            EClass actualClass = getActualClass(rs, relClass, resource);
            EAttribute idAttr = actualClass.getEIDAttribute();

            EObject relObj = null;
            String id = EcoreUtil.convertToString(idAttr.getEAttributeType(), rs.getObject(idAttr.getName()));
            URI proxyURI = EMFUtil.getProxyURI(actualClass, id);

            if (ref.isContainment()) {
                // If the object is contained, load the actual object
                relObj = loadEObject(proxyURI, actualClass, resource, rs, connection);
            } else {
                if (resource instanceof JDBCResourceImpl) {
                    relObj = ((JDBCResourceImpl)resource).findEObject(proxyURI);
                }
                if (relObj == null) {
                    // If the object an ordinary reference, just create a proxy object.
                    relObj = EMFUtil.createEObject(actualClass, id);
                    ((InternalEObject)relObj).eSetProxyURI(proxyURI);
                    if (resource instanceof JDBCResourceImpl) {
                        ((JDBCResourceImpl)resource).registerEObject(proxyURI, relObj);
                    }
                }
            }

            boolean prevDeliver = relObj.eDeliver();
            try {
                // disable notification of events since this is not really a modification,
                // we're just loading the initial state here
                relObj.eSetDeliver(false);

                EMFUtil.addReferenceToMemoryObject(eobject, ref, relObj);
            } finally {
                relObj.eSetDeliver(prevDeliver);
            }
        }
        rs.close();
    }

    private static void dbLoadContainer(EObject eObject, JDBCResource resource,
        Connection connection) throws SQLException {
        Map<EClass, Collection<EReference>> backRefs = 
            EMFUtil.getBackReferences(eObject.eClass(), resource.getRepository().getModelRegistry());
        
        for (EClass eClass : backRefs.keySet()) {
            if (!eClass.isAbstract()) {
                for (EReference reference : backRefs.get(eClass)) {
                    if (reference.isContainment()) {
                        dbLoadReferencingObjects(eObject, eClass, reference, resource, connection);
                    }
                }
            }
        }
        /*
        for (EReference reference : EMFUtil.getBackReferences(eObject.eClass())) {
            if (reference.getEOpposite() == null && reference.isContainment()) {
                dbLoadReferencingObjects(eObject, reference, resource, connection);
            }
        }
        */
    }

    //public static List<EObject> dbLoadReferencingObjects(EObject object, EReference ref,
    public static List<EObject> dbLoadReferencingObjects(EObject object, EClass relClass, EReference ref,
        JDBCResource resource, Connection connection) throws SQLException {
        //        EClass relClass = ref.getEContainingClass();

        String sql;
        if (!ref.isMany()) {
            sql = getSelectSQL(getTableName(relClass), getFKColumnName(ref), null, null, null);
        } else if (ref.getEOpposite() == null) {
            sql = getSelectSQL(getTableName(relClass), EMFUtil.findID(relClass).getName(), 
                               getJoinTableName(ref), getFKColumnName(relClass.getEIDAttribute()), 
                               getFKColumnName(ref));
        } else if (!ref.getEOpposite().isMany()) {
            sql = getSelectSQL(getTableName(relClass), EMFUtil.findID(relClass).getName(), 
                               getTableName(object.eClass()), getFKColumnName(ref.getEOpposite()), 
                               EMFUtil.findID(object.eClass()).getName());
        } else {
            sql = getSelectSQL(getTableName(relClass), EMFUtil.findID(relClass).getName(), 
                    getJoinTableName(ref, ref.getEOpposite()), 
                    getFKColumnName(ref.getEOpposite()), getFKColumnName(ref));
        }
        PreparedStatement ps = DBFactory.prepareStatement(connection, sql);

        try {
            setValueOnPreparedStatement(connection, ps, 1, object.eGet(EMFUtil.findID(object.eClass())));
            ResultSet rs = ps.executeQuery();
            return dbLoadReferencingObjects(rs, object, ref, relClass, resource, connection);
        } finally {
            ps.close();
        }
    }

    private static List<EObject> dbLoadReferencingObjects(ResultSet rs, EObject eobject, EReference ref,
        EClass relClass, JDBCResource resource, Connection connection) throws SQLException {
        List<EObject> result = new LinkedList<EObject>();
        while (rs.next()) {
            EClass actualClass = getActualClass(rs, relClass, resource);
            EAttribute idAttr = actualClass.getEIDAttribute();

            EObject relObj = null;
            String id = EcoreUtil.convertToString(idAttr.getEAttributeType(), rs.getObject(idAttr.getName()));
            URI proxyURI = EMFUtil.getProxyURI(actualClass, id);

            if (ref.isContainment()) {
                // If the object is contained, load the actual object
                relObj = loadEObject(proxyURI, actualClass, resource, rs, connection);
                result.add(relObj);
                dbLoadContainer(relObj, resource, connection);
            } else {
                if (resource instanceof JDBCResourceImpl) {
                    relObj = ((JDBCResourceImpl)resource).findEObject(proxyURI);
                }
                if (relObj != null) {
                    result.add(relObj);
                    continue;
                }

                // If the object an ordinary reference, just create a proxy object.
                relObj = EMFUtil.createEObject(actualClass, id);
                ((InternalEObject)relObj).eSetProxyURI(proxyURI);
                if (resource instanceof JDBCResourceImpl) {
                    ((JDBCResourceImpl)resource).registerEObject(proxyURI, relObj);
                }
                result.add(relObj);
            }

            boolean prevDeliver = relObj.eDeliver();
            try {
                // disable notification of events since this is not really a modification,
                // we're just loading the initial state here
                relObj.eSetDeliver(false);

                EMFUtil.addReferenceToMemoryObject(relObj, ref, eobject);
            } finally {
                relObj.eSetDeliver(prevDeliver);
            }
        }
        rs.close();
        return result;
    }

    public static String getIdentifyingColumns(EClass cls) {
        StringBuilder ret = new StringBuilder();
        ret.append(EMFSQLUtil.getTableName(cls));
        ret.append(".");
        ret.append(cls.getEIDAttribute().getName());
        if (isTableHierarchy(cls) || cls.isAbstract()) {
            ret.append(", ");
            ret.append(EMFSQLUtil.getTableName(cls));
            ret.append(".");
            ret.append(DISCRIMINATOR_COL_NAME);
        }
        return ret.toString();
    }

    public static EClass getActualClass(ResultSet rs, EClass cls, JDBCResource resource) throws SQLException {
        if (isTableHierarchy(cls) || cls.isAbstract()) {
            String clazz = (String) rs.getObject(DISCRIMINATOR_COL_NAME);
            return resource.getRepository().getModelRegistry().getEClass(clazz);
        } else {
            return cls;
        }
    }

    static List<PropertyDescriptor> getPropertyDescriptors(Class<?> clazz)
        throws IntrospectionException {
        BeanInfo bi = Introspector.getBeanInfo(clazz);
        PropertyDescriptor[] pds = bi.getPropertyDescriptors();
        List<PropertyDescriptor> result = new ArrayList<PropertyDescriptor>(pds.length);
        for (PropertyDescriptor pd : pds) {
            if (pd.getReadMethod() == null || pd.getWriteMethod() == null) {
                continue;
            }
            Method readMethod = pd.getReadMethod();
            if (readMethod.getParameterTypes().length != 0) {
                //  the read method should not have any arguments
                continue;
            }
            Method writeMethod = pd.getWriteMethod();
            Class<?>[] writeParams = writeMethod.getParameterTypes();
            if (writeParams.length != 1) {
                // the write method should have just one argument
                continue;
            }
            if (!writeParams[0].equals(readMethod.getReturnType())) {
                continue;
            }
            result.add(pd);
        }
        return result;
    }

    private static void dbLoadRelatedOneToManyObjects(EObject object, EReference ref, EReference opposite,
                                                      JDBCResource resource, Connection connection)
        throws SQLException {
        Collection<EClass> relTableClasses;
        if (isTableClass(ref.getEReferenceType())) {
            relTableClasses = Collections.singleton(ref.getEReferenceType());
        } else {
            relTableClasses = getTableClasses(ref.getEReferenceType(), 
                                              resource.getRepository().getModelRegistry());
        }
        
        for (EClass relClass : relTableClasses) {
            String sql = getSelectSQL(getTableName(relClass), getFKColumnName(opposite), null, null, null); 
            PreparedStatement ps = DBFactory.prepareStatement(connection, sql);

            try {
                Object id = object.eGet(object.eClass().getEIDAttribute());
                setValueOnPreparedStatement(connection, ps, 1, id);
                ResultSet rs = ps.executeQuery();
                dbLoadRelatedObjects(rs, object, ref, relClass, resource, connection);
            } finally {
                ps.close();
            }
        }
    }

    private static void dbLoadJoinTableOneToManyObjects(EObject object, EReference ref,
                                                        JDBCResource resource, Connection connection)
        throws SQLException {
        Collection<EClass> relTableClasses;
        if (isTableClass(ref.getEReferenceType())) {
            relTableClasses = Collections.singleton(ref.getEReferenceType());
        } else {
            relTableClasses = getTableClasses(ref.getEReferenceType(), 
                                              resource.getRepository().getModelRegistry());
        }
        
        for (EClass relClass : relTableClasses) {
            String sql = getSelectSQL(getTableName(relClass), EMFUtil.findID(relClass).getName(), 
                                      getJoinTableName(ref), getFKColumnName(ref), 
                                      getFKColumnName(EMFUtil.findID(object.eClass())));
            PreparedStatement ps = DBFactory.prepareStatement(connection, sql);

            try {
                Object id = object.eGet(object.eClass().getEIDAttribute());
                setValueOnPreparedStatement(connection, ps, 1, id);
                ResultSet rs = ps.executeQuery();
                dbLoadRelatedObjects(rs, object, ref, relClass, resource, connection);
            } finally {
                ps.close();
            }
        }
    }

    private static void dbLoadJoinTableMany2ManyObjects(EObject eobject, EReference ref, EReference opposite,
                                                        JDBCResource resource, Connection connection)
        throws SQLException {
        Collection<EClass> relTableClasses;
        if (isTableClass(ref.getEReferenceType())) {
            relTableClasses = Collections.singleton(ref.getEReferenceType());
        } else {
            relTableClasses = getTableClasses(ref.getEReferenceType(), 
                                              resource.getRepository().getModelRegistry());
        }
        
        for (EClass relClass : relTableClasses) {
            String sql = getSelectSQL(getTableName(relClass), EMFUtil.findID(relClass).getName(), 
                                      getJoinTableName(ref, opposite), 
                                      getFKColumnName(ref), getFKColumnName(opposite));
            PreparedStatement ps = DBFactory.prepareStatement(connection, sql);

            try {
                Object id = eobject.eGet(eobject.eClass().getEIDAttribute());
                setValueOnPreparedStatement(connection, ps, 1, id);
                ResultSet rs = ps.executeQuery();
                dbLoadRelatedObjects(rs, eobject, ref, relClass, resource, connection);
            } finally {
                ps.close();
            }
        }
    }

    private static String getSelectSQL(String tableName, String idColumn,
            String joinTableName, String jointTableIdColumn, String jointTableOppositeColumn) {
        StringBuilder sb = new StringBuilder("SELECT * FROM ");
        sb.append(tableName).append(" WHERE ").append(idColumn);
        if (joinTableName != null) {
            sb.append(" IN (SELECT ").append(joinTableName).append(".").append(jointTableIdColumn);
            sb.append(" FROM ").append(joinTableName);
            sb.append(" WHERE ").append(joinTableName).append(".").append(jointTableOppositeColumn);
            sb.append("= ?)");
        } else {
            sb.append("= ?");
        }
        return sb.toString().toUpperCase();
    }

    private static int dbUpdateExistingObject(EObject eObject, Connection connection) throws SQLException {
        return dbUpdateExistingObject(eObject, (List<EAttribute>)eObject.eClass().getEAllAttributes(),
                                      connection);
    }

    static int dbUpdateSingleAttribute(EObject eObject, EAttribute attr, Connection connection)
        throws SQLException {
        return dbUpdateExistingObject(eObject, Collections.singletonList(attr), connection);
    }

    private static int dbUpdateExistingObject(EObject eObject, List<EAttribute> attrs, Connection connection)
        throws SQLException {
        EAttribute id = EMFUtil.findID(eObject.eClass());

        StringBuilder sql = new StringBuilder("UPDATE " + getTableName(eObject.eClass()) + " SET ");
        List<Object> values = new LinkedList<Object>();
        List<EAttribute> multiValues = new LinkedList<EAttribute>();
        List<EAttribute> clobValues = new LinkedList<EAttribute>();

        // The following will also generate code to set the ID to its current value,
        // e.g. SET GUID=?,NAME=? WHERE GUID=?
        // this is necessary in cases where there is no data other than an ID. In these cases
        // we still want to know
        // whether the record is there because otherwise we need to insert.
        for (EAttribute attr : attrs) {
            if (attrs.size() > 1 && isClobType(attr)) {
                clobValues.add(attr);
            } else if (attr.isMany()) {
                multiValues.add(attr);
            } else if (!EMFUtil.isBuiltInType(attr)) {
                getCustomObjectInsertUpdateSQL(eObject.eGet(attr), values, sql, attr, " = ? ");
            } else {
                sql.append(getColumnForAttribute(attr) + " = ? ");
                values.add(eObject.eGet(attr));
                sql.append(", ");
            }
        }
        removeTrailingComma(sql);
        sql.append(" WHERE " + id.getName() + " = ?");

        boolean prevAutoCommit = connection.getAutoCommit();
        PreparedStatement ps = DBFactory.prepareStatement(connection, sql.toString());
        try {
            if (prevAutoCommit) {
                connection.setAutoCommit(false);
            }

            int i = 0;
            for (; i < values.size(); i++) {
                setValueOnPreparedStatement(connection, ps, i + 1, values.get(i));
            }
            setValueOnPreparedStatement(connection, ps, i + 1, eObject.eGet(id));

            int rows = ps.executeUpdate();
            if (rows > 1) {
                throw new STPRuntimeException("Data inconsistency, more than 1 row affected "
                                              + " in updated with: " + sql.toString());
            }

            // Only store these if the row actually exists.
            if (rows == 1) {
                dbStoreMultiValues(eObject, multiValues, connection);

                for (EAttribute attr : clobValues) {
                    dbUpdateSingleAttribute(eObject, attr, connection);
                }
            }

            if (prevAutoCommit) {
                connection.commit();
            }
            return rows;
        } catch (SQLException e) {
            if (prevAutoCommit) {
                connection.rollback();
            }
            throw e;
        } finally {
            ps.close();
            if (prevAutoCommit) {
                connection.setAutoCommit(prevAutoCommit);
            }
        }
    }

    private static void removeTrailingComma(StringBuilder sql) {
        int idx = sql.lastIndexOf(",");
        if (idx != -1) {
            sql.setCharAt(idx, ' ');
        }
    }

    static void dbStoreMultiValues(EObject eobject, List<EAttribute> multiValues, Connection connection)
        throws SQLException {
        for (EAttribute attr : multiValues) {
            dbStoreMultiValueTx(eobject, attr, connection);
        }
    }

    static void dbStoreMultiValue(EObject eobject, EAttribute attr, Connection connection)
        throws SQLException {
        boolean prevAutoCommit = connection.getAutoCommit();
        try {
            if (prevAutoCommit) {
                connection.setAutoCommit(false);
            }
            dbStoreMultiValueTx(eobject, attr, connection);

            if (prevAutoCommit) {
                connection.commit();
            }
        } catch (SQLException e) {
            if (prevAutoCommit) {
                connection.rollback();
            }
            throw e;
        } finally {
            if (prevAutoCommit) {
                connection.setAutoCommit(prevAutoCommit);
            }
        }
    }

    @SuppressWarnings("unchecked")
    private static void dbStoreMultiValueTx(EObject eobject, EAttribute attr, Connection connection)
        throws SQLException {
        if (connection.getAutoCommit()) {
            throw new SQLException("This method should run as part of a transaction.");
        }

        EAttribute id = EMFUtil.findID(eobject.eClass());
        Object objectID = eobject.eGet(id);
        EClass cls = (EClass)attr.eContainer();

        StringBuilder sb = new StringBuilder();
        sb.append("DELETE FROM " + getMultiValueTableName(cls, attr));
        sb.append(" WHERE " + getFKColumnName(id) + " = ?");
        PreparedStatement ps = DBFactory.prepareStatement(connection, sb.toString().toUpperCase());
        setValueOnPreparedStatement(connection, ps, 1, objectID);
        ps.executeUpdate();
        ps.close();

        int ordnr = 0;
        for (Object value : (List<Object>)eobject.eGet(attr)) {
            List<Object> valueParts = new LinkedList<Object>();

            StringBuilder sb2 = new StringBuilder();
            sb2.append("INSERT INTO " + getMultiValueTableName(cls, attr));
            sb2.append(" (" + getFKColumnName(id) + ",");
            if (!EMFUtil.isBuiltInType(attr)) {
                getCustomObjectInsertUpdateSQL(value, valueParts, sb2, attr, "");
            } else {
                sb2.append(getColumnForAttribute(attr));
                valueParts.add(value);
                sb2.append(",");
            }
            sb2.append(ORDERING_COL_NAME + ") VALUES (?");
            for (int i = 0; i < valueParts.size(); i++) {
                sb2.append(",?");
            }
            sb2.append(",?)");
            PreparedStatement ps2 = DBFactory.prepareStatement(connection, sb2.toString().toUpperCase());
            setValueOnPreparedStatement(connection, ps2, 1, objectID);
            int i = 0;
            for (; i < valueParts.size(); i++) {
                setValueOnPreparedStatement(connection, ps2, i + 2, valueParts.get(i));
            }
            setValueOnPreparedStatement(connection, ps2, i + 2, new Integer(ordnr++));
            int rows = ps2.executeUpdate();
            ps2.close();

            if (rows != 1) {
                throw new STPRuntimeException("Data inconsistency, more than 1 row "
                                              + " affected in updated with: " + sb2);
            }
        }
    }

    private static boolean isPrimitive(Object value) {

        if (value instanceof Number || value instanceof Boolean || value instanceof Date) {
            return true;
        }
        return false;
    }

    private static void setPrimitiveValue(PreparedStatement ps, int index, Object value) throws SQLException {
        if (value instanceof Byte) {
            ps.setByte(index, Byte.class.cast(value));
        } else if (value instanceof Integer) {
            ps.setInt(index, Integer.class.cast(value));
        } else if (value instanceof Boolean) {
            ps.setBoolean(index, Boolean.class.cast(value));
        } else if (value instanceof Double) {
            ps.setDouble(index, Double.class.cast(value));
        } else if (value instanceof Float) {
            ps.setFloat(index, Float.class.cast(value));
        } else if (value instanceof Long) {
            ps.setLong(index, Long.class.cast(value));
        } else if (value instanceof Short) {
            ps.setInt(index, Short.class.cast(value));
        } else if (value instanceof Date) {
            ps.setObject(index, Date.class.cast(value));
        }
    }

    private static void setValueOnPreparedStatement(Connection conn,
            PreparedStatement ps, int index, Object value) throws SQLException {

        if (value != null) {
            if (isPrimitive(value)) {
                setPrimitiveValue(ps, index, value);
            } else {
                ps.setObject(index, value.toString());
            }
        } else {
            String db = conn.getMetaData().getDatabaseProductName().toUpperCase();
            if (db.contains("ORACLE")) {
                ps.setNull(index, 0);
            } else {
                ps.setNull(index, ps.getParameterMetaData().getParameterType(index));
            }
        }
    }

    private static void dbInsertNewObject(EObject eObject, Connection connection) throws SQLException {
        List<Object> values = new LinkedList<Object>();
        List<EAttribute> multiValues = new LinkedList<EAttribute>();
        StringBuilder sql = getInsertNewObjectSQL(eObject, values, multiValues);

        boolean prevAutoCommit = connection.getAutoCommit();
        PreparedStatement ps = DBFactory.prepareStatement(connection, sql.toString());
        try {
            if (prevAutoCommit) {
                connection.setAutoCommit(false);
            }

            int i = 0;
            for (; i < values.size(); i++) {
                setValueOnPreparedStatement(connection, ps, i + 1, values.get(i));
            }
            if (isTableHierarchy(eObject.eClass())) {
                setValueOnPreparedStatement(connection, ps, i + 1, eObject.eClass().getName());
            }
            int rows = ps.executeUpdate();

            if (rows == 1) {
                dbStoreMultiValues(eObject, multiValues, connection);
            }

            if (prevAutoCommit) {
                connection.commit();
            }
        } catch (SQLException e) {
            if (prevAutoCommit) {
                connection.rollback();
            }
            throw e;
        } finally {
            ps.close();
            if (prevAutoCommit) {
                connection.setAutoCommit(prevAutoCommit);
            }
        }
    }

    private static StringBuilder getInsertNewObjectSQL(EObject eObject, List<Object> values,
                                                       List<EAttribute> multiValues) {
        StringBuilder sql = new StringBuilder("INSERT INTO " + getTableName(eObject.eClass()) + " (");
        for (EAttribute attr : (List<EAttribute>)eObject.eClass().getEAllAttributes()) {
            if (attr.isMany()) {
                multiValues.add(attr);
                continue;
            }

            if (!EMFUtil.isBuiltInType(attr)) {
                getCustomObjectInsertUpdateSQL(eObject.eGet(attr), values, sql, attr, "");
            } else {
                sql.append(getColumnForAttribute(attr));
                values.add(eObject.eGet(attr));
                sql.append(", ");
            }
        }
        removeTrailingComma(sql);
        if (isTableHierarchy(eObject.eClass())) {
            sql.append(", " + DISCRIMINATOR_COL_NAME);
        }

        sql.append(") VALUES (");
        for (int i = 0; i < values.size(); i++) {
            if (i > 0) {
                sql.append(", ");
            }
            sql.append("?");
        }
        if (isTableHierarchy(eObject.eClass())) {
            sql.append(",?");
        }
        sql.append(")");
        return sql;
    }

    private static void getCustomObjectInsertUpdateSQL(Object valueObj, List<Object> values,
                                                       StringBuilder sql, EAttribute attr, String sqlSuffix) {
        try {
            String serializationHint = EMFUtil.getAnnotationValue(
                                                            attr.getEType(),
                                                            DBHINTS_ANNOTATION,
                                                            SERIALIZATION_KEY);
            if (SERIALIZATION_TO_STRING.equals(serializationHint)) {
                sql.append(attr.getName() + sqlSuffix);
                values.add(valueObj == null ? null : valueObj.toString());
                sql.append(", ");
            } else if (valueObj != null) {
                List<PropertyDescriptor> pds = getPropertyDescriptors(valueObj.getClass());
                if (pds.size() > 0) {
                    for (PropertyDescriptor pd : pds) {
                        sql.append(attr.getName() + "_" + pd.getName() + sqlSuffix);
                        values.add(pd.getReadMethod().invoke(valueObj));
                        sql.append(", ");
                    }
                } else {
                    sql.append(attr.getName() + sqlSuffix);
                    values.add(valueObj.toString());
                    sql.append(", ");
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public static void dbEnsureObject(EObject eObject, Connection connection) throws SQLException {
        if (dbUpdateExistingObject(eObject, connection) == 0) {
            dbInsertNewObject(eObject, connection);
        }
    }

    /** Remove a particular object from the database.
     * @param eObject The object to remove.
     * @param connection The database connection to use.
     * @throws SQLException in case of trouble with the database.
     */
    public static void dbRemoveObject(EObject eObject, Connection connection) throws SQLException {
        EAttribute id = EMFUtil.findID(eObject.eClass());

        String sql = "DELETE FROM " + getTableName(eObject.eClass()) + " WHERE " + id.getName() + " = ?";

        PreparedStatement ps = DBFactory.prepareStatement(connection, sql);
        try {
            setValueOnPreparedStatement(connection, ps, 1, eObject.eGet(id));
            ps.executeUpdate();
        } finally {
            ps.close();
        }
    }

    @SuppressWarnings("unchecked")
    public static void resolveProxyURIs(EObject eobj) {
        for (EReference ref : (List<EReference>)eobj.eClass().getEReferences()) {
            Object value = eobj.eGet(ref, false);
            List refList = null;
            if (value instanceof List) {
                refList = (List)value;
            } else {
                refList = Collections.singletonList(value);
            }
            for (Object referencedObject : refList) {
                if (referencedObject instanceof InternalEObject) {
                    InternalEObject intRef = (InternalEObject)referencedObject;
                    URI oldURI = intRef.eProxyURI();
                    if (oldURI.isFile()) {
                        URI newURI = EMFUtil.getProxyURI(ref.getEReferenceType(), oldURI.fragment());
                        intRef.eSetProxyURI(newURI);
                    }
                } else {
                    throw new STPRuntimeException("Unknown reference: " + referencedObject);
                }
            }
        }
    }
    
    public static String getDatabaseName(Connection conn) throws SQLException {
        return conn.getMetaData().getDatabaseProductName().toUpperCase();
    }
}
