/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.jdbc;

import java.io.IOException;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.logging.Logger;

import com.iona.soa.repository.AbstractSOARepositoryImpl;
import com.iona.soa.repository.EMFUtil;
import com.iona.soa.repository.ModelRegistry;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.TransactionAdapter;
import com.iona.soa.repository.TransactionData;
import com.iona.soa.repository.jdbc.resource.JDBCResource;
import com.iona.soa.repository.jdbc.resource.JDBCResourceFactoryImpl;
import com.iona.soa.repository.jdbc.resource.JDBCResourceSetImpl;
import com.iona.soa.repository.util.ExceptionHelper;
import com.iona.soa.resourceset.QueryResourceSet;

import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;

public class JDBCResourceRepository extends AbstractSOARepositoryImpl {
    private static final Logger LOG = Logger.getLogger(JDBCResourceRepository.class.getName());

    private static final long LOCK_FIRST_TIMEOUT = 60;
    private static final long LOCK_TIMEOUT = 300;
    private static final TimeUnit LOCK_TIMEOUT_UNIT = TimeUnit.SECONDS;
    
    private final ModelRegistry modelRegistry;
    private final JDBCConnectionManager connectionMgr;    
    private final JDBCResourceSetImpl resourceSet;
    private final ReadWriteLock rwLock = new ReentrantReadWriteLock();
    private ThreadTrace lockStack;

    
    /**
     * Constructor.
     * @param jdbcURI the JDBC URL used to create the connection 
     * @param conn The JDBC connection to use
     * @throws IOException If an IO Error occurs.
     */
    @SuppressWarnings("unchecked")
    public JDBCResourceRepository(JDBCConnectionManager conn, 
                                  ModelRegistry registry, 
                                  Adapter ... adapters) throws IOException {
        super(adapters);
        connectionMgr = conn;
        modelRegistry = registry;
        resourceSet = new JDBCResourceSetImpl(registry);
        resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put(
                Resource.Factory.Registry.DEFAULT_EXTENSION, 
                new JDBCResourceFactoryImpl(this));
        registerTransactionAdapter(new DatabaseTransactionAdapter());
    }
    
    public void startRead(TransactionData txData) {
        acquire(rwLock.readLock());
        try {
            super.startRead(txData);
        } catch (Throwable t) {
            rwLock.readLock().unlock();
            if (t instanceof RuntimeException) {
                throw (RuntimeException)t;
            } else if (t instanceof Error) {
                throw (Error)t;
            } else {
                throw new RuntimeException(t);
            }
        }
    }
    
    public void endRead(boolean success) {
        try {
            super.endRead(success);
        } finally {
            rwLock.readLock().unlock();
        }
    }
    
    public void startWrite(TransactionData txData) {
        acquire(rwLock.writeLock());
        try {
            super.startWrite(txData);
        } catch (Throwable t) {
            rwLock.writeLock().unlock();
            if (t instanceof RuntimeException) {
                throw (RuntimeException)t;
            } else if (t instanceof Error) {
                throw (Error)t;
            } else {
                throw new RuntimeException(t);
            }
        }
    }
    

    public void endWrite(boolean success) {
        try {
            super.endWrite(success);
        } finally {
            rwLock.writeLock().unlock();
        }
    }
    
    public QueryResourceSet getResourceSet() throws IOException {
        resourceSet.load();
        return resourceSet;
    }
    
    public JDBCConnectionManager getConnectionManager() {
        return connectionMgr;
    }

    public EObject getEObject(URI uri, boolean loadOnDemand) {
        resourceSet.load();
        EObject obj = resourceSet.findEObject(uri);
        if (obj != null) {
            if (loadOnDemand) {
                if (obj.eIsProxy()) {
                    obj = EcoreUtil.resolve(obj, resourceSet);
                    if (!obj.eIsProxy()) {
                        return obj;
                    }
                }
            } else {
                return obj;
            }
        }
        return resourceSet == null ? null : resourceSet.getEObject(uri, loadOnDemand);
    }
    
    @SuppressWarnings("unchecked")
    public void unregisterObject(EObject obj) {
        URI uri = EMFUtil.getProxyURI(obj);
        for (Resource resource : (List<Resource>)resourceSet.getResources()) {
            if (resource instanceof JDBCResource) {
                ((JDBCResource)resource).unregisterEObject(uri);
            }
        }
    }

    public ModelRegistry getModelRegistry() {
        return modelRegistry;
    }
    
    /**
     * {@inheritDoc}
     * @throws IOException 
     */
    @SuppressWarnings("unchecked")
    public void persist(EObject obj) throws IOException {
        resourceSet.load();
        URI uri = EMFUtil.getProxyURI(obj);
        Resource resource = resourceSet.getResource(uri.trimFragment(), true);
        if (resource == null) {
            resource = resourceSet.createResource(uri.trimFragment()); 
            resourceSet.getResources().add(resource);
        }
        
        resource.getContents().add(obj);
    }    
    
    /**
     * {@inheritDoc}
     */
    public void close() {
        resourceSet.close();
    }
    
    private class DatabaseTransactionAdapter implements TransactionAdapter {
        public void startTransaction(SOARepository repo, TransactionData txData) throws Exception {
            getConnectionManager().startTransaction();
        }

        public void commitTransaction() throws Exception {
            getConnectionManager().commitTransaction();
        }

        public void rollbackTransaction() throws Exception {
            getConnectionManager().rollbackTransaction();
        }

        public int getPriority() {
            // highest possible priority
            return Integer.MAX_VALUE;
        }
    }
    
    private void acquire(Lock lock) {
        long total = 0;
        boolean firstRound = true;

        while (true) {
            long timeOut = firstRound ? LOCK_FIRST_TIMEOUT : LOCK_TIMEOUT;
            try {
                if (lock.tryLock(timeOut, LOCK_TIMEOUT_UNIT)) {
                    lockStack = new ThreadTrace(Thread.currentThread());
                    return;
                }
            } catch (InterruptedException e) {
                LOG.warning("interrupted exception");
            }
            total += timeOut;
            logLock(lock, total, firstRound);
            firstRound = false;
        }
    }

    private void logLock(Lock lock, long time, boolean showStackTrace) {
        StringBuilder sb = new StringBuilder()
            .append("Thread ").append(Thread.currentThread().getId())
            .append(" is blocked on lock ").append(lock)
            .append("\nTotal time blocked is ").append(time)
            .append(" ").append(LOCK_TIMEOUT_UNIT).append("\n");
        if (showStackTrace) {
            sb.append("Stack trace: ")
                .append(ExceptionHelper.getStackTraceAsString(
                        new Throwable(Thread.currentThread().getName())));
            ThreadTrace tt = lockStack;
            if (tt != null) {
                sb.append("Lock held by: ")
                    .append(ExceptionHelper.getStackTraceAsString(tt.getThread()));
                sb.append("Lock acquired at: ")
                    .append(ExceptionHelper.getStackTraceAsString(tt));
            }
            
        }
        sb.append("trying to acquire lock again...");
        LOG.warning(sb.toString());
        
    }

    @SuppressWarnings("serial")
    private static class ThreadTrace extends Throwable {
        private Thread thread;
        
        public ThreadTrace(Thread t) {
            super(t.getName());
            thread = t;
        }
        
        public Thread getThread() {
            return thread;
        }
    }
}  
