/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.jdbc;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.List;

import com.iona.repository.db.Database;
import com.iona.soa.repository.status.RepositoryInitializationException;
import com.iona.soa.repository.status.StatusMessage;
import com.iona.soa.repository.util.AbstractObjectCache;

public class MTConnectionManagerImpl implements JDBCConnectionManager {
    private final AbstractObjectCache<PooledConnection> connectionCache;
    private final Database database;
    private final ThreadLocal<PooledConnection> inProgressTransaction = new ThreadLocal<PooledConnection>();

    public MTConnectionManagerImpl(Database db) throws RepositoryInitializationException {
        database = db;
        Connection probe = null;
        try {
            probe = database.getConnection();
        } catch (Exception e) {
            throw new RepositoryInitializationException(e, StatusMessage.UNABLE_CONNECT_TO_DATABASE, 
                                                        database.getDatabaseURL());
        } 
        connectionCache = 
            new AbstractObjectCache<PooledConnection>(5, new PooledConnection(probe)) {
                @Override
                protected PooledConnection create() {
                    try {
                        return new PooledConnection(database.getConnection());
                    } catch (SQLException e) {
                        return new PooledConnection(null);
                    }
                }
            };
    }
    
    public PooledConnection getConnection() {
        PooledConnection ret = inProgressTransaction.get();
        if (ret != null) {
            return ret;
        }
        do {
            ret = connectionCache.get();
            try {
                Connection conn = ret.getConnection();
                if (conn == null || conn.isClosed()) {
                    ret = null;
                }
            } catch (SQLException ex) {
                ret = null;
            }
        } while (ret == null);
        return ret;
    }
    
    public void returnConnection(PooledConnection connection) {
        try {
            if (connection != null 
                    && inProgressTransaction.get() != connection
                    && !connection.getConnection().isClosed()) {
                connectionCache.put(connection);
            }
        } catch (SQLException ex) {
            // Ignore, connection is hosed so don't return to pool
        }
    }
    
    public void startTransaction() throws SQLException {
        if (inProgressTransaction.get() != null) {
            throw new SQLException("Transaction already in progress");
        }
        PooledConnection conn = connectionCache.get();
        if (conn == null || conn.getConnection() == null) {
            throw new SQLException("Unable to obtain JDBC Connection");
        }
        try {
            conn.getConnection().setAutoCommit(false);
        } catch (SQLException ex) {
            connectionCache.put(conn);
            throw ex;
        }
        inProgressTransaction.set(conn);
    }

    public void commitTransaction() throws SQLException {
        PooledConnection conn = inProgressTransaction.get();
        if (conn == null) {
            throw new SQLException("No Transaction in progress");
        }
        inProgressTransaction.remove();
        try {
            try {
                conn.getConnection().commit();
            } finally {
                conn.getConnection().setAutoCommit(true);
            }
        } finally {
            connectionCache.put(conn);
        }
    }

    public void rollbackTransaction() throws SQLException {
        PooledConnection conn = inProgressTransaction.get();
        if (conn == null) {
            throw new SQLException("No Transaction in progress");
        }
        inProgressTransaction.remove();
        try {
            try {
                conn.getConnection().rollback();
            } finally {
                conn.getConnection().setAutoCommit(true);
            }
        } finally {
            connectionCache.put(conn);
        }
    }

    public void shutDown() {
        List<PooledConnection> connections = connectionCache.flush();
        for (PooledConnection conn : connections) {
            try {
                conn.getConnection().close();
            } catch (SQLException e) {
                // ignore;
            }
        }
        database.stop();
    }
    
    public String getDatabaseURL() {
        return (null != database) ? database.getDatabaseURL() : "";
    }
}
