/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.jdbc;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import com.iona.soa.repository.EMFUtil;
import com.iona.soa.repository.ModelRegistry;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EReference;

public final class SQLMigrateUtil {
    private SQLMigrateUtil() {
    }

    public static List<String> getAddAttributeSQL(EAttribute attr, 
                                                  EClass tableClass,
                                                  String dbName) {
        List<String> sqls = new LinkedList<String>();
        if (attr.isMany()) {
            EMFSQLCreationUtil.getMultiValueSubTableSQL(attr, sqls, dbName);
        } else if (EMFSQLUtil.isTableClass(tableClass)) {
            StringBuilder sb = new StringBuilder();
            sb.append("ALTER TABLE ");
            sb.append(EMFSQLUtil.getTableName(tableClass));
            sb.append(" ADD ");
            if (dbName.contains("ORACLE")) {
                sb.append(" (");
            } else {
                sb.append(" COLUMN ");
            }
            EMFSQLCreationUtil.addSingleColumnDef(sb, attr, dbName);
            if (dbName.contains("ORACLE")) {
                sb.append(")");
            }
            sqls.add(sb.toString().toUpperCase());
        }
        return sqls;
    }
    
    public static List<String> getAddReferenceSQL(EReference ref, 
                                                  EClass tableClass,
                                                  String dbName) {
        List<String> sqls = new LinkedList<String>();
        if (ref.isMany()) {
            EReference opposite = ref.getEOpposite();
            if (opposite == null) {
                sqls.add(EMFSQLCreationUtil.getOneToManyTableSQL(ref, dbName));
            } else if (opposite.isMany()) {
                sqls.add(EMFSQLCreationUtil.getManyToManyTableSQL(ref, opposite, dbName));
            } else {
                return getAddReferenceSQL(opposite, opposite.getEContainingClass(), dbName);
            }
        } else if (EMFSQLUtil.isTableClass(tableClass)) {
            StringBuilder sb = new StringBuilder();
            EAttribute refID = EMFUtil.findID(ref.getEReferenceType());
            sb.append("ALTER TABLE ");
            sb.append(EMFSQLUtil.getTableName(tableClass));
            sb.append(" ADD ");
            if (dbName.contains("ORACLE")) {
                sb.append(" (");
            } else {
                sb.append(" COLUMN ");
            }
            sb.append(EMFSQLUtil.getFKColumnName(ref));
            sb.append(" ");
            sb.append(EMFSQLUtil.getSQLDataType(refID, dbName));
            if (dbName.contains("ORACLE")) {
                sb.append(")");
            }
            sqls.add(sb.toString().toUpperCase());

            if (EMFSQLUtil.isTableClass(ref.getEReferenceType())) {
                StringBuilder sb2 = new StringBuilder();
                sb2.append("ALTER TABLE " + EMFSQLUtil.getTableName(tableClass));
                sb2.append(" ADD CONSTRAINT ");
                sb2.append(EMFSQLCreationUtil.getFKConstraint(EMFSQLUtil.getUnmangledTableName(tableClass), 
                        EMFSQLUtil.normalizeReferenceName(ref)));
                sb2.append(" FOREIGN KEY (" + EMFSQLUtil.getFKColumnName(ref) + ") REFERENCES");
                sb2.append(" " + EMFSQLUtil.getTableName(ref.getEReferenceType()));
                sb2.append("(" + refID.getName() + ")");
                sqls.add(sb2.toString().toUpperCase());
            }
        }
        return sqls;
    }
    
    public static List<String> getRemoveAttributeSQL(EAttribute attr, 
                                                     EClass tableClass) {
        List<String> sqls = new LinkedList<String>();

        StringBuilder sb = new StringBuilder();
        if (attr.isMany()) {
            sb.append("DROP TABLE ");
            sb.append(EMFSQLUtil.getMultiValueTableName(attr.getEContainingClass(), attr));                
        } else if (EMFSQLUtil.isTableClass(tableClass)) {
            sb.append("ALTER TABLE ");
            sb.append(EMFSQLUtil.getTableName(tableClass));
            sb.append(" DROP COLUMN ");
            sb.append(EMFSQLUtil.getColumnForAttribute(attr));
        }
        sqls.add(sb.toString().toUpperCase());
        return sqls;
    }
    
    public static List<String> getRemoveReferenceSQL(EReference ref, 
                                                     EClass tableClass) {
        List<String> sqls = new LinkedList<String>();
        StringBuilder sb = new StringBuilder();
        if (ref.isMany()) {
            EReference opposite = ref.getEOpposite();
            if (opposite == null) {
                sb.append("DROP TABLE ");
                sb.append(EMFSQLUtil.getJoinTableName(ref));                
            } else if (opposite.isMany()) {
                sb.append("DROP TABLE ");
                sb.append(EMFSQLUtil.getJoinTableName(ref, opposite));                
            } else {
                return getRemoveReferenceSQL(opposite, null);
            }
        } else if (EMFSQLUtil.isTableClass(tableClass)) {
            sqls.addAll(getRemoveFKConstraintSQL(ref, tableClass));
            sb.append("ALTER TABLE ");
            sb.append(EMFSQLUtil.getTableName(tableClass));
            sb.append(" DROP COLUMN ");
            sb.append(EMFSQLUtil.getFKColumnName(ref));
        }
        if (sb.length() > 0) {
            sqls.add(sb.toString().toUpperCase());
        }

        return sqls;
    }
    
    public static List<String> getRemoveFKConstraintSQL(EReference ref, 
                                                        EClass tableClass) {
        List<String> sqls = new LinkedList<String>();
        if (!ref.isMany() && EMFSQLUtil.isTableClass(tableClass) 
                && EMFSQLUtil.isTableClass(ref.getEReferenceType())) {
            sqls.add(getRemoveConstraintSQL(
                    EMFSQLUtil.getTableName(tableClass), 
                    EMFSQLCreationUtil.getFKConstraint(
                            EMFSQLUtil.getUnmangledTableName(tableClass), 
                            EMFSQLUtil.normalizeReferenceName(ref))));
        }
        return sqls;
    }

    public static String getRemoveConstraintSQL(String tableName, String fkName) {
        StringBuilder sb = new StringBuilder();
        sb.append("ALTER TABLE ");
        sb.append(tableName);
        sb.append(" DROP CONSTRAINT ");
        sb.append(fkName);
        return sb.toString();
    }

    @SuppressWarnings("unchecked")
    public static List<String> getRemoveEClassSQL(EClass eClass) {
        List<String> sqls = new LinkedList<String>();
        if (EMFSQLUtil.isTableClass(eClass)) {
            StringBuilder sb = new StringBuilder();
            sb.append("DROP TABLE ");
            sb.append(EMFSQLUtil.getTableName(eClass));
            sqls.add(sb.toString().toUpperCase());
        }
        return sqls;
    }

    public static String getFKName(Connection connection, 
                                   String tableName, 
                                   String fkColName) throws SQLException {
        ResultSet rs = connection.getMetaData().getImportedKeys(null, null, tableName);
        try {
            while (rs.next()) {
                if (fkColName.equals(rs.getString("FKCOLUMN_NAME"))) {
                    return rs.getString("FK_NAME");
                }
            }
        } finally {
            rs.close();
        }
        // try again with lowercase table and fk name - postgres is case sensitive and
        // for some bizarre reason, all table structures are lowercase in postgres db's.
        rs = connection.getMetaData().getImportedKeys(null, null, tableName.toLowerCase());
        try {
            while (rs.next()) {
                if (fkColName.toLowerCase().equals(rs.getString("FKCOLUMN_NAME"))) {
                    return rs.getString("FK_NAME");
                }
            }
        } finally {
            rs.close();
        }
        throw new SQLException("Column " + fkColName + " not found in table " + tableName);
    }

    public static void updateForeignKeys(EAttribute eAttr, 
                                         Connection connection, 
                                         Collection<String> ret) throws SQLException {
        if (eAttr.isMany()) {
            EClass eClass = eAttr.getEContainingClass();
            String tableName = EMFSQLUtil.getMultiValueTableName(eClass, eAttr);
            String fkColName = EMFSQLUtil.getFKColumnName(eClass.getEIDAttribute());

            String fkName = getFKName(connection, tableName, fkColName);
            ret.add(getRemoveConstraintSQL(tableName, fkName));

            EMFSQLCreationUtil.getMultiValueSubTableFkSQL(eAttr, ret);
        }
    }

    public static void updateOneOneForeignKeys(EReference eRef, 
                                               Connection connection, 
                                               ModelRegistry registry, 
                                               Collection<String> ret) throws SQLException {
        String fkColName = EMFSQLUtil.getFKColumnName(eRef);
        for (EClass subClass : EMFSQLUtil.getTableClasses(eRef.getEContainingClass(), registry)) {
            String tableName = EMFSQLUtil.getTableName(subClass);

            String fkName = getFKName(connection, tableName, fkColName);
            ret.add(getRemoveConstraintSQL(tableName, fkName));
            EMFSQLCreationUtil.createOneToManyFkSQL(eRef, subClass, ret);
        }
    }

    public static void updateManyManyForeignKeys(EReference eRef, 
                                                 Connection connection, 
                                                 Collection<String> ret) throws SQLException {
        String tableName = EMFSQLUtil.getJoinTableName(eRef, eRef.getEOpposite());
        
        String fkColName1 = EMFSQLUtil.getFKColumnName(eRef);
        String oldFKName1 = getFKName(connection, tableName, fkColName1);
        String newFKName1 = EMFSQLCreationUtil.getFKConstraint(
                EMFSQLUtil.getUnmangledTableName(eRef.getEContainingClass()), 
                EMFSQLUtil.normalizeReferenceName(eRef.getEOpposite()), 
                EMFSQLUtil.normalizeReferenceName(eRef));

        if (!oldFKName1.equalsIgnoreCase(newFKName1)) {
            ret.add(getRemoveConstraintSQL(tableName, oldFKName1));
            ret.add(getAddFKConstraintSQL(tableName, newFKName1, fkColName1, eRef.getEReferenceType()));
        }

        String fkColName2 = EMFSQLUtil.getFKColumnName(eRef.getEOpposite());
        String oldFKName2 = getFKName(connection, tableName, fkColName2);
        String newFKName2 = EMFSQLCreationUtil.getFKConstraint(
                EMFSQLUtil.getUnmangledTableName(eRef.getEOpposite().getEContainingClass()), 
                EMFSQLUtil.normalizeReferenceName(eRef), 
                EMFSQLUtil.normalizeReferenceName(eRef.getEOpposite()));

        if (!oldFKName2.equalsIgnoreCase(newFKName2)) {
            ret.add(getRemoveConstraintSQL(tableName, oldFKName2));
            ret.add(getAddFKConstraintSQL(
                    tableName, newFKName2, fkColName2, eRef.getEOpposite().getEReferenceType()));
        }
    }

    public static void updateOnewayManyForeignKeys(EReference eRef, 
                                                   Connection connection, 
                                                   Collection<String> ret) throws SQLException {
        EAttribute id = EMFUtil.findID(eRef.getEContainingClass());
        String tableName = EMFSQLUtil.getJoinTableName(eRef);

        String fkColName1 = EMFSQLUtil.getFKColumnName(eRef);
        String oldFKName1 = getFKName(connection, tableName, fkColName1);
        String newFKName1 = EMFSQLCreationUtil.getFKConstraint(
                EMFSQLUtil.getUnmangledTableName(eRef.getEContainingClass()), 
                id.getEContainingClass().getName(), 
                EMFSQLUtil.normalizeReferenceName(eRef));

        if (!oldFKName1.equalsIgnoreCase(newFKName1)) {
            ret.add(getRemoveConstraintSQL(tableName, oldFKName1));
            ret.add(getAddFKConstraintSQL(tableName, newFKName1, fkColName1, eRef.getEReferenceType()));
        }

        if (EMFSQLUtil.isTableClass(eRef.getEContainingClass())) {
            String fkColName2 = EMFSQLUtil.getFKColumnName(id);
            String oldFKName2 = getFKName(connection, tableName, fkColName2);
            String newFKName2 = EMFSQLCreationUtil.getFKConstraint(
                    EMFSQLUtil.getUnmangledTableName(eRef.getEContainingClass()), 
                    EMFSQLUtil.normalizeReferenceName(eRef), 
                    id.getEContainingClass().getName());

            if (!oldFKName2.equalsIgnoreCase(newFKName2)) {
                ret.add(getRemoveConstraintSQL(tableName, oldFKName2));
                ret.add(getAddFKConstraintSQL(
                        tableName, newFKName2, fkColName2, eRef.getEContainingClass()));
            }
        }
    }

    private static String getAddFKConstraintSQL(String tableName, 
                                                String fkName, 
                                                String fkColName, 
                                                EClass refClass) {
        StringBuilder sb = new StringBuilder();
        sb.append("ALTER TABLE ");
        sb.append(tableName);
        sb.append(" ADD CONSTRAINT ");
        sb.append(fkName);
        sb.append(" FOREIGN KEY (" + fkColName + ") ");
        sb.append("REFERENCES " + EMFSQLUtil.getTableName(refClass)
                + "(" + EMFUtil.findID(refClass).getName() + ")");
        return sb.toString();
    }
}
