/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.jdbc.resource;

import java.io.IOException;
import java.net.SocketException;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import com.iona.soa.expression.delegate.SQLRenderer.WhereClause;
import com.iona.soa.repository.EMFUtil;
import com.iona.soa.repository.jdbc.DBFactory;
import com.iona.soa.repository.jdbc.EMFDatabaseSerializationAdapter;
import com.iona.soa.repository.jdbc.EMFSQLUtil;
import com.iona.soa.repository.jdbc.JDBCResourceRepository;
import com.iona.soa.repository.jdbc.PooledConnection;

import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.InternalEObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.emf.ecore.xmi.XMIException;
import org.eclipse.emf.ecore.xmi.XMIResource;
import org.eclipse.emf.ecore.xmi.XMLHelper;
import org.eclipse.emf.ecore.xmi.XMLSave;
import org.eclipse.emf.ecore.xmi.impl.XMIResourceImpl;
import org.eclipse.emf.ecore.xmi.impl.XMISaveImpl;

public class JDBCResourceImpl extends XMIResourceImpl  implements JDBCResource {
    public static final String COPYRIGHT = "IONA Technologies 2005-7";
    @SuppressWarnings("unused")
    private static final Logger LOG = Logger.getLogger(JDBCResourceImpl.class.getName());
    private Adapter serializationAdapter = new EMFDatabaseSerializationAdapter(this);
    
    private static class VisitedDeletableList extends ThreadLocal<Collection<EObject>> {
        public Collection<EObject> initialValue() {
            return new LinkedList<EObject>();
        }
    }
        
    private final VisitedDeletableList visitedDeletable = new VisitedDeletableList();

    // This map contains all the loaded objects (including proxies). It is used
    // to look up objects that have already been loaded to avoid the creation of 
    // duplicates in memory for what should really be a single object.
    // Davidb TODO: to avoid memory leaks, turn this one into a weak value hash map
    // Davidb TODO: synchronize the map? Maybe use one of these Java 5 concurrent maps?
    private Map<URI, EObject> loaded = new HashMap<URI, EObject>();
    
    private final JDBCResourceRepository repository;

    @SuppressWarnings("unchecked")
    public JDBCResourceImpl(URI uri, JDBCResourceRepository rep) {
        super(uri);
        repository = rep;
        eAdapters().addAll(rep.getResourceAdapters());
        eAdapters().add(serializationAdapter);
    }
    
    public EObject getEObject(EClass eClass, String id) {
        String[] fragments = new String[] {eClass.getName(), id};
        return getEObject(Arrays.asList(fragments));
    }
    
    @SuppressWarnings("unchecked")
    @Override
    protected EObject getEObject(List uriFragmentPath) {
        if (uriFragmentPath.size() < 2) {
            //          dont know how to handle
            return null; 
        }
        PooledConnection pc = repository.getConnectionManager().getConnection();
        try {
            if (uriFragmentPath.size() > 2) {
                return EMFSQLUtil.loadEObject(this, 
                                              (String) uriFragmentPath.get(0), 
                                              (String) uriFragmentPath.get(1),
                                              (String) uriFragmentPath.get(2),
                                              pc.getConnection());
            } else {
                return EMFSQLUtil.loadEObject(this, 
                                              (String) uriFragmentPath.get(0), 
                                              (String) uriFragmentPath.get(1), 
                                              pc.getConnection());
            }
        } catch (SQLException e) {
            e.printStackTrace();
            getErrors().add(new XMIException(e));
            return null;
        } catch (RuntimeException thr) {
            thr.printStackTrace();
            throw thr;
        } finally {
            repository.getConnectionManager().returnConnection(pc);
        }
    }
    
    @SuppressWarnings("unchecked")
    @Override
    protected void attachedHelper(final EObject eObject) {
        super.attachedHelper(eObject);
        for (Adapter adapter : (List<Adapter>)eAdapters()) {
            if (adapter != serializationAdapter && !eObject.eAdapters().contains(adapter)) {
                eObject.eAdapters().add(adapter);
            }
        }
    }

    public List<EObject> queryObjects(EClass type, WhereClause clause) throws SQLException {
        Collection<EClass> tableClasses;
        if (EMFSQLUtil.isTableClass(type)) {
            tableClasses = Collections.singletonList(type);
        } else {
            tableClasses = EMFSQLUtil.getTableClasses(type, getRepository().getModelRegistry());
        }
        List<EObject> ret = new LinkedList<EObject>();
        PooledConnection pc = getRepository().getConnectionManager().getConnection();
        try {
            for (EClass tableClass : tableClasses) {
                String sql = clause.getFullQuery(tableClass);
                PreparedStatement ps = DBFactory.prepareStatement(pc.getConnection(), sql);
                
                for (int i = 0; i < clause.getParams().length; i++) {
                    ps.setObject(i + 1, clause.getParams()[i]);
                }
                ret.addAll(queryObjects(tableClass, ps));
            }
        } catch (SQLException sqle) {
            if (sqle.getCause() instanceof SocketException) {
                pc.getConnection().close();
            }
            throw sqle;
        } finally {
            getRepository().getConnectionManager().returnConnection(pc);
        }
        return ret;
    }

    public List<EObject> queryObjects(EClass type, String sqlQuery) throws SQLException {
        Collection<EClass> tableClasses;
        if (EMFSQLUtil.isTableClass(type)) {
            tableClasses = Collections.singletonList(type);
        } else {
            tableClasses = EMFSQLUtil.getTableClasses(type, getRepository().getModelRegistry());
        }
        List<EObject> ret = new LinkedList<EObject>();
        PooledConnection pc = getRepository().getConnectionManager().getConnection();
        try {
            for (EClass tableClass : tableClasses) {
                String cleanedSQL = cleanQuery(tableClass, sqlQuery);
                PreparedStatement ps = DBFactory.prepareStatement(pc.getConnection(), cleanedSQL);
                ret.addAll(queryObjects(tableClass, ps));
            }
        } catch (SQLException sqle) {
            if (sqle.getCause() instanceof SocketException) {
                pc.getConnection().close();
            }
            throw sqle;
        } finally {
            getRepository().getConnectionManager().returnConnection(pc);
        }
        return ret;
    }
    
    private static String cleanQuery(EClass type, String sqlQuery) throws SQLException {
        
        // Don't allow semicolons into the SQL query to prevent injection attacks.  This
        // assumes there is no more decoding of the query to be done (eg %xx sequences).  
        if (sqlQuery.indexOf(";") != -1) {
            throw new SQLException("Invalid characters in query");
        }
        
        // TODO check for semicolons in the SQL to avoid putting in another SQL command that causes mayhem.
        sqlQuery = sqlQuery.trim();
        String ucQuery = sqlQuery.toUpperCase();
        if (!ucQuery.startsWith("SELECT")) {
            return null;
        }
        
        int idx = ucQuery.indexOf("FROM");
        if (idx == -1) {
            return null;
        }

        StringBuilder newSql = new StringBuilder();
        newSql.append("SELECT * ");
        newSql.append(sqlQuery.substring(idx));
        return newSql.toString();
    }

    private List<EObject> queryObjects(EClass type, PreparedStatement ps) 
        throws SQLException {
        EAttribute guidAttr = type.getEIDAttribute();
        try {
            List<EObject> objects = new LinkedList<EObject>();
    
            ResultSet rs = ps.executeQuery();
            while (rs.next()) {
                String id = rs.getString(guidAttr.getName());
                EClass actualType = EMFSQLUtil.getActualClass(rs, type, this);
                if (!type.isSuperTypeOf(actualType)) {
                    continue;
                }
                
                URI proxyURI = EMFUtil.getProxyURI(actualType, id);
                EObject eobj = findEObject(proxyURI);
                if (eobj == null) {                    
                    eobj = EcoreUtil.create(actualType);
                    ((InternalEObject) eobj).eSetProxyURI(proxyURI);
                    registerEObject(proxyURI, eobj);                    
                }
                if (eobj.eIsProxy()) {
                    EcoreUtil.resolve(eobj, this);
                }
                if (!objects.contains(eobj)) {
                    // Only add object if it's not already in the list.
                    // This is similar to a DISTINCT clause, but DISTINCT
                    // does not work with SELECT DISTINCT * clauses where
                    // a column is a CLOB column.
                    objects.add(eobj);
                }
            }
            
            return objects;
        } finally {
            if (ps != null) {
                ps.close();
            }
        }
    }        
    
    public List<EObject> getXRefs(EObject object, EClass type) throws SQLException {
        Set<EObject> result = new LinkedHashSet<EObject>();
        PooledConnection pc = getRepository().getConnectionManager().getConnection();
        try {
            Map<EClass, Collection<EReference>> backRefs = EMFUtil.getBackReferences(
                    type, repository.getModelRegistry());
            for (EClass eClass : backRefs.keySet()) {
                if (!eClass.isAbstract()) {
                    for (EReference reference : backRefs.get(eClass)) {
                        if (!reference.isContainment()) {
                            result.addAll(EMFSQLUtil.dbLoadReferencingObjects(
                                    object, eClass, reference, this, pc.getConnection()));
                        }

                    }
                }
            }
            return new ArrayList<EObject>(result);
        } finally {
            getRepository().getConnectionManager().returnConnection(pc);
        }
    }
    
    public JDBCResourceRepository getRepository() {
        return repository;
    }

    @Override
    public String getURIFragment(EObject eobject) {
        return EMFUtil.getProxyURIFragment(eobject);
    }

    @SuppressWarnings("unchecked")
    public void persist(EObject eobject) throws SQLException {
        registerEObject(EMFUtil.getProxyURI(eobject), eobject);

        if (!addSerializationAdapter(eobject, true)) {
            return;
        }
        
        for (EReference ref : (List<EReference>) eobject.eClass().getEAllReferences()) {
            Object reference = eobject.eGet(ref);
            List<EObject> refList = null;
            if (reference instanceof List) {
                refList = (List) reference;
            } else if (reference instanceof EObject) {
                refList = new LinkedList<EObject>();
                refList.add((EObject) reference);
            }
            
            if (refList == null) {
                continue;
            }
            
            for (EObject refObj : refList) {
                persist(refObj);
            }
            if (refList.size() > 0) {
                PooledConnection pc = repository.getConnectionManager().getConnection();
                try {
                    EMFSQLUtil.dbStoreReferences(eobject, ref, pc.getConnection());
                } finally {
                    repository.getConnectionManager().returnConnection(pc);
                }
            }
        }        
    }

    @SuppressWarnings("unchecked")
    public boolean addSerializationAdapter(EObject eobject, boolean ensureInDB) throws SQLException {
        if (eobject.eIsProxy()) {
            return false;
        }
        
        boolean alreadyPersistent = true;
        for (Adapter adapter : eAdapters()) {
            if (!eobject.eAdapters().contains(adapter)) {
                eobject.eAdapters().add(adapter);
                if (adapter == serializationAdapter) {
                    // Already persistent, no need to do it again.
                    alreadyPersistent = false;
                }
            }
        }

        if (alreadyPersistent) {
            return false;
        }
        
        EAttribute id = EMFUtil.findID(eobject.eClass());
        if (eobject.eGet(id) == null) {
            throw new RuntimeException("Object " + eobject + " does not have its ID (" 
                                       + id.getName() 
                                       + ") set.");
        }

        if (ensureInDB) {
            PooledConnection pc = getRepository().getConnectionManager().getConnection();
            try {
                EMFSQLUtil.dbEnsureObject(eobject, pc.getConnection());
            } finally {
                getRepository().getConnectionManager().returnConnection(pc);
            }
        }
        return true;
    }

    @SuppressWarnings("unchecked")
    public void unPersist(EObject eobject, boolean removeXRefs) throws SQLException {
        if (eobject == null) {
            return;
        }

        Collection<EObject> visited = visitedDeletable.get();
        visited.add(eobject);
        
        Collection<EObject> deleteObjs = new LinkedList<EObject>();

        PooledConnection pc = repository.getConnectionManager().getConnection();
        boolean prevAutoCommit = pc.getConnection().getAutoCommit();
        if (prevAutoCommit) { 
            pc.getConnection().setAutoCommit(false);
        }
        
        try {
            unPersistReferences(eobject, deleteObjs, visited);

            // Delete all list attributes next
            for (EAttribute attr : (List<EAttribute>) eobject.eClass().getEAllAttributes()) {
                if (eobject.eIsSet(attr) && attr.isMany()) {
                    eobject.eUnset(attr);
                }
            }
            
            // Delete cross references
            if (removeXRefs) {
                removeXReferences(eobject, pc);
            }
            
            doUnPersist(pc, prevAutoCommit, eobject, deleteObjs, visited);

        } finally {
            if (prevAutoCommit) {
                pc.getConnection().setAutoCommit(prevAutoCommit);
            }
            repository.getConnectionManager().returnConnection(pc);
            visited.remove(eobject);
        }
    }

    @SuppressWarnings("unchecked")
    private void doUnPersist(PooledConnection pc, boolean autoCommit,
                             EObject eobject, Collection<EObject> deleteObjs, 
                             Collection<EObject> visited) throws SQLException {
        
        try {
            EMFSQLUtil.dbRemoveObject(eobject, pc.getConnection());
            unregisterEObject(EMFUtil.getProxyURI(eobject));
            // delete non containment referenced objects
            /*
            for (EObject o : deleteObjs) {
                visited.remove(o);
                EcoreUtil.remove(o);
            }
            */
        } catch (SQLException se) {
            addException(se);
            
            if (autoCommit) {                    
                pc.getConnection().rollback();
                
                if (eobject instanceof InternalEObject) {
                    // Reload the object from the database
                    InternalEObject iobject = (InternalEObject) eobject;
                    unloaded(iobject);
                    registerEObject(iobject.eProxyURI(), iobject);
                    getEObject(getURIFragment(eobject));
                }
            }
            return;
        }
        eobject.eAdapters().clear();
        if (autoCommit) {
            pc.getConnection().commit();
        }
    }

    @SuppressWarnings("unchecked")
    private void unPersistReferences(EObject eobject, 
                                     Collection<EObject> deleteObjs, 
                                     Collection<EObject> visited) throws SQLException {

        /*
        // collect non reachable from this object referencing this object 
        // to be deleted objects
        try {
            EMFUtil.getDeletableReverseReferencingObjects(this, eobject, deleteObjs, visited);
        } catch (Exception e) {
            throw new SQLException(e.toString());
        }
        */
        for (EReference ref : (List<EReference>) eobject.eClass().getEAllReferences()) {
            if (eobject.eIsSet(ref)) {
                /*
                // collect non containment referenced to be deleted objects 
                if (!EMFUtil.getDeletableReferencedObjects(ref, eobject, 
                                                              deleteObjs, visited)) {
                    // collect non containment referenced dangling 
                    // to be deleted objects
                    EMFUtil.getDeletableReferencedDanglingObjects(ref, eobject, 
                                                                     deleteObjs, visited);
                } 
                */
                // This will automatically remove contained objects too, as the unset is 
                // caught in te EMFDatabaseSerializationAdapter
                eobject.eUnset(ref);
            }
        } 
    }

    @SuppressWarnings("unchecked")
    private void removeXReferences(EObject eobject, PooledConnection pc) throws SQLException {
        // delete all references from other objects to this object
        Map<EClass, Collection<EReference>> backRefs = EMFUtil.getBackReferences(
                eobject.eClass(), repository.getModelRegistry());
        Set<EObject> referencingObjects = new HashSet<EObject>(); 
        for (EClass eClass : backRefs.keySet()) {
            if (!eClass.isAbstract()) {
                for (EReference reference : backRefs.get(eClass)) {
                    referencingObjects.addAll(EMFSQLUtil.dbLoadReferencingObjects(
                            eobject, eClass, reference, this, pc.getConnection()));
                }
            }
        }
        for (EObject eObj : referencingObjects) {
            EObject referer;
            if (eObj.eIsProxy()) {
                URI uri = ((InternalEObject)eObj).eProxyURI();
                referer = getEObject(uri.fragment());
            } else {
                referer = eObj;
            }
            for (EReference reference : (List<EReference>)referer.eClass().getEAllReferences()) {
                if (referer.eIsSet(reference)) {
                    if (reference.isMany()) {
                        List referents = (List)referer.eGet(reference);
                        referents.remove(eobject);
                    } else if (eobject.equals(referer.eGet(reference))) {
                        referer.eUnset(reference);
                    }
                }
            }
        }
    }

    @Override
    public void load(Map<?, ?> options) throws IOException {
    }
    
    @Override
    protected void doUnload() {
        loaded.clear();
        boolean oldDeliver = eDeliver();
        try {
            eSetDeliver(false);
            getContents().clear();
        } finally {
            eSetDeliver(oldDeliver);
        }
    }
    
    public void save(Map<?, ?> options) throws IOException {
        // This is unneccessary, objects are persisted automatically
    }
    
    @SuppressWarnings("unchecked")
    public void addException(Exception e) {
        getErrors().add(new XMIException(e));
    }
    
    public EObject findEObject(URI proxyURI) {
        return loaded.get(proxyURI);
    }

    public void registerEObject(URI proxyURI, EObject relObj) {
        loaded.put(proxyURI, relObj);
    }  
    
    public void unregisterEObject(URI proxyURI) {
        loaded.remove(proxyURI);
    }  
    
    @Override
    protected XMLSave createXMLSave() {
        return new MyXMISaveImpl(createXMLHelper());
    }

    static class MyXMISaveImpl extends XMISaveImpl {

        public MyXMISaveImpl(XMLHelper helper) { 
            super(helper); 
        } 

        public MyXMISaveImpl(Map<?, ?> options, XMLHelper helper, String encoding) {
            super(options, helper, encoding);
        }

        public Object writeTopObjects(List<? extends EObject> contents) {
            /**
             * Copied implementation from XMISaveImpl wrapping pertinent
             * sections with eIsProxy() calls. XML/XMI Resource
             * implementations seem to never contain proxy objects, I guess
             * the entire model is loaded/saved in one go to/from a file.
             * However, JDBCResourceImpl does lazy load objects.
             */
            if (!toDOM) {
                doc.startElement(XMI_TAG_NS);
                Object mark = doc.mark();
                int size = contents.size();
                for (int i = 0; i < size; i++) {
                    EObject top = (EObject) contents.get(i);
                    if (!top.eIsProxy()) {
                        EClass eClass = top.eClass();
                        String name = helper.getQName(eClass);
                        doc.startElement(name);
                        root = top;
                        saveElementID(top);
                    }
                }

                doc.endElement();
                return mark;
            } else {
                // create dummy documentElement
                currentNode = document.createElementNS(XMIResource.XMI_URI,
                                                       XMI_TAG_NS);
                document.appendChild(currentNode);
                int size = contents.size();
                for (int i = 0; i < size; i++) {
                    EObject top = (EObject) contents.get(i);
                    if (!top.eIsProxy()) {
                        EClass eClass = top.eClass();
                        helper.populateNameInfo(nameInfo, eClass);
                        currentNode = 
                            currentNode.appendChild(document.createElementNS(
                                                                             nameInfo.getNamespaceURI(),
                                                                             nameInfo.getQualifiedName()));
                        handler.recordValues(currentNode, null, null, top);
                        root = top;
                        saveElementID(top);
                    }
                }
                return null;
            }
        }
    }
} 
