/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.resourceset.handler;

import java.sql.SQLException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.soa.expression.Equals;
import com.iona.soa.expression.Expression;
import com.iona.soa.expression.Operand;
import com.iona.soa.expression.Operation;
import com.iona.soa.expression.OperationType;
import com.iona.soa.expression.Term;
import com.iona.soa.expression.delegate.CrossReferencesOperationDelegate;
import com.iona.soa.expression.delegate.DependenciesOperationDelegate;
import com.iona.soa.expression.delegate.JDBCOperationDelegate;
import com.iona.soa.expression.delegate.ReadOnlyOperationDelegate;
import com.iona.soa.expression.delegate.SQLRenderer;
import com.iona.soa.expression.serialization.ExpressionSerializer;
import com.iona.soa.repository.EMFUtil;
import com.iona.soa.repository.jdbc.resource.JDBCResourceImpl;
import com.iona.soa.resourceset.QueryConstants;
import com.iona.soa.resourceset.ResourceQueryHandler;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;

public class JDBCResourceQueryHandler implements ResourceQueryHandler {
    private static final Logger LOG = Logger.getLogger(ResourceQueryHandler.class.getName());
    private static final Map<OperationType, JDBCOperationDelegate.Factory> OP_FACTORY_REGISTRY =
        new HashMap<OperationType, JDBCOperationDelegate.Factory>();
    
    static {
        OP_FACTORY_REGISTRY.put(OperationType._xrefs, 
                                     new CrossReferencesOperationDelegate.Factory());
        OP_FACTORY_REGISTRY.put(OperationType._deps, 
                                     new DependenciesOperationDelegate.Factory());
        OP_FACTORY_REGISTRY.put(OperationType._readonly, 
                                new ReadOnlyOperationDelegate.Factory());
        
    }
    public Collection<EObject> executeQuery(Resource resource, EClass type) throws Exception {
        if (!(resource instanceof JDBCResourceImpl)) {
            throw new IllegalArgumentException("Should be a JDBCResource: " + resource);
        }
        JDBCResourceImpl jdbcResource = (JDBCResourceImpl) resource;
        // A bit silly: a where clause that does nothing...
        SQLRenderer.WhereClause whereClause = new SQLRenderer.WhereClause("1=1");
        
        return jdbcResource.queryObjects(type, whereClause);
    }

    public Collection<EObject> executeQuery(Resource resource, EClass type,
            Expression expression) throws SQLException {
        if (!(resource instanceof JDBCResourceImpl)) {
            throw new IllegalArgumentException("Should be a JDBCResource: " + resource);
        }
        JDBCResourceImpl jdbcResource = (JDBCResourceImpl) resource;
        
        SQLRenderer queryRenderer = new SQLRenderer(type);
        if (queryRenderer.handle(expression)) {
            return jdbcResource.queryObjects(type, queryRenderer.getWhereClause());
        } else {
            LOG.log(Level.WARNING, "Invalid query expression: " + expression.toString());
            return Collections.emptyList();
        }
    }

    @SuppressWarnings("unchecked")
    public Collection<EObject> executeQuery(Resource resource, EClass type,
            Operation operation) throws SQLException {
        if (!(resource instanceof JDBCResourceImpl)) {
            throw new IllegalArgumentException("Should be a JDBCResource: " + resource);
        }
        JDBCResourceImpl jdbcResource = (JDBCResourceImpl) resource;

        Operand[] arguments = operation.getArguments();
        if (OperationType._GETXREFS.equals(operation.getName())) {
            if (arguments.length == 1 && arguments[0].getValue() instanceof String) {
                String guid = (String)arguments[0].getValue();
                URI uri = EMFUtil.getProxyURI(type, guid);
                EObject eObject = resource.getResourceSet().getEObject(uri, true);
                if (eObject != null) {
                    return jdbcResource.getXRefs(eObject, type);
                }
            }
        }
        return Collections.EMPTY_LIST;
    }

    @SuppressWarnings("unchecked")
    public Collection<EObject> executeQuery(Resource resource, Operation operation) throws Exception {
        if (!(resource instanceof JDBCResourceImpl)) {
            throw new IllegalArgumentException("Should be a JDBCResource: " + resource);
        }
        JDBCResourceImpl jdbcResource = (JDBCResourceImpl) resource;
        JDBCOperationDelegate.Factory factory = 
            OP_FACTORY_REGISTRY.get(operation.getName());
        if (null == factory) {
            throw new UnsupportedOperationException("The following operation is not supported: " 
                                                    + operation.getName());
        }
        JDBCOperationDelegate delegate = factory.create(jdbcResource);
        if (!delegate.handle(operation)) {
            throw delegate.getException();
        }
        return delegate.getResult();
    }

    public Collection<EObject> executeQuery(Resource resource, EClass type, String query, char language) 
        throws Exception {
        switch (language) {
        case QueryConstants.QUERY_LANGUAGE_SQL:
            if (!(resource instanceof JDBCResourceImpl)) {
                throw new IllegalArgumentException("Should be a JDBCResource: " + resource);
            }
            JDBCResourceImpl jdbcResource = (JDBCResourceImpl) resource;

            return jdbcResource.queryObjects(type, query);
        case QueryConstants.QUERY_LANGUAGE_IT_EMFQL:
            Term expr = ExpressionSerializer.deserialize(query);
            if (expr instanceof Expression) {
                return executeQuery(resource, type, (Expression) expr);                
            } else if (expr instanceof Operation) {
                return executeQuery(resource, type, (Operation) expr);
            } else {
                throw new UnsupportedOperationException("Cannot handle the following query: " + query);
            }
        default:
            throw new UnsupportedOperationException("Unsupported query language: " + language);        
        }
    }

    public EObject executeQuery(Resource resource, EClass type, Object id) throws Exception {
        EAttribute idAttr = type.getEIDAttribute();
        if (idAttr == null) {
            return null;
        }
        if (id == null) {
            throw new NullPointerException("Entity ID cannot be null: " + idAttr.getName());
        }
        
        Collection<EObject> result = executeQuery(resource, type, new Equals(idAttr.getName(), id));
        for (EObject obj : result) {
            return obj;
        }
        // if result didn't contain anything
        return null;
    }
}
