/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.db;

import java.lang.reflect.Constructor;
import java.sql.SQLException;
import java.util.StringTokenizer;


/*
 * This is a generic database wrapper resolver.  
 * Database wrappers are resolved based on jdbc url's content and a this project's package layout.
 * If no specific wrapper class for a given db type is found the generic wrapper 
 * com.iona.repository.db.DatabaseImpl is used instead
 * 
 * The class is found using the following algorithm:
 * - The second token of jdbc url is used as db type (e.g. jdbc:mysql:....)
 * - The current package is appended with the db type and the class name,
 *   which is a litteral "Embedded" and the db type that has its first letter capitalized 
 * 
 */
public class DatabaseResolver {

    private static final String JDBC_PREFIX  = new String("jdbc");
    private static final String JDBC_DELIM = new String(":");
    private static final String JDBC_PROP_DELIM = new String(";?");
    private static final String DB_CLASS_PREFIX = new String("Embedded");
    private static final StringBuffer DB_GENERIC_WRAPPER 
        = new StringBuffer(DatabaseResolver.class.getPackage().getName()).append(".DatabaseImpl");
    
    protected DatabaseResolver() {
    }
    
    public static Database resolve(
        String driver, 
        String connectUrl, 
        String name, 
        String passwd
    ) throws SQLException {
        String dbType = getDatabaseTypeFromURL(connectUrl);
        String upperCaseDbType = dbType.substring(0, 1).toUpperCase() +  dbType.substring(1, dbType.length());
        // generate the target Database wrapper class name
        StringBuilder dbClassName = new StringBuilder().append(DatabaseResolver.class.getPackage().getName());
        dbClassName.append('.').append(dbType).append('.').append(DB_CLASS_PREFIX).append(upperCaseDbType);
        Class dbClass = null;
        try {
            dbClass = Class.forName(dbClassName.toString());
        } catch (ClassNotFoundException e) {
            // resort to the genereic JDBC wrapper
            try {
                dbClass = Class.forName(DB_GENERIC_WRAPPER.toString());
            } catch (ClassNotFoundException nfe) {
                throw new SQLException("Failed to resolve database supporting runtime due to: " 
                                       + nfe.getMessage() == null ? nfe.toString() : nfe.getMessage());
            }
        } 
        try {
            Constructor ctor = 
                dbClass.getConstructor(new Class[] {String.class, String.class, String.class, String.class});
            return (Database)ctor.newInstance(new Object[] {driver, connectUrl, name, passwd});
        } catch (Exception e) {
            throw new SQLException("Failed to resolve database supporting runtime due to: " 
                                   + e.getMessage() == null ? e.toString() : e.getMessage());
        }
    }
    
    public static String getDatabaseTypeFromURL(String url) throws SQLException {
        StringTokenizer tz = new StringTokenizer(url, JDBC_DELIM);
        if (tz.countTokens() < 2 && !tz.nextToken().equals(JDBC_PREFIX)) {
            throw new SQLException("Malformed jdbc connection string: " + url);
        }

        tz.nextToken();
        return tz.nextToken();        
    }
    
    public static String getDatabaseHostFromURL(String url) throws SQLException {
        StringTokenizer tz = new StringTokenizer(url, JDBC_DELIM);
        if (tz.countTokens() < 2 && !tz.nextToken().equals(JDBC_PREFIX)) {
            throw new SQLException("Malformed jdbc connection string: " + url);
        }
        tz.nextToken();
        tz.nextToken();
        String host = tz.nextToken();
        return (null == host || -1 == host.indexOf("//")) ? "localhost" : host;
    }
    
    public static String getDatabaseNameFromURL(String url) throws SQLException {
        StringTokenizer tz = new StringTokenizer(url, JDBC_DELIM);
        if (tz.countTokens() < 3 && !tz.nextToken().equals(JDBC_PREFIX)) {
            throw new SQLException("Malformed jdbc connection string: " + url);
        }
        tz.nextToken();
        tz.nextToken();
        String name = tz.nextToken();
        // is this hostname?
        if (-1 == name.indexOf("//")) {
            name = tz.nextToken();
        }
        StringTokenizer nametz = new StringTokenizer(name, JDBC_PROP_DELIM);
        if (nametz.countTokens() < 1) {
            throw new SQLException("Jdbc connection string missing database name: " + url);
        }
        return nametz.nextToken();
    }
    
}


