/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.maven.index;

import java.io.File;
import java.io.IOException;
import java.net.SocketTimeoutException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.repository.maven.util.MavenUtils;
import com.iona.repository.search.SearchIndex;
import com.iona.soa.model.MavenRepository;
import com.iona.soa.repository.util.Out;

import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.CorruptIndexException;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.queryParser.ParseException;
import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.manager.WagonManager;
import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.artifact.repository.DefaultArtifactRepository;
import org.apache.maven.artifact.repository.layout.DefaultRepositoryLayout;
import org.apache.maven.embedder.MavenEmbedder;
import org.apache.maven.embedder.MavenEmbedderException;
import org.apache.maven.model.Dependency;
import org.apache.maven.model.Model;
import org.apache.maven.wagon.Wagon;
import org.apache.maven.wagon.WagonException;
import org.apache.maven.wagon.providers.file.FileWagon;
import org.apache.maven.wagon.repository.Repository;
import org.codehaus.plexus.util.xml.pull.XmlPullParserException;

public class Maven2Indexer extends SearchIndex {

    private static final Logger LOG = Logger.getLogger(Maven2Indexer.class.getName());

    private static final String TRAILING_SLASH = "/";
    
    private static final String[] CHECKSUM_IDS = {"md5", "sha1"};

    private static final String ROOT_DIR = "/";

    private static final String PARENT_DIR = "../";
    
    private String mavenSettingsPath;
    
    public Maven2Indexer(String indexStorageDir, String settingsPath) throws IOException {
        super(indexStorageDir);
        this.mavenSettingsPath = settingsPath;
    }

    @SuppressWarnings("deprecation")
    public void index(MavenRepository mavenRepo) throws ParseException {
        index(mavenRepo, null);
    }

    @SuppressWarnings("deprecation")
    public void index(MavenRepository mavenRepo, String specificDirIndex) throws ParseException {
        
        LOG.info("The Indexer may take a while , " 
                + "depending on the size of the repository(s) specified in your configuration");
        String repoURL = mavenRepo.getUrl().toString();
        
//        LOG.info("Creating the index at " + getIndexDir(mavenRepo));
        
        try {
            
            ArtifactRepository artifactRepository 
                = new DefaultArtifactRepository(mavenRepo.getId(),
                                            repoURL,
                                            new DefaultRepositoryLayout());
        
            MavenEmbedder embedder = createEmbedder();
        
            WagonManager wagonManager =
                (WagonManager)embedder.getPlexusContainer().lookup(WagonManager.class);
        
            Repository wagonRepository = new Repository(mavenRepo.getId(), 
                                                        removeTrailingSlash(repoURL));
            Wagon wagon = wagonManager.getWagon(wagonRepository);
            wagon.connect(wagonRepository);
        
            createIndex(mavenRepo,
                        embedder, 
                        artifactRepository, 
                        wagon,
                        specificDirIndex);
        } catch (Exception ex) {
            LOG.log(Level.SEVERE, "Error during indexing", ex);
        }
        
//        return getIndexDir(mavenRepo);
    }

    @SuppressWarnings("unchecked")
    void createIndex(MavenRepository mavenRepo,
                     MavenEmbedder mavenEmbedder,
                     ArtifactRepository artifactRepository,
                     Wagon wagon) {
        createIndex(mavenRepo, mavenEmbedder, artifactRepository, wagon, null);
    }
    
    @SuppressWarnings("unchecked")
    void createIndex(MavenRepository mavenRepo,
                     MavenEmbedder mavenEmbedder,
                     ArtifactRepository artifactRepository,
                     Wagon wagon,
                     String specificDirIndex) {
        
        IndexWriter writer = null;
        try {
            writer = openIndexWriter();
        } catch (IOException ioe) {
            LOG.log(Level.WARNING, "Error opening lucene index writer" , ioe);
            return;
        }
        
        final Date startTime = new Date();
        Out<Integer> numOfArtifacts = new Out<Integer>(0);
        
        try {
            LOG.log(Level.INFO,
                    "Commencing index of {0} ({1})",
                    new Object[] {mavenRepo.getName(), mavenRepo.getUrl()});
            
            final Collection<ArtifactRepository> artifactRepos 
                = Collections.unmodifiableList(Collections.singletonList(artifactRepository));
            final String repoURL = removeTrailingSlash(mavenRepo.getUrl().toString());
            final String repoGuid = mavenRepo.getGuid();
            
            List<String> toSearch = initialiseToSearchDir(specificDirIndex);
            Collection<String> alreadySearched = new HashSet<String>();
            Collection<String> visitedPoms = new ArrayList<String>();

    
            int foldersSearched = 1;
            
            while (toSearch.size() > 0) {
                final String folderPath = (String)toSearch.remove(0);
                
                logFolderInfo(foldersSearched, toSearch.size() + 1, numOfArtifacts.get(),
                    wagon.getRepository().getUrl() + folderPath.toString());
                
                // get the first element from the to be searched list
                // mark the URL as searched (we want this one way or the other)
                alreadySearched.add(folderPath);
                
                Collection<String> fileList = null;
                try {
                    foldersSearched++;
                    fileList = wagon.getFileList(folderPath);
                } catch (Throwable t) {
                    if (t.getCause() instanceof SocketTimeoutException) {
                        LOG.log(Level.WARNING, "Timeout reading directory " + repoURL + folderPath);
                    } else if (LOG.isLoggable(Level.FINE)) {
                        LOG.log(Level.WARNING, "Error reading directory " + repoURL + folderPath, t);
                    } else {
                        LOG.log(Level.WARNING, 
                                "Error reading directory " + repoURL + folderPath + " - " + t);
                    }
                    continue;
                }
                for (String file : fileList) {
                    String filePath = folderPath + file;

                    if (!alreadySearched.contains(filePath)
                        && !toSearch.contains(filePath)
                        && !addDirectory(file, filePath, toSearch, wagon)
                        && !visitedPoms.contains(filePath)
                        && file.endsWith("pom")) {
                        visitedPoms.add(filePath);

                        try {
                            File pom = retrievePom(mavenEmbedder, wagon, folderPath, filePath);
                            Model model = null;
                            try {
                                model = mavenEmbedder.readModel(pom);
                            } catch (XmlPullParserException ex) {
                                LOG.log(Level.FINE, "Problem reading pom model for " + filePath, ex);
                            }


                            if (model == null || !model.getPackaging().equals("pom")) {
                                Document doc = processPom(repoURL, 
                                                          repoGuid, 
                                                          model, 
                                                          mavenEmbedder, 
                                                          artifactRepos, 
                                                          filePath,
                                                          fileList);
                                storeDocument(writer, doc, numOfArtifacts);     
                            }
                        } catch (Exception ex) {
                            if (ex.getCause() instanceof SocketTimeoutException) {
                                LOG.log(Level.WARNING, "Timeout processing pom " + repoURL + filePath);
                            } else if (LOG.isLoggable(Level.FINE)) {
                                LOG.log(Level.WARNING, "Error processing pom " + repoURL + filePath, ex);
                            } else {
                                LOG.log(Level.WARNING, 
                                        "Error processing pom " + repoURL + filePath + " - " + ex);
                            }
                        }
                    }
                }
            }
        } finally {
            try {
                closeIndexWriter();
            } catch (IOException ex) {
                LOG.log(Level.WARNING, "Error closing lucene index writer" , ex);
            }
            
            logIndexingSummary(mavenRepo, startTime, numOfArtifacts.get());
        }
    }
    
    private boolean addDirectory(String file, String filePath, List<String> toSearch, Wagon wagon) {
        boolean isDirectory = false;
        
        if (file.endsWith(TRAILING_SLASH) && !file.endsWith(PARENT_DIR)) {
            isDirectory = true;
        } else if (wagon instanceof FileWagon) { 
            // Check to see if this is a directory when using FileWagon, maybe we
            // should use this to cover all transports
            try {
                 
                filePath += "/";
                wagon.getFileList(filePath);
                isDirectory = true;
            } catch (Exception rdnee) {
                isDirectory = false;
            }
            
        }
        if (isDirectory) {
            toSearch.add(filePath);
        }
        
        return isDirectory;
    }

    private void storeDocument(IndexWriter writer, Document doc,
                               Out<Integer> numOfArtifacts) throws CorruptIndexException, IOException {
        if (doc != null) {
            if (LOG.isLoggable(Level.INFO)) {
                StringBuilder sb = new StringBuilder();
                sb.append("Added document to index:\n");
            
                for (Object o : doc.getFields()) {
                    Field f = (Field) o;
                    sb.append("    ").append(f.name())
                       .append(" =\"").append(f.stringValue())
                       .append("\"");            
                }
                LOG.info(sb.toString());
            }
            writer.addDocument(doc);
            numOfArtifacts.set(numOfArtifacts.get() + 1);
        }
    }

    private List<String> initialiseToSearchDir(String specificDirIndex) {
        List<String> toSearch = new ArrayList<String>();
        if (specificDirIndex == null) {
            toSearch.add(ROOT_DIR);
        } else {
            toSearch.add(ROOT_DIR + specificDirIndex);
        }
        return toSearch;
    }
    
    private File retrievePom(MavenEmbedder mavenEmbedder,
            org.apache.maven.wagon.Wagon wagon, String strURL, String fullPath)
        throws WagonException {
        File mavenRepository = new File(mavenEmbedder.getLocalRepository().getBasedir());
        String abPath = mavenRepository.getAbsolutePath();
        
        // pomDir
        File pomDir = new File(abPath + strURL);
        pomDir.mkdirs();
        // pom file 
        String pomFileUrl = abPath + fullPath;
        File pom = new File(pomFileUrl);
        if (!pom.exists()) {
            pomDir.mkdirs();
        }
        wagon.get(fullPath, pom);
        return pom;
    }
    
    private void logIndexingSummary(MavenRepository mavenRepo, 
                                    Date startTime, 
                                    int numOfArtifacts) {
        Date endTime = new Date();
        long totalTime = endTime.getTime() - startTime.getTime();
        LOG.log(Level.INFO,
                "Completed indexing for ", 
                mavenRepo.getName());
        LOG.log(Level.INFO,
                "Indexed {0} artifacts in {1} ms", 
                 new Object[] {numOfArtifacts, totalTime});
    }
    
    private void logFolderInfo(int foldersSearched, int vectorSize, int numOfArtifacts,
                String folderBeingIndexed) {

        LOG.info("Folders Searched = " + foldersSearched);
        LOG.info("Known folders left to search = " + vectorSize);
        LOG.info("Number of artifacts indexed so far = " + numOfArtifacts);
        LOG.info("Indexing " + folderBeingIndexed);
    }
    
    @SuppressWarnings("unchecked")
    private Document processPom(String repoURL,
                               String repoGuid,
                               Model inModel,
                               MavenEmbedder mavenEmbedder,
                               Collection<ArtifactRepository> repos,
                               String filePath,
                               Collection<String> fileList) {
        String groupId = null;
        String version = null;
        String packaging = null;
        Collection<Dependency> dependencies = new ArrayList<Dependency>();
        String artifactId = null;
        if (inModel != null) {
            groupId = inModel.getGroupId();
            version = inModel.getVersion();
            packaging = inModel.getPackaging();
            dependencies = (List<Dependency>)inModel.getDependencies();
            artifactId = inModel.getArtifactId();
            
        }
        
        if (groupId == null) {
            if (inModel != null && inModel.getParent() != null) {
                groupId = inModel.getParent().getGroupId();  
            } else {
                groupId = getGroupIdFromPath(filePath, groupId); 
            }
            
        } else if (groupId.contains("${") && groupId.contains("}")) {
            groupId = getGroupIdFromPath(filePath, groupId);
        }
        
        
        if (version == null) {
            if (inModel != null && inModel.getParent() != null) {
                version = inModel.getParent().getVersion();
            } else {
                version = getVersionFromPath(filePath);
            }
        } else if (version.contains("${") && version.contains("}")) {
            version = getVersionFromPath(filePath);
        }
        
        if (packaging == null || packaging.equals("bundle")) {
            packaging = "jar";
        }

        if (artifactId == null) {
            artifactId = getArtifactIdFromPath(filePath);
        }
        Artifact artifact = null;
        String fileExtension = MavenUtils.guessFileExtensionFromMavenPackaging(packaging);

        if (packaging.equals("jbi-component")) {
            artifact = mavenEmbedder.createArtifactWithClassifier(groupId,
                                                                  artifactId,
                                                                  version,
                                                                  fileExtension,
                                                                  "installer");
        } else if (packaging.equals("jbi-service-assembly")) {
            artifact = mavenEmbedder.createArtifact(groupId,
                                                    artifactId,
                                                    version,
                                                    "compile",
                                                    fileExtension);
            
        } else {
            artifact = mavenEmbedder.createArtifact(groupId,
                                                    artifactId,
                                                    version,
                                                    "compile",
                                                    fileExtension);
        }
        
        
        int startIndex = filePath.lastIndexOf("/") + 1;
        int endIndex = filePath.lastIndexOf(".pom");
        String fileNoExtention = filePath.substring(startIndex == -1 ? 0 : startIndex, 
                    endIndex == -1 ? 1 : endIndex);
        

        String bestGuess = fileNoExtention + "." + packaging;
        
        if (!fileList.contains(bestGuess)) {
            for (String name : fileList) {
                // Is it a directory?
                if (name.endsWith(TRAILING_SLASH) || name.endsWith(CHECKSUM_IDS[0]) 
                        || name.endsWith(CHECKSUM_IDS[1])
                        || name.endsWith("pom")) {
                    continue;
                }
                
                if (name.startsWith(fileNoExtention + ".")) {
                    bestGuess = name;
                    break;
                }
                
            }
        }
        
        if (fileList.contains(bestGuess)) {
        
            String artifactLocation = getArtifactLocation(repoURL, 
                    filePath, 
                    bestGuess);
            try {
                return createDocument(artifactLocation,
                           artifact,
                           packaging,
                           dependencies,
                           repoGuid);
            } catch (IOException ioe) {
                LOG.warning("Error adding artifact to index : " + groupId
                            + ":" + artifact.getArtifactId()
                            + ":" + version
                            + "\n " + ioe.toString());
            }
        }
        return null;
    }

    private String getVersionFromPath(String filePath) {
        String version;
        int endIndex = filePath.lastIndexOf("/");
        String intermPath = filePath.substring(0, endIndex == -1 ? 1 : endIndex);
        int startIndex = intermPath.lastIndexOf("/") == -1 ? 0 : intermPath.lastIndexOf("/");
        version = filePath.substring(startIndex == -1 ? 0 : startIndex + 1, 
                endIndex == -1 ? 0 : endIndex - 1);
        return version;
    }
    
    private String getArtifactIdFromPath(String filePath) {
        String artifactId;
        String version = getVersionFromPath(filePath);
        int startIndex = filePath.lastIndexOf("/");
        int endIndex = filePath.lastIndexOf(version);
        artifactId = filePath.substring(startIndex + 1, endIndex - 1);
        return artifactId;
    }

    private String getGroupIdFromPath(String filePath, String groupId) {
        int index = filePath.lastIndexOf("/");
        if (index != -1) {
            groupId = filePath.substring(0, index);
            index = groupId.lastIndexOf("/");
            if (index != -1) {
                groupId = filePath.substring(0, index);
                groupId.replace("/", ".");
            }
        }
        return groupId;
    }

    private String getArtifactLocation(String repoURL, String pomLocation, String artifactName) {
        int breakAt = pomLocation.lastIndexOf('/') + 1;
        String location = repoURL + pomLocation.substring(0, breakAt) + artifactName;
        return location;
    }
    
    private Document createDocument(String strURL,
                            Artifact artifact,
                            String artifactPackaging,
                            Collection<Dependency> deps,
                            String repoGuid) throws IOException {
        Document doc = new Document();
        doc.add(new Field("artifactId", artifact.getArtifactId(), Field.Store.YES, Field.Index.ANALYZED));
        
        String groupId = artifact.getGroupId();
        doc.add(new Field("groupId", groupId, Field.Store.YES, Field.Index.ANALYZED));
        
        String version = artifact.getVersion();
        doc.add(new Field("version", version, Field.Store.YES, Field.Index.ANALYZED));
        
        doc.add(new Field("type", artifact.getType(), Field.Store.YES, Field.Index.ANALYZED));
        
        if (artifactPackaging != null) {
            doc.add(new Field("packaging", artifactPackaging, Field.Store.YES, Field.Index.ANALYZED));
        }
        
        String classifier = artifact.getClassifier();
        if (classifier != null) {
            doc.add(new Field("classifier", classifier, Field.Store.YES, Field.Index.ANALYZED));
        }
        
        doc.add(new Field("url", strURL, Field.Store.YES, Field.Index.ANALYZED));
        
        String dependencies = "";
        for (Dependency dep : deps) {
            dependencies += dep.getGroupId() + ":" 
                         + dep.getArtifactId() + ":"
                         + dep.getVersion() + ":"
                         + dep.getClassifier() + ":"
                         + dep.getType() + ":"
                         + dep.getScope() + ","; 
        }
        doc.add(new Field("dependencies", dependencies, Field.Store.YES, Field.Index.NO));
        
        doc.add(new Field("mavenRepoGUID", repoGuid, Field.Store.YES, Field.Index.ANALYZED));
        return doc;
    }
    
    String removeTrailingSlash(String path) {
        if (path.lastIndexOf(TRAILING_SLASH) == (path.length() - 1)) {
            path = path.substring(0, path.length() - 1);
        }
        return path;
    }
    
    private MavenEmbedder createEmbedder() throws MavenEmbedderException {
        return MavenUtils.getInstance().getMavenEmbedder(mavenSettingsPath);
    }

}
