/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.maven.osgi;

import java.io.IOException;
import java.sql.Timestamp;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.repository.maven.index.Maven2Indexer;
import com.iona.repository.maven.resolve.Resolver;
import com.iona.repository.maven.settings.MavenServiceSettings;

import com.iona.soa.model.IndexStates;
import com.iona.soa.model.MavenRepository;
import com.iona.soa.model.SOA_NetworkPackage;
import com.iona.soa.repository.EMFUtil;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.SOARepositoryService;
import com.iona.soa.repository.TransactionData;
import com.iona.soa.repository.configuration.ConfigurationService;
import com.iona.soa.resourceset.ResourceSetQueryHandler;
import com.iona.soa.security.SecurityService;

import org.apache.lucene.document.Document;
import org.apache.lucene.queryParser.ParseException;
import org.apache.maven.shared.dependency.tree.DependencyNode;
import org.codehaus.plexus.component.repository.exception.ComponentLookupException;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.springframework.beans.factory.InitializingBean;

public class RepositoryMavenServiceBean implements IRepositoryMavenService, InitializingBean {
    private static final Logger LOG = Logger.getLogger(RepositoryMavenServiceBean.class.getName());

    // initialise parameters
    String configFile = System.getProperty(MavenServiceSettings.REPO_CONFIG_KEY);

    private String indexStorageDir;
    private String settings;
    private ConfigurationService configurationService;
    private Resolver resolver;
    private int maxThreadCount = 1;
    private int threadCount;
    private SOARepository soaRepo;
    private Map<String, Thread> repoIndexThread = new HashMap<String, Thread>();
    private Map<String, Maven2Indexer> indices = new HashMap<String, Maven2Indexer>();
    
    public void setRepositoryService(SOARepositoryService repositoryService) {
        soaRepo = repositoryService.getRepository();
        LOG.log(Level.FINE, "SOARepository: {0}", soaRepo);
    }
    
    public void setSOARepository(SOARepository soaRepository) {
        LOG.fine("Setting the repository explicitly");
        this.soaRepo = soaRepository;
    }
    
    
    
    public void setIndexDestinationDir(String indexDir) {
        indexStorageDir = indexDir;
    }
   
    public ConfigurationService getConfigurationService() {
        return this.configurationService;
    }
    
    public void setConfigurationService(ConfigurationService configurationService) {
        this.configurationService = configurationService;
        LOG.log(Level.FINE, "configurationService: {0}", configurationService);
        indexStorageDir = configurationService.getConfiguration()
                                                  .get(MavenServiceSettings.INDEX_DIRECTORY_KEY);
        settings = configurationService.getConfiguration()
                                       .get(MavenServiceSettings.MAVEN_SETTINGS_LOCATION_KEY);
    }

    public void afterPropertiesSet() throws Exception {
        resolver = new Resolver();
        threadCount = 0;
        
        if (soaRepo != null) {
            // reset maven repo states
            List<EObject> mvnReposEobjs = 
                ResourceSetQueryHandler.executeQuery(soaRepo.getResourceSet(), 
                                                     SOA_NetworkPackage.eINSTANCE.getMavenRepository(), 
                                                     true);
            TransactionData txData = new TransactionData(SecurityService.INTERNAL_USER,
                                                         SecurityService.INTERNAL_TOKEN);
            soaRepo.startWrite(txData);
            boolean txPassed = false;
            try {
                for (EObject eobject : mvnReposEobjs) {
                    MavenRepository mavenRepo = (MavenRepository)eobject;
                    if (mavenRepo.getIndexState() == IndexStates.INDEXING) {
                        mavenRepo.setIndexState(IndexStates.FAILED);
                    }
                }
                txPassed = true;
            } finally {
                soaRepo.endWrite(txPassed);
            }
        }
    }
    
    public void shutdown() {
        synchronized (indices) {
            for (String key : indices.keySet()) {
                Maven2Indexer indexer = indices.remove(key);
                try {
                    indexer.close();
                } catch (IOException e) {
                    // ignore
                }
            }
        }
        
    }
    
    public DependencyNode resolve(String groupid,
                                  String artifactId,
                                  String version,
                                  String scope,
                                  List<MavenRepository> repos) throws ParseException, IOException {
        try {
            return resolver.resolve(settings, groupid, artifactId, version, scope, repos);
        } catch (ComponentLookupException e) {
            LOG.log(Level.WARNING, "Unable to resolve artifact",  e);
            return null;
        }
    }

    public void waitForIndexThreadToFinish(String guid) {
        Thread indexThread = repoIndexThread.get(guid);
        try {
            indexThread.join();
        } catch (InterruptedException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    public void index(String user, String token, String guid, boolean async) throws ParseException {
        index(user, token, guid, null, async);
    }
    
    public void index(String user, String token, String guid, String specifiedDir, boolean async) 
        throws ParseException {
        
        if (async && threadCount > maxThreadCount) {
            return;
        }

        IndexThread index = new IndexThread(new TransactionData(user, token), guid, specifiedDir);
        if (async) {
            // start a new thread 
            repoIndexThread.put(guid, index);
            index.start();
        } else {
            index.run();
        }
    }

    public Iterator<Document> search(String query, MavenRepository mavenRepo) {
        if (mavenRepo.getIndexState().equals(IndexStates.READY)) {
            LOG.log(Level.INFO, "Querying being passed executed: " + query);
            try {
                return getSearchIndex(mavenRepo.getId()).queryIndex(query);
            } catch (Exception e) {
                LOG.log(Level.WARNING, "Exception while searching index", e);
            }
        }
        return null;
    }
    
    public Maven2Indexer getSearchIndex(String id) throws IOException {
        synchronized (indices) {
            Maven2Indexer index = indices.get(id);
            if (index == null) {
                index = new Maven2Indexer(indexStorageDir + "/" + id, settings);
                indices.put(id, index);
            }
            return index;
        }
    }

    private class IndexThread extends Thread {
        private String mavenRepositoryGuid;
        private TransactionData trnsData;
        private String specificDirToIndex;

        public IndexThread(TransactionData td, String guid, String specificDir) {
            mavenRepositoryGuid = guid;
            trnsData = td;
            specificDirToIndex = specificDir;
        }

        protected MavenRepository getMavenRepo(String id) {
            EClass eclass = SOA_NetworkPackage.eINSTANCE.getMavenRepository();
            URI uri = EMFUtil.getProxyURI(eclass, id);
            try {
                MavenRepository ret = (MavenRepository)soaRepo.getResourceSet().getEObject(uri, true);
                if (ret == null) {
                    LOG.log(Level.WARNING, "Unable to load maven repo " + id);
                }
                return ret;
            } catch (IOException e) {
                LOG.log(Level.WARNING, "Unable to load maven repo " + id, e);
            }
            return null;
        }

        public void run() {
            LOG.log(Level.FINE, "Running Index");

            MavenRepository mvnRepo = getMavenRepo(mavenRepositoryGuid);
            if (mvnRepo == null) {
                return;
            }
            if (mvnRepo.getIndexState().equals(IndexStates.INDEXING)) {
                LOG.info("Maven repo index already in progress");
                return;
            }

            boolean txPassed = false;
            soaRepo.startWrite(trnsData);
            try {
                mvnRepo.setIndexState(IndexStates.INDEXING);
                txPassed = true;
            } finally {
                soaRepo.endWrite(txPassed);
            }

            boolean indexPassed = false;
            try {
                getSearchIndex(mvnRepo.getId()).index(mvnRepo, specificDirToIndex);
                indexPassed = true;
            } catch (Exception e) {
                LOG.log(Level.WARNING, "Unable to index maven repository", e);
            } finally {
                txPassed = false;
                
                // use internal token once the indexing is done since the user SSO Token 
                // may have timed out.
                
                TransactionData stopTrans = new TransactionData(SecurityService.INTERNAL_USER,
                        SecurityService.INTERNAL_TOKEN);
                soaRepo.startWrite(stopTrans);

                try {
                    if (indexPassed) {
                        mvnRepo.setIndexState(IndexStates.READY);
                        mvnRepo.setLastIndexed(new Timestamp(new Date().getTime()));
                    } else {
                        mvnRepo.setIndexState(IndexStates.FAILED);
                    }
                    txPassed = true;
                } finally {
                    soaRepo.endWrite(txPassed);
                }
            }
        }
    }
}
