/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.model.util;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import com.iona.soa.expression.Expression;
import com.iona.soa.expression.delegate.BeanExpressionHandler;
import com.iona.soa.expression.serialization.ExpressionSerializer;
import com.iona.soa.model.repository.Property;
import com.iona.soa.model.repository.RepositoryFactory;
import com.iona.soa.model.repository.Version;
import com.iona.soa.repository.ModelRegistry;
import com.iona.soa.repository.SchemaVersionChecker;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

public final class ModelHelper {
    private static final ModelRegistry MODEL_REGISTRY_SINGLETON = new ModelRegistry(
            SchemaVersionChecker.getCodeSchemaVersion(),
            getEPackages("com.iona.soa.model.repository.RepositoryPackage",
                         "com.iona.soa.model.folder.FolderPackage",
                         "com.iona.soa.model.security.SecurityPackage",
                         "com.iona.soa.model.userattributes.UserAttributesPackage",
                         "com.iona.soa.model.notification.NotificationPackage",
                         "com.iona.soa.model.collaboration.CollaborationPackage",
                         "com.iona.soa.model.scheduling.SchedulingPackage",
                         "com.iona.soa.model.SOA_NetworkPackage"));
    
    private ModelHelper() {
    }
    
    public static ModelRegistry getSOAModelRegistry() {
        return MODEL_REGISTRY_SINGLETON;
    }

    public static Map<String, String> getProperties(EObject obj) {
        Map<String, String> map = new HashMap<String, String>();
        for (Property property : getPropertiesFromEObject(obj)) {
            map.put(property.getName(), property.getValue());
        }
        return map;
    }
    
    public static String getProperty(EObject obj, String propertyName, String defaultVal) {
        for (Property property : getPropertiesFromEObject(obj)) {
            if (propertyName.equals(property.getName())) {
                return property.getValue();
            }
        }
        return defaultVal;
    }
    
    public static boolean hasProperty(EObject obj, String propertyName) {
        for (Property property : getPropertiesFromEObject(obj)) {
            if (propertyName.equals(property.getName())) {
                return true;
            }
        }
        return false;
    }
    
    public static Property setProperty(EObject obj, String name, String value) {
        return setProperty(obj, newProperty(name, value));
    }
    
    public static Property setProperty(EObject obj, Property property) {
        Property previous = null;
        List<Property> properties = getPropertiesFromEObject(obj);
        for (Iterator<Property> iter = properties.iterator(); iter.hasNext();) {
            Property p = iter.next();
            if (property.getName().equals(p.getName())) {
                previous = p;
                iter.remove();
                break;
            }
        }
        
        properties.add(property);
        return previous;
    }
    
    @SuppressWarnings("unchecked")
    private static List<Property> getPropertiesFromEObject(EObject obj) {
        EStructuralFeature feature = obj == null ? null : obj.eClass().getEStructuralFeature("properties");
        if (feature instanceof EReference) {
            return (List<Property>)obj.eGet(feature);
        }
        return Collections.EMPTY_LIST;
    }

    public static com.iona.soa.model.repository.Property newProperty(String name, String value) {
        Property property = RepositoryFactory.eINSTANCE.createProperty();
        property.setName(name);
        property.setValue(value);
        return property;
    }

    // Set a Version object from its stringified representation
    public static Version setVersionProperties(Version version,
                                               String versionNumber, String versionQualifier) {
        if (version == null) {
            return null;
        }
        
        if (versionNumber != null) {

            String[] versionComponents = versionNumber.split("\\.");
            
            if (versionComponents.length > 0) {
                version.setMajor(Integer.valueOf(versionComponents[0]));
            }
            if (versionComponents.length > 1) {
                version.setMinor(Integer.valueOf(versionComponents[1]));
            }
            if (versionComponents.length > 2) {
                version.setMicro(Integer.valueOf(versionComponents[2]));
            }
            if (versionComponents.length > 3) {
                version.setNano(Integer.valueOf(versionComponents[3]));
            }
        }

        if (versionQualifier != null) {
            version.setQualifier(versionQualifier);
        }
        
        return version;
    }

    public static boolean checkEMFQLCondition(EObject obj, String expression) {
        Expression expr = (Expression) ExpressionSerializer.deserialize(expression);
        if (expr == null) {
            return false;
        }
        
        BeanExpressionHandler handler = new BeanExpressionHandler(Collections.singletonList(obj));
        handler.handle(expr);
        
        return handler.getQualifying().size() > 0;
    }

    @SuppressWarnings("unchecked")    
    private static Collection<EPackage> getEPackages(String ... classNames) {
        Collection<EPackage> ret = new LinkedList<EPackage>();
        for (String className : classNames) {
            try {
                Class clazz = Class.forName(className);
                if (clazz != null 
                    && clazz.isInterface() 
                    && clazz.getInterfaces().length == 1 
                    && clazz.getInterfaces()[0].equals(EPackage.class)) {
                    EPackage pkg = (EPackage)clazz.getDeclaredField("eINSTANCE").get(null);
                    ret.add(pkg);
                }
            } catch (Exception ex) {                
                RuntimeException re = 
                    new RuntimeException("Unable to find required repository class: " + className, ex);
                re.printStackTrace();
                throw re;
            }
        }
        return ret;
    }
}
