/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.notification;

import java.io.UnsupportedEncodingException;
import java.util.Collections;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.mail.Session;
import javax.mail.internet.InternetAddress;

import com.iona.repository.notification.jet.CreatedEMailTextTemplate;
import com.iona.repository.notification.jet.DeletedEMailTextTemplate;
import com.iona.repository.notification.jet.UpdatedEMailHtmlTemplate;
import com.iona.repository.notification.jet.UpdatedEMailTextTemplate;
import com.iona.soa.model.events.DataChangeEvent;
import com.iona.soa.model.events.DataChangeKind;
import com.iona.soa.model.notification.NotificationScheme;
import com.iona.soa.model.security.User;
import com.iona.soa.repository.SOARepository;

import org.apache.commons.mail.Email;
import org.apache.commons.mail.EmailException;
import org.apache.commons.mail.HtmlEmail;

class EmailDispatcher implements INotificationDispatcher {
    private static final Logger LOG = Logger.getLogger(EmailDispatcher.class.getName());

    private static final String PROPERTY_FROM_NAME = "repository.mail.fromName";
    
    private static final String PROPERTY_FROM_ADDRESS = "repository.mail.fromAddress";

    private static final String PROPERTY_TEXT_FORMAT = "repository.mail.includeTextFormat";

    private static final String PROPERTY_HTML_FORMAT = "repository.mail.includeHtmlFormat";
    
    private static final String DEFAULT_FROM_NAME = "Repository Administrator";
    
    private static final String DEFAULT_FROM_ADDRESS = "repository_admin@example.org";
    
    private SOARepository repository;
    
    private Session mailSession;
    
    private String fromMailName;
    
    private String fromMailAddress;
    
    private boolean includeTextFormat;

    private boolean includeHtmlFormat;
    
    EmailDispatcher(SOARepository repo) {
        repository = repo;
    }

    public void start(Map<String, String> properties) {
        Properties p = convert2Properties(properties);
        mailSession = Session.getInstance(p, null);
        // If not already set via "mail.debug" property
        if (!mailSession.getDebug()) {
            mailSession.setDebug(LOG.isLoggable(Level.FINE));
        }
        
        fromMailName = getProperty(properties, PROPERTY_FROM_NAME, DEFAULT_FROM_NAME);        
        fromMailAddress = getProperty(properties, PROPERTY_FROM_ADDRESS, DEFAULT_FROM_ADDRESS);
        includeTextFormat = Boolean.parseBoolean(getProperty(properties, PROPERTY_TEXT_FORMAT, 
                                                             Boolean.toString(false)));
        includeHtmlFormat = Boolean.parseBoolean(getProperty(properties, PROPERTY_HTML_FORMAT, 
                                                             Boolean.toString(true)));
    }
    



    public void stop() {        
    }
 
    public void sendNotification(NotificationScheme scheme,
                                 DataChangeEvent dataEvent, 
                                 NotificationTargets to) {
        if (to.isEmpty()) {
            return;
        }

        final String subject = getSubject(scheme, dataEvent);
        final String textContent = includeTextFormat ? getTextContent(scheme, dataEvent) : null;
        final String htmlContent = includeHtmlFormat ? getHtmlContent(scheme, dataEvent) : null;

        for (String s : to.getEmailAddresses()) {
            try {
                InternetAddress address = new InternetAddress();
                address.setAddress(s);
                Email mail = createMessage(subject, htmlContent, textContent);
                mail.setTo(Collections.singletonList(address));
                mail.send();
            } catch (EmailException ex) {
                LOG.log(Level.SEVERE, "Failed to send email message to " + s, ex);
                // TODO - log to AUDIT log somehow
            }
        }

        for (User user : to.getUsers()) {
            if (user.getEmail() == null || "".equals(user.getEmail())) {
                LOG.log(Level.INFO, "User {0} has no email address, not sending notification",
                        user.getUserid());
                continue;
            }

            InternetAddress address = new InternetAddress();
            address.setAddress(user.getEmail());
            if (user.getName() != null && !"".equals(user.getName())) {
                try {
                    address.setPersonal(user.getName());
                } catch (UnsupportedEncodingException ex) {
                    LOG.log(Level.WARNING, 
                            "Failed to set personal information (name) for" + user.getUserid(), ex);
                }
            }
            
            try {
                Email mail = createMessage(subject, htmlContent, textContent);
                mail.setTo(Collections.singletonList(address));
                mail.send();
            } catch (EmailException ex) {
                LOG.log(Level.SEVERE, "Failed to send email message to " + user.getUserid(), ex);
                // TODO - log to AUDIT log somehow
            }
        }
    }

    String getSubject(NotificationScheme scheme, DataChangeEvent dataEvent) {
        // Example
        // [AR/R] Updated: Service \"Ticket Service\" has been updated
        StringBuffer sb = new StringBuffer();
        sb.append(scheme.getEmailSubjectPrefix());
        sb.append(" ");
        
        switch(dataEvent.getChangeKind().getValue()) {
        case DataChangeKind.ADD:
            sb.append("Created:");
            break;
        case DataChangeKind.REMOVE:
            sb.append("Deleted:");
            break;
        case DataChangeKind.MODIFY:
            sb.append("Updated:");
            break;
        default:
            break;
        }
        
        String typeName = NotificationEngine.getChangeObjectTypeName(repository, dataEvent);
        if (typeName != null) {
            sb.append(" ");
            sb.append(typeName);
            String name = dataEvent.getTargetName() != null ? dataEvent.getTargetName() : "<unknown>";
            if (name != null) {
                sb.append(" ");
                sb.append("\"");
                sb.append(name);
                sb.append("\"");
            }
        }
    
        return sb.toString();
    }

    String getTextContent(NotificationScheme scheme, DataChangeEvent dataEvent) {
        JetArgumentForEMail arg = new JetArgumentForEMail(dataEvent);
        
        switch(dataEvent.getChangeKind().getValue()) {
        case DataChangeKind.ADD:
            CreatedEMailTextTemplate createdT = new CreatedEMailTextTemplate();
            return createdT.generate(arg);
        case DataChangeKind.REMOVE:
            DeletedEMailTextTemplate deletedT = new DeletedEMailTextTemplate();
            return deletedT.generate(arg);
        case DataChangeKind.MODIFY:
            UpdatedEMailTextTemplate updatedT = new UpdatedEMailTextTemplate();
            return updatedT.generate(arg);
        default:
            return "";
        }
    }

    String getHtmlContent(NotificationScheme scheme, DataChangeEvent dataEvent) {
        JetArgumentForEMail arg = new JetArgumentForEMail(dataEvent);
        
        switch(dataEvent.getChangeKind().getValue()) {
        case DataChangeKind.ADD:
            CreatedEMailTextTemplate createdT = new CreatedEMailTextTemplate();
            return createdT.generate(arg);
        case DataChangeKind.REMOVE:
            DeletedEMailTextTemplate deletedT = new DeletedEMailTextTemplate();
            return deletedT.generate(arg);
        case DataChangeKind.MODIFY:
            UpdatedEMailHtmlTemplate updatedT = new UpdatedEMailHtmlTemplate();
            return updatedT.generate(arg);
        default:
            return "";
        }
    }

    private HtmlEmail createMessage(String subject,
                                    String htmlContent,
                                    String textContent) throws EmailException {
        HtmlEmail email = new HtmlEmail();
        email.setFrom(fromMailAddress, fromMailName);
        email.setMailSession(mailSession);
        email.setSubject(subject);
        if (htmlContent != null) {
            email.setHtmlMsg(htmlContent);
        }
        if (textContent != null) {
            email.setTextMsg(textContent);
        }
        return email;
    }
    
    private String getProperty(Map<String, String> properties,
                               String name, String defaultValue) {
        String v = properties.get(name);
        if (v == null) { 
            return defaultValue;
        }
        return v;
    }
    

    private Properties convert2Properties(Map<String, String> properties) {
        Properties p = new Properties();
        for (String k : properties.keySet()) {
            p.setProperty(k, properties.get(k));
        }
        return p;
    }
}
