/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.notification;

import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.soa.expression.Equals;
import com.iona.soa.expression.Expression;
import com.iona.soa.expression.Operand;
import com.iona.soa.model.collaboration.CollaborationPackage;
import com.iona.soa.model.collaboration.ICollaboration;
import com.iona.soa.model.events.DataChangeEvent;
import com.iona.soa.model.events.DataChangeFieldDelta;
import com.iona.soa.model.notification.CustomRecipient;
import com.iona.soa.model.notification.NotificationDefinition;
import com.iona.soa.model.notification.NotificationFactory;
import com.iona.soa.model.notification.NotificationScheme;
import com.iona.soa.model.notification.impl.CustomRecipientImpl;
import com.iona.soa.model.repository.RepositoryPackage;
import com.iona.soa.model.security.SecurityPackage;
import com.iona.soa.model.security.User;
import com.iona.soa.repository.SOARepository;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.xmi.impl.XMIResourceImpl;

/**
 *  Listens for events of type DataChangeEvent.
 *  <p><p>
 *  For an event
 *  <ul> 
 *  <li>Obtain NotificationScheme associated with the changed object</li>
 *  <li>For each NotificationDefinition in the scheme check if the content of the event matches</li>
 *  <li>If an event matches, add the NotificationRecipients</li>
 *  <li>Once all definitions have been visited, send a notification to the 
 *  cumulative RecipientSet (if any</li>
 *  </ul>
 */
public final class NotificationEngine {
    private static final Logger LOG = Logger.getLogger(NotificationEngine.class.getName());
    
    private static final String DEFAULT_SCHEME_FILE = "/default.notification_scheme";
    
    private NotificationScheme defaultScheme;
    
    private final INotificationDispatcher emailDispatcher;
    
    private final SOARepository repository;
    
    public NotificationEngine(SOARepository repo) {
        repository = repo;
        emailDispatcher = new EmailDispatcher(repository);        
    }
    
    /**
     * The notification engine will extract configuration from the supplied properties
     * object.  Configuration includes the mail.* properties that are used
     * to configure email sending using the JavaMail API.
     * 
     * @param properties containing mail and other relevant configuration for
     * the notification engine.
     */
    public void start(Map<String, String> properties) {
        emailDispatcher.start(properties);
    }
    
    public void stop() {
        emailDispatcher.stop();
    }
    
    public void dataChanged(DataChangeEvent event) {
        NotificationTargets targets = calculateRecipients(event);
        if (!targets.isEmpty()) {
            emailDispatcher.sendNotification(getScheme(event), event, targets);
        }
    }
    
    @SuppressWarnings("unchecked")
    NotificationTargets calculateRecipients(DataChangeEvent event) {
        NotificationTargets targets = new NotificationTargets();
        
        List<User> subscribers = getSubscribers(repository, event);
        boolean interestingChangeMade = false;
        // get copy of NotificationDefinitions to prevent ConcurrentModificationException 
        // see REPO-549
        NotificationDefinition[] defs = 
            getScheme(event).getNotificationDefinitions().toArray(new NotificationDefinition[0]);
        for (NotificationDefinition def : defs) {
            if (def.evaluate(event)) {
                LOG.log(Level.FINE, "NotificationDefinition {0} triggered", def.getRule());
                interestingChangeMade = true;
                
                if (def.getRecipients().isNotifySubscribers()) {
                    targets.mergeUsers(subscribers);
                }
                
                // Include custom recipients if applicable
                CustomRecipient[] crecs = 
                    def.getRecipients().getCustomRecipients().toArray(new CustomRecipient[0]);
                for (CustomRecipient cr : crecs) {
                    targets.mergeUsers(cr.getUsers(event));
                }
            }
        }
        
        if (interestingChangeMade) {
            // Owner is not on subscribers list by default, we include
            // them if there are any events.  
            // TODO - use a boolean like isNotifySubscribers in the model on each 
            // definition (isNotifyOwner)? 
            User owner = getOwner(repository, event);
            if (owner != null && !owner.getUserid().equals(event.getUserId())) {
                targets.mergeUsers(Collections.singletonList(owner));
            }
        }
        return targets;
    }
    
    @SuppressWarnings("unchecked")
    NotificationScheme getDefaultScheme() {
        if (defaultScheme == null) {
            Resource res = new XMIResourceImpl();
            try {
                InputStream is = getClass().getResourceAsStream(DEFAULT_SCHEME_FILE); 
                res.load(is, Collections.EMPTY_MAP);
                defaultScheme = (NotificationScheme)res.getContents().get(0);
            } catch (IOException ex) {
                LOG.log(Level.SEVERE, "Failed to load " + DEFAULT_SCHEME_FILE, ex);
                throw new RuntimeException(ex);
                // return NotificationFactory.eINSTANCE.createNotificationScheme();
            }
            
            defaultScheme.getNotificationDefinitions().add(createOwnerManagingDefinition());
        }
        return defaultScheme;        
    }
    
    private NotificationScheme getScheme(DataChangeEvent event) {
        // Later, use info in event to determine scheme
        return getDefaultScheme();
    }
    
    private NotificationDefinition createOwnerManagingDefinition() {
        NotificationDefinition def = NotificationFactory.eINSTANCE.createNotificationDefinition();
        def.setName("Owner");
        def.setRule("changeKind = DataChangeKind::MODIFY and "
                    + "updatedFields->exists(d: DataChangeFieldDelta | d.fieldDesc.name = 'owner')");
        def.setRecipients(NotificationFactory.eINSTANCE.createNotificationRecipients());
        
        // This CustomRecipient crap needs to be changed, say to be an interface
        // that is also notified of an event.  Say a list of them per scheme.
        CustomRecipient ownerChangeRecipient = new CustomRecipientImpl() {
            @Override
            public EList<User> getUsers(DataChangeEvent event) {
                EList<User> users = new BasicEList<User>();
                for (DataChangeFieldDelta delta : event.getUpdatedFields()) {
                    if (CollaborationPackage.Literals.ICOLLABORATION__OWNER.equals(delta.getFieldDesc())) {
                        String oldUserId = delta.getOldValue();
                        String newUserId = delta.getNewValue();
                        if (oldUserId != null 
                                && !"".equals(oldUserId)
                                && !oldUserId.equals(event.getUserId())) {
                            User oldUser = getUser(repository, oldUserId);
                            if (oldUser != null) {
                                users.add(oldUser);
                            }
                        }
                        
                        if (newUserId != null 
                                && !"".equals(newUserId)
                                && !newUserId.equals(event.getUserId())) {
                            User newUser = getUser(repository, newUserId);
                            if (newUser != null) {
                                users.add(newUser);
                            }
                        }
                    }
                } 
                return users;
            }
        };
        def.getRecipients().getCustomRecipients().add(ownerChangeRecipient);
        return def;
    }
    
    /**
     * TODO - Move these static getters into DataChangeEvent as derived methods
     * that we implement ourselves (via a DataChangeEventHelper @see NotificationDefinitionHelper
     */
    static String getChangeObjectTypeName(SOARepository repository, DataChangeEvent event) {
        // TODO - may have to use gen.edit code to get the readble version
        // of this.
        return event.getTargetType().getName();
    }
    
    private static EObject getObjectFromRepository(SOARepository repository, DataChangeEvent event) {
        EObject ret = null;
        boolean success = false;
        repository.startRead(null);
        try {
            Operand guidOper = new Operand(RepositoryPackage.Literals.IPERSISTABLE_OBJECT__GUID.getName());
            Operand valueOper = new Operand(event.getTargetGuid());
            Expression expr = new Equals(guidOper, valueOper);
            Collection<EObject> objects = repository.getResourceSet().executeQuery(event.getTargetType(), 
                                                                                   expr, 
                                                                                   true);
            if (!objects.isEmpty()) {
                ret = objects.iterator().next();
            }
            success = true;
        } catch (Exception ex) {
            LOG.log(Level.SEVERE, "Error occured retrieving object <" + event.getTargetGuid() + ">", ex);
        } finally {
            repository.endRead(success);
        }
        return ret;
    }
    
    private static User getUser(SOARepository repository, String userId) {
        try {
            Operand idOper = new Operand(SecurityPackage.Literals.USER__USERID.getName());
            Operand valueOper = new Operand(userId);
            Expression expr = new Equals(idOper, valueOper);
            EClass eClass = SecurityPackage.Literals.USER;
            Collection<EObject> objects = repository.getResourceSet().executeQuery(eClass, 
                                                                                   expr, 
                                                                                   true);
            if (!objects.isEmpty()) {
                return (User)objects.iterator().next();
            }
        } catch (Exception ex) {
            LOG.log(Level.SEVERE, "Error occured retrieving user <" + userId + ">", ex);
        }
        return null;        
    }
    
    @SuppressWarnings("unchecked")
    private static List<User> getSubscribers(SOARepository repository, DataChangeEvent event) {
        EObject changed = getObjectFromRepository(repository, event);
        if (changed instanceof ICollaboration) {
            ICollaboration c = (ICollaboration)changed;
            return c.getSubscribers();
        }
        return Collections.emptyList();
    }
    
    private static User getOwner(SOARepository repository, DataChangeEvent event) {
        EObject changed = getObjectFromRepository(repository, event);
        if (changed instanceof ICollaboration) {
            ICollaboration c = (ICollaboration)changed;
            return c.getOwner();
        }
        return null;
    }
}
