/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.rest.atom;

import java.util.logging.Logger;

import javax.xml.namespace.QName;


import com.iona.repository.rest.Constants;
import com.iona.repository.rest.resource.Attribute;
import com.iona.repository.rest.resource.Base;
import com.iona.repository.rest.resource.Reference;
import com.iona.repository.rest.resource.Resource;
import com.iona.repository.rest.resource.ResourceFactory;
import com.iona.repository.rest.resource.ResourceFactoryImpl;

import org.apache.abdera.Abdera;
import org.apache.abdera.factory.Factory;
import org.apache.abdera.model.Element;
import org.apache.abdera.model.Entry;
import org.apache.abdera.model.Feed;
import org.apache.abdera.model.Link;
import org.apache.abdera.util.Configuration;

/**
 * Utility class for converting between Abdera Elements and resource objects.
 *
 */
public final class AbderaUtil {

    private static final Logger LOG = Logger.getLogger(AbderaUtil.class.getName());
  
    private static final String TAG_AUTHORITY = "com.iona";
    
    private static final String TAG_DATE = "2008-02-10";
    
    private static Abdera abdera;
    private static Factory abderaFactory;
    
    private static ResourceFactory resourceFactory = new ResourceFactoryImpl();
    
    private static final QName XML_BASE_QNAME = new QName("http://www.w3.org/XML/1998/namespace", "base");
    
    static {
        // Configure single instance of Abdrea configured to able to run
        // inside an OSGi container.
        Configuration myConfig = new AbderaOSGiConfig();
        abdera = new Abdera(myConfig);
        abderaFactory = abdera.getFactory();        
    }
    
    private AbderaUtil() {
        // Complete.
    }
   

    public static Abdera getAbdera() {
        return abdera;
    }
    
    public static Factory getAbderaFactory() {
        return abderaFactory;
    }
    
    /**
     * Returns an Abdera Element representing a resource object.
     * @param resource the resource object
     * @return the Abdera Element representing the resource
     */
    public static Element resourceToElement(Resource resource) {
        return resourceToElement(resource, null);
    }
    
    /**
     * Returns an Abdera Element representing a resource object.
     * @param resource the resource object
     * @return the Abdera Element representing the resource
     */
    private static Element resourceToElement(Resource resource, Element parent) {
        Element root;
        if (parent != null) {
            root = abderaFactory.newElement(Resource.RESOURCE_QNAME, (org.apache.abdera.model.Base)parent);
        } else {
            root = abderaFactory.newElement(Resource.RESOURCE_QNAME);
        }

        addBaseAttributes(root, resource);
        addResourceAttribtues(root, resource);
        for (Attribute attribute : resource.getAllAttributes()) {
            Element element = abderaFactory.newElement(Resource.ATTRIBUTE_QNAME, root);
            
            addBaseAttributes(element, attribute);
            addAttributeAttributes(element, attribute);
        }
        
        for (Reference reference : resource.getAllReferences()) {
            Element element = abderaFactory.newElement(Resource.REFERENCE_QNAME, root);
            addBaseAttributes(element, reference);
            addReferenceAttributes(element, reference);
        }
        
        for (Resource childResource : resource.getResources()) {
            resourceToElement(childResource, root);
        }
        
        return root;
    }




    /**
     * Returns a resource object represented by an Abdera Element.
     * @param element the Abdera element
     * @return the resource object representing the Abdera Element.
     * @throws IOException if there is an error during conversion
     * @throws ResourceParseException if there is an error during
     *  conversion
     */
    public static Resource elementToResource(Element element) 
        throws ResourceParseException {
        
        QName elementQName = element.getQName();
        if (!elementQName.equals(Resource.RESOURCE_QNAME)) {
            LOG.severe("Expected " + Resource.RESOURCE_QNAME + " when constructing resource "
                      + "from Atom element but got " + elementQName);
            throw new ResourceParseException("Invalid QName " + elementQName
                    + "; expected " + Resource.RESOURCE_QNAME); 
                    
        }
        String name = element.getAttributeValue("name");
        
        Resource resource = resourceFactory.createResource(name);
        addBaseAttributes(resource, element);
        addResourceAttributes(resource, element);
        
        for (Element e : element.getElements()) {
            QName qn = e.getQName();
            if (qn.equals(Resource.ATTRIBUTE_QNAME)) {
                Attribute attribute = elementToAttribute(e);
                resource.addAttribute(attribute);
            } else if (qn.equals(Resource.REFERENCE_QNAME)) {
                Reference reference = elementToReference(e);
                resource.addReference(reference);                
            } else if (qn.equals(Resource.RESOURCE_QNAME)) {
                Resource child = elementToResource(e);
                resource.addResource(child);                
            } else {
                LOG.severe("Unknown element " + qn + " found while when constructing"
                          + " resource from Atom element " + e);
                throw new ResourceParseException("Invalid Qname " + qn
                        + "; expected one of " + Resource.RESOURCE_QNAME 
                        + ", " + Resource.ATTRIBUTE_QNAME + " or "
                        + Resource.REFERENCE_QNAME);
            }
        }
        
        return resource;
    }


 


    /**
     * Sets the baseUri attribute on a feed/entry.  This is here to workaround
     * an Axiom bug (WSCOMMONS-281) where the XML is serialized with a xmlns:xml
     * attribute which causes problems for some XML parsers. 
     * @param base Abdera entry/feed
     * @param uri Base URI
     * @return base
     */
    public static org.apache.abdera.model.Base setBaseUri(org.apache.abdera.model.Base base, String uri) {
        if (base instanceof Entry) {
            Entry e = (Entry) base;
            e.setAttributeValue("xml:base", uri);
        } else if (base instanceof Feed) {
            Feed f = (Feed) base;
            f.setAttributeValue("xml:base", uri);
        }
        return base;
    }
    
    /**
     * Gets the baseUri attribute from an Abdera feed/entry.  This is to workaround 
     * an Axiom bug (WSCOMMONS-281).
     * @param base Abdera feed/entry
     * @return the "xml:base" attribute or null if not present.
     */
    public static String getBaseUri(org.apache.abdera.model.Base base) {
        if (base instanceof Element) {
            return getBaseUriAttribute((Element) base);
        } else {
            return null;
        }
    }

    private static String getBaseUriAttribute(Element element) {
        if (element == null) {
            return null;
        }
        // First try the "xml:base" attribute
        String base = element.getAttributeValue("xml:base");
        if (base == null) {
            // Next try using the proper QName.
            base = element.getAttributeValue(XML_BASE_QNAME);
        }
        if (base == null) {
            // Finally try the parent.
            return getBaseUri(element.getParentElement());
        }
        return base;
    }
    
    /**
     * Returns a valid URL for an Atom feed/entry ID.  This uses tag URLs
     * (see RFC-4151).
     * @param baseUri Base URI of the feed/entry
     * @param id the ID of the entry/feed
     * @return a valid URL string for the ID
     */
    public static String idToUrl(String baseUri, String id) {
        // TODO: base the authority off the baseUri domain
        
        return new StringBuilder()
            .append("tag:")
            .append(TAG_AUTHORITY)
            .append(",")
            .append(TAG_DATE)
            .append(":")
            .append(id)
            .toString();
          
    }
    
    
    public static void addEntry(Feed feed, Resource resource) {
        // TODO Auto-generated method stub
        
    }

    private static void addBaseAttributes(Element element, Base base) {
        String v = base.getBaseUri();
        if (v != null) {
            element.setAttributeValue("base", v);
        }
        
        v = base.getName();
        if (v != null) {
            element.setAttributeValue("name", v);
        }
        
        v = base.getEClass();
        if (v != null) {
            element.setAttributeValue("eclass", v);
        }
        
        int m = base.getMin();
        if (m != 1) {
            element.setAttributeValue("min", Integer.toString(m));
        }
        
        m = base.getMax();
        if (m != 1) {
            element.setAttributeValue("max", Integer.toString(m));
        }
        
        v = base.getType();
        if (v != null) {
            element.setAttributeValue("type", v);
        }
        
        v = base.getText();
        element.setText(v);
    }
    
    
    private static void addResourceAttribtues(Element element, Resource resource) {
        String id = resource.getId();
        if (id != null && !"".equals(id)) {
            element.setAttributeValue("id", id);
        }        
        String collection = resource.getCollection();
        if (collection != null && !"".equals(collection)) {
            element.setAttributeValue("collection", collection);
        }        
    }

    private static void addAttributeAttributes(Element element, Attribute attribute) {
        if (attribute.isReadonly()) {
            element.setAttributeValue("readonly", "true");
        }
        if (attribute.isHidden()) {
            element.setAttributeValue("hidden", "true");
        }
    }

    private static void addReferenceAttributes(Element element, Reference reference) {
        
        if (reference.isReadonly()) {
            element.setAttributeValue("readonly", "true");
        }
        
        String href = reference.getHref();
        if (href != null) {
            element.setAttributeValue("href", href);
        }        
        
        href = reference.getCollectionHref();
        if (href != null) {
            element.setAttributeValue("collectionHref", href);
        }
        
        int size = reference.getSize();
        if (size > 0) {
            element.setAttributeValue("size", Integer.toString(size));
        }
    }



    
    private static Reference elementToReference(Element element) {
        
        String s = element.getAttributeValue("name");
        Reference reference = resourceFactory.createReference(s);
        
        addBaseAttributes(reference, element);

        s = element.getAttributeValue("href");
        reference.setHref(s);
        
        s = element.getAttributeValue("collectionHref");
        reference.setCollectionHref(s);
        
        s = element.getAttributeValue("collectionHref");
        
        int i = toInt(element.getAttributeValue("size"));
        reference.setSize(i);

        return reference;
    }


    private static Attribute elementToAttribute(Element element) {

        String name = element.getAttributeValue("name");
        Attribute attribute = resourceFactory.createAttribute(name);
        addBaseAttributes(attribute, element);
        
        if ("true".equals(element.getAttributeValue("readonly"))) {
            attribute.setReadonly(true);
        }
        
        if ("true".equals(element.getAttributeValue("hidden"))) {
            attribute.setHidden(true);
        }

        return attribute;
    }


    private static void addBaseAttributes(Base base, Element element) {
        
        String s = element.getAttributeValue("base");
        base.setBaseUri(s);
        
        s = element.getAttributeValue("type");
        base.setType(s);
        
        s = element.getAttributeValue("eclass");
        base.setEClass(s);
        
        s = element.getText();
        base.setText(s);
        
        int i = toInt(element.getAttributeValue("min"));
        base.setMin(i);
        
        i = toInt(element.getAttributeValue("max"));
        base.setMax(i);
        
    }
    


    private static void addResourceAttributes(Resource resource, Element element) {        
        String v = element.getAttributeValue("id");
        resource.setId(v);
        v = element.getAttributeValue("collection");
        resource.setCollection(v);
    }
    
    /**
     * Returns the resource object contained within an Abdera Entry document
     * @param entry the document containing the resource
     * @return the resource object if found; null otherwise
     */
    public static Resource getResource(Entry entry) {
        Element el = findXmlContent(entry);
        if (el == null) {
            return null;
        }
        try {
            return elementToResource(el);
        } catch (ResourceParseException e) {
            LOG.severe("Failed to convert Atom entry to resource; exception "
                    + e);
            return null;
        }
    }

    /**
     * Returns the Abdera Element representing a resource object within an
     * Abdera Entry document.
     * @param entry the Entry document containing the resource Element.
     * @return the Abdera Element containing the resource if present; otherwise
     * null
     */
    public static Element findXmlContent(Entry entry) {
        for (Link link : entry.getLinks(Constants.RESOURCE_LINK_REL)) {
            // Strictly speaking there should only be one link but search for
            // the first one with content.
            if (link.getAttributeValue("type").equals(Constants.RESOURCE_CONTENT_TYPE)) {
                for (Element e : link.getExtensions(Resource.NAMESPACE)) {
                    return e;
                }
            }
        }
        return null;
    }

    private static int toInt(String v) {
        
        if (v == null) {
            return 0;
        }
        
        try {
            return Integer.parseInt(v);
        } catch (NumberFormatException e) {
            LOG.warning("Attribute " + v + " is not a valid number; assuming 0");
        }
        return 0;        
    }
}
