/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.rest.atom;

import java.util.Date;
import java.util.UUID;
import java.util.logging.Logger;

import com.iona.repository.rest.Constants;
import com.iona.repository.rest.resource.Attribute;
import com.iona.repository.rest.resource.Resource;

import org.apache.abdera.factory.Factory;
import org.apache.abdera.model.AtomDate;
import org.apache.abdera.model.Element;
import org.apache.abdera.model.Entry;
import org.apache.abdera.model.Link;
import org.apache.abdera.model.Person;

/**
 * Abstract class for creating Abdera Entry documents containing a resource object.
 *
 */
public abstract class AbstractAtomEntryFactory implements AtomEntryFactory {

    private static final Logger LOG = Logger.getLogger(AbstractAtomEntryFactory.class.getName());

    private static final AtomEntryOptions DEFAULT_OPTIONS =  new AtomEntryOptions();    

    private static final AtomDate DEFAULT_UPDATE_DATE = new AtomDate();
    
    protected Factory abderaFactory;    

    public AbstractAtomEntryFactory() {
        abderaFactory = AbderaUtil.getAbderaFactory();
    }

    public Entry createEntry(Resource resource, String baseUri, String path) {
        return createEntry(resource, baseUri, path, DEFAULT_OPTIONS);
    }

    public Entry createEntry(Resource resource, String baseUri, String path, AtomEntryOptions options) {

        Entry entry = abderaFactory.newEntry();
        // TODO: Might be a good idea to only set the base URI if there is no
        // parent feed or if there is a feed but it uses a different baseUri.
        
        AbderaUtil.setBaseUri(entry, baseUri);
        
        entry.addAuthor(getAuthor(resource));
        entry.setId(AbderaUtil.idToUrl(baseUri, getId(resource)));
        entry.setTitle(getTitle(resource));
        
        Date updated = getUpdated(resource);
        entry.setUpdated(updated);
        
        if (addEditedDate()) {
            entry.setEdited(updated);
        }
        
        if (path != null) {
            Link link = getSelfLink(resource, baseUri, path);
            if (link != null) {
                entry.addLink(link);
            }
            link = getEditLink(resource, baseUri, path);
            if (link != null) {
                entry.addLink(link);
            }
            link = getDeleteLink(resource, baseUri, path);
            if (link != null) {
                entry.addLink(link);
            }
        }

        addContent(entry, resource, baseUri, path, options);
        
        if (options.includeRelatedLinks()) {
            addReferenceLinks(entry, resource, baseUri);
        }

        return entry;
    }



    protected void addContent(Entry entry,
                              Resource resource,
                              String baseUri, 
                              String path,
                              AtomEntryOptions options) {

        Link contentLink = getResourceLink(resource, baseUri, path);

        if (options.includeResource() || options.includeXhtml()) {

            Element element = getResourceElement(resource);

            if (element != null) {              
                if (options.includeResource()) {
                    // Standalone entry
                    contentLink.addComment("Inline XML representation of repository resource");
                    contentLink.addExtension(element);
                }
                if (options.includeXhtml()) {
                    String xhtml = getXhtml(entry.getTitle(), element);
                    if (xhtml != null) {
                        entry.setContentAsXhtml(xhtml);
                    }
                }
            }
        }

        if (options.includeSummary()) {
            String summary = getSummary(resource);
            if (summary != null) {
                entry.setSummary(summary);
            }
        }

        entry.addLink(contentLink);
    }


    protected void addReferenceLinks(Entry entry, Resource resource, String baseUri) {
    }

    protected Element getResourceElement(Resource resource) {
        return AbderaUtil.resourceToElement(resource);
    }

    protected String getSummary(Resource resource) {
        return null;
    }

    protected String getId(Resource resource) {
        
        // Need a unique identifier for the resource 
        String id = resource.getId();
        if (!"".equals(id)) {
            return id;
        }
        
        LOG.warning("Resource " + resource.getName() + " has no ID; generating GUID");
        return UUID.randomUUID().toString();
    }

    protected Person getAuthor(Resource resource) {
        Person author = abderaFactory.newAuthor();
        author.setName(Constants.ATOM_ENTRY_AUTHOR);
        return author;
    }

    protected String getTitle(Resource resource) {
        
        // First try the resource name, then the "name" attribute
        // and finally the text content.
        
        String s = resource.getName();
        if (s != null) {
            return s;
        }
        
        Attribute a = resource.getFirstAttribute("name");
        if (a == null) {
            return "";
        }
        
        String name = a.getText();
        if (name == null) {
            return "";
        }
        
        return name;
    }

    protected Date getUpdated(Resource resource) {
        
        Attribute a = resource.getFirstAttribute("timestamp");
        if (a == null) {
            return DEFAULT_UPDATE_DATE.getDate();
        }
        Object o = a.getObject();
        if (o instanceof Date) {
            return (Date) o;
        }
        String dateStr = a.getText();
        if (dateStr == null || "".equals(dateStr)) {
            return DEFAULT_UPDATE_DATE.getDate();
        }
        
        // Try to avoid this last resort as there can be conversion problems
        // when parsing dates.
        return new AtomDate(dateStr).getDate();
        
    }

    protected Link getSelfLink(Resource resource, String baseUri, String path) {
        return abderaFactory.newLink()
               .setHref(new StringBuilder().append(path).append("/").append(getId(resource)).toString())
               .setRel("self")
               .setMimeType(Constants.ATOM_CONTENT_TYPE);
    }

    protected Link getEditLink(Resource resource, String baseUri, String path) {
        return null;
    }
    
    protected Link getDeleteLink(Resource resource, String baseUri, String path) {
        return null;
    }
    
    protected boolean addEditedDate() {
        return false;
    }


    protected Link getResourceLink(Resource resource, String baseUri, String path) {

        Link link = abderaFactory.newLink()
                    .setRel(Constants.RESOURCE_LINK_REL)
                    .setMimeType(Constants.RESOURCE_CONTENT_TYPE)
                    .setTitle(resource.getName());
        if (path != null) {
            link.setHref(new StringBuilder().append(path).append("/").append(getId(resource)).toString());
        }
        return link;
    }
    
    protected String getXhtml(String title, Element element) {
        return null;
    }
}
