/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.rest.resource;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class ResourceImpl extends BaseImpl implements Resource {

    private String collection;
    private String id;
    private Map<String, List<Attribute>> attributes;
    private Map<String, List<Reference>> references;
    private Map<String, Object> properties;
    private List<Resource> resourceList = new ArrayList<Resource>();
    private boolean isEditable = true;
    private boolean isDeletable = true;
    protected ResourceImpl(String name) {
        this(name, null, null);
    }

    protected ResourceImpl(String name, String collection, String id) {
        super(name);
        this.collection = collection;
        this.id = id;
    }
    
    public String getId() {
        return id;
    }
    
    public Resource setId(String i) {
        id = i;
        return this;
    }
    
    public String getCollection() {
        return collection;
    }
    
    public Resource setCollection(String c) {
        collection = c;
        return this;
    }
    
    public List<Attribute> getAllAttributes() {
        List<Attribute> all = new ArrayList<Attribute>();
        for (List<Attribute> list : getAttributesMap().values()) {
            for (Attribute a : list) {
                all.add(a);
            }
        }
        return all;
    }

    public List<Attribute> getAttributes(String name) {
        return getAttributesList(name);
    }

    public Attribute getFirstAttribute(String name) {
        List<Attribute> atts = getAttributesList(name);
        if (atts.size() > 0) {
            return atts.get(0);
        }
        return null;
    }
    
    public Attribute addAttribute(String name) {
        Attribute att = new AttributeImpl(name);
        getAttributesList(name).add(att);
        return att;
    }

    public Attribute addAttribute(Attribute att) {
        getAttributesList(att.getName()).add(att);
        return att;
    }

    
    public List<Reference> getAllReferences() {
        List<Reference> all = new ArrayList<Reference>();
        for (List<Reference> list : getReferencesMap().values()) {
            for (Reference a : list) {
                all.add(a);
            }
        }
        return all;
    }

    public List<Reference> getReferences(String name) {
        return getReferencesList(name);
    }    
    
    public Reference getFirstReference(String name) {
        List<Reference> refs = getReferencesList(name);
        if (refs.size() > 0) {
            return refs.get(0);
        }
        return null;
    }


    public Reference addReference(String name) {
        Reference ref = new ReferenceImpl(name, this);
        getReferencesList(name).add(ref);
        return ref;
    }

    public Reference addReference(Reference ref) {
        getReferencesList(ref.getName()).add(ref);
        ref.setParent(this);
        return ref;
    }

    
    public List<Resource> getResources() {

        return resourceList;
    }
    
    public void addResource(Resource resource) {
        resourceList.add(resource);
    } 
    
    
    public String toString() {
        
        StringBuilder sb = new StringBuilder()
            .append("Resource \"").append(getName())
            .append("\"\n")
            .append("         id:  ").append(getId()).append("\n")
            .append(" collection:  ").append(getCollection()).append("\n")
            .append("     eclass:  ").append(getEClass()).append("\n")
            .append("   editable:  ").append(isEditable).append("\n")
            .append("  deletable:  ").append(isDeletable).append("\n");
        
        for (Attribute a : getAllAttributes()) {
            String t = a.getText();
            sb.append("  attribute: \"")
                .append(a.getName())
                .append("\" [")
                .append(a.getType())
                .append(", ")
                .append(a.getMin())
                .append(", ")
                .append(a.getMax());
            if (a.isHidden()) {
                sb.append(", hidden");
            }
            if (a.isReadonly()) {
                sb.append(", read-only");
            }
            sb.append("] {")            
                .append(t == null ? "<null>" : t).append("}\n");
        }
        for (Reference r : getAllReferences()) {
            sb.append("  reference: \"")
                .append(r.getName())
                .append("\" [min ")
                .append(r.getMin())
                .append(", max ")
                .append(r.getMax())
                .append(", size ")
                .append(r.getSize())
                .append(", eclass: ")
                .append(r.getEClass());
            if (r.getHref() != null) {
                sb.append(", href ").append(r.getHref());
            }
            if (r.getCollectionHref() != null) {
                sb.append(", collectionHref ")
                    .append(r.getCollectionHref());
            }
            sb.append("] {");            
            String t = r.getText();
            sb.append(t == null ? "" : t).append("}\n");
        }
        return sb.toString();
    }

    public Object getProperty(String name) {
        return getProperties().get(name);
    }

    public Resource setProperty(String name, Object value) {
        getProperties().put(name, value);
        return this;
    }


    private Map<String, List<Attribute>> getAttributesMap() { 
        if (attributes == null) {
            attributes = new HashMap<String, List<Attribute>>();
        }       
        return attributes;
    }

    private List<Attribute> getAttributesList(String name) {
        Map<String, List<Attribute>> map = getAttributesMap();
        List<Attribute> list = map.get(name);
        if (list == null) {
            list = new ArrayList<Attribute>();
            map.put(name, list);
        }
        return list;
    }
    
    private Map<String, List<Reference>> getReferencesMap() { 
        if (references == null) {
            references = new HashMap<String, List<Reference>>();
        }       
        return references;
    }

    private List<Reference> getReferencesList(String name) {
        Map<String, List<Reference>> map = getReferencesMap();
        List<Reference> list = map.get(name);
        if (list == null) {
            list = new ArrayList<Reference>();
            map.put(name, list);
        }
        return list;
    }
    
    private Map<String, Object> getProperties() {
        if (properties == null) {
            properties = new HashMap<String, Object>();
        }
        return properties;
    }

    public boolean isEditable() {
        return isEditable;
    }

    public void setEditable(boolean editable) {
        isEditable = editable;
        if (isEditable && isPrototype()) {
            throw new IllegalStateException("A prototype object cannot have an edit link");
        }
    }

    public boolean isDeletable() {
        return isDeletable;
    }

    public void setDeletable(boolean deletable) {
        isDeletable = deletable;
        if (isDeletable && isPrototype()) {
            throw new IllegalStateException("A prototype object cannot have a delete link");
        }
    }

    public boolean isPrototype() {
        return getFirstAttribute("isPrototype") == null ? false
                : Boolean.parseBoolean(getFirstAttribute("isPrototype").getText()); 
    }
    
    public void markAsPrototype() {
        if (!isPrototype()) {
            isEditable = false;
            isDeletable = false;
            addAttribute("isPrototype").setText(Boolean.toString(true));
        }
    }
}
