/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.rest.rt.media;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ResourceBundle;
import java.util.logging.Logger;

import com.iona.repository.rest.RESTContext;
import com.iona.repository.rest.RESTException;
import com.iona.repository.rest.RESTQuery;
import com.iona.repository.rest.rt.osgi.MessageBundle;
import com.iona.repository.rest.rt.repo.RestEMFUtil;
import com.iona.repository.util.i18n.Message;
import com.iona.soa.model.Artifact;
import com.iona.soa.repository.util.InputStreamHelper;

import org.eclipse.emf.ecore.EObject;

/*
 * This is a temporary workaround for the REST server until the document store
 * is extracted from the FileStorageProvider. 
 */
public class MediaStore {
    
    private static final Logger LOG = Logger.getLogger(MediaStore.class.getName());
    
    private static final ResourceBundle BUNDLE = MessageBundle.getInstance();

    private static MediaStore instance;

    private File rootDirectory;
    
    public MediaStore(String rootDirectoryName) {
        
        rootDirectory  = new File(rootDirectoryName);
        
        if (!rootDirectory.exists()) {
            LOG.info("Creating media store root directory " + rootDirectory);
            if (!rootDirectory.mkdirs()) {
                LOG.severe("Error creating media store root directory " + rootDirectory);
            }
        }
    }
    
    public static void init(String directory) {
        instance = new MediaStore(directory);
    }

    public static MediaStore getInstance() {
        return instance;
    }

    public void addMedia(RESTContext ctx, Artifact artifact) 
        throws RESTException {
        
        
        String id = RestEMFUtil.getResourceId(artifact);
        File mediaResourceFile = getMediaResourceFile(id);
        
        if (mediaResourceFile.exists()) {
            throw new RESTException(
                    RESTException.INTERNAL_SERVER_ERROR,
                    new Message(BUNDLE, "MEDIA_RESOURCE_EXISTS", 
                                mediaResourceFile.getParentFile().getAbsolutePath()));
        }
        
        if (!mediaResourceFile.getParentFile().mkdirs()) {
            throw new RESTException(
                    RESTException.INTERNAL_SERVER_ERROR,
                    new Message(BUNDLE, "CREATE_DIRS_FAILED", 
                                mediaResourceFile.getParentFile().getAbsolutePath()));
        }
        
        URI uri = null;
        try {
            
            // TODO: Using the resolved URI here is a problem since it depends 
            // on the location of the client doing the post (eg could be "localhost").
            uri = new URI(makeUri(ctx, id));
        } catch (URISyntaxException e) {
            throw new RESTException(RESTException.INTERNAL_SERVER_ERROR, e);
        }       

        LOG.info("Saving media resource to file " + mediaResourceFile);
        
        try {
            saveToDisk(ctx.getInputStream(), mediaResourceFile);
        } catch (IOException e) {
            throw new RESTException(RESTException.INTERNAL_SERVER_ERROR, e);
        }
        
        artifact.setUri(uri); 
    }
    

    public InputStream getInputStream(RESTContext ctx, EObject doc) throws RESTException {

        String id = RestEMFUtil.getResourceId(doc);
        File mediaResourceFile = getMediaResourceFile(id);
        
        LOG.info("Getting media resource file " + mediaResourceFile);

        try {
            if (!mediaResourceFile.exists()) {
                throw new FileNotFoundException();
            }

            return new FileInputStream(mediaResourceFile);
            
        } catch (FileNotFoundException e) {
            throw new RESTException(
                    RESTException.NOT_FOUND,
                    new Message(BUNDLE, "MEDIA_RESOURCE_NOT_FOUND", id));
        }        
    }

    public void updateMedia(RESTContext ctx, EObject doc)
        throws RESTException {
        
        String id = RestEMFUtil.getResourceId(doc);
        File mediaResourceFile = getMediaResourceFile(id);
        
        LOG.info("Updating media resource file " + mediaResourceFile);

        try {
            if (!mediaResourceFile.exists()) {
                throw new FileNotFoundException();
            }
                        
            try {
                saveToDisk(ctx.getInputStream(), mediaResourceFile);
            } catch (IOException e) {
                throw new RESTException(RESTException.INTERNAL_SERVER_ERROR, e);
            }
        } catch (FileNotFoundException e) {
            throw new RESTException(
                    RESTException.NOT_FOUND,
                    new Message(BUNDLE, "MEDIA_RESOURCE_NOT_FOUND", id));
        }        
    }
    
    public void deleteMedia(RESTContext ctx, EObject doc) 
        throws RESTException {
        
        String id = RestEMFUtil.getResourceId(doc);
        File mediaResourceFile = getMediaResourceFile(id);
        File dir = mediaResourceFile.getParentFile();             
        
        if (mediaResourceFile.exists()) {
            LOG.info("Deleting media resource file " + mediaResourceFile);
            if (!mediaResourceFile.delete()) {
                LOG.warning("Failed to delete media resource file " + mediaResourceFile);
            }
        } else {
            LOG.info("Media resource file " + mediaResourceFile + " alread deleted");
        }
        
        if (dir.exists()) {
            LOG.info("Deleting media resource directory " + dir);
            if (!dir.delete()) {
                LOG.warning("Failed to delete media resource directory " + dir);
            }
        }
    }
    
    public String getResourceUri(RESTContext ctx, EObject doc) {
        try {
            String id = RestEMFUtil.getResourceId(doc);
            return makeUri(ctx, id);
        } catch (RESTException e) {
            return null;
        }
    }

    // Used during tests.
    public void deleteMediaStore() {        
        if (rootDirectory.exists()) {
            // Delete the directory since this is only a temporary filestore.
            deleteDirectoryHierarchy(rootDirectory);
            if (rootDirectory.exists()) {
                LOG.warning("Media store directory " + rootDirectory + " was not deleted");
            }
        }
    }

    private File getMediaResourceFile(String id) {
        return new File(new File(rootDirectory, id), "media");
    }
    
    private void saveToDisk(InputStream is, File mediaResourceFile)
        throws RESTException {
        
        OutputStream os = null;
        try {
            if (is == null) {
                throw new RESTException(
                        RESTException.BAD_REQUEST,
                        new Message(BUNDLE, "NO_MEDIA_CONTENT"));
            }

            os = new BufferedOutputStream(new FileOutputStream(mediaResourceFile));
            InputStreamHelper.drain(is, os);
            
        } catch (IOException e) {
            throw new RESTException(RESTException.INTERNAL_SERVER_ERROR, e);
        } finally {
            InputStreamHelper.close(is);          
            InputStreamHelper.close(os);
        }
    }
    
    private String makeUri(RESTContext ctx, String id) {
        String base = ctx.getBaseUri();
        String collection = ctx.getCollection();
        if (base == null || collection == null) {
            return null;
        }

        return new StringBuilder()
            .append(base)
            .append("/")
            .append(collection)
            .append("/")
            .append(id)
            .append("?")
            .append(RESTQuery.QUERY_MEDIA).toString();
    }
    
    void deleteDirectoryHierarchy(File dir) {
        
        for (String d : dir.list()) {
            File file = new File(dir, d);
            if (file.isDirectory()) {
                deleteDirectoryHierarchy(file);
            }
            LOG.info("Deleting " + file);
            file.delete();
        }
        dir.delete();
    }
}
