/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.rest.rt.osgi;

import java.util.Collection;
import java.util.Collections;
import java.util.Hashtable;
import java.util.Map;
import java.util.logging.Logger;

import com.iona.repository.maven.osgi.IRepositoryMavenService;
import com.iona.repository.rest.osgi.RESTCollectionService;
import com.iona.repository.rest.rt.atom.ServerAtomEntryFactory;
import com.iona.repository.rest.rt.atom.ServerAtomFeedFactory;
import com.iona.repository.rest.rt.media.MediaStore;
import com.iona.repository.rest.rt.repo.RepositoryRESTAdapter;
import com.iona.repository.rest.rt.server.RESTAdapter;
import com.iona.repository.rest.rt.server.RESTProvider;
import com.iona.repository.rest.rt.server.RepositoryProvider;
import com.iona.repository.rest.rt.server.RepositoryServlet;
import com.iona.repository.taskservice.TaskService;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.SOARepositoryService;
import com.iona.soa.repository.configuration.ConfigurationKeys;
import com.iona.soa.repository.configuration.ConfigurationService;
import com.iona.soa.repository.util.ExceptionHelper;
import com.iona.soa.search.SearchService;
import com.iona.soa.security.SecurityService;

import org.osgi.framework.BundleContext;
import org.osgi.service.http.HttpService;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.osgi.context.BundleContextAware;
import org.springframework.osgi.service.importer.OsgiServiceLifecycleListener;

public class RestServiceBean implements InitializingBean, OsgiServiceLifecycleListener, BundleContextAware {
    private static final Logger LOG = Logger.getLogger(RestServiceBean.class.getName());
    
    RepositoryRESTAdapter restAdapter;

    private BundleContext bundleContext;
    private SOARepositoryService repositoryService;
    private HttpService httpService;
    private ConfigurationService configService;
    private IRepositoryMavenService mavenSpiderService;
    private Collection<TaskService> taskServices = Collections.emptySet(); // injected
    
    public RestServiceBean() { }

    public IRepositoryMavenService getMavenSpiderService() {
        return mavenSpiderService;
    }

    public void setBundleContext(BundleContext bc) {
        bundleContext = bc;        
    }

    public void setMavenSpiderService(IRepositoryMavenService mavenSpiderService) {
        LOG.fine("Spider service available");
        this.mavenSpiderService = mavenSpiderService;
    }
    
    public void setRepositoryService(SOARepositoryService repositoryService) {
        LOG.fine("Repository service available");
        this.repositoryService = repositoryService;
    }
        
    public void setHttpService(HttpService httpService) {
        LOG.fine("HTTP service available");
        this.httpService = httpService;
    }
        
    public void setConfigurationService(ConfigurationService c) {
        LOG.fine("Configuration service available");
        configService = c;
    }
    
    public Collection<TaskService> getTaskServices() {
        return taskServices;
    }
    
    public void setRestCollectionServices(Collection<RESTCollectionService> rcs) {
        // these are currently handled via the listener
    }    
    
    public void setTaskServices(Collection<TaskService> ts) {
        taskServices = ts;
    }
    
    public void afterPropertiesSet() throws Exception {
        init();
    }
    
    private void init() {
        
        LOG.fine("Initializing REST service");
                
        Map<String, String> config = configService.getConfiguration();
                
        String servletAlias = config.get(ConfigurationKeys.ATOM_SERVLET_ALIAS);
        if (servletAlias == null) {
            servletAlias = "/repository/atom";
        }
        LOG.info("Creating REST servlet using alias " + servletAlias);


        SOARepository soaRepo = null;
        try {
            soaRepo = repositoryService.getRepository();
        } catch (Throwable t) {
            LOG.severe("Got exception " + t + " trying to get SOA repository object");
            LOG.info(ExceptionHelper.getStackTraceAsString(t));
        }
        
       
        if (soaRepo == null) {
            LOG.severe("SOA Repository object is null");
        }

        SecurityService securityService = repositoryService.getSecurityService();

        String isInsecure = config.get("atom.insecure");
        if ("true".equals(isInsecure)) {
            securityService = null;
        }
        
        String mediaStoreDir = config.get(ConfigurationKeys.MEDIA_STORE_DIR);
        if (mediaStoreDir == null) {
            mediaStoreDir = "media-store";
        }
        MediaStore.init(mediaStoreDir);
        
        SearchService searchService = repositoryService.getSearchService();
        initRESTAdapter(config, soaRepo, securityService, searchService);        
        
        RepositoryProvider provider = 
            new RepositoryProvider(
                    restAdapter,
                    securityService,
                    new ServerAtomFeedFactory(),
                    new ServerAtomEntryFactory(),
                    servletAlias);

        RepositoryServlet servlet = new RepositoryServlet(provider);

        try {
            httpService.registerServlet(servletAlias, servlet, null, null);
            LOG.info("Registered REST servlet using alias " + servletAlias);

        } catch (Exception e) {
            LOG.severe("Failed to register REST servlet; exception " + e);
        }
    }

    void initRESTAdapter(Map<String, String> config,
            SOARepository soaRepo, SecurityService securityService,
            SearchService searchService) {
        restAdapter = new RepositoryRESTAdapter(soaRepo, 
                                                securityService,
                                                searchService,
                                                config,
                                                mavenSpiderService,
                                                this);
        bundleContext.registerService(
            new String [] {RESTAdapter.class.getName(), RESTProvider.class.getName()},
            restAdapter, new Hashtable<Object, Object>());
    }

    @SuppressWarnings("unchecked")
    public void bind(Object service, Map properties) throws Exception {
        restAdapter.registerCollectionService((RESTCollectionService) service);
    }

    @SuppressWarnings("unchecked")
    public void unbind(Object service, Map properties) throws Exception {
        restAdapter.unregisterCollectionService((RESTCollectionService) service);
    }
}
