/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.rest.rt.repo;

import java.io.IOException;
import java.sql.Timestamp;
import java.util.ResourceBundle;
import java.util.logging.Logger;

import com.iona.repository.rest.Constants;
import com.iona.repository.rest.RESTContext;
import com.iona.repository.rest.RESTException;
import com.iona.repository.rest.resource.Attribute;
import com.iona.repository.rest.resource.Reference;
import com.iona.repository.rest.resource.Resource;
import com.iona.repository.rest.rt.osgi.MessageBundle;
import com.iona.repository.rest.rt.server.RESTProvider;
import com.iona.repository.util.i18n.Message;
import com.iona.soa.model.repository.IPersistableObject;
import com.iona.soa.repository.EMFUtil;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.util.EcoreUtil;

public abstract class AbstractEFeatureHandler implements EFeatureHandler {

    protected static final Logger LOG = Logger.getLogger(AbstractEFeatureHandler.class.getName());

    protected static final ResourceBundle BUNDLE = MessageBundle.getInstance();
    protected RESTProvider restProvider;

    public AbstractEFeatureHandler(RESTProvider a) {
        restProvider = a;
    }
    
    protected void addAttributeToResource(String qualName,
                                          EAttribute att, 
                                          Object value,
                                          Resource resource) throws RESTException {

        String name = att.getName();
        Attribute resourceAtt = resource.addAttribute(name);

        resourceAtt.setMin(att.getLowerBound());
        resourceAtt.setMax(att.getUpperBound());
        boolean isHidden = restProvider.isHidden(qualName);
        if (isHidden) {
            resourceAtt.setHidden(true);
        }
        if (restProvider.isReadonly(qualName)) {
            resourceAtt.setReadonly(true);
        }

        resourceAtt.setType(restProvider.getXsdType(att));
        if (value != null && !isHidden) {
            EDataType type = att.getEAttributeType();

            // Special case for timestamp since this can be used for sorting 
            // a collection and parsing text based dates can cause problems.
            if (type.getInstanceClass().equals(Timestamp.class)) {               
                resourceAtt.setObject(value);
            } else {                
                resourceAtt.setText(EcoreUtil.convertToString(type, value));
            }
        }
    }

    protected void addReferenceToResource(EReference ref,
                                          EObject eobject,
                                          int min,
                                          int max,
                                          Resource resource) throws RESTException {

        // Get either the class of the actual object or the 
        // referenced class if there is no object.
        EClass refClass = null;
        if (eobject == null) {
            refClass = ref.getEReferenceType();
        } else {
            refClass = eobject.eClass();
        }
        
        String className = RestEMFUtil.getQualifiedClassName(refClass);
        if (restProvider.isReferenceable(className)) {

            Reference reference = resource.addReference(ref.getName());
            reference.setMin(min);
            reference.setMax(max);
            reference.setType(Constants.ATOM_ENTRY_CONTENT_TYPE);
            reference.setEClass(RestEMFUtil.getQualifiedClassName(ref.getEReferenceType()));


            String colHref = restProvider.getCollectionHref(className);
            if (colHref != null && !ref.isContainment()) {
                reference.setCollectionHref(colHref);
            }

            if (eobject != null) {                
                String href = getHref(eobject, colHref);
                reference.setHref(href);

                String name = RestEMFUtil.getResourceName(eobject);
                reference.setText(name);                
            }
        } else {
            restProvider.internalError(
                    "reference " + RestEMFUtil.getQualifiedName(ref)
                    + " of class " + className + " is not referenceable");
        }
    }

    @SuppressWarnings("unchecked")
    protected void addSubReferenceToResource(String qualName, EObject eobject,
                                             EReference ref, Resource resource)
        throws RESTException {

        EClass refClass = eobject.eClass();
        
        String qualClassName = RestEMFUtil.getQualifiedClassName(refClass);
        if (restProvider.isReferenceable(qualClassName)) {

            int size = ref.isMany()
                       ? ((EList)eobject.eGet(ref)).size()
                       : (eobject.eIsSet(ref) ? 1 : 0);
            Reference reference = resource.addReference(ref.getName());
            reference.setMin(ref.getLowerBound());
            reference.setMax(ref.getUpperBound());
            reference.setSize(size);
            reference.setType(Constants.ATOM_FEED_CONTENT_TYPE);
            reference.setEClass(RestEMFUtil.getQualifiedClassName(ref.getEReferenceType()));

            String colHref = restProvider.getCollectionHref(qualClassName);

            String href = getHref(eobject, colHref, ref.getName());
            reference.setHref(href);
            
        } else {
            restProvider.internalError(
                    "reference " + RestEMFUtil.getQualifiedName(ref)
                    + " of class " + qualClassName + " is not referenceable");
        }
    }
    
    protected void getAttributeChanges(RESTContext ctx, EAttribute att, EObject eobject,
                                       Resource resource) {
        
        Attribute resourceAtt = resource.getFirstAttribute(att.getName());
        if (resourceAtt != null) {
            
            Object newValue = getObject(resourceAtt, att.getEAttributeType());
            Object oldValue = eobject.eGet(att);
            
            if ((oldValue == null && newValue != null)
                || (oldValue != null && !oldValue.equals(newValue))) {
                eobject.eSet(att, newValue);
            }
        }        
    }


    protected void getReferenceChanges(RESTContext ctx, EReference ref, EObject eobject, 
                                   EObject newObject) {
        eobject.eSet(ref, newObject);
    }
        
    protected Object getObject(Attribute att, EDataType type) {
        if (att.getText() == null || att.getText().length() == 0) {
            return null;
        } else {
            return EcoreUtil.createFromString(type, att.getText());
        }
    }

    protected EObject getReferencedObject(Reference reference, RESTContext ctx) throws RESTException {
        
        String href = reference.getHref();
        if (href == null) {
            return null;
        }
        String[] parts = href.split("/");
        if (parts.length != 2) {
            throw new RESTException(
                    RESTException.BAD_REQUEST,
                    new Message(BUNDLE, "BAD_REFERENCE", reference.getHref()));
        }

        EClass eclass = restProvider.getEClass(parts[0]);
        if (eclass == null) {
            throw new RESTException(
                    RESTException.BAD_REQUEST,
                    new Message(BUNDLE, "UNKNOWN_REFERENCE_COLLECTION",
                            reference.getHref()));
        }

        String eid = RestEMFUtil.getObjectID(eclass, parts[1]);
        URI uri = EMFUtil.getProxyURI(eclass, eid);        
        EObject eobject = restProvider.getRepository().getEObject(uri, true);
        if (eobject == null) {
            throw new RESTException(
                    RESTException.INTERNAL_SERVER_ERROR,
                    new Message(BUNDLE, "MISSING_OBJECT",
                            reference.getHref()));               
        }

        return eobject;

    }

    @SuppressWarnings("unchecked")
    public void getAttributeListChanges(RESTContext ctx, EAttribute att,
                                        EObject eobject, EList newList, 
                                        EList oldList) {
        // TODO: what about ordering/uniqueness?
        EList addList = new BasicEList();
        EList removeList = new BasicEList();
        
        for (Object o : newList) {
            if (!oldList.contains(o)) {
                addList.add(o);
            }
        }
        for (Object o : oldList) {
            if (!newList.contains(o)) {
                removeList.add(o);
            }
        }

        if (addList.size() == 0 && removeList.size() == 0) {
            return;
        }
        
        EList values = (EList)eobject.eGet(att);
        
        if (addList.size() > 0) {
            for (Object o : addList) {
                values.add(o);
            }
        }
        
        if (removeList.size() > 0) {
            for (Object o : removeList) {
                values.remove(o);
            }
        }
    }


    @SuppressWarnings("unchecked")
    public void getReferenceListChanges(RESTContext ctx, EReference ref,
                                        EObject eobject, EList<EObject> newList, 
                                        EList<EObject> oldList) {
        // TODO: what about ordering/uniqueness?
        EList<EObject> addList = new BasicEList<EObject>();
        EList<EObject> removeList = new BasicEList<EObject>();
        
        for (EObject o : newList) {
            if (!oldList.contains(o)) {
                addList.add(o);
            }
        }
        for (EObject o : oldList) {
            if (!newList.contains(o) && !RestEMFUtil.isInternal(o)) {
                removeList.add(o);
            }
        }

        if (addList.size() == 0 && removeList.size() == 0) {
            return;
        }
        
        EList<EObject> values = (EList<EObject>)eobject.eGet(ref);

        if (addList.size() > 0) {
            for (EObject o : addList) {
                values.add(o);
            }
        }
        
        if (removeList.size() > 0) {
            for (EObject o : removeList) {
                values.remove(o);
            }
        }
    }
    
    protected void addNewObject(EObject eobject) throws RESTException {
        // TODO what's the best way of doing this?
        try {
            restProvider.getRepository().persist(eobject);
        } catch (IOException e) {
            throw new RESTException(RESTException.INTERNAL_SERVER_ERROR, e);
        }        
    }
    
    protected String getHref(EObject eobject, String collectionHref) throws RESTException {
        IPersistableObject po = (IPersistableObject) eobject;
        String id = RestEMFUtil.getResourceId(po);
        return collectionHref + "/" + id;
    }
    
    private String getHref(EObject eobject, String collHref, String subname) throws RESTException {
        
        IPersistableObject po = (IPersistableObject) eobject;
        String id = RestEMFUtil.getResourceId(po);
        return collHref + "/" + id + ";" + subname;
    }
    
    protected long getVersion(Object object) {
        if (object instanceof EObject) {
            return EMFUtil.getGenerationNumber((EObject)object, -1L);
        }
        return -1;
        
        /*
        EAttribute genNrAttr = EMFUtil.getGenerationNumberAttribute(eobject());
        Long version = (Long) eobject.eGet(genNrAttr);
        */
    }
    
}
