/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.rest.rt.repo;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.repository.maven.osgi.IRepositoryMavenService;
import com.iona.repository.rest.RESTContext;
import com.iona.repository.rest.RESTException;
import com.iona.repository.rest.RESTQuery;
import com.iona.repository.rest.resource.Attribute;
import com.iona.repository.rest.resource.Reference;
import com.iona.repository.rest.resource.Resource;
import com.iona.soa.expression.And;
import com.iona.soa.expression.Equals;
import com.iona.soa.expression.Expression;
import com.iona.soa.model.Artifact;
import com.iona.soa.model.ArtifactType;
import com.iona.soa.model.MavenRepository;
import com.iona.soa.model.SOA_NetworkFactory;
import com.iona.soa.model.SOA_NetworkPackage;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.TransactionData;
import com.iona.soa.resourceset.QueryResourceSet;
import com.iona.soa.resourceset.ResourceSetQueryHandler;
import com.iona.soa.search.SearchService;
import com.iona.soa.security.SecurityContext;

import org.apache.maven.shared.dependency.tree.DependencyNode;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;

public class ArtifactResolveExecutor extends AbstractMavenQueryExecutor {

    public static final String RSC_ATTR_DEPENDS = "depends";
    public static final String RSC_ATTR_DEPENDS_IDS = "dependsIDs";

    private static final Logger LOG = Logger.getLogger(ArtifactResolveExecutor.class.getName());

    private IRepositoryMavenService mavenService;
    private RepositoryRESTAdapter restProvider;
    private Map<URI, MavenRepository> urlToReposMap;

    public ArtifactResolveExecutor(SOARepository soaRepo, SearchService searchService,
                                   IRepositoryMavenService mvnService, RepositoryRESTAdapter restProvider) {
        super(soaRepo, searchService);
        this.mavenService = mvnService;
        this.restProvider = restProvider;
    }

    @SuppressWarnings("unchecked")
    @Override
    public Collection<Resource> queryResources(RESTContext ctx, TransactionData trnsData, RESTQuery query)
        throws RESTException {

        String artifactId = query.getDecodedParameter(RESTQuery.QUERY_MVN_ARTIFACT_ID);
        String groupId = query.getDecodedParameter(RESTQuery.QUERY_MVN_GROUP_ID);
        String version = query.getDecodedParameter(RESTQuery.QUERY_MVN_VERSION);
        String mavenGuid = query.getDecodedParameter(RESTQuery.QUERY_MVN_REPOS_GUID);
        String scope = query.getDecodedParameter(RESTQuery.QUERY_MVN_SCOPE);
        if ((artifactId != null) && (groupId != null) && (version != null)) {
            List<Resource> foundDep = resolveMavenDependencies(groupId, 
                    artifactId, version, mavenGuid, scope, ctx);
            return foundDep;
        }

        // TODO: Log error

        return new ArrayList();
    }

    @SuppressWarnings("unchecked")
    private List<Resource> resolveMavenDependencies(String groupId, String artifactId, String version,
        String mavenGuid, String scope, RESTContext ctx) throws RESTException {
        try {
            List<MavenRepository> mavenRepoList = new ArrayList();
            MavenRepository specifiedRepo = null;
            if (mavenGuid != null) {
                specifiedRepo = getMavenRepo(mavenGuid);
                if (specifiedRepo != null) {
                    mavenRepoList.add(specifiedRepo);
                }
            }

            // TODO: Replace this code when the SOARepository is available in
            // the
            // RepositoryMavenServiceBean
            List<EObject> reposObjs = ResourceSetQueryHandler.executeQuery(getSOARepository()
                .getResourceSet(), SOA_NetworkPackage.eINSTANCE.getMavenRepository(), true);
            for (EObject obj : reposObjs) {
                MavenRepository currentRepo = (MavenRepository)obj;
                if (specifiedRepo == null || !currentRepo.getGuid().equals(specifiedRepo.getGuid())) {
                    mavenRepoList.add(currentRepo);
                }
            }

            urlToReposMap = new HashMap<URI, MavenRepository>(mavenRepoList.size());

            for (MavenRepository repos : mavenRepoList) {
                // Unfortunately this required field can be null
                if (repos.getUrl() != null) {
                    urlToReposMap.put(repos.getUrl(), repos);
                }
            }

            Map<String, ArtifactType> knownTypes = MavenRepoQueryExecutor
                .getArtifactTypesKnownToDepot(getSOARepository().getResourceSet());
            DependencyNode depNode;

            depNode = mavenService.resolve(groupId, artifactId, version, scope, mavenRepoList);
            if (depNode == null) {
                return new ArrayList<Resource>(); 
            }


            Map<String, Resource> resources = new HashMap<String, Resource>();
            Resource resource = constructDepResourceStruct((DependencyNode)depNode.getChildren().get(0),
                                                           null, resources, knownTypes, ctx);
            // mark the top level node... might be useful to the client...
            resource.addAttribute("depRoot").setText("true");

            return new ArrayList<Resource>(resources.values());

        } catch (Exception e) {
            throw new RESTException(RESTException.INTERNAL_SERVER_ERROR, e);
        }
    }

    @SuppressWarnings("unchecked")
    private Resource constructDepResourceStruct(DependencyNode depRoot,
                                                List<Artifact> deepRootRepEObjCandidates,
                                                Map<String, Resource> allDepResources,
                                                Map<String, ArtifactType> knownTypes, RESTContext ctx)
        throws RESTException, URISyntaxException {

        Artifact eObjArtifact = null;

        // 1st check the resource processed already...
        Resource depRsc = findDepInProcessedResources(depRoot, allDepResources, ctx);

        String mvnVersion = depRoot.getArtifact().getVersion();
        String mvnArtifactId = depRoot.getArtifact().getArtifactId();
        String mvnGroupId = depRoot.getArtifact().getGroupId();
        String mvnArtifactType = depRoot.getArtifact().getArtifactHandler().getPackaging();
        String mvnDownloadUrl = depRoot.getArtifact().getDownloadUrl();
        MavenRepository foundIn = findContainingRepos(depRoot, ctx);

        // 2nd check the list of eobj candidates (obtained from the "depends"
        // ref of the parent eobj, if any)
        if (depRsc == null && deepRootRepEObjCandidates != null && deepRootRepEObjCandidates.size() > 0) {

            for (Artifact candidate : deepRootRepEObjCandidates) {
                if (candidate.getName().equals(mvnArtifactId) && candidate.getNamespace().equals(mvnGroupId)
                    && candidate.getVersion().equals(mvnVersion)
                    && (foundIn == null || candidate.getDepends().contains(foundIn))) {
                    eObjArtifact = candidate;
                    depRsc = restProvider.doGetResource(null, null, restProvider.getDescriptor("artifacts"),
                                                        candidate);
                    break;
                }
            }
        }

        // 3rd search the model for a matching artifact eObj.
        if (depRsc == null) {
            try {

                QueryResourceSet rs = restProvider.getRepository().getResourceSet();
                Expression expr = constructQueryExpression(mvnVersion, mvnArtifactId, mvnGroupId, foundIn);

                EClass ec = SOA_NetworkPackage.eINSTANCE.getArtifact();
                List<EObject> eobjs = ResourceSetQueryHandler.executeQuery(rs, ec, expr, true);
                if (eobjs.size() == 1) {
                    eObjArtifact = (Artifact)eobjs.get(0);
                    depRsc = restProvider.doGetResource(ctx, RepositoryRESTAdapter.getSecurityContext(ctx),
                                                        restProvider.getDescriptor("artifacts"),
                                                        eObjArtifact, true);
                    allDepResources.put(depRsc.getId(), depRsc);
                }
            } catch (IOException e) {
                LOG.log(Level.FINE, "exception caught", e);
            } catch (Exception e) {
                LOG.log(Level.FINE, "exception caught", e);
            }
        }

        // 4th create a ad-hoc artifact eObj to represent the dependancy...
        if (depRsc == null) {

            eObjArtifact = SOA_NetworkFactory.eINSTANCE.createArtifact();
            eObjArtifact.setName(mvnArtifactId);
            eObjArtifact.setNamespace(mvnGroupId);
            eObjArtifact.setUri(new URI(mvnDownloadUrl));
            eObjArtifact.setFileName(mvnDownloadUrl.substring(mvnDownloadUrl.lastIndexOf('/') + 1));
            eObjArtifact.setVersion(mvnVersion);

            if (foundIn != null) {
                eObjArtifact.getFoundIn().add(foundIn);
            }

            MavenRepoQueryExecutor.guessArtifactType(eObjArtifact, mvnArtifactType, knownTypes);

            eObjArtifact.setModifiable(false);
            eObjArtifact.setGuid(null);

            // generating rsc out of obj
            depRsc = restProvider.doGetResource(ctx, RepositoryRESTAdapter.getSecurityContext(ctx),
                                                restProvider.getDescriptor("artifacts"), eObjArtifact);
            depRsc.setId(MavenRepoQueryExecutor.constructRscKey(eObjArtifact));

            allDepResources.put(depRsc.getId(), depRsc);
        }

        // adding "resolved" flag to indicated that the details of the artifact
        // were retrieved sucessfully
        boolean isResolved = !mvnDownloadUrl.startsWith("null");

        Attribute resolvedAttr = depRsc.addAttribute("resolved");
        resolvedAttr.setText("" + isResolved);
        resolvedAttr.setType("xsd:boolean");
        resolvedAttr.setMax(1);
        resolvedAttr.setMin(0);
        resolvedAttr.setReadonly(true);

        processNextLevelDeps(depRsc, eObjArtifact, depRoot.getChildren(), allDepResources, knownTypes, ctx);

        return depRsc;
    }

    @SuppressWarnings("unchecked")
    private void processNextLevelDeps(Resource parentResource, Artifact parentEObj,
                                      List<DependencyNode> nextLevelDeps,
                                      Map<String, Resource> allDepResources,
                                      Map<String, ArtifactType> knownTypes, RESTContext ctx)
        throws RESTException, URISyntaxException {

        if (nextLevelDeps == null || nextLevelDeps.size() == 0) {
            // no sub dep : getting out ...
            return;
        }

        for (DependencyNode dep : nextLevelDeps) {
            Resource depRsc = constructDepResourceStruct(dep, parentEObj == null ? null : parentEObj
                .getDepends(), allDepResources, knownTypes, ctx);

            // adding dep ref to parent rsc
            Reference refTemplate = parentResource.getFirstReference(RSC_ATTR_DEPENDS);
            Reference depRef = parentResource.addReference(RSC_ATTR_DEPENDS_IDS);
            depRef.setText(depRsc.getName());
            depRef.setHref(depRsc.getId());
            depRef.setCollectionHref(refTemplate.getCollectionHref());
            depRef.setEClass(refTemplate.getEClass());
            depRef.setMax(-1);
            depRef.setMin(0);
            depRef.setType(refTemplate.getType());

        }
    }

    private Resource findDepInProcessedResources(DependencyNode dep, Map<String, Resource> allRscs,
                                                 RESTContext ctx) throws URISyntaxException {
        String extention = dep.getArtifact().getDownloadUrl().substring(
                                                                        dep.getArtifact().getDownloadUrl()
                                                                            .lastIndexOf("."));
        String depKey = MavenRepoQueryExecutor.constructRscKey(dep.getArtifact().getArtifactId(), dep
            .getArtifact().getGroupId(), dep.getArtifact().getVersion(), extention, findContainingRepos(dep,
                                                                                                        ctx));
        // no point to generate duplicate resources
        Resource depRsc = allRscs.get(depKey);
        return depRsc;
    }

    private MavenRepository findContainingRepos(DependencyNode dep, RESTContext ctx)
        throws URISyntaxException {
        String downloadUrl = dep.getArtifact().getDownloadUrl();
        MavenRepository repo = null;
        if (downloadUrl != null) {
            for (URI reposUri : urlToReposMap.keySet()) {
                if (downloadUrl.startsWith(reposUri.toString())) {
                    repo = urlToReposMap.get(reposUri);
                }
            }

            if (repo == null) {
                boolean passed = false;
                String groupId = dep.getArtifact().getGroupId();
                String groupUrlPart = groupId.replace(".", "/");
                String[] repoUrls = downloadUrl.split(groupUrlPart);
                if (repoUrls.length > 0) {
                    String repoUrl = repoUrls[0];
                    if (repoUrl != null) {
                        SecurityContext securityCtx = RepositoryRESTAdapter.getSecurityContext(ctx);
                        restProvider.getRepository().endRead(true);
                        restProvider.getRepository().startWrite(RepositoryRESTAdapter
                                                                .getTransactionData(securityCtx));
                        try {
                            repo = SOA_NetworkFactory.eINSTANCE.createMavenRepository();
                            repo.setUrl(new URI(repoUrl));
                            repo.setName(repo.getUrl().getHost() + " Repository");
                            repo.setId(repo.getUrl().getHost());
                            restProvider.getRepository().persist(repo);
                            urlToReposMap.put(repo.getUrl(), repo);
                            passed = true;
                        } catch (IOException ioe) {
                            LOG.warning("Error persisting maven repository with url: " + repoUrl);  
                        } finally {
                            restProvider.getRepository().endWrite(passed);
                            restProvider.getRepository().startRead(RepositoryRESTAdapter
                                                                   .getTransactionData(securityCtx));
                        }
                    }
                }

            }
        }

        return repo;
    }

    private Expression constructQueryExpression(String mvnVersion, String mvnArtifactId, String mvnGroupId,
                                                MavenRepository foundIn) {
        Expression expr = new And(new Equals("name", mvnArtifactId), new Equals("namespace", mvnGroupId),
                                  new Equals("version", mvnVersion));

        if (foundIn != null) {
            expr = new And(expr, new Equals("foundIn.guid", foundIn.getGuid()));
        }

        return expr;
    }

}
