/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.rest.rt.repo;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import com.iona.repository.rest.Constants;
import com.iona.repository.rest.RESTContext;
import com.iona.repository.rest.RESTException;
import com.iona.repository.rest.resource.Attribute;
import com.iona.repository.rest.resource.Reference;
import com.iona.repository.rest.resource.Resource;
import com.iona.repository.rest.resource.ResourceFactory;
import com.iona.repository.rest.resource.ResourceFactoryImpl;
import com.iona.repository.rest.rt.media.MediaStore;
import com.iona.repository.rest.rt.server.RESTProvider;
import com.iona.soa.model.Artifact;
import com.iona.soa.model.SOA_NetworkFactory;
import com.iona.soa.repository.EMFUtil;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EObject;


public class ArtifactsHandler extends DefaultEClassHandler {
    private static final String UNKNOWN_FILENAME = "unknown";
    private static final String[] ACCEPTS = {
        "application/atom+xml",
        "*/*"
    };

    private static final SOA_NetworkFactory FACTORY = SOA_NetworkFactory.eINSTANCE;
    private static ResourceFactory resourceFactory = new ResourceFactoryImpl();
    
    MediaStore mediaStore = MediaStore.getInstance();

    public ArtifactsHandler(RESTProvider a) {
        super(a);
    }
    
    @Override
    public String[] getAccepts() {
        return ACCEPTS;
    }

    @Override
    public void getMedia(RESTContext ctx, EObject eobject) throws RESTException {   
        
        ctx.setContentType("application/octet-stream");
        InputStream is = mediaStore.getInputStream(ctx, eobject);
        ctx.setInputStream(is);
    }

    @Override
    public EObject updateMedia(RESTContext ctx, EObject eobject)
        throws RESTException {
        mediaStore.updateMedia(ctx, eobject);
        return eobject;
    }
    
    @Override
    public void deleteObject(RESTContext ctx, EObject eobject) 
        throws RESTException {
        
        super.deleteObject(ctx, eobject);
        mediaStore.deleteMedia(ctx, eobject);
    }

    protected Attribute makeReadonly(Resource resource, String name) {
        Attribute att = resource.getFirstAttribute(name);
        if (att == null) {
            return null;
        }
        return att.setReadonly(true);
    }

    @Override
    public void addObjectToResource(RESTContext ctx, EObject eobject, Resource resource)
        throws RESTException {
        
        super.addObjectToResource(ctx, eobject, resource);
        Artifact art = (Artifact) eobject;

        if (!art.isExternal()) {
            Attribute att = makeReadonly(resource, "uri");
            String uri = mediaStore.getResourceUri(ctx, art);
            if (uri != null && att != null) {
                resource.setProperty("MediaContentType", "application/octet-stream");
                resource.setProperty("MediaUri", uri);
               
                // Dynamically override the value of `documentUri` so that 
                // it is accessable from the client's location.   (ie it might
                // be stored using "localhost" in the repository.              
                att.setText(uri);                
            }
        }
    }


    @Override
    public EObject createObjectWithMedia(RESTContext ctx, EFactory efactory, EClass eclass)
        throws RESTException {
        
        try {
            Artifact art = FACTORY.createArtifact();
            ctx.getRepo().persist(art);
            
            String name = ctx.getSlug();
            if (name == null || "".equals(name)) {
                name = UNKNOWN_FILENAME;
            }

            art.setName(name);
            art.setFileName(name);
            art.setExternal(false);
            
            mediaStore.addMedia(ctx, art);
            
            return art;
        } catch (Exception e) {
            throw new RESTException(RESTException.INTERNAL_SERVER_ERROR, e);
        }        
    }

    @Override
    public EObject createObject(RESTContext ctx,
                                EFactory efactory,
                                EClass eclass, 
                                String name,
                                Resource resource,
                                boolean persist) throws RESTException {
        
        Artifact art = (Artifact) super.createObject(ctx, efactory, eclass, name, null, persist);
        art.setExternal(true);
        // change descriptor needs to know the url to the resources's parent
        
        if (resource != null) {
            // go through the list create a changes for the depend - 
            // push in attribute to for the depends and then let the getOBjecty
            getObjectChanges(ctx, art, resource);
        }
        
        removeVersionFromFileName(art);

        return art;
    }

    @Override
    public Collection<? extends EObject> createObjects(RESTContext ctx,
                                                       EFactory efactory,
                                                       EClass eclass,
                                                       Collection<Resource> resources,
                                                       boolean persist) throws RESTException {

        if (resources != null && resources.size() > 0) {

            Map<String, Artifact> persistedArtifactByRscId = new HashMap<String, Artifact>();
            Map<String, Resource> requestRscsById = new HashMap<String, Resource>();

            for (Resource rsc : resources) {
                requestRscsById.put(rsc.getId(), rsc);
            }

            RecursionCtx rc = new RecursionCtx(ctx, eclass, efactory, persist);
            Collection<String> empty = Collections.emptyList();
            for (Resource rsc : resources) {
                doCreateArtifactTree(rsc,
                        requestRscsById,
                        persistedArtifactByRscId,
                        empty,
                        rc);
            }

            return persistedArtifactByRscId.values();
        }

        return Collections.emptyList();
    }
    
    private void doCreateArtifactTree(Resource targetRsc,
                                      Map<String, Resource> requestRscsById,
                                      Map<String, Artifact> persistedArtifactByRscId,
                                      Collection<String> depIdsPath,
                                      RecursionCtx rc) throws RESTException {
        
        String targetrscId = targetRsc.getId();
        
        if (requestRscsById.get(targetrscId) == null
                || persistedArtifactByRscId.get(targetrscId) != null
                || depIdsPath.contains(targetrscId)) {
            return;
        }
        
        
        // 1st create the dependencies (listed in an ad-hoc attribute "dependsIDs" by the maven resolver)
        // and add their ref to the attribute "depends" (as per emf model)
        String attrName = ArtifactResolveExecutor.RSC_ATTR_DEPENDS_IDS;
        for (Reference dependsByIdAttrs : targetRsc.getReferences(attrName)) {
            String id = dependsByIdAttrs.getHref();
            
            if (requestRscsById.get(id) != null) {
                
                Collection<String> incrementedDepIdsPath = new ArrayList<String>();
                incrementedDepIdsPath.addAll(depIdsPath);
                incrementedDepIdsPath.add(targetrscId);
                
                doCreateArtifactTree(requestRscsById.get(id),
                                     requestRscsById,
                                     persistedArtifactByRscId,
                                     incrementedDepIdsPath,
                                     rc);
                addDepRefToArtifactResource(targetRsc, persistedArtifactByRscId.get(id));
            } else {
                
                
                String eid = RestEMFUtil.getObjectID(rc.getEclass(), id);
                URI uri = EMFUtil.getProxyURI(rc.getEclass(), eid);        
                if (restProvider.getRepository().getEObject(uri, true) != null) {
                    addDepRefToArtifactResource(targetRsc, id);
                }
            }
        }
        
        Artifact artifactEObj = null;
        if (targetRsc.getFirstAttribute("guid") != null
                && targetRsc.getFirstAttribute("guid").getText() != null
                && targetRsc.getFirstAttribute("guid").getText().trim().length() > 0) {
            try {
                String eid = RestEMFUtil.getObjectID(rc.getEclass(), targetrscId);
                URI uri = EMFUtil.getProxyURI(rc.getEclass(), eid);        
                artifactEObj = (Artifact)restProvider.getRepository().getEObject(uri, true);
            } catch (Exception e) { /* doesn't matter */ }
        }
        
        if (artifactEObj == null) {
            String name = targetRsc.getFirstAttribute("name").getText();
            artifactEObj = (Artifact) createObject(rc.getCtx(),
                                                   rc.getEfactory(),
                                                   rc.getEclass(),
                                                   name,
                                                   null,
                                                   rc.isPersist());
        }
        
        // change descriptor needs to know the url to the resources's parent
        
        // go through the list create a changes for the depend - 
        // push in attribute to for the depends and then let the getOBjecty
        getObjectChanges(rc.getCtx(), artifactEObj, targetRsc);
        
        removeVersionFromFileName(artifactEObj);

        persistedArtifactByRscId.put(targetrscId, artifactEObj);
    }
    
    private void addDepRefToArtifactResource(Resource artifactRsc,
                                             Artifact childArtifact) throws RESTException {
        
        if (artifactRsc == null || childArtifact == null) {
            return;
        }
        addDepRefToArtifactResource(artifactRsc, RestEMFUtil.getResourceId(childArtifact));
    }

    private void addDepRefToArtifactResource(Resource artifactRsc, String depId) {
        if (artifactRsc == null || depId == null || depId.trim().length() == 0) {
            return;
        }

        for (Reference ref : artifactRsc.getReferences("depends")) {
            if (ref.getHref() != null && ref.getHref().equals("artifacts/" + depId)) {
                return;
            }
        }
        
        if (artifactRsc.getReferences("depends").size() == 1
                && (artifactRsc.getFirstReference("depends").getHref() == null
                        || artifactRsc.getFirstReference("depends").getHref().trim().length() == 0)) {
            artifactRsc.getReferences("depends").remove(0);
        }
        
        Reference depends = resourceFactory.createReference("depends");
        //depends.set
        depends.setMin(0);
        depends.setMax(-1);
        depends.setSize(-1);
        depends.setType(Constants.ATOM_ENTRY_CONTENT_TYPE);
        
        depends.setHref("artifacts/" + depId);            
        artifactRsc.addReference(depends);
    }
    
    private void removeVersionFromFileName(Artifact art) {
        if (art.getFileName() != null && art.getVersion() != null) {
            art.setFileName(art.getFileName().replace("-" + art.getVersion(), ""));
        }
    }

    private class RecursionCtx {
        private RESTContext ctx;
        private EFactory efactory;
        private EClass eclass;
        private boolean persist;
        
        
        public RecursionCtx(RESTContext ctx, EClass eclass, EFactory efactory, boolean persist) {
            super();
            this.ctx = ctx;
            this.eclass = eclass;
            this.efactory = efactory;
            this.persist = persist;
        }

        protected RESTContext getCtx() { return ctx; }
        protected EFactory getEfactory() { return efactory; }
        protected EClass getEclass() { return eclass; }
        protected boolean isPersist() { return persist; }
    }

}
