/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.rest.rt.repo;

import java.io.IOException;
import java.util.Collection;
import java.util.LinkedList;
import java.util.logging.Level;

import com.iona.repository.rest.Constants;
import com.iona.repository.rest.RESTContext;
import com.iona.repository.rest.RESTException;
import com.iona.repository.rest.repo.EClassHandler;
import com.iona.repository.rest.resource.Attribute;
import com.iona.repository.rest.resource.Reference;
import com.iona.repository.rest.resource.Resource;
import com.iona.repository.rest.rt.server.RESTProvider;
import com.iona.repository.util.i18n.Message;
import com.iona.soa.model.security.IProtectedRepositoryObject;
import com.iona.soa.model.security.ObjectPermission;
import com.iona.soa.model.security.Role;
import com.iona.soa.model.security.SecurityArea;
import com.iona.soa.model.security.SecurityFactory;
import com.iona.soa.model.security.SecurityPackage;
import com.iona.soa.resourceset.ResourceSetQueryHandler;
import com.iona.soa.security.SecurityUtil;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;


public class DefaultEClassHandler extends AbstractEClassHandler implements EClassHandler {
    public static final String DEFAULT_PERMISSION_FEATURE = "defaultPermission";
    public static final EClass OBJ_PERM_CLASS = SecurityPackage.eINSTANCE.getObjectPermission();

    private static final String[] DEFAULT_ACCEPTS = {
        "application/atom+xml",
    };
    
    protected RESTProvider restProvider;
    protected DefaultEFeatureHandler defaultEFeatureHandler;

    public DefaultEClassHandler(RESTProvider a) {
        restProvider = a;
        defaultEFeatureHandler = new DefaultEFeatureHandler(restProvider);
    }
    
    public String[] getAccepts() {
        return DEFAULT_ACCEPTS;
    }

    public void addObjectToResource(RESTContext ctx,
                                    EObject eobject,
                                    Resource resource) throws RESTException {

        EClass eclass = eobject.eClass();
        String qualName = RestEMFUtil.getQualifiedClassName(eclass);
        resource.setEClass(qualName);
        
        for (EStructuralFeature feature : eclass.getEAllStructuralFeatures()) {
            addFeatureToResource(feature, eobject, resource);
        }
        
        if (SecurityPackage.eINSTANCE.getIProtectedRepositoryObject().isSuperTypeOf(eclass)) {
            addDefaultPermissionReference(ctx, (IProtectedRepositoryObject)eobject, resource);
        }
    }

    public void getObjectChanges(RESTContext ctx,
                                 EObject eobject,
                                 Resource resource) throws RESTException {
        
        if (LOG.isLoggable(Level.FINE)) {
            LOG.fine("Input resource: " + resource);
        }
        
        EClass eclass = eobject.eClass();        
        for (EStructuralFeature feature : eclass.getEAllStructuralFeatures()) {
            getFeatureChanges(ctx, feature, eobject, resource);
        }   
    }

    public EObject createObject(RESTContext ctx,
                                EFactory efactory,
                                EClass eclass, 
                                String name,
                                Resource resource,
                                boolean persist) throws RESTException {
        
        EObject eobject = efactory.create(eclass);
        RestEMFUtil.setEObjectName(eobject, name);
        initRequiredAttributes(eobject);
        
        if (persist) {
            try {
                ctx.getRepo().persist(eobject);
            } catch (IOException e) {
                throw new RESTException(RESTException.INTERNAL_SERVER_ERROR, e);
            }
        }

        if (resource != null) {
            getObjectChanges(ctx, eobject, resource);        
        }
        
        return eobject;        
    }
    
    @Override
    public Collection<? extends EObject> createObjects(RESTContext ctx,
                                                       EFactory efactory,
                                                       EClass eclass,
                                                       Collection<Resource> resources,
                                                       boolean persist) throws RESTException {
        
        Collection<EObject> createdEObjs = new LinkedList<EObject>();
        if (resources != null && resources.size() > 0) {
            for (Resource rsc : resources) {
                Attribute nameAttr = rsc.getFirstAttribute("name");
                String name = (nameAttr != null && nameAttr.getText() != null) ? nameAttr.getText() : "";
                createdEObjs.add(createObject(ctx, efactory, eclass, name, rsc, persist));
            }
        }
        return createdEObjs;
    }

    public EObject createObjectWithMedia(RESTContext ctx,
                                         EFactory efactory,
                                         EClass eclass) throws RESTException {
        
        // Just return null and let the rest-adapter class raise an exception
        // since the message will be "nicer".
        return null;     
    }
    
    public void getMedia(RESTContext ctx, EObject eobject) throws RESTException {        
        // Just return null and let the rest-adapter class raise an exception
        // since the message will be "nicer".
        ctx.setInputStream(null);
    }


    public EObject updateMedia(RESTContext ctx, EObject eobject)
        throws RESTException {

        // Just return null and let the rest-adapter class raise an exception
        // since the message will be "nicer".
        return null;        
    }
    
    public void addFeatureToResource(EStructuralFeature feature,
                                     EObject eobject, 
                                     Resource resource) throws RESTException {

        String qualName = RestEMFUtil.getQualifiedName(feature);

        if (restProvider.isFeatureExcluded(qualName)) {            
            return;
        }
        

        EFeatureHandler handler = restProvider.getEFeatureHandler(qualName);
        if (handler == null) {
            handler = defaultEFeatureHandler;
        }

        handler.addFeatureToResource(qualName, feature, eobject, resource);

    }
    
    @SuppressWarnings("unchecked")
    public EList getContainedObjects(RESTContext ctx, EObject eobject,
                                     String refName) 
        throws RESTException {
        
        if (DEFAULT_PERMISSION_FEATURE.equals(refName)) {
            IProtectedRepositoryObject ipo = (IProtectedRepositoryObject) eobject;
            ctx.setCollectionTitle("Default Permissions " + RestEMFUtil.getResourceName(eobject));
            return getDefaultPermission(ctx, ipo);
        } 

        EStructuralFeature feature = eobject.eClass().getEStructuralFeature(refName);
        if (feature == null) {
            throw new RESTException(RESTException.BAD_REQUEST,
                    new Message(BUNDLE, "NO_SUBCOLLECTION", refName));
        }        
        
        if (!(feature instanceof EReference)) {
            throw new RESTException(RESTException.BAD_REQUEST,
                    new Message(BUNDLE, "SUBCOLLECTION_ERROR", refName));
        } else if (feature.isMany()) {
            return (EList)eobject.eGet(feature);
        } else {
            EList elist = new BasicEList();
            if (eobject.eIsSet(feature)) { 
                elist.add(eobject.eGet(feature));
            }
            return elist;
        }
    }

    public EClass getContainedClass(RESTContext ctx, EObject eobject,
                                    String refName) 
        throws RESTException {
                
        if (DEFAULT_PERMISSION_FEATURE.equals(refName)) {
            return OBJ_PERM_CLASS;
        } 

        EStructuralFeature feature = eobject.eClass().getEStructuralFeature(refName);        
        if (feature == null) {
            throw new RESTException(RESTException.BAD_REQUEST,
                    new Message(BUNDLE, "NO_SUBCOLLECTION", refName));
        }
        
        if (!(feature instanceof EReference)) {
            throw new RESTException(RESTException.BAD_REQUEST,
                    new Message(BUNDLE, "SUBCOLLECTION_ERROR", refName));            
        }        
        EReference ref = (EReference) feature;
        return ref.getEReferenceType();
    }

    @SuppressWarnings("unchecked")
    public EObject createContainedObject(RESTContext ctx,
                                         EObject eobject,
                                         String refName) throws RESTException {
        
        EStructuralFeature feature = eobject.eClass().getEStructuralFeature(refName);        
        if (feature == null) {
            throw new RESTException(RESTException.BAD_REQUEST,
                                    new Message(BUNDLE, "NO_SUBCOLLECTION", refName));
        }
        
        if (!(feature instanceof EReference)) {
            throw new RESTException(RESTException.BAD_REQUEST,
                                    new Message(BUNDLE, "SUBCOLLECTION_ERROR", refName));            
        }        
        EReference ref = (EReference) feature;
        EClass refClass = ref.getEReferenceType();
        
        CollectionDescriptor desc = restProvider.getDescriptor(refClass);
        refClass = desc.getEClass();
        
        EFactory efactory = desc.getEPackage().getEFactoryInstance();
        EObject newEObject = desc.getHandler().createObject(ctx,
                                                            efactory,
                                                            refClass,
                                                            ctx.getSlug(),
                                                            ctx.getResource(),
                                                            false);

        if (ref.isMany()) {
            EList<EObject> oldList = new BasicEList<EObject>();
            EList<EObject> newList = new BasicEList<EObject>();
            newList.add(newEObject);
            getReferenceListChanges(ctx, ref, eobject, newList, oldList);
        } else {
            eobject.eSet(ref, newEObject);
        }
        return newEObject;
                
    }
 
    protected void getFeatureChanges(RESTContext ctx,
                                     EStructuralFeature feature, 
                                     EObject eobject,
                                     Resource resource)
        throws RESTException {
        
        String qualName = RestEMFUtil.getQualifiedName(feature);
        
        if (restProvider.isFeatureExcluded(qualName) 
            || restProvider.isReadonly(qualName)) {
            return;
        }
        
        EFeatureHandler handler = restProvider.getEFeatureHandler(qualName);
        if (handler == null) {
            handler = defaultEFeatureHandler;
        }
        
        handler.getFeatureChanges(ctx, qualName, feature, eobject, resource);
    }
    
    
    @SuppressWarnings("unchecked")
    private void getReferenceListChanges(RESTContext ctx, EReference ref,
                                         EObject eobject, EList<EObject> newList,
                                         EList<EObject> oldList) 
        throws RESTException {
        
        String qualName = RestEMFUtil.getQualifiedName(ref);
                
        EFeatureHandler handler = restProvider.getEFeatureHandler(qualName);
        if (handler == null) {
            handler = defaultEFeatureHandler;
        }
        
        handler.getReferenceListChanges(ctx, ref, eobject, newList, oldList);
    }
    
    protected void initRequiredAttributes(EObject eobject) throws RESTException {        
        for (EAttribute att : eobject.eClass().getEAllAttributes()) {
            if (att.isRequired() && !eobject.eIsSet(att) && !att.isMany()) {
                eobject.eSet(att, att.getDefaultValue());
            }
        }
    }
    
    private void addDefaultPermissionReference(RESTContext ctx,
                                               EObject eobject, 
                                               Resource resource) throws RESTException {
        Reference reference = resource.addReference(DEFAULT_PERMISSION_FEATURE);
        reference.setMin(1);
        reference.setMax(1);
        reference.setSize(1);
        reference.setType(Constants.ATOM_FEED_CONTENT_TYPE);
        reference.setEClass(RestEMFUtil.getQualifiedClassName(OBJ_PERM_CLASS));
        
        String id = RestEMFUtil.getResourceId(eobject);
        reference.setHref(resource.getCollection() + "/" + id + ";" + DEFAULT_PERMISSION_FEATURE);
    }

    @SuppressWarnings("unchecked")
    private EList getDefaultPermission(RESTContext ctx, IProtectedRepositoryObject ipo) {
        EList elist = new BasicEList();
        Collection<EObject> roles = null;
        try {
            roles = ResourceSetQueryHandler.executeQuery(ctx.getRepo().getResourceSet(), 
                                                         SecurityPackage.eINSTANCE.getRole(),
                                                         true);
        } catch (Exception ex) {
            String msg = "Unable to retrieve Roles " + ex;
            LOG.warning(msg);
            return elist;
        }
        Collection<String> secAreas = SecurityUtil.getSecurityAreas(ipo, true);
        ObjectPermission op = SecurityFactory.eINSTANCE.createObjectPermission();
        op.setGuid("nonexist");
        
        for (EObject eObj : roles) {
            Role role = (Role)eObj;
            for (SecurityArea area : role.getCanRead()) {
                if (secAreas.contains(area.getAreaId())) {
                    op.getRolesRead().add(role);
                    break;
                }
            }
            for (SecurityArea area : role.getCanUpdate()) {
                if (secAreas.contains(area.getAreaId())) {
                    op.getRolesUpdate().add(role);
                    break;
                }
            }
            for (SecurityArea area : role.getCanDelete()) {
                if (secAreas.contains(area.getAreaId())) {
                    op.getRolesDelete().add(role);
                    break;
                }
            }
        }
        elist.add(op);
        return elist;
    }

    
}
