/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.rest.rt.repo;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.repository.maven.osgi.IRepositoryMavenService;
import com.iona.repository.rest.RESTContext;
import com.iona.repository.rest.RESTException;
import com.iona.repository.rest.RESTQuery;
import com.iona.repository.rest.resource.Resource;
import com.iona.soa.model.Artifact;
import com.iona.soa.model.ArtifactRepository;
import com.iona.soa.model.ArtifactType;
import com.iona.soa.model.IndexStates;
import com.iona.soa.model.MavenRepository;
import com.iona.soa.model.SOA_NetworkFactory;
import com.iona.soa.model.SOA_NetworkPackage;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.TransactionData;
import com.iona.soa.resourceset.QueryResourceSet;
import com.iona.soa.resourceset.ResourceSetQueryHandler;
import com.iona.soa.search.SearchService;
import com.iona.soa.security.SecurityContext;

import org.apache.lucene.document.Document;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;


public class MavenRepoQueryExecutor extends RESTQueryExecutor {

    public static final String OBJ_FIELD_GROUP = "namespace";
    public static final String OBJ_FIELD_ARTIFACT = "name";
    public static final String OBJ_FIELD_VERSION = "version";
    public static final String OBJ_FIELD_URL = "uri";
    //public static final String OBJ_FIELD_URL = "documentUri";
    private static final String DEFAULT_ARTIFACT_NAME = "<Unknown>";
    
    private static final Logger LOG = Logger.getLogger(MavenRepoQueryExecutor.class.getName());
    
    private IRepositoryMavenService mavenService; 
    private RepositoryRESTAdapter restProvider;
    
    public MavenRepoQueryExecutor(SOARepository soaRepo, 
                                  SearchService searchService, 
                                  IRepositoryMavenService mvnService,
                                  RepositoryRESTAdapter restProvider) {
        super(soaRepo, searchService);

        this.mavenService = mvnService;
        this.restProvider = restProvider;
    }
    
    @Override
    public Collection<Resource> queryResources(RESTContext ctx,
                                               TransactionData trnsData,
                                               RESTQuery query) throws RESTException {
        
        String searchQueryStr = query.getDecodedParameter(RESTQuery.QUERY_SEARCH);
        if (searchQueryStr == null) {
            return Collections.emptyList();
        }
        String reposFilterStr = query.getDecodedParameter(RESTQuery.QUERY_ACCPTED_REPOS);
        List<String> acceptedReposGuids = new ArrayList<String>();
        if (reposFilterStr != null && reposFilterStr.trim().length() > 0) {
            StringTokenizer tokenizer = new StringTokenizer(reposFilterStr, ",");
            while (tokenizer.hasMoreTokens()) {
                String tok = tokenizer.nextToken().trim();
                if (tok.length() > 0) {
                    acceptedReposGuids.add(tok);
                }
            }
        }
        
        Map<String, Resource> mavenResults = queryMavenArtifacts(searchQueryStr, acceptedReposGuids);
        
        queryDepotArtifacts(ctx, query, acceptedReposGuids, mavenResults);
        return mavenResults.values();
    }

    private Map<String, Resource> queryMavenArtifacts(String query,
                                                      List<String> acceptedReposGuids) throws RESTException {

        Map<String, Resource> resources = new HashMap<String, Resource>();
        try {
            
    
            Map<String, ArtifactType> artifactTypes = 
                getArtifactTypesKnownToDepot(getSOARepository().getResourceSet());

            List<EObject> mvnReposEobjs = 
                ResourceSetQueryHandler.executeQuery(getSOARepository().getResourceSet(), 
                                                     SOA_NetworkPackage.eINSTANCE.getMavenRepository(), 
                                                     true);
            
            for (EObject eobject : mvnReposEobjs) {
                MavenRepository mavenRepo = (MavenRepository)eobject;
                
                // apply filter if one is specified...
                if (acceptedReposGuids.size() == 0 || acceptedReposGuids.contains(mavenRepo.getGuid())) {
                
                    //getLogger().log(Level.FINE, "querying maven repos: " + mavenRepo.getName());
                    getLogger().log(Level.FINE, "querying maven repos: " + mavenRepo.getName());
                    if (mavenRepo.getIndexState().equals(IndexStates.READY)) {
                        getLogger().log(Level.FINE, "Querying being passed executed: " + query);
    
                        Iterator<Document> docs = mavenService.search(query, mavenRepo);
    
                        if (docs == null) {
                            return resources;
                        }
    
                        int count = 0;
                        while (docs.hasNext()) {
                            Document doc = docs.next();
                            count++;
                            
                            /*System.out.println("\n===========================\n"
                                               + doc
                                               + "\n===========================\n");*/
                            String groupId = doc.get("groupId") != null
                                                ? doc.get("groupId")
                                                : doc.get("parent groupId");
                            String version = doc.get("version") != null
                                                ? doc.get("version")
                                                : doc.get("parent version");
                                                
                            String packaging = doc.get("packaging") != null
                                                   ? doc.get("packaging")
                                                   : doc.get("parent packaging");
                            
                            Artifact modelArtifact = SOA_NetworkFactory.eINSTANCE.createArtifact();
                            modelArtifact.setName(doc.get("artifactId"));
                            modelArtifact.setNamespace(groupId);
                            String urlStr = doc.get("url");
                            modelArtifact.setUri(new URI(urlStr));
                            modelArtifact.setFileName(urlStr.substring(urlStr.lastIndexOf('/') + 1));
                            modelArtifact.setVersion(version);

                            guessArtifactType(modelArtifact, packaging, artifactTypes);
                            
                            modelArtifact.setModifiable(false);
                            modelArtifact.setGuid(null);
                            modelArtifact.getFoundIn().add(mavenRepo);

                            Resource resource = constructResource(null,
                                                                  modelArtifact,
                                                                  resources,
                                                                  null);
                            resource.setEditable(false);
                            resource.setDeletable(false);
                            getLogger().log(Level.FINE, "rsc ready: " + resource.getId());
    
                        }
                        getLogger().log(Level.FINE, "Querying hits returned: {0}", count);
                        
                    }
                }
            }
            return resources;
            
        } catch (Exception e) {
            e.printStackTrace();
            getLogger().log(Level.SEVERE, "Error querying lucene index ", e);
            // Just return an empty collection instead of raising an exception.
            return resources;
        }
    }

    private void queryDepotArtifacts(RESTContext ctx,
                                     RESTQuery query,
                                     List<String> acceptedReposGuids,
                                     Map<String, Resource> mavenResults) throws RESTException {
        
        EClass artifactEClass = SOA_NetworkPackage.eINSTANCE.getArtifact();
        List<EObject> eobjs = queryObjects(ctx, query, artifactEClass);
        
        SecurityContext securityCtx = RepositoryRESTAdapter.getSecurityContext(ctx);
        
        for (EObject eobj : eobjs) {
            if (!artifactEClass.equals(eobj.eClass()))  {
                continue;
            }

            Artifact artifactObj = (Artifact)eobj;
            if (acceptedReposGuids.size() == 0
                || (artifactObj.getFoundIn().size() > 0
                    && acceptedReposGuids.contains(artifactObj.getFoundIn().get(0).getGuid()))) {

                if (restProvider.isReadAuthorized(securityCtx, eobj)) {
                    constructResource(ctx, eobj, mavenResults, securityCtx);
                    
                } else if (eobj instanceof Artifact) {
                    mavenResults.remove(constructRscKey(artifactObj));
                }
            }
        }
    }

    private Resource constructResource(RESTContext ctx,
                                       EObject eobj,
                                       Map<String, Resource> mavenResults,
                                       SecurityContext securityCtx) throws RESTException {
        
        Resource resource = restProvider.doGetResource(ctx,
                                                       securityCtx,
                                                       restProvider.getDescriptor("artifacts"),
                                                       eobj);
        if (resource != null) {
            Artifact artifactObj = (Artifact)eobj;
            if (artifactObj.getGuid() == null) {
                resource.setId(constructRscKey(artifactObj));
            }
            mavenResults.put(constructRscKey(artifactObj), resource);
        }
        return resource;

    }
    
    static String constructRscKey(Artifact artifactEObj) {
        
        String extString = artifactEObj.getFileName();
        if (extString == null)     {
            URI uri = artifactEObj.getUri();
            if (uri != null) {
                extString = uri.getPath();                
            }
        }

        return constructRscKey(artifactEObj.getName(),
                               artifactEObj.getNamespace(),
                               artifactEObj.getVersion(),
                               getExtension(extString),
                               null);
    }


    static String constructRscKey(String name,
                                  String group,
                                  String version,
                                  String ext,
                                  ArtifactRepository foundInObj) {
        StringBuilder sb = new StringBuilder();
        if (name == null) {
            sb.append(DEFAULT_ARTIFACT_NAME);
        } else {
            sb.append(name);
        }
        if (group != null) {
            sb.append("-").append(group);
        }
        if (version != null) {
            sb.append("-").append(version);
        }
        if (ext != null) {
            sb.append(".").append(ext);
        }

        try {
            return URLEncoder.encode(sb.toString(), "UTF-8");
        } catch (UnsupportedEncodingException e) {
            LOG.log(Level.SEVERE, "problem while url encoding", e);
            return sb.toString().replaceAll(" ", "_");            
        }
    }
    
    static Map<String, ArtifactType> getArtifactTypesKnownToDepot(QueryResourceSet  rs) throws Exception,
                                                                                               IOException {
        
        List<EObject> mvnArtifactTypesEobjs = 
            ResourceSetQueryHandler.executeQuery(rs, 
                                                 SOA_NetworkPackage.eINSTANCE.getArtifactType(), 
                                                 true);
        //System.out.println("+++++++++++++ model artifact types: " + mvnArtifactTypesEobjs.size());
        Map<String, ArtifactType> artifactTypes =
            new HashMap<String, ArtifactType>(mvnArtifactTypesEobjs.size());
        for (EObject eobject : mvnArtifactTypesEobjs) {
            ArtifactType type = (ArtifactType)eobject;
            //System.out.println("+++++++++++++ model artifact type: " + type.getName());
            artifactTypes.put(type.getName(), type);
        }
        return artifactTypes;
    }
    
    static ArtifactType guessArtifactType(Artifact modelArtifact,
                                          String packaging,
                                          Map<String, ArtifactType> artifactTypes) {
        if (artifactTypes != null
                && artifactTypes.size() > 0
                && packaging != null
                && packaging.length() > 0) {
            
            packaging = packaging.toLowerCase().trim();
            for (ArtifactType at  : artifactTypes.values()) {
                for (String marker : at.getMarkers()) {
                    if (marker.toLowerCase().trim().equals(packaging)) {
                        modelArtifact.setType(at);
                        return at;
                    }
                }
            }
        }
        
        return null;
    }
    
    
    private static String getExtension(String path) {
        if (path == null) {
            return null;
        } 
        int i = path.lastIndexOf(".");
        if (i == -1) {
            return null;
        }
        
        return path.substring(i + 1);        
    }


}
