/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.rest.rt.repo;

import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.ResourceBundle;
import java.util.Set;
import java.util.logging.Logger;

import com.iona.repository.rest.RESTContext;
import com.iona.repository.rest.RESTException;
import com.iona.repository.rest.RESTQuery;
import com.iona.repository.rest.resource.Attribute;
import com.iona.repository.rest.resource.Resource;
import com.iona.repository.rest.resource.ResourceFactory;
import com.iona.repository.rest.resource.ResourceFactoryImpl;
import com.iona.repository.rest.rt.osgi.MessageBundle;
import com.iona.repository.util.i18n.Message;
import com.iona.soa.expression.Expression;
import com.iona.soa.expression.Term;
import com.iona.soa.expression.serialization.ExpressionSerializer;
import com.iona.soa.model.SOA_NetworkPackage;
import com.iona.soa.model.StoredQuery;
import com.iona.soa.repository.ModelRegistry;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.TransactionData;
import com.iona.soa.repository.jdbc.EMFSQLUtil;
import com.iona.soa.repository.jdbc.resource.JDBCResource;
import com.iona.soa.repository.status.RepositoryException;
import com.iona.soa.resourceset.QueryResourceSet;
import com.iona.soa.resourceset.ResourceSetQueryHandler;
import com.iona.soa.search.SearchService;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;

public class RESTQueryExecutor {

    private static final ResourceBundle BUNDLE = MessageBundle.getInstance();
    private static final Logger LOG = Logger.getLogger(RESTQueryExecutor.class.getName());
    private static final ResourceFactory RESOURCE_FACTORY = new ResourceFactoryImpl();

    private SOARepository soaRepo;
    private SearchService searchService;

    private Collection<EClass> eclasses;
    private Set<EClass> querableClasses;


    public RESTQueryExecutor(SOARepository soaRepo, 
                             SearchService searchService) {
        this.soaRepo = soaRepo;
        this.searchService = searchService;
        

    }
    
    public Logger getLogger() {
        return LOG;
    }
    
    public void setEclasses(Collection<EClass> inEclasses) {
        this.eclasses = inEclasses;
    }
    
    public SOARepository getSOARepository() {
        return soaRepo;
    }
    
    ResourceFactory getResourceFactory() {
        return RESOURCE_FACTORY;
    }
    
    
    public List<EObject> queryObjects(RESTContext ctx, RESTQuery query, EClass eclass) throws RESTException {
        
        String queryStr = query.getDecodedParameter(RESTQuery.QUERY_EMFQL);                
        if (queryStr != null) {
            return queryEmfql(ctx, eclass, queryStr);
        }
        
        queryStr = query.getDecodedParameter(RESTQuery.QUERY_SQL);
        if (queryStr != null) {
            return querySql(ctx, eclass, queryStr);
        }

        queryStr = query.getDecodedParameter(RESTQuery.QUERY_SEARCH);                
        if (queryStr != null) {
            return textSearch(ctx, queryStr);
        }
        
        queryStr = query.getParameter(RESTQuery.QUERY_ROOT);
        if (queryStr != null) {
            return rootContents(ctx, queryStr);
        }

        if (eclass == null) {
            LOG.warning("no eclass supplied; returning no objects");
            return new ArrayList<EObject>();            
        }
        try {
            QueryResourceSet rs = soaRepo.getResourceSet();
            List<EObject> eobjs = 
                ResourceSetQueryHandler.executeQuery(rs, eclass, true);
            return eobjs;
        } catch (Exception e) {
            LOG.warning("Exception " + e);
            throw new RESTException(RESTException.INTERNAL_SERVER_ERROR, e);
        }
    }
    


    public Collection<Resource> queryResources(RESTContext ctx,
                                               TransactionData trnsData,
                                               RESTQuery query) throws RESTException {

        return new ArrayList<Resource>();
    }
    
    @SuppressWarnings("unchecked")
    public EList runStoredQuery(RESTContext ctx, EClass eclass, StoredQuery query)
        throws RESTException {
        
        // TODO: better if this returned List<EObject>.
        
        String type = query.getType();
        if (RESTQuery.QUERY_EMFQL.equals(type)) {
            List<EObject> elist =  queryEmfql(ctx, eclass, query.getQuery());
            return new BasicEList(elist);
            
        } else if (RESTQuery.QUERY_SQL.equals(type)) {
            List<EObject> elist = querySql(ctx, eclass, query.getQuery());
            return new BasicEList(elist);

        } else if (RESTQuery.QUERY_SEARCH.equals(type)) {
            List<EObject> elist = textSearch(ctx, query.getQuery());
            return new BasicEList(elist);
        }
        
        LOG.warning("Unknown query type \"" + type + "\"");
        return new BasicEList();
    }

    private List<EObject> textSearch(RESTContext ctx, String query) throws RESTException {
        
        LOG.fine("Search query " + query);
        if (searchService == null) {
            LOG.warning("No search service present; returning empty collection");
            return new ArrayList<EObject>();
        }

        ctx.setCollectionTitle("Search: " + query);
        
        try {
            List<EObject> result = searchService.search(query);
            if (result == null) {
                result = Collections.emptyList();
            }
            return result;
        } catch (RepositoryException e) {
            throw new RESTException(500, e);
        }
    }
    
    
    private List<EObject> queryEmfql(RESTContext ctx, EClass eclass, String query) 
        throws RESTException {
        
        List<EObject> elist = new ArrayList<EObject>();
        
        Expression expr = getExpression(query);
        
        Collection<EClass> queryClasses = null;
        if (eclass != null) {
            queryClasses = Collections.singleton(eclass);
        } else {
            queryClasses = getAllQuerableClasses();
        }
        
        for (EClass ec : queryClasses) {
            try {
                QueryResourceSet rs = soaRepo.getResourceSet();
                List<EObject> eobjs = 
                    ResourceSetQueryHandler.executeQuery(rs, ec, expr, true);
                LOG.fine("Query on " + ec.getName() + ": " + eobjs.size());
                elist.addAll(eobjs);
            
            } catch (SQLException e) {
                // Most likely a result of a result of a mismatch between the query
                // and the eclass.  Just ignore.
            } catch (Exception e) {
                throw new RESTException(RESTException.INTERNAL_SERVER_ERROR, e);
            }
        }
        return elist;
    }

    private List<EObject> querySql(RESTContext ctx, EClass eclass, String query) 
        throws RESTException {

        List<EObject> elist = new ArrayList<EObject>();

        Collection<EClass> queryClasses = null;
        if (eclass != null) {
            queryClasses = Collections.singleton(eclass);
        } else {
            // TODO: This can cause problems with both the query and trying to load
            // objects into a different class.  May need to disable this type of query.
            queryClasses = getAllQuerableClasses();
        }
        
        JDBCResource jdbcResource = getJdbcResource();
        
        
        for (EClass ec : queryClasses) {
            try {
                List<EObject> eobjs = jdbcResource.queryObjects(ec, query);
                LOG.fine("Query on " + ec.getName() + ": " + eobjs.size());
                elist.addAll(eobjs);

            } catch (SQLException e) {
                // Most likely a result of a result of a mismatch between the query
                // and the eclass.  Just ignore.
                LOG.fine("Ignoring SQL exception " + e);
            } catch (Exception e) {
                throw new RESTException(RESTException.INTERNAL_SERVER_ERROR, e);
            }
        }
        return elist;
    }

    
    private List<EObject> rootContents(RESTContext ctx, String queryStr) 
        throws RESTException {
        
        // Get contents of root folder (currently this cannot be handled by
        // a simple EMFQL or SQL query).
        
        EClass folderClass = SOA_NetworkPackage.eINSTANCE.getFolder();
        List<EObject> rootFolders = queryEmfql(ctx, folderClass, "isnull:parent");
        LOG.fine("Root folders: " + rootFolders.size());
        
        List<EObject> contents = queryEmfql(ctx, null, "isnull:belongsTo");
        LOG.fine("Root contents: " + contents.size());
        
        contents.addAll(rootFolders);
        return contents;
    }
    private JDBCResource getJdbcResource() {
        
        // This assumes there is only one JDBCResource in the resource set.
        try {
            for (Object resource : soaRepo.getResourceSet().getResources()) {
                if (resource instanceof JDBCResource) {
                    return (JDBCResource) resource;
                }
            }
        } catch (IOException e) {
            LOG.warning("Exception getting JDBC Resource; " + e);
        }
        return null;
    }

    private synchronized Collection<EClass> getAllQuerableClasses() {
        if (querableClasses == null) {

            querableClasses = new HashSet<EClass>();

            ModelRegistry reg = soaRepo.getModelRegistry();
            for (EClass eclass : eclasses) {
                querableClasses.addAll(EMFSQLUtil.getTableClasses(eclass, reg));        
            }
        }        
        return querableClasses;
    }
    
    private Expression getExpression(String query) throws RESTException {

        try {
            Term term = ExpressionSerializer.deserialize(query);
            if (!(term instanceof Expression)) {
                LOG.info("EMFQL Term: " + (term == null ? "<null>" : term.getClass()));
                throw new RESTException(RESTException.BAD_REQUEST,
                                        new Message(BUNDLE, "EMFQL_BAD_TERM", query));
            }
            return (Expression) term;
        
        } catch (Exception e) {
            LOG.fine("Error parsing EMFQL query: " + e);
            throw new RESTException(RESTException.BAD_REQUEST,
                                    new Message(BUNDLE, "EMFQL_PARSE_ERROR", query));            
        }
    }

    protected void addAttribute(Resource resource, String name, String value, String type) {
        if (value != null) {
            Attribute newAttrib = RESOURCE_FACTORY.createAttribute(name);
            newAttrib.setText(value);
            newAttrib.setType(type);
            resource.addAttribute(newAttrib);
        }
    }


}
