/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.rest.rt.repo;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.repository.maven.osgi.IRepositoryMavenService;
import com.iona.repository.rest.RESTContext;
import com.iona.repository.rest.RESTException;
import com.iona.repository.rest.RESTQuery;
import com.iona.repository.rest.osgi.RESTCollectionService;
import com.iona.repository.rest.repo.EClassHandler;
import com.iona.repository.rest.resource.Resource;
import com.iona.repository.rest.resource.ResourceFactory;
import com.iona.repository.rest.resource.ResourceFactoryImpl;
import com.iona.repository.rest.rt.osgi.MessageBundle;
import com.iona.repository.rest.rt.osgi.RestServiceBean;
import com.iona.repository.rest.rt.server.RESTAdapter;
import com.iona.repository.rest.rt.server.RESTProvider;
import com.iona.repository.util.i18n.Message;
import com.iona.soa.model.SOA_NetworkPackage;
import com.iona.soa.model.collaboration.CollaborationPackage;
import com.iona.soa.model.folder.FolderPackage;
import com.iona.soa.model.repository.RepositoryPackage;
import com.iona.soa.model.security.SecurityPackage;
import com.iona.soa.model.userattributes.UserAttributesPackage;
import com.iona.soa.repository.EMFUtil;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.TransactionData;
import com.iona.soa.repository.configuration.ConfigurationKeys;
import com.iona.soa.search.SearchService;
import com.iona.soa.security.AuthorizationException;
import com.iona.soa.security.SecurityContext;
import com.iona.soa.security.SecurityService;
import com.iona.soa.security.SecurityUtil;

import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.notify.impl.AdapterImpl;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;

public class  RepositoryRESTAdapter implements RESTAdapter, RESTProvider {

    private static final ResourceBundle BUNDLE = MessageBundle.getInstance();

    private static final Logger LOG = Logger.getLogger(RepositoryRESTAdapter.class.getName());

    private static final String UNKNOWN_USER = "unknown";

    private static final Collection<String> DOWNLOAD_AREAS = 
        Collections.singletonList(SecurityUtil.DOWNLOAD_SECAREA_ID);


    private Map<String, CollectionDescriptor> collections;
    private Map<String, String> eClassPaths;
    private Map<String, EClass> collectionEClasses;

    private Set<String> excludedFeatures;   
    private Set<String> hiddenFeatures = new HashSet<String>();

    private Set<String> readonlyFeatures;
    private Map<String, String> xsdTypeMap;
    private Map<String, EFeatureHandler> featureHandlers;


    private SOARepository soaRepo;
    private SecurityService securityService;

    private ResourceFactory resourceFactory = new ResourceFactoryImpl();

    private RESTQueryExecutor defaultQueryExecutor;

    
    public RepositoryRESTAdapter(SOARepository soaRepo, 
                                 SecurityService securityService,
                                 SearchService searchService, 
                                 Map<String, String> config,
                                 IRepositoryMavenService mavenService,
                                 RestServiceBean restServiceBean) {
        this.soaRepo = soaRepo;
        this.securityService = securityService;
        collections = new HashMap<String, CollectionDescriptor>();
        eClassPaths = new HashMap<String, String>();
        collectionEClasses = new HashMap<String, EClass>();
        defaultQueryExecutor = new RESTQueryExecutor(soaRepo, 
                searchService);
        
        EClassHandler defaultEClassHandler = new DefaultEClassHandler(this);
        EClassHandler searchHandler = new SearchHandler(this);
        String spiderDirectory = config.get(ConfigurationKeys.SPIDER_STORAGE_DIR);
        
        addCollaborationCollections(defaultEClassHandler); 
        addRepositoryCollections(defaultEClassHandler);
        addUserAttributeCollections(defaultEClassHandler);
        addSecurityCollections(defaultEClassHandler);
        addSchedulingCollections(searchService, restServiceBean, defaultEClassHandler);
        addSOANCollections(defaultEClassHandler);
        addSearchCollections(searchHandler);
        
        addMavenResourceCollections(soaRepo, searchService, mavenService, searchHandler, spiderDirectory);
        
        makeReferenceable(FolderPackage.eINSTANCE.getIFolderContent(), null);
        makeReferenceable(FolderPackage.eINSTANCE.getIFolder(), "folders");
        makeReferenceable(FolderPackage.eINSTANCE.getICategorizable(), null);
        makeReferenceable(FolderPackage.eINSTANCE.getICategory(), "categories");


        hiddenFeatures.add("security.User.password");
        //hiddenFeatures.add("scheduling.ScheduledTask.token");
        hiddenFeatures.add("scheduling.ScheduledTask.password");
        
        markReadOnlyFeatures();
        
        initXSDTypeMap();         
        initializeExclusions();        
                       
        featureHandlers = new HashMap<String, EFeatureHandler>();
        featureHandlers.put("repository.IVersionedObject.version", new VersionHandler(this));
        featureHandlers.put("userattributes.ITagged.tags", new TagsHandler(this));
        featureHandlers.put("repository.IPersistableObject.timestamp", new TimestampHandler(this));
        featureHandlers.put("security.IProtectedRepositoryObject.objectControl", 
                            new ObjectControlHandler(this));
        featureHandlers.put("security.User.password", new PasswordHandler(this, true));
        featureHandlers.put("security.User.userid", new UserIdHandler(this));
        featureHandlers.put("scheduling.ScheduledTask.password", new PasswordHandler(this, false));

        List<String> collectionList = getCollections();
        for (String collection : collectionList) {
            try {
                getDescriptor(collection).getQueryExecutor().setEclasses(collectionEClasses.values());
            } catch (RESTException e) {
                e.printStackTrace();
            }
        }
    }

    private void markReadOnlyFeatures() {
        readonlyFeatures = new HashSet<String>();
        
        readonlyFeatures.add("repository.IPersistableObject.timestamp");
        readonlyFeatures.add("repository.IPersistableObject.lastEditedBy");
        readonlyFeatures.add("repository.IPersistableObject.guid");
        readonlyFeatures.add("repository.IPersistableObject.createdBy");
        readonlyFeatures.add("repository.IPersistableObject.generationNumber");
        
        readonlyFeatures.add("model.Artifact.external");
        
        readonlyFeatures.add("model.MavenRepository.lastIndexed");
        readonlyFeatures.add("model.MavenRepository.nextIndex");
        readonlyFeatures.add("model.MavenRepository.indexState");
        
        readonlyFeatures.add("scheduling.ScheduledTask.lastRun");
        readonlyFeatures.add("scheduling.ScheduledTask.duration");
        readonlyFeatures.add("scheduling.ScheduledTask.state");
    }
    
    private void initXSDTypeMap() {
        xsdTypeMap = new HashMap<String, String>();
        xsdTypeMap.put("EString", "xsd:normalizedString");
        xsdTypeMap.put("Timestamp", "xsd:dateTime");      
        xsdTypeMap.put("URI", "xsd:anyURI");
        xsdTypeMap.put("ELong", "xsd:int");
        xsdTypeMap.put("EInt", "xsd:int");
        xsdTypeMap.put("EIntegerObject", "xsd:integer");
        xsdTypeMap.put("EBoolean", "xsd:boolean");
        xsdTypeMap.put("EShort", "xsd:short");
    }

    private void addMavenResourceCollections(SOARepository soaRepository,
                                             SearchService searchService,
                                             IRepositoryMavenService mavenService,
                                             EClassHandler searchHandler,
                                             String spiderDirectory) {
        
        MavenRepoQueryExecutor mavenQueryExecutor = new MavenRepoQueryExecutor(soaRepository, 
                                                                               searchService,
                                                                               mavenService, 
                                                                               this);
        
        ArtifactResolveExecutor artifactResolverExecutor = new ArtifactResolveExecutor(soaRepository, 
                                                                                       searchService, 
                                                                                       mavenService,
                                                                                       this);
        
        SpiderActionExecutor spiderActionExecutor = new SpiderActionExecutor(soaRepository, 
                                                                             searchService, 
                                                                             mavenService);

        addResourceCollection("mavenreposearch", searchHandler, mavenQueryExecutor);
        addResourceCollection("artifactResolve", searchHandler, artifactResolverExecutor);
        addResourceCollection("spiderAction", searchHandler, spiderActionExecutor);
    }

    private void addResourceCollection(String name,
                                       EClassHandler searchHandler,
                                       RESTQueryExecutor queryExecutor) {

        CollectionDescriptor cd = new CollectionDescriptor(name).setEPackage(null)
                                                                .setEClass(null)
                                                                .setHandler(searchHandler)
                                                                .setIsResourceBased(true)
                                                                .setQueryExecutor(queryExecutor);
        
        collections.put(name, cd);
        
        if (cd.getEClass() != null) {
            String className = RestEMFUtil.getQualifiedClassName(cd.getEClass());
            eClassPaths.put(className, name);
            collectionEClasses.put(name, cd.getEClass());
        }     
        
    }

    private void addRepositoryCollections(EClassHandler defaultEClassHandler) {
        addCollection("properties",
                RepositoryPackage.eINSTANCE,
                RepositoryPackage.eINSTANCE.getProperty(),
                new PropertyHandler(this));
        addCollection("auditevents",
                      RepositoryPackage.eINSTANCE,
                      RepositoryPackage.eINSTANCE.getAuditEvent(),
                      defaultEClassHandler,
                      false, false);
    }

    private void addCollaborationCollections(EClassHandler defaultEClassHandler) {
        addCollection("comments", 
                CollaborationPackage.eINSTANCE, 
                CollaborationPackage.eINSTANCE.getComment(), 
                defaultEClassHandler);
    }

    private void addUserAttributeCollections(EClassHandler defaultEClassHandler) {
        addCollection("tags",
                UserAttributesPackage.eINSTANCE,
                UserAttributesPackage.eINSTANCE.getUserAttributeDefinition(),
                defaultEClassHandler);
        addCollection("useratts",
                UserAttributesPackage.eINSTANCE,
                UserAttributesPackage.eINSTANCE.getUserAttributeValue(),
                defaultEClassHandler);
    }

    private void addSearchCollections(EClassHandler searchHandler) {
        addCollection("search", null, null, searchHandler);
    }

    private void addSecurityCollections(EClassHandler defaultEClassHandler) {
        addCollection("users",
                      SecurityPackage.eINSTANCE,
                      SecurityPackage.eINSTANCE.getUser(),
                      new UserHandler(this));
        addCollection("roles", 
                      SecurityPackage.eINSTANCE, 
                      SecurityPackage.eINSTANCE.getRole(), 
                      new RoleHandler(this));
        addCollection("securityareas",
                      SecurityPackage.eINSTANCE,
                      SecurityPackage.eINSTANCE.getSecurityArea(),
                      defaultEClassHandler,
                      false, false);
        addCollection("objectpermissions",
                      SecurityPackage.eINSTANCE,
                      SecurityPackage.eINSTANCE.getObjectPermission(),
                      new ObjectPermissionHandler(this));
    }
    
    private void addSchedulingCollections(SearchService searchService, 
                                          RestServiceBean restServiceBean,
                                          EClassHandler defaultEClassHandler) {
        
        TaskServiceQueryExecutor taskServiceQueryExecutor = new TaskServiceQueryExecutor(soaRepo, 
                                                                                         searchService,
                                                                                         restServiceBean);
        addResourceCollection("taskservices", defaultEClassHandler, taskServiceQueryExecutor);
    }    

    private void addSOANCollections(EClassHandler defaultEClassHandler) {
        addCollection("artifacts",
                SOA_NetworkPackage.eINSTANCE,
                SOA_NetworkPackage.eINSTANCE.getArtifact(),
                new ArtifactsHandler(this),
                true, true);
        addCollection("artifacttypes",
                SOA_NetworkPackage.eINSTANCE,
                SOA_NetworkPackage.eINSTANCE.getArtifactType(),
                defaultEClassHandler);
        addCollection("applicationtypes",
                SOA_NetworkPackage.eINSTANCE,
                SOA_NetworkPackage.eINSTANCE.getApplicationType(),
                defaultEClassHandler);
        addCollection("folders", 
                SOA_NetworkPackage.eINSTANCE, 
                SOA_NetworkPackage.eINSTANCE.getFolder(), 
                new FolderHandler(this));
        addCollection("categories", 
                SOA_NetworkPackage.eINSTANCE, 
                SOA_NetworkPackage.eINSTANCE.getCategory(), 
                new CategoryHandler(this));
        addCollection("applications",
                SOA_NetworkPackage.eINSTANCE,
                SOA_NetworkPackage.eINSTANCE.getApplication(),
                defaultEClassHandler);
        addCollection("profiles",
                SOA_NetworkPackage.eINSTANCE,
                SOA_NetworkPackage.eINSTANCE.getDeploymentProfile(),
                new DeploymentProfileHandler(this));
        addCollection("versionedprofiles",
                SOA_NetworkPackage.eINSTANCE,
                SOA_NetworkPackage.eINSTANCE.getVersionedProfile(),
                new VersionedProfileHandler(this));
        addCollection("queries",
                SOA_NetworkPackage.eINSTANCE,
                SOA_NetworkPackage.eINSTANCE.getStoredQuery(),
                new StoredQueryHandler(this));
        addCollection("containers",
                SOA_NetworkPackage.eINSTANCE,
                SOA_NetworkPackage.eINSTANCE.getContainer(),
                new ContainerHandler(this));  
        addCollection("artifactrepositories",
                SOA_NetworkPackage.eINSTANCE,
                SOA_NetworkPackage.eINSTANCE.getArtifactRepository(),
                defaultEClassHandler,
                true, true);
        addCollection("mavenrepositories",
                SOA_NetworkPackage.eINSTANCE,
                SOA_NetworkPackage.eINSTANCE.getMavenRepository(),
                defaultEClassHandler);
        addCollection("installedartifacts",
                SOA_NetworkPackage.eINSTANCE,
                SOA_NetworkPackage.eINSTANCE.getInstalledArtifact(),
                defaultEClassHandler);
        addCollection("installedapplications",
                SOA_NetworkPackage.eINSTANCE,
                SOA_NetworkPackage.eINSTANCE.getInstalledApplication(),
                defaultEClassHandler);
        
        addCollection("configurationOverrides",
                SOA_NetworkPackage.eINSTANCE,
                SOA_NetworkPackage.eINSTANCE.getConfigurationOverride(),
                defaultEClassHandler);
    }

    private void initializeExclusions() {
        excludedFeatures = new HashSet<String>();
        excludedFeatures.add("repository.IPersistableObject.generationNumber");
        excludedFeatures.add("repository.IPersistableObject.internal");
        excludedFeatures.add("repository.IPersistableObject.modifiable");
        excludedFeatures.add("repository.Version.tracker");
        excludedFeatures.add("model.Container.adminPolicies");
        excludedFeatures.add("model.Container.builtInServices");
        excludedFeatures.add("model.Container.deployedParticipants");
        excludedFeatures.add("scheduling.ScheduledTask.affectedObjects");
        excludedFeatures.add("scheduling.ScheduledTask.token");
    }
    
    public List<String> getCollections() {
        return new ArrayList<String>(collections.keySet());
    }
    
    public String[] getCollectionAccepts(String collection) throws RESTException {
        CollectionDescriptor desc = getDescriptor(collection);
        return desc.getHandler().getAccepts();         
    }
    
    public void getCollection(RESTContext ctx, String collection, RESTQuery query) throws RESTException {
        
        CollectionDescriptor desc = getDescriptor(collection);
        
        List<Resource> resources = new ArrayList<Resource>();

        SecurityContext securityCtx = getSecurityContext(ctx);
        soaRepo.startRead(getTransactionData(securityCtx));
        
        try {

            if (desc.isResourceBased()) {
                TransactionData trnsData = getTransactionData(securityCtx);
                resources.addAll(desc.getQueryExecutor().queryResources(ctx, trnsData, query));  
            
            } else {
                ctx.setEditable(isCreateAuthorized(securityCtx, desc.getEClass(), null));
                List<EObject> eobjs = desc.getQueryExecutor().queryObjects(ctx, query, desc.getEClass());
                
                handleErrorsAndWarnings();
                
                List<EObject> filteredEObjs = applyEclassFilter(ctx, eobjs);
                
                for (EObject eobj : filteredEObjs) {
                    if (isReadAuthorized(securityCtx, eobj)) {
                        Resource resource = 
                            doGetResource(ctx, securityCtx, desc, eobj, desc.isFilterEnabled());
                        if (resource != null) {
                            resources.add(resource);
                        }
                    }
                }
            }
            if (ctx.getCollectionTitle() == null) {
                ctx.setCollectionTitle(desc.getCollectionTitle());
            }
        } catch (Exception e) {
            throwException(e);
            
        } finally {
            soaRepo.endRead(true);
        }
        
        ctx.setResourceList(resources);
    }

    @SuppressWarnings("unchecked")
    public void getSubCollection(RESTContext ctx,
                                 String collection,
                                 String id,
                                 String subcollection) throws RESTException {
        
        CollectionDescriptor desc = getDescriptor(collection);
        ctx.setRepo(soaRepo);

        List<Resource> resources = new ArrayList<Resource>();

        SecurityContext securityCtx = getSecurityContext(ctx);
        soaRepo.startRead(getTransactionData(securityCtx));
        
        try {

            EObject containingEObject = getEObject(desc.getEClass(), collection, id);
            ctx.setEditable(isCreateAuthorized(securityCtx, desc.getEClass(), containingEObject));
            

            EList eobjs = desc.getHandler().getContainedObjects(ctx, containingEObject, subcollection);
            
            List<EObject> filteredEObjs = applyEclassFilter(ctx, eobjs);
            
            for (Object obj : filteredEObjs) {                
                EObject eobject = (EObject) obj;                
                if (isReadAuthorized(securityCtx, eobject)) {
                    Resource resource = doGetResource(ctx, securityCtx, null, eobject);
                    if (resource != null) {
                        resources.add(resource);
                    }
                }
            }
            
            if (ctx.getCollectionTitle() == null) {
                String name = RestEMFUtil.getResourceName(containingEObject);
                ctx.setCollectionTitle(desc.getCollectionTitle() + ": " + name);
            }
        } catch (Exception e) {
            throwException(e);
            
        } finally {
            soaRepo.endRead(true);
        }
        
        ctx.setResourceList(resources);
    }

    @SuppressWarnings("unchecked")
    static List<EObject> applyEclassFilter(RESTContext ctx, List eobjs) {
        
        if (ctx.getRestQuery() == null 
                || !ctx.getRestQuery().isSet(RESTQuery.QUERY_ECLASS_FILTER)) {
            return eobjs;
        }
        
        String commaDelimitedEClasses = ctx.getRestQuery().getParameter(RESTQuery.QUERY_ECLASS_FILTER);
        if (commaDelimitedEClasses == null || commaDelimitedEClasses.trim().length() == 0) {
            return eobjs;
        }
        
        StringTokenizer tokenizer = new StringTokenizer(commaDelimitedEClasses, ",");
        List<String> acceptedECLasses = new ArrayList<String>();
        while (tokenizer.hasMoreElements()) {
            acceptedECLasses.add(tokenizer.nextToken());
        }
        
        List<EObject> acceptedEObjs = new ArrayList<EObject>();
        
        for (Object obj : eobjs) {                
            EObject eobject = (EObject) obj;  
            if (acceptedECLasses.contains(RestEMFUtil.getQualifiedClassName(eobject.eClass()))) {
                acceptedEObjs.add(eobject);
            }
        }
        
        return acceptedEObjs;
        
    }

    public void getResource(RESTContext ctx, String collection, String id) throws RESTException {
        
        CollectionDescriptor desc = getDescriptor(collection);
        SecurityContext securityCtx = getSecurityContext(ctx);

        if (Resource.PROTOTYPE_RESOURCE_ID.equals(id)) {
            // If we decide do any reading to say, default a name value 
            // then we'll need to use a transaction.
            EClass eclass = desc.getEClass();
            EFactory efactory = desc.getEPackage().getEFactoryInstance();
            EObject eobject = desc.getHandler().createObject(ctx,
                                                             efactory,
                                                             eclass,
                                                             ctx.getSlug(),
                                                             ctx.getResource(),
                                                             false);
            Resource r = doGetResource(ctx, securityCtx, desc, eobject);
            r.markAsPrototype();
            ctx.setResource(r);
        } else {
            soaRepo.startRead(getTransactionData(securityCtx));
            
            try {
                EObject eobject = getEObject(desc.getEClass(), collection, id);
                
                if (!isReadAuthorized(securityCtx, eobject)) { 
                    throw new RESTException(RESTException.FORBIDDEN,
                            new Message(BUNDLE, "UNAUTH_READ", securityCtx.getUser(), 
                                        RestEMFUtil.getResourceName(eobject)));
                }
                            
                ctx.setResource(doGetResource(ctx, securityCtx, desc, eobject));
                
            } catch (Exception e) {
                throwException(e);
                
            } finally {
                soaRepo.endRead(true);
            }
        }        
    }

    public void createResource(RESTContext ctx, String collection) throws RESTException {
        
        CollectionDescriptor desc = getDescriptor(collection);
        
        if (!desc.isWritable()) {
            throw new RESTException(RESTException.FORBIDDEN, 
                                    new Message(BUNDLE, "READONLY_COLLECTION", "create in"));
        }
        
        ctx.setRepo(soaRepo);
        
        SecurityContext securityCtx = getSecurityContext(ctx);
        soaRepo.startWrite(getTransactionData(securityCtx));
        boolean passed = false;
        try {

            EClass eclass = desc.getEClass();
            EFactory efactory = desc.getEPackage().getEFactoryInstance();
            EObject eobject = desc.getHandler().createObject(ctx,
                                                             efactory,
                                                             eclass,
                                                             ctx.getSlug(),
                                                             ctx.getResource(),
                                                             true);
            
            ctx.setResource(doGetResource(ctx, securityCtx, desc, eobject));
            passed = true;
        
        } catch (Exception e) {
            throwException(e);
        } finally {
            soaRepo.endWrite(passed);
        }
    }        
    
    public void createResources(RESTContext ctx, String collection) throws RESTException {
        
        CollectionDescriptor desc = getDescriptor(collection);
        
        if (!desc.isWritable()) {
            throw new RESTException(RESTException.FORBIDDEN, 
                                    new Message(BUNDLE, "READONLY_COLLECTION", "create in"));
        }
        
        ctx.setRepo(soaRepo);
        
        SecurityContext securityCtx = getSecurityContext(ctx);
        soaRepo.startWrite(getTransactionData(securityCtx));
        boolean passed = false;
        try {

            EClass eclass = desc.getEClass();
            EFactory efactory = desc.getEPackage().getEFactoryInstance();
            Collection<? extends EObject> eobjects = desc.getHandler().createObjects(ctx,
                                                                                     efactory,
                                                                                     eclass,
                                                                                     ctx.getResourceList(),
                                                                                     true);
            ctx.setResourceList(new ArrayList<Resource>(eobjects.size()));
            for (EObject eObj : eobjects) {
                ctx.getResourceList().add(doGetResource(ctx, securityCtx, desc, eObj));
            }
            
            passed = true;
        
        } catch (Exception e) {
            throwException(e);
        } finally {
            soaRepo.endWrite(passed);
        }
    }        
        
    public void createSubcollectionResource(RESTContext ctx,
                                            String collection,
                                            String id,
                                            String subcollection) throws RESTException {
        
        CollectionDescriptor desc = getDescriptor(collection);
        
        if (!desc.isWritable()) {
            throw new RESTException(RESTException.FORBIDDEN, 
                                    new Message(BUNDLE, "READONLY_COLLECTION", "create in"));
        }
        
        ctx.setRepo(soaRepo);
        SecurityContext securityCtx = getSecurityContext(ctx);
        soaRepo.startWrite(getTransactionData(securityCtx));
        boolean passed = false;
        try {
            EObject containingEObject = getEObject(desc.getEClass(), collection, id);
            
            EObject newObject = desc.getHandler().createContainedObject(ctx,
                                                                        containingEObject,
                                                                        subcollection);
            
            desc = this.getDescriptor(newObject.eClass());
                                    
            ctx.setResource(doGetResource(ctx, securityCtx, desc, newObject));

            passed = true;
        
        } catch (Exception e) {
            throwException(e);
        
        } finally {
            soaRepo.endWrite(passed);
        }
        
    }


    public void updateResource(RESTContext ctx, String collection, String id) throws RESTException {
        
        CollectionDescriptor desc = getDescriptor(collection);
        
        if (!desc.isWritable()) {
            throw new RESTException(RESTException.FORBIDDEN, 
                                    new Message(BUNDLE, "READONLY_COLLECTION", "update in"));
        }

        ctx.setRepo(soaRepo);
        SecurityContext securityCtx = getSecurityContext(ctx);
        soaRepo.startWrite(getTransactionData(securityCtx));
        boolean passed = false;
        try {
            EObject eobject = null;

            if (ctx.getResource().isPrototype()) {
                EClass eclass = desc.getEClass();
                EFactory efactory = desc.getEPackage().getEFactoryInstance();
                eobject = desc.getHandler().createObject(ctx,
                                                         efactory,
                                                         eclass,
                                                         ctx.getSlug(),
                                                         ctx.getResource(),
                                                         true);
            } else {
                eobject = getEObject(desc.getEClass(), collection, id);
                Adapter o = new AdapterImpl();
                eobject.eAdapters().add(o);

                desc.getHandler().getObjectChanges(ctx, eobject, ctx.getResource());
            }
            
            ctx.setResource(doGetResource(ctx, securityCtx, desc, eobject));
            
            passed = true;
        } catch (Exception e) {
            throwException(e);
        } finally {
            soaRepo.endWrite(passed);
        }
    }

 
    public void deleteResource(RESTContext ctx, String collection, String id) throws RESTException {
        
        CollectionDescriptor desc = getDescriptor(collection);
        
        if (!desc.isWritable()) {
            throw new RESTException(RESTException.FORBIDDEN, 
                                    new Message(BUNDLE, "READONLY_COLLECTION", "delete from"));
        }
        
        ctx.setRepo(soaRepo);
        SecurityContext securityCtx = getSecurityContext(ctx);
        soaRepo.startWrite(getTransactionData(securityCtx));
        boolean passed = false;
        try {                        
            String eid = RestEMFUtil.getObjectID(desc.getEClass(), id);
            URI uri = EMFUtil.getProxyURI(desc.getEClass(), eid);

            EObject eobject = soaRepo.getEObject(uri, true);
            if (eobject == null) {
                // TODO: is there a distinction between deleting a
                // previously deleted object (204) and deleting an object
                // that never existed (404)?
                return;
            }
            
            desc.getHandler().deleteObject(ctx, eobject);

            passed = true;
        
        } catch (Exception e) {
            throwException(e);
        } finally {
            soaRepo.endWrite(passed);
        }
    }
    


    public void createMediaResource(RESTContext ctx, String collection) throws RESTException {

        CollectionDescriptor desc = getDescriptor(collection);
        
        if (!desc.isWritable()) {
            throw new RESTException(RESTException.FORBIDDEN, 
                                    new Message(BUNDLE, "READONLY_COLLECTION", "create in"));
        }

        ctx.setRepo(soaRepo);
        SecurityContext securityCtx = getSecurityContext(ctx);
        soaRepo.startWrite(getTransactionData(securityCtx));
        boolean passed = false;
        try {
            EClass eclass = desc.getEClass();
            EFactory efactory = desc.getEPackage().getEFactoryInstance();
            EObject eobject = desc.getHandler()
                .createObjectWithMedia(ctx, efactory, eclass);
            
            if (eobject == null) {
                throw new RESTException(415,
                        new Message(BUNDLE, "MEDIA_NOT_SUPPORTED", desc.getName()));
            }
            
            ctx.setResource(doGetResource(ctx, securityCtx, desc, eobject));

            passed = true;
        
        } catch (Exception e) {
            throwException(e);
        } finally {
            soaRepo.endWrite(passed);
        }
    }        

    
    public void getMediaResource(RESTContext ctx, String collection, String id)
        throws RESTException {
        
        CollectionDescriptor desc = getDescriptor(collection);

        SecurityContext securityCtx = getSecurityContext(ctx);
        soaRepo.startRead(getTransactionData(securityCtx));
        
        try {
            EObject eobject = getEObject(desc.getEClass(), collection, id);
            
            if (!isReadAuthorized(securityCtx, eobject) && !isDownloadAuthorized(securityCtx, eobject)) { 
                throw new RESTException(RESTException.FORBIDDEN,
                        new Message(BUNDLE, "UNAUTH_READ", securityCtx.getUser(), 
                                    RestEMFUtil.getResourceName(eobject)));
            }

            desc.getHandler().getMedia(ctx, eobject);
            
            if (ctx.getInputStream() == null) {
                throw new RESTException(415,
                        new Message(BUNDLE, "MEDIA_NOT_SUPPORTED", desc.getName()));
            }
            
        } catch (Exception e) {
            throwException(e);
        } finally {
            soaRepo.endRead(true);
        }
        
    }

    public void updateMediaResource(RESTContext ctx, String collection,
                                    String id) 
        throws RESTException {
        CollectionDescriptor desc = getDescriptor(collection);
        
        if (!desc.isWritable()) {
            throw new RESTException(RESTException.FORBIDDEN, 
                                    new Message(BUNDLE, "READONLY_COLLECTION", "update in"));
        }

        ctx.setRepo(soaRepo);
        SecurityContext securityCtx = getSecurityContext(ctx);
        soaRepo.startWrite(getTransactionData(securityCtx));
        boolean passed = false;
        try {
            EObject eobject = getEObject(desc.getEClass(), collection, id);
            
            eobject = desc.getHandler().updateMedia(ctx, eobject);
            
            if (eobject == null) {
                throw new RESTException(415,
                        new Message(BUNDLE, "MEDIA_NOT_SUPPORTED", desc.getName()));
            }
       
            ctx.setResource(doGetResource(ctx, securityCtx, desc, eobject));

            passed = true;
        } catch (Exception e) {
            throwException(e);
        } finally {
            soaRepo.endWrite(passed);
        }
    }
    
    public boolean isReferenceable(String className) {
        return eClassPaths.containsKey(className);
    }

    public String getCollectionHref(String className) { 
        return eClassPaths.get(className);
    }
       
    public boolean isFeatureExcluded(String qualName) {
        return excludedFeatures.contains(qualName);
    }
    
    public boolean isHidden(String qualName) {
        return hiddenFeatures.contains(qualName);
    }

    public boolean isReadonly(String qualName) {
        return readonlyFeatures.contains(qualName);
    }
    
    public EClass getEClass(String name) {
        return collectionEClasses.get(name);
    }

    public EFeatureHandler getEFeatureHandler(String qualName) {
        return featureHandlers.get(qualName);
    }
    
    public String getXsdType(EAttribute att) throws RESTException {        
        EDataType etype = att.getEAttributeType();        
        if (etype.getInstanceClass() == String.class && RestEMFUtil.isClob(att)) {
            // Special case: for strings annotated with CLOB use xsd:string
            return "xsd:string";
        } else if (etype instanceof EEnum) {
            return "xsd:normalizedString";
        }
        
        String xsdType = xsdTypeMap.get(etype.getName());
        
        if (xsdType == null) {
            internalError("No mapping for type " + etype.getName());
        }
        
        return xsdType;
    }      
    


    public void internalError(String msg) throws RESTException {
        throw new RESTException(
                RESTException.NOT_IMPLEMENTED,
                new Message(BUNDLE, "INTERNAL_ERROR", msg));        
    }
    
    public SOARepository getRepository() {
        return soaRepo;
    }
    
    
    private CollectionDescriptor addCollection(String name,
                                               EPackage epackage,
                                               EClass eclass,
                                               EClassHandler handler,
                                               boolean enableFilter,
                                               boolean writabe) {
        CollectionDescriptor cd = addCollection(name, epackage, eclass, handler)
            .setFilterEnabled(enableFilter)
            .setWritable(writabe);
        return cd;
    }
    
    private CollectionDescriptor addCollection(String name,
                                               EPackage epackage,
                                               EClass eclass,
                                               EClassHandler handler) {
        CollectionDescriptor cd = new CollectionDescriptor(name)
            .setEPackage(epackage)
            .setEClass(eclass)
            .setHandler(handler)
            .setIsResourceBased(false)
            .setQueryExecutor(defaultQueryExecutor);
        collections.put(name, cd);
        if (eclass != null) {
            String className = RestEMFUtil.getQualifiedClassName(eclass);
            eClassPaths.put(className, name);
            collectionEClasses.put(name, eclass);
        }
        return cd;
    }

    private void removeCollection(String name) {
        collections.remove(name);
    } 
    
    private void makeReferenceable(EClass eclass, String path) {
        String className = RestEMFUtil.getQualifiedClassName(eclass);
        eClassPaths.put(className, path);
        
    }
    
    public CollectionDescriptor getDescriptor(String collection) throws RESTException {
        CollectionDescriptor desc = collections.get(collection);
        if (desc == null) {
            throw new RESTException(404,
                    new Message(BUNDLE, "UNKNOWN_COLLECTION", collection));
        }
        return desc;
    }
    
    public CollectionDescriptor getDescriptor(EClass eclass) throws RESTException {
        String collection = eClassPaths.get(RestEMFUtil.getQualifiedClassName(eclass));
        if (collection == null) {
            throw new RESTException(404,
                    new Message(BUNDLE, 
                            "UNKNOWN_COLLECTION_CLASS", 
                            RestEMFUtil.getQualifiedClassName(eclass)));
        }
        return getDescriptor(collection);
    }
    
    public CollectionDescriptor findDescriptor(EClass eclass) throws RESTException {
        String collection = eClassPaths.get(RestEMFUtil.getQualifiedClassName(eclass));
        if (collection == null) {
            return null;
        }
        return getDescriptor(collection);
    }
    
    public RESTQueryExecutor getQueryExecutor(String collection) throws RESTException {
        return getDescriptor(collection).getQueryExecutor();
    }

    private EObject getEObject(EClass eclass, String collection, String id) 
        throws RESTException {

        String eid = RestEMFUtil.getObjectID(eclass, id);
        URI uri = EMFUtil.getProxyURI(eclass, eid);
        EObject eobject = soaRepo.getEObject(uri, true);

        if (eobject == null) {
            throw new RESTException(
                    RESTException.NOT_FOUND,
                    new Message(BUNDLE, "NOT_FOUND", collection,  id));
        }
        return eobject;
    }


    Resource doGetResource(RESTContext ctx,
                           SecurityContext securityCtx, 
                           CollectionDescriptor desc,
                           EObject eobject) throws RESTException {
                
        String name = RestEMFUtil.getResourceName(eobject);
        String id = RestEMFUtil.getResourceId(eobject);
        
        if (desc == null) {
            // This means the collection is not tied to a particular eclass.
            desc = findDescriptor(eobject.eClass());
            if (desc == null) {
                LOG.fine("cannot add object " + eobject + " to collection");
                return null;
            }
        }
        
        Resource resource = resourceFactory.createResource(name, desc.getName(), id);
        desc.getHandler().addObjectToResource(ctx, eobject, resource);
        resource.setEditable(isEditAuthorized(securityCtx, eobject));
        resource.setDeletable(isDeleteAuthorized(securityCtx, eobject));
                
        if (LOG.isLoggable(Level.FINE)) {
            LOG.fine(resource.toString());
        }
            
        return resource; 

    }
    
    
    Resource doGetResource(RESTContext ctx,
                           SecurityContext securityCtx, 
                           CollectionDescriptor desc,
                           EObject eobject,
                           boolean omitSubclasses) throws RESTException {
        
        // Only return resource if eclass matches.  Might be better to use a query
        // to restrict the range of objects returned.
        if (omitSubclasses) {
            if (eobject.eClass().equals(desc.getEClass())) {
                return doGetResource(ctx, securityCtx, desc, eobject);
            }
        } else if (desc == null || desc.getEClass() == null 
                || desc.getEClass().isSuperTypeOf(eobject.eClass())) {
            return doGetResource(ctx, securityCtx, null, eobject);
        }
        
        if (LOG.isLoggable(Level.FINE)) {
            LOG.fine("ignoring eobject " + eobject + "; eclass is not " + desc.getEClass());
        }
        return null;
    }   
    
    static SecurityContext getSecurityContext(RESTContext ctx) {
        Object o = ctx.getRequestAttribute(RESTContext.SECURITY_CONTEXT);
        if (o instanceof SecurityContext) {
            return (SecurityContext) o;
        }
        
        return null;
    }
    
    static TransactionData getTransactionData(SecurityContext ctx) {
        String ssoToken = null;
        String userId = UNKNOWN_USER;
        if (ctx != null) {
            ssoToken = ctx.getSSOToken();
            userId = ctx.getUser();            
        }
        return new TransactionData(userId, ssoToken);
    }
    
    boolean isDownloadAuthorized(SecurityContext ctx, EObject eobject) {
        if (securityService == null) {
            return true;
        }
        if (LOG.isLoggable(Level.FINE)) {
            LOG.fine("Authorizing download " + ctx.getUser() + "/" + ctx.getSSOToken());
        }
        if (!securityService.authorizeRead(ctx.getSSOToken(), eobject, DOWNLOAD_AREAS)) {
            LOG.fine("User \"" + ctx.getUser()
                    + "\" is not authorized to download object contents");  
            return false;
        }
        
        return true;
    }
    
    boolean isReadAuthorized(SecurityContext ctx, EObject eobject) {
        if (RestEMFUtil.isInternal(eobject)) {
            return false;
        } else if (securityService == null) {
            return true;
        }
        if (LOG.isLoggable(Level.FINE)) {
            LOG.fine("Authorizing read " + ctx.getUser() + "/" + ctx.getSSOToken());
        }
        
        if (eobject.eResource() == null && eobject.eAdapters().size() == 0) {
            // not a persisted object. E.g. a synthesised ObjectPermission instance 
            // used to represent default permissions
            return true;
        }
        
        if (!securityService.authorizeRead(ctx.getSSOToken(), eobject)) {
            LOG.fine("User \"" + ctx.getUser()
                    + "\" is not authorized to access object " + eobject);  
            return false;
        }
        
        return true;
    }
    
    private boolean isCreateAuthorized(SecurityContext ctx, EClass eClass, EObject container) {
        if (securityService == null) {
            return true;
        }
        if (LOG.isLoggable(Level.FINE)) {
            LOG.fine("Authorizing create " + ctx.getUser() + "/" + ctx.getSSOToken());
        }
        
        return securityService.authorizeCreate(ctx.getSSOToken(), eClass, container);
    }
    
    private boolean isEditAuthorized(SecurityContext ctx, EObject eobject) {
        if (!RestEMFUtil.isModifiable(eobject)) {
            return false;
        } else if (securityService == null) {
            return true;
        }
        if (LOG.isLoggable(Level.FINE)) {
            LOG.fine("Authorizing edit " + ctx.getUser() + "/" + ctx.getSSOToken());
        }
        
        if (eobject.eResource() == null && eobject.eAdapters().size() == 0) {
            // not a persisted object. E.g. a synthesised ObjectPermission instance 
            // used to represent default permissions
            return false;
        }
        
        return securityService.authorizeUpdate(ctx.getSSOToken(), eobject, null);
    }
    
    private boolean isDeleteAuthorized(SecurityContext ctx, EObject eobject) {
        if (!RestEMFUtil.isModifiable(eobject)) {
            return false;
        } else if (securityService == null) {
            return true;
        }
        if (LOG.isLoggable(Level.FINE)) {
            LOG.fine("Authorizing delete " + ctx.getUser() + "/" + ctx.getSSOToken());
        }
        
        if (eobject.eResource() == null && eobject.eAdapters().size() == 0) {
            // not a persisted object. E.g. a synthesised ObjectPermission instance 
            // used to represent default permissions
            return false;
        }
        
        return securityService.authorizeDelete(ctx.getSSOToken(), eobject);
    }


    private void handleErrorsAndWarnings() {
        // TODO: see same function in AbstractInfrastructureProvider
    }

    public void registerCollectionService(RESTCollectionService service) {        
        EClass eclass = service.getEClass();        
        EPackage epkg;
        if (eclass != null) {
            epkg = eclass.getEPackage();            
        } else {
            epkg = null;
        }
        addCollection(service.getName(), epkg, eclass, 
            service.getHandler(), !service.includeSubclasses(), true);
    }

    public void unregisterCollectionService(RESTCollectionService service) {
        removeCollection(service.getName());
    }

    private void throwException(Exception e) throws RESTException {
        if (e instanceof RESTException) {
            throw (RESTException) e;
        }
        
        if (e instanceof AuthorizationException) {
            AuthorizationException ex = (AuthorizationException) e;
            throw new RESTException(RESTException.FORBIDDEN, ex.getI18nMessage()); 
        }    
        
        throw new RESTException(RESTException.INTERNAL_SERVER_ERROR, e);   

    }
    
    
}


