/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.rest.rt.repo;

import java.util.List;
import java.util.logging.Logger;

import com.iona.repository.rest.RESTContext;
import com.iona.repository.rest.RESTException;
import com.iona.repository.rest.resource.Attribute;
import com.iona.repository.rest.resource.Resource;
import com.iona.soa.expression.Equals;
import com.iona.soa.expression.Expression;
import com.iona.soa.model.userattributes.TagDictionary;
import com.iona.soa.model.userattributes.UserAttributeDefinition;
import com.iona.soa.model.userattributes.UserAttributesFactory;
import com.iona.soa.model.userattributes.UserAttributesPackage;
import com.iona.soa.resourceset.QueryResourceSet;
import com.iona.soa.resourceset.ResourceSetQueryHandler;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

public class TagsHandler extends AbstractEFeatureHandler implements
        EFeatureHandler {

    private static final Logger LOG = Logger.getLogger(TagsHandler.class.getName());
    private static final String DEFAULT_DICTIONARY = "Default";
    private TagDictionary tagDictionary;
    

    public TagsHandler(RepositoryRESTAdapter a) {
        super(a);
    }

    @SuppressWarnings("unchecked")
    public void addFeatureToResource(String qualName, EStructuralFeature feature, 
                                     EObject eobject, Resource resource)
        throws RESTException {

        EList elist = (EList) eobject.eGet(feature);
        if (elist.size() == 0) {
            resource.addAttribute("tags")
                .setMin(feature.getLowerBound())
                .setMax(feature.getUpperBound())
                .setType("xsd:normalizedString")
                .setText(null);
        } else {
            for (Object o : elist) {
                if (!(o instanceof UserAttributeDefinition)) {
                    restProvider.internalError("tag is not a UserAttributeDefinition");
                }
                UserAttributeDefinition tag = (UserAttributeDefinition) o;
                resource.addAttribute("tags")
                    .setMin(feature.getLowerBound())
                    .setMax(feature.getUpperBound())
                    .setType("xsd:normalizedString")
                    .setText(tag.getName());
            }
        }
    }

    @SuppressWarnings("unchecked")
    public void getFeatureChanges(RESTContext ctx, String qualName,
                                  EStructuralFeature feature, EObject eobject, 
                                  Resource resource)
        throws RESTException {

        EReference ref = (EReference) feature;
        
        EList<EObject> tags = (EList<EObject>) eobject.eGet(ref);
        
        List<Attribute> atts = resource.getAttributes("tags");
        // If only attribute is empty remove it.
        if (atts.size() == 1 && atts.get(0).getText() == null) {
            atts.clear();
        }

        if (atts.size() == 0 && tags.size() == 0) {
            return;
        }
        
        try {      
            
            EList<EObject> newList = new BasicEList<EObject>();
            for (Attribute att : atts) {
                newList.add(findTag(ctx, att.getText()));
            }

            this.getReferenceListChanges(ctx, ref, eobject, newList, tags);
            
        } catch (Exception e) {
            throw new RESTException(RESTException.INTERNAL_SERVER_ERROR, e);
        }        
    }

    private UserAttributeDefinition findTag(RESTContext ctx, String name)
        throws RESTException {

        LOG.fine("looking for tag: " + name);

        TagDictionary dict = findTagDictionary(ctx, DEFAULT_DICTIONARY);
        // TODO: better searching/caching needed here.
        for (UserAttributeDefinition tag : dict.getAttributeDefs()) {
            if (tag.getName().equals(name)) {
                return tag;
            }
        }
        return createTag(ctx, name);
    }

    private UserAttributeDefinition createTag(RESTContext ctx, String name) 
        throws RESTException {

        LOG.fine("creating tag: " + name);
        UserAttributeDefinition tag = UserAttributesFactory.eINSTANCE.createUserAttributeDefinition();
        tag.setName(name);

        return addTag(ctx, tag);
    }


    private UserAttributeDefinition addTag(RESTContext ctx, UserAttributeDefinition tag) 
        throws RESTException {

        TagDictionary dict = findTagDictionary(ctx, DEFAULT_DICTIONARY);
        dict.getAttributeDefs().add(tag);
        return tag;
    }


    private TagDictionary findTagDictionary(RESTContext ctx, String name) 
        throws RESTException {

        // Find a tag dictionary called name.

        if (tagDictionary == null) {
            try {
                EClass eclass = UserAttributesPackage.Literals.TAG_DICTIONARY;

                Expression expr = new Equals("name", name);
                QueryResourceSet rs = restProvider.getRepository().getResourceSet();
                List<EObject> dicts =  
                    ResourceSetQueryHandler.executeQuery(rs, eclass, expr, true);
                for (EObject eobject : dicts) {
                    if (eobject instanceof TagDictionary) {
                        TagDictionary dict = (TagDictionary) eobject;
                        if (dict.getName().equals(name)) {
                            tagDictionary = dict;
                        }
                    }
                }

                if (tagDictionary == null) {
                    tagDictionary = UserAttributesFactory.eINSTANCE.createTagDictionary();
                    tagDictionary.setName(name);
                    ctx.getRepo().persist(tagDictionary);
                }

            } catch (Exception e) {
                throw new RESTException(RESTException.INTERNAL_SERVER_ERROR, e);
            }
        }
        return tagDictionary;
    }
}
