/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.rest.rt.repo;

import java.util.ArrayList;
import java.util.List;

import com.iona.repository.rest.RESTContext;
import com.iona.repository.rest.RESTException;
import com.iona.repository.rest.repo.EClassHandler;
import com.iona.repository.rest.resource.Resource;
import com.iona.repository.rest.rt.server.RESTProvider;
import com.iona.soa.model.Application;
import com.iona.soa.model.ConfigurationOverride;
import com.iona.soa.model.DeploymentProfile;
import com.iona.soa.model.SOA_NetworkPackage;
import com.iona.soa.model.VersionedProfile;
import com.iona.soa.model.repository.RepositoryPackage;
import com.iona.soa.repository.EMFUtil;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;

public class VersionedProfileHandler extends DefaultEClassHandler implements
        EClassHandler {

    public VersionedProfileHandler(RESTProvider a) {
        super(a);
    }

    @Override
    public void getObjectChanges(RESTContext ctx,
                                 EObject eobject,
                                 Resource resource) throws RESTException {
        
        if (!(eobject instanceof VersionedProfile)) {
            return;
        }
        
        VersionedProfile toUpdate = (VersionedProfile)eobject;
        DeploymentProfile dp = (DeploymentProfile) toUpdate.eContainer();

        if (EcoreUtil.getID(toUpdate).equals(EcoreUtil.getID(dp.getCurrent()))) {
            if (dp.eIsSet(SOA_NetworkPackage.eINSTANCE.getDeploymentProfile_Active())) {
                // If the profile being modified is the active one then 
                //   * create a duplicate of the active 
                //   * add it to the version list 
                //   * get the changes in the resource applied to the new version
                //     (by redirecting toUpdate)
                //
                VersionedProfile active = dp.getActive();
                if (EcoreUtil.getID(active).equals(EcoreUtil.getID(toUpdate))) {
                    toUpdate = (VersionedProfile)EMFUtil.copyObjectWithFreshGuids(active);
                    
                    dp.getVersions().add(toUpdate);
                    dp.setCurrent(toUpdate);
                }
            }

        } else {
            // The profile being modified is not the current one replace current
            // with a copy of the one being updated and apply the changes there.
            // This occurs when a profile from the "versions list" is edited, as 
            // opposed to the "current" one.
            // Note - this is not "revert".  Revert is handled by the DeploymentProfileHandler
            // by its responding to submitted changes to the "active" reference.
            toUpdate  = (VersionedProfile)EMFUtil.copyObjectWithFreshGuids(toUpdate);
            dp.getVersions().remove(dp.getCurrent());
            dp.getVersions().add(toUpdate);
            dp.setCurrent(toUpdate);
        }
        
        // Apply changes described in resource to the VersionedProfile.
        super.getObjectChanges(ctx, toUpdate, resource);

        // personal hygiene: remove any config overrides that doesn't map to an assigned application
        List<String> appIds = new ArrayList<String>();
        for (EObject app : toUpdate.getApplications()) {
            appIds.add(EcoreUtil.getID(app));
        }
        List<EObject> cfgOverridesToRemove = new ArrayList<EObject>();
        for (EObject override : toUpdate.getConfigurationOverride()) {
            Application app = ((ConfigurationOverride)override).getApplication();
            if (app == null || !appIds.contains(EcoreUtil.getID(app))) {
                cfgOverridesToRemove.add(override);
            }
        }
        for (EObject toRemove : cfgOverridesToRemove) {
            toUpdate.getConfigurationOverride().remove(toRemove);
        }
        
        
        // Clients edit a VersionedProfile - keep the DeploymentProfile name synchronized
        if (toUpdate.eIsSet(RepositoryPackage.eINSTANCE.getIRepositoryObject_Name())) {
            dp.setName(toUpdate.getName());
        }
        
        // Set the active link to the object being updated
        if (toUpdate.isPublishNow()) {
            dp.setActive(toUpdate);
            // Should not be required as this feature is transient and therefore
            // should not be persisted.  But leaving in as a safeguard.
            toUpdate.setPublishNow(false);
        }
    }
    
    
}
