/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.rest.rt.server;

import java.util.logging.Logger;

import com.iona.repository.rest.RESTContext;
import com.iona.repository.rest.RESTException;
import com.iona.soa.repository.util.ExceptionHelper;
import com.iona.soa.security.AuthenticationException;
import com.iona.soa.security.SecurityContext;
import com.iona.soa.security.SecurityService;

import org.apache.abdera.protocol.server.Filter;
import org.apache.abdera.protocol.server.FilterChain;
import org.apache.abdera.protocol.server.ProviderHelper;
import org.apache.abdera.protocol.server.RequestContext;
import org.apache.abdera.protocol.server.RequestContext.Scope;
import org.apache.abdera.protocol.server.ResponseContext;


public class AuthFilter implements Filter {
    
    private static final Logger LOG = Logger.getLogger(AuthFilter.class.getName());
    
    private SecurityService securityService;

    public AuthFilter(SecurityService repositoryProtector) {
        this.securityService = repositoryProtector;
        
        if (repositoryProtector == null) {
            LOG.warning("\n\n\tWARNING: Security over REST disabled\n\n");
        }
    }

    public ResponseContext filter(RequestContext request, FilterChain chain) {
        
        if (securityService == null) {
            return chain.next(request);
        }

        boolean isWebClient = false;
        try {
            SecurityContext ctx;
            String depotCookie = request.getHeader("Cookie");
            if (depotCookie != null && depotCookie.contains(SecurityService.TOKEN_TYPE)) {
                int index1 = depotCookie.indexOf(SecurityService.TOKEN_TYPE);
                int index2 = depotCookie.indexOf(';', index1);
                if (index2 == -1) {
                    depotCookie = depotCookie.substring(index1);
                } else {
                    depotCookie = depotCookie.substring(index1, index2);
                }
                depotCookie = depotCookie.replaceFirst("=", " ");
                depotCookie = depotCookie.replaceAll("\"", "");
                isWebClient = true;
                ctx = securityService.authenticate(depotCookie);
            } else {
                ctx = securityService.authenticate(request.getAuthorization());
            }
            request.setAttribute(Scope.REQUEST, RESTContext.SECURITY_CONTEXT, ctx);

        } catch (AuthenticationException e) {
            logAuthFailure(e, request);
            return authFailureResponse(request, e.getLocalizedMessage(), !isWebClient);
        } catch (Throwable t) {
            logAuthFailure(t, null);
            return authFailureResponse(request, "Exception during authentication; " + t, !isWebClient);
        }

        return chain.next(request);
    }
    

    private void logAuthFailure(Throwable t, RequestContext request) {
        
        if (t instanceof RESTException) {
            LOG.warning(new StringBuilder()
                .append("Authorization failed: ")
                .append(t.getMessage())
                .append("\nRequest: ")
                .append(request.getMethod())
                .append(", ")
                .append(request.getResolvedUri())
                .append("\nUser-Agent: ")
                .append(request.getHeader("User-Agent"))
                .toString());
    
        } else {
            LOG.warning("Authorization failed due to exception " + t);
            LOG.fine(ExceptionHelper.getStackTraceAsString(t));
        }
    }
    
    private ResponseContext authFailureResponse(RequestContext request, String msg, boolean addBscAuthHdr) {
        ResponseContext rc = ProviderHelper.unauthorized(request, msg);
        if (addBscAuthHdr) {
            rc.addHeader(SecurityService.AUTHENTICATE_RESPONSE_HEADER,
                         SecurityService.BASIC_AUTH_TYPE);
        }
        return rc;
    }
}
