/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.rest.rt.server;

import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.MimeType;

import org.apache.abdera.protocol.server.Filter;
import org.apache.abdera.protocol.server.FilterChain;
import org.apache.abdera.protocol.server.RequestContext;
import org.apache.abdera.protocol.server.RequestContext.Scope;
import org.apache.abdera.protocol.server.ResponseContext;
import org.apache.abdera.protocol.server.Target;
import org.apache.abdera.util.EntityTag;

public class LoggingFilter implements Filter {

    private static final Logger LOG = Logger.getLogger(LoggingFilter.class.getName());

    
    public ResponseContext filter(RequestContext request, FilterChain chain) {
        if (!LOG.isLoggable(Level.INFO)) {
            return chain.next(request);
        }
        logRequest(request);
        ResponseContext response = chain.next(request);
        logResponse(response);
        return response;
    }
    
    private void logRequest(RequestContext request) {
        try {
            if (LOG.isLoggable(Level.FINE)) {
                LOG.fine(getDetailedRequest(request));
            } else {
                LOG.info(getSummaryRequest(request));
            }
        } catch (Throwable t) {
            LOG.warning("Exception when logging request: " + t);
        }
    }
    
    
    private void logResponse(ResponseContext response) {
        try {
            if (LOG.isLoggable(Level.FINE)) {
                LOG.fine(getDetailedResponse(response));
            } else {
                String status = response.getStatusText();
                MimeType mt = response.getContentType();
                LOG.info(new StringBuilder()
                        .append(response.getStatus())
                        .append(" ")
                        .append(status == null ? "" : status)
                        .append(" [")
                        .append(mt == null ? "?" : mt.toString())
                        .append("]")
                        .toString());
            }
        } catch (Throwable t) {
            LOG.warning("Exception when logging response: " + t);
        }
        
    }


    private String getSummaryRequest(RequestContext request) {
        Object user = request.getAttribute(Scope.REQUEST, "username");
        StringBuilder sb = new StringBuilder()
            .append(request.getMethod())
            .append(" ")
            .append(request.getResolvedUri())
            .append(" [")
            .append(user == null ? "?" : user)
            .append(", ")
            .append(request.getTarget().getType())
            .append("]\nUser-Agent: ").append(request.getHeader("User-Agent"));
        return sb.toString();
    }

    private String getDetailedRequest(RequestContext request) {
        StringBuilder sb = new StringBuilder();
        sb.append("\nMessage Headers:");
        for (String s : request.getHeaderNames()) {
            sb.append("\n  ").append(s).append(":");
            for (Object v : request.getHeaders(s)) {
                sb.append(" ").append(v);
            }
        }
        sb.append("\nMessage:")
            .append("\n  Slug:                ").append(request.getSlug())
            .append("\n  Cache-Control:       ").append(request.getCacheControl())
            .append("\n  Content-Type:        ").append(request.getContentType())
            .append("\n  Content-Location:    ").append(request.getContentLocation())
            .append("\n  Content-Language:    ").append(request.getContentLanguage())
            .append("\n  Max Age:             ").append(request.getMaxAge())
            .append("\n  No-Cache:            ").append(request.isNoCache())
            .append("\n  No-Store:            ").append(request.isNoStore())
            .append("\n  No-Transform:        ").append(request.isNoTransform())
            .append("\nRequest:")
            .append("\n  Accept:              ").append(request.getAccept())
            .append("\n  Accept Charset:      ").append(request.getAcceptCharset())
            .append("\n  Accept Encoding:     ").append(request.getAcceptEncoding())
            .append("\n  Accept Language:     ").append(request.getAcceptLanguage())
            .append("\n  Authorization:       ").append(request.getAuthorization())
            .append("\n  If-Modified-Since:   ").append(request.getIfModifiedSince())
            .append("\n  If-Unmodified-Since: ").append(request.getIfUnmodifiedSince())
            .append("\n  Max Stale:           ").append(request.getMaxStale())
            .append("\n  Min Fresh:           ").append(request.getMinFresh())
            .append("\n  Only-If-Cached:      ").append(request.isOnlyIfCached())
            .append("\n  If-Match:            ");
        for (EntityTag e : request.getIfMatch()) {
            sb.append(e).append(" ");
        }
        sb.append("\nRequest Context: ")
            //.append("\n  Subject:             ").append(request.getSubject())
            .append("\n  Locale:              ").append(request.getPreferredLocale())
            .append("\n  Method:              ").append(request.getMethod())
            .append("\n  URI:                 ").append(request.getUri())
            .append("\n  Resolved URI:        ").append(request.getResolvedUri())
            .append("\n  BaseURI:             ").append(request.getBaseUri())
            .append("\n  Context Path:        ").append(request.getContextPath())
            .append("\n  Target Path:         ").append(request.getTargetPath())
            .append("\n  Target Base Path:    ").append(request.getTargetBasePath());

        sb.append("\nParameters:");
        for (String p : request.getParameterNames()) {
            sb.append("\n    ").append(p).append(":");
            for (String v : request.getParameters(p)) {
                sb.append(" ").append(v);
            }
        }
        Target target = request.getTarget();
        sb.append("\nTarget:")
            .append("\n  Target Type:         ").append(target.getType())
            .append("\n  Identity:            ").append(target.getIdentity())
            .append("\nTarget Parameters:");
        for (String s : target.getParameterNames()) {
            sb.append("\n  ").append(s).append(": ")
              .append(target.getParameter(s));
        }
        sb.append("\nAttribtues (request scope):");
        for (String s : request.getAttributeNames(Scope.REQUEST)) {
            sb.append("\n  ").append(s)
                .append(":   ").append(request.getAttribute(Scope.REQUEST, s));
        }
        sb.append("\n");
        return sb.toString();
    }

    private String getDetailedResponse(ResponseContext response) {
        StringBuilder sb = new StringBuilder();
        sb.append("Status: ")
            .append(response.getStatus())
            .append(" ")
            .append(response.getStatusText());
        
        sb.append("\nMessage Headers:");
        for (String s : response.getHeaderNames()) {
            sb.append("\n  ").append(s).append(":");
            for (Object v : response.getHeaders(s)) {
                sb.append(" ").append(v);
            }
        }
        return sb.toString();
    }


}
