/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.rest.rt.server;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.SequenceInputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.Enumeration;
import java.util.List;
import java.util.ResourceBundle;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.MimeType;
import javax.mail.BodyPart;
import javax.mail.Header;
import javax.mail.MessagingException;
import javax.mail.internet.MimeMessage;
import javax.mail.internet.MimeMultipart;

import com.iona.repository.rest.Constants;
import com.iona.repository.rest.RESTContext;
import com.iona.repository.rest.RESTException;
import com.iona.repository.rest.RESTQuery;
import com.iona.repository.rest.atom.AbderaUtil;
import com.iona.repository.rest.atom.AtomEntryFactory;
import com.iona.repository.rest.atom.AtomEntryOptions;
import com.iona.repository.rest.resource.Resource;
import com.iona.repository.rest.resource.ResourceComparator;
import com.iona.repository.rest.rt.atom.AtomFeedFactory;
import com.iona.repository.rest.rt.osgi.MessageBundle;
import com.iona.repository.util.i18n.Message;
import com.iona.soa.repository.util.ExceptionHelper;

import org.apache.abdera.Abdera;
import org.apache.abdera.ext.opensearch.OpenSearchConstants;
import org.apache.abdera.factory.Factory;
import org.apache.abdera.model.Base;
import org.apache.abdera.model.Document;
import org.apache.abdera.model.Element;
import org.apache.abdera.model.Entry;
import org.apache.abdera.model.Feed;
import org.apache.abdera.model.Link;
import org.apache.abdera.protocol.server.ProviderHelper;
import org.apache.abdera.protocol.server.RequestContext;
import org.apache.abdera.protocol.server.ResponseContext;
import org.apache.abdera.protocol.server.Target;
import org.apache.abdera.protocol.server.context.EmptyResponseContext;
import org.apache.abdera.protocol.server.context.MediaResponseContext;
import org.apache.abdera.protocol.server.context.ResponseContextException;
import org.apache.abdera.protocol.server.impl.AbstractCollectionAdapter;

public class RepositoryAdapter extends AbstractCollectionAdapter {

    private static final ResourceBundle BUNDLE = MessageBundle.getInstance();

    private static final Logger LOG = Logger.getLogger(RepositoryAdapter.class.getName());

    private static final String DEFAULT_ORDERING_ATTRIBUTE = "timestamp";

    private static final Date OLTEST_DATE = new Date(0);

    private static final AtomEntryOptions DEFAULT_ENTRY_OPTIONS = new AtomEntryOptions();
    private static final AtomEntryOptions MINIMAL_ENTRY_OPTIONS = 
        new AtomEntryOptions()
            .includeRelatedLinks(false)
            .includeSummary(false)
            .includeXhtml(false)
            .includeResource(true);


    private static Abdera abdera = AbderaUtil.getAbdera();

    private AtomFeedFactory feedFactory;
    private AtomEntryFactory entryFactory;
    private RESTAdapter restAdapter;
    private String basePath;
    private String fullBasePath;


    public RepositoryAdapter(RESTAdapter restAdapter,
                             AtomFeedFactory feedFactory, 
                             AtomEntryFactory entryFactory,
                             String basePath) {

        this.restAdapter = restAdapter;
        this.feedFactory = feedFactory;
        this.entryFactory = entryFactory;
        this.basePath = basePath;
        fullBasePath = "/" + basePath;
    }


    @Override
    public String getId(RequestContext request) {
        return null;
    }

    public String getTitle(RequestContext request) {
        return new Message(BUNDLE, "COLLECTION_TITLE").toString();
    }

    @Override 
    public String getHref() {
        return null;
    }

    @Override
    public ResponseContext getCategories(RequestContext request) {
        return ProviderHelper.notfound(request,
                new Message(BUNDLE, "NO_CATAGORIES").toString());
    }

    @Override
    public String getAuthor(RequestContext request) throws ResponseContextException {
        return new Message(BUNDLE, "DEFAULT_AUTHOR").toString();
    }
    
    public ResponseContext getFeed(RequestContext request) {

        try {
            Target target = request.getTarget();
            String baseUri = getBaseUri(request);
            String collection = getCollection(request);
            String id = target.getParameter("id");
            String subcollection = target.getParameter("subcollection");
            RESTQuery query = new RESTQueryImpl(target.getParameter("query"));

            AtomEntryOptions options = DEFAULT_ENTRY_OPTIONS;
            if (query.isSet(RESTQuery.QUERY_MIN)) {
                options = MINIMAL_ENTRY_OPTIONS;
            }                        
            
            RESTContext ctx = new RESTContextImpl()
                .setRequest(request)
                .setCollection(collection)
                .setRestQuery(query);
            
            Feed feed = null;
            if (id == null && subcollection == null) {
                
                restAdapter.getCollection(ctx, collection, query);                
                feed = feedFactory.createFeed(baseUri, collection, ctx.getCollectionTitle());
                
                Comparator<? super Resource> comparator = 
                    getComparator(query.getParameter(RESTQuery.QUERY_ORDERBY),
                                  query.getParameter(RESTQuery.QUERY_ASCENDING));
                if (comparator != null) {
                    Collections.sort(ctx.getResourceList(), comparator);
                }

                Date last = addResourcesToFeed(request, feed, ctx, baseUri, collection, options, query);
                feed.setUpdated(last);
       
            } else if (id != null && subcollection != null) {
                String path = new StringBuilder()
                    .append(collection)
                    .append("/")
                    .append(id)
                    .append(";")
                    .append(subcollection)
                    .toString();

                restAdapter.getSubCollection(ctx, collection, id, subcollection);
                feed = feedFactory.createFeed(baseUri, path, ctx.getCollectionTitle());

                Comparator<? super Resource> comparator = 
                    getComparator(query.getParameter(RESTQuery.QUERY_ORDERBY),
                                  query.getParameter(RESTQuery.QUERY_ASCENDING));
                if (comparator != null) {
                    Collections.sort(ctx.getResourceList(), comparator);
                }
                
                Date last = addResourcesToFeed(request, feed, ctx, baseUri, path, options, query);
                feed.setUpdated(last);
                
            } else {
                throw new RESTException(
                        RESTException.BAD_REQUEST,
                        request,
                        new Message(BUNDLE, 
                                "BAD_SUBCOLLECTION",
                                request.getMethod(),
                                request.getResolvedUri()));
            }

            return okDocResponse(feed, feed.getUpdated());

        } catch (RESTException e) {
            return errorResponse(e);

        } catch (Throwable t) {
            return errorResponse(request, t);
        }
    }


    public ResponseContext getEntry(RequestContext request) {       
        try {
            String baseUri = getBaseUri(request);

            String collection = getCollection(request);
            String id = getResourceId(request);

            Target target = request.getTarget();
            
            RESTQuery query = new RESTQueryImpl(target.getParameter("query"));
            
            if (query.isSet(RESTQuery.QUERY_MEDIA)) {
                return getMedia(request);
            }

            AtomEntryOptions options = DEFAULT_ENTRY_OPTIONS;
            
            if (query.isSet(RESTQuery.QUERY_MIN)) {
                options = MINIMAL_ENTRY_OPTIONS;
            }
            
            RESTContext ctx = new RESTContextImpl()
                .setRequest(request)
                .setCollection(collection)
                .setRestQuery(query);
            
            restAdapter.getResource(ctx, collection, id);
            
            Entry entry = createAtomEntry(ctx.getResource(), baseUri, options);
            
            if (query.isSet(RESTQuery.QUERY_FEED)) {
                return convertEntryToFeed(ctx, entry);
            }

            return okDocResponse(entry, entry.getUpdated());
            
        } catch (RESTException e) {
            return errorResponse(e);

        } catch (Throwable t) {
            return errorResponse(request, t);
        }
    }




    public ResponseContext postEntry(RequestContext request) {

        try {
            LOG.fine("Rest Adapter : handling POST");
            Target target = request.getTarget();
            String baseUri = getBaseUri(request);
            String collection = getCollection(request);
            String id = target.getParameter("id");
            String subcollection = target.getParameter("subcollection");

            RESTQuery query = new RESTQueryImpl(target.getParameter("query"));

            // TODO: return 405 for POST on a collection
            LOG.fine("Rest Adapter : collection >" + collection + "<");
            LOG.fine("Rest Adapter : subcollection >" + subcollection + "<");
            
            AtomEntryOptions options = DEFAULT_ENTRY_OPTIONS;
            if (query.isSet(RESTQuery.QUERY_MIN)) {
                options = MINIMAL_ENTRY_OPTIONS;
            }
            
            Feed queryFeed = getAtomFeedFromRequest(request);
            if (queryFeed != null) {
                
                List<Resource> resources = new ArrayList<Resource>(queryFeed.getEntries().size());
                for (Entry entry : queryFeed.getEntries()) {
                    Resource resource = AbderaUtil.getResource(entry);
                    if (resource == null) {
                        throw new RESTException(RESTException.BAD_REQUEST,
                                                new Message(BUNDLE, "RESOURCE_NOT_IN_ENTRY"));
                    }
                    resources.add(resource);
                }
                
                RESTContext ctx = new RESTContextImpl().setRequest(request)
                                                       .setResourceList(resources)
                                                       .setCollection(collection)
                                                       .setRestQuery(query);

                restAdapter.createResources(ctx, collection);
            
                
                Feed responseFeed = feedFactory.createFeed(baseUri, collection, ctx.getCollectionTitle());
                /*
                Comparator<? super Resource> comparator = 
                    getComparator(query.getParameter(RESTQuery.QUERY_ORDERBY),
                                  query.getParameter(RESTQuery.QUERY_ASCENDING));
                if (comparator != null) {
                    Collections.sort(ctx.getResourceList(), comparator);
                }
                */
                Date last =
                    addResourcesToFeed(request, responseFeed, ctx, baseUri, collection, options, query);
                responseFeed.setUpdated(last);
                
                String location = baseUri + responseFeed.getLink("self").getHref().toString();
                return okDocResponse(responseFeed, responseFeed.getUpdated(), 201).addHeader("Location",
                                                                                             location);
                
            } else {
                Entry entry = getAtomEntry(request);
    
                Resource resource = AbderaUtil.getResource(entry);
                if (resource == null) {
                    throw new RESTException(RESTException.BAD_REQUEST,
                                            new Message(BUNDLE, "RESOURCE_NOT_IN_ENTRY"));
                }
    
                RESTContext ctx = new RESTContextImpl().setRequest(request)
                                                       .setResource(resource)
                                                       .setCollection(collection)
                                                       .setRestQuery(query);
    
                if (id == null && subcollection == null) {
                    restAdapter.createResource(ctx, collection);
                    
                } else if (id != null && subcollection != null) {
                    LOG.fine("Rest Adapter : need to create sub collection obj");
                    restAdapter.createSubcollectionResource(ctx, collection, id, subcollection);
                    
                } else {
                    throw new RESTException(RESTException.BAD_REQUEST,
                                            request,
                                            new Message(BUNDLE, 
                                                        "BAD_SUBCOLLECTION",
                                                        request.getMethod(),
                                                        request.getResolvedUri()));
                }
                
                
                entry = createAtomEntry(ctx.getResource(), baseUri, options);
             
                // TODO: Content-Location headers?
                String location = baseUri + entry.getLink("self").getHref().toString();
                return okDocResponse(entry, entry.getUpdated(), 201).addHeader("Location", location);
            }            
            


        } catch (RESTException e) {
            return errorResponse(e);

        } catch (Throwable t) {
            return errorResponse(request, t);
        }
    }


    public ResponseContext putEntry(RequestContext request) {

        try {

            String baseUri = getBaseUri(request);
            String collection = getCollection(request);
            String id = getResourceId(request);

            Target target = request.getTarget();
            RESTQuery query = new RESTQueryImpl(target.getParameter("query"));
            
            if (query.isSet(RESTQuery.QUERY_MEDIA)) {
                return putMedia(request);
            }
            
            AtomEntryOptions options = DEFAULT_ENTRY_OPTIONS;
            if (query.isSet(RESTQuery.QUERY_MIN)) {
                options = MINIMAL_ENTRY_OPTIONS;
            }
           
            
            Entry entry = getAtomEntry(request);
            Resource resource = AbderaUtil.getResource(entry);
            if (resource == null) {
                throw new RESTException(
                        RESTException.BAD_REQUEST,
                        new Message(BUNDLE, "RESOURCE_NOT_IN_ENTRY"));
            }

            RESTContext ctx = new RESTContextImpl()
                .setRequest(request)
                .setResource(resource)
                .setCollection(collection)
                .setRestQuery(query);
            
            restAdapter.updateResource(ctx, collection, id);

            entry = createAtomEntry(ctx.getResource(), baseUri, options);

            return okDocResponse(entry, entry.getUpdated());

        } catch (RESTException e) {
            return errorResponse(e);

        } catch (Throwable t) {
            return errorResponse(request, t);
        }
    }
    
    @Override
    public ResponseContext getMedia(RequestContext request) {
        
        try {
            String collection = getCollection(request);
            String id = getResourceId(request);

            RESTContext ctx = new RESTContextImpl()
                .setRequest(request)
                .setCollection(collection);
            
            restAdapter.getMediaResource(ctx, collection, id);

            InputStream is = ctx.getInputStream();
            //EntityTag etag = EntityTag.generate("foo");
            ResponseContext response = new MediaResponseContext(is, 200)
                .addHeader("Content-Type", ctx.getContentType());

            // TODO: Add cache, location and entity tag headers
            return response;
            
        } catch (RESTException e) {
            return errorResponse(e);

        } catch (Throwable t) {
            return errorResponse(request, t);
        }
    }

    @Override
    public ResponseContext postMedia(RequestContext request) {
        
        try {
            String baseUri = getBaseUri(request);
            String collection = getCollection(request);
            RESTContext ctx = new RESTContextImpl()
                .setRequest(request)
                .setCollection(collection);

            Target target = request.getTarget();
            RESTQuery query = new RESTQueryImpl(target.getParameter("query"));

            AtomEntryOptions options = DEFAULT_ENTRY_OPTIONS;
            if (query.isSet(RESTQuery.QUERY_MIN)) {
                options = MINIMAL_ENTRY_OPTIONS;
            }
            
            MimeType mimeType = request.getContentType();
            LOG.info("Content-Type: " + mimeType);

            // If there is no content type assume we just want to create
            // a default resource. 
            if (mimeType == null) {
                ctx.setResource(null);
                
                LOG.info("No Content-Type; assuming empty resource");
                
                restAdapter.createResource(ctx, collection);

                Entry entry = createAtomEntry(ctx.getResource(), baseUri, options);

                return okDocResponse(entry, entry.getUpdated(), 201);
                
            } else {
                
                ctx.setInputStream(request.getInputStream());                                
                
                boolean isFormData = false;
                if (mimeType != null) {
                    ctx.setContentType(mimeType.toString());
                    if (mimeType.getPrimaryType().equals("multipart")) {
                        processMultipart(ctx, mimeType);
                        isFormData = true;
                    }
                }
                
                restAdapter.createMediaResource(ctx, collection);
                
                if (isFormData) {
                    // Just return text content that is the URI of the resource.
                    String uri = baseUri + collection + "/" + ctx.getResource().getId();
                    LOG.info("Returning: " + uri);
                    return textResponse(request, 201, uri);
                    
                } else {

                    Entry entry = createAtomEntry(ctx.getResource(), baseUri, options);

                    // TODO: Content-Location headers?
                    String location = baseUri + entry.getLink("self").getHref().toString();
                    return okDocResponse(entry, entry.getUpdated(), 201)
                        .addHeader("Location", location);
                }
            }


        } catch (RESTException e) {
            return errorResponse(e);

        } catch (Throwable t) {
            return errorResponse(request, t);
        }
    }





    @Override
    public ResponseContext putMedia(RequestContext request) {

        try {
            String baseUri = getBaseUri(request);
            String collection = getCollection(request);
            String id = getResourceId(request);
            RESTContext ctx = new RESTContextImpl()
                .setRequest(request)
                .setCollection(collection);

            Target target = request.getTarget();
            RESTQuery query = new RESTQueryImpl(target.getParameter("query"));

            AtomEntryOptions options = DEFAULT_ENTRY_OPTIONS;
            if (query.isSet(RESTQuery.QUERY_MIN)) {
                options = MINIMAL_ENTRY_OPTIONS;
            }
           
            ctx.setInputStream(request.getInputStream());
            MimeType mimeType = request.getContentType();
            if (mimeType != null) {
                ctx.setContentType(mimeType.toString());
            }
            restAdapter.updateMediaResource(ctx, collection, id);
            
            Entry entry = createAtomEntry(ctx.getResource(), baseUri, options);

            // TODO: Content-Location headers?
            String location = baseUri + entry.getLink("self").getHref().toString();
            return okDocResponse(entry, entry.getUpdated())
                   .addHeader("Location", location);

        } catch (RESTException e) {
            return errorResponse(e);

        } catch (Throwable t) {
            return errorResponse(request, t);
        }
    }


    public ResponseContext deleteEntry(RequestContext request) {

        try {
            String collection = getCollection(request);
            String id = getResourceId(request);

            RESTContext ctx = new RESTContextImpl().setRequest(request);
            restAdapter.deleteResource(ctx, collection, id);

            return emptyResponse(request,
                    new Message(BUNDLE, 
                            "RESOURCE_DELETED", 
                            collection, id));

        } catch (RESTException e) {
            return errorResponse(e);

        } catch (Throwable t) {
            return errorResponse(request, t);
        }
    }
    
    @SuppressWarnings("unchecked")
    private void processMultipart(RESTContext ctx, MimeType mimeType) 
        throws RESTException {
        
        // The GWT web client uses a form post to upload documents.  Need to extract
        // the content from a multipart message.

        LOG.info("multipart message, mime type: " + mimeType);
        if (!mimeType.getSubType().equals("form-data")) {
            throw new RESTException(RESTException.BAD_REQUEST,
                    new Message(BUNDLE, "MULTIPART_NOT_SUPPORTED"));
        }
        
        
        try {
            LOG.info("multipart message, mime type: " + mimeType);
            String boundary = mimeType.getParameter("boundary");

            // Abdera has consumed part of the messasge so this must be reconstructed.
            
            String header = ""
                + "Content-Type: multipart/form-data; boundary=" + boundary + "\r\n"
                + "\r\n";
            InputStream headerInputStream = new ByteArrayInputStream(header.getBytes());
            SequenceInputStream is = new SequenceInputStream(headerInputStream, ctx.getInputStream());

            MimeMessage msg = new MimeMessage(null, is);
            msg.setHeader("Content-Type", mimeType.toString());
            MimeMultipart mp = (MimeMultipart) msg.getContent();
            
            
            if (mp.getCount() == 0) {
                throw new RESTException(RESTException.BAD_REQUEST,
                        new Message(BUNDLE, "NO_PARTS"));            
            }

            boolean foundContent = false;
            
            for (int i = 0; i < mp.getCount(); i++) {
                BodyPart part = mp.getBodyPart(i);
                
                if (LOG.isLoggable(Level.FINE)) {
                    Enumeration e = part.getAllHeaders();
                    while (e.hasMoreElements()) {
                        Header h = (Header) e.nextElement();
                        LOG.fine("Header - " + h.getName() + " = " + h.getValue());
                    }
                }
                String name = getPartDispositionName(part);
                LOG.info("Part disposition name " + name);
                
                if ("artifactFile".equals(name)) {
                    ctx.setContentType(part.getContentType());
                    ctx.setInputStream(part.getInputStream());
                    foundContent = true;
                } else if ("simpleFileName".equals(name)) {
                    ctx.setSlug(readStringContent(part.getInputStream()));
                } else {
                    LOG.warning("Ignoring message part " + name);
                }
            }
 
            if (!foundContent) {
                throw new RESTException(RESTException.BAD_REQUEST,
                        new Message(BUNDLE, "MULTIPART_MISSING_CONTENT"));
            }
            
                        
        } catch (Exception e) {
            throw new RESTException(RESTException.INTERNAL_SERVER_ERROR, e);
        }
    }


    private String getPartDispositionName(BodyPart part) {
        try {
            // NOTE: no way of extracting "name" paramater from BodyPart.
            String headers[] = part.getHeader("Content-Disposition");
            if (headers.length == 0) {
                return null;
            } else {
                // TODO: very basic parsing!
                String name = headers[0];
                int i = name.indexOf("name=\"");
                if (i == -1) {
                    return null;
                }
                name = name.substring(i + 6);
                i = name.indexOf("\"");
                if (i == -1) {
                    return null;
                }
                name = name.substring(0, i);
                return name;
            }
        } catch (MessagingException e) {
            e.printStackTrace();
        }

        return null;
    }


    private Date addResourcesToFeed(RequestContext request,
                                    Feed feed, 
                                    RESTContext ctx,
                                    String baseUri, 
                                    String path,
                                    AtomEntryOptions options, 
                                    RESTQuery query) throws RESTException {

        String relHref = getRelativeUri(request, baseUri);
        feed.addLink(relHref, "self").setMimeType(Constants.ATOM_CONTENT_TYPE);
        if (ctx.isEditable()) {
            feed.addLink(relHref, "create").setMimeType(Constants.ATOM_CONTENT_TYPE);
        }
        Date youngestResource = OLTEST_DATE;

        int count = query.getIntParameter(RESTQuery.QUERY_COUNT, -1);
        if (count == 0) {
            throw new RESTException(RESTException.BAD_REQUEST, new Message(BUNDLE, "ZERO_COUNT"));
        }
        
        List<Resource> resources = ctx.getResourceList();
        int page = 0;
        if (count < 0) {
            for (Resource resource : resources) {
                Entry entry = createAtomEntry(resource, baseUri, options);
                feed.addEntry(entry);
                if (entry.getUpdated().after(youngestResource)) {
                    youngestResource = entry.getUpdated();
                }
            }
        } else {

            int size = resources.size();
            page = query.getIntParameter(RESTQuery.QUERY_PAGE, 0);
            int base = count * page;
            
            // Add OpenSearch elements (should this always be done and not 
            // just for paging).
            Factory abderaFactory = AbderaUtil.getAbderaFactory();
            Element e = abderaFactory.newElement(OpenSearchConstants.TOTAL_RESULTS);
            e.setText(Integer.toString(size));
            feed.addExtension(e);

            e = abderaFactory.newElement(OpenSearchConstants.ITEMS_PER_PAGE);
            e.setText(Integer.toString(count));
            feed.addExtension(e);
            
            e = abderaFactory.newElement(OpenSearchConstants.START_INDEX);
            e.setText(Integer.toString(base));
            feed.addExtension(e);
            // TODO: OpenSearch.QUERY
           
            LOG.fine("count = " + count + ", page = " + page + ", size " + size + ", base " + base);
            
            for (int i = 0; i < count; i++) {
                if (base + i >= size) {
                    break;
                }
                Resource r = resources.get(base + i);
                Entry entry = createAtomEntry(r, baseUri, options);
                feed.addEntry(entry);
                if (entry.getUpdated().after(youngestResource)) {
                    youngestResource = entry.getUpdated();
                }
            }

            
            // Add paging links.
            StringBuilder sb = new StringBuilder().append(path).append("?");
            for (String name : query.getParameterNames()) {
                if (!RESTQuery.QUERY_PAGE.equals(name)) {
                    sb.append(name).append("=").append(query.getParameter(name)).append("&");
                }
            }
            String baseuri = sb.toString();

            feed.addLink(baseuri + "_page=0", "first")
                .setMimeType(Constants.ATOM_FEED_CONTENT_TYPE);

            int last = size / count;               
            feed.addLink(baseuri + "_page=" + last, "last")
                .setMimeType(Constants.ATOM_FEED_CONTENT_TYPE);

            if (page > 0) {
                feed.addLink(baseuri + "_page=" + (page - 1), "previous")
                    .setMimeType(Constants.ATOM_FEED_CONTENT_TYPE);                
            }

            if (page < last) {
                feed.addLink(baseuri + "_page=" + (page + 1), "next")
                    .setMimeType(Constants.ATOM_FEED_CONTENT_TYPE);                
            }           
        }        
        return youngestResource;
    }


    private Entry createAtomEntry(Resource resource, String baseUri, 
                                  AtomEntryOptions options) {

        Entry entry = entryFactory.createEntry(resource, 
                                               baseUri, 
                                               resource.getCollection(),
                                               options);
      
        Object contentType = resource.getProperty("MediaContentType");
        Object uri = resource.getProperty("MediaUri");
        if (contentType != null && uri != null) {
            entry.addLink(uri.toString(), "edit-media").setMimeType(contentType.toString());
        }
        
        Link selfLink = entry.getSelfLink();
        if (selfLink != null) {
            entry.addLink(entry.getSelfLink().getHref() + "?" + RESTQuery.QUERY_FEED, 
                          Constants.FEED_LINK_REL).setMimeType(Constants.ATOM_FEED_CONTENT_TYPE);
        }
        return entry;
    }
    

    private ResponseContext convertEntryToFeed(RESTContext ctx, Entry entry) {

        // Convert an Atom entry into an Atom feed.
        
        Resource resource = ctx.getResource();
        String title = resource.getName() + " (" + ctx.getCollection() + ")";
        Feed feed = feedFactory.createFeed(ctx.getBaseUri(), 
                                           resource.getCollection() + "/" + resource.getId(),  
                                           title);
        feed.addEntry(entry);
        feed.setUpdated(entry.getEdited());
        
        return okDocResponse(feed, feed.getUpdated());
    }    
    
    private Entry getAtomEntry(RequestContext request) throws RESTException {
        try {
            Document<Entry> entryDoc = request.getDocument(abdera.getParser());
            if (entryDoc == null || !"entry".equals(entryDoc.getRoot().getQName().getLocalPart())) {
                throw new RESTException(RESTException.BAD_REQUEST,
                                        request,
                                        new Message(BUNDLE, "ATOM_ENTRY_MISSING"));
            }
            return entryDoc.getRoot();

        } catch (Exception e) {
            throw new RESTException(RESTException.INTERNAL_SERVER_ERROR, request, e);
        }
    }

    
    private Feed getAtomFeedFromRequest(RequestContext request) throws RESTException {
        try {
            Document<Feed> feedDoc = request.getDocument(abdera.getParser());
            if (feedDoc != null && "feed".equals(request.getDocument().getRoot().getQName().getLocalPart())) {
                return feedDoc.getRoot();
            }
        } catch (Exception e) {
            throw new RESTException(RESTException.INTERNAL_SERVER_ERROR, request, e);
        }
        
        return null;
    }

    private String getResourceId(RequestContext request) 
        throws RESTException {

        Target target = request.getTarget();
        String id = target.getParameter("id");
        if (id == null || "".equals(id)) {
            throw new RESTException(
                    RESTException.BAD_REQUEST,
                    request,
                    new Message(BUNDLE, 
                            "MISSING_RESOURCE_ID", 
                            request.getMethod(),
                            request.getResolvedUri()));
        }
        return id;
    }

    private String getCollection(RequestContext request) 
        throws RESTException {

        Target target = request.getTarget();
        String collection = target.getParameter("collection");

        if (collection == null || "".equals(collection)) {
            throw new RESTException(
                    RESTException.BAD_REQUEST,
                    request,
                    new Message(BUNDLE, 
                            "NO_COLLECTION", 
                            request.getMethod(),
                            request.getResolvedUri()));
        }
        return collection;
    }

    private Comparator<? super Resource> getComparator(String orderBy, String ascending) {
        return new ResourceComparator(orderBy == null ? DEFAULT_ORDERING_ATTRIBUTE : orderBy,
                                      ascending != null);
    }


    private String getBaseUri(RequestContext request) {
        return request.getBaseUri().toString() + basePath;
    }
    
    private String getRelativeUri(RequestContext request, String baseUri) {
        String relUri = request.getUri().toString();
        if (relUri.startsWith(fullBasePath)) {
            relUri = relUri.substring(fullBasePath.length());
        }
        return relUri;
    }



    private ResponseContext okDocResponse(Base base, Date updated) {
        return okDocResponse(base, updated, 200);
    }

    private ResponseContext okDocResponse(Base base, Date updated, int code) {
        
        return ProviderHelper.returnBase(base,
                                         code,
                                         updated)
            .setEntityTag(ProviderHelper.calculateEntityTag(base))
            .addHeader("Expires", "0")
            .addHeaders("Cache-Control", "max-age=0", "no-store");
        // TODO: need to refine these cache control/expires headers
        // (required for the GWT client).
    }
    

    private ResponseContext emptyResponse(RequestContext request, Message msg) {
        return new EmptyResponseContext(204, msg.toString());
    }

    private ResponseContext textResponse(RequestContext request, int code, String text) {
        
        // NOTE: content-type is text/html to keep GWT happy.
        LOG.fine("Content: " + text);
        InputStream is = new ByteArrayInputStream(text.getBytes());
        ResponseContext response = new MediaResponseContext(is, code)
            .addHeader("Content-Type", "text/html");
        return response;
    }


    private ResponseContext errorResponse(RESTException ex) {

        Level logLevel = ex.isServerError() ? Level.SEVERE : Level.WARNING;
        String prefix = ex.isServerError() ? "Server Error" : "Client Error";
        
        if (LOG.isLoggable(Level.FINE)) {
            LOG.fine(ExceptionHelper.getStackTraceAsString(ex));
        }

        String msg = ex.getMessage();
        if (msg == null) {
            msg = "";
        }
        int code = ex.getCode();
        LOG.log(logLevel, prefix + " " + code + ": " + msg);

        Throwable cause = ex.getCause();
        if (cause != null) {
            LOG.log(logLevel, "Cause: ", cause);
        }
        
        return ProviderHelper.createErrorResponse(abdera, code, msg);
    }

    private ResponseContext errorResponse(RequestContext request,  
            Throwable t) {
        return errorResponse(
                new RESTException(RESTException.INTERNAL_SERVER_ERROR,
                        request,
                        t));        
    }    

    private String readStringContent(InputStream is) throws Exception { 
        BufferedReader br = new BufferedReader(new InputStreamReader(is));

        StringBuilder sb = new StringBuilder();
        String line = null;
        line = br.readLine();
        while (line != null) {
            sb.append(line);
            line = br.readLine();
        }
        br.close();
        return sb.toString();
    }

}
