/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.rest.rt.server;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.logging.Logger;

import com.iona.repository.rest.RESTException;
import com.iona.repository.rest.atom.AtomEntryFactory;
import com.iona.repository.rest.rt.atom.AtomFeedFactory;
import com.iona.repository.rest.rt.osgi.MessageBundle;
import com.iona.repository.util.i18n.Message;
import com.iona.soa.security.SecurityService;

import org.apache.abdera.Abdera;
import org.apache.abdera.protocol.server.CollectionAdapter;
import org.apache.abdera.protocol.server.CollectionInfo;
import org.apache.abdera.protocol.server.Provider;
import org.apache.abdera.protocol.server.RequestContext;
import org.apache.abdera.protocol.server.TargetType;
import org.apache.abdera.protocol.server.WorkspaceInfo;
import org.apache.abdera.protocol.server.filters.MethodOverrideFilter;
import org.apache.abdera.protocol.server.impl.AbstractWorkspaceProvider;
import org.apache.abdera.protocol.server.impl.RegexTargetResolver;
import org.apache.abdera.protocol.server.impl.SimpleCollectionInfo;



public class RepositoryProvider extends AbstractWorkspaceProvider implements Provider {

    private static final ResourceBundle BUNDLE = MessageBundle.getInstance();
    private static final Logger LOG = Logger.getLogger(RepositoryProvider.class.getName());



    private RepositoryAdapter adapter;
    private SecurityService securityService;
    private RESTAdapter restAdapter;
    private AtomFeedFactory feedFactory;
    private AtomEntryFactory entryFactory;
    private String path;
    private String adjustedPath;

    public RepositoryProvider(RESTAdapter restAdapter,
                              SecurityService securityService,
                              AtomFeedFactory feedFactory,
                              AtomEntryFactory entryFactory,
                              String path) {
        
        this.restAdapter = restAdapter;
        this.securityService = securityService;
        this.feedFactory = feedFactory;
        this.entryFactory = entryFactory;
        this.path = path;
        
        // Move the leading "/" to the end of the path
        adjustedPath = path.substring(1) + "/";
    }


    @Override
    public void init(Abdera abdera, Map<String, String> properties) {
        super.init(abdera, properties);

        LOG.info("Atom servlet path: " + path);
        
        // The target patterns are as follows:
        //
        //  <base>/<collection>
        //  <base>/<collection>/<id>
        //  <base>/<collection>/<id>;<sub-collection>
        //

        setTargetResolver(new RegexTargetResolver()
            .setPattern(path + "/?(\\?[^#]*)?", 
                        TargetType.TYPE_SERVICE)
            .setPattern(path + "/([^/#?;]+)", 
                        TargetType.TYPE_COLLECTION, 
                        "collection")
            .setPattern(path + "/([^/#?;]+)\\?(.*)", 
                        TargetType.TYPE_COLLECTION, 
                        "collection", "query")
            .setPattern(path + "/([^/#?;]+)/([^/#?;]+)", 
                        TargetType.TYPE_ENTRY,
                        "collection", "id")
            .setPattern(path + "/([^/#?;]+)/([^/#?;]+)\\?(.*)", 
                        TargetType.TYPE_ENTRY,
                        "collection", "id", "query")
            .setPattern(path + "/([^/#?;]+)/([^/#?;]+);([^/#?;]+)",
                        TargetType.TYPE_COLLECTION, 
                        "collection", "id", "subcollection")
            .setPattern(path + "/([^/#?;]+)/([^/#?;]+);([^/#?;]+)\\?(.*)",
                        TargetType.TYPE_COLLECTION, 
                        "collection", "id", "subcollection", "query")
        );

       // Create adapter for handling all requests.
        adapter = new RepositoryAdapter(restAdapter,
                                        feedFactory,
                                        entryFactory,
                                        adjustedPath);


        // Add support for Overriding POST for PUT & DELETE
        addFilter(new MethodOverrideFilter());
        addFilter(new AuthFilter(securityService));
        addFilter(new LoggingFilter());

        // Output the collections available.
        StringBuilder sb = new StringBuilder();
        for (String name : restAdapter.getCollections()) {
            sb.append("\n    ").append(name).append("   [");
            String[] accepts = null;
            try {
                accepts = restAdapter.getCollectionAccepts(name);
            } catch (RESTException e) {
                // Complete
            }
            boolean first = true;
            for (String a : accepts) {
                if (!first) {
                    sb.append(", ");
                }
                sb.append(a);
                first = false;
            }
            sb.append("]");
        }
        LOG.info("Atom Collections available: " + sb.toString());

    }

    public CollectionAdapter getCollectionAdapter(RequestContext request) {
        return adapter;
    }
        
    @Override
    public Collection<WorkspaceInfo> getWorkspaces(RequestContext request) {
        List<WorkspaceInfo> workspaces = new ArrayList<WorkspaceInfo>();
        
        RepositoryWorkspaceInfo workspace = new RepositoryWorkspaceInfo();
        workspace.setTitle(new Message(BUNDLE, "WORKSPACE_TITLE").toString());
        
        for (String name : restAdapter.getCollections()) {
            String[] accepts = null;
            try {
                accepts = restAdapter.getCollectionAccepts(name);
            } catch (RESTException e) {
                // Complete
            }   
            CollectionInfo info = createCollectionInfo(request.getBaseUri().toString(),
                    name, accepts);
            workspace.addCollection(info);
            LOG.info("Added \"" + info.getTitle(null)
                    + "\" " + info.getHref(null));

        }
        workspaces.add(workspace);
        return workspaces;
    }
    
    private CollectionInfo createCollectionInfo(String baseUri, String name, String... accepts) {
        return new SimpleCollectionInfo(name, 
                                        baseUri + adjustedPath + name,
                                        accepts);
    }

}
