/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.search;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;

import com.iona.soa.model.repository.RepositoryPackage;
import com.iona.soa.model.scheduling.SchedulingPackage;
import com.iona.soa.model.security.SecurityPackage;
import com.iona.soa.repository.EMFUtil;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.status.RepositoryException;
import com.iona.soa.repository.util.ExceptionHelper;
import com.iona.soa.resourceset.ResourceSetQueryHandler;

import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.Term;
import org.eclipse.emf.common.util.EMap;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAnnotation;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.util.EcoreUtil;

public class EObjectSearchIndex extends SearchIndex {
    private static final String ID_FIELD = "id";
    private static final String URI_FIELD = "uri";
    private static final Collection<EAttribute> UNINDEXED_ATTRS;

    private SOARepository repository;
    private boolean indexRepo;

    static {
        UNINDEXED_ATTRS = new ArrayList<EAttribute>();
        UNINDEXED_ATTRS.add(RepositoryPackage.eINSTANCE.getIPersistableObject_GenerationNumber());
        UNINDEXED_ATTRS.add(RepositoryPackage.eINSTANCE.getIPersistableObject_Guid());
        UNINDEXED_ATTRS.add(RepositoryPackage.eINSTANCE.getIPersistableObject_Internal());
        UNINDEXED_ATTRS.add(RepositoryPackage.eINSTANCE.getIPersistableObject_Modifiable());
        UNINDEXED_ATTRS.add(SecurityPackage.eINSTANCE.getUser_Password());
        UNINDEXED_ATTRS.add(SchedulingPackage.eINSTANCE.getScheduledTask_Password());
    }

    public EObjectSearchIndex(String indexDir, SOARepository repo) throws IOException {
        super(indexDir);
        repository = repo;
        if (indexRepo) {
            IndexWriter iwriter = openIndexWriter();
            try {
                for (EClass eClass : repository.getModelRegistry().getAllEClasses()) {
                    if (!eClass.isAbstract() && eClass.getEIDAttribute() != null) {
                        try {
                            for (EObject eObj : ResourceSetQueryHandler.executeQuery(
                                    repository.getResourceSet(), eClass, true)) {
                                Document doc = eObjtoDoc(eObj);
                                iwriter.addDocument(doc);
                            }
                        } catch (Exception e) {
                            LOG.log(Level.WARNING, "Exception adding eclass " + eClass + " to index", e);
                        }
                    }
                }
            } finally {
                closeIndexWriter();
            }
        }
    }

    public List<EObject> search(String query) throws RepositoryException {
        if (query == null) {
            LOG.warning("<null> search query not supported");
            return null;
        }
        Iterator<Document> docs = null;
        try {
            LOG.log(Level.FINE, "Querying being passed executed: {0}", query);
            docs = queryIndex(query);
        } catch (Exception e) {
            LOG.warning("Lucene query for '" + query + "' aucsed an error: " + e);
            LOG.fine(ExceptionHelper.getStackTraceAsString(e));
            return null;
        }
        if (docs == null) {
            return null;
        }
        List<EObject> result = new LinkedList<EObject>();
        while (docs.hasNext()) {
            Document doc = docs.next();
            URI uri = URI.createURI(doc.get("uri"));
            EObject theObject = repository.getEObject(uri, true);
            if (theObject != null && !result.contains(theObject)) {
                result.add(theObject);
            }
        }
        return result;
    }

    public void addObject(EObject eObj) {
        if (eObj == null) {
            return;
        }
        LOG.log(Level.FINE, "Adding EObject in index: {0}", eObj);
        Document doc = eObjtoDoc(eObj);
        try {
            addToIndex(doc);
        } catch (Exception ioe) {
            LOG.warning("Error updating object " + EMFUtil.getLongName(eObj) + " to index: " + ioe);
            LOG.fine(ExceptionHelper.getStackTraceAsString(ioe));
        }
    }

    public void updateObject(EObject eObj) {
        if (eObj == null) {
            return;
        }
        LOG.log(Level.FINE, "Updating EObject in index: {0}", eObj);
        Document doc = eObjtoDoc(eObj);
        Term term = new Term(ID_FIELD, EMFUtil.getGuid(eObj));
        try {
            IndexWriter iwriter = openIndexWriter();
            try {
                iwriter.updateDocument(term, doc);
            } finally {
                closeIndexWriter();
            }
        } catch (Exception ioe) {
            LOG.warning("Error adding object " + EMFUtil.getLongName(eObj) + " to index: " + ioe);
            LOG.fine(ExceptionHelper.getStackTraceAsString(ioe));
        }
    }

    public void deleteObject(String guid) {
        if (guid == null) {
            return;
        }
        LOG.log(Level.FINE, "Deleting EObject from index: {0}", guid);
        Term term = new Term(ID_FIELD, guid);
        try {
            removeFromIndex(term);
        } catch (Exception ioe) {
            LOG.warning("Error removing object " + guid + " from index: " + ioe);
            LOG.fine(ExceptionHelper.getStackTraceAsString(ioe));
        }
    }

    public SOARepository getRepository() {
        return repository;
    }
    
    @Override
    protected void initialize() throws IOException {
        super.initialize();
        indexRepo = true;
    }

    @SuppressWarnings("unchecked")
    private Document eObjtoDoc(EObject eObj) {
        EClass cls = eObj.eClass();
        Document doc = new Document();
        doc.add(new Field(ID_FIELD, EMFUtil.getGuid(eObj),
                          Field.Store.YES, Field.Index.NOT_ANALYZED));
        doc.add(new Field(URI_FIELD, EMFUtil.getProxyURI(eObj).toString(),
                Field.Store.YES, Field.Index.NOT_ANALYZED));

        for (EAttribute attr : cls.getEAllAttributes()) {
            if (attr.equals(cls.getEIDAttribute()) || UNINDEXED_ATTRS.contains(attr) || !eObj.eIsSet(attr)) {
                continue;
            }
            String attrName = attr.getName();
            Collection<?> vals;
            if (attr.isMany()) {
                vals = (Collection<?>)eObj.eGet(attr);
            } else {
                vals = Collections.singletonList(eObj.eGet(attr));
            }
            for (Object val : vals) {
                if (val != null) {
                    String strVal = EcoreUtil.convertToString((EDataType)attr.getEType(), val);
                    LOG.log(Level.FINE, "adding this to index {0} value: {1}",
                            new Object[] {attrName, strVal});
                    doc.add(new Field(attrName, strVal, Field.Store.NO, Field.Index.ANALYZED));
                }
            }
        }

        EAnnotation ann = eObj.eClass().getEAnnotation("searchHints");
        if (ann != null) {
            EMap<String, String> annotationDetails = ann.getDetails();
            for (String referenceName : annotationDetails.keySet()) {
                EStructuralFeature reference = eObj.eClass().getEStructuralFeature(referenceName);
                if (reference instanceof EReference) {
                    Collection<EObject> objs;
                    if (reference.isMany()) {
                        objs = (Collection<EObject>)eObj.eGet(reference);
                    } else {
                        objs = Collections.singletonList((EObject)eObj.eGet(reference));
                    }
                    for (EObject referred : objs) {
                        String attributeName = (String) annotationDetails.get(referenceName);
                        EStructuralFeature attr = referred.eClass().getEStructuralFeature(attributeName);
                        String attributeValue = null;
                        if (attr instanceof EAttribute) {
                            attributeValue = referred.eGet(attr).toString();
                        } else {
                            LOG.log(Level.WARNING,
                                    "Attribute {0} found in searchHints annotation does not exist in {1}",
                                    new Object[] {attributeName, referred});
                        }
                        if (attributeValue != null) {
                            doc.add(new Field(referenceName, attributeValue, Field.Store.NO,
                                              Field.Index.ANALYZED));
                        }
                    }
                } else {
                    LOG.log(Level.WARNING,
                            "Reference {0} found in searchHints annotation does not exist in {1}",
                            new Object[] {referenceName, eObj});
                }
            }
        }
        return doc;
    }
}
