/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.search;

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.logging.Level;
import java.util.logging.Logger;


import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.Term;
import org.apache.lucene.queryParser.MultiFieldQueryParser;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.BooleanQuery;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.TopDocCollector;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory;
import org.apache.lucene.store.NativeFSLockFactory;

/** @see http://wiki.apache.org/jakarta-lucene/LuceneFAQ
 */
public class SearchIndex {
    static final Logger LOG = Logger.getLogger(SearchIndex.class.getName());

    private final Directory indexDir;
    private final Object mutex = new Object();
    private IndexWriter iWriter;
    private int writerUserCount;

    static {
        IndexWriter.setDefaultWriteLockTimeout(5000);

        // TODO - this should change.  It's a fix for REPO-534.
        // There is a unit test in Maven2IndexerTest.
        BooleanQuery.setMaxClauseCount(Integer.MAX_VALUE);
    }

    public SearchIndex(File indexDir) throws IOException {
        if (!indexDir.exists()) {
            indexDir.mkdirs();
        }
        this.indexDir = FSDirectory.getDirectory(indexDir, new NativeFSLockFactory(indexDir));
        if (!isInitialized()) {
            initialize();
        }
    }
    
    public SearchIndex(String indexDir) throws IOException {
        this(new File(indexDir));
    }
    
    public Directory getIndexDir() {
        return indexDir;
    }
    
    public void close() throws IOException {
        try {
            closeIndexWriter(true);
        } finally {
            indexDir.close();
        }
    }
    
    public boolean isInitialized() {
        try {
            String[] list = indexDir.list();
            return list != null && list.length > 0;
        } catch (IOException e) {
            return false;
        }
    }

    @SuppressWarnings("unchecked")
    public Iterator<Document> queryIndex(String queryString) throws IOException, ParseException {
        final IndexReader reader = IndexReader.open(indexDir);
        final IndexSearcher searcher = new IndexSearcher(reader);
        TopDocCollector collector = new TopDocCollector(searcher.maxDoc());
        Collection<String> allFields = reader.getFieldNames(IndexReader.FieldOption.ALL);
        String[] searchFields = allFields.toArray(new String[0]);
        final ScoreDoc[] hits;
        Query query;
        if (queryString == null || queryString.length() == 0) {
            Collection<Document> ret = Collections.emptyList();
            return ret.iterator();
        } else if (queryString.contains(":")) {
            QueryParser parser = new QueryParser("id", new StandardAnalyzer());
            query = parser.parse(queryString);
        } else {
            MultiFieldQueryParser parser = new MultiFieldQueryParser(searchFields, new StandardAnalyzer());
            parser.setAllowLeadingWildcard(true);
            parser.setDefaultOperator(QueryParser.Operator.OR); 
            query = parser.parse(queryString);
        }            
        LOG.log(Level.FINE, "query [" + query.toString() + "]");     
        searcher.search(query, collector);
        hits = collector.topDocs().scoreDocs;
        return new Iterator<Document>() {
            private int index;
            public boolean hasNext() {
                return index < hits.length;
            }
            public Document next() {
                Document ret = null;
                if (hasNext()) {
                    try {
                        ret = searcher.doc(hits[index++].doc);
                        if (!hasNext()) {
                            try { 
                                searcher.close(); 
                            } finally { 
                                reader.close();
                            }
                        }
                    } catch (IOException ioe) {
                        LOG.log(Level.WARNING, "Error retrieving document from lucene index", ioe);
                    }
                }
                return ret;
            }
            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }
    
    public void addToIndex(Document doc) throws IOException {
        IndexWriter iwriter = openIndexWriter();
        try {
            iwriter.addDocument(doc);
        } finally {
            closeIndexWriter();
        }
    }
    
    public void addToIndex(Iterator<Document> docIter) throws IOException {
        IndexWriter iwriter = openIndexWriter();
        try {
            while (docIter.hasNext()) {
                iwriter.addDocument(docIter.next());
            }
        } finally {
            closeIndexWriter();
        }
    }
    
    public void removeFromIndex(Term term) throws IOException {
        IndexWriter iwriter = openIndexWriter();
        try {
            iwriter.deleteDocuments(term);
        } finally {
            closeIndexWriter();
        }
    }
    
    protected void initialize() throws IOException {
        synchronized (mutex) {
            IndexWriter temp = new IndexWriter(indexDir, new StandardAnalyzer(), true, 
                                               IndexWriter.MaxFieldLength.UNLIMITED);
            try {
                temp.optimize();
            } finally {
                temp.close();
            }
            
        }
    }
    
    protected IndexWriter openIndexWriter() throws IOException {
        synchronized (mutex) {
            if (writerUserCount++ == 0) {
                iWriter = new IndexWriter(indexDir, new StandardAnalyzer(), false, 
                                          IndexWriter.MaxFieldLength.UNLIMITED);
            }
        }
        return iWriter;
    }
    
    protected void closeIndexWriter() throws IOException {
        closeIndexWriter(false);
    }
    
    private void closeIndexWriter(boolean force) throws IOException {
        synchronized (mutex) {
            if (iWriter != null && (force || --writerUserCount == 0)) {
                IndexWriter temp = iWriter;
                iWriter = null;
                try {
                    temp.optimize();
                } catch (IOException ioe) {
                    // ignore
                }
                try {
                    temp.close();
                } finally {
                    if (IndexWriter.isLocked(indexDir)) {
                        IndexWriter.unlock(indexDir);
                    }
                }
            }
        }
    }
}
