/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.security.impl;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.repository.util.i18n.Message;
import com.iona.soa.expression.Equals;
import com.iona.soa.expression.Expression;
import com.iona.soa.model.security.SecurityPackage;
import com.iona.soa.model.security.User;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.util.Base64;
import com.iona.soa.repository.util.InputStreamHelper;
import com.iona.soa.repository.util.Tuple;
import com.iona.soa.resourceset.ResourceSetQueryHandler;
import com.iona.soa.security.AuthenticationException;
import com.iona.soa.security.SecurityUtil;

import org.eclipse.emf.ecore.EObject;

public class AuthenticationServiceImpl implements AuthenticationService {
    private static final Logger LOG = 
        Logger.getLogger(AuthenticationServiceImpl.class.getPackage().getName());

    private SOARepository soaRepo;
    private Map<String, Tuple<String, Long>> tokenMap = new HashMap<String, Tuple<String, Long>>();
    private long sessionTimeout = 1000L * 60L * 60L * 3L; // defaults to 3 hours
    private File authFile = new File("../etc/storage/security/authentication.properties");
    
    /* (non-Javadoc)
     * @see com.iona.soa.security.impl.AuthenticationService#setSoaRepo(com.iona.soa.repository.SOARepository)
     */
    public void setSoaRepo(SOARepository soaRepo) {
        this.soaRepo = soaRepo;
    }

    /* (non-Javadoc)
     * @see com.iona.soa.security.impl.AuthenticationService#setConfig(java.util.Map)
     */
    public void setConfig(Map<String, String> config) {
        String timeout = config.get("authentication.token.timeout");
        if (timeout != null) {
            try {
                sessionTimeout = parseTimeoutString(timeout);
            } catch (NumberFormatException pe) {
                LOG.log(Level.WARNING, "Invalid date/time format in config item "
                                       + "'authentication.token.timeout'. Expected 'ddd:hh:mm:ss'. " + pe);
            }
        } else {
            timeout = config.get("authentication.token.timeout.minutes");
            if (timeout != null) {
                try {
                    sessionTimeout = Long.parseLong(timeout) * 60000L;
                } catch (NumberFormatException nfe) {
                    LOG.log(Level.WARNING, "Invalid value in config item "
                                           + "'authentication.token.timeout.minutes'. Expected long value. "
                                           + nfe.toString());
                }
            }
        }
        String val = config.get("security.storage.dir");
        if (val != null) {
            authFile = new File(val, "authentication.properties");
        }
        Properties props = loadProperties(authFile);
        boolean modified = false;
        for (Object key : props.keySet().toArray()) {
            String userId = key.toString();
            try {
                // make user user exists
                getUser(userId);
                String token = props.getProperty(userId);
                if (userId != null && token != null) {
                    tokenMap.put(token, new Tuple<String, Long>(userId, Long.MAX_VALUE));
                }
            } catch (AuthenticationException e) {
                props.remove(key);
                modified = true;
            }
        }
        if (modified) {
            saveProperties(props, authFile);
        }
    }

    /* (non-Javadoc)
     * @see com.iona.soa.security.impl.AuthenticationService#getUserId(java.lang.String)
     */
    public String getUserId(String token) throws AuthenticationException {
        Tuple<String, Long> data = tokenMap.get(token);
        if (data == null) {
            throw new AuthenticationException(new Message(SecurityServiceImpl.BUNDLE, "AUTH_BAD_TOKEN"));
        } else if (System.currentTimeMillis() > data.getRhs()) {
            tokenMap.remove(token);
            throw new AuthenticationException(new Message(SecurityServiceImpl.BUNDLE, "AUTH_BAD_TOKEN"));
        }
        return data.getLhs();
    }
    
    /* (non-Javadoc)
     * @see com.iona.soa.security.impl.AuthenticationService#authenticate(java.lang.String)
     */
    public String authenticate(String token) throws AuthenticationException {
        Tuple<String, Long> data = tokenMap.get(token);
        if (data == null) {
            throw new AuthenticationException(new Message(SecurityServiceImpl.BUNDLE, "AUTH_BAD_TOKEN"));
        } else if (System.currentTimeMillis() > data.getRhs()) {
            tokenMap.remove(token);
            throw new AuthenticationException(new Message(SecurityServiceImpl.BUNDLE, "AUTH_BAD_TOKEN"));
        }
        return token;
    }
    
    /* (non-Javadoc)
     * @see AuthenticationService#authenticatePersistent(java.lang.String, java.lang.String)
     */
    public String authenticatePersistent(String userId, String password) throws AuthenticationException {
        authenticateInternal(userId, password);
        Properties props = loadProperties(authFile);
        if (props.containsKey(userId)) {
            return props.getProperty(userId);
        } else {
            String token = createNewToken(userId, Long.MAX_VALUE);
            props.put(userId, token);
            saveProperties(props, authFile);
            return token;
        }
    }
    
    /* (non-Javadoc)
     * @see com.iona.soa.security.impl.AuthenticationService#authenticate(java.lang.String, java.lang.String)
     */
    public String authenticate(String userId, String password) throws AuthenticationException {
        authenticateInternal(userId, password);
        return createNewToken(userId, System.currentTimeMillis() + sessionTimeout);
    }

    /* (non-Javadoc)
     * @see com.iona.soa.security.impl.AuthenticationService#logout(java.lang.String)
     */
    public boolean logout(String token) {
        Tuple<String, Long> data = tokenMap.remove(token);
        if (data == null) {
            return false;
        } 
        return data.getRhs() > System.currentTimeMillis();
    }
    
    public long getSessionTimeout() {
        return sessionTimeout;
    }
    
    private void authenticateInternal(String userId, String password) throws AuthenticationException {
        User user = getUser(userId);
        if ((user.getPassword() == null || user.getPassword().length() == 0) 
                && (password == null || password.length() == 0)) {
            String msg = "No password set for user '" + userId + "'";
            LOG.fine(msg);
        } else if (!SecurityUtil.calculateDigestAsHex(password).equals(user.getPassword())) {
            String msg = "Incorrect password provided for user '" + userId + "'";
            LOG.fine(msg);
            throw new AuthenticationException(new Message(SecurityServiceImpl.BUNDLE, 
                    "AUTH_FAILED", userId));
        }

        if (user.getRoles().size() == 0) {
            String msg = "User '" + userId + "' provided a valid password " 
                         + "but has no roles assigned, hence the authentication can not be approved. "
                         + "Contact the Administrator for more information.";
            LOG.warning(msg);
            throw new AuthenticationException(new Message(SecurityServiceImpl.BUNDLE, 
                    "AUTH_FAILED", userId));
        }
    }

    private User getUser(String userId) throws AuthenticationException {
        Collection<EObject> users = null;
        try {
            Expression expr = new Equals(SecurityPackage.eINSTANCE.getUser_Userid().getName(), userId);
            users = ResourceSetQueryHandler.executeQuery(soaRepo.getResourceSet(), 
                    SecurityPackage.eINSTANCE.getUser(), 
                    expr, 
                    true);
        } catch (Exception ex) {
            LOG.severe("Error executing a query: " + ex.getMessage());
            throw new AuthenticationException(new Message(SecurityServiceImpl.BUNDLE, 
                    "AUTH_FAILED", userId), ex);
        }
        if (users == null || users.size() == 0) {
            String msg = "UserId '" + userId + "' does not exist."; 
            LOG.fine(msg);
            throw new AuthenticationException(new Message(SecurityServiceImpl.BUNDLE, 
                    "AUTH_FAILED", userId));
        }
        if (users.size() > 1) {
            String msg = "Multiple user with userId '" + userId + "' found."; 
            LOG.warning(msg);
            throw new AuthenticationException(new Message(SecurityServiceImpl.BUNDLE, 
                    "AUTH_FAILED", userId));
        }
        User user = (User)users.iterator().next();
        return user;
    }

    private String createNewToken(String userId, long expiry) {
        byte[] digest = SecurityUtil.calculateDigest(Long.toString(expiry) + userId);
        String token = Base64.encodeBytes(digest);
        tokenMap.put(token, new Tuple<String, Long>(userId, expiry));
        return token;
    }

    private static Properties loadProperties(File propsFile) {
        Properties props = new Properties();
        if (!propsFile.isFile()) {
            return props;
        }
        
        InputStream is = null;
        try {
            is = new BufferedInputStream(new FileInputStream(propsFile));
            props.load(is);
        } catch (IOException e) {
            // ignore
        } finally {
            InputStreamHelper.close(is);
        }
        return props;
    }

    private static void saveProperties(Properties props, File propsFile) {
        propsFile.getParentFile().mkdirs();
        OutputStream os = null;
        try {
            os = new BufferedOutputStream(new FileOutputStream(propsFile));
            props.store(os, null);
        } catch (IOException e) {
            // ignore
        } finally {
            InputStreamHelper.close(os);
        }
    }

    private static long parseTimeoutString(String toStr) {
        String[] fields =  toStr.split(":");
        if (fields.length == 0 || fields.length > 4) {
            throw new NumberFormatException(toStr);
        }
        long ret = 0L;
        int[] multipliers = new int[] {24, 60, 60, 1000};
        int mi = 4 - fields.length;
        for (String field : fields) {
            ret += Integer.parseInt(field);
            ret *= multipliers[mi++];
        }
        return ret;
    }
}
