/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.security.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.soa.expression.Equals;
import com.iona.soa.expression.Expression;
import com.iona.soa.model.repository.IPersistableObject;
import com.iona.soa.model.security.ObjectPermission;
import com.iona.soa.model.security.OperationType;
import com.iona.soa.model.security.Role;
import com.iona.soa.model.security.SecurityArea;
import com.iona.soa.model.security.SecurityPackage;
import com.iona.soa.model.security.User;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.resourceset.ResourceSetQueryHandler;
import com.iona.soa.security.SecurityUtil;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;

public class AuthorizationServiceImpl implements AuthorizationService {
    private static final Logger LOG = Logger.getLogger(AuthorizationServiceImpl.class.getPackage().getName());
    private static final String NO_AUTH = "Not authorized: ";
    
    /**
     * Map between SecurityArea IDs (areaId) and guids  
     */
    private Map<String, String> securityAreas;
    private SOARepository soaRepo;
    
    /* (non-Javadoc)
     * @see com.iona.soa.security.impl.AuthorizationService#setSoaRepo(com.iona.soa.repository.SOARepository)
     */
    public void setSoaRepo(SOARepository soaRepo) {
        this.soaRepo = soaRepo;
    }

    /* (non-Javadoc)
     * @see com.iona.soa.security.impl.AuthorizationService#setConfig(java.util.Map)
     */
    public void setConfig(Map<String, String> config) {
    }

    /* (non-Javadoc)
     * @see AuthorizationService#isAuthorized(java.lang.String, 
     * \                                      com.iona.soa.model.security.OperationType, 
     * \                                      java.util.Collection)
     */
    @SuppressWarnings("unchecked")
    public boolean isAuthorized(String userId, 
                                OperationType operation,
                                Collection<String> areaIds) throws AuthorizationException {
        User user = getUser(userId);
        if (areaIds == null || areaIds.size() == 0) {
            String msg = "Authorization failed: Operation '" + operation 
                         + "' has no security areas.";
            LOG.severe(msg);
            throw new AuthorizationException(msg);
        } else if (areaIds.contains(SecurityUtil.NONE_SECAREA_ID)) {
            return true;
        } else {
            return doDefaultAuthorization(operation, user, areaIds);
        }
    }
    
    /* (non-Javadoc)
     * @see AuthorizationService#isAuthorized(java.lang.String, 
     * \                                      com.iona.soa.model.security.OperationType, 
     * \                                      com.iona.soa.model.security.ObjectPermission)
     */
    @SuppressWarnings("unchecked")
    public boolean isAuthorized(String userId, 
                                OperationType operation,
                                ObjectPermission objPerm) throws AuthorizationException {
        User user = getUser(userId);
        return doObjectAuthorization(operation, user, objPerm);
    }

    private User getUser(String userId) throws AuthorizationException {
        if (soaRepo == null) {
            String msg = NO_AUTH + "SecurityService not initialized";
            LOG.severe(msg);
            throw new AuthorizationException(msg);
        }

        Collection<EObject> users = null;
        try {
            Expression expr = new Equals(SecurityPackage.eINSTANCE.getUser_Userid().getName(), userId);
            users = ResourceSetQueryHandler.executeQuery(soaRepo.getResourceSet(), 
                                                         SecurityPackage.eINSTANCE.getUser(), expr, true);
        } catch (Exception ex) {
            String msg = NO_AUTH + "Unable to retrieve Users " + ex;
            LOG.log(Level.SEVERE, msg, ex);
            throw new AuthorizationException(msg);
        }

        if (users == null || users.size() == 0) {
            String msg = NO_AUTH + "Unable to retrieve User with id " + userId;
            LOG.fine(msg);
            throw new AuthorizationException(msg);
        }
        if (users.size() > 1) {
            String msg = NO_AUTH + "Multiple user with userId '" + userId + "' found."; 
            LOG.warning(msg);
            throw new AuthorizationException(msg);
        }
        User user = (User)users.iterator().next();
        
        Collection<Role> roles = user.getRoles();
        for (Role role : roles) {
            if (role.eIsProxy()) {
                EcoreUtil.resolve(role, user.eResource());
            }
        }
        return user;
    }

    private boolean doDefaultAuthorization(OperationType operation, User user, Collection<String> areIds) 
        throws AuthorizationException {
        if (securityAreas == null) {
            loadSecurityAreas();
        }
        
        Collection<String> areaGuids = new ArrayList<String>(areIds.size());
        for (String areaId : areIds) {
            String areaGuid = securityAreas.get(areaId);
            if (areaGuid == null) {
                String msg = "Invalid security areaId '" + areaId + "' not found.";
                LOG.severe(msg);
                throw new AuthorizationException(msg);
            }
            areaGuids.add(areaGuid);
        }
        
        for (Role role : user.getRoles()) {
            Collection<SecurityArea> roleAreas = new HashSet<SecurityArea>();
            switch (operation) {
            case READ:
                roleAreas.addAll(role.getCanRead());
                // fall through
            case DELETE:
                roleAreas.addAll(role.getCanDelete()); 
                // fall through
            case UPDATE:
                roleAreas.addAll(role.getCanUpdate()); 
                // fall through
            default:
                roleAreas.addAll(role.getCanCreate()); 
            }
            for (SecurityArea roleArea : roleAreas) {
                if (areaGuids.contains(roleArea.getGuid())) {
                    return true;
                }
            }
        }
        return false;
    }

    private boolean doObjectAuthorization(OperationType operation, User user, ObjectPermission objPerm) {
        Collection<Role> allowedRoles = new HashSet<Role>();
        Collection<User> allowedUsers = new HashSet<User>();
        switch (operation) {
        case READ:
            allowedRoles.addAll(objPerm.getRolesRead());
            allowedUsers.addAll(objPerm.getUsersRead());
            // fall through
        case DELETE:
            allowedRoles.addAll(objPerm.getRolesDelete());
            allowedUsers.addAll(objPerm.getUsersDelete());
            // fall through
        case UPDATE:
            allowedRoles.addAll(objPerm.getRolesUpdate());
            allowedUsers.addAll(objPerm.getUsersUpdate());
            break;
        default:
            return false;
        }
        return findIntersection(allowedUsers, Collections.singletonList(user))
            || findIntersection(allowedRoles, user.getRoles());
    }
    
    private boolean findIntersection(Collection<? extends IPersistableObject> list1, 
                                     Collection<? extends IPersistableObject> list2) {
        Collection<String> guids = new ArrayList<String>(list1.size());
        for (IPersistableObject ipo : list1) {
            guids.add(ipo.getGuid());
        }
        for (IPersistableObject ipo : list2) {
            if (guids.contains(ipo.getGuid())) {
                return true;
            }
        }
        return false;
    }

    private synchronized void loadSecurityAreas() throws AuthorizationException {
        if (securityAreas != null) {
            return;
        }
        
        Map<String, String> tempMap = new HashMap<String, String>();
        
        Collection<EObject> areas = null;
        try {
            areas = ResourceSetQueryHandler.executeQuery(soaRepo.getResourceSet(), 
                                                         SecurityPackage.eINSTANCE.getSecurityArea(),
                                                         true);
        } catch (Exception ex) {
            String msg = NO_AUTH + "Unable to retrieve SecurityAreas " + ex;
            LOG.log(Level.SEVERE, msg, ex);
            throw new AuthorizationException(msg);
        }
        
        for (EObject eObj : areas) {
            SecurityArea area = (SecurityArea)eObj;
            String areaId = area.getAreaId();
            
            if (areaId == null) {
                String msg = "Missing areaId in SecurityArea " + area;
                LOG.severe(msg);
                throw new AuthorizationException(msg);
            }
            
            if (tempMap.containsKey(areaId)) {
                String msg = "Duplicate SecurityArea with id '" + areaId + "'";
                LOG.severe(msg);
                throw new AuthorizationException(msg);
            }
            
            tempMap.put(areaId, area.getGuid());
        }
        
        securityAreas = tempMap;
    }
}
