/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.security.impl;

import java.util.Collection;
import java.util.LinkedList;
import java.util.ResourceBundle;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.repository.util.i18n.Message;
import com.iona.soa.model.security.IProtectedRepositoryObject;
import com.iona.soa.model.security.ObjectPermission;
import com.iona.soa.model.security.OperationType;
import com.iona.soa.model.security.SecurityPackage;
import com.iona.soa.model.security.User;
import com.iona.soa.repository.EMFUtil;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.configuration.ConfigurationService;
import com.iona.soa.security.AuthenticationException;
import com.iona.soa.security.MessageBundle;
import com.iona.soa.security.SecurityContext;
import com.iona.soa.security.SecurityContextImpl;
import com.iona.soa.security.SecurityService;
import com.iona.soa.security.SecurityUtil;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

public class SecurityServiceImpl implements SecurityService {
    static final ResourceBundle BUNDLE = MessageBundle.getInstance();
    private static final Logger LOG = Logger.getLogger(SecurityService.class.getName());
    
    private String txUser;
    private Collection<EObject> allowedUpdates = new LinkedList<EObject>();
    private AuthenticationService authService = new AuthenticationServiceImpl();
    private AuthorizationService a11nService = new AuthorizationServiceImpl();

    public void setSOARepository(SOARepository repo) {
        authService.setSoaRepo(repo);
        a11nService.setSoaRepo(repo);
    }
    
    public void setConfigurationService(ConfigurationService configurationService) {
        authService.setConfig(configurationService.getConfiguration());
        a11nService.setConfig(configurationService.getConfiguration());
    }
    
    public void startTransaction(String userId) {
        txUser = userId;
    }
    
    public void endTransaction() {
        txUser = null;
        allowedUpdates.clear();
    }

    public boolean authorizeRead(String ssoToken, EObject eObject) {
        boolean ret;
        if (eObject == null) {
            ret = false;
        } else if (eObject instanceof User && getUserId(ssoToken).equals(((User)eObject).getUserid())) {
            // users are ok to read their own data, just need to authenticate token
            ret = ssoToken.equals(authenticateTokenInternal(ssoToken));
        } else {
            ret = doAuthorize(OperationType.READ, null, eObject, ssoToken);
        }
        if (LOG.isLoggable(Level.FINE)) {
            String id = eObject == null ? "<null>" : EMFUtil.getName(eObject);
            LOG.fine("User " + getUserId(ssoToken) + (ret ? " allowed" : " denied") 
                     + " to read object " + id);
        }
        return ret;
    }
    
    public boolean authorizeRead(String ssoToken, EObject eObject, Collection<String> areas) {
        return authorize(OperationType.READ, areas, ssoToken, getObjectPermission(eObject));
    }

    public boolean authorizeCreate(String ssoToken, EClass eClass, EObject container) {
        boolean ret;
        if (eClass == null) {
            ret = false;
        } else {
            if (txUser != null && !txUser.equals(getUserId(ssoToken))) {
                throw new SecurityException("Wrong user! Tx started by: " + txUser 
                                            + " current: " + getUserId(ssoToken));
            }
            ret = doAuthorize(OperationType.CREATE, eClass, container, ssoToken);
        }
        if (LOG.isLoggable(Level.FINE)) {
            String id = eClass == null ? "<null>" : eClass.getName();
            LOG.fine("User " + getUserId(ssoToken) + (ret ? " allowed" : " denied") 
                     + " to create class " + id);
        }
        return ret;
    }

    public boolean authorizeUpdate(String ssoToken, EObject eObject, EStructuralFeature feature) {
        boolean ret;
        String userId = getUserId(ssoToken);
        if (eObject == null) {
            ret = false;
        } else if (txUser == null) {
            if (eObject instanceof User  && ((User)eObject).getUserid().equals(userId)) {
                // users are ok to request to change their object, just need to authenticate token
                ret = ssoToken.equals(authenticateTokenInternal(ssoToken));
            } else {
                ret = doAuthorize(OperationType.UPDATE, null, eObject, ssoToken);
            }
        } else {
            if (!txUser.equals(userId)) {
                throw new SecurityException("Wrong user! Tx started by: " + txUser + " current: " + userId);
            }
            if (allowedUpdates.contains(eObject)) {
                ret = true;
            } else if (eObject instanceof User 
                       && userId.equals(((User)eObject).getUserid()) 
                       && SecurityPackage.eINSTANCE.getUser_Password().equals(feature)) {
                // users are ok to actually change their password, just need to authenticate token
                ret = ssoToken.equals(authenticateTokenInternal(ssoToken));
            } else {
                ret = doAuthorize(OperationType.UPDATE, null, eObject, ssoToken);
                if (ret) {
                    allowedUpdates.add(eObject);
                }
            }
        }
        if (LOG.isLoggable(Level.FINE)) {
            String id = eObject == null ? "<null>" : EMFUtil.getName(eObject);
            LOG.fine("User " + getUserId(ssoToken) + (ret ? " allowed" : " denied") 
                     + " to update object " + id);
        }
        return ret;
    }

    public boolean authorizeDelete(String ssoToken, EObject eObject) {
        boolean ret;
        if (eObject == null) {
            ret = false;
        } else {
            if (txUser != null && !txUser.equals(getUserId(ssoToken))) {
                throw new SecurityException("Wrong user! Tx started by: " + txUser 
                                            + " current: " + getUserId(ssoToken));
            }
            ret = doAuthorize(OperationType.DELETE, null, eObject, ssoToken);
        }
        if (LOG.isLoggable(Level.FINE)) {
            String id = eObject == null ? "<null>" : EMFUtil.getName(eObject);
            LOG.fine("User " + getUserId(ssoToken) + (ret ? " allowed" : " denied") 
                     + " to delete object " + id);
        }
        return ret;
    }

    
    // REST (HTTP) based Authentication
    public SecurityContext authenticate(String credentials) throws AuthenticationException {
        if (credentials == null) {
            throw new AuthenticationException(new Message(BUNDLE, "AUTH_NO_DATA"));
        }
        
        String[] credData = SecurityUtil.parseAuthHeader(credentials);
            
        if (credData == null) {
            throw new AuthenticationException(new Message(BUNDLE, "AUTH_BAD_DATA"));
        }
    
        if (TOKEN_TYPE.equals(credData[0])) {
            // Authentication Token
            if (INTERNAL_TOKEN.equals(credData[1])) {
                throw new AuthenticationException(new Message(BUNDLE, "AUTH_INTERNAL_TOKEN"));
            }
            String ssoToken = authenticateTokenInternal(credData[1]);
            if (ssoToken == null) {
                throw new AuthenticationException(new Message(BUNDLE, "AUTH_BAD_TOKEN"));
            } 
            String username = getUserId(ssoToken);
            if (username == null) {
                throw new AuthenticationException(new Message(BUNDLE, "AUTH_BAD_USER_TOKEN"));
            }
            LOG.fine("User " + username + " authenticated");
            return new SecurityContextImpl(username, ssoToken);    
        } else if (BASIC_AUTH_TYPE.equals(credData[0])) {
            // Basic Auth
            return authenticate(credData[1], credData[2]);
        } else {
            throw new AuthenticationException(new Message(BUNDLE, "AUTH_UNKNOWN_TYPE", credData[0]));
        }

    }

    // Scheduler authentication
    public SecurityContext authenticatePersistent(String userName, String password) 
        throws AuthenticationException {
        if (userName == null || password == null) {
            return null;
        }
        String token = authService.authenticatePersistent(userName, password);
        return new SecurityContextImpl(userName, token);
    }

    // Web authentication
    @SuppressWarnings("unchecked")
    public SecurityContext authenticate(String userName, String password) throws AuthenticationException {
        if (userName == null || password == null) {
            return null;
        }
        String token = authService.authenticate(userName, password);
        return new SecurityContextImpl(userName, token);
    }
    
    public boolean logoutUser(String ssoToken) {
        return authService.logout(ssoToken);
    }

    void setAuthenticationService(AuthenticationService service) {
        authService = service;
    }

    void setAuthorizationService(AuthorizationService service) {
        a11nService = service;
    }

    String getUserId(String ssoTokenValue) {
        if (INTERNAL_TOKEN.equals(ssoTokenValue)) {
            return INTERNAL_USER;
        }
        try {
            return authService.getUserId(ssoTokenValue);
        } catch (AuthenticationException e) {
            return null;
        }
    }
    
    private boolean doAuthorize(OperationType operation, 
                                EClass eClass,
                                EObject object,
                                String ssoToken) {
        Collection<String> securityAreas;
        switch (operation) {
        case CREATE:
            securityAreas = SecurityUtil.getSecurityAreas(eClass);
            if (securityAreas.size() == 0 && object != null) {
                // An type with no security area is created inside a container object
                // Treat this as an update to the container object instead: 
                return authorizeUpdate(ssoToken, object, null);
            }
            break;
        case READ:
            securityAreas = SecurityUtil.getSecurityAreas(object, true);
            break;
        default:
            securityAreas = SecurityUtil.getSecurityAreas(object, false);
            if (securityAreas.size() == 0 && object.eContainer() != null) {
                // An object with no security area is updated or deleted, which is contained in a 
                // container object. Treat this as an update to the container object instead. 
                return authorizeUpdate(ssoToken, object.eContainer(), null);
            }
        }
        if (securityAreas.size() == 0) {
            String message = "No security area has been identified for " 
                             + (eClass == null ? object.eClass().getName() : eClass.getName());
            LOG.log(operation == OperationType.CREATE ? Level.FINE : Level.WARNING, message);
            return false;
        }

        ObjectPermission objectPermission = operation != OperationType.CREATE 
                                            ? getObjectPermission(object) : null;
        
        return authorize(operation, securityAreas, ssoToken, objectPermission);         
    }

    private boolean authorize(OperationType operation, 
                              Collection<String> areas, 
                              String ssoToken, 
                              ObjectPermission oControl) {
        if (ssoToken == null) {
            return false;
        }
        if (ssoToken.equals(INTERNAL_TOKEN)) {
            return true;
        }
        
        String userId = getUserId(ssoToken);
        
        if (userId == null) {
            return false;
        }
        
        try {
            if (oControl == null) {
                return a11nService.isAuthorized(userId, operation, areas);
            } else {
                return a11nService.isAuthorized(userId, operation, oControl);
            }
        } catch (AuthorizationService.AuthorizationException ae) {
            return false;
        }
    }

    private String authenticateTokenInternal(String ssoTokenValue) {

        try {
            return authService.authenticate(ssoTokenValue);
        } catch (AuthenticationException e) {
            LOG.log(Level.FINE, "SSO Authentication failed. Token invalid or expired.", e);
            return null;
        }
    }

    private ObjectPermission getObjectPermission(EObject object) {
        return object instanceof IProtectedRepositoryObject 
               ? ((IProtectedRepositoryObject)object).getObjectControl() 
               : null;
    }
}
