/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.serialization.v1;

import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringWriter;
import java.io.Writer;
import java.net.URL;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import com.iona.soa.repository.ModelRegistry;
import com.iona.soa.repository.util.InputStreamHelper;
import com.iona.soa.repository.util.Tuple;
import com.iona.soa.serialization.AbstractSerializationBase;
import com.iona.soa.serialization.AttributeHandler;
import com.iona.soa.serialization.Deserializer;
import com.iona.soa.serialization.ObjectResolver;
import com.iona.soa.serialization.SerializationHelper;
import com.iona.soa.serialization.SerializationType;
import com.iona.soa.serialization.XMLConstants;
import com.iona.soa.serialization.XMLParseException;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

import org.jdom.Attribute;
import org.jdom.Element;
import org.jdom.Namespace;

public class DeserializerImpl extends AbstractSerializationBase implements Deserializer {
    private static final Map<SerializationType, Collection<String>> ALLOWED_META_ATTRIBUTES 
        = new HashMap<SerializationType, Collection<String>>();
    private Element rootElement;
    private ModelRegistry modelRegistry;
    private ObjectResolver eObjectResolver;
    private Map<String, EPackage> nsPackageMap; 
    private Map<String, EObject> objectsById;
    
    static {
        ALLOWED_META_ATTRIBUTES.put(SerializationType.ADD, 
                                    Arrays.asList(new String[] {XMLConstants.ACTION_ATTR,
                                                                XMLConstants.CONTAINER_ATTR,
                                                                XMLConstants.CONTAINER_TYPE_ATTR,
                                                                XMLConstants.CONTAINER_FEATURE_ATTR,
                                    }));
        ALLOWED_META_ATTRIBUTES.put(SerializationType.MODIFY, 
                                    Arrays.asList(new String[] {XMLConstants.ACTION_ATTR,
                                                                XMLConstants.HREF_ATTR,
                                                                XMLConstants.CONTAINER_ATTR,
                                                                XMLConstants.CONTAINER_TYPE_ATTR,
                                                                XMLConstants.CONTAINER_FEATURE_ATTR,
                                    }));
        ALLOWED_META_ATTRIBUTES.put(SerializationType.DELETE, 
                                    Arrays.asList(new String[] {XMLConstants.ACTION_ATTR,
                                                                XMLConstants.HREF_ATTR,
                                    }));
    }
    
    /*
     * Package private ctor. Instances are usually created via the factory.
     */
    DeserializerImpl(Element root, ObjectResolver resolver) {
        rootElement = root;
        eObjectResolver = resolver;
    }
    
    public void setModelRegistry(ModelRegistry registry) {
        modelRegistry = registry;
    }
    
    public ModelRegistry getModelRegistry() {
        if (modelRegistry == null) {
            modelRegistry = ModelRegistry.getDefaultRegistry();
        }
        return modelRegistry;
    }
    
    @SuppressWarnings("unchecked")
    public void deserialize() throws XMLParseException {
        String version = rootElement.getAttributeValue(XMLConstants.VERSION_ATTR);
        if (!SerializerImpl.XML_VERSION.equals(version)) {
            throw new XMLParseException("Incompatible XML version " + version);
        }
        
        Map<Element, Tuple<EObject, SerializationType>> objectsByElem = 
            new LinkedHashMap<Element, Tuple<EObject, SerializationType>>();
        objectsById = new HashMap<String, EObject>();
        for (Element childElem : (List<Element>)rootElement.getChildren()) {
            Tuple<EObject, SerializationType> child = parseElement(childElem);
            if (child != null) {
                objectsByElem.put(childElem, child);
            }
        }
        
        for (Element childElem : objectsByElem.keySet()) {
            Tuple<EObject, SerializationType> tuple = objectsByElem.get(childElem);
            EObject eObj = tuple.getLhs();
            SerializationType type = tuple.getRhs();
            switch (type) {
            case ADD:
            case MODIFY:
                deserializeElementData(childElem, eObj, type);
                deserializeContainer(childElem, eObj);
                break;
            case DELETE:
                deserializeDelete(childElem, eObj);
                break;
            default:
                // never reached
                throw new XMLParseException("Unsupported serialization type " + type);
            }
        }
    }

    private Tuple<EObject, SerializationType> parseElement(Element elem) throws XMLParseException {
        SerializationType type;
        String typeVal = elem.getAttributeValue(XMLConstants.ACTION_ATTR, XMLConstants.META_NS);
        if (typeVal == null) {
            type = SerializationType.ADD;
        } else {
            try {
                type = SerializationType.valueOf(typeVal);
            } catch (Exception ex) {
                throw new XMLParseException("Invalid " + XMLConstants.TYPE_ATTR + " attribute: " + typeVal);
            }
        }
        
        EObject eObj = null;
        String objectId = null;
        EClass eClass = getEClass(elem.getNamespace().getPrefix(), elem.getName());
        switch (type) {
        case ADD:
            EAttribute idAttr = eClass.getEIDAttribute();
            String idVal = elem.getAttributeValue(idAttr.getName());
            if (idVal != null) {
                Object id = SerializationHelper.fromString(idAttr, idVal);
                if (typeVal == null && getOption(OPTION_ALLOW_MODIFY_ON_ADD)) {
                    URI uri = URI.createURI("#" + id.toString());
                    eObj = eObjectResolver.getEObject(eClass, uri, true);
                }
                if (eObj == null) {
                    eObj = eObjectResolver.createEObject(eClass, id);
                }
                if (eObj == null) {
                    throw new XMLParseException("Unable to add object with id " + id);
                }
                objectId = idVal;
                break;
            }
            throw new XMLParseException("Missing id attribute");
        case MODIFY:
        case DELETE:
            String hrefVal = elem.getAttributeValue(XMLConstants.HREF_ATTR, XMLConstants.META_NS);
            if (hrefVal != null) {
                URI uri = URI.createURI(hrefVal);
                eObj = eObjectResolver.getEObject(eClass, uri, true);
                objectId = hrefVal;
                break;
            }
            throw new XMLParseException("Missing href attribute");
        default:
            // never reached
            throw new XMLParseException("Unsupported serialization type " + type);
        }

        if (objectsById.containsKey(objectId)) {
            throw new XMLParseException("Duplicate object with id " + objectId);
        } else if (eObj != null) {
            objectsById.put(objectId, eObj);
            return new Tuple<EObject, SerializationType>(eObj, type);
        } else if (getOption(OPTION_STRICT_MODE)) {
            throw new XMLParseException("Unable to resolve object with id " + objectId);
        } else {
            return null;
        }
    }

    @SuppressWarnings("unchecked")
    private void deserializeDelete(Element elem, EObject eObj) throws XMLParseException {
        for (Attribute attribute : (Collection<Attribute>)elem.getAttributes()) {
            if (attribute.getNamespace() != XMLConstants.META_NS 
                || !ALLOWED_META_ATTRIBUTES.get(SerializationType.DELETE).contains(attribute.getName())) {
                throw new XMLParseException("Invalid attribute " + attribute);
            }
        }
        for (Element featureElem : (Collection<Element>)elem.getChildren()) {
            if (getOption(OPTION_STRICT_MODE)) {
                throw new XMLParseException("Invalid XML element " + featureElem);
            }
        }
        if (!eObjectResolver.deleteFromResource(eObj) && getOption(OPTION_STRICT_MODE)) {
            throw new XMLParseException("Unable to delete object with id " 
                                        + SerializationHelper.getIdString(eObj));
        }
    }

    @SuppressWarnings("unchecked")
    private void deserializeElementData(Element elem, 
                                        EObject eObj, 
                                        SerializationType type) throws XMLParseException {
        for (Attribute attribute : (Collection<Attribute>)elem.getAttributes()) {
            if (attribute.getNamespace() == Namespace.NO_NAMESPACE) {
                String name = attribute.getName();
                String strValue = attribute.getValue();
                setFeature(eObj, name, strValue, false, false, null);
            } else if (attribute.getNamespace() != XMLConstants.META_NS 
                       || !ALLOWED_META_ATTRIBUTES.get(type).contains(attribute.getName())) {
                throw new XMLParseException("Invalid attribute " + attribute);
            }
        }
        for (Element featureElem : (Collection<Element>)elem.getChildren()) {
            if (featureElem.getNamespace() == Namespace.NO_NAMESPACE) {
                String name = featureElem.getName();
                String strValue = featureElem.getText();
                boolean unset = false;
                Attribute unsetAttr = featureElem.getAttribute(XMLConstants.UNSET_ATTR, 
                                                               XMLConstants.META_NS);
                if (unsetAttr != null) {
                    if (type == SerializationType.MODIFY) {
                        unset = Boolean.parseBoolean(unsetAttr.getValue());
                    } else {
                        throw new XMLParseException("Invalid attribute " + unsetAttr);
                    }
                }
                boolean remove = false;
                Attribute removeAttr = featureElem.getAttribute(XMLConstants.REMOVE_ATTR, 
                                                                XMLConstants.META_NS);
                if (removeAttr != null) {
                    if (type == SerializationType.MODIFY) {
                        remove = Boolean.parseBoolean(removeAttr.getValue());
                    } else {
                        throw new XMLParseException("Invalid attribute " + removeAttr);
                    }
                }
                String typeHint = 
                    featureElem.getAttributeValue(XMLConstants.TYPE_ATTR, XMLConstants.META_NS);
                setFeature(eObj, name, strValue, unset, remove, typeHint);
            } else if (getOption(OPTION_STRICT_MODE)) {
                throw new XMLParseException("Invalid XML element " + featureElem);
            }
        }
    }

    private void deserializeContainer(Element elem, EObject eObj) throws XMLParseException {
        Attribute containerAttr = 
            elem.getAttribute(XMLConstants.CONTAINER_ATTR, XMLConstants.META_NS);
        if (containerAttr != null) {
            String featureVal = 
                elem.getAttributeValue(XMLConstants.CONTAINER_FEATURE_ATTR, XMLConstants.META_NS);
            if (featureVal == null) {
                throw new XMLParseException("Missing " + XMLConstants.CONTAINER_FEATURE_ATTR 
                                            + " attribute in Element " + elem);
            }
            String typeHint = 
                elem.getAttributeValue(XMLConstants.CONTAINER_TYPE_ATTR, XMLConstants.META_NS);
            String strValue = containerAttr.getValue();
            setContainer(eObj, featureVal, strValue, typeHint);
        } else if (!eObjectResolver.addToResource(eObj) && getOption(OPTION_STRICT_MODE)) {
            throw new XMLParseException("Unable to add object with id " 
                                        + SerializationHelper.getIdString(eObj));
        }
    }    

    @SuppressWarnings("unchecked")
    private void setContainer(EObject eObj, 
                              String featureVal, 
                              String strValue, 
                              String typeHint) throws XMLParseException {
        EObject container;
        if (objectsById.containsKey(strValue)) {
            container = objectsById.get(strValue);
        } else {
            EClass eClass = null;
            if (typeHint != null) {
                eClass = getEClass(typeHint);
            } else if (getOption(OPTION_STRICT_MODE)) {
                throw new XMLParseException("Missing type for container " + strValue);
            }
            container = eObjectResolver.getEObject(eClass, URI.createURI(strValue), true);
            if (container != null) {
                objectsById.put(strValue, container);
            } else if (getOption(OPTION_STRICT_MODE)) {
                throw new XMLParseException("Unable to resolve " + strValue);
            } else {
                return;
            }
        }
        EStructuralFeature containerRef = container.eClass().getEStructuralFeature(featureVal);
        if (!(containerRef instanceof EReference)) {
            throw new XMLParseException("Invalid feature " + featureVal);
        }
        if (!eObjectResolver.addToContainer(eObj, container, (EReference)containerRef) 
                && getOption(OPTION_STRICT_MODE)) {
            throw new XMLParseException("Unable to add object with id " 
                                        + SerializationHelper.getIdString(eObj)
                                        + " to container " + SerializationHelper.getIdString(container));
        }
    }

    private void setFeature(EObject eObj, 
                            String featureName, 
                            String strValue, 
                            boolean unset,
                            boolean remove,
                            String typeHint) throws XMLParseException {
        EStructuralFeature feature = eObj.eClass().getEStructuralFeature(featureName);
        if (feature != null) {
            if (unset) {
                eObj.eUnset(feature);
                if (strValue != null && strValue.length() > 0 && getOption(OPTION_STRICT_MODE)) {
                    throw new XMLParseException("Invalid text '" + getShotenedToString(strValue)
                                                + "' in unset feature " + featureName);
                }
            } else if (feature instanceof EAttribute) {
                Object object = SerializationHelper.fromString((EAttribute)feature, strValue);
                if (remove) {
                    removeFeature(eObj, feature, object);
                } else {
                    setFeature(eObj, feature, object);
                }
            } else {
                EObject refObj;
                if (objectsById.containsKey(strValue)) {
                    refObj = objectsById.get(strValue);
                } else {
                    EClass eClass = getReferenceType((EReference)feature, typeHint);
                    refObj = eObjectResolver.getEObject(eClass, URI.createURI(strValue), false);
                    if (refObj != null) {
                        objectsById.put(strValue, refObj);
                    } else if (getOption(OPTION_STRICT_MODE)) {
                        throw new XMLParseException("Unable to resolve " + strValue);
                    } else {
                        return;
                    }
                }
                if (remove) {
                    removeFeature(eObj, feature, refObj);
                } else {
                    setFeature(eObj, feature, refObj);
                }
            }
        } else if (featureName.endsWith(XMLConstants.HREF_ATTR_SUFFIX)) {
            featureName = featureName.substring(0, featureName.length() - 4);
            feature = eObj.eClass().getEStructuralFeature(featureName);
            if (feature instanceof EAttribute) {
                EAttribute eAttr = (EAttribute)feature;
                try {
                    URI uri = URI.createURI(strValue);
                    URIConverter converter = getOption(OPTION_URL_CONVERTER);
                    if (converter != null) {
                        uri = converter.convert(uri);
                    }
                    AttributeHandler handler = getHandler(eAttr);
                    if (handler != null) {
                        setFeature(eObj, eAttr, handler.deserialize(eObj, uri));
                    } else if (getOption(OPTION_ENABLE_INTERNAL_URL_READER)) {
                        Reader isr = new InputStreamReader(new URL(uri.toString()).openStream(), "UTF-8");
                        Writer sw = new StringWriter();
                        InputStreamHelper.drain(isr, sw);
                        InputStreamHelper.close(isr);
                        Object val = SerializationHelper.fromString(eAttr, sw.toString());
                        setFeature(eObj, eAttr, val);
                    } else {
                        throw new XMLParseException("Missing AttributeHandler for EAttribute " + eAttr);
                    }
                } catch (IOException ioe) {
                    throw new XMLParseException(ioe);
                }
            } else {
                throw new XMLParseException("Attribute feature not found " + featureName);
            }
        } else if (getOption(OPTION_STRICT_MODE)) {
            throw new XMLParseException("Unknown Feature '" + featureName + "'");
        }
    }

    @SuppressWarnings("unchecked")
    private void setFeature(EObject eObj, EStructuralFeature feature, Object object) {
        if (feature.isMany()) {
            ((Collection)eObj.eGet(feature, false)).add(object);
        } else {
            eObj.eSet(feature, object);
        }
    }

    @SuppressWarnings("unchecked")
    private void removeFeature(EObject eObj, 
                               EStructuralFeature feature, 
                               Object object) throws XMLParseException {
        if (feature.isMany()) {
            ((Collection)eObj.eGet(feature, false)).remove(object);
        } else if (getOption(OPTION_STRICT_MODE)) {
            throw new XMLParseException("Unable to remove item '" + getShotenedToString(object) 
                                        + "' from feature " + feature.getName()
                                        + " of object " + SerializationHelper.getIdString(eObj));
        }
    }

    private EClass getEClass(String typeHint) throws XMLParseException {
        String[] parts = typeHint.split(":");
        if (parts.length != 2) {
            throw new XMLParseException("Invalid type hint " + typeHint);
        }
        return getEClass(parts[0], parts[1]);
    }
    
    @SuppressWarnings("unchecked")
    private synchronized EClass getEClass(String pkgName, String clsName) throws XMLParseException {
        if (nsPackageMap == null) {
            nsPackageMap = new HashMap<String, EPackage>();
            for (Namespace ns : (Collection<Namespace>)rootElement.getAdditionalNamespaces()) {
                EPackage ePkg = getModelRegistry().getPackage(ns.getURI());
                if (ePkg != null) {
                    nsPackageMap.put(ns.getPrefix(), ePkg);
                } else {
                    throw new XMLParseException("Unable to find EPackage for namespace " + ns.getURI());
                }
            }
        }
        
        EPackage pkg = nsPackageMap.get(pkgName);
        if (pkg == null) {
            throw new XMLParseException("Unknown Namespace " 
                                        + rootElement.getNamespace(pkgName).getURI()
                                        + " [" + pkgName + "]");
        }

        EClassifier classifier = pkg.getEClassifier(clsName);
        if (!(classifier instanceof EClass)) {
            throw new XMLParseException("Unknown EClass " + clsName);
        }
        return (EClass)classifier;
    }

    private EClass getReferenceType(EReference reference, String typeHint) throws XMLParseException {
        EClass eClass = reference.getEReferenceType();
        if (typeHint != null) {
            EClass hintedClass = getEClass(typeHint);
            if (eClass.isSuperTypeOf(hintedClass)) {
                eClass = hintedClass;
            } else if (getOption(OPTION_STRICT_MODE)) {
                throw new XMLParseException("Invalid type " + hintedClass.getName() 
                                            + " for reference " + reference.getName());
            }
        }
        return eClass;
    }

    private String getShotenedToString(Object object) {
        String val  = object.toString();
        if (val.length() > 25) {
            val = val.substring(0, 22) + "...";
        }
        return val;
    }
}