/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.serialization.v1;

import java.io.IOException;
import java.io.OutputStream;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Map;

import com.iona.soa.serialization.AbstractSerializationBase;
import com.iona.soa.serialization.AttributeHandler;
import com.iona.soa.serialization.ProxyURIFactory;
import com.iona.soa.serialization.SerializationHelper;
import com.iona.soa.serialization.SerializationType;
import com.iona.soa.serialization.Serializer;
import com.iona.soa.serialization.XMLConstants;
import com.iona.soa.serialization.XMLSerializationException;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.InternalEObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;

import org.jdom.Attribute;
import org.jdom.Element;
import org.jdom.Namespace;
import org.jdom.output.XMLOutputter;

public class SerializerImpl extends AbstractSerializationBase implements Serializer {
    static final String XML_VERSION = "1.0";
    private Collection<EObject> addedObjects = new LinkedList<EObject>();
    private Map<EObject, ModificationData> modifiedObjects = new LinkedHashMap<EObject, ModificationData>();
    private Collection<EObject> deletedObjects = new LinkedList<EObject>();
    private ProxyURIFactory uriFactory;

    /*
     * Package private ctor. Instances are usually created via the factory.
     */
    SerializerImpl() {
    }
    
    public void addObjects(EObject ... objects) throws XMLSerializationException {
        boolean inclChildren = getOption(OPTION_INCLUDE_CHILDREN);
        boolean inclContainer = getOption(OPTION_INCLUDE_CONTAINER);
        for (EObject eObj : objects) {
            registerObject(eObj, SerializationType.ADD, null, null, inclChildren, inclContainer);
        }
    }

    public void deleteObjects(EObject... objects) throws XMLSerializationException {
        for (EObject eObj : objects) {
            registerObject(eObj, SerializationType.DELETE, null, null, true, false);
        }
    }
    
    public void modifyObject(EObject object, 
                             String featureName, 
                             Object ... values) throws XMLSerializationException {
        EStructuralFeature feature = object.eClass().getEStructuralFeature(featureName);
        if (feature == null) {
            throw new XMLSerializationException("Invalid feature name " + featureName);
        }
        modifyObject(object, feature, values);
    }

    public void modifyObject(EObject object, 
                             EStructuralFeature feature, 
                             Object ... values) throws XMLSerializationException {
        registerObject(object, SerializationType.MODIFY, feature, values, false, false);
    }

    @SuppressWarnings("unchecked")
    private void registerObject(EObject eObj,
                                SerializationType type,
                                EStructuralFeature feature, 
                                Object[] values,
                                boolean includeChildren, 
                                boolean includeContainer) throws XMLSerializationException {
        if (eObj.eIsProxy()) {
            return;
        }
        registerObject(eObj, type);
        if (includeContainer) {
            EObject container = eObj.eContainer();
            while (container != null) {
                registerObject(container, type);
                container = container.eContainer();
            }
        }
        if (includeChildren) {
            for (EObject child : (Collection<EObject>)eObj.eContents()) {
                registerObject(child, type, null, null, true, false);
            }
        }
        if (SerializationType.MODIFY == type) {
            modifiedObjects.get(eObj).addChangeEntry(feature, values);
        }
    }

    private void registerObject(EObject eObj, SerializationType type) throws XMLSerializationException {
        switch (type) {
        case ADD:
            if (modifiedObjects.containsKey(eObj) || deletedObjects.contains(eObj)) {
                throw new XMLSerializationException("Object " + SerializationHelper.getIdString(eObj) 
                        + " already registered");
            }
            addedObjects.add(eObj);
            break;
        case MODIFY:
            if (addedObjects.contains(eObj) || deletedObjects.contains(eObj)) {
                throw new XMLSerializationException("Object " + SerializationHelper.getIdString(eObj) 
                        + " already registered");
            } else if (!modifiedObjects.containsKey(eObj)) {
                modifiedObjects.put(eObj, new ModificationData());
            }
            break;
        case DELETE:
            if (addedObjects.contains(eObj) || modifiedObjects.containsKey(eObj)) {
                throw new XMLSerializationException("Object " + SerializationHelper.getIdString(eObj) 
                        + " already registered");
            }
            deletedObjects.add(eObj);
            break;
        default:
            throw new RuntimeException("Invalid serialization type");
        }
    }
    
    public ProxyURIFactory getURIFactory() {
        if (uriFactory == null) {
            uriFactory = new ProxyURIFactory() {

                public URI getURI(EObject eObject) {
                    if (eObject.eIsProxy()) {
                        return ((InternalEObject)eObject).eProxyURI();
                    } else if (eObject.eResource() != null) {
                        Resource res = eObject.eResource();
                        URI uri = res.getURI() == null ? URI.createURI("") : res.getURI();
                        return uri.appendFragment(res.getURIFragment(eObject));
                    } else {
                        return EcoreUtil.getURI(eObject);
                    }
                }

            };
        }
        return uriFactory;
    }

    public void setURIFactory(ProxyURIFactory factory) {
        uriFactory = factory;
    }

    public String serialize() throws XMLSerializationException {
        Element elem = serializeInternal();
        return new XMLOutputter(getOption(OPTION_XML_FORMAT)).outputString(elem);
    }

    public void serialize(Writer out) throws XMLSerializationException, IOException {
        Element elem = serializeInternal();
        new XMLOutputter(getOption(OPTION_XML_FORMAT)).output(elem, out);
    }

    public void serialize(OutputStream out) throws XMLSerializationException, IOException {
        Element elem = serializeInternal();
        new XMLOutputter(getOption(OPTION_XML_FORMAT)).output(elem, out);
    }

    private Element serializeInternal() throws XMLSerializationException {
        Element topLevel = new Element(XMLConstants.COLLECTION_ELEM, XMLConstants.META_NS);
        topLevel.setAttribute(XMLConstants.VERSION_ATTR, XML_VERSION);
        Collection<Namespace> namespaces = new LinkedHashSet<Namespace>();

        FeatureFilter filter = getOption(OPTION_FEATURE_FITLER);

        for (EObject eObj : addedObjects) {
            Element childElem = serializeObjectInternal(eObj, SerializationType.ADD, namespaces, filter);
            if (childElem != null) {
                topLevel.addContent(childElem);
            }
        }
        for (EObject eObj : modifiedObjects.keySet()) {
            Element childElem = serializeObjectInternal(eObj, SerializationType.MODIFY, namespaces, filter);
            if (childElem != null) {
                topLevel.addContent(childElem);
            }
        }
        for (EObject eObj : deletedObjects) {
            Element childElem = serializeObjectInternal(eObj, SerializationType.DELETE, namespaces, filter);
            if (childElem != null) {
                topLevel.addContent(childElem);
            }
        }
        for (Namespace ns : namespaces) {
            topLevel.addNamespaceDeclaration(ns);
        }
        return topLevel;
    }
    
    private Element serializeObjectInternal(EObject eObj, 
                                            SerializationType type,
                                            Collection<Namespace> namespaces,
                                            FeatureFilter filter) throws XMLSerializationException {
        EClass eClass = eObj.eClass();
        EPackage ePkg = eClass.getEPackage(); 
        Namespace childNS = Namespace.getNamespace(ePkg.getNsPrefix(), ePkg.getNsURI());
        namespaces.add(childNS);
        Element childElem = new Element(eClass.getName(), childNS);
        childElem.setAttribute(XMLConstants.ACTION_ATTR, type.toString(), XMLConstants.META_NS);
        if (type == SerializationType.ADD) {
            for (EAttribute eAttr : getAttributes(eObj.eClass())) {
                addAttribute(childElem, eObj, eAttr, filter);
            }
            for (EReference eRef : getReferences(eObj.eClass())) {
                addReference(childElem, eObj, eRef, filter, namespaces);
            }
            Namespace ns = addContainer(childElem, eObj);
            if (ns != null) {
                namespaces.add(ns);
            }
        } else if (type == SerializationType.DELETE) {
            if (!deletedObjects.contains(eObj.eContainer())) {
                addProxyInfo(childElem, eObj);
            } else {
                // skip this object
                return null;
            }
        } else if (type == SerializationType.MODIFY) {
            addProxyInfo(childElem, eObj);
            serializeModifications(childElem, eObj, namespaces);
        }
        return childElem;
    }
    
    private void serializeModifications(Element elem, 
                                        EObject eObj, 
                                        Collection<Namespace> namespaces) throws XMLSerializationException {
        ModificationData entry = modifiedObjects.get(eObj);
        for (EAttribute eAttr : entry.getChangedAttributes()) {
            AttributeHandler handler = getHandler(eAttr);
            if (!eObj.eIsSet(eAttr)) {
                Collection<Attribute> attributes = 
                    Collections.singletonList(new Attribute(XMLConstants.UNSET_ATTR, 
                                                            Boolean.toString(true), 
                                                            XMLConstants.META_NS));
                addAttributeData(elem, eObj, eAttr, null, handler, attributes);
            } else if (eAttr.isMany()) {
                Collection currVals = (Collection)eObj.eGet(eAttr, false);
                for (Object value : entry.getManyAttributeValues(eAttr)) {
                    Collection<Attribute> attributes = new LinkedList<Attribute>();
                    if (!currVals.contains(value)) {
                        attributes.add(new Attribute(XMLConstants.REMOVE_ATTR, 
                                                     Boolean.toString(true), 
                                                     XMLConstants.META_NS));
                    }
                    addAttributeData(elem, eObj, eAttr, value, handler, attributes);
                }
            } else {
                Object value = eObj.eGet(eAttr, false);
                Collection<Attribute> attributes = Collections.emptyList();
                addAttributeData(elem, eObj, eAttr, value, handler, attributes);
            }
        }
        for (EReference eRef : entry.getChangedReferences()) {
            if (!eObj.eIsSet(eRef)) {
                Collection<Attribute> attributes = 
                    Collections.singletonList(new Attribute(XMLConstants.UNSET_ATTR, 
                                                            Boolean.toString(true), 
                                                            XMLConstants.META_NS));
                addReferenceData(elem, eRef, null, attributes);
            } else if (eRef.isMany()) {
                Collection currRefs = (Collection)eObj.eGet(eRef, false);
                for (EObject refObj : entry.getManyReferenceValues(eRef)) {
                    Collection<Attribute> attributes = 
                        getTypeXMLAttribute(eRef, refObj, namespaces);
                    if (!currRefs.contains(refObj)) {
                        attributes.add(new Attribute(XMLConstants.REMOVE_ATTR, 
                                                     Boolean.toString(true), 
                                                     XMLConstants.META_NS));
                    }
                    addReferenceData(elem, eRef, refObj, attributes);
                }
            } else {
                EObject refObj = (EObject)eObj.eGet(eRef, false);
                Collection<Attribute> attributes = getTypeXMLAttribute(eRef, refObj, namespaces);
                addReferenceData(elem, eRef, refObj, attributes);
            }
        }
    }

    private Namespace addContainer(Element elem, EObject eObj) throws XMLSerializationException {
        EObject eContainer = eObj.eContainer();
        if (eContainer == null) {
            return null;
        } else if (addedObjects.contains(eContainer) && !getOption(OPTION_FORCE_CONTAINER_REF)) {
            return null;
        }
        
        EReference feature = eObj.eContainmentFeature();
        String id;
        boolean addType = false;
        if (addedObjects.contains(eContainer)) {
            id = SerializationHelper.getIdString(eContainer);
        } else {
            URI uri = getURIFactory().getURI(eContainer);
            if (uri == null) {
                if (getOption(OPTION_STRICT_MODE)) {
                    throw new XMLSerializationException("URIFactory returned null URI");
                } else {
                    return null;
                }
            }
            id = uri.toString();
            addType = true;
        }
        elem.setAttribute(XMLConstants.CONTAINER_ATTR, id, XMLConstants.META_NS);
        elem.setAttribute(XMLConstants.CONTAINER_FEATURE_ATTR, 
                feature.getName(), 
                XMLConstants.META_NS);
        if (addType) {
            EPackage refPkg = eContainer.eClass().getEPackage();
            Namespace refNS = 
                Namespace.getNamespace(refPkg.getNsPrefix(), refPkg.getNsURI());
            elem.setAttribute(XMLConstants.CONTAINER_TYPE_ATTR, 
                    refPkg.getNsPrefix() + ":" + eContainer.eClass().getName(),
                    XMLConstants.META_NS);
            return refNS;
        }
        return null;
    }

    private void addReference(Element elem, 
                              EObject eObj, 
                              EReference eRef, 
                              FeatureFilter filter,
                              Collection<Namespace> namespaces) throws XMLSerializationException {
        if (filter != null && filter.filterAll(eRef)) {
            return;
        } else if (!eObj.eIsSet(eRef)) {
            return;
        }

        Collection<EObject> refObjs;
        if (eRef.isMany()) {
            refObjs = getListRefVal(eObj, eRef);
        } else {
            refObjs = Collections.singletonList((EObject)eObj.eGet(eRef, false));
        }
        for (EObject refObj : refObjs) {
            if (filter != null && filter.filter(eRef, eObj, refObj)) {
                continue;
            }
            Collection<Attribute> attributes = getTypeXMLAttribute(eRef, refObj, namespaces);
            addReferenceData(elem, eRef, refObj, attributes);
        }
    }

    private void addReferenceData(Element elem, 
                                  EReference eRef, 
                                  EObject refObj, 
                                  Collection<Attribute> attributes) throws XMLSerializationException {
        String value;

        if (refObj == null) {
            value = null;
        } else if (addedObjects.contains(refObj)) {
            value = SerializationHelper.getIdString(refObj);
        } else {
            URI uri = getURIFactory().getURI(refObj);
            if (uri == null) {
                if (getOption(OPTION_STRICT_MODE)) {
                    throw new XMLSerializationException("URIFactory returned null URI");
                } else {
                    return;
                }
            }
            value = uri.toString();
        }
        
        if (eRef.isMany() || attributes.size() > 0 || value == null) {
            Element subElem = new Element(eRef.getName());
            if (value != null) {
                subElem.setText(value);
            }
            for (Attribute attrib : attributes) {
                subElem.setAttribute(attrib);
            }
            elem.addContent(subElem);
        } else {
            elem.setAttribute(eRef.getName(), value);
        }
    }

    private Collection<Attribute> getTypeXMLAttribute(EReference eRef, 
                                                      EObject refObj, 
                                                      Collection<Namespace> namespaces) {
        Collection<Attribute> attributes = new LinkedList<Attribute>();
        if (eRef.getEReferenceType() != refObj.eClass() && !addedObjects.contains(refObj)) {
            EPackage refPkg = refObj.eClass().getEPackage();
            namespaces.add(Namespace.getNamespace(refPkg.getNsPrefix(), refPkg.getNsURI()));
            attributes.add(new Attribute(XMLConstants.TYPE_ATTR, 
                                         refPkg.getNsPrefix() + ":" + refObj.eClass().getName(),
                                         XMLConstants.META_NS));
        }
        return attributes;
    }

    private void addAttribute(Element elem, 
                              EObject eObj, 
                              EAttribute eAttr, 
                              FeatureFilter filter) throws XMLSerializationException {
        if (filter != null && filter.filterAll(eAttr)) {
            return;
        }
        if (!eObj.eIsSet(eAttr)) {
            return;
        }
        AttributeHandler handler = getHandler(eAttr);
        Collection<Object> values;
        if (eAttr.isMany()) {
            values = getListAttribVal(eObj, eAttr);
        } else {
            values = Collections.singletonList(eObj.eGet(eAttr, false));
        }
        for (Object obj : values) {
            if (filter != null && filter.filter(eAttr, eObj, obj)) {
                continue;
            }
            Collection<Attribute> attributes = Collections.emptyList();
            addAttributeData(elem, eObj, eAttr, obj, handler, attributes);
        }
    }
    
    private void addAttributeData(Element elem, 
                                  EObject eObj,
                                  EAttribute eAttr, 
                                  Object value,
                                  AttributeHandler handler,
                                  Collection<Attribute> attributes) throws XMLSerializationException {
        String xmlKey;
        String xmlValue;
        if (handler != null) {
            xmlKey = eAttr.getName() + XMLConstants.HREF_ATTR_SUFFIX;
            URI uri = handler.serialize(eObj, value);
            xmlValue = uri == null ? null : uri.toString();
        } else {
            xmlKey = eAttr.getName();
            xmlValue = SerializationHelper.toString(eAttr, value);
        }
        if (eAttr.isMany() || attributes.size() > 0 || xmlValue == null) {
            Element valElem = new Element(xmlKey);
            valElem.setText(xmlValue);
            for (Attribute attrib : attributes) {
                valElem.setAttribute(attrib);
            }
            elem.addContent(valElem);
        } else {
            elem.setAttribute(xmlKey, xmlValue);
        }
    }

    private void addProxyInfo(Element elem, EObject eObj) throws XMLSerializationException {
        URI uri = getURIFactory().getURI(eObj);
        if (uri == null) {
            throw new XMLSerializationException("URIFactory returned null URI");
        }
        String value = uri.toString();
        elem.setAttribute(XMLConstants.HREF_ATTR, value, XMLConstants.META_NS);
    }

    @SuppressWarnings("unchecked")
    private static Collection<EAttribute> getAttributes(EClass eClass) {
        Collection<EAttribute> allAttrs = (Collection<EAttribute>)eClass.getEAllAttributes();
        Collection<EAttribute> ret = new ArrayList<EAttribute>(allAttrs.size());
        for (EAttribute eRef : allAttrs) {
            if (!eRef.isTransient()) {
                ret.add(eRef);
            }
        }
        return ret;
    }

    @SuppressWarnings("unchecked")
    private static Collection<EReference> getReferences(EClass eClass) {
        Collection<EReference> allRefs = (Collection<EReference>)eClass.getEAllReferences();
        Collection<EReference> ret = new ArrayList<EReference>(allRefs.size());
        for (EReference eRef : allRefs) {
            if (!eRef.isTransient()) {
                ret.add(eRef);
            }
        }
        return ret;
    }

    @SuppressWarnings("unchecked")
    private Collection<EObject> getListRefVal(EObject eObj, EReference eRef) {
        return (Collection<EObject>)eObj.eGet(eRef, false);
    }

    @SuppressWarnings("unchecked")
    private Collection<Object> getListAttribVal(EObject eObj, EAttribute eAttr) {
        return (Collection<Object>)eObj.eGet(eAttr, false);
    }
}
