/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository;

import java.io.File;
import java.io.FilenameFilter;
import java.io.PrintStream;
import java.net.URL;
import java.text.DateFormat;
import java.text.MessageFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.repository.adapters.ChangeAuthorizationAdapter;
import com.iona.repository.adapters.ChangeValidationAdapter;
import com.iona.repository.adapters.PerTransactionUpdatesAdapter;
import com.iona.repository.db.Database;
import com.iona.repository.db.DatabaseResolver;
import com.iona.repository.event.DataChangeAdapter;
import com.iona.repository.event.DataChangeEventObserver;
import com.iona.repository.notification.NotificationEngine;
import com.iona.soa.configgen.util.DBCreate;
import com.iona.soa.model.events.DataChangeEvent;
import com.iona.soa.repository.ModelRegistry;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.SOARepositoryService;
import com.iona.soa.repository.SchemaVersionChecker;
import com.iona.soa.repository.configuration.ConfigurationService;
import com.iona.soa.repository.configuration.HttpConfiguration;
import com.iona.soa.repository.jdbc.DBFactory;
import com.iona.soa.repository.jdbc.JDBCConnectionManager;
import com.iona.soa.repository.jdbc.JDBCResourceRepository;
import com.iona.soa.repository.jdbc.LoggerOutputStream;
import com.iona.soa.repository.jdbc.MTConnectionManagerImpl;
import com.iona.soa.repository.jdbc.PooledConnection;
import com.iona.soa.repository.model.util.ModelHelper;
import com.iona.soa.repository.security.PasswordProviderHandler;
import com.iona.soa.repository.util.URLUtil;
import com.iona.soa.repository.util.VersionType;
import com.iona.soa.search.SearchService;
import com.iona.soa.security.SecurityService;

import org.eclipse.emf.ecore.resource.ResourceSet;
import org.osgi.framework.BundleContext;
import org.springframework.osgi.context.BundleContextAware;

public class RepositoryServiceImpl implements SOARepositoryService, BundleContextAware {
    
    public static class ExitStatus {
        private int code;

        private String message = "";

        private Throwable exception;

        public void setCode(int code) {
            this.code = code;
        }

        public int getCode() {
            return code;
        }

        public void setMessage(String msg) {
            this.message = msg;
        }

        public String getMessage() {
            return message;
        }

        public void setException(Throwable e) {
            exception = e;
        }

        public Throwable getException() {
            return exception;
        }
    }

    public static final int DEFAULT_REPOSITORY_PORT = 4968;

    public static final String EMBEDDED_UDDI_ENABLED = "uddi.embedded";

    public static final String DB_DRIVER_PROP = "database.driver";

    public static final String DB_URL_PROP = "database.url";

    public static final String USERNAME_PROP = "database.user";

    public static final String PASSWORD_PROP = "database.password";

    public static final String HOST_PROP = "repository.url.host";

    public static final String PORT_PROP = "repository.url.port";

    public static final String SECURITY_ENABLED = "provider.securityEnabled";

    public static final String CHANGE_VALIDATOR_ENABLED = "repository.changeValidatorEnabled";
    
    protected static final MessageFormat VERSION_MESSAGE_FMT = new MessageFormat("{0}    {1}");

    private static final Logger LOG = Logger.getLogger("RepositoryServer");

    // TO DO: This really should be called DATABASE as there should really be no
    // difference
    // between an embeded database and an extrnal one from the Repository point
    // of view
    private static final String EMBEDDED_DATABASE = "embeddedDatabase";

    private static final String APPLICATION_NAME = "FUSE Depot Server";

    private static Map<String, Object> compMap = new HashMap<String, Object>();

    private static ModelRegistry modelRegistry;

    private static NotificationEngine notificationEngine;

    private static DataChangeEventObserver notificationEngineAsObserver;

    private static ExitStatus exitStatus = new ExitStatus();

    private static JDBCConnectionManager connectionMgr;

    private static SOARepository repository;

    private BundleContext osgiBundleContext;
    
    private SecurityService securityService;

    private SearchService searchService;

    private ConfigurationService configService;

    private void init(Map<String, String> props) throws Exception {

        Boolean enableChangeValidator = Boolean.TRUE;
        String v = props.get(CHANGE_VALIDATOR_ENABLED);
        if (v != null) {
            enableChangeValidator = Boolean.parseBoolean(v);
        }

        if (securityService != null) {
            repository = new JDBCResourceRepository(connectionMgr, modelRegistry,
                                                    new ChangeValidationAdapter(enableChangeValidator),
                                                    new PerTransactionUpdatesAdapter(),
                                                    DataChangeAdapter.getInstance(),
                                                    new ChangeAuthorizationAdapter(securityService));
            securityService.setSOARepository(repository);
        } else {
            repository = new JDBCResourceRepository(connectionMgr, modelRegistry,
                                                    new ChangeValidationAdapter(enableChangeValidator),
                                                    new PerTransactionUpdatesAdapter(),
                                                    DataChangeAdapter.getInstance());
        }
        if (searchService != null) {
            searchService.setSOARepository(repository);
        }
        osgiBundleContext.registerService(ResourceSet.class.getName(), repository.getResourceSet(), 
                                          new Hashtable<String, String>());
        createNotificationEngine(props);
    }

    public static ExitStatus getExistStatus() {
        return exitStatus;
    }

    public SOARepository getRepository() {
        return repository;
    }

    public void start() throws Exception {
        try {
          
            Map<String, String> props = configService.getConfiguration();
            initializeEMF();

            connectionMgr = createJDBCConectionManager(props);
            
            if (props.containsKey("-version")) {
                System.out.println(APPLICATION_NAME);
                System.out.println("Model schema version: " + SchemaVersionChecker.getCodeSchemaVersion());
                if (connectionMgr != null) {
                    List<VersionType> dbVersions;
                    PooledConnection pc = connectionMgr.getConnection();
                    try {
                        dbVersions = SchemaVersionChecker.getAllSchemaVersionsFromDb(pc.getConnection());
                    } finally {
                        connectionMgr.returnConnection(pc);
                    }
                    if (dbVersions.size() > 0) {
                        System.out.println("Database schema version: "
                                           + dbVersions.get(dbVersions.size() - 1));
                    }
                    if (dbVersions.size() > 1) {
                        System.out.print("Previous database schema versions: ");
                        for (int i = 0; i < dbVersions.size() - 1; i++) {
                            System.out.print(i == 0 ? "" : ", ");
                            System.out.print(dbVersions.get(i));
                        }
                        System.out.println();
                    }
                }
                System.out.println("Component versions: \n" + getRepositoryVersion(getJarFile()));
                return;
            }

            if (connectionMgr == null) {
                LOG.severe("unable to connect to database");
                return;
            }

            modelRegistry = ModelHelper.getSOAModelRegistry();
            ModelRegistry.setDefaultRegistry(modelRegistry);

            String v = props.get("depot.startup.messages");
            boolean verbose = v != null && "true".equals(v);

            if (verbose) {                
                System.out.print(APPLICATION_NAME);
                v = props.get("depot.build.version");                    
                System.out.print(" (" + (v == null ? "?" : v));
                v = props.get("depot.build.timestamp");        
                if (v != null) {
                    System.out.print(", " + v);
                }
                System.out.println(")");
            }

            if (createDatabase(connectionMgr, configService.getHttpConfiguration(), verbose)) {
                init(props);
                if (verbose) {
                    System.out.println(APPLICATION_NAME + " ready...");
                }
            }
        } catch (Exception e) {
            LOG.log(Level.SEVERE, "Error initializing server", e);
            exitStatus.setCode(1);
            exitStatus.setMessage("Exception caught during the server initialization : " + e.getMessage());
            exitStatus.setException(e);
            LOG.severe(exitStatus.getMessage());
            stop();
        }
    }

    private void initializeEMF() {
        System.setProperty("org.eclipse.emf.ecore.EPackage.Registry.INSTANCE",
                           "org.eclipse.emf.ecore.impl.EPackageRegistryImpl");
    }

    public void createNotificationEngine(Map<String, String> props) throws Exception {
        notificationEngine = new NotificationEngine(repository);
        notificationEngineAsObserver = new DataChangeEventObserver() {
            public void dataChanged(DataChangeEvent event) {
                notificationEngine.dataChanged(event);
            }
        };
        DataChangeAdapter.getInstance().registerObserver(notificationEngineAsObserver);
        notificationEngine.start(props);
    }

    private boolean createDatabase(JDBCConnectionManager mgr, HttpConfiguration httpConfig, boolean verbose) 
        throws Exception {
        
        boolean println = false;
        try {
            DBFactory.setLogger(new PrintStream(new LoggerOutputStream(LOG, Level.FINE)));

            VersionType codeSchemaVersion = SchemaVersionChecker.getCodeSchemaVersion();
            if (codeSchemaVersion != null) {
                LOG.config(APPLICATION_NAME + " model schema version " + codeSchemaVersion);

                if (!DBCreate.tablesAlreadyCreated(mgr)) {
                    if (verbose) {
                        System.out.print("Initializing database...");
                        println = true;
                    }
                    LOG.fine("Creating db tables");
                    DBCreate.createTables(mgr, modelRegistry);
                    LOG.fine("Populating database");
                    DBCreate.createNetwork(codeSchemaVersion, mgr, modelRegistry, httpConfig);
                    if (verbose) {
                        System.out.println("done.");
                    }
                } else {
                    VersionType dbSchemaVersion;
                    PooledConnection pc = mgr.getConnection();
                    try {
                        dbSchemaVersion = SchemaVersionChecker.getSchemaVersionFromDb(pc.getConnection());
                    } finally {
                        mgr.returnConnection(pc);
                    }
                    if (SchemaVersionChecker.isCompatible(codeSchemaVersion, dbSchemaVersion)) {
                        LOG.config(APPLICATION_NAME + " database schema version " + dbSchemaVersion);
                        return true;
                    } else {
                        LOG.severe("Incompatible repository schema version found in existing database"
                                   + ": is " + dbSchemaVersion + ", server requires " + codeSchemaVersion);
                        return false;
                    }
                }
            } else {
                LOG.severe("No schema version available to Repository Server.");
                return false;
            }
        } catch (Exception ex) {
            if (verbose && println) {
                System.out.println();
            }
            LOG.log(Level.SEVERE, "Error initialising database", ex);
            throw ex;
        }
        return true;
    }

    private JDBCConnectionManager createJDBCConectionManager(Map<String, String> props) throws Exception {
        String url = props.get(DB_URL_PROP);
        String driver = props.get(DB_DRIVER_PROP);
        if (null == url || null == driver) {
            LOG.warning("database connectivity required properties are missing: " + DB_URL_PROP + " and "
                        + DB_DRIVER_PROP);
            return null;
        }

        String user = props.get(USERNAME_PROP);
        String password = props.get(PASSWORD_PROP);
        if (password == null) {
            // check password provider
            password = PasswordProviderHandler.getPassword(props.get(PASSWORD_PROP + ".provider"));
        }

        Database db = DatabaseResolver.resolve(driver, url, user, password);

        // store this Database reference in the Map used for resolving resources
        compMap.put(EMBEDDED_DATABASE, db);

        return new MTConnectionManagerImpl(db);
    }


    public File getJarFile() throws Exception {
        String resourceName = "/" + getClass().getName().replace(".", "/") + ".class";
        URL url = getClass().getResource(resourceName);
        if (!URLUtil.isJarURL(url)) {
            return null;
        }

        URL baseUrl = URLUtil.getBaseURLForClass(getClass());
        return new File(baseUrl.toURI());
    }

    public String getRepositoryVersion(File jarFile) throws Exception {
        if (jarFile == null) {
            return "Server is running from class files - no version information available";
        }
        File jarDir = jarFile.getParentFile();
        File[] itFiles = jarDir.listFiles(new FilenameFilter() {
            public boolean accept(File dir, String name) {
                return name.startsWith("it-");
            }
        });

        if (itFiles == null) {
            LOG.warning("Unable to read " + jarDir);
            return "";
        }

        StringBuilder versions = new StringBuilder();
        for (int i = 0; i < itFiles.length; i++) {
            if (i > 0) {
                versions.append("\n");
            }
            String name = itFiles[i].getName();
            if (name.endsWith(".jar")) {
                name = name.substring(0, name.length() - 4);
            }

            Date date = new Date(itFiles[i].lastModified());
            DateFormat dForm = DateFormat.getDateTimeInstance();
            Object[] args = {dForm.format(date), name};
            versions.append(VERSION_MESSAGE_FMT.format(args));
        }
        return versions.toString();
    }

    public void stop() {
        LOG.info("stopping server");
        synchronized (RepositoryServiceImpl.class) {
            if (notificationEngine != null) {
                DataChangeAdapter.getInstance().unregisterObserver(notificationEngineAsObserver);
                notificationEngine.stop();
            }
            if (connectionMgr != null) {
                LOG.fine("stopping database");
                connectionMgr.shutDown();
            }
        }
        LOG.info("server stopped");
    }
    
    public BundleContext getBundleContext() {
        return osgiBundleContext;
    }
    
    public void setBundleContext(BundleContext ctx) {
        osgiBundleContext = ctx;
    }

    public SearchService getSearchService() {
        return searchService;
    }

    // Injected by Spring
    public void setSearchService(SearchService searchServ) {
        searchService = searchServ;
        if (repository != null) {
            searchServ.setSOARepository(repository);
        }
    }

    public SecurityService getSecurityService() {
        return securityService;
    }

    // Injected by Spring
    public void setSecurityService(SecurityService secServ) {
        securityService = secServ;
        if (repository != null) {
            securityService.setSOARepository(repository);
            repository.registerTransactionAdapter(new ChangeAuthorizationAdapter(securityService));
        }
    }
    
    // Injected by Spring
    public void setConfigurationService(ConfigurationService c) {
        configService = c;
    }

}
