/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;

import com.iona.soa.repository.util.VersionType;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;

public final class ModelRegistry {
    private static ModelRegistry defaultRegistry;
    private final Map<String, EPackage> registeredModelPackages = new LinkedHashMap<String, EPackage>(); 
    private final Map<String, EClass> classNameMap = new HashMap<String, EClass>();
    private final Map<String, EEnum> enumNameMap = new HashMap<String, EEnum>();
    private VersionType schemaVersion;
    
    public ModelRegistry(EPackage ... packages) {
        this(Arrays.asList(packages));
    }

    public ModelRegistry(VersionType schemaVersion, EPackage ... packages) {
        this(schemaVersion, Arrays.asList(packages));
    }

    public ModelRegistry(Collection<EPackage> packages) {
        this(new VersionType(1, 0), packages);
    }
    
    @SuppressWarnings("unchecked")
    public ModelRegistry(VersionType schemaVersion, Collection<EPackage> packages) {
        this.schemaVersion = schemaVersion;
        Map<EPackage, Set<EPackage>> packageMap = new HashMap<EPackage, Set<EPackage>>();
        for (EPackage ePackage : packages) {
            packageMap.put(ePackage, new HashSet<EPackage>());
            for (EClass eClass : getEClasses(ePackage)) {
                classNameMap.put(eClass.getName(), eClass);

                for (EClass superClass : (Collection<EClass>)eClass.getESuperTypes()) {
                    EPackage superPkg = superClass.getEPackage();
                    if (packages.contains(superPkg) && superPkg != ePackage) {
                        packageMap.get(ePackage).add(superPkg);
                    }
                }
                for (EReference ref : (Collection<EReference>)eClass.getEReferences()) {
                    EPackage refPkg = ref.getEReferenceType().getEPackage();
                    if (packages.contains(refPkg) && refPkg != ePackage) {
                        packageMap.get(ePackage).add(refPkg);
                    }
                }
            }
            for (EEnum eEnum : getEEnums(ePackage)) {
                enumNameMap.put(eEnum.getName(), eEnum);
            }
        }
        while (!packageMap.isEmpty()) {
            Collection<EPackage> rootPackages = new LinkedList<EPackage>();
            for (EPackage ePackage : packageMap.keySet()) {
                if (packageMap.get(ePackage).isEmpty()) {
                    rootPackages.add(ePackage);
                }
            }
            if (rootPackages.isEmpty()) {
                throw new RuntimeException("Circular EPackage dependencies!");
            }
            for (EPackage ePackage : rootPackages) {
                packageMap.remove(ePackage);
                registeredModelPackages.put(ePackage.getNsURI(), ePackage);
            }
            for (EPackage ePackage : packageMap.keySet()) {
                packageMap.get(ePackage).removeAll(rootPackages);
            }
        }
    }

    public static ModelRegistry getDefaultRegistry() {
        if (defaultRegistry != null) {
            return defaultRegistry;
        }
        Collection<EPackage> packages = new HashSet<EPackage>();
        do {
            for (Object obj : new ArrayList<Object>(EPackage.Registry.INSTANCE.values())) {
                if (obj instanceof EPackage) {
                    packages.add((EPackage)obj);
                } else if (obj instanceof EPackage.Descriptor) {
                    EPackage.Descriptor ePackageDescriptor = (EPackage.Descriptor)obj;
                    packages.add(ePackageDescriptor.getEPackage());
                }
            }
        } while(EPackage.Registry.INSTANCE.values().size() != packages.size());
        return new ModelRegistry(packages);
    }    
    
    public static void setDefaultRegistry(ModelRegistry reg) {
        defaultRegistry = reg;
    }

    public static Collection<EClass> getEClasses(EPackage ePackage) {
        Collection<EClass> ret = new ArrayList<EClass>(ePackage.eContents().size());
        for (Object obj : ePackage.eContents()) {
            if (obj instanceof EClass) {
                ret.add((EClass)obj);
            }
        }
        return ret;
    }

    public static Collection<EEnum> getEEnums(EPackage ePackage) {
        Collection<EEnum> ret = new ArrayList<EEnum>(ePackage.eContents().size());
        for (Object obj : ePackage.eContents()) {
            if (obj instanceof EEnum) {
                ret.add((EEnum)obj);
            }
        }
        return ret;
    }
    
    public Collection<EPackage> getPackages() {
        return registeredModelPackages.values();
    }
    
    public EPackage getPackage(String nsURI) {
        return registeredModelPackages.get(nsURI);
    }
    
    public Collection<EClass> getAllEClasses() {
        return classNameMap.values();
    }
    
    public EClass getEClass(String className) {
        return classNameMap.get(className);
    }

    public EClass getEClass(EClassifierDescriptor descriptor) {
        EClass eClass = getEClass(descriptor.getName());
        if (eClass == null) {
            return null;
        } else if (eClass.getEPackage().getNsURI().equals(descriptor.getPackageNsURI())) {
            return eClass;
        } else {
            EPackage ePkg = getPackage(descriptor.getPackageNsURI());
            if (ePkg != null) {
                for (EClass pkgClass : ModelRegistry.getEClasses(ePkg)) {
                    if (pkgClass.getName().equals(descriptor.getName())) {
                        return pkgClass;
                    }
                }
            }
            return null;
        }
    }

    public Collection<EEnum> getAllEEnums() {
        return enumNameMap.values();
    }
    
    public EEnum getEEnum(String enumName) {
        return enumNameMap.get(enumName);
    }

    public EEnum getEEnum(EClassifierDescriptor descriptor) {
        EEnum eEnum = getEEnum(descriptor.getName());
        if (eEnum == null) {
            return null;
        } else if (eEnum.getEPackage().getNsURI().equals(descriptor.getPackageNsURI())) {
            return eEnum;
        } else {
            EPackage ePkg = getPackage(descriptor.getPackageNsURI());
            if (ePkg != null) {
                for (EEnum pkgEnum : ModelRegistry.getEEnums(ePkg)) {
                    if (pkgEnum.getName().equals(descriptor.getName())) {
                        return pkgEnum;
                    }
                }
            }
            return null;
        }
    }

    public VersionType getSchemaVersion() {
        return schemaVersion;
    }
}
