/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Collections;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;
import java.util.jar.Attributes;
import java.util.jar.JarFile;
import java.util.jar.Manifest;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.soa.repository.util.InputStreamHelper;
import com.iona.soa.repository.util.SQLHelper;
import com.iona.soa.repository.util.URLUtil;
import com.iona.soa.repository.util.VersionType;

public final class SchemaVersionChecker {
    public static final String REPOSITORY_SCHEMA_PROPERTY = "Repository-Schema-Version";
    public static final VersionType DEFAULT_SCHEMA_VERSION = new VersionType(1, 0);
    private static final Logger LOGGER = Logger.getLogger(SchemaVersionChecker.class.getName());

    private SchemaVersionChecker() {
    }
    
    /**
     * Get the schema version from the manifest of the soamodel jar.
     * The ClassLoader that this class was loaded with is used to locate the manifest.
     * @return a string containing the version number of the schema as embedded in 
     * the manifest, or <tt>null</tt> if no version information exists
     * @throws IOException 
     */
    public static VersionType getCodeSchemaVersion() {
        Class clazz;
        try {
            clazz = Class.forName("com.iona.soa.model.repository.ModelMetadata");
        } catch (ClassNotFoundException ex) {
            LOGGER.severe("Unable to locate model class ModelMetadata");
            return null;
        }
        URL baseUrl = URLUtil.getBaseURLForClass(clazz);
        if (baseUrl == null) {
            LOGGER.severe("Unable to locate model class ModelMetadata");
            return null;
        } else {
            return getCodeSchemaVersion(clazz.getClassLoader());
        }
    }    
    
    private static VersionType getCodeSchemaVersion(ClassLoader classLoader) {
        try {
            Enumeration<URL> resources = classLoader.getResources("META-INF/MANIFEST.MF");
            while (resources.hasMoreElements()) {
                URL mfURL = resources.nextElement();
                InputStream is = null;
                try {
                    is = mfURL.openStream();
                    Manifest mf = new Manifest(is);
    
                    Attributes attrs = mf.getMainAttributes();
                    String val = attrs.getValue(REPOSITORY_SCHEMA_PROPERTY);
                    if (val != null) {
                        LOGGER.log(Level.INFO, "Using model schema version: " + val);
                        return VersionType.parseVersion(val);
                    }
                } finally {
                    InputStreamHelper.close(is);                    
                }
            }
        } catch (IOException e) {
            try {
                LOGGER.log(Level.SEVERE, "Cannot open model manifest file from location: " 
                        + Collections.list(classLoader.getResources("META-INF/MANIFEST.MF")));
            } catch (IOException ioe) {
                ioe.printStackTrace();
            }
        }
        LOGGER.log(Level.SEVERE, "Could not find the model version, reverting to default version: " 
                + DEFAULT_SCHEMA_VERSION);
        return DEFAULT_SCHEMA_VERSION;        
    }

    /**
     * Get the schema version from the manifest of the soamodel jar.
     * The specified ClassLoader is used to locate the manifest.
     * @return a string containing the version number of the schema as embedded in 
     * the manifest, or <tt>null</tt> if no version information exists
     */
    public static VersionType getCodeSchemaVersion(URL resourceURL) {
        File resourceFile = null;
        try {
            resourceFile = new File(resourceURL.toURI());
        } catch (Exception ex) {
            LOGGER.log(Level.SEVERE, "Unable to locate file for URL " + resourceURL, ex);
            return null;
        }
        
        if (resourceFile == null || !resourceFile.exists()) {
            LOGGER.severe("Unable to locate file for URL " + resourceURL);
            return null;
        } else if (resourceFile.isFile()) {
            try {
                JarFile jarFile = new JarFile(resourceFile);
                Manifest mf = jarFile.getManifest();
                Attributes attrs = mf.getMainAttributes();
                String val = attrs.getValue(REPOSITORY_SCHEMA_PROPERTY);
                if (val != null) {
                    return VersionType.parseVersion(val);
                }
            } catch (Exception e) {
                // ignore
            }
            LOGGER.warning("Cannot extract schema version from manifest (" + resourceFile + ")");
        } else {    
            File manifest = new File(resourceFile, "META-INF/MANIFEST.MF");
            InputStream mfStream = null;
            try {
                if (manifest.exists() && manifest.canRead()) {
                    // The manifest is available in the file system with the
                    // classes, search for the version there.
                    mfStream = new FileInputStream(manifest);
                    Properties mfProps = new Properties();
                    mfProps.load(mfStream);
                    if (mfProps.containsKey(REPOSITORY_SCHEMA_PROPERTY)) {
                        return VersionType.parseVersion(mfProps.getProperty(REPOSITORY_SCHEMA_PROPERTY));
                    }
                }
                
            } catch (IOException e) {
                // ignore
            } finally {
                InputStreamHelper.close(mfStream);
            }
            LOGGER.warning("Cannot extract schema version from manifest "
                           + manifest + "- server is running with class files");
        }
        return DEFAULT_SCHEMA_VERSION;
    }
        
    public static boolean isCompatible(VersionType v1, VersionType v2) {
        int[] i1 = new int[] {
                v1.getMajor() == null ? 0 : v1.getMajor(), 
                v1.getMinor() == null ? 0 : v1.getMinor(), 
        };
        int[] i2 = new int[] {
                v2.getMajor() == null ? 0 : v2.getMajor(), 
                v2.getMinor() == null ? 0 : v2.getMinor(), 
        };
        return i1[0] == i2[0] && i1[1] == i2[1];
    }

    public static VersionType getSchemaVersionFromDb(Connection conn) {
        List<VersionType> allVersions = getAllSchemaVersionsFromDb(conn);
        if (allVersions.size() > 0) {
            return allVersions.get(allVersions.size() - 1);
        }
        return null;
    }
    
    public static List<VersionType> getAllSchemaVersionsFromDb(Connection conn) {
        ResultSet rs = null;
        List<VersionType> ret = new LinkedList<VersionType>();
        try {
            Statement stmt = conn.createStatement();
            stmt.execute("select * from " + SQLHelper.getNonSQLKeyword("MODELMETADATA"));
            rs = stmt.getResultSet();
            
            while (rs.next()) {
                VersionType version = null;
                try {
                    String versionVal = rs.getString(SQLHelper.getNonSQLKeyword("SCHEMAVERSION"));
                    version = VersionType.parseVersion(versionVal);
                } catch (Exception ex) {
                    // ignore
                }
                if (version != null) {
                    ret.add(version);
                }
            }
        } catch (SQLException e) {
            LOGGER.log(Level.SEVERE, "Error accessing schema information in database", e);
        }  finally {
            if (rs != null) {
                try {
                    rs.close();
                } catch (SQLException ex) {
                    // ignore
                }
            }
        }
        
        Collections.sort(ret);
        
        return ret;
    }
}
