/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.status;

import java.io.StringReader;
import java.util.ArrayList;
import java.util.List;

import com.iona.soa.repository.util.Base64;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Verifier;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;

public final class StatusSerializer {
    private static final String TAG_ROOT = "reply";
    private static final String TAG_DETAILS = "details";
    private static final String ATTR_LEVEL = "level";
    private static final String TAG_TYPE = "type";
    private static final String TAG_MESSAGE = "message";
    private static final String TAG_TRACE = "trace";
    private static final String TAG_HREF = "href";
    private static final String ATTR_ENCODING = "encoding";
    private static final String BASE64_ENC = "base64";

    private StatusSerializer() {
    }

    public static String serialize(RepositoryStatus status) {
        Element root = new Element(TAG_ROOT);
        root.setAttribute(ATTR_LEVEL, status.getLevel().name());
        if (status.getType() != null) {
            Element e = new Element(TAG_TYPE);
            e.setText(status.getType());
            root.addContent(e);
        }
        for (RepositoryStatusMessage statusMessage : status.getMessages()) {
            Element statusElem = new Element(TAG_DETAILS);
            if (statusMessage.getMessage() != null) {
                Element e = new Element(TAG_MESSAGE);
                encodeElement(e, statusMessage.getMessage());
                statusElem.addContent(e);
            }
            if (statusMessage.getTrace() != null) {
                Element e = new Element(TAG_TRACE);
                encodeElement(e, statusMessage.getTrace());
                statusElem.addContent(e);
            }
            if (statusMessage.getHrefs() != null) {
                for (String href : statusMessage.getHrefs()) {
                    Element e = new Element(TAG_HREF);
                    e.setText(href);
                    statusElem.addContent(e);
                }
            }
            root.addContent(statusElem);
        }
        return new XMLOutputter().outputString(root);
    }

    @SuppressWarnings("unchecked")
    public static RepositoryStatus deserialize(String msg) {
        try {
            StringReader reader = new StringReader(msg);
            Document doc = new SAXBuilder().build(reader);
            Element root = doc.getRootElement();
            MessageLevel level = MessageLevel.valueOf(root.getAttributeValue(ATTR_LEVEL));
            String type = root.getChildText(TAG_TYPE);
            List<Element> detailElems = root.getChildren(TAG_DETAILS);
            RepositoryStatusMessage[] statusMessages = new RepositoryStatusMessage[detailElems.size()];
            int index = 0;
            for (Element details : detailElems) {
                String message = decodeElement(details.getChild(TAG_MESSAGE));
                String trace = decodeElement(details.getChild(TAG_TRACE));
                List<String> hrefs = new ArrayList<String>(details.getChildren(TAG_HREF).size());
                for (Element href : (List<Element>)details.getChildren(TAG_HREF)) {
                    hrefs.add(href.getText());
                }
                statusMessages[index++] = new RepositoryStatusMessage(message, trace, hrefs);
            }
            return new RepositoryStatus(level, type, statusMessages);
        } catch (Exception ex) {
            RepositoryStatusMessage statusMessage = new RepositoryStatusMessage(ex);
            RepositoryStatus status = 
                new RepositoryStatus(MessageLevel.ERROR, StatusType.INVALID_REPLY, statusMessage);
            return status;
        }
    }
    
    private static void encodeElement(Element elem, String text) {
        if (Verifier.checkCharacterData(text) != null) {
            byte[] bytes;
            try {
                bytes = text.getBytes("UTF-8");
            } catch (Exception ex) {
                bytes = text.getBytes();
            }
            text = Base64.encodeBytes(bytes);
            elem.setAttribute(ATTR_ENCODING, BASE64_ENC);
        }
        elem.setText(text);
    }
    
    private static String decodeElement(Element elem) {
        if (elem != null) {
            String text = elem.getText();
            if (BASE64_ENC.equals(elem.getAttributeValue(ATTR_ENCODING))) {
                byte[] bytes = null;
                try {
                    bytes = Base64.decode(text);
                    text = new String(bytes, "UTF-8");
                } catch (Exception ex) {
                    text = new String(bytes == null ? new byte[0] : bytes);
                }
            }
            return text;
        }
        return null;
    }

    public static RepositoryException deserialize(int httpCode, String msg) {
        RepositoryStatus status = deserialize(msg);
        return new RepositoryException(httpCode, status);
    }
}
