/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.util;

import java.lang.ref.Reference;
import java.lang.ref.ReferenceQueue;
import java.lang.ref.SoftReference;
import java.util.LinkedList;
import java.util.List;

public abstract class AbstractObjectCache<T> {
    private List<Reference<T>> secondaryCache = new LinkedList<Reference<T>>();
    private ReferenceQueue<T> refQueue = new ReferenceQueue<T>();
    private T[] primaryCache;
    private int primaryIndex;
    private int secondaryMaxSize;
    
    @SuppressWarnings("unchecked")
    public AbstractObjectCache(int primaryCacheSize) {
        this(primaryCacheSize, -1);
    }

    
    @SuppressWarnings("unchecked")
    public AbstractObjectCache(int primaryCacheSize, int secondaryCacheSize) {
        primaryCache = (T[])new Object[primaryCacheSize];
        primaryIndex = 0;
        secondaryMaxSize = secondaryCacheSize;
    }

    public AbstractObjectCache(int primaryCacheSize, T ... preallocated) {
        this(primaryCacheSize);
        if (preallocated != null) {
            for (T t : preallocated) {
                put(t);
            }
        }
    }

    protected abstract T create();
    
    public T get() {
        T ret = getCachedObject();
        return ret == null ? create() : ret;
    }

    public void put(T t) {
        synchronized (secondaryCache) {
            if (primaryIndex < primaryCache.length) {
                primaryCache[primaryIndex++] = t;
            } else {
                flushQueue();
                if (secondaryMaxSize < 0 || secondaryCache.size() < secondaryMaxSize) {
                    secondaryCache.add(new SoftReference<T>(t, refQueue));
                }
            }
        }
    }
    
    public List<T> flush() {
        List<T> ret = new LinkedList<T>();
        
        T elem;
        while ((elem = getCachedObject()) != null) {
            ret.add(elem);
        }
        
        return ret;
    }

    private T getCachedObject() {
        T ret = null;
        synchronized (secondaryCache) {
            if (primaryIndex > 0) {
                ret = primaryCache[--primaryIndex];
                primaryCache[primaryIndex] = null;
            } else {
                flushQueue();
                while (ret == null && secondaryCache.size() > 0) {
                    ret = secondaryCache.remove(0).get();
                }
            }
        }
        return ret;
    }

    private void flushQueue() {
        Reference<? extends T> ref;
        while ((ref = refQueue.poll()) != null) {
            secondaryCache.remove(ref);
        }
    }
}
