/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Enumeration;
import java.util.Properties;
import java.util.logging.ConsoleHandler;
import java.util.logging.Formatter;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;

public final class ArgumentParser {
    public static final String PROPERTIES_FILE_ARG = "properties";

    private static final Logger LOG = Logger.getLogger(ArgumentParser.class.getName());

    private String appName;

    public ArgumentParser(String applicationName) {
        appName = applicationName;
    }
    
    public Properties parseArguments(String[] args) throws IOException {
        boolean verbose = false;
        for (String arg : args) {
            if ("-verbose".equals(arg) || "-v".equals(arg)) {
                verbose = true;
                break;
            }
        }
        if (!verbose) {
            Logger log = LOG;
            while (log != null) {
                for (Handler handler : log.getHandlers()) {
                    if (handler instanceof ConsoleHandler) {
                        if (handler.getLevel() == Level.INFO) {
                            handler.setLevel(Level.WARNING);
                        }
                        handler.setFormatter(new VerySimpleFormatter());
                    }
                }
                log = log.getParent();
            }
        }

        Properties props = new Properties();

        String propsFileName = findPropertiesFileName(args);
        if (propsFileName != null) {
            loadCustomPropertiesFile(propsFileName, props);
        } else {
            LOG.warning("No properties file specified");
        }

        // finally overwrite any properties read from files with their
        // command line args equivalents
        for (int i = 0; i < args.length; i++) {
            String arg = args[i];
            if (arg.startsWith("-")) {
                if (i + 1 < args.length && !args[i + 1].startsWith("-")) {
                    String key = arg.substring(1);
                    props.setProperty(key, args[i + 1]);
                } else {
                    props.setProperty(arg, arg);
                }
            }
        }

        if (LOG.isLoggable(Level.CONFIG)) {
            StringBuilder msg = new StringBuilder(appName);
            msg.append(" configuration: ");
            for (Enumeration<?> it = props.propertyNames(); it
                    .hasMoreElements();) {
                String name = (String) it.nextElement();
                String value = props.getProperty(name);
                msg.append(name + "=" + value + "; ");
            }
            LOG.config(msg.toString());
        }
        return props;
    }
    
    private String findPropertiesFileName(String[] args) {
        int i = 0;
        String thePropsSwitch = "-" + PROPERTIES_FILE_ARG;
        for (i = 0; i < args.length - 1; i++) {
            if (thePropsSwitch.equals(args[i])) {
                return args[++i];
            }
        }
        return null;
    }

    private void loadCustomPropertiesFile(String propsFileName,
            Properties props) throws IOException {

        File propsFile = new File(propsFileName);
        if (propsFile.exists()) {
            InputStream is = new FileInputStream(propsFile);
            props.load(is);
            is.close();
            if (appName != null) {
                LOG.info(appName + " loading properties from " + propsFile);
            }
        } else {
            LOG.warning("Properties file does not exist: " + propsFile);
            return;
        }
    }

    private class VerySimpleFormatter extends Formatter {
        private String lineSeparator = System.getProperty("line.separator", "\n");
        
        @Override
        public String format(LogRecord record) {
            StringBuffer sb = new StringBuffer();
            String message = formatMessage(record);
            sb.append(record.getLevel().getLocalizedName());
            sb.append(": ");
            sb.append(message);
            sb.append(lineSeparator);
            if (record.getThrown() != null) {
                try {
                    StringWriter sw = new StringWriter();
                    PrintWriter pw = new PrintWriter(sw);
                    record.getThrown().printStackTrace(pw);
                    pw.close();
                    sb.append(sw.toString());
                } catch (Exception ex) {
                    // ignore
                }
            }
            return sb.toString();
        }

    }
}
