/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.util;

import java.io.ByteArrayOutputStream;
import java.io.Closeable;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Reader;
import java.io.StringWriter;
import java.io.Writer;
import java.nio.charset.Charset;

/** This utility class contains helper functions for dealing with inputstreams.
 */
public final class InputStreamHelper {
    public static final Charset UTF8 = Charset.forName("UTF-8");
    
    static final int BLOCK_SIZE = 16 * 1024;
    
    private InputStreamHelper() {
        
    }
    
    /** Reads everything from an <tt>InputStream</tt> into an <tt>OutputStream</tt>.
     * @param r The <tt>InputStream</tt> to be read from.
     * @param w The <tt>OutputStream</tt> to write to.
     * @throws IOException In case of trouble reading/writing.
     */
    public static void drain(InputStream r, OutputStream w) throws IOException {
        
        byte[] bytes = new byte[BLOCK_SIZE];

        try {
            int length = 0;
            int offset = 0;

            while ((length = r.read(bytes, offset, bytes.length - offset)) != -1) {
                offset += length;

                if (offset == bytes.length) {
                    w.write(bytes, 0, bytes.length);
                    offset = 0;
                }
            }
            if (offset != 0) {
                w.write(bytes, 0, offset);
            }
        } finally {
            bytes = null;
        }
    }

    /** Reads everything from an <tt>Reader</tt> into a <tt>Writer</tt>.
     * @param r The <tt>Reader</tt> to be read from.
     * @param w The <tt>Writer</tt> to write to.
     * @throws IOException In case of trouble reading/writing.
     */
    public static void drain(Reader r, Writer w) throws IOException {
        
        char[] chars = new char[BLOCK_SIZE];

        try {
            int length = 0;
            int offset = 0;

            while ((length = r.read(chars, offset, chars.length - offset)) != -1) {
                offset += length;

                if (offset == chars.length) {
                    w.write(chars, 0, chars.length);
                    offset = 0;
                }
            }
            if (offset != 0) {
                w.write(chars, 0, offset);
            }
        } finally {
            chars = null;
        }
    }

    /** Reads everything from an <tt>InputStream</tt> into a byte array.
     * @param r The <tt>InputStream</tt> to read from. 
     * @return The byte array containing the contents of the input stream.
     * @throws IOException In case of trouble reading.
     */
    public static byte[] drain(InputStream r) throws IOException {
        int avail = r.available();
        ByteArrayOutputStream bytes;
        if (avail > 0) {
            bytes = new ByteArrayOutputStream(avail);
        } else {
            bytes = new ByteArrayOutputStream();
        }
        drain(r, bytes);
        return bytes.toByteArray();
    }
    
    /** Reads everything from an <tt>Reader</tt> into a String.
     * @param r The <tt>Reader</tt> to read from. 
     * @return The String containing the contents of the Reader.
     * @throws IOException In case of trouble reading.
     */
    public static String drain(Reader reader) throws IOException {
        StringWriter writer = new StringWriter();
        drain(reader, writer);
        return writer.toString();
    }
    
    public static void close(Closeable cs) {
        if (cs != null) {
            try {
                cs.close();
            } catch (IOException ioe) {
                // can't really do much with this...
                ioe.printStackTrace();
            }
        }
    }
    
    public static void closeAll(Closeable ... closeables) {
        for (Closeable cs : closeables) {
            close(cs);
        }
    }
}

