/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.util;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.JarInputStream;

import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.Type;
import org.apache.bcel.util.ClassLoaderRepository;

public class JarIntrospector {
    private URLClassLoader classLoader;
    private ClassLoaderRepository repository;
    private List<String> files;
    private List<String> allClasses;
    
    private boolean incompatibleClassesDetected;

    public JarIntrospector(String jarFileOrDirectory) throws IOException {
        this(new File(jarFileOrDirectory));
    }

    public JarIntrospector(File ... jarFilesOrDirectories) throws IOException {
        this(getURLs(jarFilesOrDirectories));
    }
    
    public JarIntrospector(URL ... urls) throws IOException {
        this(new URLClassLoader(urls, JarIntrospector.class.getClassLoader()));
    }

    public JarIntrospector(URLClassLoader classLoader) throws IOException {
        this.classLoader = classLoader;
        repository = new ClassLoaderRepository(classLoader);
    }
    
    public boolean wereIncompatibleClassesDetected() {
        return incompatibleClassesDetected;
    }
 
    public boolean exists(Class clazz) {
        return getAllClasses().contains(clazz.getName());
    }

    public boolean exists(String className) {
        return getAllClasses().contains(className);
    }

    public List<String> getAllClasses() {
        if (allClasses != null) {
            return allClasses;
        }
        
        allClasses = new LinkedList<String>();
        for (Iterator<String> iter = findAllFiles().iterator(); iter.hasNext();) {
            String fileName = iter.next();
            if (fileName.endsWith(".class")) {
                allClasses.add(convertResourceToClassName(fileName));
            }
        }
        return allClasses;
    }
    
    public String[] findClasses(Class clazz) {
        return findClasses(clazz.getName());
    }

    public String[] findClasses(String interfaceName) {
        List<String> ret = new LinkedList<String>();
        Map<String, Boolean> classCache = new HashMap<String, Boolean>();

        incompatibleClassesDetected = false;
        
        for (String className : getAllClasses()) {
            try {
                if (instanceOf(classCache, className, interfaceName)) {
                    ret.add(className);
                }
            } catch (UnsupportedClassVersionError e) {
                incompatibleClassesDetected = true;
            } catch (Throwable t) {
                System.err.println("Problem with " + className + ": " + t);
                t.printStackTrace();
            }
            resetBCEL();
        }
        return ret.toArray(new String[0]);
    }

    private void resetBCEL() {
        classLoader = new URLClassLoader(classLoader.getURLs(), classLoader.getParent());
        repository = new ClassLoaderRepository(classLoader);
    }

    private String convertResourceToClassName(String fileName) {
        return fileName.replaceAll("/", ".").substring(0, fileName.length() - 6);
    }

    protected List<String> findAllFiles() {
        if (files != null) {
            return files;
        }

        files = new LinkedList<String>();
        for (URL url : classLoader.getURLs()) {
            
            if (url.getProtocol().toLowerCase().equals("file")) {
                URI uri = null;
                try { 
                    uri = url.toURI();
                } catch (URISyntaxException ex) {
                    // ignore
                }
                
                File file;
                if (uri != null) {
                    try {
                        file = new File(uri);
                    } catch (IllegalArgumentException ex) {
                        String fileName = uri.toString().substring(5).replace('/', File.separatorChar);
                        file = new File(fileName);
                    }
                } else {
                    file = new File(url.getPath());
                }
                if (file.isDirectory()) {
                    files.addAll(listFilesInDir(file, ""));
                } else {
                    files.addAll(listFilesInJar(file));
                }
            } else {
                files.addAll(listFilesInArchivedJar(url));
            }
        }
        return files;
    }

    public String[] findClassesByMethodSignature(Class<?> returnType, String name, Class<?>[] args)
        throws ClassNotFoundException, UnsupportedClassVersionError {
        try {
            List<String> matchingClasses = new ArrayList<String>();
            MethodSignature sig = new MethodSignature(returnType, name, args);
            for (String fileName : findAllFiles()) {
                if (fileName.endsWith(".class")) {
                    String className = convertResourceToClassName(fileName);
                    JavaClass clz = repository.loadClass(className);
                    for (Method m : clz.getMethods()) {
                        if (sig.match(m)) {
                            matchingClasses.add(className);
                        }
                    }
                }
            }
            return matchingClasses.toArray(new String[matchingClasses.size()]);
        } finally {
            resetBCEL();
        }
    }

    public boolean isInterface(String className) {
        // try classloader first
        try {
            Class clazz = classLoader.loadClass(className);
            return clazz != null && clazz.isInterface();
        } catch (ClassNotFoundException cnfe) {
            // ignore
        } catch (NoClassDefFoundError ncdfe) {
            // ignore
        }

        // now try bytecode analysis with BCEL
        try {
            JavaClass clazz = repository.loadClass(className);
            return clazz != null && clazz.isInterface();
        } catch (ClassNotFoundException cnfe) {
            // ignore 
        } finally {
            resetBCEL();
        }

        return false;
    }

    
    private boolean instanceOf(Map<String, Boolean> cache, String className, final String classOrInterface) {
        if (cache.containsKey(className)) {
            return cache.get(className);
        } else if (className.equals(classOrInterface)) {
            cache.put(className, true);
            return true;
        }

        // try classloader first
        try {
            Class clazz = classLoader.loadClass(className);
            if (clazz.getSuperclass() != null) {
                String superClass = clazz.getSuperclass().getName();
                if (instanceOf(cache, superClass, classOrInterface)) {
                    cache.put(className, true);
                    return true;
                }
            }
            Class[] interfaces = clazz.getInterfaces();
            for (int i = 0; i < interfaces.length; i++) {
                if (instanceOf(cache, interfaces[i].getName(), classOrInterface)) {
                    cache.put(className, true);
                    return true;
                }
            }
        } catch (ClassNotFoundException cnfe) {
            // ignore
        } catch (NoClassDefFoundError ncdfe) {
            // ignore
        }

        // now try bytecode analysis with BCEL
        boolean ret = false;
        try {
            JavaClass clazz = repository.loadClass(className);
            String superName = clazz.getSuperclassName();
            if (superName != null && !className.equals(superName)
                && instanceOf(cache, superName, classOrInterface)) {
                ret = true;
            } else {
                String[] intfNames = clazz.getInterfaceNames();
                for (int i = 0; i < intfNames.length; i++) {
                    if (!className.equals(intfNames[i]) 
                        && instanceOf(cache, intfNames[i], classOrInterface)) {
                        ret = true;
                        break;
                    }
                }
            }
        } catch (ClassNotFoundException cnfe) {
            // ignore 
        }
        cache.put(className, ret);
        return ret;
    }

    private List<String> listFilesInDir(File dir, String prefix) {
        List<String> allFilesInDir = new LinkedList<String>();
        File[] allFiles = dir.listFiles();
        for (int i = 0; i < allFiles.length; i++) {
            if (allFiles[i].isFile()) {
                allFilesInDir.add(prefix + allFiles[i].getName());
            } else if (allFiles[i].isDirectory()) {
                allFilesInDir.addAll(listFilesInDir(allFiles[i], prefix + allFiles[i].getName() + "/"));
            }
        }
        return allFilesInDir;
    }

    private List<String> listFilesInJar(File jar) {
        List<String> filesInJar = new LinkedList<String>();
        JarFile jarFile = null;
        try {
            jarFile = new JarFile(jar);
            Enumeration<JarEntry> entries = jarFile.entries();
            while (entries.hasMoreElements()) {
                JarEntry entry = entries.nextElement();
                if (!entry.isDirectory()) {
                    filesInJar.add(entry.getName());
                }
            }
        } catch (IOException ioe) {
            // ignore
        } finally {
            if (jarFile != null) {
                try {
                    jarFile.close();
                } catch (IOException ioe) {
                    // ignore
                }
            }
        }
        return filesInJar;
    }
    
    private List<String> listFilesInArchivedJar(URL jarUrl) {
        List<String> filesInJar = new LinkedList<String>();
        JarInputStream jarFile = null;
        try {
            jarFile = new JarInputStream(jarUrl.openStream());
            JarEntry entry = jarFile.getNextJarEntry();
            while (entry != null) {
                if (!entry.isDirectory()) {
                    filesInJar.add(entry.getName());
                }
                entry = jarFile.getNextJarEntry();
            }
        } catch (IOException ioe) {
            // ignore
        } finally {
            if (jarFile != null) {
                try {
                    jarFile.close();
                } catch (IOException ioe) {
                    // ignore
                }
            }
        }
        return filesInJar;
    }

    private static URL[] getURLs(File ... files) throws IOException {
        URL[] urls = new URL[files.length];
        for (int i = 0; i < urls.length; i++) {
            urls[i] = files[i].toURI().toURL();
        }
        return urls;
    }

    static class MethodSignature {
        private final Type returnType;
        private final String methodName;
        private final Type[] argTypes;

        public MethodSignature(Class<?> ret, String name, Class<?>[] args) {
            returnType = Type.getType(ret);
            methodName = name;
            argTypes = new Type[args.length];
            for (int i = 0; i < args.length; i++) {
                argTypes[i] = Type.getType(args[i]);
            }
        }

        public Type[] getArgTypes() {
            return argTypes;
        }

        public String getMethodName() {
            return methodName;
        }

        public Type getReturnType() {
            return returnType;
        }

        public boolean match(Method m) {

            boolean match = false;
            if (methodName.equals(m.getName()) && returnType.equals(m.getReturnType())
                && argTypes.length == m.getArgumentTypes().length) {
                match = true;
                for (int i = 0; i < argTypes.length; i++) {
                    if (!argTypes[i].equals(m.getArgumentTypes()[i])) {
                        match = false;
                        break;
                    }
                }

            }

            return match;
        }
    }
}
