/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.util;


import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.math.BigInteger;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

public final class ObjectComparator {
    
    private static final String GETTER_PREFIX = "get";
    private static final String VALUE_OF_METHOD_NAME = "valueOf";
    
    private static final Class BOXED_PRIMITIVES[] = {String.class, Boolean.class, Character.class, 
                                                     BigDecimal.class, BigInteger.class, Byte.class, 
                                                     Double.class, Float.class, Integer.class, 
                                                     Long.class, Short.class};
    
    private static final Map<Class, Class> PRIMITIVE_TO_BOXED_MAP;
    
    static {
        PRIMITIVE_TO_BOXED_MAP = new HashMap<Class, Class>();
        PRIMITIVE_TO_BOXED_MAP.put(boolean.class, Boolean.class);
        PRIMITIVE_TO_BOXED_MAP.put(char.class, Character.class);
        PRIMITIVE_TO_BOXED_MAP.put(byte.class, Byte.class);
        PRIMITIVE_TO_BOXED_MAP.put(double.class, Double.class);
        PRIMITIVE_TO_BOXED_MAP.put(float.class, Float.class);
        PRIMITIVE_TO_BOXED_MAP.put(int.class, Integer.class);
        PRIMITIVE_TO_BOXED_MAP.put(long.class, Long.class);
        PRIMITIVE_TO_BOXED_MAP.put(short.class, Short.class);
    }
    
    private ObjectComparator() {
        // complete
    }
    
    public static boolean equals(Object s, Object t) throws Exception {
        // compare pointers
        if (s == t) {
            return true;
        }
        // compare (null && not null) case
        if (null == s || null == t) {
            return false;
        }
        // compare types
        try {
            s.getClass().cast(t);
        } catch (ClassCastException e) {
            try {
                t.getClass().cast(s);
            } catch (ClassCastException ex) {
                return false;
            }
        }

        
        for (int p = 0; p < BOXED_PRIMITIVES.length; ++p) {
            if (BOXED_PRIMITIVES[p].isAssignableFrom(s.getClass())) {
                return s.equals(t);
            }
        }
                
        Method[] sms = s.getClass().getDeclaredMethods();
        Method[] tms = s.getClass().getDeclaredMethods();
        
        for (int i = 0; i < sms.length; ++i) {
            Method sm = sms[i];
            Method tm = tms[i];
            if (sm.getName().startsWith(GETTER_PREFIX)) {
                Class<?> returnClass = sm.getReturnType();
                Class boxed = PRIMITIVE_TO_BOXED_MAP.get(returnClass);
                if (null != boxed) {
                    if (!comparePrimitives(s, t, sm, tm, returnClass, boxed)) {
                        return false;
                    }
                    continue;
                }
                if (Enum.class.isAssignableFrom(returnClass)) {
                    if (!compareEnums(s, t, sm, tm)) {
                        return false;
                    }
                    continue;
                } 
                if (Collection.class.isAssignableFrom(returnClass)) {
                    if (!compareCollections(s, t, sm, tm)) {
                        return false;
                    }
                    continue;
                }
                if (Object.class.isAssignableFrom(returnClass)) {
                    if (!compareObjects(s, t, sm, tm)) {
                        return false;
                    }
                }
            }
        }
        return true;
    }
    
    private static boolean compareEnums(Object s, Object t, Method sm, Method tm) throws Exception {
        Enum sv = (Enum)sm.invoke(s, (Object[])null);
        Enum tv = (Enum)tm.invoke(t, (Object[])null);
        if (sv == tv) {
            return true;
        }
        if (null == sv || null == tv) {
            return false;
        }        
        return 0 == sv.compareTo(tv);
    }

    private static boolean compareCollections(Object s, Object t, Method sm, Method tm) throws Exception {
        Collection sc = (Collection)sm.invoke(s, (Object[])null);
        Collection tc = (Collection)tm.invoke(t, (Object[])null);
        
        if (null != sc && null != tc && sc.size() != tc.size()) {
            return false;
        }
        
        if (null != sc && null != tc) {
            Iterator sit = sc.iterator();
            Iterator tit = tc.iterator();
            while (sit.hasNext()) {
                Object so = sit.next(); 
                Object to = tit.next();
                if (!equals(so, to)) {
                    return false;
                }
            }
        }
        return true;
    }

    private static boolean comparePrimitives(Object s, 
                                             Object t, 
                                             Method sm, 
                                             Method tm, 
                                             Class primitive, 
                                             Class boxed) throws Exception {
        Method valueOf = boxed.getMethod(VALUE_OF_METHOD_NAME, new Class[] {primitive});
        Object sv = valueOf.invoke(null, sm.invoke(s, (Object[])null));
        Object tv = valueOf.invoke(null, tm.invoke(t, (Object[])null));
        return equals(sv, tv);
    }
    
    private static boolean compareObjects(Object s, Object t, Method sm, Method tm) throws Exception {
        Object sv = sm.invoke(s, (Object[])null);
        Object tv = tm.invoke(t, (Object[])null);
        return equals(sv, tv);
    }
}

