/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.util;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

public final class SQLHelper {

    private static final Logger LOG = Logger.getLogger(SQLHelper.class.getName()); 
    private static final String KEYWORDS_FILE = "/sql92keywords.txt";
    private static final String COMMENT_CHAR = "#";
    private static final String MANGLE_SUFFIX = "1";
    private static final int MAX_SQL_LENGTH = 30;
    
    private static Set<String> sqlKeywords = new HashSet<String>();

    static {
        loadSQLKeywords();
    }
    

    private SQLHelper() {
        //empty
    }
    
    public static String cleanSQLKeyword(String kw) {
        String ret = kw;
        if (kw.endsWith(MANGLE_SUFFIX)) {
            // get the root keyword without any mangling
            ret = kw.substring(0, kw.length() - MANGLE_SUFFIX.length());
            // double check that the demangled word is in our list of keywords. 
            // If it's not then we did not mangle it in the first place
            if (!sqlKeywords.contains(ret.toUpperCase())) {
                ret = kw;
            }
        }
        return ret;
    }
    
    public static void addSQLKeywords(Connection conn) {
        try {
            String[] keywords = conn.getMetaData().getSQLKeywords().split(",");
            for (String kw : keywords) {
                addKeyword(kw);
            }
        } catch (SQLException ex) {
            LOG.log(Level.WARNING, "unable to get keywords from JDBC connection", ex);
        }
        
    }
    
    public static String getNonSQLKeyword(String ... strs) {
        if (strs == null || strs.length == 0) {
            throw new IllegalArgumentException("Need at least 1 argument");
        }
        
        int spaceAvailable = MAX_SQL_LENGTH - strs.length + 1;
        int totalLength = 0;
        for (String str : strs) {
            int len = str.length();
            if (len <= 2) {
                spaceAvailable -= len;
            } else {
                totalLength += len;
            }
        }
        float roundFactor = 0.49f;
        while (totalLength > spaceAvailable) {
            float shortenFactor = (float)spaceAvailable / (float)totalLength;
            totalLength = 0;
            for (int i = 0; i < strs.length; i++) {
                int oldSize = strs[i].length();
                if (oldSize <= 2) {
                    continue;
                }
                int newSize = (int)(shortenFactor * oldSize + roundFactor);
                totalLength += newSize;
                if (oldSize != newSize) {
                    strs[i] = strs[i].substring(0, newSize);
                }
            }
            roundFactor -= .1f;
        }
        StringBuffer buf = new StringBuffer(strs[0]);
        for (int i = 1; i < strs.length; i++) {
            buf.append("_");
            buf.append(strs[i]);
        }
        if (sqlKeywords.contains(buf.toString().toUpperCase())) {
            if (buf.length() + MANGLE_SUFFIX.length() > MAX_SQL_LENGTH) {
                buf.setLength(MAX_SQL_LENGTH - MANGLE_SUFFIX.length());
            }
            buf.append(MANGLE_SUFFIX);
        }
        return buf.toString().toUpperCase();
    }

    public static void close(Statement stmt) {
        try {
            if (stmt != null) {
                stmt.close();
            }
        } catch (SQLException ex) {
            LOG.log(Level.WARNING, "error closing statement", ex);
        }
    }
    
    private static void loadSQLKeywords() {
        InputStream in = SQLHelper.class.getResourceAsStream(KEYWORDS_FILE);
        try {
            BufferedReader reader = new BufferedReader(new InputStreamReader(in));
            String line = null;
            while ((line = reader.readLine()) != null) {
                addKeyword(line);
            }
        } catch (IOException ex) {
            ex.printStackTrace();
        } finally { 
            InputStreamHelper.close(in);
        }
    }
    
    private static void addKeyword(String kw) {
        String entry = kw.trim().toUpperCase();
        if (!entry.startsWith(COMMENT_CHAR) && !sqlKeywords.contains(entry)) {
            sqlKeywords.add(entry);
        }
        
    }
}
