/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.util;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

public final class TempUtil {
    public interface CleanupTask {
        void run();
    }
    
    private static CleanupThread cleaner;
    private static final File SYS_TEMP_DIR = new File(System.getProperty("java.io.tmpdir", "."));
    private static final String PREFIX = "ionatmp";
        
    private TempUtil() {
    }
    
    public static File createTempDir() {
        return createTempDir(true, null);
    }

    public static File createTempDir(CleanupTask task) {
        return createTempDir(true, task);
    }

    public static File createTempDir(boolean automaticCleanup) {
        return createTempDir(automaticCleanup, null);
    }
    
    private static synchronized File createTempDir(boolean automaticCleanup, CleanupTask task) {
        File tempDir = null;
        long count = System.currentTimeMillis();
        do {
            tempDir = new File(SYS_TEMP_DIR, PREFIX + (count++));
        } while (tempDir.exists());
        tempDir.mkdirs();
        if (automaticCleanup) {
            if (cleaner == null) {
                cleaner = new CleanupThread();
                Runtime.getRuntime().addShutdownHook(cleaner);
            }
            cleaner.addDir(tempDir);
            if (task != null) {
                cleaner.addTask(task);
            }
        }
        return tempDir;
    }
    
    public static File createTempFile() throws IOException {
        return createTempFile(true);
    }
    
    public static File createTempFile(boolean automaticCleanup) throws IOException {
        return createTempFile(PREFIX, automaticCleanup);          
    }
    
    public static File createTempFile(String name, boolean automaticCleanup) throws IOException {
        return File.createTempFile(name, null, createTempDir(automaticCleanup));
    }
    
    public static synchronized void removeTempFiles() {
        if (cleaner != null) {
            cleaner.run();
        }
    }
    
    public static synchronized void removeTempFile(File fileOrDir) {
        if (cleaner != null) {
            cleaner.removeDir(fileOrDir);
        }
        try {
            removeAll(fileOrDir);
        } catch (IOException ioe) {
            throw new RuntimeException(ioe);
        }
    }
    
    private static void removeAll(File f) throws IOException {
        if (f.exists()) {
            if (f.isDirectory()) {
                File[] children = f.listFiles();
                for (int i = 0; i < children.length; i++) {
                    removeAll(children[i]);
                }
            }
            if (!f.delete()) {
                throw new IOException("Unable to delete " + f.getAbsolutePath());
            }
        }
    }
    
    private static final class CleanupThread extends Thread {
        private Map<File, String> dirs = new HashMap<File, String>();
        private Collection<CleanupTask> tasks = new LinkedList<CleanupTask>();
        
        public void addDir(File dir) {
            String stack = getStack(new Exception("Directory created at:"));
            dirs.put(dir, stack);
        }
        
        public void removeDir(File dir) {
            dirs.remove(dir);
        }
        
        public void addTask(CleanupTask task) {
            tasks.add(task);
        }

        public void run() {
            for (CleanupTask task : tasks) {
                task.run();
            }
            StringBuffer buf = new StringBuffer();
            for (File dir : dirs.keySet()) {
                try {
                    removeAll(dir);
                } catch (IOException e) {
                    buf.append(getStack(e));
                    buf.append(dirs.get(dir));
                }
            }
            dirs.clear();
            if (buf.length() > 0) {
                throw new RuntimeException(buf.toString());
            }
        }
        
        private String getStack(Throwable t) {
            StringWriter sw = new StringWriter();
            PrintWriter pw = new PrintWriter(sw);
            t.printStackTrace(pw);
            return sw.toString();
        }
    }
}
