/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.util;

import java.util.Arrays;
import java.util.Collection;

/** 
 * <p>
 * Version identifiers have six components.
 * <ol>
 * <li>Major version. A non-negative integer.</li>
 * <li>Minor version. A non-negative integer.</li>
 * <li>Micro version. A non-negative integer.</li>
 * <li>Nano version. A non-negative integer.</li>
 * <li>Qualifier1. A text string. See <code>Version(String)</code> for the
 * format of the qualifier string.</li>
 * </ol>
 * 
 * <p>
 * <code>Version</code> objects are mutable.
 * 
 * @version 1.0
 */

public class VersionType implements Comparable<VersionType> {
    /**
     * The empty version "0". Equivalent to calling
     * <code>new Version(0,null,null,null)</code>.
     */
    public static final VersionType EMPTY_VERSION = new VersionType(0, null, null, null);
    public static final char SEPARATOR = '.';
    private static final Collection<Character> SEPARATOR_CHARS = Arrays.asList(SEPARATOR, '-');

    private Integer major;
    private Integer minor;
    private Integer micro;
    private Integer nano;
    private String qualifier;
    private Integer hashCode;
    
    /**
     * Creates a version identifier from the specified numerical components.
     * 
     * <p>
     * The qualifier is set to the empty string.
     * 
     * @param major Major component of the version identifier.
     * @param minor Minor component of the version identifier.
     * @throws IllegalArgumentException If the numerical components are
     *         negative.
     */
    public VersionType(Integer major, Integer minor) {
        this(major, minor, null, null, null);
    }

    /**
     * Creates a version identifier from the specified numerical components.
     * 
     * <p>
     * The qualifier is set to the empty string.
     * 
     * @param major Major component of the version identifier.
     * @param minor Minor component of the version identifier.
     * @param micro Micro component of the version identifier.
     * @throws IllegalArgumentException If the numerical components are
     *         negative.
     */
    public VersionType(Integer major, Integer minor, Integer micro) {
        this(major, minor, micro, null, null);
    }

    /**
     * Creates a version identifier from the specified numerical components.
     * 
     * <p>
     * The qualifier is set to the empty string.
     * 
     * @param major Major component of the version identifier.
     * @param minor Minor component of the version identifier.
     * @param micro Micro component of the version identifier.
     * @param nano Nano component of the version identifier.
     * @throws IllegalArgumentException If the numerical components are
     *         negative.
     */
    public VersionType(Integer major, Integer minor, Integer micro, Integer nano) {
        this(major, minor, micro, nano, null);
    }

    /**
     * Creates a version identifier from the specifed components.
     * 
     * @param major Major component of the version identifier.
     * @param minor Minor component of the version identifier. If
     *        <code>null</code> is specified, then the base version
     *        is equivalent to <code>emptyVersion</code>
     * @param micro Micro component of the version identifier.
     * @param nano Nano component of the version identifier.
     * @param qualifier1 Qualifier1 component of the version identifier. If
     *        <code>null</code> is specified, then the qualifier will be set
     *        to the empty string.
     * @param qualifier2 Qualifier2 component of the version identifier. If
     *        <code>null</code> is specified, then the qualifier will be set
     *        to the empty string.
     * @throws IllegalArgumentException If the numerical components are negative
     *         or the qualifier string is invalid.
     */
    public VersionType(Integer major, Integer minor, Integer micro, Integer nano, String qualifier) {
        init(major, minor, micro, nano, qualifier);
    }

    /**
     * Created a version identifier from the specified string.
     * 
     * <p>
     * Here is the grammar for version strings.
     * 
     * <pre>
     * version ::= major('.'minor('.'micro('.'nano('.'qualifier1)?)?)?
     * major ::= digit+
     * minor ::= digit+
     * micro ::= digit+
     * nano ::= digit+
     * qualifier1 ::= (alpha|digit|'_'|'@')+
     * digit ::= [0..9]
     * alpha ::= [a..zA..Z]
     * </pre>
     * 
     * There must be no whitespace in version.
     * 
     * @param version String representation of the version identifier.
     * @throws IllegalArgumentException If <code>version</code> is improperly
     *         formatted.
     */
    protected VersionType(String version) {
        Integer[] numbers = new Integer[4];
        StringBuffer quali = null;
        int numberIndex = 0;
        int beginIndex = 0;
        
        for (int endIndex = 0; endIndex <= version.length(); endIndex++) {
            Character c = endIndex == version.length() ? null : version.charAt(endIndex);
            if (c == null || SEPARATOR_CHARS.contains(c)) {
                if (endIndex == 0) {
                    throw new IllegalArgumentException("Incorrectly formatted version. Leading separator: "
                            + version);
                } else if (beginIndex == version.length()) {
                    if (quali == null) {
                        throw new IllegalArgumentException("Incorrectly formatted version. "
                                + "Trailing separator: " + version);
                    }
                }
                String candidate = c == null 
                                   ? version.substring(beginIndex) 
                                   : version.substring(beginIndex, endIndex);
                int intVal = parseIntegerString(candidate);
                if (intVal < 0) {
                    if (numberIndex == 0) {
                        throw new IllegalArgumentException("Incorrectly formatted version. Missing numbers: "
                                                           + version);
                    } else if (quali == null) {
                        quali = new StringBuffer(candidate);
                    } else {
                        quali.append(candidate);
                    }
                    if (c != null) {
                        quali.append(c);
                    }
                } else {
                    if (numberIndex == 4 || quali != null) {
                        throw new IllegalArgumentException("Incorrectly formatted version. Number at index " 
                                                           + beginIndex + ": " + version);
                    } else {
                        numbers[numberIndex++] = intVal;
                    }
                }
                beginIndex = endIndex + 1;
            } else if (Character.isWhitespace(c) || Character.isISOControl(c)) {
                throw new IllegalArgumentException("Incorrectly formatted version. Bad character at index " 
                        + beginIndex + ": " + version);
            }

        }

        init(numbers[0], numbers[1], numbers[2], numbers[3], quali == null ? null : quali.toString());
    }       
    
    private int parseIntegerString(String str) {
        try {
            return Integer.parseInt(str);
        } catch (NumberFormatException nfe) {
            return -1;
        }
    }

    /**
     * Parses a version identifier from the specified string.
     * 
     * <p>
     * See <code>Version(String)</code> for the format of the version string.
     * 
     * @param version String representation of the version identifier. Leading
     *        and trailing whitespace will be ignored.
     * @return A <code>Version</code> object representing the version
     *         identifier. If <code>version</code> is <code>null</code> or
     *         the empty string then <code>emptyVersion</code> will be
     *         returned.
     * @throws IllegalArgumentException If <code>version</code> is improperly
     *         formatted.
     */
    public static VersionType parseVersion(String version) {
        version = version.trim();
        return new VersionType(version);
    }

    /**
     * Returns the major component of this version identifier.
     * 
     * @return The major component.
     */
    public Integer getMajor() {
        return major;
    }

    /**
     * Returns the minor component of this version identifier.
     * 
     * @return The minor component.
     */
    public Integer getMinor() {
        return minor;
    }

    /**
     * Returns the micro component of this version identifier.
     * 
     * @return The micro component.
     */
    public Integer getMicro() {
        return micro;
    }

    /**
     * Returns the nano component of this version identifier.
     * 
     * @return The nano component.
     */
    public Integer getNano() {
        return nano;
    }
    
    /**
     * Returns the qualifier1 component of this version identifier.
     * 
     * @return The qualifier1 component.
     */
    public String getQualifier() {
        return qualifier;
    }

    /**
     * Returns the string representation of this version identifier.
     * 
     * <p>
     * The format of the version string will be <code>major.minor.micro.nano</code>
     * if qualifier1 or qualifier2 are empty strings or
     * <code>major.minor.micro.qualifier1.qualifier2</code> otherwise.
     * 
     * @return The string representation of this version identifier.
     */
    public String toString() {
        String base = major.toString();
        if (minor != null) {
            base += SEPARATOR + minor.toString();
        }
        if (micro != null && micro != 0 || nano != null && nano != 0) {
            base += SEPARATOR + micro.toString();
        }
        if (nano != null && nano != 0) {
            base += SEPARATOR + nano.toString();
        }
        if (qualifier != null) {
            base += SEPARATOR + qualifier;
        }
        return base;
    }
    
    /**
     * Returns a hash code value for the object.
     * 
     * @return An integer which is a hash code value for this object.
     */
    public int hashCode() {
        if (hashCode == null) {
            hashCode = toString().hashCode();
        }
        return hashCode;
    }

    /**
     * Compares this <code>Version</code> object to another object.
     * 
     * <p>
     * A version is considered to be <b>equal to </b> another version if the
     * major, minor, micro and nano components are equal and the qualifier1 and
     * qualifier2 components are equal (using <code>String.equals</code>).
     * 
     * @param object The <code>Version</code> object to be compared.
     * @return <code>true</code> if <code>object</code> is a
     *         <code>Version</code> and is equal to this object;
     *         <code>false</code> otherwise.
     */
    public boolean equals(Object object) {
        if (object == this) { // quicktest
            return true;
        }

        if (!(object instanceof VersionType)) {
            return false;
        }

        VersionType other = (VersionType) object;
        if ((major == null ? 0 : major) != (other.major == null ? 0 : other.major)
            || (minor == null ? 0 : minor) != (other.minor == null ? 0 : other.minor)
            || (micro == null ? 0 : micro) != (other.micro == null ? 0 : other.micro)
            || (nano == null ? 0 : nano) != (other.nano == null ? 0 : other.nano)) {
            return false;
        }
        return qualifier == null ? other.qualifier == null : qualifier.equals(other.qualifier);
    }

    /**
     * Compares this <code>Version</code> object to another object.
     * 
     * <p>
     * A version is considered to be <b>less than </b> another version if its
     * major component is less than the other version's major component, or the
     * major components are equal and its minor component is less than the other
     * version's minor component, or the major and minor components are equal
     * and its micro component is less than the other version's micro component,
     * or the major, minor and micro components are equal and its nano component 
     * is less than the other version's nano component, or the major, minor, micro 
     * and nano components are equal and it's qualifier1 component is less than 
     * the other version's qualifier1 component (using <code>String.compareTo</code>).
     * <p>
     * A version is considered to be <b>equal to</b> another version if the
     * major, minor, micro and nano components are equal and the qualifier1 and qualifier2
     * component2 are equal (using <code>String.compareTo</code>).
     * 
     * @param object The <code>Version</code> object to be compared.
     * @return A negative integer, zero, or a positive integer if this object is
     *         less than, equal to, or greater than the specified
     *         <code>Version</code> object.
     * @throws ClassCastException If the specified object is not a
     *         <code>Version</code>.
     */
    public int compareTo(VersionType other) {
        if (other == this) { // quicktest
            return 0;
        }

        int[] vals1 = new int[] { 
            major == null ? 0 : major, minor == null ? 0 : minor, 
            micro == null ? 0 : micro, nano == null ? 0 : nano };
        int[] vals2 = new int[] { 
            other.major == null ? 0 : other.major, other.minor == null ? 0 : other.minor, 
            other.micro == null ? 0 : other.micro, other.nano == null ? 0 : other.nano };
        for (int i = 0; i < 4; i++) {
            if (vals1[i] - vals2[i] != 0) {
                return vals1[i] - vals2[i];
            }
        }
        if (qualifier != null && other.qualifier != null) {
            return qualifier.compareTo(other.qualifier);
        } else if (qualifier == null && other.qualifier == null) {
            return 0;
        } else {
            return qualifier != null ? 1 : -1;
        }
    }

    private void init(Integer maj, Integer min, Integer mic, Integer nan, String qual) {
        if (maj == null) {
            major = 0;
            minor = null;
            micro = null;
            nano = null;
            qualifier = null;
        } else {
            major = maj;
            minor = min;
            if (min == null) {
                micro = null;
                nano = null;
            } else if (mic == null) {
                micro = mic;
                nano = null;
            } else {
                micro = mic;
                nano = nan;
            }
            if (qual != null && qual.trim().length() > 0) {
                qualifier = qual.trim();
            }
        }
    }
}
