/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.repository.util;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

public final class ZipUtil {
    private ZipUtil() {
    }

    public static void unpackZipFile(File zipFile, File targetDir) throws IOException {
        unpackZipURL(zipFile.toURL(), targetDir);
    }
    
    public static void unpackZipResource(String resourceName, File targetDir) throws IOException {
        URL url;
        if (resourceName.startsWith("/")) {
            url = ZipUtil.class.getResource(resourceName);
        } else {
            url = ZipUtil.class.getResource("/" + resourceName);
        }
        if (url == null) {
            throw new IOException("Unable to access resource " + resourceName);
        }
        unpackZipURL(url, targetDir);
    }
    
    public static void unpackZipURL(URL url, File targetDir) throws IOException {
        InputStream is = url.openStream();
        if (is == null) {
            throw new IOException("Unable to access URL " + url);
        }
        unpackZipStream(is, targetDir);
    }
    
    public static void unpackZipStream(InputStream is, File targetDir) throws IOException {
        ZipInputStream zis = new ZipInputStream(is);
        try {
            unpackZipStream(zis, targetDir);
        } finally {
            InputStreamHelper.close(zis);
        }
    }
    
    private static void unpackZipStream(ZipInputStream zis, File targetDir) throws IOException {
        ZipEntry entry;
        while ((entry = zis.getNextEntry()) != null) {
            if (entry.isDirectory()) {
                new File(targetDir, entry.getName()).mkdirs();
            } else {
                File outFile = new File(targetDir, entry.getName());
                outFile.getParentFile().mkdirs();
                OutputStream os = new BufferedOutputStream(new FileOutputStream(outFile));
                try {
                    InputStreamHelper.drain(zis, os);
                } finally {
                    InputStreamHelper.close(os);
                }
            }
        }
    }
    
    public static void createZipFile(File source, File outputFile) throws IOException {
        ZipOutputStream zos = new ZipOutputStream(new BufferedOutputStream(new FileOutputStream(outputFile)));
        try {
            addFilesToZip(source, null, zos);
        } finally {
            InputStreamHelper.close(zos);
        }
        
    }

    private static void addFilesToZip(File file, String entryName, ZipOutputStream zos) throws IOException {
        if (file.isDirectory()) {
            for (File child : file.listFiles()) {
                String name = entryName == null ? child.getName() : entryName + "/" + child.getName();
                addFilesToZip(child, name, zos);
            }
        } else {
            String name = entryName == null ? file.getName() : entryName;
            ZipEntry entry = new ZipEntry(name);
            zos.putNextEntry(entry);
            InputStream is = new BufferedInputStream(new FileInputStream(file));
            try {
                InputStreamHelper.drain(is, zos);
            } finally {
                InputStreamHelper.close(is);
                zos.closeEntry();
            }
        }
    }
}
