/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.web.repository.base.client.atom;

import java.util.ArrayList;
import java.util.List;

import com.allen_sauer.gwt.log.client.Log;
import com.google.gwt.http.client.Request;
import com.google.gwt.http.client.RequestBuilder;
import com.google.gwt.http.client.RequestBuilder.Method;
import com.google.gwt.http.client.RequestCallback;
import com.google.gwt.http.client.RequestException;
import com.google.gwt.http.client.Response;
import com.google.gwt.http.client.URL;
import com.iona.soa.web.repository.base.client.model.ReposObj;
import com.iona.soa.web.repository.base.client.model.ReposObjAttribute;
import com.iona.soa.web.repository.base.client.model.ReposObjChangeNotifier;
import com.iona.soa.web.repository.base.client.model.ReposObjCollection;
import com.iona.soa.web.repository.base.client.remote.ICollectionCallBack;
import com.iona.soa.web.repository.base.client.remote.IDeletionCallBack;
import com.iona.soa.web.repository.base.client.remote.IObjCallBack;
import com.iona.soa.web.repository.base.client.remote.IRemoteErrorCallBack;
import com.iona.soa.web.repository.base.client.remote.PagingInfo;
import com.iona.soa.web.repository.base.client.remote.RemoteClient;
import com.iona.soa.web.repository.base.client.remote.RemoteError;

public class AtomClient implements RemoteClient {
    
    public static final int STATUS_CODE_OK = 200;
    public static final int STATUS_CODE_CREATED_OK = 201;

    private static final int TIME_OUT_MILLIS = 600000;

    private static final String QUERY_COMPACT = "_min";
    private static final String QUERY = "?";
    private static final String CONCAT = "&";
    private static final String SUBCOLLECTION = ";";
    //private static final String EMFQL = "emfql";
    private static final String SEARCH = "search";
    
    public AtomClient() {
        
    }
    
    public void searchReposObjList(String baseUrl, 
                                   String queryFilter,
                                   PagingInfo pagingInfo,
                                   final ICollectionCallBack aCallBack) {
        
        
        String searchUrl = baseUrl.endsWith("/") ? baseUrl + SEARCH : baseUrl + "/" + SEARCH;
        getReposObjList(searchUrl, queryFilter, null, aCallBack);
    }

    private String appendFilterQueryToUrl(String baseUrl, String queryFilter) {
        if (queryFilter == null || queryFilter.trim().length() == 0) {
            return baseUrl;
        }

        queryFilter = URL.encode(queryFilter);
        return appendQueryString(baseUrl, SEARCH, queryFilter);
    }

    private String appendPagingQueriesToUrl(String baseUrl, PagingInfo pagingInfo) {
        if (pagingInfo == null) {
            return baseUrl;
        }

        if (pagingInfo.getPageSize() != -1 && pagingInfo.getPageNo() != -1) {
            baseUrl = appendQueryString(baseUrl, "_count", "" + pagingInfo.getPageSize());
            baseUrl = appendQueryString(baseUrl, "_page", "" + pagingInfo.getPageNo());
        }
        
        if (pagingInfo.isAscending()) {
            baseUrl = appendQueryString(baseUrl, "_ascending", null);
        }
        
        if (PagingInfo.SORT_BY_TITLE.equals(pagingInfo.getOrdering())) {
            baseUrl = appendQueryString(baseUrl, "_orderBy", "name");
        } else if (pagingInfo.getOrdering() != null) {
            baseUrl = appendQueryString(baseUrl, "_orderBy", pagingInfo.getOrdering());
        }
        
        
        return baseUrl;
    }
    
    private String prepareUrl(String objUrl) {
        if (objUrl.indexOf(SUBCOLLECTION) > 0) {
            return objUrl;
        } 
        return appendQueryString(objUrl, QUERY_COMPACT, null);
    }

    private String appendQueryString(String baseUrl, String queryStringName, String queryStringValue) {
        return baseUrl
                + (baseUrl.indexOf(QUERY) > 0 ? CONCAT : QUERY)
                + queryStringName 
                + (queryStringValue == null ? "" : ("=" + queryStringValue));
    }

    public void getReposObjList(String baseUrl,
                                String collectionName,
                                String filterQuery,
                                PagingInfo pagingInfo,
                                final ICollectionCallBack aCallBack) {
        String feedUrl = baseUrl.endsWith("/") ? baseUrl + collectionName : baseUrl + "/" + collectionName;
        getReposObjList(feedUrl, filterQuery, pagingInfo, aCallBack);
    }
    
    public void getReposObjList(String fullUrl,
                                String filterQuery,
                                PagingInfo pagingInfo,
                                final ICollectionCallBack aCallBack) {
        try {
            String preparedUrl =
                prepareUrl(appendPagingQueriesToUrl(appendFilterQueryToUrl(fullUrl, filterQuery),
                                                    pagingInfo));
            Log.debug("\n\nGET (List URL)      ==> " + preparedUrl);
            RequestBuilder rb = createRequestBuilder(RequestBuilder.GET, preparedUrl);
            
            rb.sendRequest(null, new RequestCallback() {

                public void onError(Request request, Throwable exception) {
                    Log.debug("an error occured:", exception);
                    aCallBack.processConnectionError(exception);
                }

                public void onResponseReceived(Request request, Response response) {
                    Log.debug("GET (STATUS CODE) ==>" + response.getStatusCode());
                    Log.debug("GET (STATUS TEXT) ==>" + response.getStatusText());
                    Log.debug("GET (RESPONSE) ==>" + response.getText());
                    
                    AtomParser parser = new AtomParser(response.getText());

                    if (parser.isFeed()) {
                        aCallBack.processReposObjCollection(parser.parseFeed());
                    } else if (parser.isError()) {
                        RemoteError err = parser.parseError();
                        aCallBack.processResponseError(err.getErrorCode(), err.getErrorMsg());
                    } else {
                        aCallBack.processResponseError(Integer.toString(response.getStatusCode()), 
                                                       response.getText());
                    }
                }
            });
        } catch (RequestException e) {
            Log.debug("GET (ERROR) ==> " + e.getMessage(), e);
            aCallBack.processConnectionError(e);
        }
    }


    public void getReposObj(String objUrl, final IObjCallBack aCallBack) {
        
        Log.debug("\n\nGET (Object URL)      ==> " + objUrl);
        
        try {
            RequestBuilder rb = createRequestBuilder(RequestBuilder.GET, prepareUrl(objUrl));

            rb.sendRequest(null, new RequestCallback() {

                public void onError(Request request, Throwable exception) {
                    Log.debug("an error occured:");
                    aCallBack.processConnectionError(exception);
                }

                public void onResponseReceived(Request request, Response response) {
                    Log.debug("GET (STATUS CODE) ==>" + response.getStatusCode());
                    Log.debug("GET (STATUS TEXT) ==>" + response.getStatusText());
                    Log.debug("GET (RESPONSE) ==>" + response.getText());
                    
                    AtomParser parser = new AtomParser(response.getText());
                    
                    if (parser.isEntry()) {
                        aCallBack.processReposObj(parser.parseEntry());
                    } else if (parser.isError()) {
                        RemoteError err = parser.parseError();
                        aCallBack.processResponseError(err.getErrorCode(), err.getErrorMsg());
                    } else {
                        aCallBack.processResponseError(Integer.toString(response.getStatusCode()), 
                                                       response.getText());
                    }
                }
            });
        } catch (RequestException e) {
            Log.debug("GET (ERROR) ==> " + e.getMessage());
            aCallBack.processConnectionError(e);
        }
    }
    

    public void getReposObjs(String[] fullUrls, ICollectionCallBack callBack) {
        List<ReposObj> objects = new ArrayList<ReposObj>(fullUrls.length);
        getReposObjsInternal(fullUrls, objects, 0, callBack);
    }
    
    private void getReposObjsInternal(final String[] urls, 
                                      final List<ReposObj> objects, 
                                      final int index, 
                                      final ICollectionCallBack callBack) {
        if (index == urls.length) {
            ReposObjCollection collection = new ReposObjCollection(null);
            collection.setTheReposObjs(objects);
            callBack.processReposObjCollection(collection);
        } else {
            getReposObj(urls[index],
                        new IObjCallBack() {
                    public void processReposObj(ReposObj queriedObj) {
                        objects.add(index, queriedObj);
                        getReposObjsInternal(urls, objects, index + 1, callBack);
                    }
                    public void processConnectionError(Throwable e) {
                        callBack.processConnectionError(e);
                    }
                    public void processResponseError(String anErrorCode, String anErrorMsg) {
                        callBack.processResponseError(anErrorCode, anErrorMsg);
                    }
                });
        }
    }

    
    public void updateReposObj(final ReposObj updatedObj, final IObjCallBack aCallBack) {
        if (updatedObj.getEditLink() == null) {
            aCallBack.processResponseError(Integer.toString(IRemoteErrorCallBack.ERR_CODE_UNAUTHORIZED), 
                                           "The user is not authorized to update this object");
            return;
        }

        Log.debug("\n\nPUT (URL)      ==> " + updatedObj.getEditLink());
        
        String serializedEntry = (new AtomSerializer()).serializeEntry(updatedObj);
        Log.debug("PUT (DATA)     ==> " + serializedEntry);

        try {
            RequestBuilder rb = createRequestBuilder("PUT", updatedObj.getEditLink());
            rb.setHeader("Content-Type", "application/atom+xml; type=entry");
            
            rb.sendRequest(serializedEntry, new RequestCallback() {

                public void onError(Request request, Throwable exception) {
                    Log.debug("an error occured:");
                    aCallBack.processConnectionError(exception);
                }

                public void onResponseReceived(Request request, Response response) {
                    Log.debug("PUT (STATUS CODE) ==>" + response.getStatusCode());
                    Log.debug("PUT (STATUS TEXT) ==>" + response.getStatusText());
                    Log.debug("PUT (RESPONSE) ==> " + response.getText());
                    
                    AtomParser parser = new AtomParser(response.getText());
                    
                    if (parser.isEntry()) {
                        ReposObj updatedObj = parser.parseEntry();
                        aCallBack.processReposObj(updatedObj);
                        ReposObjChangeNotifier.INSTANCE.reposObjUpdated(updatedObj);
                    } else if (parser.isError()) {
                        RemoteError err = parser.parseError();
                        aCallBack.processResponseError(err.getErrorCode(), err.getErrorMsg());
                    } else {
                        aCallBack.processResponseError(Integer.toString(response.getStatusCode()), 
                                                       response.getText());
                    }
                }
            });
        } catch (RequestException e) {
            Log.debug("PUT (ERROR) ==> " + e.getMessage());
            aCallBack.processConnectionError(e);
        }
    }
    
    public void updateReposObjs(List<ReposObj> updatedObjs, final ICollectionCallBack aCallBack) {
        updateReposObjsInternal(updatedObjs, 0, aCallBack);
    }
    
    private void updateReposObjsInternal(final List<ReposObj> updatedObjs, 
                                         final int index, 
                                         final ICollectionCallBack callBack) {
        if (index == updatedObjs.size()) {
            ReposObjCollection callBackCollection = new ReposObjCollection(null);
            callBackCollection.setTheReposObjs(updatedObjs);
            callBack.processReposObjCollection(callBackCollection);
        } else {
            updateReposObj(updatedObjs.get(index),
                           new IObjCallBack() {
                    public void processReposObj(ReposObj anObj) {
                        updatedObjs.add(index, anObj);
                        updateReposObjsInternal(updatedObjs, index + 1, callBack);
                    }
                    public void processConnectionError(Throwable e) {
                        callBack.processConnectionError(e);
                    }
                    public void processResponseError(String anErrorCode, String anErrorMsg) {
                        callBack.processResponseError(anErrorCode, anErrorMsg);
                    }
                });
        }
    }

    public void createReposObj(String baseUrl,
                               String collectionName,
                               ReposObj defaultValue,
                               IObjCallBack aCallBack) {
        
        String feedUrl = baseUrl.endsWith("/") ? baseUrl + collectionName : baseUrl + "/" + collectionName;
        createReposObj(feedUrl, defaultValue, aCallBack);
    }

    public void createReposObj(final String fullUrl,
                               ReposObj defaultValue,
                               final IObjCallBack aCallBack) {
        
        
        try {
            String preparedUrl = prepareUrl(fullUrl);
            Log.debug("\n\nCREATE (URL)      ==> " + fullUrl);
            RequestBuilder rb = createRequestBuilder(RequestBuilder.POST, preparedUrl);

            String requestData = null;
            if (defaultValue != null) {
                AtomSerializer serializer = new AtomSerializer();
                requestData = serializer.serializeEntry(defaultValue);
                rb.setHeader("Content-Type", "application/atom+xml; type=entry");
                Log.debug("\n\nPOST DATA (URL)      ==> " + requestData);
            }
            
            rb.sendRequest(requestData, new RequestCallback() {

                public void onError(Request request, Throwable exception) {
                    Log.debug("an error occured:", exception);
                    aCallBack.processConnectionError(exception);
                }

                public void onResponseReceived(Request request, Response response) {
                    Log.debug("nCREATE (STATUS CODE) ==>" + response.getStatusCode());
                    Log.debug("nCREATE (STATUS TEXT) ==>" + response.getStatusText());
                    Log.debug("nCREATE (RESPONSE) ==>" + response.getText());
                    
                    AtomParser parser = new AtomParser(response.getText());

                    if (STATUS_CODE_CREATED_OK == response.getStatusCode()) {
                        ReposObj createdObj = parser.parseEntry();
                        aCallBack.processReposObj(createdObj);
                        ReposObjChangeNotifier.INSTANCE.reposObjCreated(createdObj, fullUrl, -1);

                    } else if (parser.isError()) {
                        RemoteError err = parser.parseError();
                        aCallBack.processResponseError(err.getErrorCode(), err.getErrorMsg());
                    } else {
                        aCallBack.processResponseError(Integer.toString(response.getStatusCode()), 
                                                       response.getText());
                    }
                }
            });
        } catch (RequestException e) {
            Log.debug("GET (ERROR) ==> " + e.getMessage(), e);
            aCallBack.processConnectionError(e);
        }
    }

    public void createReposObjs(String baseUrl,
                               String collectionName,
                               List<ReposObj> theObjs,
                               final ICollectionCallBack aCallBack) {
        
        String feedUrl = baseUrl.endsWith("/") ? baseUrl + collectionName : baseUrl + "/" + collectionName;
        createReposObjs(feedUrl, theObjs, aCallBack);
    }

    public void createReposObjs(final String fullUrl,
                                List<ReposObj> theObjs,
                                final ICollectionCallBack aCallBack) {
        
        
        try {
            String preparedUrl = prepareUrl(fullUrl);
            Log.debug("\n\nCREATE (URL)      ==> " + fullUrl);
            RequestBuilder rb = createRequestBuilder(RequestBuilder.POST, preparedUrl);

            if (theObjs == null || theObjs.size() == 0) {
                aCallBack.processResponseError("NO_CONTENT", "at least one repos obj need to be specified");
                return;
            }
            
            ReposObjCollection coll = new ReposObjCollection(null);
            for (ReposObj obj : theObjs) {
                coll.getTheReposObjs().add(obj);
            }
            AtomSerializer serializer = new AtomSerializer();
            String requestData = serializer.serializeFeed(coll);
            rb.setHeader("Content-Type", "application/atom+xml; type=entry");
            Log.debug("\n\nPOST DATA (URL)      ==> " + requestData);
            
            
            rb.sendRequest(requestData, new RequestCallback() {

                public void onError(Request request, Throwable exception) {
                    Log.debug("an error occured:", exception);
                    aCallBack.processConnectionError(exception);
                }

                public void onResponseReceived(Request request, Response response) {
                    Log.debug("nCREATE (STATUS CODE) ==>" + response.getStatusCode());
                    Log.debug("nCREATE (STATUS TEXT) ==>" + response.getStatusText());
                    Log.debug("nCREATE (RESPONSE) ==>" + response.getText());
                    
                    AtomParser parser = new AtomParser(response.getText());

                    if (STATUS_CODE_CREATED_OK == response.getStatusCode()) {
                        ReposObjCollection createdObjs = parser.parseFeed();
                        aCallBack.processReposObjCollection(createdObjs);
                        
                        for (ReposObj createdObj : createdObjs.getTheReposObjs()) {
                            ReposObjChangeNotifier.INSTANCE.reposObjCreated(createdObj, fullUrl, -1);
                        }

                    } else if (parser.isError()) {
                        RemoteError err = parser.parseError();
                        aCallBack.processResponseError(err.getErrorCode(), err.getErrorMsg());
                    } else {
                        aCallBack.processResponseError(Integer.toString(response.getStatusCode()), 
                                                       response.getText());
                    }
                }
            });
        } catch (RequestException e) {
            Log.debug("GET (ERROR) ==> " + e.getMessage(), e);
            aCallBack.processConnectionError(e);
        }
    }

    public void deleteReposObj(final ReposObj deletedObj, final IDeletionCallBack aCallBack) {
        if (deletedObj.getDeleteLink() == null) {
            aCallBack.processResponseError(Integer.toString(IRemoteErrorCallBack.ERR_CODE_UNAUTHORIZED), 
                                           "The user is not authorized to delete this object");
            return;
        }

        Log.debug("\n\nDELETE (URL)      ==> " + deletedObj.getDeleteLink());
        
        try {
            RequestBuilder rb = createRequestBuilder("DELETE", deletedObj.getDeleteLink());
            rb.sendRequest(null, new RequestCallback() {

                public void onError(Request request, Throwable exception) {
                    Log.debug("an error occured:");
                    aCallBack.processConnectionError(exception);
                }

                public void onResponseReceived(Request request, Response response) {
                    Log.debug("DELETE (STATUS CODE) ==>" + response.getStatusCode());
                    Log.debug("DELETE (STATUS TEXT) ==>" + response.getStatusText());
                    Log.debug("DELETE (RESPONSE) ==> " + response.getText());
                    
                    AtomParser parser = new AtomParser(response.getText());
                    
                    if ((response.getStatusCode() >= 200 && response.getStatusCode() < 300)
                            || response.getStatusCode() == 1223) { //because IE replaces 204 with 1223 :'( ...
                        ReposObjChangeNotifier.INSTANCE.reposObjDeleted(deletedObj);
                        aCallBack.processSuccess();
                        
                    } else if (parser.isError()) {
                        RemoteError err = parser.parseError();
                        aCallBack.processResponseError(err.getErrorCode(), err.getErrorMsg());
                    } else {
                        aCallBack.processResponseError(Integer.toString(response.getStatusCode()), 
                                                       response.getText());
                    }
                }
            });
        } catch (RequestException e) {
            Log.debug("PUT (ERROR) ==> " + e.getMessage());
            aCallBack.processConnectionError(e);
        }
    }

    public void deleteReposObj(String objUrl, final IDeletionCallBack callBack) {
        getReposObj(objUrl, new IObjCallBack() {
            public void processReposObj(ReposObj anObj) {
                deleteReposObj(anObj, callBack);
            }
            public void processConnectionError(Throwable e) {
                callBack.processConnectionError(e);
            }
            public void processResponseError(String anErrorCode,
                    String anErrorMsg) {
                callBack.processResponseError(anErrorCode, anErrorMsg);
            }
        });
    }
    
    public void deleteReposObjs(List<ReposObj> deletedObjs, IDeletionCallBack callBack) {
        deleteReposObjsInternal(deletedObjs, 0, callBack);
    }
    
    private void deleteReposObjsInternal(final List<ReposObj> deletedObjs, 
                                         final int index, 
                                         final IDeletionCallBack callBack) {
        if (index == deletedObjs.size()) {
            callBack.processSuccess();
        } else {
            deleteReposObj(deletedObjs.get(index),
                           new IDeletionCallBack() {
                    
                    public void processSuccess() {
                        deleteReposObjsInternal(deletedObjs, index + 1, callBack);
                    }
                    public void processConnectionError(Throwable e) {
                        callBack.processConnectionError(e);
                    }
                    public void processResponseError(String anErrorCode, String anErrorMsg) {
                        callBack.processResponseError(anErrorCode, anErrorMsg);
                    }
                });
        }
    }

    public void deleteReposObjs(String[] objUrls, final IDeletionCallBack callBack) {
        if (objUrls.length == 0) {
            callBack.processSuccess();
            return;
        }
        getReposObjs(objUrls, new ICollectionCallBack() {
            public void processReposObjCollection(ReposObjCollection list) {
                List<ReposObj> objs = list.getTheReposObjs();
                deleteReposObjs(objs, callBack);
            }
            public void processConnectionError(Throwable e) {
                callBack.processConnectionError(e);
            }
            public void processResponseError(String anErrorCode,
                    String anErrorMsg) {
                callBack.processResponseError(anErrorCode, anErrorMsg);
            }
        });
    }
    
    public Object parseServerMessage(String aMessage) throws Exception {
        if (aMessage == null || aMessage.trim().length() == 0) {
            throw new Exception("the message was empty");
        }
        
        AtomParser parser = new AtomParser(aMessage);

        if (parser.isFeed()) {
            return parser.parseFeed();

        } else if (parser.isEntry()) {
            return parser.parseEntry();
        
        } else if (parser.isError()) {
            return parser.parseError();
        }

        throw new Exception("the message was unrecognized: " + aMessage);
    }

    private RequestBuilder createRequestBuilder(Method m, String url) {
        RequestBuilder rb = new RequestBuilder(m, url);
        rb.setHeader("Cache-Control", "max-age=0, no-store");
        rb.setTimeoutMillis(TIME_OUT_MILLIS);
        return rb;
    }

    private RequestBuilder createRequestBuilder(String method, String url) {
        RequestBuilder rb = createRequestBuilder(RequestBuilder.POST, url);
        rb.setHeader("X-HTTP-Method-Override", method);
        return rb;
    }

    public void resolveMavenDependencies(String reposBaseUrl,
                                         ReposObj artifact,
                                         ICollectionCallBack callBack) {
        StringBuffer fullUrl = new StringBuffer(reposBaseUrl);
        if (!reposBaseUrl.endsWith("/")) {
            fullUrl.append("/");
        }
        
        String artifactId = artifact.getAttributeValue("name");
        String groupId = artifact.getAttributeValue("namespace");
        String version = artifact.getAttributeValue("version");
        
        fullUrl.append("artifactResolve");
        char sepChar = '?';
        if (artifactId != null) {
            fullUrl.append(sepChar);
            sepChar = '&';
            fullUrl.append("artifactId=");
            fullUrl.append(artifactId);
        }
        if (groupId != null) {
            fullUrl.append(sepChar);
            sepChar = '&';
            fullUrl.append("groupId=");
            fullUrl.append(groupId);
        }
        if (version != null) {
            fullUrl.append(sepChar);
            sepChar = '&';
            fullUrl.append("version=");
            fullUrl.append(version);
        }
        
        ReposObjAttribute foundIn = artifact.getAttribute("foundIn"); 
        if (foundIn != null && foundIn.getValue() != null) {
            fullUrl.append(sepChar);
            fullUrl.append("mavenRepoGuid=");
            
            String href = foundIn.getValue();
            String guid = href.replaceFirst("mavenrepositories/", "MavenRepository-");
            fullUrl.append(guid);
        }
        
        getReposObjList(fullUrl.toString(), null, null, callBack);
    }

    public void createIndex(String reposBaseUrl, String mavenRepoGuid,
            ICollectionCallBack callBack) {
        
        String fullUrl = reposBaseUrl;
        if (!fullUrl.endsWith("/")) {
            fullUrl += "/";
        }
        fullUrl += "spiderAction?mavenRepoGuid="  + mavenRepoGuid;  
        Log.debug("\n\nIn create index      ==> " + fullUrl);
        getReposObjList(fullUrl, null, null, callBack);
        
    }

}
