/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.web.repository.base.client.atom;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.google.gwt.xml.client.Attr;
import com.google.gwt.xml.client.Document;
import com.google.gwt.xml.client.Element;
import com.google.gwt.xml.client.Node;
import com.google.gwt.xml.client.NodeList;
import com.google.gwt.xml.client.XMLParser;
import com.iona.soa.web.repository.base.client.model.ReposObj;
import com.iona.soa.web.repository.base.client.model.ReposObjAttribute;
import com.iona.soa.web.repository.base.client.model.ReposObjCollection;

public class AtomSerializer {
    
    private static final String RSC_SCHEMA_NS = "http://schemas.iona.com/repository/2008/02/resource";

    private static final String XML_ENTRY_SKELETON = 
        "<entry xmlns=\"http://www.w3.org/2005/Atom\" xmlns:app=\"http://www.w3.org/2007/app\">"
        + "    <link rel=\"alternate\" type=\"application/atom+xml; type=entry\">"
        + "        <resource xmlns=\"" + RSC_SCHEMA_NS + "\"/>"
        + "    </link>"
        + "</entry>";

    private static final String XML_FEED_SKELETON = 
        "<feed xmlns=\"http://www.w3.org/2005/Atom\" xml:base=\"http://localhost:9090/repository/atom/\">"
        + "    <link rel=\"alternate\" type=\"application/atom+xml; type=entry\">"
        + "        <resource xmlns=\"" + RSC_SCHEMA_NS + "\"/>"
        + "    </link>"
        + "</feed>";
    

    public String serializeFeed(ReposObjCollection theFeed) {
        
        if (theFeed == null || theFeed.getTheReposObjs().size() == 0) {
            return null;
        }
        
        if (theFeed.getRepresentingNode() == null) {
            addMockRepresentingNode(theFeed);            
        }
        
        
        Document repNode = (Document)theFeed.getRepresentingNode();
        for (ReposObj anObj : theFeed.getTheReposObjs()) {
            synchRepNodeWithData(anObj);
            
            Node objNode = anObj.getRepresentingNode();
            if (objNode instanceof Document) {
                objNode = ((Document)objNode).getDocumentElement();
            }
            repNode.getDocumentElement().appendChild(objNode);
        }
        
        String serializedEntryElt = repNode.toString();
        
        return fixSerializedXml(serializedEntryElt);
    }


    public String serializeEntry(ReposObj theObj) {
        
        if (theObj == null) {
            return null;
        }

        synchRepNodeWithData(theObj);
        
        String serializedEntryElt = theObj.getRepresentingNode().toString();
        
        return fixSerializedXml(serializedEntryElt);
    }

    private void synchRepNodeWithData(ReposObj theObj) {
        Document xmlDoc;
        
        if (theObj.getRepresentingNode() == null) {
            xmlDoc = XMLParser.parse(XML_ENTRY_SKELETON);
            theObj.setRepresentingNode(xmlDoc);
        } else {
            
            // necessary if we try to reuse an entry that we got from a feed...
            Element representingElt = (Element)theObj.getRepresentingNode();
            
            String nsURI = representingElt.getNamespaceURI();
            if (nsURI == null || nsURI.length() == 0) {
                Attr parentXmlnsNode = 
                    (Attr)theObj.getRepresentingNode().getParentNode().getAttributes().getNamedItem("xmlns");
                representingElt.setAttribute("xmlns", parentXmlnsNode.getValue());
            }
            
            xmlDoc = representingElt.getOwnerDocument();
        }
        
        Node resourceNode = xmlDoc.getElementsByTagName("resource").item(0);
        for (ReposObjAttribute attr : theObj.getAttributes().values()) {
            
            String attrName = attr.getName();

            if (attr.isReadOnly()) {
                // do nothing
            } else {
                
                Element representingNode = (Element)attr.getRepresentingNode();

                if (representingNode == null) {
                    String nodeName = attr.isReference() ? "reference" : "attribute";
                    representingNode = xmlDoc.createElement(nodeName);
                    
                    representingNode.setAttribute("name", attr.getName());
                    //representingNode.setAttribute("xmlns", RSC_SCHEMA_NS);
                    
                    resourceNode.appendChild(representingNode);
                } 
                
                if (attr.isMulti()) {
                    // 1st remove all the nodes representing individual values for that attribute
                    NodeList allAttributes = resourceNode.getChildNodes();
                    List<Node> nodesMarkedForRemoval = new ArrayList<Node>();
                    for (int j = 0; j < allAttributes.getLength(); j++) {
                        
                        Node item = allAttributes.item(j);
                        if ((item instanceof Element)
                                && attrName.equals(((Element)item).getAttribute("name"))) {
                            nodesMarkedForRemoval.add(item);
                        }
                    }
                    for (Iterator<Node> rmvIter = nodesMarkedForRemoval.iterator(); rmvIter.hasNext();) {
                        resourceNode.removeChild(rmvIter.next());
                    }
                    

                    // 2nd re-add as many node as required to cover the items in the list
                    
                    // note : unlike for GETs, UPDATEs do not require the entry to contain any 
                    // "attribute/reference" node for empty/deleted attributes (no value)
                    
                    for (int j = 0; j < attr.getValues().length; j++) {
                        String collectionValue = attr.getValues()[j];
                        
                        Element listValueRepresentingNode = (Element)representingNode.cloneNode(true);
                        updateAttributeNode(xmlDoc,
                                            listValueRepresentingNode,
                                            attr.isReference()
                                                ? attr.getValueCaption(collectionValue)
                                                : collectionValue,
                                            attr.isReference() ? collectionValue : null);
                        resourceNode.appendChild(listValueRepresentingNode);
                    }
                    
                } else {
                    String caption = attr.isReference()
                                        ? attr.getValueCaption(attr.getValue())
                                        : attr.getValue();
                    String href = attr.isReference() ? attr.getValue() : null;
                    
                    updateAttributeNode(xmlDoc,
                                        representingNode,
                                        caption,
                                        href);
                }
                
            }
        }
    }

    
    /**
     * add a mock XML node representing the specified <code>ReposObjCollection</code>.
     * This happens when the ReposObjCollection was NOT obtain from parsing an XML feed/entry, for example 
     * when we want to provide initial data to a creation request. See the <code>RemoteClient</code> API for
     * details.
     * @param aCollection
     * @return 
     */
    private static Document addMockRepresentingNode(ReposObjCollection aCollection) {
        Document doc = XMLParser.parse(XML_FEED_SKELETON);
        aCollection.setRepresentingNode(doc);
        return doc;
    }

    private void updateAttributeNode(Document ownerDoc,
                                     Element representingNode,
                                     String value,
                                     String valueHref) {

        // clear any potential old data
        if (representingNode.getFirstChild() != null) {
            representingNode.removeChild(representingNode.getFirstChild());
        }
        representingNode.removeAttribute("href");
        

        if (value != null && value.trim().length() != 0) {
            representingNode.appendChild(ownerDoc.createTextNode(value));
        }

        if (valueHref != null) {
            representingNode.setAttribute("href", valueHref);
        }
    }
    
    private String fixSerializedXml(String serializedEntryElt) {
        // for some reason the GWT XML API escapes ";" so we need to un-escape them
        String fixEntryXml = serializedEntryElt.replaceAll("&semi;", ";");

        // for some reason the GWT XML API set a blank NS for created Node so we need to clear it 
        // in order to get the element inherits its parent NS.
        fixEntryXml = serializedEntryElt.replaceAll(" xmlns=\"\"", "");
        return fixEntryXml;
    }



}
