/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.web.repository.base.client.model;

import java.util.HashMap;
import java.util.Map;

import com.google.gwt.xml.client.Node;

/**
 * represents a generic ARR objects.
 * Note that the data object keeps a reference to the XML node that initially represented the ARR entity (just
 * in case...)
 */
public class ReposObj {

    private Node representingNode;
    private String serverBaseLink;
    private String selfLink;
    private String editLink;
    private String deleteLink;
    private String feedSubscriptionLink;
    
    private String id;
    private String caption;
    private String timeStamp;
    private String eClass;
    private Map<String, ReposObjAttribute> attributes;
    
    /**
     * Allow UI code to store temporary data here
     */
    private Map<String, Object> userValues;
    
    
    
    public ReposObj() {
        this(null);
    }
    
    public ReposObj(Node aNode) {
        representingNode = aNode;
        attributes = new HashMap<String, ReposObjAttribute>();
    }
    
    /**
     * retrieves the URL to use to re-fetch the object from the server
     * @return the URL
     */
    public String getSelfLink() {
        return selfLink;
    }
    /**
     * sets the URL to use to re-fetch the object from the server
     * @param selfLink the URL
     */
    public void setSelfLink(String aLink) {
        selfLink = aLink;
    }
    
    /**
     * retrieves the URL to use to commit changes on the object to the server
     * @return the URL
     */
    public String getEditLink() {
        return editLink;
    }

    /**
     * sets the URL to use to commit changes on the object to the server
     * @param aLink the URL
     */
    public void setEditLink(String aLink) {
        editLink = aLink;
    }
    
    /**
     * retrieves the URL to use to remove the object from the server
     * @return the URL
     */
    public String getDeleteLink() {
        return deleteLink;
    }

    /**
     * sets the URL to use to remove the object from the server
     * @param aLink the URL
     */
    public void setDeleteLink(String aLink) {
        deleteLink = aLink;
    }
    
    /**
     * retrieves the URL to register with RSS/ATOM readers to track changes to this repos object
     * @return the URL
     */
    public String getFeedSubscriptionLink() {
        return feedSubscriptionLink;
    }

    /**
     * sets the URL to register with RSS/ATOM readers to track changes to this repos object
     * @param aLink the URL
     */
    public void setFeedSubscriptionLink(String aLink) {
        feedSubscriptionLink = aLink;
    }
    
    public String getId() {
        return id;
    }
    public void setId(String anId) {
        id = anId;
    }

    public String getCaption() {
        return caption;
    }
    
    public void setCaption(String aCaption) {
        caption = aCaption;
    }
    
    public String getLastUpdatedTimeStamp() {
        return timeStamp;
    }
    public void setLastUpdatedTimeStamp(String aTimeStamp) {
        timeStamp = aTimeStamp;
    }
    
    /**
     * retrieves the attributes of the ARR object
     * @return Map<String, ReposObjAttribute> where the string is the name/label of the attribute and the 
     * <code>ReposObjAttribute</code> the data object representing it.
     */
    public Map<String, ReposObjAttribute> getAttributes() {
        return attributes;
    }
    
    /**
     * convenience method to retrieve a specific attribute
     */
    public ReposObjAttribute getAttribute(String attributeName) {
        return (ReposObjAttribute)attributes.get(attributeName);
    }
    
    /**
     * convenience method to retrieve a specific attribute
     */
    public String getAttributeValue(String attributeName) {
        ReposObjAttribute attr = (ReposObjAttribute)attributes.get(attributeName);
        if (attr != null) {
            return attr.getValue();
        } else {
            return null;
        }
    }

    /**
     * add an attribute to the current obj. Note that the <code>ReposObjAttribute</code> needs to have its
     *  name set, otherwise it will not be added to the current <code>ReposObj</code>.
     * Also Note that any other attribute with the same name previously set will be dropped. 
     * @param anAttr
     */
    public void addAttribute(ReposObjAttribute anAttr) {
        if (anAttr == null || anAttr.getName() == null) {
            return;
        }
        anAttr.setDefiningEClass(getEClass());
        attributes.put(anAttr.getName(), anAttr);
        anAttr.setParent(this);
    }

    public void removeAttribute(String attributeName) {
        attributes.remove(attributeName);

    }

    public Node getRepresentingNode() {
        return representingNode;
    }
    public void setRepresentingNode(Node aNode) {
        representingNode = aNode;
    }

    public String getEClass() {
        return eClass;
    }
    public void setEClass(String className) {
        eClass = className;
    }

    public String getServerBaseLink() {
        return serverBaseLink;
    }

    public void setServerBaseLink(String aLink) {
        serverBaseLink = aLink;
    }

    public ReposObj deepCopy() {
        ReposObj clone = new ReposObj();
        
        clone.setRepresentingNode(getRepresentingNode());
        clone.setServerBaseLink(getServerBaseLink());
        clone.setSelfLink(getSelfLink());
        clone.setEditLink(getEditLink());
        clone.setDeleteLink(getDeleteLink());
        clone.setFeedSubscriptionLink(getFeedSubscriptionLink());
        clone.setId(getId());
        clone.setCaption(getCaption());
        clone.setLastUpdatedTimeStamp(getLastUpdatedTimeStamp());
        clone.setEClass(getEClass());
        
        for (ReposObjAttribute attr : getAttributes().values()) {
            clone.addAttribute(attr.deepCopy());
        }

        return clone;
    }

    public boolean isEditAllowed() {
        return editLink != null;
    }

    public boolean isDeleteAllowed() {
        return deleteLink != null;
    }
    
    public final void setUserValue(String key, Object value) {
        if (userValues == null) {
            userValues = new HashMap<String, Object>();
        }
        userValues.put(key, value);
    }
    
    public final Object getUserValue(String key) {
        return userValues == null ? null : userValues.get(key);
    }
    
    public final void clearUserValues() {
        if (userValues != null) {
            userValues.clear();
        }
    }
    
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        } else if (!(obj instanceof ReposObj)) {
            return false;
        }
        ReposObj other = (ReposObj)obj;
        if (eq(getRepresentingNode(), other.getRepresentingNode())
            && eq(getServerBaseLink(), other.getServerBaseLink())
            && eq(getSelfLink(), other.getSelfLink())
            && eq(getEditLink(), other.getEditLink())
            && eq(getDeleteLink(), other.getDeleteLink())
            && eq(getFeedSubscriptionLink(), other.getFeedSubscriptionLink())
            && eq(getId(), other.getId())
            && eq(getCaption(), other.getCaption())
            && eq(getLastUpdatedTimeStamp(), other.getLastUpdatedTimeStamp())
            && eq(getEClass(), other.getEClass())
            && eq(getAttributes(), other.getAttributes())) {
            return true;
        }
        return false;
    }
    
    private static boolean eq(Object o1, Object o2) {
        return o1 == o2 || (o1 != null && o2 != null && o1.equals(o2));
    }
}
