/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.web.repository.base.client.model;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;

import com.google.gwt.xml.client.Node;

/**
 * represents a generic ARR object attribute.
 * Note that the data object keeps a reference to the XML node that initially represented the ARR entity (just
 * in case...)
 * It also contains some level of self describing meta-data to help the client deal with the information
 * carried
 */
public class ReposObjAttribute {

    /**
     * a string expected to get quite long (e.g. comments, descriptions...)
     */
    public static final String SIMPLE_TYPE_LONG_STRING = "xsd:string";
    
    /**
     * A short string (likely <254 chars) that does not contain line feeds, carriage returns, or tabs
     */
    public static final String SIMPLE_TYPE_SHORT_STRING = "xsd:normalizedString";

    /**
     * A very short string that does not contain line feeds, carriage returns, tabs, leading or trailing 
     * spaces, or multiple spaces... so basically a word...
     */
    public static final String SIMPLE_TYPE_TOKEN_STRING = "xsd:token";
    
    /**
     * for URLs and the like...
     */
    public static final String SIMPLE_TYPE_URI = "xsd:anyURI";

    public static final String SIMPLE_TYPE_INTEGER = "xsd:integer";
    public static final String SIMPLE_TYPE_BOOLEAN = "xsd:boolean";
    //public static final String SIMPLE_TYPE_DATE_TIME = "xsd:dateTime";
    
    private Node representingNode;
    
    private String name;
    
    private boolean isReadOnly;
    private boolean isOptional;
    private boolean isHidden;
    
    private String refEClass;
    private String primitiveType;
    
    private String defEClass;
    
    private boolean isMany;
    private boolean isReference;
    private boolean isCollection;
    private int collectionSize;
    
    private String collectionHref;
    
    private ReposObj parent;
        
    /**
     * stores the values set for the attribute by mapping the actual literal value/href to its displayable
     * caption. the caption part is only meaningful for referenced values as their href isn't very human 
     * friendly. literals can use their value itself as a caption
     */
    private Map<String, String> values;

    public ReposObjAttribute() {
        this(null);
    }
    
    public ReposObjAttribute(Node representingNode) {
        super();
        this.representingNode = representingNode;
        clearValues();
    }

    public Node getRepresentingNode() {
        return representingNode;
    }

    public ReposObjAttribute setRepresentingNode(Node aNode) {
        representingNode = aNode;
        return this;
    }

    public String getName() {
        return name;
    }

    public ReposObjAttribute setName(String aName) {
        name = aName;
        return this;
    }

    public String getPrimitiveType() {
        return primitiveType;
    }

    public ReposObjAttribute setPrimitiveType(String aType) {
        primitiveType = aType;
        return this;
    }

    public String getRefEClass() {
        return refEClass;
    }

    public ReposObjAttribute setRefEClass(String eClass) {
        refEClass = eClass;
        return this;
    }

    public String getDefiningEClass() {
        return defEClass;
    }

    public ReposObjAttribute setDefiningEClass(String eClass) {
        defEClass = eClass;
        return this;
    }

    public boolean isMulti() {
        return isMany;
    }

    public ReposObjAttribute setMulti(boolean isIt) {
        isMany = isIt;
        return this;
    }

    public boolean isReadOnly() {
        return isReadOnly;
    }

    public ReposObjAttribute setReadOnly(boolean isIt) {
        isReadOnly = isIt;
        return this;
    }

    public boolean isHidden() {
        return isHidden;
    }

    public ReposObjAttribute setHidden(boolean isIt) {
        isHidden = isIt;
        return this;
    }
    
    /**
     * retrieves all the values (or the HREFs representing those values in case of referenced objects)
     * set for that attribute. 
     * Note: for "non-multi" attributes, the user might as well use <code>getValue()</code> 
     * @return the array of values/HREFs
     */
    public String[] getValues() {
        return (String[])values.keySet().toArray(new String[values.size()]);
    }

    /**
     * convenience method to retrieve the single allowed value for "non-multi" attributes. 
     * Effectively retrieves the 1st value (or the HREF representing that value in case of a referenced
     *  object)
     * set for that attribute.
     * @return the 1st value/HREF set for that attribute
     */
    public String getValue() {
        if (values == null || values.size() == 0) {
            return null;
        }
        return getValues()[0];
    }

    /**
     * adds a referenced value to the list already set values.
     * Note: for "non-multi" attributes, the user might as well use <code>setValue()</code> 
     * @param aRefObj the object to which we want to add a reference to. Note that the object must already be
     * stored by the server (i.e you'll need to obtain it from a query to the server)
     */
    public ReposObjAttribute addRefValue(ReposObj aRefObj) throws InvalidOperationException {
        
        if (aRefObj.getSelfLink() == null
                || aRefObj.getServerBaseLink() == null) {
            throw new InvalidOperationException("a repos obj can only be added as a reference if it is"
                                                + " already stored by the server");
        }
        String href = aRefObj.getSelfLink().indexOf(aRefObj.getServerBaseLink()) > -1 
                        ? aRefObj.getSelfLink().substring(aRefObj.getServerBaseLink().length())
                        : aRefObj.getSelfLink();
        addRefValue(href, aRefObj.getCaption());
        
        return this;
    }
    
    /**
     * adds a referenced value to the list already set values.
     * Note: for "non-multi" attributes, the user might as well use <code>setValue()</code> 
     * @param href the HREF representing the referenced value/object (note that the href should be relative
     * to the server'd base URL)
     * @param valueCaption the caption for that value/object
     */
    public ReposObjAttribute addRefValue(String href, String valueCaption) throws InvalidOperationException {
        if (!isReference()) {
            throw new InvalidOperationException("referenced values can only be added to"
                                                + " a referenced Attribute");
        }
        values.put(href, valueCaption);
        return this;
    }

    /**
     * convenience method to set the single allowed referenced value for "non-multi" attributes. 
     * effectively clear whatever  values were already set and add a single value to the attribute
     * @param aRefObj the object to which we want to add a reference to. Note that the object must already be
     * stored by the server (i.e you'll need to obtain it from a query to the server)
     */
    public ReposObjAttribute setRefValue(ReposObj aRefObj) throws InvalidOperationException {
        
        if (aRefObj.getSelfLink() == null
                || aRefObj.getServerBaseLink() == null) {
            throw new InvalidOperationException("a repos obj can only be added as a reference if it is"
                                                + " already stored by the server");
        }
        String href = aRefObj.getSelfLink().indexOf(aRefObj.getServerBaseLink()) > -1 
                        ? aRefObj.getSelfLink().substring(aRefObj.getServerBaseLink().length())
                        : aRefObj.getSelfLink();
        setRefValue(href, aRefObj.getCaption());
        setRefEClass(aRefObj.getEClass());
        return this;
    }

    /**
     * convenience method to set the single allowed referenced value for "non-multi" attributes. 
     * effectively clear whatever  values were already set and add a single value to the attribute
     * @param href the HREF representing the referenced value/object (note that the href should be relative
     * to the server'd base URL)
     * @param valueCaption the caption for that value/object
     */
    public ReposObjAttribute setRefValue(String href, String valueCaption) throws InvalidOperationException {
        if (!isReference()) {
            throw new InvalidOperationException("referenced values can only be added to"
                                                + " a referenced Attribute");
        }
        clearValues();
        addRefValue(href, valueCaption);
        return this;
    }

    /**
     * adds a non-referenced value to the list already set values.
     * Note: for "non-multi" attributes, the user might as well use <code>setValue()</code> 
     * @param literalValue the literal value to set
     */
    public ReposObjAttribute addSimpleValue(String literalValue) throws InvalidOperationException {
        if (isReference()) {
            throw new InvalidOperationException("referenced values can only be added to"
                                                + " a referenced Attribute");
        }
        values.put(literalValue, literalValue);
        return this;
    }

    /**
     * remove a value to the list already set values.
     * Note: for "non-multi" attributes, the user might as well use <code>clearValues()</code> 
     * @param literalValue the literal value to set
     */
    public ReposObjAttribute removeValue(String literalValue) throws InvalidOperationException {
        values.remove(literalValue);
        return this;
    }

    public ReposObjAttribute removeRefValue(String literalValue) throws InvalidOperationException {
        if (!isReference()) {
            throw new InvalidOperationException("Only valid for references");
        }
        values.remove(literalValue);
        return this;
    }
    
    /**
     * convenience method to set the single allowed non-referenced value for "non-multi" attributes. 
     * effectively clear whatever  values were already set and add a single value to the attribute
     * @param literalValue the literal value to set
     */
    public ReposObjAttribute setSimpleValue(String literalValue) throws InvalidOperationException {
        if (isReference()) {
            throw new InvalidOperationException("referenced values can only be added to"
                                                + " a referenced Attribute");
        }
        clearValues();
        addSimpleValue(literalValue);
        return this;
    }

    public ReposObjAttribute clearValues() {
        values  = new LinkedHashMap<String, String>();
        return this;
    }

    /**
     * retrieves the caption for the given value/HREF (only meaningful for referenced object/values)
     * @param href the value (or the HREF representing that value in case of a referenced object)
     * @return the caption
     */
    public String getValueCaption(String href) {
        return isReference() ? values.get(href) : href;
    }


    public boolean isReference() {
        return isReference;
    }

    public ReposObjAttribute setReference(boolean isIt) {
        isReference = isIt;
        return this;
    }

    /**
     * indicates if this attribute point to an entry/object or to a feed/collection
     * Note that:
     *  - only reference attributes can also be collection attributes
     *  - for collection attributes, the relevant value (i.e. link to the collection) is accessed via 
     * <code>[get/set]CollectionLink()</code> 
     *  - for NON-collection attributes, the relevant value (i.e. link to the collection) is stored in 
     * <code>[get/set]RefValue()</code> and <code>[get/set]Values()</code>  
     * @return true if this attribute points to a feed/collection and false if it's an actual entry/object 
     */
    public boolean isCollection() {
        return isCollection;
    }

    /**
     * indicates if this attribute point to an entry/object or to a feed/collection.
     * Note that:
     *  - only reference attributes can also be collection attributes
     *  - for collection attributes, the relevant value (i.e. link to the collection) is accessed via 
     * <code>[get/set]CollectionLink()</code> 
     *  - for NON-collection attributes, the relevant value (i.e. link to the collection) is stored in 
     * <code>[get/set]RefValue()</code> and <code>[get/set]Values()</code>  
     * @param isIt true if this attribute points to a feed/collection and false if it's an actual entry/object
     */
    public ReposObjAttribute setCollection(boolean isIt) {
        isCollection = isIt;
        return this;
    }

    public ReposObjAttribute setCollectionSize(int size) {
        if (!isCollection) {
            throw new InvalidOperationException("Attribute is not a collection");
        }
        collectionSize = size;
        return this;
    }
    
    public int getCollectionSize() {
        if (!isCollection) {
            throw new InvalidOperationException("Attribute is not a collection");
        }
        return collectionSize;
    }
    
    public boolean isOptional() {
        return isOptional;
    }

    public ReposObjAttribute setOptional(boolean isIt) {
        isOptional = isIt;
        return this;
    }

    /**
     * has a different meaning depending on the value returned by <code>isCollection()</code>
     *  - for collection attributes: this is the relevant value of those attributes as it stores the link 
     *  to the actual collections represented by those links  
     *  - for NON-collection attributes: this is the URL to the collection/pool for possible values for that
     *  attribute. the actual values of the attribute is/are accessed via <code>[get/set]RefValue()</code> and
     *   <code>[get/set]Values()</code> 
     * @return
     */
    public String getCollectionLink() {
        return collectionHref;
    }

    /**
     * @see <code>getCollectionLink()</code>
     * @param aLink
     */
    public ReposObjAttribute setCollectionLink(String aLink) {
        collectionHref = aLink;
        return this;
    }

    public ReposObj getParent() {
        return parent;
    }

    protected ReposObjAttribute setParent(ReposObj theParent) {
        parent = theParent;
        return this;
    }
    
    
    public ReposObjAttribute deepCopy() {
        ReposObjAttribute clone = new ReposObjAttribute();
        
        clone.setRepresentingNode(getRepresentingNode());
        clone.setName(getName());
        clone.setReadOnly(isReadOnly());
        clone.setOptional(isOptional());
        clone.setHidden(isHidden());
        clone.setRefEClass(getRefEClass());
        clone.setPrimitiveType(getPrimitiveType());
        clone.setDefiningEClass(getDefiningEClass());
        clone.setMulti(isMulti());
        clone.setReference(isReference());
        clone.setCollection(isCollection());
        
        if (isCollection()) {
            clone.setCollectionSize(getCollectionSize());
        }
        
        clone.setCollectionLink(getCollectionLink());
        clone.setParent(getParent());

        if (isReference()) {
            for (String value : getValues()) {
                clone.addRefValue(value, getValueCaption(value));
            }
        } else {
            for (String value : getValues()) {
                clone.addSimpleValue(value);
            }

        }
        return clone;
    }
    
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        } else if (!(obj instanceof ReposObjAttribute)) {
            return false;
        }
        ReposObjAttribute other = (ReposObjAttribute)obj;
        if (eq(getRepresentingNode(), other.getRepresentingNode())
            && eq(getName(), other.getName())
            && isReadOnly() == other.isReadOnly()
            && isOptional() == other.isOptional()
            && isHidden() == other.isHidden()
            && eq(getRefEClass(), other.getRefEClass())
            && eq(getPrimitiveType(), other.getPrimitiveType())
            && eq(getDefiningEClass(), other.getDefiningEClass())
            && isMulti() == other.isMulti()
            && isReference() == other.isReference()
            && isCollection() == other.isCollection()
            && collectionSize == other.collectionSize
            && eq(getCollectionLink(), other.getCollectionLink())
            && (getParent() == null && other.getParent() == null || eq(getParent().getId(), other.getParent().getId()))
            && eq(values, other.values)) {
            return true;
        }
        return false;
    }
    
    private static boolean eq(Object o1, Object o2) {
        if (o1 == o2) {
            return true;
        } else if (o1 == null || o2 == null) {
            return false;
        } else if (o1 instanceof ReposObj && o2 instanceof ReposObj) {
            return eq(((ReposObj)o1).getId(), ((ReposObj)o2).getId());
        } else {
            return o1.equals(o2);
        }
    }
}
