/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.web.repository.base.client.remote;

import java.util.List;

import com.iona.soa.web.repository.base.client.model.ReposObj;

/**
 * The basic interface to communicate with an ARR remote server.
 * Note that all the operations in this interface are asynchronous.  
 * 
 * @see also com.iona.soa.web.repository.base.client.model.ReposObjChangeNotifier for
 * a way to listen on data changes (why it was implemented totally independently 
 * of this class I do not know, sigh!)
 */
public interface RemoteClient {
    
    /**
     * retrieves a full collection of ARR objects
     * @param baseUrl the base URL of the server 
     * @param collectionName the name of the collection we want to retrieve
     * @param filterQuery an optional query to filter the collection
     * @param pagingInfo optional
     * @param aCallBack the call back handler to process the response to the request
     */
    void getReposObjList(String baseUrl,
                         String collectionName,
                         String filterQuery,
                         PagingInfo pagingInfo,
                         final ICollectionCallBack aCallBack);

    /**
     * retrieves a full collection of ARR objects
     * @param fullUrl the full URL pointing to the collection content 
     * @param filterQuery an optional query to filter the collection
     * @param pagingInfo optional
     * @param aCallBack the call back handler to process the response to the request
     */
    void getReposObjList(String fullUrl,
                         String filterQuery,
                         PagingInfo pagingInfo,
                         final ICollectionCallBack aCallBack);

    /**
     * run a fuzzy search against the ARR objects
     * @param reposBaseUrl the base URL of the server
     * @param queryFilter the filter to apply to the collection
     * @param pagingInfo TODO
     * @param aCallBack the call back handler to process the response to the request
     */
    void searchReposObjList(String reposBaseUrl,
                            String queryFilter,
                            PagingInfo pagingInfo, final ICollectionCallBack aCallBack);
    
    void resolveMavenDependencies(String reposBaseUrl,
                                  ReposObj artifact,
                                  final ICollectionCallBack aCallBack);
    
    void createIndex(String reposBaseUrl,
                     String mavenRepoGuid,
                     final ICollectionCallBack aCallBack);
    
    /**
     * retrieves a specific ARR object
     * @param fullUrl the full URL pointing to the object targeted
     * @param aCallBack the call back handler to process the response to the request
     */
    void getReposObj(String fullUrl, final IObjCallBack aCallBack);

    /**
     * retrieves a number of ARR objects and returns them as a collection
     * @param fullUrls the URLs pointing to the objects targeted
     * @param aCallBack the call back handler to process the response to the request
     */
    void getReposObjs(String[] fullUrls, final ICollectionCallBack aCallBack);

    /**
     * commits the changes made to a given ARR object
     * @param updatedObj the updated object to commit to the server
     * @param aCallBack the call back handler to process the response to the request
     */
    void updateReposObj(ReposObj updatedObj, final IObjCallBack aCallBack);
    
    /**
     * commits the changes made to the given ARR objects
     * @param updatedObjs the updated objects to commit to the server
     * @param aCallBack the call back handler to process the response to the request
     */
    void updateReposObjs(List<ReposObj> updatedObjs, final ICollectionCallBack aCallBack);

    /**
     * creates a new ARR object of a specific type/collection
     * @param baseUrl the base URL of the server 
     * @param collectionName the name of the type/collection of the object we want to create
     * @param defaultValue (OPTIONAL) <code>ReposObj</code> specifying the defaults values to add to newly 
     * created object. it only needs to contain <code>reposObjAttribute</code> for the fields we want to
     * initialise, for any other field the server will come up with its own defaults values. Note that this
     * parameter should be optional and if left null, the server control the initialisation.
     * @param aCallBack the call back handler to process the response to the request
     */
    void createReposObj(String baseUrl,
                        String collectionName,
                        ReposObj defaultValue,
                        final IObjCallBack aCallBack);

    /**
     * creates a new ARR object of a specific type/collection
     * @param fullUrl the full URL of the type/collection of the object we want to create
     * @param defaultValue (OPTIONAL) <code>ReposObj</code> specifying the defaults values to add to newly 
     * created object. it only needs to contain <code>reposObjAttribute</code> for the fields we want to
     * initialise, for any other field the server will come up with its own defaults values. Note that this
     * parameter should be optional and if left null, the server control the initialisation.
     * @param aCallBack the call back handler to process the response to the request
     */
    void createReposObj(String fullUrl,
                        ReposObj defaultValue,
                        final IObjCallBack aCallBack);

    /**
     * creates several new ARR objects of a specific type/collection
     * @param baseUrl the base URL of the server 
     * @param collectionName the name of the type/collection of the object we want to create
     * @param theObjs <code>List<ReposObj></code> specifying the defaults values to add to newly 
     * created objects. it only needs to contain <code>reposObjAttribute</code> for the fields we want to
     * initialise for each object, for any other field the server will come up with its own defaults values.
     * Note that this parameter should not be optional and if left null, the request will be cancelled and an
     *  error thrown.
     * Also the "eclass" field set on individual reposObj is irrelevant as the eclass of the targeted 
     * collection will be used anyway.
     * @param aCallBack the call back handler to process the response to the request
     */
    void createReposObjs(String baseUrl,
                         String collectionName,
                         List<ReposObj> theObjs,
                         final ICollectionCallBack aCallBack);

    /**
     * creates several new ARR objects of a specific type/collection
     * @param fullUrl the full URL of the type/collection of the object we want to create
     * @param theObjs <code>List<ReposObj></code> specifying the defaults values to add to newly 
     * created objects. it only needs to contain <code>reposObjAttribute</code> for the fields we want to
     * initialise for each object, for any other field the server will come up with its own defaults values.
     * Note that this parameter should not be optional and if left null, the request will be cancelled and an
     *  error thrown.
     * Also the "eclass" field set on individual reposObj is irrelevant as the eclass of the targeted 
     * collection will be used anyway.
     * @param aCallBack the call back handler to process the response to the request
     */
    void createReposObjs(String fullUrl,
                         List<ReposObj> theObjs,
                         final ICollectionCallBack aCallBack);

    /**
     * deletes a given object from the ARR server
     * @param deletedObj the object to delete
     * @param aCallBack the call back handler to process the response to the request
     */
    void deleteReposObj(ReposObj deletedObj, IDeletionCallBack aCallBack);
    
    /**
     * deletes a given object from the ARR server
     * @param objUrl the url of the object to delete
     * @param aCallBack the call back handler to process the response to the request
     */
    void deleteReposObj(String objUrl, IDeletionCallBack aCallBack);
    
    /**
     * deletes the given objects from the ARR server
     * @param deletedObjs the objects to delete
     * @param aCallBack the call back handler to process the response to the request
     */
    void deleteReposObjs(List<ReposObj> deletedObjs, IDeletionCallBack aCallBack);
    
    /**
     * deletes the given objects from the ARR server
     * @param objUrls the urls of the objects to delete
     * @param aCallBack the call back handler to process the response to the request
     */
    void deleteReposObjs(String[] objUrls, IDeletionCallBack aCallBack);
    
    
    
    /**
     * attempts to parse a message from the server
     * @param aMessage
     * @return a <code>ReposObj</code>, a <code>ReposObjCollectoion</code> or a <code>RemoteError</code> 
     * depending of the nature of the message
     * @throws Exception if the message could not be parsed (malformed of otherwise...)
     */
    Object parseServerMessage(String aMessage) throws Exception;
}
