/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.web.repository.main.client.base.ui;

import com.extjs.gxt.ui.client.Events;
import com.extjs.gxt.ui.client.Style.HorizontalAlignment;
import com.extjs.gxt.ui.client.data.ModelData;
import com.extjs.gxt.ui.client.event.GridEvent;
import com.extjs.gxt.ui.client.event.Listener;
import com.extjs.gxt.ui.client.store.ListStore;
import com.extjs.gxt.ui.client.widget.grid.ColumnConfig;
import com.extjs.gxt.ui.client.widget.grid.ColumnData;
import com.extjs.gxt.ui.client.widget.grid.Grid;
import com.extjs.gxt.ui.client.widget.grid.GridCellRenderer;
import com.google.gwt.core.client.GWT;
import com.google.gwt.dom.client.Element;
import com.iona.soa.web.repository.main.client.HtmlSnippets;
import com.iona.soa.web.repository.main.client.base.ui.SimpleGridConfigurator.SimpleColumnConfig;

/**
 * Intended for use where the contents of the column will be a action link,
 * for example "Edit" or "Delete".  Note that the caller will have to implement 
 * a GridCellListener on the grid and provide an onCellClick implementation
 * that implements actions for any ActionColumns in the grid. 
 */
public class ActionColumnConfig<D extends ModelData> extends SimpleColumnConfig {

    protected final HtmlSnippets snippets = GWT.create(HtmlSnippets.class);

    private String actionLabel;
    
    private ActionColumnConfig.ActionColumnCallback<D> actionCallback;

    private final Grid<D> gridPanel;
    
    /**
     * Clients must call <code>setActionLabel</code> and <code>setActionCallback</code> before 
     * the grid is rendered.
     *  
     * @param grid
     * @param columnId
     */
    public ActionColumnConfig(Grid<D> grid, String columnId) {
        super(columnId, "");
        gridPanel = grid;
        setSortable(false);
        setAlignment(HorizontalAlignment.RIGHT);
        setRenderer(createRenderer());
        addCellClickListener(grid);
        
    }
    
    public final ActionColumnConfig<D> setActionLabel(String label) {
        actionLabel = label;
        setWidth(label.length() * 8);
        return this;
    }
    
    public final ActionColumnConfig<D> setActionCallback(ActionColumnCallback<D> callback) {
        actionCallback = callback;
        return this;
    }

    /**
     * Can be overridden to supply the label that appears in the hyperlink
     * 
     * @return label that will be rendered as a hyperlink
     */
    protected String getActionLabel(D record) {
        return actionLabel;
    }
    
    protected GridCellRenderer<D> createRenderer() {
        return new ActionColumnRenderer();        
    }

    protected final Grid<D> getGrid() {
        return gridPanel;
    }
    
    /**
     * Used to identify the anchor tag for an action within a row.
     * 
     * @return a unique column css class. 
     */
    protected final String getAnchorTagCssClass() {
        return "arr-action-link arr-" + getId() + "-link";
    }

    private void addCellClickListener(Grid<D> grid) {
        
        final ColumnConfig thisCol = this;
        grid.addListener(Events.CellClick, new Listener<GridEvent>() {  

            @SuppressWarnings("unchecked")
            public void handleEvent(GridEvent ge) {
                
                String targetStyleNames = "" + ge.getTargetEl().getStyleName();
                if (ge.getTargetEl().is("A")
                        && targetStyleNames.indexOf("arr-action-link") >= 0
                        && ge.grid.getColumnModel().getColumn(ge.colIndex) == thisCol) {
                    Element anchorElement = ge.grid.getView().getCell(ge.rowIndex, ge.colIndex);
                    if (anchorElement != null) {
                        ModelData record = ge.grid.getStore().getAt(ge.rowIndex);
                        actionCallback.actionClicked((D)record, anchorElement);
                    }                            
                }
            }
        });
    }
    
    
    /**
     * A column represented by an ActionColumnConfig invokes on actionClicked 
     * when the associated anchor is clicked on by the user. 
     */
    public static interface ActionColumnCallback<D extends ModelData> {
        /**
         * @param record the record representing the selected row
         * @param obj the Repository object represented by the selected row
         * @param anchorElement the element representing the html anchor element
         */
        void actionClicked(D record, Element anchorElement);
    };
    
    private class ActionColumnRenderer implements GridCellRenderer<D> {
        public String render(D record,
                             String property,
                             ColumnData config,
                             int rowIndex,
                             int colIndex,
                             ListStore<D> store) {
            
            String actualActionLabel = getActionLabel(record);

            // Only provide html for records that provide data
            if (record == null || actualActionLabel == null) {
                return "";
            }
                            
            String html = snippets.anchorTagWithCssCls(actualActionLabel, getAnchorTagCssClass());
            
            return html;
        }
    }    
}