/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.web.repository.main.client.base.ui.editor;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.allen_sauer.gwt.log.client.Log;
import com.google.gwt.http.client.Request;
import com.google.gwt.http.client.RequestBuilder;
import com.google.gwt.http.client.RequestCallback;
import com.google.gwt.http.client.RequestException;
import com.google.gwt.http.client.Response;
import com.google.gwt.xml.client.Document;
import com.google.gwt.xml.client.Element;
import com.google.gwt.xml.client.Node;
import com.google.gwt.xml.client.NodeList;
import com.google.gwt.xml.client.XMLParser;
import com.iona.soa.web.repository.base.client.model.ReposObj;
import com.iona.soa.web.repository.base.client.util.PropertyCallBack;
import com.iona.soa.web.repository.base.client.util.PropertyHelper;

public final class EditorLayoutRegistry {

    public interface EditorLayoutRegistryCallBack {
        void layoutLoaded(EditorLayoutDescriptor aDesc);
    }

    public static final EditorLayoutRegistry INSTANCE = new EditorLayoutRegistry();

    private static final int TIME_OUT_MILLIS = 0;


    private static Map<String, String> typeToLayoutIdMapping;
    
    private Map<String, EditorLayoutDescriptor> descriptors;

    private EditorLayoutRegistry() { /* singleton */
        descriptors = new HashMap<String, EditorLayoutDescriptor>();
    }

    public void getLayout(final ReposObj anObj, final EditorLayoutRegistryCallBack layoutCallBack) {

        if (typeToLayoutIdMapping != null) {
            doGetLayout((String)typeToLayoutIdMapping.get(anObj.getEClass()), layoutCallBack);
        
        } else {
            PropertyHelper.loadProperties("editorCfg/layoutmapping.properties", new PropertyCallBack() {
                public void propertiesLoaded(Map<String, String> properties) {
                    typeToLayoutIdMapping = properties;
                    doGetLayout((String)typeToLayoutIdMapping.get(anObj.getEClass()), layoutCallBack);
                }
            });
        }
    }

    private void doGetLayout(String layoutId, EditorLayoutRegistryCallBack callBack) {
        
        if (layoutId == null) {
            callBack.layoutLoaded(null);

        } else if (descriptors.get(layoutId) != null) {
            callBack.layoutLoaded((EditorLayoutDescriptor)descriptors.get(layoutId));
        
        } else {
            loadDescriptor(layoutId, callBack);
        }
    }

    private void loadDescriptor(final String layoutId, final EditorLayoutRegistryCallBack callBack) {
        
        Log.debug("\n\nGET layout descriptor       ==> " + layoutId);
        
        try {
            RequestBuilder rb = new RequestBuilder(RequestBuilder.GET, "editorCfg/" + layoutId + ".xml");
            rb.setHeader("Cache-Control", "max-age=0, no-store");
            rb.setTimeoutMillis(TIME_OUT_MILLIS);

            rb.sendRequest(null, new RequestCallback() {

                public void onError(Request request, Throwable exception) {
                    Log.debug("an error occured:", exception);
                    callBack.layoutLoaded(null);
                }

                public void onResponseReceived(Request request, Response response) {
                    Log.debug("GET (STATUS CODE) ==>" + response.getStatusCode());
                    Log.debug("GET (STATUS TEXT) ==>" + response.getStatusText());
                    Log.debug("GET (RESPONSE) ==>" + response.getText());
                    
                    EditorLayoutDescriptor parsedDesc = parseDesc(response.getText());
                    callBack.layoutLoaded(parsedDesc);
                    descriptors.put(layoutId, parsedDesc);
                }
            });
            

        } catch (RequestException e) {
            Log.debug("GET (ERROR) ==> " + e.getMessage(), e);
            callBack.layoutLoaded(null);
        }

    }

    private EditorLayoutDescriptor parseDesc(String xmlDesc) {
        Document document = XMLParser.parse(xmlDesc);
        
        EditorLayoutDescriptor objDesc = new EditorLayoutDescriptor();
        
        NodeList ignoreElts = document.getElementsByTagName("ignore");
        for (int sectionIndex = 0; sectionIndex < ignoreElts.getLength(); sectionIndex++) {
            Element ignoreElt = (Element)ignoreElts.item(sectionIndex);
            objDesc.addIgnoredField(ignoreElt.getAttribute("name"));
            
        }
        
        NodeList sectionElts = document.getElementsByTagName("section");
        for (int sectionIndex = 0; sectionIndex < sectionElts.getLength(); sectionIndex++) {
            EditorSection sectionObj = new EditorSection();
            
            Element sectionElt = (Element)sectionElts.item(sectionIndex);
            sectionObj.setCaption(sectionElt.getAttribute("caption"));
            
            NodeList colElts = sectionElt.getElementsByTagName("column");
            for (int columnIndex = 0; columnIndex < colElts.getLength(); columnIndex++) {
                EditorColumn colObj = new EditorColumn();
                
                Element colElt = (Element)colElts.item(columnIndex);
                
                NodeList itemElts = colElt.getChildNodes();
                for (int itemIndex = 0; itemIndex < itemElts.getLength(); itemIndex++) {
                    
                    Node itemNode = itemElts.item(itemIndex);
                    
                    if (itemNode instanceof Element) {
                        Element itemElt = (Element)itemNode;
                        
                        String itemType = itemElt.getNodeName();
                        if ("field".equals(itemType)) {
                            colObj.addField(itemElt.getAttribute("name"));
                        
                        } else if ("fieldGroup".equals(itemType)) {
                            
                            FieldGroup fieldGroupObj = new FieldGroup();
                            fieldGroupObj.setCaption(itemElt.getAttribute("caption"));
                            
                            NodeList fieldElts = itemElt.getElementsByTagName("field");
                            for (int fieldIndex = 0; fieldIndex < fieldElts.getLength(); fieldIndex++) {
                                
                                Element fieldElt = (Element)fieldElts.item(fieldIndex);
                                fieldGroupObj.addField(fieldElt.getAttribute("name"));
                            }
                            colObj.addFieldGroup(fieldGroupObj);
                        }
                    }
                }
                sectionObj.addColumn(colObj);
            }
            objDesc.addSection(sectionObj);
        }
        
        return objDesc;
    }

    class EditorLayoutDescriptor {
        
        private List<EditorSection> sections;
        private List<String> ignoredFields = new ArrayList<String>();
        
        private int index;
        
        EditorLayoutDescriptor() {
            addIgnoredField("timestamp");
            addIgnoredField("lastEditedBy");
        }
        
        public void resetIndex() { index = 0; }
        
        public EditorSection nextSection() { 
            return (sections == null || index == sections.size())
                        ? null
                        : (EditorSection)sections.get(index++);
        }
        
        public int getSectionCount() {
            return sections == null ? 0 : sections.size();
        }
        
        public boolean hasNext() {
            return sections == null ? false : index < sections.size();
        }

        protected void addSection(EditorSection aSection) {
            if (aSection == null) { return; }
            if (sections == null) { sections = new ArrayList<EditorSection>(); }
            sections.add(aSection);
        }
        
        protected void addIgnoredField(String anIgnoredField) {
            if (anIgnoredField == null) { return; }
            if (ignoredFields == null) { ignoredFields = new ArrayList<String>(); }
            ignoredFields.add(anIgnoredField);
        }

        protected List<String> getIgnoredFields() {
            return ignoredFields;
        }
    }

    public static class EditorSection {
        private String caption;
        private List<EditorColumn> columns;

        private int index;
        
        public void resetIndex() { index = 0; }
        
        public EditorColumn nextColumn() { 
            return (columns == null || index == columns.size()) ? null : (EditorColumn)columns.get(index++);
        }
        
        public int getColumnCount() {
            return columns == null ? 0 : columns.size();
        }
        
        public boolean hasNext() {
            return columns == null ? false : index < columns.size();
        }

        protected void addColumn(EditorColumn aColumn) {
            if (aColumn == null) { return; }
            if (columns == null) { columns = new ArrayList<EditorColumn>(); }
            columns.add(aColumn);
        }
        
        public String getCaption() { return caption; }
        protected void setCaption(String aCaption) { caption = aCaption; }

    }

    public static class EditorColumn {
        private List<Object> columnItems;
        private int index;

        public void resetIndex() { index = 0; }
        
        /**
         * @return the next item in the column, could a String (field id) or a FieldGroup
         */
        public Object nextItem() { 
            return (columnItems == null || index == columnItems.size()) ? null : columnItems.get(index++);
        }
        
        public int getItemCount() {
            return columnItems == null ? 0 : columnItems.size();
        }
        
        public boolean hasNext() {
            return columnItems == null ? false : index < columnItems.size();
        }

        protected void addField(String fieldId) {
            if (fieldId == null || fieldId.length() == 0) { return; }
            if (columnItems == null) { columnItems = new ArrayList<Object>(); }
            columnItems.add(fieldId);
        }

        protected void addFieldGroup(FieldGroup aGroup) {
            if (aGroup == null) { return; }
            if (columnItems == null) { columnItems = new ArrayList<Object>(); }
            columnItems.add(aGroup);
        }
    }
    
    public static class FieldGroup {
        private String caption;
        private List<String> fieldIds;

        private int index;
        
        public void resetIndex() { index = 0; }
        
        public String nextFieldId() { 
            return (fieldIds == null || index == fieldIds.size()) ? null : (String)fieldIds.get(index++);
        }
        
        public int getFieldCount() {
            return fieldIds == null ? 0 : fieldIds.size();
        }

        public boolean hasNext() {
            return fieldIds == null ? false : index < fieldIds.size();
        }

        protected void addField(String fieldId) {
            if (fieldId == null || fieldId.length() == 0) { return; }
            if (fieldIds == null) { fieldIds = new ArrayList<String>(); }
            fieldIds.add(fieldId);
        }

        public String getCaption() { return caption; }
        protected void setCaption(String aCaption) { caption = aCaption; }
    }
}
