/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.repository.scheduler.taskservice;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.repository.taskservice.TaskService;
import com.iona.soa.repository.EMFUtil;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.SOARepositoryService;
import com.iona.soa.repository.TransactionData;
import com.iona.soa.resourceset.ResourceSetQueryHandler;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;



public class ModelModificationTaskService implements TaskService {
    public static final String SET_COMMAND = "SET";
    public static final String ATTRIBUTE_OP = "ATTRIBUTE";
    public static final String REFERENCE_OP = "REFERENCE";

    private static final Logger LOG = Logger.getLogger(ModelModificationTaskService.class.getName());
    
    SOARepositoryService repoService; // injected
    private SOARepository soaRepository;

    
    public SOARepositoryService getRepositoryService() {
        return repoService;
    }

    public void setRepositoryService(SOARepositoryService rs) {
        repoService = rs;
        soaRepository = repoService.getRepository();
    }

    public SOARepository getSoaRepository() {
        return soaRepository;
    }

    public void setSoaRepository(SOARepository soaRepo) {
        soaRepository = soaRepo;
    }
    
    public String getTaskTypeID() {
        return ModelModificationTaskService.class.getName();
    }

    // command style:
    // SET nsURI DeploymentProfile 12345 ATTRIBUTE attr1 blah
    // SET nsURI DeploymentProfile 23456 RELATION rel1 OtherObject_guid
    public void run(String user, String externalToken, String... parameters) throws Exception {        
        //tempSSOToken = ssoToken;
        List<String> params = Arrays.asList(parameters);
        if (parameters.length != 7) {
            throw new IllegalArgumentException("Incorrect number of arguments to " + getTaskTypeID()
                + ". Expected 7 arguments, but received: " + params);
        }
     
        List<String> remaining = new ArrayList<String>(params);
        String command = remaining.remove(0);
        TransactionData transData = new TransactionData(user, externalToken);
        if (SET_COMMAND.equals(command)) {
            executeSetCommand(remaining, transData);
        } else {
            throw new IllegalArgumentException("Command not understood: " + params);
        }
    }

    private void executeSetCommand(List<String> remaining, TransactionData transData) throws Exception {
        String nsURI = remaining.remove(0);
        String entity = remaining.remove(0);
        String id = remaining.remove(0);
        String op = remaining.remove(0);
     
        if (nsURI == null || entity == null || id == null || op == null) {
            throw new IllegalArgumentException("Invalid SET command. nsURI: " + nsURI 
                    + " EClass: " + entity + " ID: " + id + " Op:" + op);
        }
        
        EPackage pkg = EPackage.Registry.INSTANCE.getEPackage(nsURI);
        if (pkg == null) {
            throw new IllegalArgumentException("Could not find package with URI: " + nsURI);
        }
        EClassifier classifier = pkg.getEClassifier(entity);
        if (!(classifier instanceof EClass)) {
            throw new IllegalArgumentException("Could not find entity type: " + entity);
        }
        
        soaRepository.startRead(transData);
        EObject eo;
        boolean success = true;
        try {
            eo = ResourceSetQueryHandler.executeQuery(
                    soaRepository.getResourceSet(), (EClass) classifier, id);
        } catch (Exception e) {
            success = false;
            throw e;
        } finally {
            soaRepository.endRead(success);  
        }

        
        if (eo == null) {
            throw new IllegalArgumentException("Could not find " + entity + " with id " + id);
        }
        
        if (ATTRIBUTE_OP.equals(op)) {
            executeSetAttributeCommand(eo, remaining, transData);
        } else if (REFERENCE_OP.equals(op)) {
            executeSetReferenceCommand(eo, remaining, transData);
        }
    }

    private void executeSetAttributeCommand(EObject eo, List<String> remaining,
            TransactionData transData) throws Exception {
        String attrName = remaining.remove(0);
        String newValue = remaining.remove(0);
        
        EStructuralFeature attribute = eo.eClass().getEStructuralFeature(attrName);
        if (!(attribute instanceof EAttribute)) {
            throw new IllegalArgumentException(attrName + " is not an attribute on " + eo);
        }

        boolean success = false;
        soaRepository.startWrite(transData);
        try {
            eo.eSet(attribute, newValue);
            success = true;
        } finally {
            soaRepository.endWrite(success);  
            if (success) {
                LOG.log(Level.INFO, "Set {0} on {1} to {2} successfully",
                        new Object[] {attrName, EMFUtil.getGuid(eo), newValue});
            } else {
                LOG.log(Level.WARNING, "Failed to set {0} on {1} to {2}",
                        new Object[] {attrName, EMFUtil.getGuid(eo), newValue});
            }
        }
    }

    private void executeSetReferenceCommand(EObject eo, 
                                            List<String> remaining,
                                            TransactionData transData) throws Exception {
        String refName = remaining.remove(0);
        String refGuid = remaining.remove(0);
        
        EStructuralFeature reference = eo.eClass().getEStructuralFeature(refName);
        if (!(reference instanceof EReference)) {
            throw new IllegalArgumentException(refName + " is not an reference on " + eo);
        }
        
        if (!(reference.getEType() instanceof EClass)) {
            throw new IllegalArgumentException(
                refName + " on " + eo + " does no refer to an EClass: " + reference);
        }
        
        EObject refObj = ResourceSetQueryHandler.executeQuery(soaRepository.getResourceSet(),
                (EClass) reference.getEType(), refGuid);
        soaRepository.startWrite(transData);
        boolean success = false;
        try {
            eo.eSet(reference, refObj);
            success = true;
        } finally {
            soaRepository.endWrite(success);  
            if (success) {
                LOG.log(Level.INFO, "Set {0} on {1} to {2} successfully",
                        new Object[] {refName, EMFUtil.getGuid(eo), refGuid});
            } else {
                LOG.log(Level.WARNING, "Failed to set {0} on {1} to {2}",
                        new Object[] {refName, EMFUtil.getGuid(eo), refGuid});
            }
        }
    }

    public String getTaskDescription() {
        return "ModelModificationTaskService";
    }

    public String getTaskName() {
        return "ModelModificationTaskService";
    }

    public String[] getExampleParameterList() {
        String[] paramList = new String[7];
        paramList[0] = new String("{Set}");
        paramList[1] = new String("{Package Namespace Uri}");
        paramList[2] = new String("{Rule}");
        paramList[3] = new String("{pt2}");
        paramList[4] = new String("{Reference}");
        paramList[5] = new String("{PolicyTemplate}");
        paramList[6] = new String("{Guid}");

        return paramList;
    }
}
