/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.scheduler;

import java.io.IOException;
import java.sql.Timestamp;
import java.util.Calendar;
import java.util.logging.Level;

import com.iona.repository.rest.RESTContext;
import com.iona.repository.rest.RESTException;
import com.iona.repository.rest.osgi.RESTCollectionService;
import com.iona.repository.rest.repo.EClassHandler;
import com.iona.repository.rest.resource.Resource;
import com.iona.repository.rest.rt.repo.DefaultEClassHandler;
import com.iona.repository.rest.rt.repo.RestEMFUtil;
import com.iona.repository.rest.rt.server.RESTProvider;
import com.iona.soa.model.scheduling.OnceOffTask;
import com.iona.soa.model.scheduling.ScheduledTask;
import com.iona.soa.model.scheduling.SchedulingPackage;
import com.iona.soa.model.scheduling.TaskStates;
import com.iona.soa.model.security.User;
import com.iona.soa.repository.SOARepositoryService;
import com.iona.soa.security.SecurityContext;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EObject;

public abstract class AbstractSchedulerCollectionService implements RESTCollectionService {
    EClassHandler handler;
    Scheduler scheduler;
    SOARepositoryService repositoryService; // injected
    
    public  SOARepositoryService getRepositoryService() {
        return repositoryService;
    }
    public void setRepositoryService(SOARepositoryService rs) {
        repositoryService = rs;
        
    }
    
 
    public EClass getEClass() {
        return SchedulingPackage.eINSTANCE.getScheduledTask();
    }

    public EClassHandler getHandler() {
        return handler;
    }

    public void setRESTAdapter(RESTProvider ra) {
        handler = new SchedulerHandler(ra);
    }
    
    public void setScheduler(Scheduler s) {
        scheduler = s;
    }
    

    protected String getTransientToken(ScheduledTask ts, RESTContext ctx) {
        String transientSSOToken = null; 
        Object o = ctx.getRequestAttribute(RESTContext.SECURITY_CONTEXT);
        if (o instanceof SecurityContext) {
            SecurityContext secCtx = (SecurityContext)o;
            if (secCtx.getSSOToken() != null) {
                transientSSOToken = secCtx.getSSOToken();
            }
        }
        return transientSSOToken;
    }
    
    protected String getTransientUser(ScheduledTask ts, RESTContext ctx) {
        String transientUser = null; 
        Object o = ctx.getRequestAttribute(RESTContext.SECURITY_CONTEXT);
        if (o instanceof SecurityContext) {
            SecurityContext secCtx = (SecurityContext)o;
            if (secCtx.getUser() != null) {
                transientUser = secCtx.getUser();
            }
        }
        return transientUser;
    }
    
    
    private class SchedulerHandler extends DefaultEClassHandler implements EClassHandler {
        public SchedulerHandler(RESTProvider ra) {
            super(ra);
        }
        
        @Override
        public EObject createObject(RESTContext ctx, EFactory efactory, EClass eclass, 
                String name, Resource resource, boolean persist)
            throws RESTException {
            
            EObject eobject = efactory.create(eclass);
            
            RestEMFUtil.setEObjectName(eobject, name);
            initRequiredAttributes(eobject);
            initDefaultValues(eobject);
            
            if (persist) {
                try {
                    ctx.getRepo().persist(eobject);
                } catch (IOException e) {
                    throw new RESTException(RESTException.INTERNAL_SERVER_ERROR, e);
                }
            }
            
            if (resource != null) {
                getObjectChanges(ctx, eobject, resource);        
            }
            
            return eobject;        
        }

        @Override
        public void getObjectChanges(RESTContext ctx, EObject eobject,
                Resource resource) throws RESTException {

  
            super.getObjectChanges(ctx, eobject, resource);

            if (scheduler == null) {
                LOG.warning("Scheduler not set, could not take the schedule changes into effect");
                return;
            }
            
            if (eobject instanceof ScheduledTask) {
                ScheduledTask st = (ScheduledTask) eobject;
                storeAuthToken(resource, st, ctx);

                if (st.getAction() != null && st.getAction().length() > 0) {                    
                    scheduler.performAction(st, st.getAction());
                    st.setAction("");
                } else {
                    scheduler.taskUpdated(st);
                }
            }
        }

        private void storeAuthToken(Resource resource, ScheduledTask ts, RESTContext ctx) {
            if (ts.isRunNow()) {
                String transientUser = getTransientUser(ts, ctx);
                if (ts.getUser() == null) {
                    String transientSSOToken = getTransientToken(ts, ctx);
                    if (ts.getToken() == null && transientSSOToken != null) {
                        ts.setToken(transientSSOToken);
                        ts.setState(TaskStates.AUTHENICATED);
                        return;
                    }
                } else if (ts.getUser().getUserid().equals(transientUser)) {
                    return;
                } else {
                    LOG.log(Level.WARNING, "Task Service User and Runner must be the same");
                }
            }
            
            if (resource.getFirstReference("user") != null 
                    && ts.getPassword() != null) {
                User user = ts.getUser();
                String password = ts.getPassword();
                try {
                    SecurityContext cxt = getRepositoryService().getSecurityService().
                                              authenticatePersistent(user.getUserid(), password);
                    if (cxt == null) {
                        LOG.log(Level.WARNING, "Failed to authenticate user {0}", user.getUserid());
                    } else {
                        ts.setToken(cxt.getSSOToken());
                        ts.setState(TaskStates.CREATED); 
                        return;
                    }
                } catch (Throwable e) {
                    LOG.log(Level.WARNING, "Not Authenicated", e);
                    
                } finally {
                    // We don't want it to be persisted - ever!
                    // It is marked as transient in the ecore model, but just in 
                    // case that changes..
                    ts.setPassword(null); 
                }
            } 
            
            ts.setState(TaskStates.UNAUTHENICATED);
            ts.setToken(null);

            /*

            */


        }


        
        private void initDefaultValues(EObject task) {
            if (task instanceof OnceOffTask) {
                Calendar c = Calendar.getInstance();
                c.set(c.get(Calendar.YEAR),
                      c.get(Calendar.MONTH), 
                      c.get(Calendar.DAY_OF_MONTH), 
                      23, 59, 0);
                Timestamp tonight = new Timestamp(c.getTimeInMillis());
                ((OnceOffTask)task).setScheduleDate(tonight);
            }
        }
    }
}
