/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.scheduler;

import java.sql.Timestamp;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.iona.repository.taskservice.TaskService;
import com.iona.soa.model.scheduling.OnceOffTask;
import com.iona.soa.model.scheduling.ScheduledTask;
import com.iona.soa.model.scheduling.SchedulingPackage;
import com.iona.soa.model.scheduling.TaskStates;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.TransactionData;
import com.iona.soa.resourceset.ResourceSetQueryHandler;
import com.iona.soa.security.SecurityService;

import org.eclipse.emf.ecore.EObject;

public class ExecutableTask implements Runnable {
    private static final Logger LOG = Logger.getLogger(ExecutableTask.class.getName());
    final String scheduledTaskGuid;
    private final Scheduler scheduler;
    private SOARepository soaRepository;
    private SecurityService securityService;
    
    public ExecutableTask(Scheduler s, String st, SOARepository soaRepo, 
            SecurityService securityService) {
        scheduler = s;
        scheduledTaskGuid = st;
        soaRepository = soaRepo;
        this.securityService = securityService;
    }

    public void run() {
        ScheduledTask scheduledTask = null;
        TaskService ts = null;
        
        soaRepository.startWrite(getNewTxData());
        boolean success = false;

        try {
            scheduledTask = getScheduledTask();
            
            scheduledTask.setState(TaskStates.RUNNING);
            // use scheduler.now() instead of System.currentTimeMillis() to facilitate
            // testing. The real Scheduler.now() invokes System.currentTimeMillis().
            scheduledTask.setLastRun(new Timestamp(scheduler.now()));            
            
            ts = findTask(scheduledTask.getTypeID());
            if (ts == null) {
                setTaskToErrorState(scheduledTask);
                LOG.severe("Could not find task with ID " 
                        + scheduledTask.getTypeID() 
                        + " for task " + scheduledTask);
                return;
            }            
            success = true;
        } catch (Exception e) {
            LOG.log(Level.SEVERE, "Error preparing task execution " + scheduledTask, e);
            setTaskToErrorState(scheduledTask);            
            return;
        } finally {
            soaRepository.endWrite(success);
        }
        
        try {
            String user = null;
            String authToken = scheduledTask.getToken();

            if (scheduledTask.getUser() != null) {
                user = scheduledTask.getUser().getUserid();
            }
            
            if (authToken != null && authToken.equals(SecurityService.INTERNAL_TOKEN)) {
                user = SecurityService.INTERNAL_USER;
            }

            ts.run(user,
                    authToken, scheduledTask.getParameters().toArray(new String [] {})); 

        } catch (Exception e) {
            LOG.log(Level.SEVERE, "Error executing task " + scheduledTask, e);
            seTaskToErrorStateTx(scheduledTask);
            return;
        }
        
        soaRepository.startWrite(getNewTxData());
        boolean success2 = false;
        try {
            setDuration(scheduledTask);

            if (scheduledTask instanceof OnceOffTask) {
                OnceOffTask oot = (OnceOffTask)scheduledTask;
                if (oot.isCleanupOnCompletion()) {
                    // remove the task here 
                    EObject scheduledTaskObject = scheduledTask;
                    scheduledTaskObject.eResource().getContents().remove(scheduledTaskObject);
                    LOG.log(Level.INFO, "Removing scheduled object " + oot.getName());
                    
                } else {
                    scheduledTask.setState(TaskStates.FINISHED);
                    scheduledTask.setNextRun(null);
                }
                
            } else {                
                scheduler.reschedule(scheduledTask);
            }
            success2 = true;
        } finally {
            soaRepository.endWrite(success2);
        }
    }

    private TransactionData getNewTxData() {
        return new TransactionData(SecurityService.INTERNAL_USER, SecurityService.INTERNAL_TOKEN);
    }
    
    private void setDuration(ScheduledTask st) {
        long duration = scheduler.now() - st.getLastRun().getTime();
        st.setDuration((int) duration);        
    }

    ScheduledTask getScheduledTask() throws Exception {
        return (ScheduledTask) ResourceSetQueryHandler.executeQuery(
                soaRepository.getResourceSet(),
                SchedulingPackage.eINSTANCE.getScheduledTask(), scheduledTaskGuid);
    }

    private void setTaskToErrorState(ScheduledTask scheduledTask) {
        setDuration(scheduledTask);
        scheduledTask.setState(TaskStates.ERROR);
        scheduledTask.setNextRun(null);
    }    

    private void seTaskToErrorStateTx(ScheduledTask scheduledTask) {
        soaRepository.startWrite(getNewTxData());
        boolean success = false;
        try {
            setTaskToErrorState(scheduledTask);
            success = true;
        } finally {
            soaRepository.endWrite(success);
        }
    }

    private TaskService findTask(String typeID) {
        for (TaskService ts : scheduler.getTaskServices()) {
            if (typeID.equals(ts.getTaskTypeID())) {
                return ts;
            }
        }
        return null;
    }
}
