/**
 * Copyright (C) [2009]  Progress Software Corporation.
 * FUSE DEPOT 1.0
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.iona.soa.scheduler;

import java.io.IOException;
import java.sql.Timestamp;
import java.util.Calendar;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import com.iona.repository.taskservice.TaskService;
import com.iona.soa.model.scheduling.DailyRepeatedTask;
import com.iona.soa.model.scheduling.IntervalRepeatedTask;
import com.iona.soa.model.scheduling.MonthlyRepeatedTask;
import com.iona.soa.model.scheduling.OnceOffTask;
import com.iona.soa.model.scheduling.ScheduledTask;
import com.iona.soa.model.scheduling.SchedulingPackage;
import com.iona.soa.model.scheduling.TaskStates;
import com.iona.soa.model.scheduling.Weekdays;
import com.iona.soa.model.scheduling.WeeklyRepeatedTask;
import com.iona.soa.repository.SOARepository;
import com.iona.soa.repository.SOARepositoryService;
import com.iona.soa.repository.TransactionData;
import com.iona.soa.resourceset.ResourceSetQueryHandler;
import com.iona.soa.security.SecurityService;


import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.InitializingBean;

// Turn this into a bean that is injected with OSGi services
public class Scheduler implements InitializingBean, DisposableBean {
    public static final String RUN_NOW_ACTION = "RUN_NOW";
    private static final Logger LOG = Logger.getLogger(Scheduler.class.getName());    

    ResourceSet resourceSet; // injected
    Collection<TaskService> taskServices = Collections.emptySet(); // injected
        
    private ScheduledExecutorService executor = Executors.newSingleThreadScheduledExecutor();
    private Map<String, Future<?>> scheduledTasks = new HashMap<String, Future<?>>(); 

    private SOARepository soaRepository;
    private SecurityService securityService;

    public void setRepositoryService(SOARepositoryService repositoryService) throws IOException {
        LOG.fine("Repository service available");

        soaRepository = repositoryService.getRepository();
        resourceSet = repositoryService.getRepository().getResourceSet();   
        securityService = repositoryService.getSecurityService();
    }
    
    public void setUpScheduler(SOARepository soaRepoIn, 
                               SecurityService secSrv) throws IOException {
        LOG.fine("Setting the repository explicitly");
        LOG.fine("Setting the repository protector explicitly");
        securityService = secSrv;
        soaRepository = soaRepoIn;
        resourceSet = soaRepository.getResourceSet();
    }    
    
    public void afterPropertiesSet() throws Exception {
        scheduleAllJobs();
    }

    public void scheduleAllJobs() throws Exception {
        TransactionData trnx = new TransactionData(SecurityService.INTERNAL_USER, 
                                                   SecurityService.INTERNAL_TOKEN);
        
        soaRepository.startWrite(trnx);
        boolean success = false;
        try {
            for (EObject obj : ResourceSetQueryHandler.executeQuery(resourceSet, 
                    SchedulingPackage.eINSTANCE.getScheduledTask(), true)) {
                if (obj instanceof ScheduledTask) {
                    ScheduledTask st = (ScheduledTask) obj;                
                    submit(st, true);
                }
            }
            success = true;
        } finally {
            soaRepository.endWrite(success);
        }
    }

    private void submit(ScheduledTask st, boolean rescheduleIfNeeded) {
        if (st.getState() == TaskStates.DISABLED
            || st.getState() == TaskStates.ERROR
            || st.getState() == TaskStates.UNAUTHENICATED) {
            return;
        }
        
        if (st.getNextRun() == null && rescheduleIfNeeded) {
            reschedule(st);
            return;
        }
        
        long delay = st.getNextRun().getTime() - now();
        if (delay < 0 && rescheduleIfNeeded) {
            LOG.info("Rescheduling task as it missed its slot: " + st);
            reschedule(st);            
            return;
        }
        ScheduledFuture<?> future = 
            getExecutor().schedule(new ExecutableTask(this, st.getGuid(), 
                    soaRepository, securityService), delay, TimeUnit.MILLISECONDS);
        
        Map<String, Future<?>> tasks = getScheduledTasks();
        synchronized (tasks) {
            tasks.put(st.getGuid(), future);
        }        
        
        st.setState(TaskStates.SCHEDULED);
    }

    private Future<?> cancel(ScheduledTask task) {
        Map<String, Future<?>> tasks = getScheduledTasks();
        synchronized (tasks) {
            Future<?> future = tasks.get(task.getGuid());            
            if (future != null && !future.isDone()) {
                future.cancel(false);
                return future;
            }
        }
        return null;
    }

    public void reschedule(ScheduledTask st) {
        if (st.isRunNow()) {
            runTaskNow(st);
            st.setRunNow(false);
            return;
        }
        if (st instanceof IntervalRepeatedTask) {
            rescheduleIntervalTask((IntervalRepeatedTask) st);
        } else if (st instanceof OnceOffTask) {
            rescheduleOnceOffTask((OnceOffTask) st);
        } else if (st instanceof WeeklyRepeatedTask) {
            rescheduleWeeklyTask((WeeklyRepeatedTask) st);
        } else if (st instanceof MonthlyRepeatedTask) {
            rescheduleMonthlyTask((MonthlyRepeatedTask) st);
        } else if (st instanceof DailyRepeatedTask) {
            rescheduleDailyTask((DailyRepeatedTask) st);
        }
        
        // Do some cleanup while we're here
        cleanUpFinishedTasks();
    }


    private void cleanUpFinishedTasks() {
        Map<String, Future<?>> tasks = getScheduledTasks();
        synchronized (tasks) {
            for (String id : tasks.keySet().toArray(new String[0])) {
                Future<?> future = tasks.get(id);
                if (future.isDone() || future.isCancelled()) {
                    tasks.remove(id);
                }
            }
        }
    }

    private void rescheduleIntervalTask(IntervalRepeatedTask irt) {
        long add = 0;
        switch (irt.getTimeUnit()) {
        case HOURS:
            add = (long) irt.getInterval() * 60L * 60L * 1000L; 
            break;
        case MINUTES:
            add = (long) irt.getInterval() * 60L * 1000L; 
            break;
        default:
            return;
        }
        
        irt.setNextRun(new Timestamp(now() + add));
        submit(irt, false);        
    }
    
    private void rescheduleOnceOffTask(OnceOffTask oot) {
        if (oot.getScheduleDate() == null) {
            return;
        }
        oot.setNextRun(oot.getScheduleDate());
        if (oot.getScheduleDate().getTime() > now()) {
            submit(oot, false);        
        }
    }

    private void rescheduleDailyTask(DailyRepeatedTask drt) {
        Calendar c = getNextDailyDate(drt);
        
        drt.setNextRun(new Timestamp(c.getTimeInMillis()));
        submit(drt, false);
    }

    private Calendar getNextDailyDate(DailyRepeatedTask drt) {
        Calendar c = Calendar.getInstance();
        c.setTimeInMillis(now());
        
        if (c.get(Calendar.HOUR_OF_DAY) > drt.getHour()
            || (c.get(Calendar.MINUTE) >= drt.getMinute() 
                && c.get(Calendar.HOUR_OF_DAY) == drt.getHour())) {
            c.add(Calendar.DATE, 1);
        }
        c.set(Calendar.HOUR_OF_DAY, drt.getHour());
        c.set(Calendar.MINUTE, drt.getMinute());
        c.set(Calendar.SECOND, 0);
        return c;
    }

    private void rescheduleWeeklyTask(WeeklyRepeatedTask wrt) {
        Calendar c = getClosestWeekDay(wrt);
        
        wrt.setNextRun(new Timestamp(c.getTimeInMillis()));
        submit(wrt, false);
    }
    
    private Calendar getClosestWeekDay(WeeklyRepeatedTask wrt) {
        SortedSet<Calendar> dates = new TreeSet<Calendar>(); 
        
        for (Weekdays wd : wrt.getWeekdays()) {
            dates.add(getClosestDate(wrt, wd));
        }
        
        if (dates.size() == 0) {
            return null;
        }
        // Take the first date, since this is a sorted set
        return dates.iterator().next();
    }
    
    private Calendar getClosestDate(WeeklyRepeatedTask wrt, Weekdays wd) {
        Calendar c = getNextDailyDate(wrt);
        
        int weekday = c.get(Calendar.DAY_OF_WEEK);
        int newWeekDay = weekday2CalendarWeekday(wd);
        if (weekday != newWeekDay) {
            int addDays = newWeekDay - weekday;
            if (addDays < 0) {
                addDays += 7;
            }
            c.add(Calendar.DATE, addDays);
        }
        
        return c;
    }

    private static int weekday2CalendarWeekday(Weekdays wd) {
        int rv = Calendar.SUNDAY;
        switch(wd) {
        case SUNDAY: 
            rv = Calendar.SUNDAY;
            break;
        case MONDAY:
            rv =  Calendar.MONDAY;
            break;
        case TUESDAY:
            rv =  Calendar.TUESDAY;
            break;
        case WEDNESDAY:
            rv =  Calendar.WEDNESDAY;
            break;
        case THURSDAY:
            rv =  Calendar.THURSDAY;
            break;
        case FRIDAY:
            rv =  Calendar.FRIDAY;
            break;
        case SATURDAY:
            rv =  Calendar.SATURDAY;
            break;
        default:
            break;
        }
        return rv;
    }
        
    private void rescheduleMonthlyTask(MonthlyRepeatedTask mrt) {
        Calendar c = getClosestMonthDay(mrt);
        
        mrt.setNextRun(new Timestamp(c.getTimeInMillis()));
        submit(mrt, false);
    }

    private Calendar getClosestMonthDay(MonthlyRepeatedTask mrt) {
        SortedSet<Calendar> dates = new TreeSet<Calendar>();
        
        for (int day : mrt.getDays()) {
            Calendar cd = getClosestDate(mrt, day);
            if (cd != null) {
                dates.add(cd);
            }
        }

        if (dates.size() == 0) {
            return null;
        }
        
        // Take the first date, since this is a sorted set
        return dates.iterator().next();        
    }

    private Calendar getClosestDate(MonthlyRepeatedTask mrt, int day) {
        if (day > 31 || day < 1) {
            return null;
        }
        
        Calendar c = getNextDailyDate(mrt);
        
        int date = c.get(Calendar.DATE);
        if (date > day) {
            c.add(Calendar.MONTH, 1);
        }
        
        c.set(Calendar.DATE, day);
        // The following handles cases like when it's February and the date
        // is 31. The previous call will cause the date to roll over to the next month 
        // e.g. 31 February is 2 or 3 March. In this case we need to skip that date and
        // roll over to the next month, i.e. 31 March.
        while (c.get(Calendar.DATE) != day) {
            c.set(Calendar.DATE, day);
        }
        
        return c;
    }

    public long now() {
        return System.currentTimeMillis();
    }

    public void taskUpdated(ScheduledTask task) {
        cancel(task);
        reschedule(task);
    }

    public void performAction(ScheduledTask st, String action) {
        if (RUN_NOW_ACTION.equals(action)) {
            runTaskNow(st);
        } else {
            LOG.warning("Unrecognized action '" + action + "' on task " + st);
        }
    }
    
    private void runTaskNow(ScheduledTask st) {
        cancel(st);
        
        Future<?> future = getExecutor().submit(
                new ExecutableTask(this, st.getGuid(), soaRepository, securityService));
        Map<String, Future<?>> tasks = getScheduledTasks();
        synchronized (tasks) {
            tasks.put(st.getGuid(), future);
        }        

    }

    public void destroy() throws Exception {
        getExecutor().shutdownNow();
    }

    // ----------------------------------------------------------------------
    // accessors
    // ----------------------------------------------------------------------

    public ResourceSet getResourceSet() {
        return resourceSet;
    }

    public void setResourceSet(ResourceSet rs) {
        resourceSet = rs;
    }

    /* (non-Javadoc)
     * @see com.iona.soa.scheduler.osgi.IScheduler#getTaskServices()
     */
    public Collection<TaskService> getTaskServices() {
        return taskServices;
    }

    public void setTaskServices(Collection<TaskService> ts) {
        taskServices = ts;
    }

    public void setExecutor(ScheduledExecutorService executor) {
        this.executor = executor;
    }

    public ScheduledExecutorService getExecutor() {
        return executor;
    }

    public void setScheduledTasks(Map<String, Future<?>> scheduledTasks) {
        synchronized (this.scheduledTasks) {
            this.scheduledTasks = scheduledTasks;
        }
    }

    public Map<String, Future<?>> getScheduledTasks() {
        return scheduledTasks;
    }    
}
